// ==== METHODS Anova1.cc ====
#include "Anova1.hh"

/**
 * \file
 * One-way analysis of variance.
 * Havlicek, L.L. and Crain, R.D.: Practical Statistics for the Physical Sciences.
 * ACS Professional Reference Book, Washington, 1988.
 * \author Andras Aszodi
 * \date 9-Oct-2001.
 */

// ==== MODULE HEADERS ====

#include "Statprob.hh"
using namespace RazorBack;

// ==== METHODS ====

/// Clears the calling object and prepares it for a new analysis.
void Anova1_::clear()
{
    Xgroups.clear();
    fill_n(Sumsq.begin(), 3, 0.0);
    fill_n(Meansq.begin(), 3, 0.0);
    fill_n(Degfree.begin(), 3, 0);
    Sumx=Sumx2=Sumgroup=0.0;
    Groupno=Groupsize=0; Ntotal=0;
    Equalgroups=true;
}
// END of clear()

// Updates the internal sum-of-squares after a new group has been added
// by add_group(). Sx and Sx2 are the sums and sums-squared of the data points
// in the new group, n>=1 is the number of data points in the group.
// Hidden
void Anova1_::update_sumsq(double Sx, double Sx2, unsigned int n)
{
    // summing
    Sumx+=Sx;
    Sumx2+=Sx2;
    Ntotal+=n;
    Sumgroup+=(Sx*Sx/n);
    
    // check if all groups have the same size
    Groupno=Xgroups.size();
    if (Groupno==1) // first group
    {
	Groupsize=n; Equalgroups=true;
    }
    else
	Equalgroups=Equalgroups && (Groupsize==n);
    
    // re-calc ANOVA sums and degrees of freedom
    // fabs() is needed because little differences can be negative...
    Sumsq[BETWEEN]=fabs(Sumgroup-(Sumx*Sumx)/Ntotal);
    Degfree[BETWEEN]=Groupno-1;
    Meansq[BETWEEN]=(Degfree[BETWEEN])? Sumsq[BETWEEN]/Degfree[BETWEEN]: 0.0;
    
    Sumsq[WITHIN]=fabs(Sumx2-Sumgroup);
    Degfree[WITHIN]=Ntotal-Groupno;
    Meansq[WITHIN]=(Degfree[WITHIN])? Sumsq[WITHIN]/Degfree[WITHIN]: 0.0;
    
    // this is done "for completeness only" (not used later)
    Sumsq[TOTAL]=Sumsq[WITHIN]+Sumsq[BETWEEN]; Degfree[TOTAL]=Ntotal-1;
    Meansq[TOTAL]=(Degfree[TOTAL])? Sumsq[TOTAL]/Degfree[TOTAL]: 0.0;
}
// END of update_sumsq(Sx, Sx2, n)

/**
 * Returns the significance value that the two groups pointed to by G1, G2
 * have different means using the Scheffe test.
 * A small returned value means that the means are significantly different.
 */
double Anova1_::scheffe_test(vector<Group_>::const_iterator G1,
    	vector<Group_>::const_iterator G2) const
{
    if (Groupno<2)
	throw Toofewexc_("Anova1_::scheffe_test(G1, G2)", "Number of groups", Groupno);
    
    double Diff=G1->mean()-G2->mean();
    
    // no within-group variance, any mean diff other than 0.0 is significant
    if (Meansq[WITHIN]==0.0)
	return (Diff==0.0)? 1.0: 0.0;
    
    double F;
    if (Equalgroups)
	F=Groupsize*Diff*Diff/(2.0*(Groupno-1)*Meansq[WITHIN]);
    else
    {
	unsigned int n1=G1->size(), n2=G2->size();
	F=Diff*Diff/(Degfree[BETWEEN]*Meansq[WITHIN]*(1.0/n1+1.0/n2));
    }
    return f_test(F);
}
// END of scheffe_test(G1, G2)

/**
 * Returns the significance value for the test that the two meta-groups of groups
 * between [F1, L1) and [F2, L2) have different means according to the
 * Scheffe procedure. [F1,L1) and [F2,L2) are supposed to be valid ranges
 * in the groups given to the Anova1_ object before. Currently, only
 * continuous ranges are supported (not any combination of groups).
 * A small returned value means that the means are significantly different.
 */
double Anova1_::scheffe_test(vector<Group_>::const_iterator F1,
    	    	    	    vector<Group_>::const_iterator L1,
    	    	    	    vector<Group_>::const_iterator F2,
    	    	    	    vector<Group_>::const_iterator L2) const
{
    if (Groupno<2)
	throw Toofewexc_("Anova1_::scheffe_test(F1, L1, F2, L2)", "Number of groups", Groupno);
    
    double Metaavg1=0.0, Metaavg2=0.0, Diff, F;
    unsigned int n, Metan1=0, Metan2=0;
    vector<Group_>::const_iterator G;
    for (G=F1; G!=L1; ++G)
    {
	n=G->size();
	Metaavg1+=n*G->mean();
	Metan1+=n;
    }
    Metaavg1/=Metan1;
    for (G=F2; G!=L2; ++G)
    {
	n=G->size();
	Metaavg2+=n*G->mean();
	Metan2+=n;
    }
    Metaavg2/=Metan2;
    Diff=Metaavg1-Metaavg2;
    
    // no within-group variance, any mean diff other than 0.0 is significant
    if (Meansq[WITHIN]==0.0)
	return (Diff==0.0)? 1.0: 0.0;
    
    F=Diff*Diff/(Degfree[BETWEEN]*Meansq[WITHIN]*(1.0/Metan1+1.0/Metan2));
    return f_test(F);
}
// END of scheffe_test(F1, L1, F2, L2)

/**
 * Returns the significance value for the test that the two meta-groups of groups
 * defined in the vectors of vector<Group_> iterators have different means according to the
 * Scheffe procedure. The iterators in Metagroup1,2 are supposed to be valid ranges
 * in the groups given to the Anova1_ object before. 
 * A small returned value means that the means are significantly different.
 */
double Anova1_::scheffe_test(const vector< vector<Group_>::const_iterator >& Metagroup1,
    	    	    	    const vector< vector<Group_>::const_iterator >& Metagroup2) const
{
    if (Groupno<2)
	throw Toofewexc_("Anova1_::scheffe_test(Metagroup1, Metagroup2)", "Number of groups", Groupno);
    
    double Metaavg1=0.0, Metaavg2=0.0, Diff, F;
    unsigned int n, Metan1=0, Metan2=0;
    
    vector< vector<Group_>::const_iterator >::const_iterator Mit;
    vector<Group_>::const_iterator G;
    for (Mit=Metagroup1.begin(); Mit!=Metagroup1.end(); ++Mit)
    {
	G=(*Mit);
	n=G->size();
	Metaavg1+=n*G->mean();
	Metan1+=n;
    }
    Metaavg1/=Metan1;
    for (Mit=Metagroup2.begin(); Mit!=Metagroup2.end(); ++Mit)
    {
	G=(*Mit);
	n=G->size();
	Metaavg2+=n*G->mean();
	Metan2+=n;
    }
    Metaavg2/=Metan2;
    Diff=Metaavg1-Metaavg2;
    
    // no within-group variance, any mean diff other than 0.0 is significant
    if (Meansq[WITHIN]==0.0)
	return (Diff==0.0)? 1.0: 0.0;
    
    F=Diff*Diff/(Degfree[BETWEEN]*Meansq[WITHIN]*(1.0/Metan1+1.0/Metan2));
    return f_test(F);
}
// END of scheffe_test(Metagroup1, Metagroup2)

/**
 * Returns the significance value that the two pointed to by G1, G2
 * have different means using the Tukey A test.
 * A small returned value means that the means are significantly different.
 */
double Anova1_::tukey_atest(vector<Group_>::const_iterator G1,
    	vector<Group_>::const_iterator G2) const
{
    if (Groupno<2)
	throw Toofewexc_("Anova1_::tukey_atest(G1, G2)", "Number of groups", Groupno);
    
    double Diff=fabs(G1->mean()-G2->mean());
    
    // no within-group variance, any mean diff other than 0.0 is significant
    if (Meansq[WITHIN]==0.0)
	return (Diff==0.0)? 1.0: 0.0;
    
    double q, n;
    if (!Equalgroups)
    {
	double Harmsum=0.0;
	for (vector<Group_>::const_iterator Git=Xgroups.begin();
		Git!=Xgroups.end(); ++Git)
		    Harmsum+=1.0/Git->size();
	n=Groupno/Harmsum;
    }
    else n=Groupsize;
    q=Diff/sqrt(Meansq[WITHIN]/n);
    return q_test(q);
}
// END of tukey_atest(G1, G2)

// ---- classical tests ----

// Returns the significance level for the test that the
// between/within F ratio is larger than 1.0.
// If the value returned by f_test() is low, then the probability is large
// that f_ratio() is high.
// Hidden
double Anova1_::f_test(double Fratio) const
{
    double x=Degfree[WITHIN]/(Degfree[WITHIN]+Degfree[BETWEEN]*Fratio);
    return Statprob_::ibeta_regul(0.5*Degfree[WITHIN], 0.5*Degfree[BETWEEN], x);
}

// Returns the significance that t is larger than 0.0 using the internal
// settings necessary for Tukey's A-test. A small value indicates high
// significance. Based on the "Studentized Range Critical Values".
// Hidden
double Anova1_::q_test(double q) const
{
    return Statprob_::studrange_prob(q, Degfree[WITHIN], Groupno);
}
// END of q_test(q)

// ---- aux ----

/*
 * Returns an iterator to the group called "Name" or groups().end()
 * if not found.
 */
vector<Anova1_::Group_>::const_iterator
Anova1_::find_group(const string& Name) const
{
    vector<Group_>::const_iterator Git;
    for (Git=Xgroups.begin(); Git!=Xgroups.end(); ++Git)
	if (Git->descr()==Name)
	    break;
    return Git;
}
// END of find_group(Name)

// ==== END OF METHODS Anova1.cc ====
