#ifndef BASEMAT_TMPL_HEADER
#define BASEMAT_TMPL_HEADER

// ==== HEADER Basemat.hh ====

/**
 * \file Basemat.hh
 * Abstract base class for the matrix template hierarchy.
 */
 
// 8-Dec-2000. Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// Exception support is assumed.
// This module depends on the Standard Template Library.
 
// ---- COMPILER COMMAND LINE MACROS ----

// Define the macro INCLUDE_TMPL_DEFS if your compiler
// does not support automatic template source inclusion.

// ---- STANDARD HEADERS ----

#include <stdlib.h>
#include <math.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#  include <iostream>
#  include <sstream>
#  include <vector>
#  include <algorithm>
#  include <numeric>
#  include <functional>
#else
#  include <iostream.h>
#  include <strstream.h>
#  define istringstream istrstream 
#  define ostringstream ostrstream 
#  include <vector.h>
#  include <algo.h>
#  include <function.h>
#endif

using namespace std;

// ---- MODULE HEADERS ----

#include "Refstorage.hh"
#include "Vector.hh"
#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Matstorage_: helper class that contains an STL vector
 * and its length. All matrices have their elements represented
 * in this STL vector, usually in row major order.
 * Objects of this class are stored in Refstorage_ templates
 * which support copy-on-write (see below).
 */
template <class T_>
class Matstorage_
{
    // data
    vector<T_> Elems;   // STL vector of matrix elements
    size_t Eno;  // total no. of elements
    
    // methods
    public:
    
    /// default ctor
    explicit Matstorage_(size_t N=0): Elems(N), Eno(N) {}
    
    /// access
    const vector<T_>& elems() const { return Elems; }
    vector<T_>& elems() { return Elems; }
    size_t eno() const { return Eno; }
    size_t eno(size_t Enew)
    {
	if (Enew==Eno) return Eno;
	size_t Eold=Eno;
	if (Enew>Eold)
	    Elems.insert(Elems.end(), Enew-Eold, T_(0));    // grows
	else
	    Elems.erase(Elems.begin()+Enew, Elems.end());   // shrinks
	Eno=Enew;
	return Eold;
    }
};
// END OF CLASS Matstorage_

/**
 * Basemat_: implements the matrix element storage supporting
 * copy-on-write and all matrix operations which can be performed
 * on all matrix elements in parallel, irrespective of the layout.
 * Serves as an abstract base class for other matrix classes.
 * Copy-on-write is inherited from the Refstorage_ template.
 */
template <class T_>
class Basemat_: protected Refstorage_< Matstorage_<T_> >
{

    // data
    bool Ftnidx;    	//< use [1..N] FORTRAN indexing when true, see operator()
    bool Check;     	//< enable range checking when true, see operator()
    
    /// methods
    public:
    
    /// default ctor: proper setup is done in derived classes
    Basemat_(): Ftnidx(false), Check(false) {}
    
    /// destructor
    virtual ~Basemat_() {}
        
    	/// access
    
    /**
    * rno(), cno(): return the number of rows and columns.
    * Pure virtual at this stage (layout-dependent).
    */
    virtual size_t rno() const =0;
    virtual size_t cno() const =0;
    
    /**
    * The C-style [] operator provides unchecked access.
    * The two-argument function call operator provides optional checked access
    * when Check is true, and FORTRAN-style indexing when Ftnidx is true.
    * These are pure virtual as they depend on the matrix layout.
    */
    virtual typename vector<T_>::const_iterator operator[](size_t Idx) const =0;
    virtual typename vector<T_>::iterator operator[](size_t Idx) =0;
    virtual const T_& operator()(size_t Ridx, size_t Cidx) const =0;
    virtual T_& operator()(size_t Ridx, size_t Cidx) =0;
    
    /**
    * row(), col(): read or write the Idx:th row or column
    * as a Vector_<T_> object.
    * Pure virtual: depends on layout.
    */
    virtual Vector_<T_> row(size_t Idx) const =0;
    virtual void row(const Vector_<T_>& Vec, size_t Idx) =0;
    virtual Vector_<T_> col(size_t Idx) const =0;
    virtual void col(const Vector_<T_>& Vec, size_t Idx) =0;
    
    /// set_values(): sets all elements to X (0 by default).
    Basemat_<T_>& set_values(const T_& X=T_(0))
    {
	clone();
    	fill(elems().begin(), elems().end(), X);
    	return(*this);
    }

    /**
    * ftn_idx(): returns true if FORTRAN indexing is used, false otherwise.
    * ftn_idx(Fi): sets the FORTRAN indexing to Fi, returns old value.
    * The Ftnidx status is ignored when [] indexing is used.
    */
    bool ftn_idx() const { return Ftnidx; }
    bool ftn_idx(bool Fi) { bool Of=Ftnidx; Ftnidx=Fi; return Of; }
    
    /**
    * range_check(): returns true if range checking is enabled in operator().
    * range_check(Ci): sets the range_check status, returns old value.
    * Range checking is ignored when [] indexing is used.
    */
    bool range_check() const { return Check; }
    bool range_check(bool Ci) { bool Oc=Check; Check=Ci; return Oc; }
    
    /**
    * simple arithmetics
    * Multiplication and division by a "scalar" (*=, /=).
    * Consider Mat*=(1.0/X) instead of Mat/=X if X is floating-point.
    */
    Basemat_<T_>& operator*=(const T_& Factor);
    Basemat_<T_>& operator/=(const T_& Div);
    
    	/// Diagonals
    
    /**
    * diag(): copies the diagonal into a Vector_ object.
    * diag(Vec): sets the diagonal to the values in the Vector_ object Vec.
    */
    virtual Vector_<T_> diag() const =0;
    virtual void diag(const Vector_<T_>& Vec) =0;

    /**
    * diag_matrix(Dval): turns the calling object into a diagonal matrix 
    * with the value Dval in all diagonal positions.
    */
    virtual void diag_matrix(const T_& Dval) =0;

    /// trace(): returns the sum of the diagonal elements.
    virtual T_ trace() const =0;

    /**
    * transform(): this method allows the modification of each
    * element of the calling matrix object "in place" using the
    * STL function object Unaryfunc. Could be used for things like
    * taking the square root of each matrix element.
    */
    void transform(unary_function<T_, T_> Unaryfunc)
    {
	clone();
	transform(elems().begin(), elems().end(),
		elems().begin(), Unaryfunc);
    }

	// I/O
    
    /**
     * read_matrix(In):
     * reads a matrix from the stream In.
     * The appropriate version will be invoked by the overloaded >> method
     * in each derived class.
     */
    virtual void read_matrix(istream& In) =0;
    
    /**
     * Lists the calling object to the stream Out.
     * The parameters provide formatting facilities,
     * normally you would just use the overloaded << method
     * that invokes list_matrix() with the default parameters.
     */
    virtual void list_matrix(ostream& Out, 
	    unsigned int Prec=6, unsigned int Width=13, 
	    unsigned int Linewidth=80) const;

    // hidden methods
    protected:
	    
    // direct data access (no non-const cloning!)
    const vector<T_>& elems() const { return data().elems(); }
    vector<T_>& elems() { return data().elems(); }
    const T_& elems(size_t Idx) const { return data().elems()[Idx]; }
    T_& elems(size_t Idx) { return data().elems()[Idx]; }
    size_t eno() const { return data().eno(); }
    size_t eno(size_t E) { return data().eno(E); }

    virtual void print_rows(ostream& Out, size_t Rw, size_t Jbeg, 
	size_t Items, size_t Width, size_t Prec) const =0;

    private:
    static size_t prt_width(int N) { return int(ceil(log10(double(N)))); }
};
// END OF CLASS Basemat_

// ---- OUT-OF-CLASS FUNCTIONS ----

// These were originally friend templates but EGCS 1.1.2
// did not like them. 12-Jul-1999.

/// Reads a matrix from In.
template <class T_>
istream& operator >> (istream& In, Basemat_<T_>& Mat)
{
    Mat.read_matrix(In); return(In);
}

/// Writes a matrix to Out.
template <class T_>
ostream& operator << (ostream& Out, const Basemat_<T_>& Mat)
{
    Mat.list_matrix(Out); return(Out);
}

} // RazorBack

// ==== END OF HEADER Basemat.hh ====

#ifdef INCLUDE_TMPL_DEFS
#  include "Basemat.cc"
#endif

#endif	    // BASEMAT_TMPL_HEADER
