#ifndef DERIVHDL_TMPL_HEADER
#define DERIVHDL_TMPL_HEADER

// ==== TEMPLATE HEADER Derivhdl.hh ====

/**
 * \file Derivhdl.hh
 * Handler and reference counting class template
 * for class hierarchies.
 * Its main use is to provide an adapter for storing a mixture
 * of classes derived from a common base in STL containers.
 */

// 22-Jan-2002, Andras Aszodi

// ==== STANDARD HEADERS ====

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>

// ==== MODULE HEADERS ====

#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * This class implements a reference-counted storage
 * for objects derived from the same base class.
 * The Derivhdl_<BaseT_> provides the external interface,
 * and holds a pointer to a Refcount_<BaseT_> object.
 * The Refcount_<BaseT_> class may be pointed to by
 * several Derivhdl_ classes. Refcount_<BaseT_> holds one BaseT_ base type pointer
 * and owns the object pointed to by BaseT_. That object itself
 * may be of type BaseT_ or any class derived from BaseT_.
 * \verbatim
   Derivhdl_<BaseT_>  Derivhdl_<BaseT_>
       |                 |
       +------+   +------+
              |...|
   	      V   V
   	 Refcount_<BaseT_>
   	      |
   	      V
   	    BaseT_
   \endverbatim
 * The purpose of all this is to provide an interface
 * for storing a mixture of class objects derived from the
 * same BaseT_ in STL containers. My other reference counter
 * storage template (see "Refstorage.hh") was designed for
 * storing objects of the same type.
 * BaseT_ class requirements:-
 * \verbatim
   1) The BaseT_ class must have a virtual constructor method
   that looks like
   
   BaseT_* T_::virtual_ctor() const { return new T_(*this); }
   
   where T_ is a type possibly derived from BaseT_. This function
   shall return a BaseT_ pointer to a freshly allocated
   copy of the calling object.
   
   2) BaseT_ shall provide the less-than comparison operator <
   to be used by the STL. This is called via the comparison operator
   of Derivhdl_.
   \endverbatim
 */
template <class BaseT_>
class Derivhdl_
{
    // data
    private:
	    
    /**
     * Refcount_: member classlet for reference-counted storage
     * and ownership of objects pointed to by a BaseT_* pointer.
     */
    struct Refcount_
    {
	// data
	int Refcount;   //< reference counting
	BaseT_*  Ptr;    //< pointer to base type

	/**
	 * Init with a BaseT_ object (deep copy)
	 * pointed to by Bptr. If NULL (the default),
	 * init to empty
	 */
	explicit Refcount_(const BaseT_* Bptr): 
		Refcount(0), Ptr(Bptr==NULL? NULL: Bptr->virtual_ctor()) {}

	/// copy ctor, for deep copy
	Refcount_(const Refcount_& Rc): 
		Refcount(0), Ptr(Rc.Ptr==NULL? NULL: Rc.Ptr->virtual_ctor()) {}

	/// destructor
	~Refcount_() { delete Ptr; }

    };

    Refcount_ *Dataptr;	    //< pointer to reference-counted internal object
    
    // methods
    public:
	    
    /// default ctor: init to empty
    Derivhdl_(): Dataptr(NULL) {}
    
    /// init with a BaseT_ object, stores a deep copy: also BaseT_ -> wrapper conversion
    Derivhdl_(const BaseT_& B): Dataptr(new Refcount_(&B))
    {
	Dataptr->Refcount++;	// sets to 1
    }
    
    /// copy ctor, implements shallow copy
    Derivhdl_(const Derivhdl_& D): Dataptr(D.Dataptr)
    {
	if (Dataptr!=NULL) Dataptr->Refcount++;
    }
    
    /// destructor
    ~Derivhdl_() { unlink_old(); }
    
    /// assignment
    Derivhdl_& operator=(const Derivhdl_& Rhs);
    
    /// empty()
    bool empty() const { return (Dataptr==NULL); }
    
    /**
    * ref(): access to the underlying BaseT_ object.
    * Throws Emptyexc_ if the calling object is empty.
    */
    const BaseT_& ref() const;
    BaseT_& ref();
    
    /**
    * ptr(): returns a pointer to the underlying BaseT_ object.
    * Returns NULL if the calling object is empty.
    */
    const BaseT_* ptr() const;
    BaseT_* ptr();

    /**
     * real_ptr(): returns a pointer to the wrapped object of type DerivedT_.
     * dynamic_cast must be able to convert from BaseT_* to DerivedT_*.
     * In general, you should manipulate the wrapped objects only through
     * their virtual methods. However, under some circumstances it might be
     * useful to know and use the "real" type of the polymorphic
     * object wrapped in Derivhdl_ (for instance when doing I/O). If the conversion
     * was unsuccessful or the wrapper was empty then NULL is returned.
     * Use this at least as carefully as you would dynamic_cast! Always test
     * the returned pointer!
     */
    template <class DerivedT_>
    const DerivedT_* real_ptr() const
    {
	return dynamic_cast<const DerivedT_*>(ptr());
    }
    template <class DerivedT_>
    DerivedT_* real_ptr()
    {
	return dynamic_cast<DerivedT_*>(ptr());
    }
    
    /**
    * <: comparison for STL algorithms
    * Empty objects are always "smaller" than non-empty ones
    * Uses the < operator of BaseT_
    */
    bool operator<(const Derivhdl_<BaseT_>& Rhs) const
    {
	return (empty() || !Rhs.empty() && ref()<Rhs.ref()); 
    }
    
    // hidden methods
    protected:
    
    void unlink_old();
};
// END OF CLASS Derivhdl_

} // RazorBack

// ==== END OF TEMPLATE HEADER Derivhdl.hh ====

#ifdef INCLUDE_TMPL_DEFS
#include "Derivhdl.cc"
#endif

#endif		// DERIVHDL_TMPL_HEADER
