// ==== METHODS Distrapprox.cc ====
#include "Distrapprox.hh"

// Class for approximating univariate continuous distributions.

// 26-Sep-2001. Andras Aszodi

// ---- STANDARD HEADERS ----

#ifdef HAVE_STD
#include <sstream>
using namespace std;
#else
#include <strstream.h>
#define istringstream istrstream 
#define ostringstream ostrstream 
#endif

// ---- MODULE HEADERS ---- 
using namespace RazorBack;

// ==== METHODS ====

/*
 * reset(W):
 * Resets the calling object so that it is ready to do an approximation
 * with a bin width W. All previous data are lost.
 * If W<0.0, then its absolute value is used.
 * If W==0.0, then a Badvalexc_ is thrown.
 */
void Distrapprox_::reset(double W)
{
    W=fabs(W);
    if (W==0.0)
	throw Badvalexc_("Distrapprox_::reset()", W);
    Width=W;
    clear();
}
// END of reset(W)

/*
 * clear():
 * clears the calling object. All data are lost except the histogram bin width.
 */
void Distrapprox_::clear()
{
    Binlist.clear();
    Binno=0; Low=High=0.0;
    Pdf.clear(); Cdf.clear();
    Dirty=true;
    Stat.clear();
}
// END of clear()

/// +=: adds the value X (a sample from the distribution to be approximated).
Distrapprox_& Distrapprox_::operator+=(double X)
{
    int Xidx=bin_index(X);  // would fall into the Xidx:th bin
    Bin_ Searchbin(Xidx);
    Biniter_ Pos=lower_bound(Binlist.begin(), Binlist.end(), Searchbin);
    if (Pos->index()==Xidx)
	++(*Pos);   // found, increase count
    else
    	Binlist.insert(Pos, ++Searchbin);	// with count 1
    Stat+=X;
    Dirty=true;     // evaluation will be necessary
    return *this;
}
// END of +=

// evaluate(): normalise the counts and make the PDF and CDF.
void Distrapprox_::evaluate()
{
    if (!Dirty) return; // clean
    unsigned int Total=Stat.data_no();
    if (!Total) return;   // empty
    
    register double Pdfnorm=1.0/(Width*Total), Cdfnorm=1.0/Total,
	    Pdfval, Cdfval;
    register unsigned int i, Sum=0, Pdfcnt;
    
    // set the low and high limits of X covered by the bins
    Low=Binlist.begin()->index()*Width-0.5*Width;
    High=(--Binlist.end())->index()*Width+0.5*Width;
    
    // set up the PDF and CDF arrays
    Binno=(--Binlist.end())->index()-Binlist.begin()->index()+1;
    Pdf.reserve(Binno); fill_n(Pdf.begin(), Binno, 0.0);
    Cdf.reserve(Binno); fill_n(Cdf.begin(), Binno, 0.0);
    
    // normalisation
    Biniter_ Bp, Bpnext;
    unsigned int Idx, Nextidx;
    for (Idx=0, Bp=Binlist.begin(); Bp!=Binlist.end(); )
    {
	Pdfcnt=Bp->count();
	Pdfval=Pdfcnt*Pdfnorm;
	Sum+=Pdfcnt;
	Cdfval=Sum*Cdfnorm;
	Pdf[Idx]=Pdfval;
	Cdf[Idx]=Cdfval;
	Bpnext=Bp;
	++Bpnext;   // Bp+1 is only for random access iterators
	if (Bpnext!=Binlist.end())    // lookahead before last
	{
	    // if Nextidx>Idx+1, there are empty bins coming!
	    Nextidx=Idx+Bpnext->index()-Bp->index();
	    for (i=Idx+1; i<Nextidx; ++i)
	    {
		Pdf[i]=0.0; 	// emptiness is zero PDF
		Cdf[i]=Cdfval;	// CDF stays constant over the empty bins
	    }
	    Idx=Nextidx;    // step
	}
	Bp=Bpnext;  // stepping Bp, not in the for clause though
    }
    Dirty=false;
}
// END of evaluate()

// inv_cdf(Y): approximates and returns X for which CDF(X)=Y.
// Throws Forgottenexc_ if evaluate() was not invoked beforehand.
double Distrapprox_::inv_cdf(double Y) const
{
    if (Y<0.0 || Y>1.0)
	throw Argexc_("Distrapprox_::inv_cdf()", Y, 0.0, 1.0);
    
    if (Y==0.0) return Low;
    if (Y==1.0) return High;
    
    if (Dirty)
	throw Forgottenexc_("Distrapprox_::inv_cdf(Y)", "Distrapprox_::evaluate()");
    
    register unsigned int Idxlow, Idxup;
    register double Xlow, Ylow, Yup;
    
    // interval halving to find the CDF bin for Y
    if (Y<=Cdf[0])    // Y in first half of first bin
    {
	Xlow=Low; Ylow=0.0;
	Yup=Cdf[0];
    }
    else if (Y>Cdf[Binno-1]) // Y in second half of last bin
    {
	Xlow=bin_midpoint(Binno-1); Ylow=Cdf[Binno-1];
	Yup=1.0;
    }
    else    // general case
    {
	register unsigned int Idxm;
	register double Ym;
	Idxlow=0; Idxup=Binno-1;	// first and last bins
	do
	{
	    Idxm=(Idxlow+Idxup)/2;
	    Ym=Cdf[Idxm];
	    if (Ym>=Y) Idxup=Idxm;  // new bracket from above
	    else Idxlow=Idxm;  	    // or from below
	}
	while (Idxup-Idxlow>1);
	Xlow=bin_midpoint(Idxlow); Ylow=Cdf[Idxlow];
	Yup=Cdf[Idxup];
    }
    
    // interpolate X
    double X=Width*(Y-Ylow)/(Yup-Ylow)+Xlow;
    return(X);
}
// END of inv_cdf(Y)

// entropy(): approximates the entropy of the distribution.
// Throws Forgottenexc_ if evaluate() was not invoked beforehand.
double Distrapprox_::entropy() const
{
    if (Dirty)
	throw Forgottenexc_("Distrapprox_::entropy()", "Distrapprox_::evaluate()");
    
    register double Sum=0.0, p, Temp;
    register unsigned int k;
    
    // simple trapezoidal rule
    for (k=0; k<Binno; k++)
    {
	p=Pdf[k];
	Temp=(p==0.0)? 0.0: p*log(p);
	Sum+=2*Temp;
    }
    Sum*=(High-Low)/(2*(Binno+2));
    return(-Sum);
}
// END of entropy()

// ---- I/O (serialization) ----

// The format of a serialized Distrapprox_ object.
// DISTRAPPROX
// WIDTH = 0.1
// STAT
// N = 1000
// MIN = -4.7362
// MAX = 6.30655
// SUM = 957.413
// SUM2 = 4575.58
// SUM3 = 11291.8
// SUM4 = 57072.3
// ENDSTAT
// -47 1
// -45 3
// ...
// 61 2
// 63 1
// ENDDISTRAPPROX

// friend fn defs must be in scope
namespace RazorBack {

ostream& operator<<(ostream& Out, const Distrapprox_& D)
{
    Iovar_<double> Wio("WIDTH", D.Width);
    Out<<"DISTRAPPROX"<<endl;
    Out<<Wio<<endl;
    Out<<D.Stat;    // supplies endl
    list<Distrapprox_::Bin_>::const_iterator Bp;
    for (Bp=D.Binlist.begin(); Bp!=D.Binlist.end(); ++Bp)
	Out<<(*Bp)<<endl;
    Out<<"ENDDISTRAPPROX"<<endl;
    return Out;
}
// END of <<

istream& operator>>(istream& In, Distrapprox_& D)
{
#define LINELEN 120
    char Line[LINELEN];
    bool Dseen=false, Endseen=false;
    
    // read In line-by-line, look for a line that begins with "DISTRAPPROX"
    while (In.getline(Line, LINELEN, '\n').good())
    {
	if (!strncmp(Line, "DISTRAPPROX", 11))
	{
	    Dseen=true;
	    break;
	}
    }
    if (!Dseen) return In;	// not found
    
    // get the bin width
    Iovar_<double> Wio("WIDTH");
    In>>Wio;
    if (!In) return In;     // sg went wrong
    
    // get the full statistics object
    Stat_ S;
    In>>S;
    if (!In) return In;     // could not read the stat part
    
    // get the bins: read until the "ENDDISTRAPPROX" token is seen
    list<Distrapprox_::Bin_> Tempbins;
    Distrapprox_::Bin_ B;
    while (In.getline(Line, LINELEN, '\n').good())
    {
	if (!strncmp(Line, "ENDDISTRAPPROX", 14))
	{
	    Endseen=true;
	    break;
	}
	istringstream Lineistr(Line);
	Lineistr>>B;
	if (!In.good()) continue;    // skip as it was probably not OK
	Tempbins.push_back(B);
    }
    if (!Endseen || !In) return In;     // problems with the bin read part
    
    // looks OK, so fill up the D object
    D.Width=Wio.value();
    D.Stat=S;
    D.Binlist=Tempbins;
    D.Dirty=true;   // must be re-evaluated as PDF,CDF are not serialized
    return In;
#undef LINELEN
}
// END of >>

} // RazorBack

// ---- aux ----

// lin_intp(X, YY, YYmax): linear interpolation for X.
// YY contains the Y values, YYmax the value of YY if X>=Up.
// Call lin_intp(X, Pdf, 0.0) for PDF approximation,
// and lin_intp(X, Cdf, 1.0) for CDF approximation.
// Throws Forgottenexc_ if the data are dirty.
// Protected
double Distrapprox_::lin_intp(double X, const vector<double>& YY, double YYmax) const 
{
    if (Dirty)
	throw Forgottenexc_("Distrapprox_::lin_intp(...)", "Distrapprox_::evaluate()");
    
    if (X<=Low) return 0.0;
    if (X>=High) return YYmax;
    
    register int Idx=locate_index(X);
    register double Xmid, Xlow, Xup, Ylow, Yup;
    
    Xmid=bin_midpoint(Idx);	// midpoint abscissa of the bin of X
    
    // find the coordinates of the midpoints bracketing X
    if (X>=Xmid)    // X is in upper half of bin
    {
	Xlow=Xmid; Ylow=YY[Idx];
	if (Idx==Binno-1)   // last bin
	{
	    Xup=High; Yup=YYmax;	// PDF set to 0.0, CDF 1.0 
	}
	else
	{
	    Xup=Xlow+Width; Yup=YY[Idx+1];
	}
    }
    else    	    // X is in lower half of bin
    {
	Xup=Xmid; Yup=YY[Idx];
	if (Idx==0) 	// first bin
	{
	    Xlow=Low; Ylow=0.0;	// PDF/CDF set to 0.0 outside range
	}
	else
	{
	    Xlow=Xup-Width; Ylow=YY[Idx-1];
	}
    }
    
    // linear interpolation between (Xlow,Ylow)..X..(Xup, Yup)
    double Y=(Yup-Ylow)*(X-Xlow)/Width+Ylow;
    return Y;
}
// END of lin_intp(X, YY, YYmax)

// locate_index(X): returns the index of the PDF/CDF arrays containing X
// (in the range 0..Binno-1) or -1 if X is outside [Low..High].
// Protected
int Distrapprox_::locate_index(double X) const
{
    int Idx=-1;
    if (Low<=X && X<=High)
    {
	Idx=static_cast<int>(floor(Binno*(X-Low)/(High-Low)));
	if (Idx==Binno) Idx--;
    }
    return Idx;
}
// END of locate_index(X)

// ==== END OF METHODS Distrapprox.cc ====
