#ifndef DISTRBASE_HEADER
#define DISTRBASE_HEADER

// ==== HEADER Distrbase.hh ====

/**
 * \file Distrbase.hh
 * Contains an abstract base class for univariate continuous distributions.
 */

// 11-May-2001. Andras Aszodi

// ==== MODULE HEADERS ====

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Distrbase_: an abstract base class for univariate continuous
 * distributions. Classes derived from this base shall be able to
 * return the probability density function (PDF) and the
 * cumulative distribution function (CDF) at a given point,
 * calculate the probability of a given interval,
 * calculate the mean, variance, skewness, kurtosis,entropy etc. of the distribution,
 * generate random numbers from the distribution.
 */
class Distrbase_
{
    // no data
    
    // methods
    public:
    
    virtual ~Distrbase_() {}
    
    /// pdf(X): return the PDF value at X.
    virtual double pdf(double X) const =0;
    
    /// cdf(X): return the CDF value at X.
    virtual double cdf(double X) const =0;
    
    /// inv_cdf(Y): return the inverse CDF at Y.
    virtual double inv_cdf(double Y) const =0;
    
    /// Returns the mean of the distribution.
    virtual double mean() const =0;
    
    /// Returns the variance of the distribution.
    virtual double var() const =0;
    
    /// Returns the skewness of the distribution.
    virtual double skew() const =0;
    
    /// Returns the kurtosis of the distribution.
    virtual double kurt() const =0;
    
    /// Returns the entropy of the distribution.
    virtual double entropy() const =0;
    
    /// random(): generate a random number from the distribution.
    virtual double random() const =0;
};
// END OF CLASS Distrbase_

// ==== END OF HEADER Distrbase.hh ====

// NOTE: there is no associated methods (*.cc) file.

} // RazorBack

#endif	// DISTRBASE_HEADER
