// ==== FUNCTIONS Factanal.cc ====
#include"Factanal.hh"

// Factor analysis module.

// 25-Feb-2000. Andras Aszodi

// ---- STANDARD HEADERS ----

#ifdef HAVE_STD
#include <algorithm>
#include <functional>
using namespace std;
#else
#include <algo.h>
#endif

// ---- MODULE HEADERS ----

#include"Stat2.hh"
#include "Vector.hh"
#include "Rsymdiag.hh"
using namespace RazorBack;

/* ==== FUNCTIONS ==== */

// std_data(X, Mean, Sd): standardizes the data vector X
// by calculating the mean and SD, and then subtracting the
// mean from each entry and divide the centralised value by
// the SD. Returns the original mean in Mean, the SD in Sd.
// Return value: the no. of entries converted or 0 on error.
unsigned int std_data(vector<double>& X, double& Mean, double& Sd)
{
    Stat_ Sr;	    // univariate statistics
    register unsigned int i, N=X.size();
    
    // build the stat
    for (i=0; i<N; i++)
	Sr+=X[i];
    Mean=Sr.avg(); Sd=Sr.sd();
    
    // normalise
    register double S;
    if (Sd<1e-30)
    {
	cerr<<"\n? std_data(): SD="<<Sd<<" small, not used\n";
	S=1.0;
    }
    else S=1.0/Sd;
    for (i=0; i<N; i++)
	X[i]=(X[i]-Mean)*S;
    return(N);
}
// END of std_data(X, Mean, Sd)

// correl_data(X, Y): calculates the correlation coefficient
// between the data vectors X and Y.
// X and Y are assumed to have been standardised by
// std_data() before the call.
// Return value: the correlation coefficient (-1..+1).
double correl_data(const vector<double>& X, const vector<double>& Y)
{
    Stat2_ Sr;   // bivariate statistics
    register unsigned int i, N;
    
    // sanity check
    if (X.size() != Y.size())
    {
	cerr<<"\n? correl_data(): X.size()="<<X.size()<<" != Y.size()="
		<<Y.size()<<", using min\n";
	N=min(X.size(), Y.size());
    }
    else N=X.size();
    
    for (i=0; i<N; i++)
	Sr.add(X[i], Y[i]);
    return(Sr.corr());
}
// END of correl_data(X, Y)

// calc_loadings(Correl, Load, Tol, Qual):
// decomposes the correlation matrix Correl
// so that the loadings will be returned in Load.
// Return value: the number of factors which belong
// to eigenvalues higher than Tol*N, where N is the size of Correl.
// If Qual!=NULL, then a quality coefficient is returned in *Qual.
int calc_loadings(const Symmat_<double>& Correl, Rectmat_<double>& Load,
        double Tol, double *Qual)
{
    Rectmat_<double> Evec; 
    Vector_<double> Eval;
    register unsigned int i, j, N=Correl.rno();
    
    // Get eigenvalues and eigenvectors of the correlation matrix.
    // The eigenvalues should be above Tol*N.
    // Eigenvectors are in the columns of Evec, sorted in descending order
    Rsymdiag_ Diag(Correl);
    int Factno=Diag.evalevec_above(Tol*N, Eval, Evec, true);
    if (Factno<=0)
    {
	cerr<<"\n? calc_loadings(): No eigenvalue above "
		<<(Tol*N)<<endl;
	return Factno;
    }

    // calculate quality if needed: still uses "original" Eval
    if (Qual!=NULL)
    {
    	double Abseval, Abssum=0.0, Newsum=0.0;
	for (i=0; i<Factno; i++)
	{
	    Abseval=fabs(Eval[i]);
	    Abssum+=Abseval; Newsum+=Abseval;
	}
	for (i=Factno; i<N; i++)
	    Abssum+=fabs(Eval[i]);
	*Qual=Newsum/Abssum;
    }
    
    // make up loading vectors in a Factno-dimensional subspace.
    // For this we need the square roots of the eigenvalues
    transform(Eval.begin(), Eval.begin()+Factno, Eval.begin(), ptr_fun<double,double>(sqrt)); // STL

    Load.size(N, Factno);
    for (i=0; i<N; i++)
        for (j=0; j<Factno; j++)
            Load[i][j]=Eval[j]*Evec[i][j];

    return(Factno);
}
// END of calc_loadings(Correl, Load, Tol, Qual)

// ==== END OF FUNCTIONS Factanal.cc ====
