#ifndef FPARCH_TMPL_METHODS
#define FPARCH_TMPL_METHODS

// ==== TEMPLATE METHODS Fparch.cc ====

/*
 * Machine floating-point architecture constants. 
 * Based on freeware from NETLIB
 * originally written by W. J. Cody, Argonne National Laboratory
 * based on an algorithm due to M. Malcolm, Comm. ACM 15 (1972), pp. 949-951,
 * plus some improvements suggested by M. Gentleman and S. Marovich,
 * Comm. ACM 17 (1974), pp. 276-277.
 * \author Andras Aszodi
 * \date 14-Nov-2001.
 */

// ==== METHODS ====
namespace RazorBack {

// ---- Static init ----

template <class REAL>
const REAL Fparch_<REAL>::ZERO=static_cast<REAL>(0);
template <class REAL>
const REAL Fparch_<REAL>::ONE=static_cast<REAL>(1);
template <class REAL>
const REAL Fparch_<REAL>::TWO=static_cast<REAL>(2);

// The default constructor sets all variables. Had they been declared static,
// it would have been necessary to init them like ZERO,ONE,TWO above. However,
// there is a quite complex interdependency between some of them and it would
// have been very tedious to break up the algorithm below into independent pieces.
// Therefore I decided to keep all members non-static and init them in one
// huge constructor.
template <class REAL>
Fparch_<REAL>::Fparch_()
{
    uname(&Utsname);	// UNIX system description
    
    REAL a, b, Temp;
    int i;
    
    // find radix (base of the representation, usually 2)
    a=ONE;
    do
    {
	a=a+a;
	Temp=a+ONE;
	Temp-=a; Temp-=ONE;
    }
    while (Temp==ZERO);
    b=ONE;
    do
    {
	b=b+b;
	Temp=a+b;
	Temp-=a;
    }
    while (Temp==ZERO);
    Radix=static_cast<int>((a+b)-a);

    // find the number of significant Radix-base digits in mantissa
    REAL Beta=static_cast<REAL>(Radix);
    Digits=0;  b=ONE;
    do
    {
	Digits++;
	b*=Beta; Temp=b+ONE;
	Temp-=b; Temp-=ONE;
    }
    while (Temp==ZERO);
    
    // find out floating-point rounding properties
    REAL Beta2=Beta/TWO;
    Rnd=0;
    Temp=a+Beta2;
    Temp-=a;
    if (Temp!=ZERO) Rnd=1;
    REAL Abeta=a+Beta;
    Temp=Abeta+Beta2;
    if (!Rnd && Temp-Abeta !=ZERO) Rnd=2;
    
    // find Negep
    Negep=Digits+3;
    REAL Betainv=ONE/Beta;
    a=ONE;
    for (i=0; i<Negep; ++i)
	a*=Betainv;
    b=a;
    while (true)    // exit on break
    {
	Temp=ONE-a;
	if (Temp-ONE != ZERO) break;
	a*=Beta; --Negep;
    }
    Negep*= -1;
    
    // find Epsneg and Machep
    Epsneg=a;
    Machep=-Digits-3;
    a=b;
    while (true)    // exit on break
    {
	Temp=ONE+a;
	if (Temp-ONE != ZERO) break;
	a*=Beta;
	++Machep;
    }
    if (Radix!=2 && Rnd>0)
    {
	a=(a*(ONE+a))/TWO;
	Temp=ONE-a;
	Temp-=ONE;
	if (Temp==ZERO) Epsneg=a;
    }
    
    // find Eps
    Eps=a;
    if (Radix!=2 && Rnd>0)
    {
	a=(a*(ONE+a))/TWO;
	Temp=ONE+a;
	Temp-=ONE;
	if (Temp!=ZERO) Eps=a;
    }
    
    // find Nguard
    REAL Epsone=Eps+ONE;
    Nguard=0;
    Temp=Epsone;
    Temp*=ONE;
    Temp-=ONE;
    if (!Rnd && Temp!=ZERO) Nguard=1;
    
    // find Expdigits, Minexp, Smallest
    int k, mx, nx=0;
    REAL z=Betainv, y, Tempbeta;
    for (i=0, k=1; /* exit via break */; ++i, k+=k)
    {
	y=z; z=y*y;
	a=z*ONE;
	Temp=z*Epsone;
	if (a+a==ZERO || absolute(z)>=y) break; // underflow here
	Tempbeta=Temp*Betainv;
	if (Tempbeta*Beta==z) break;
    }
    if (Radix!=10)
    {
	Expdigits=i+1;
	mx=k+k;
    }
    else
    {
	// decimal machines only
	Expdigits=2;
	int iz=Radix;
	while (k>=iz)
	{
	    iz*=Radix;
	    ++Expdigits;
	}
	mx=iz+iz-1;
    }
    while(true)	// exit via break-s
    {
	Smallest=y;
	y*=Betainv;
	a=y*ONE;
	Temp=y*Epsone;
	if (a+a==ZERO || absolute(y)>=Smallest) break;
	Tempbeta=Temp*Betainv;
	++k;
	if (Tempbeta*Beta==y && Temp!=y)
	{
	    nx=3; Smallest=y;
	    break;
	}
    }
    Minexp=-k;
    
    // find Maxexp, Largest, adjust Expdigits
    if (mx<=k+k-3 && Radix!=10)
    {
	mx+=mx; ++Expdigits;
    }
    Maxexp=mx+Minexp;
    Rnd+=nx;
    if (Rnd>=2) Maxexp-=2;
    i=Maxexp+Minexp;
    if (Radix==2 && !i) --Maxexp;
    if (i>20) --Maxexp;
    if (a!=y) Maxexp-=2;
    Largest=ONE-Epsneg;
    Temp=Largest*ONE;
    if (Temp!=Largest)
	Largest=ONE-Beta*Epsneg;
    Largest/=(Beta*Beta*Beta*Smallest);
    i=Maxexp+Minexp+3;
    for (int j=0; j<i; ++j)
    {
	if (Radix==2) Largest+=Largest;
	else Largest*=Beta;
    }
}
// END of ctor

} // RazorBack

// ==== END OF TEMPLATE METHODS Fparch.cc ====

#endif	    // FPARCH_TMPL_METHODS
