#ifndef IOVAR_TMPL_METHODS
#define IOVAR_TMPL_METHODS

// ==== TEMPLATE METHODS Iovar.cc ====

// Template class for representing variables
// for nicely formatted ASCII I/O.

// 12-Feb-2002, Andras Aszodi

// ==== STANDARD HEADERS ====

#include <ctype.h>

#ifdef HAVE_STD
#include <iomanip>
using namespace std;
#else
#include <iomanip.h>
#endif

// ==== METHODS ====
namespace RazorBack {

// ---- Ctors ----

// Inits to store a variable with description token Dtok.
// Specification of Dtok is mandatory, the ctor throws
// a "fatal" Emptyexc_ if Dtok is NULL or "".
// Defval is the default value of the variable, T_(0) by default.
// Noc=true if the description token is meant to be case-insensitive,
// this is false by default (exact match required).
// Eq (by default '=') is the "equal" character, if you want
// to omit it, then specify '\0'.
template <class T_>
Iovar_<T_>::Iovar_(const char *Dtok, T_ Defval, bool Noc, char Eq):
    	Nocase(Noc), Eqchar(Eq), Value(Defval)
{
    if (Dtok==NULL || *Dtok=='\0')
	throw Emptyexc_("Iovar_(<empty>)", true);
    Descr=Dtok;
    if (nocase())
    {
	// case-insensitive comparison will be done via uppercase
	string::iterator Sp;
	for (Sp=Descr.begin(); Sp!=Descr.end(); ++Sp)
	    *Sp=toupper(*Sp);
    }
}

// ---- Access ----

// match_descr(Str, Nocase): returns true if Str begins with descr().
// If nocase()==true, then a case-insensitive comparison is done,
// otherwise (the default) an exact match is required.
template <class T_>
bool Iovar_<T_>::match_descr(const string& Str) const
{
    if (!nocase())  // case-sensitive comparison
	return (NPOS!=Str.find(descr()));
    else    // case-insensitive
    {
	// convert the beginning of Str on the fly
	if (Str.size()<dlen()) return false;
	for (register size_t i=0; i<dlen(); ++i)
	    if (Descr[i]!=toupper(Str[i]))
		return false;
	return true;
    }
}
// END of match_descr(Str, Nocase)

// ---- I/O ----

// Input: reads an Iov object from In accepting the formats
// "Descr Value" or "Descr=Value" (where '=' is eqchar() and
// it can be surrounded by >=0 whitespace chars on either side).
// On error, no warnings are printed but In's failbit is set
// and the stream is rewound to its original position.
template <class T_>
istream& operator>>(istream& In, Iovar_<T_>& Iov)
{
    if (!In) return In; // stream is in bad shape, no input
    
#if defined(__GNUC__) && (__GNUC__ <=2) && (__GNUC_MINOR__ <=95)
    size_t Width=Iov.dlen()+1;	// not quite OK in old GNU
#else
    size_t Width=Iov.dlen();
#endif
    string Buf(Width, ' ');   // token buffer
    char Eq='\0';   // eqchar buffer
    
    streampos Origpos=In.tellg();   // save stream position
    bool Error=false;
    
    // look for the token
    while (!Error)
    {
	In>>setw(Width)>>Buf;
	Error=!In;
	if (!Error && Iov.match_descr(Buf))   // found, OK
	    break;
    }

    // look for the equal-char if needed
    if (!Error)
    {
	if (Iov.eqchar()!='\0')
	{
	    In>>ws>>Eq;
	    Error=(!In || Eq!=Iov.eqchar());
	}
	else    // at least one ws is needed
	{
	    Eq=In.get();    // gets next char
	    Error=(!In || !isspace(Eq));
	}
    }

    // look for the value
    if (!Error)
    {
	Error=!(Iov.read_value(In));
    }
    
    if (Error) // failure
    {
	In.seekg(Origpos);
	In.clear(ios::failbit|In.rdstate());
    }
    return In;
}
// END of >>

// Output: uses the format "Descr Value" or "Descr = Value"
// when Eqchar (here '=') is specified. No newline at the end.
template <class T_>
ostream& operator<<(ostream& Out, const Iovar_<T_>& Iov)
{
    Out<<Iov.descr()<<" ";
    if (Iov.eqchar()!='\0') Out<<Iov.eqchar()<<" ";
    return Iov.write_value(Out);
}
// END of <<

// read_value(): reads a value of T_ from In.
template <class T_>
istream& Iovar_<T_>::read_value(istream& In)
{
    if (!In) return In;
    T_ Tempval=value();
    In>>ws>>Tempval;
    if (In) value(Tempval); // set value if OK
    return In;
}

template <class T_>
ostream& Iovar_<T_>::write_value(ostream& Out) const
{
    Out<<value();
    return Out;
}

// ---- Template specialisations for certain Iovar_ types ----

#ifndef RAZORBACK_LIB_COMPILE

// I/O for bool: the case-insensitive tokens "true","on","yes" are accepted
// in addition to "1", and "false","off","no" in addition to 0.
// The output strings are always lowercase "true" and "false".
// The matches are case-independent.

template<> istream& Iovar_<bool>::read_value(istream& In)
{
    if (!In) return In;
    
    string Boolbuf;
    
    In>>setw(6)>>Boolbuf;
    string::iterator Sp;    // uppercase
    for (Sp=Boolbuf.begin(); Sp!=Boolbuf.end(); ++Sp)
	*Sp=toupper(*Sp);
    if (Boolbuf[0]=='0' 
	    || Boolbuf=="FALSE"
	    || Boolbuf=="NO"
	    || Boolbuf=="OFF")
	Value=false;
    else if (Boolbuf[0]=='1' 
	    || Boolbuf=="TRUE"
	    || Boolbuf=="YES"
	    || Boolbuf=="ON")
	Value=true;
    else In.clear(ios::failbit|In.rdstate());	// was something else
    return In;
}

template<> ostream& Iovar_<bool>::write_value(ostream& Out) const
{
    Out<<(value()? "true": "false");
    return Out;
}

#endif

} // RazorBack

// ==== END OF TEMPLATE METHODS Iovar.cc ====

#endif	    // IOVAR_TMPL_METHODS
