// ==== MODULES MTmgr.cc ====
#include"MTmgr.hh"

// Multithread manager classes.

// 20-Sep-2001. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <unistd.h>
#include <string.h>

#ifdef HAVE_STD
#include <iostream>
#include <fstream>
#include <iomanip>
#include <algorithm>
using namespace std;
#else
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <algo.h>
#endif

// ==== MODULE HEADERS ====

using namespace RazorBack;

// ==== METHODS ====

// ---- MTmgrbase_ ----

// is_running(ID):
// returns true if the thread identified by ID is running.
bool MTmgrbase_::is_running(pthread_t ID) const
{
    if (empty()) return false;
    list<pthread_t>::const_iterator P=find(Threads.begin(), Threads.end(), ID);
    return (P!=Threads.end());
}
// END of is_running(ID)

// get_cpuno():
// returns the number of CPUs. Very architecture- and OS-dependent.
size_t MTmgrbase_::get_cpuno()
{
    // Linux is not POSIX-compliant in this yet.
#if defined(__linux)
    #define LINELEN 128
    char Line[LINELEN+1];
    ifstream Info;
    size_t Ncpus=0;
    
    // read "/proc/cpuinfo" and count the "processor" lines
    Info.open("/proc/cpuinfo");
    if (!Info)	// could not open, but
    { 
	Ncpus=1;    // there must be a CPU if this is running :-)
	cerr<<"\n? get_cpuno(): cannot open\"/proc/cpuinfo\""<<endl;
    }
    else
    {
	while (Info.getline(Line, LINELEN, '\n').good())
	    if (!strncmp(Line, "processor", 9)) Ncpus++;
	Info.close();
    }
    return Ncpus;
    #undef LINELEN    
#else
    // POSIX-compliant systems (SGI, Alpha, Sun,...)
    // SGI uses _SC_NPROC_ONLN so we define it as the "real POSIX"
    // constant _SC_NPROCESSORS_ONLN if necessary
    #if defined(_SC_NPROCESSORS_ONLN) && !defined(_SC_NPROC_ONLN)
    	#define _SC_NPROC_ONLN _SC_NPROCESSORS_ONLN
    #endif
    return sysconf(_SC_NPROC_ONLN);     // POSIX no. of online CPUs
#endif
}
// END of get_cpuno()

// ---- MTmgrjoin_ ----

// launch(Threadno, Startfn, Arg):
// launches Threadno joinable threads. These start execution in
// the global C function pointed to by Startfn, with argument Arg.
// The threads will have default attributes.
// Returns the number of threads actually launched.
size_t MTmgrjoin_::launch(size_t Threadno, Threadstartfn_ Startfn, void *Arg)
{
    size_t Tno;
    pthread_t ID;
    int Err;
    
    for (Tno=0; Tno<Threadno; ++Tno)
    {
	Err=pthread_create(&ID, NULL, Startfn, Arg);
	if (Err) break;
	Threads.push_back(ID);
    }
    
    if (Err)
	throw MTmgrexc_("MTmgrjoin_::launch(...)", Tno, Err);
    return Tno;
}
// END of launch(Threadno, Startfn, Arg)

// reap(): 
// reaps all threads launched by the calling object.
// This happens by performing a pthread_join() on all thread IDs
// that are in the internal list of the calling object,
// i.e. invocation of this method means that the invoking thread
// must wait until all threads are reaped.
void MTmgrjoin_::reap()
{
    if (empty()) return;    // don't bother
    
    list<pthread_t>::const_iterator Idp;
    for (Idp=Threads.begin(); Idp!=Threads.end(); ++Idp)
    {
	pthread_join(*Idp, NULL); // waits on it, discards return value
    }
    Threads.clear();	// wipe the list clean
}
// END of reap()

// reap(ID):
// performs a join operation on thread ID (waits on it if necessary).
// Does nothing if the thread ID is invalid.
void MTmgrjoin_::reap(pthread_t ID)
{
    if (empty()) return;
    
    list<pthread_t>::iterator Idp;
    Idp=find(Threads.begin(), Threads.end(), ID);
    if (Idp!=Threads.end())
    {
	pthread_join(*Idp, NULL);
	Threads.erase(Idp);
    }
}
// END of reap(ID)

// ---- MTmgrdetach_ ----

// launch(Threadno, Startfn, Arg):
// launches Threadno detached threads. These start execution in
// the global C function pointed to by Startfn, with argument Arg.
// The threads will have default attributes otherwise,
// in particular cancellation state enabled, cancellation type deferred.
// Returns the number of threads actually launched.
size_t MTmgrdetach_::launch(size_t Threadno, Threadstartfn_ Startfn, void *Arg)
{
    size_t Tno;
    pthread_t ID;
    int Err;
    
    for (Tno=0; Tno<Threadno; ++Tno)
    {
	Err=pthread_create(&ID, &Attr, Startfn, Arg);
	if (Err) break;
	Threads.push_back(ID);
    }
    
    if (Err)
	throw MTmgrexc_("MTmgrdetach_::launch(...)", Tno, Err);
    return Tno;
}
// END of launch(Threadno, Startfn, Arg)

// reap(): 
// reaps all threads launched by the calling object.
// This happens by sending a pthread_cancel() request on all thread IDs
// that are in the internal list of the calling object.
// The threads will be canceled only if they did not change
// their original cancellation state!
// The method returns immediately (no wait on other threads).
void MTmgrdetach_::reap()
{
    if (empty()) return;    // don't bother
    
    list<pthread_t>::const_iterator Idp;
    for (Idp=Threads.begin(); Idp!=Threads.end(); ++Idp)
    {
	pthread_cancel(*Idp);
    }
    Threads.clear();	// wipe the list clean
}
// END of reap()

// reap(ID):
// performs a cancel operation on thread ID.
// Does nothing if the thread ID is invalid.
// The threads will be canceled only if they did not change
// their original cancellation state!
// The method returns immediately (no wait on other threads).
void MTmgrdetach_::reap(pthread_t ID)
{
    if (empty()) return;
    
    list<pthread_t>::iterator Idp;
    Idp=find(Threads.begin(), Threads.end(), ID);
    if (Idp!=Threads.end())
    {
	pthread_cancel(*Idp);
	Threads.erase(Idp);
    }
}
// END of reap(ID)

// ---- MTmgrexc_ ----

/**
 * Constructs an exception that is thrown by the method called Thnm,
 * the index of the offending thread is in Tno,
 * the errno value is in Eno. F, the fatality indicator,
 * is false by default (ie. only a warning is issued)
 */
MTmgrexc_::MTmgrexc_(const char *Thnm, int Tno, int Eno, bool F)
	: Utilsexc_(Thnm, F), Errno(Eno)
{
    ostringstream Ost;
    print_prefix(Ost);
    Ost<<"while launching thread #"<<Threadno<<": ";
    switch (Errno)
    {
	case EAGAIN:
	    Ost<<"Too many threads or not enough memory";
	break;
    	case EPERM:
	    Ost<<"Insufficient privileges";
	break;
	default:
	    Ost<<strerror(Errno);	// whatever system message
    }
    Message=Ost.str();
}

// ==== END OF MTmgr.cc ====
