#ifndef MTMGR_HEADER
#define MTMGR_HEADER

// Warn if vital macros are missing. Platform-dependent.
#if defined(__linux) || defined(__alpha)
#ifndef _REENTRANT
    	#error Macro _REENTRANT must be defined for POSIX threads!
#endif
#endif

#ifdef __sgi
#if !defined(_POSIX_C_SOURCE) || !defined(_PTHREADS)
    	#error Macros _POSIX_C_SOURCE=199506L and _PTHREADS are needed for POSIX threads under IRIX!
#endif
#endif
// END of warnings

// ==== HEADER MTmgr.hh ====

/** 
 * \file MTmgr.hh
 * POSIX thread manager classes.
 */

// 20-Sep-2001. Andras Aszodi

// ==== STANDARD HEADERS ====

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <pthread.h>	// POSIX threads, must be included first on Alphas
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_STD
#  include <list>
  using namespace std;
#else
#  include <list.h>
#endif

// ==== MODULE HEADERS ====

#include "Utilsexc.hh"

// ==== TYPEDEFS ====

namespace RazorBack {

/**
 * Threadstartfn_: signature of a global C function
 * where a newly launched thread starts executing.
 */
typedef void *(*Threadstartfn_)(void *);

// ==== CLASSES ====

/**
 * Class MTmgrbase_:
 * abstract base class for the management of threads.
 * Classes derived from this ABC can launch new threads
 * and stop them (all or some).
 */
class MTmgrbase_
{
    // data
    // no copy or assignment!
    protected:
	    
    list<pthread_t> Threads;	///< list of thread IDs
    
    /// methods
    public:
	   
    /// Init to empty 
    MTmgrbase_(): Threads() {}
    
    /// dtor: reap all threads still running
    virtual ~MTmgrbase_() {}
    
    	/// job control
    
    /**
    * launch(Threadno, Startfn, Arg):
    * launches Threadno threads. These start execution in
    * the global C function pointed to by Startfn, with argument Arg.
    * Returns the number of threads actually launched.
    */
    virtual size_t launch(size_t Threadno, Threadstartfn_ Startfn, void *Arg) =0;
    
    /**
    * reap(): reap all threads currently running.
    * reap(ID): reap the thread identified by ID. Do nothing
    * if that ID is nonexistent.
    */
    virtual void reap() =0;
    virtual void reap(pthread_t ID) =0;
    
    	/// access 
    
    /**
    * get_cpuno():
    * returns the number of CPUs. Very architecture- and OS-dependent.
    */
    static size_t get_cpuno();
    
    /// empty(): returns true when no threads are running.
    bool empty() const { return Threads.empty(); }
    
    /// thread_no(): returns the number of currently running threads.
    size_t thread_no() const { return Threads.size(); }
    
    /**
    * is_running(ID):
    * returns true if the thread identified by ID is running.
    */
    bool is_running(pthread_t ID) const;
    
    /**
    * run_list():
    * enables const access to the list of thread IDs.
    */
    const list<pthread_t>& run_list() const { return Threads; }
    
    private:
    MTmgrbase_(const MTmgrbase_&);
    MTmgrbase_& operator=(const MTmgrbase_&);
};
// END OF CLASS MTmgrbase_

/**
 * Class MTmgrjoin_:
 * manages joinable threads. Most useful in situation where
 * the original thread just has to launch a few "working" threads
 * and then collects them before the program exits.
 */
class MTmgrjoin_: public MTmgrbase_
{
    // methods
    public:
	    
    /// default ctor
    MTmgrjoin_(): MTmgrbase_() {}
    
    /// dtor: reap all threads still running
    virtual ~MTmgrjoin_() { reap(); }
    
    /**
    * launch(Threadno, Startfn, Arg):
    * launches Threadno joinable threads. These start execution in
    * the global C function pointed to by Startfn, with argument Arg.
    * The threads will have default attributes.
    * Returns the number of threads actually launched.
    */
    virtual size_t launch(size_t Threadno, Threadstartfn_ Startfn, void *Arg);
    
    /**
    * reap(): 
    * reaps all threads launched by the calling object.
    * This happens by performing a pthread_join() on all thread IDs
    * that are in the internal list of the calling object,
    * i.e. invocation of this method means that the invoking thread
    * must wait until all threads are reaped.
    */
    virtual void reap();
    
    /**
    * reap(ID):
    * performs a join operation on thread ID (waits on it if necessary).
    * Does nothing if the thread ID is invalid.
    */
    virtual void reap(pthread_t ID);
    
    private:
    MTmgrjoin_(const MTmgrjoin_&);
    MTmgrjoin_& operator=(const MTmgrjoin_&);
};
// END OF CLASS MTmgrjoin_

/**
 * Class MTmgrdetach_:
 * class for launching detached threads.
 * The MTmgrdetach_ object sends cancel signals to them when reaping
 * and does not wait for them.
 */
class MTmgrdetach_: public MTmgrbase_
{
    // data
    private:
	    
    pthread_attr_t Attr;    ///< POSIX thread attribute
    
    // methods
    public:
	    
    /// ctor
    MTmgrdetach_(): MTmgrbase_()
    {
	pthread_attr_init(&Attr);
	pthread_attr_setdetachstate(&Attr, PTHREAD_CREATE_DETACHED);
    }
    
    /// dtor
    virtual ~MTmgrdetach_() { reap(); pthread_attr_destroy(&Attr); }
	    
    /**
    * launch(Threadno, Startfn, Arg):
    * launches Threadno detached threads. These start execution in
    * the global C function pointed to by Startfn, with argument Arg.
    * The threads will have default attributes otherwise,
    * in particular cancellation state enabled, cancellation type deferred.
    * Returns the number of threads actually launched.
    */
    size_t launch(size_t Threadno, Threadstartfn_ Startfn, void *Arg);

    /**
    * reap(): 
    * reaps all threads launched by the calling object.
    * This happens by sending a pthread_cancel() request on all thread IDs
    * that are in the internal list of the calling object.
    * The threads will be canceled only if they did not change
    * their original cancellation state!
    * The method returns immediately (no wait on other threads).
    */
    void reap();

    /**
    * reap(ID):
    * performs a cancel operation on thread ID.
    * Does nothing if the thread ID is invalid.
    * The threads will be canceled only if they did not change
    * their original cancellation state!
    * The method returns immediately (no wait on other threads).
    */
    void reap(pthread_t ID);
    
    private:
    MTmgrdetach_(const MTmgrdetach_&);
    MTmgrdetach_& operator=(const MTmgrdetach_&);
};
// END OF CLASS MTmgrdetach_

/**
 * Exception class for MTmgr. Gets thrown when for some reason 
 * the threads could not be started. Stores the errno value
 * set by the POSIX thread routines.
 */
class MTmgrexc_: public Utilsexc_
{
    // data
    private:
	    
    int Errno;	///< stores the errno variable from <errno.h>
    int Threadno;   ///< stores the idx of the thread that caused trouble
    
    // methods
    public:
	    
    /**
     * Constructs an exception that is thrown by the method called Thnm,
     * the index of the offending thread is in Tno,
     * the errno value is in Eno. F, the fatality indicator,
     * is false by default (ie. only a warning is issued)
     */
    explicit MTmgrexc_(const char *Thnm, int Tno, int Eno, bool F=false);
    
    /// Returns the index of the thread that caused trouble.
    int thread_idx() const { return Threadno; }
    
    /// Returns the errno that caused the exception.
    int err_no() const { return Errno; }
    
    private:
    MTmgrexc_();    // no default ctor
};
// END OF CLASS MTmgrexc_

} // RazorBack

// ==== END OF HEADER MTmgr.hh ====

#endif	    // MTMGR_HEADER
