#ifndef MTQUEUE_HEADER
#define MTQUEUE_HEADER

// ==== MTqueue.hh ====

/**
 * \file MTqueue.hh
 * Classes that represent a "job queue" for multiple threads.
 * A job object in the queue shall contain all information necessary
 * for a thread to carry out its task.
 * The queue is fed by threads that request certain operations to be
 * carried out, and "working threads" at the other and of the queue
 * perform the requests. Jobs can be prioritised according to their
 * estimated complexity and the job queue
 * does job balancing by sending the "biggest" jobs to the working threads first.
 */

// 18-May-2000, Andras Aszodi

// ==== STANDARD HEADERS ====

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>

#ifdef HAVE_STD
#  include <queue>
#  include <vector>
#  include <list>
  using namespace std;
#else
#  include <stack.h>
#  include <vector.h>
#  include <list.h>
#endif

// ==== MODULE HEADERS ====

#include"MTsync.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Class MTjobbase_:
 * abstract base class for job objects.
 * The derived objects shall define the methods
 * for carrying out the task and a measure for
 * calculation complexity for setting up a priority queue for jobs.
 * The class of user job objects must always be derived from MTjobbase_.
 */
class MTjobbase_
{
    // data
    private:
	    
    double Complexity;	///< some sort of measure of the complexity of the task
    
    // methods
    public:
    
    /// default ctor
    MTjobbase_(): Complexity(0.0) {}
    
    /// copy ctor
    MTjobbase_(const MTjobbase_& Jb): Complexity(Jb.complexity()) {}
    
    /// dtor
    virtual ~MTjobbase_() {}
    
    public:
    
    /**
    * task():
    * this virtual method shall carry out all computations 
    * in the thread that got the job.
    * All necessary data shall be member variables of
    * the job class derived from MTjobbase_,
    * no extra parameters are accepted.
    * The method may choose to return information
    * pointed to by this dirty void*, or (this is preferred)
    * the derived job class shall be laid out so that
    * the return info is also contained in it.
    */
    virtual void* task() =0;
    
    /**
    * complexity([C]):
    * returns [sets] a measure of the task complexity [to C].
    * [Only derived classes may set this value.]
    */
    double complexity() const { return Complexity; }
    
    /// comparison operator required by the job queue
    bool operator<(const MTjobbase_& Rhs) const
    {
	return (complexity()<Rhs.complexity()); 
    }
    
    // hidden methods
    protected:
	    
    /// Sets the complexity of the job to C.
    void complexity(double C) { Complexity=C; }
};
// END OF CLASS MTjobbase_

/**
 * Class MTqueue_:
 * this class sets up a priority queue that pipes "jobs"
 * to a set of working threads. The jobs must be independent of each other.
 */
class MTqueue_
{
    /// member classlets
    public:
    
    /**
    * Jobq_: this is the object that is held in the job priority queue.
    * Contains a ptr to the Derivhdl_-wrapped actual job description object
    * that is submitted by a "producer". Can also be sorted.
    */
    struct Jobq_
    {
	/// data
	MTjobbase_ *Jobptr; 	///< base ptr to somebody's job
	Semaphore_ *Doneptr;  	///< points to a semaphore to be signalled upon job completion
	
	/**
	 * Inits with a job pointed to by Jptr and a semaphore
	 * pointed to by Dptr.
	 */
	explicit Jobq_(MTjobbase_ *Jptr=NULL, Semaphore_ *Dptr=NULL): 
		Jobptr(Jptr), Doneptr(Dptr) {}
	bool operator<(const Jobq_& Rhs) const
	{
	    return Jobptr->complexity()<Rhs.Jobptr->complexity();
	}
	bool empty() const { return Jobptr==NULL; }
    };
    // END OF STRUCT Jobq_
	
    // data
    private:
	   
    priority_queue<Jobq_> Jobqueue;	///< jobs queued for the working threads
    mutable Mutex_ Qmutex;  	///< controls access to the queue
    mutable Condvar_ Todo;	///< to alert working threads to do some job
    
    // methods
    public:
	    
    /// Init to empty
    MTqueue_(): Jobqueue(), Qmutex(), Todo() {}
    
    /**
    * submit_jobs(First, Last):
    * pushes a range of job objects between the iterators First and Last
    * onto the priority queue and waits until all jobs
    * have been completed by the working threads.
    * The job objects may be stored in any STL sequence,
    * vectors and lists are good choices.
    * It is assumed that the objects in the sequence between First and Last
    * are of a type derived from MTjobbase_.
    */
    template <class Fwditer_>
    void submit_jobs(Fwditer_ First, Fwditer_ Last);

    /**
    * wait_4job(): this is to be used by the working threads
    * to wait until someone pushes a job onto the queue.
    */
    void wait_4job() const { Todo.wait(); }
    
    /**
    * pop_job():
    * takes a job object wrapper from the priority queue,
    * removes it from the queue and returns it.
    * If the queue was empty then an empty wrapper is returned.
    */
    MTqueue_::Jobq_ pop_job();

};
// END OF CLASS MTqueue_

// being a template fn, the definition is needed here.
template <class Fwditer_>
void MTqueue_::submit_jobs(Fwditer_ First, Fwditer_ Last)
{
    // Not the job objects themselves, but the Jobq_ wrappers
    // will be pushed onto the job queue
    Semaphore_ Done;	// signals when all jobs are done
    size_t Jobno;
    Qmutex.lock();
    for (Jobno=0; First!=Last; ++First, ++Jobno)
    {
	Jobq_ Jq(&(*First), &Done);	// init queue wrapper
	Jobqueue.push(Jq);  // push wrapper onto queue
	Todo.signal();   // alert working threads
    }
    Done.increment(Jobno);  // set semaphore
    Qmutex.unlock();

    // now we start waiting on the semaphore
    // must receive wakeup signal when all jobs are done
    Done.wait();
}

// ==== GLOBAL FUNCTIONS ====

/**
 * MTqueue_start_workthread(Qptr):
 * this is a global C function that must be executed by the
 * "working threads" that process the jobs submitted to the queue.
 * The working threads run indefinitely if not cancelled explicitly,
 * so launch them via MTmgrdetach_ as detached threads.
 * The MTqueue object supplying the jobs is pointed to by Qptr;
 * type safety cannot be guaranteed here, be careful!
 * Returns NULL when the thread finished.
 */
void* MTqueue_start_workthread(void* Qptr);

} // RazorBack

// ==== END OF HEADER MTqueue.hh ====

#endif	    // MTQUEUE_HEADER
