// ==== METHODS MTsync.cc ====
#include"MTsync.hh"

// Simple C++ wrappers to POSIX thread synchronisation primitives.

// 11-May-2000, Andras Aszodi

// ==== MODULE HEADERS ====

using namespace RazorBack;

// ==== METHODS ====

// ---- Condvar_ ----

// reset():
// sets the Signal to false.
void Condvar_::reset()
{
    pthread_mutex_lock(&Mtx);
    Signal=false;
    pthread_mutex_unlock(&Mtx);
}

// wait():
// locks the associated mutex, reads the condition variable,
// then unlocks Mutex and starts waiting on the condition.
void Condvar_::wait()
{
    pthread_mutex_lock(&Mtx);
    while (!Signal)
    {
	pthread_cond_wait(&Cond, &Mtx);
    }
    pthread_mutex_unlock(&Mtx);
}
// END of wait()

// timed_wait(Sec, Nanosec):
// does the same as wait() but times out after Sec seconds and Nanosec
// nanoseconds have elapsed.
void Condvar_::timed_wait(long Sec, long Nanosec)
{
    struct timespec Abstime;
    Abstime.tv_sec=Sec; Abstime.tv_nsec=Nanosec;
    pthread_mutex_lock(&Mtx);
    while (!Signal)
    {
	pthread_cond_timedwait(&Cond, &Mtx, &Abstime);
    }
    pthread_mutex_unlock(&Mtx);
}
// END of timed_wait(Sec, Nanosec)

// signal(): indicates that the condition is fulfilled and
// one thread waiting on the condition variable can be unblocked.
void Condvar_::signal()
{
    pthread_mutex_lock(&Mtx);
    Signal=true;
    pthread_cond_signal(&Cond);
    pthread_mutex_unlock(&Mtx);
}
// END of signal()

// broadcast(): indicates that the condition is fulfilled and
// all threads waiting on the condition variable may be unblocked.
void Condvar_::broadcast()
{
    pthread_mutex_lock(&Mtx);
    pthread_cond_broadcast(&Cond);
    pthread_mutex_unlock(&Mtx);
}
// END of broadcast()

// ---- Semaphore_ ----

// reset():
// wakes up all threads waiting on the semaphore
// and then sets the value to 0.
void Semaphore_::reset()
{
    pthread_mutex_lock(&Mtx);
    if (Value>0)
	pthread_cond_broadcast(&Cond);
    Value=0;
    pthread_mutex_unlock(&Mtx);
}
// END of reset()

// increment(Inc):
// increases the semaphore value by Inc (1 by default).
void Semaphore_::increment(size_t Inc)
{
    pthread_mutex_lock(&Mtx);
    Value+=Inc;
    pthread_mutex_unlock(&Mtx);
}
// END of increment(Inc)

// wait():
// Waits until the semaphore value is 0.
void Semaphore_::wait()
{
    pthread_mutex_lock(&Mtx);
    while (Value>0)
    {
	pthread_cond_wait(&Cond, &Mtx);
    }
    pthread_mutex_unlock(&Mtx);
}
// END of wait()

// decrement_and_signal(Dec):
// decreases the semaphore value by Dec (1 by default)
// and signals a thread waiting on the semaphore
// when the value goes 0 (or below).
void Semaphore_::decrement_and_signal(size_t Dec)
{
    pthread_mutex_lock(&Mtx);
    Value-=Dec;
    if (Value<=0)
	pthread_cond_signal(&Cond);
    pthread_mutex_unlock(&Mtx);
}
// END of decrement_and_signal(Dec)

// decrement_and_broadcast(Dec):
// same as decrement_and_signal(Dec) but broadcasts
// to all threads waiting on the semaphore when Value goes to 0.
void Semaphore_::decrement_and_broadcast(size_t Dec)
{
    pthread_mutex_lock(&Mtx);
    Value-=Dec;
    if (Value<=0)
	pthread_cond_broadcast(&Cond);
    pthread_mutex_unlock(&Mtx);
}
// END of decrement_and_broadcast(Dec)

// ==== END OF METHODS MTsync.cc ====


