#ifndef MTSYNC_HEADER
#define MTSYNC_HEADER

// Warn if vital macros are missing. Platform-dependent.
#if defined(__linux) || defined(__alpha)
#ifndef _REENTRANT
    	#error Macro _REENTRANT must be defined for POSIX threads!
#endif
#endif

#ifdef __sgi
#if !defined(_POSIX_C_SOURCE) || !defined(_PTHREADS)
    	#error Macros _POSIX_C_SOURCE=199506L and _PTHREADS are needed for POSIX threads under IRIX!
#endif
#endif
// END of warnings

// ==== HEADER MTsync.hh ====

/**
 * \file MTsync.hh
 * C++ wrappers to POSIX thread synchronisation primitives.
 * Note that read-write locks are not implemented yet.
 */

// 9-Apr-2001, Andras Aszodi

// ==== STANDARD HEADERS ====

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <pthread.h>	// POSIX threads, must be included first on Alphas
#include <stdlib.h>
#include <errno.h>

// ==== CLASSES ====

namespace RazorBack {

/**
 * Class Mutex_:
 * wrapper class for POSIX mutexes. Initialisation, destroying
 * and the lock/unlock operations are implemented.
 */
class Mutex_
{
    // data
    private:
	    
    pthread_mutex_t Mutex;  ///< a POSIX mutex
    
    // methods
    public:
	    
    /// Init the mutex to default settings.
    Mutex_() { pthread_mutex_init(&Mutex, NULL); }
    
    /// dtor
    ~Mutex_()
    {
	pthread_mutex_destroy(&Mutex);
    }
    
    /// lock(): locks the calling mutex object.
    void lock() { pthread_mutex_lock(&Mutex); }
    
    /**
    * trylock(): tries to lock the calling mutex object
    * without blocking. Returns true if the locking was successful.
    */
    bool trylock()
    {
	return (EBUSY!=pthread_mutex_trylock(&Mutex));
    }
    
    /// unlock(): unlocks the calling mutex object.
    void unlock() { pthread_mutex_unlock(&Mutex); }
    
    private:
    Mutex_(const Mutex_&);
    Mutex_& operator=(const Mutex_&);
};
// END OF CLASS Mutex_

/**
 * Class Condvar_:
 * wrapper class for condition variables.
 */
class Condvar_
{
    // data
    private:
	    
    pthread_cond_t Cond;    ///< the condition variable
    pthread_mutex_t Mtx;    ///< the associated mutex
    bool Signal;    	    ///< true if somebody signalled
    	    	    	    // (to avoid spurious returns from wait())
    // methods
    public:
	    
    /// Init to default
    Condvar_(): Signal(false)
    {
	pthread_cond_init(&Cond, NULL);
	pthread_mutex_init(&Mtx, NULL);
    }
    
    /// dtor
    ~Condvar_()
    {
	pthread_cond_destroy(&Cond);
	pthread_mutex_destroy(&Mtx);
    }
    
    /**
    * reset():
    * sets the Signal to false.
    */
    void reset();
    
    /**
    * wait():
    * locks the associated mutex, reads the condition variable,
    * then unlocks Mutex and starts waiting on the condition.
    */
    void wait();
    
    /**
    * timed_wait(Sec, Nanosec):
    * does the same as wait() but times out after Sec seconds and Nanosec
    * nanoseconds have elapsed.
    */
    void timed_wait(long Sec, long Nanosec);
    
    /**
    * signal(): indicates that the condition is fulfilled and
    * one thread waiting on the condition variable can be unblocked.
    */
    void signal();
    
    /**
    * broadcast(): indicates that the condition is fulfilled and
    * all threads waiting on the condition variable may be unblocked.
    */
    void broadcast();
    
    private:
    Condvar_(const Condvar_&);
    Condvar_& operator=(const Condvar_&);
};
// END OF CLASS Condvar_

/**
 * Semaphore_: wrapper class for POSIX thread semaphores.
 */
class Semaphore_
{
    // data
    private:
	    
    int Value;
    pthread_cond_t Cond;    ///< condition variable for Value=0
    pthread_mutex_t Mtx;    ///< the associated mutex

    // methods
    /// no copy or assignment!
    public:
	    
    /// Init to default
    Semaphore_(): Value(0)
    {
	pthread_cond_init(&Cond, NULL);
	pthread_mutex_init(&Mtx, NULL);
    }
    
    /// dtor
    ~Semaphore_()
    {
	pthread_cond_destroy(&Cond);
	pthread_mutex_destroy(&Mtx);
    }
    
    /**
    * reset():
    * wakes up all threads waiting on the semaphore
    * and then sets the value to 0.
    */
    void reset();
    
    /**
    * increment(Inc):
    * increases the semaphore value by Inc (1 by default).
    */
    void increment(size_t Inc=1);
    
    /**
    * wait():
    * Waits until the semaphore value is 0.
    */
    void wait();
    
    /**
    * decrement_and_signal(Dec):
    * decreases the semaphore value by Dec (1 by default)
    * and signals a thread waiting on the semaphore
    * when the value goes 0 (or below).
    */
    void decrement_and_signal(size_t Dec=1);
    
    /**
    * decrement_and_broadcast(Dec):
    * same as decrement_and_signal(Dec) but broadcasts
    * to all threads waiting on the semaphore when Value goes to 0.
    */
    void decrement_and_broadcast(size_t Dec=1);
    
    private:
    Semaphore_(const Semaphore_&);
    Semaphore_& operator=(const Semaphore_&);
};
// END OF CLASS Semaphore_

} // RazorBack

// ==== END OF HEADER MTsync.hh ====

#endif	    // MTSYNC_HEADER
