// ==== METHODS Nonlinreg.cc ====
#include "Nonlinreg.hh"

// Nonlinear multiple least-squares regression.

// 19-Sep-2001, Andras Aszodi

// ---- MODULE HEADER ----

using namespace RazorBack;

// ==== METHODS ====

// ---- Nonlinfunc_ METHODS ----

// jac(X, P, Y, J): calculates the Jacobian matrix J=dY/dP
// by a simple approximation. Y==f(X, P). The Jacobian is
// calculated subject to P normalisation, i.e.
// J[i][k]=dY[i]/dP[k]*p[k].
// Return value: 0 if OK, non-0 on error.
// Not const because derived classes may modify internal data.
int Nonlinfunc_::jac(const Vecdbl_& X, const Vecdbl_& P, Vecdbl_& Y, Rmatdbl_& J)
{
    static const double DERIV_COEFF=0.001, DERIV_MINSTEP=1E-6;

    Vecdbl_ Pdelta(np()), Ydelta(ny());
    register double Dp;
    register unsigned int ip;
    int Error=0;

    // numerical derivation
    for (ip=0; ip<np(); ++ip)
    {
	Dp=DERIV_COEFF*fabs(P[ip])+DERIV_MINSTEP;	// shift ip:th par
	Pdelta=P; Pdelta[ip]+=Dp;
	Error=f(X, Pdelta, Ydelta); // new Y at shifted P
	if (Error) break;	// problem

	Ydelta-=Y;	// change in Y
	J.col(Ydelta*(P[ip]/Dp), ip);   // normed Jacobian column
    }
    return Error;
}
// END of jac(X, P, Y, J)
	
// ---- Nonlinreg_ METHODS ----

// nonlin_reg(Xmeas, Ymeas [,Userwgt], Initpar, Funcobj, Itmax, Steplim):
// Least-squares multiple nonlinear regression for the function
// Y=F(X, P) where Y is a Ny-dim. vector, X is a Nx-dim vector,
// and P is a Np-dim. vector of parameters, F a nonlinear
// vector-vector function.
// Routine parameters:-
// Xmeas: Nm x Nx matrix of independent variables
// Ymeas: Nm x Ny matrix of dependent variables
// Userwgt: optional Nm x Ny matrix of weights
//  	can be omitted if equal or relative weights are used
// Initpar: Np-dim. vector of the initial parameter estimates
// Funcobj: a nonlinear function object (instance of class Nonlinfunc_)
//  	that can calculate F and the Jacobian dF/dP
// Itmax: on input, the maximal number of iterations,
//  	on output, the actual number of iterations
// Steplim: minimal stepsize (default 1e-5).
// Return value: the residual deviation.
double Nonlinreg_::nonlin_reg(const Rmatdbl_& Xmeas, const Rmatdbl_& Ymeas, const Rmatdbl_& Userwgt,
    const Vecdbl_& Initpar, Nonlinfunc_& Funcobj, int& Itmax, double Steplim)
{
    // size checks
    adjust_dimensions(Xmeas, Ymeas, Userwgt, Initpar, Funcobj);
    setup_weight(Userwgt);
    
    // iteration parameter checks
    Itmax=abs(Itmax); if (!Itmax) Itmax=100;
    Steplim=fabs(Steplim);
    
    // initial quality
    double Q, Qold;
    P=Initpar;
    Qold=quality(Xmeas, Ymeas, P, Funcobj);
    
    // main iteration cycle
    // Each step is a linear regression.
    // Xtx contains J'WJ, Xty contains J'W(Y-F(X,P))
    Rmatdbl_ Jac(Ny, Np);
    Smatdbl_ Xtxold(Np);
    Vecdbl_ Y(Ny), Dy(Ny), DeltaB(Np), Pnew(Np);
    int Error, Itno=0;
    register unsigned int im, iy, ip, jp;
    double Lambda=0.01, Steplen, Stepfac;
    bool Grow;
    
    do
    {
	Xtx.set_values(); Xty.set_values();
	
	// calculate J'WJ and J'W(Y-F(X,P)) from observed data
	for (im=0; im<Nm; im++)	    // all measurement points
	{
	    Error=Funcobj.f(Xmeas.row(im), P, Y);
	    if (Error)
	    {
		cerr<<"\n? Nonlinreg_::nonlin_reg(): Error "
			<<Error<<" in f()\n";
		return -1.0;
	    }
	    Dy=Ymeas.row(im)-Y; // obs. vs. calc.
	    
	    // numerical derivation for the Jacobian
	    Funcobj.jac(Xmeas.row(im), P, Y, Jac);
	    
	    // now J'WJ and the "right-hand side" J'WDy
	    for (iy = 0; iy < Ny; iy++)
		for (ip = 0; ip < Np; ip++)
		{
		    for (jp = 0; jp <= ip; jp++)
			Xtx[ip][jp] += Jac[iy][ip] * W[im][iy] * Jac[iy][jp];
		    Xty[ip] += Jac[iy][ip] * W[im][iy] * Dy[iy];
		}
	}	// for im
	
	// normalise J'WJ to correlation-matrix
	normalise();
	
	// carry out linear regression step with Marquardt's modif.
	// repeat until quality improves
	Xtxold=Xtx; // save original J'WJ
	Grow=false;
	do
	{
	    if (Grow) Xtx=Xtxold;   // get back J'WJ
	    Lambda=solve(DeltaB, Lambda, 10.0);	// relative step
	    
	    // check stepsize
	    Steplen=vec_len(DeltaB);
	    Stepfac=1.0;
	    for (ip=0; ip<Np; ++ip)
	    {
		if (Stepfac*DeltaB[ip]<-0.95)
		    Stepfac=-0.95/DeltaB[ip];
	    }
	    DeltaB*=Stepfac; Steplen*=Stepfac;
	    
	    // get new parameters (update with corrected step)
	    for (ip = 0; ip < Np; ip++)
		Pnew[ip] = P[ip] * (1.0 + Stepfac * DeltaB[ip]);
	    
	    // get new quality
	    Q=quality(Xmeas, Ymeas, Pnew, Funcobj);
	    
	    // increase Marquardt's lambda if Q grew
	    Grow=(Q>=Qold);
	    if (Grow)
		Lambda=(Lambda>0.0)? 10.0*Lambda: 0.01;
	    else
	    {
		// successful step
		Lambda=(Lambda > 1e-6)? 0.1*Lambda: 0.0;
		P=Pnew; Qold=Q;
	    }
	}
	while (Grow && Steplen >= Steplim / 10);	// linreg step
	Itno++;
	report(cout, Itno, Q, Lambda, P);
    }
    while (Steplen >= Steplim && Itno <= Itmax);	// main cycle end
    Itmax=Itno;     // give back actual no. of iterations
    
    // calc the final statistics
    sdev_correl(Q);

    return Q;
}
double Nonlinreg_::nonlin_reg(const Rmatdbl_& Xmeas, const Rmatdbl_& Ymeas,
    const Vecdbl_& Initpar, Nonlinfunc_& Funcobj, int& Itmax, double Steplim)
{
    if (Weighting==USER)
    {
	cerr<<"\n? Nonlinreg_::nonlin_reg(Xmeas, Ymeas, ...): no user weights, EQUAL weighting used\n";
	Weighting=EQUAL;
    }
    
    // invoke weighted routine with Userwgt==Ymeas
    // will be ignored for EQUAL, and used for RELATIVE
    return nonlin_reg(Xmeas, Ymeas, Ymeas, Initpar, Funcobj, Itmax, Steplim);
}
// END of nonlin_reg(Xmeas, Ymeas [,Userwgt], Initpar, Funcobj, Itmax, Steplim)

// adjust_dimensions(Xmeas, Ymeas, Userwgt, Initpar, Funcobj):
// checks the input sizes, returns false on error, true if OK.
// Adjusts internal sizes except for the weight matrix.
// Hidden
void Nonlinreg_::adjust_dimensions(const Rmatdbl_& Xmeas, const Rmatdbl_& Ymeas,
    const Rmatdbl_& Userwgt, const Vecdbl_& Initpar, const Nonlinfunc_& Funcobj)
{
    // Nx check
    if (Xmeas.cno()!=Funcobj.nx())
	throw Dimexc_("Nonlinreg_::adjust_dimensions(): Nx mismatch", 
		Xmeas.cno(), Funcobj.nx(),true);
    Nx=Xmeas.cno();
    
    // Np check
    if (Initpar.dim()!=Funcobj.np())
	throw Dimexc_("Nonlinreg_::adjust_dimensions(): Np mismatch", 
		Initpar.dim(), Funcobj.np(), true);
    Np=Initpar.dim();
    
    // user weight matrix check
    if (Weighting==USER)
    {
	if (Ymeas.rno()!=Userwgt.rno())
	    throw Dimexc_("Nonlinreg_::adjust_dimensions(): Y|W row mismatch", 
		    Ymeas.rno(), Userwgt.rno(), true);
	if (Ymeas.cno()!=Userwgt.cno())
	    throw Dimexc_("Nonlinreg_::adjust_dimensions(): Y|W col mismatch", 
		    Ymeas.cno(), Userwgt.cno(), true);		
    }
	
    // Ny check
    if (Ymeas.cno()!=Funcobj.ny())
	throw Dimexc_("Nonlinreg_::adjust_dimensions(): Ny mismatch", 
		Ymeas.cno(), Funcobj.ny(), true);
    Ny=Ymeas.cno();
    
    // Nm check
    if (Xmeas.rno()!=Ymeas.rno())
	throw Dimexc_("Nonlinreg_::adjust_dimensions(): Nm mismatch", 
		Xmeas.rno(), Ymeas.rno(), true);
    Nm=Xmeas.rno();
    
    // set DOF
    Nf=Nm*Ny-Np;
    if (Nf<=0)
	throw Toofewexc_("Nonlinreg_::adjust_dimensions()", "Number of measurements", Nm, true);
    
    resize();
}
// END of adjust_dimensions(Xmeas, Ymeas, Initpar, Funcobj)

// setup_weight(YW): inits the weight matrix according to Weighting:
// if EQUAL, then W[i][j]=1, YW ignored;
// if RELATIVE, then YW is interpreted as Ymeas;
// if USER, then YW is interpreted as a user-defined weight matrix.
// Hidden
void Nonlinreg_::setup_weight(const Rmatdbl_& YW)
{
    static const double WMIN=1e-6;
    double Wtemp, Wsum=0;  // for relative weights
    register unsigned int iy, im;
    
    W.size(Nm, Ny);	// set weight matrix size
    switch (Weighting)
    {
	case EQUAL:
	    W.set_values(1.0);
	break;
	case RELATIVE:
	    for (iy=0; iy<Ny; ++iy)
	    {
		Wsum=0.0;
		for (im=0; im<Nm; ++im)
		{
		    Wtemp=fabs(YW[im][iy])+WMIN;
		    Wtemp=1.0/(Wtemp*Wtemp);
		    W[im][iy]=Wtemp;
		    Wsum+=Wtemp;
		}
		for (im=0; im<Nm; ++im)
		    W[im][iy]*=(Nm/Wsum);
	    }
	break;
	case USER:
	    for (iy=0; iy<Ny; ++iy)
	    {
		Wsum=0.0;
		for (im=0; im<Nm; ++im)
		{
		    Wtemp=fabs(YW[im][iy])+WMIN;
		    W[im][iy]=Wtemp;
		    Wsum+=Wtemp;
		}
		for (im=0; im<Nm; ++im)
		    W[im][iy]*=(Nm/Wsum);
	    }
	break;
    }
}
// END of setup_weight(YW)

// quality(Xmeas, Ymeas, Par, Funcobj): calculates the actual
// target function.
// Hidden
double Nonlinreg_::quality(const Rmatdbl_& Xmeas, const Rmatdbl_& Ymeas,
    const Vecdbl_& Par, Nonlinfunc_& Funcobj) const
{
    Vecdbl_ Y(Ny);
    register unsigned int im, iy;
    double Q=0.0;
    int Error;
    
    for (im=0; im<Nm; ++im)
    {
	Error=Funcobj.f(Xmeas.row(im), Par, Y);
	if (Error)
	    throw Utilsexc_("Nonlinreg_::quality()",
		    false, "Error in function evaluation");
	Y-=Ymeas.row(im);   // calc-obs difference
	for (iy=0; iy<Ny; ++iy)
	    Q+=Y[iy]*W[im][iy]*Y[iy];
    }
    return Q;
}
// END of quality(Xmeas, Ymeas, Par, Funcobj)

// report(Out, Itno, Q, Lambda, Par): report the progress
// of the iteration to Out depending on the Verb status.
// SILENT reports nothing, TERSE reports Itno and Q,
// VERBOSE reports all.
// Hidden
void Nonlinreg_::report(ostream& Out, int Itno, double Q, double Lambda,
	const Vecdbl_& Par) const
{
    if (Verb==SILENT) return;
    if (Verb==TERSE || Verb==VERBOSE)
	    Out<<"Step "<<Itno<<", Q="<<Q<<endl;
    if (Verb==VERBOSE)
	    Out<<"Lambda="<<Lambda<<"\nParameters:\n"<<Par<<endl;
}
// END of report(Out, Itno, Q, Lambda, Par)

// ==== END OF METHODS Nonlinreg.cc ====
