// ==== METHODS Orthpoly.cc ====
#include "Orthpoly.hh"

// Parameter estimation with orthogonal polynomials.

// 19-Sep-2001, Andras Aszodi

// ---- STANDARD HEADERS ----

#include <stdlib.h>
#include <math.h>
#include <float.h>

#ifdef HAVE_STD
#include <iostream>
#include <iomanip>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#endif

// ---- MODULE HEADER ----

using namespace RazorBack;

// ==== METHODS ====

// data(Xm, Ym, (Wgt|Wtype)): resets the calling object and
// stores the data points with the weights in it.
// If the weighting type is specified instead as the last parameter,
// then the weight vector will be calculated appropriately from Ym.
// Return value: the number of data points.
int Orthpoly_::data(const Vecdbl_& Xm, const Vecdbl_& Ym,
    const Vecdbl_& Wgt)
{
    Nm=check_dim(Xm, Ym, Wgt);
    Xmeas=Xm; Ymeas=Ym; W=Wgt;
    reset();
    return Nm;
}
int Orthpoly_::data(const Vecdbl_& Xm, const Vecdbl_& Ym,
    Weighting_ Wtype)
{
    setup_weight(Ym, Wtype);
    Nm=check_dim(Xm, Ym, W);
    Xmeas=Xm; Ymeas=Ym;
    reset();
    return Nm;
}
// END of data(Xm, Ym, (Wgt|Wtype))

// setup_weight(Ymeas, Weighting): inits the weight vector according to Weighting:
// if EQUAL, then W={1, 1, ...,1}, Ymeas ignored;
// if RELATIVE, then W[i]=1/Ymeas[i]^2,
// if POISSION, then W[i]=1/Ymeas[i]
// Hidden
void Orthpoly_::setup_weight(const Vecdbl_& Ymeas, Weighting_ Weighting)
{
    static const double WMIN=1e-6;
    double Wtemp, Wsum=0;  // for relative and Poisson weights
    register unsigned int im, Nm;
    
    W.dim(Nm=Ymeas.dim());	// set weight vector size
    switch (Weighting)
    {
	case EQUAL:
	    W.set_values(1.0);
	break;
	case RELATIVE:
	    for (im=0; im<Nm; ++im)
	    {
		Wtemp=fabs(Ymeas[im])+WMIN;
		W[im]=1.0/(Wtemp*Wtemp);
		Wsum+=W[im];
	    }
	    W*=(Nm/Wsum);
	break;
	case POISSON:
	    for (im=0; im<Nm; ++im)
	    {
		Wtemp=fabs(Ymeas[im])+WMIN;
		W[im]=1.0/Wtemp;
		Wsum+=W[im];
	    }
	    W*=(Nm/Wsum);
	break;
    }
}
// END of setup_weight(Ymeas)

// poly_fit(Deg): fits orthogonal polynomials up to degree Deg
// to the data previously supplied via the data(...) method.
// Returns the "goodness-of-fit" parameter.
// Throws Toofewexc_ if there are not enough data points.
double Orthpoly_::poly_fit(unsigned int Deg)
{
    if ((Deg+1)>=Nm)
	throw Toofewexc_("Orthpoly_::poly_fit(Deg)", "Number of measurements", Nm);
    if (Deg==Degree)
    {
	// do nothing
	return good_fit();
    }
    
    register int p;
    resize_param(Deg);	// set output data sizes
    if (Deg>Degree) 	// go up, do higher-order fit
    {
	if (Deg>Maxdegree)  // make more room
	{
	    resize_orth(Deg);	// (increase) orthogonal data sizes
	    Maxdegree=Deg;
	}
	for (p=Degree+1; p<=Deg; ++p)
	{
	    make_orthpoly(p);   // construct polynomials
	    make_orthcoeff(p);	// estimate orthogonal coefficients
	}
    }
    else
    {
	for (p=Degree-1; p>=(int)Deg; --p)
	    back_orthcoeff(p);
    }
    Degree=Deg;
    Makepolycoeff=true; // recalculate "real" polynomial coefficients
    
    return good_fit();
}
// END of poly_fit(Deg)

// make_orthcoeff(Deg): calculates the Deg:th orthogonal polynomial
// coefficient, provided everything up to Deg-1 has been done.
// Hidden
void Orthpoly_::make_orthcoeff(unsigned int Deg)
{
    register unsigned int im, p;
    register double Temp, Ycalc;
    
    // get the Deg:th orthogonal coefficient
    Temp=0.0;
    for (im=0; im<Nm; ++im)
	Temp+=W[im]*Ymeas[im]*Polyval[Deg][im];
    Orthcoeff[Deg]=Temp/Norm[Deg];
    
    // get the residual for the Deg:th polynomial
    Temp=0.0;
    for (im=0; im<Nm; ++im)
    {
	Ycalc=0.0;
	for (p=0; p<=Deg; ++p)
	    Ycalc+=Orthcoeff[p]*Polyval[p][im];
	Ycalc-=Ymeas[im];
	Temp+=W[im]*Ycalc*Ycalc;
    }
    Dres[Deg]=Temp;
    
    // get degrees of freedom, residual deviation
    Nf=im-Deg-1;
    Sres=Dres[Deg]/Nf;
    
    // adjust the variances of the parameters
    if (Deg>0)
    {
	register double Sratio=(Dres[Deg]*(Nf+1))/(Dres[Deg-1]*Nf);
	for (p=0; p<=Deg-1; ++p)
	    Varorth[p]*=Sratio;
    }
    Varorth[Deg]=Sres/Norm[Deg];
}
// END of make_orthcoeff(Deg)

// back_orthcoeff(Deg): assuming the previous polynom degree was Deg+1,
// this method reduces the degree to Deg. This means that the
// degrees of freedom and the polynomial coefficient variances change.
// Hidden
void Orthpoly_::back_orthcoeff(unsigned int Deg)
{
    register unsigned int p;
    register double Sratio;
    
    // degrees of freedom, residual deviation
    Nf=Nm-Deg-1;
    Sres=Dres[Deg]/Nf;
    
    // adjust the variances of the parameters
    Sratio=(Dres[Deg+1]*Nf)/(Dres[Deg]*(Nf-1));
    for (p=0; p<=Deg; ++p)
	Varorth[p]/=Sratio;
}
// END of back_orthcoeff(Deg)

// make_polycoeff(): once the user is satisfied with the orthogonal fit,
// then this method should be called to calculate the parameters and
// standard deviations for the non-orthogonal estimation
// Y = p[0]+p[1]*x+...+p[Degree]*x^Degree.
void Orthpoly_::make_polycoeff()
{
    if (!Makepolycoeff) return; // no need to recalculate
    
    register unsigned int p, q, r;
    register double b0, c0, Sij;
    
    for (r=0; r<=Degree; ++r)
    {
	// calculate the parameters
	b0=0.0;
	for (p=r; p<=Degree; ++p)
	    b0+=Poly[p][r]*Orthcoeff[p];
	Param[r]=b0;
	
	// calculate the parameter covariances
	for (q=0; q<=r; ++q)
	{
	    c0=0.0;
	    for (p=r; p<=Degree; ++p)
		c0+=Poly[p][r]*Poly[p][q]*Varorth[p];
	    Correl[r][q]=c0;
	}
	Sdev[r]=sqrt(Correl[r][r]);	// r:th parameter SD
    }
    
    // normalise covariances to get correlation matrix
    for (p=0; p<=Degree; ++p)
    {
	Correl[p][p]=1.0;
	for (q=0; q<p; ++q)
	{
	    Sij=Sdev[p]*Sdev[q];
	    if (Sij<FLT_EPSILON) Correl[p][q]=0.0;
	    else Correl[p][q]/=Sij;
	}
    }    
    Makepolycoeff=false;    // done
}
    
// approx(X): approximate Y(X) with the current polynomial.
double Orthpoly_::approx(double X) const
{
    register double Temp=Param[Degree];
    register int i;
    for (i=Degree-1; i>=0; --i)
	Temp=Temp*X+Param[i];
    return Temp;
}
// END of approx(X)

// ---- Auxiliaries ----

// check_dim(Xm, Ym, Wgt): returns the common length of the arguments.
// Throws Dimexc_ on dim mismatch.
// Static hidden
int Orthpoly_::check_dim(const Vecdbl_& Xm, const Vecdbl_& Ym,
    const Vecdbl_& Wgt)
{
    int Nm=Xm.dim();
    if (Ym.dim()!=Nm)
	throw Dimexc_("Orthpoly_::check_dim(...), X|Y", Nm, Ym.dim());
    if (Wgt.dim()!=Nm)
	throw Dimexc_("Orthpoly_::check_dim(...), X|W", Nm, Wgt.dim());
    return Nm;	// OK
}
// END of check_dim(Xm, Ym, Wgt)

// reset(): clears the internal data except the data points.
// Prepares to store 0th order polynomials, constructs them as well.
// Hidden
void Orthpoly_::reset()
{
    Degree=0; 
    resize_orth(Degree); resize_param(Degree);
    make_orthpoly(Degree);
    make_orthcoeff(Degree);
}
// END of reset()

// resize_orth(Maxdeg): changes the sizes of the orthogonal data structures
// so that Maxdeg-degree polynomials can be stored.
// resize_param(Maxdeg): does the same thing to the output params only.
// Hidden
void Orthpoly_::resize_orth(unsigned int Maxdeg)
{
    ++Maxdeg;
    Dres.dim(Maxdeg); Norm.dim(Maxdeg);
    Poly.size(Maxdeg); Polyval.size(Maxdeg, Nm);
    Orthcoeff.dim(Maxdeg); Varorth.dim(Maxdeg);
}
void Orthpoly_::resize_param(unsigned int Maxdeg)
{
    ++Maxdeg;
    Param.dim(Maxdeg); Sdev.dim(Maxdeg);
    Correl.size(Maxdeg);
}
// END of resize(Maxdeg)

// make_orthpoly(Deg): constructs the Deg-degree orthogonal polynomial,
// provided the Deg-1:th and Deg-2:th are already available (not checked).
// Hidden
void Orthpoly_::make_orthpoly(size_t Deg)
{
    register size_t p, q, im;
    double Gpp, Gpq, Wxt, Temp;
    
    // special cases
    if (Deg==0)
    {
	Poly[0][0]=1.0;
	Temp=0.0;
	for (im=0; im<Nm; ++im)
	{
	    Polyval[0][im]=1.0;
	    Temp+=W[im];
	}
	Norm[0]=Temp;
	return;
    }
    if (Deg==1)
    {
	Poly[1][1]=1.0;
	Temp=0.0;
	for (im=0; im<Nm; ++im)
	    Temp-=W[im]*Xmeas[im];
	Poly[1][0]=Temp/Norm[0];
	
	double Ptemp=0.0;
	Temp=0.0;
	for (im=0; im<Nm; ++im)
	{
	    Polyval[1][im]=Ptemp=Xmeas[im]+Poly[1][0];
	    Temp+=W[im]*Ptemp*Ptemp;
	}
	Norm[1]=Temp;
	return;
    }
	
    // general case, Deg>=2
    p=Deg-1; Wxt=0.0;
    for (im=0; im<Nm; ++im)
    {
	Temp=Polyval[p][im];
	Wxt+=W[im]*Xmeas[im]*Temp*Temp;
    }
    Gpp=Wxt/Norm[p]; Gpq=Norm[p]/Norm[p-1];
    Poly[Deg][0]=-Gpp*Poly[p][0]-Gpq*Poly[p-1][0];
    for (q=1; q<p; ++q)
    	Poly[Deg][q]=Poly[p][q-1]-Gpp*Poly[p][q]-Gpq*Poly[p-1][q];
    Poly[Deg][p]=Poly[p][p-1]-Gpp;
    Poly[Deg][Deg]=1.0;
    Norm[Deg]=0.0;
    for (im=0; im<Nm; ++im)
    {
    	Temp=Polyval[Deg][im]
		=(Xmeas[im]-Gpp)*Polyval[p][im]-Gpq*Polyval[p-1][im];
    	Norm[Deg]+=W[im]*Temp*Temp;
    }
}
// END of make_orthpoly(Deg)

// ==== END OF METHODS Orthpoly.cc ====
