// ==== METHODS Paramest.cc ====
#include"Paramest.hh"

// Linear and nonlinear parameter estimation.

// 5-Feb-1999, Andras Aszodi

// ---- MODULE HEADER ----

using namespace RazorBack;

// ==== Paramest_ METHODS ====

// tcrit_95(DOF) returns the value of the t-distribution
// at 95% significance level with Nf degrees of freedom.
// Use for confidence intervals.
float Paramest_::tcrit_95() const
{
  double t, TEMP;

  if (Nf <= 5)
    t = 6.415 + (5 - Nf) *
	  (0.289 + (4 - Nf) * (0.0575 + (3 - Nf) * (0.022 + (2 - Nf) * 0.020552)));
  else if (Nf <= 30)
    t = 7.6278 - 0.2316 * Nf + 0.00421 * Nf * Nf - 0.186 * sin(0.2214 * Nf) -
	0.0116 * sin(0.4428 * Nf) + 0.0186 * cos(0.4428 * Nf);
  else if (Nf < 40)
    t = 4.4067 + 0.0296 * (30 - Nf);
  else if (Nf < 60)
    t = 4.1108 + 0.021 * (40 - Nf);
  else
    t = 3.6888 + 0.0116 * (60 - Nf);
  t = exp(t);
  if (Nf == 7)
    t += 2;
  if ((unsigned long)Nf < 32 && ((1L << Nf) & 0x100a100L) != 0)
    t++;
  modf(t, &TEMP);
  return ((float)(TEMP / 1000.0));   /* round to x.xxx */
}
// END of tcrit_95()

// normalise(): normalises the X'WX matrix, returns trace.
// Hidden
double Paramest_::normalise()
{
    register double Trace=Xtx.trace();
    register unsigned int ip, jp;
    
    Trace/=(Np*1000.0);	    // no norm with pivots smaller than this
    Norm=Xtx.diag();
    for (ip=0; ip<Np; ++ip)
	Norm[ip]=(Norm[ip]>Trace)? sqrt(Norm[ip]): 1.0;

    for (ip=0; ip<Np; ++ip)
	for (jp=0; jp<=ip; ++jp)
	    Xtx[ip][jp]/=(Norm[ip]*Norm[jp]);
    return Trace;
}
// END of normalise()

// solve(Sol, Lambda0, Lambdafact): 
// solves the system of linear least-squares equations.
// The matrix to be inverted is always Xtx, the rhs is Xty.
// Uses a ridge parameter for linear problems, or
// Marquardt's lambda for nonlinear problems: it gets increased
// until X'WX can be inverted.
// Lambda0 is the initial Lambda value, Lambdafact is the
// factor by which Lambda should be increased. If Lambdafact==0.0,
// then Lambda will be additively increased by 0.01.
// Return value: the last Lambda value used.
// Hidden
double Paramest_::solve(Vecdbl_& Sol, double Lambda0, double Lambdafact)
{
    // inversion (Lambda is increased if Xtx is singular)
    double Lambda=fabs(Lambda0);
    register unsigned int ip, jp;
    Smatdbl_ Xtxold(Xtx);
    
    while(!posdef_inv(Xtx)) // could not invert?
    {
	Xtx=Xtxold;	// get back old Xtx
	if (Lambdafact!=0.0)
	    Lambda = (Lambda==0.0)? 0.01: Lambdafact*Lambda;
	else
	    Lambda+=0.01;	// increment ridge parameter
	for (ip=0; ip<Np; ++ip)
	    Xtx[ip][ip]+=Lambda;	// and add to diagonal
    }

    // solution
    for (ip=0; ip<Np; ++ip)	// norm back: divide again
	for (jp=0; jp<=ip; ++jp)
	    Xtx[ip][jp]/=(Norm[ip]*Norm[jp]);

    Sol=Xtx*Xty;	// this is the solution
    return Lambda;
}
// END of solve(Rhs, Sol, Lambda0, Lambdafact)

// posdef_inv(A): inverts the positive definite symmetric matrix A.
// Since the inverse is symmetric, too, it is returned in A
// (original values are overwritten). 
// Return value: true if OK, false if A was singular.
// Hidden static
bool Paramest_::posdef_inv(Smatdbl_& A)
{
  register int i, j, k, N=A.rno();
  double *H=new double [N];
  register double At;

  // main elimination cycle
  for (k = N; k >= 1; k--)
  {  
    if (fabs(A[0][0]) < DBL_MIN)    // singular?
    {
	delete [] H;
	return false;
    }
    H[N - 1] = 1 / A[0][0];
    for (i = 2; i <= N; i++)
    {
      At = A[i - 1][0] * H[N - 1];
	H[i - 2] = (i>k)? At: -At;
      for (j = 2; j <= i; j++)
	A[i - 2][j - 2] = A[i - 1][j - 1] + A[i - 1][0] * H[j - 2];
    }
    for (i = 0; i < N; i++)
      A[N - 1][i] = H[i];
  }
  delete [] H; return true;
}
// END of posdef_inv()

// sdev_correl(Q): calculates the standard deviations
// and correlations of the parameters from the residual deviation Q.
// Hidden
void Paramest_::sdev_correl(double Q)
{
    double Sij;
    register unsigned int ip, jp;
    
    Correl=Xtx;
    Correl*=(Q*Q);	// covariance matrix
    Sdev=Correl.diag();	// variances
    for (ip=0; ip<Np; ++ip)
	Sdev[ip]=sqrt(Sdev[ip]);    // std. deviations
    
    for (ip=0; ip<Np; ++ip)	// norm to correlation
    {
	Correl[ip][ip]=1.0;
	for (jp=0; jp<ip; ++jp)
	{
	    Sij=Sdev[ip]*Sdev[jp];
	    if (Sij<FLT_EPSILON) Correl[ip][jp]=0.0;
	    else Correl[ip][jp]/=Sij;
	}
    }
}
// END of sdev_correl(Q)

// ==== END OF METHODS Paramest.cc ====
