// ==== METHODS Random.cc ====
#include"Random.hh"

/*
 * Uniformly distributed random numbers using the 32-bit generator of
 * L'Ecuyer, P. Efficient and portable combined random number
 * generators, C.A.C.M., vol. 31, 742-749, June 1988.
 * Period is very long, about 2.3e18.
 * Plus the Box/Mueller method for standard normal variates.
 * \author Andras Aszodi
 * \date 21-Nov-2001.
 */

// ---- STANDARD HEADERS ----

#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <unistd.h>
#include <time.h>

// ---- MODULE HEADERS ----

using namespace RazorBack;

// ==== Randombase_ METHODS ====

// ---- Preparation ----

/*
 * reset(S1, S2): resets the generator with S1, S2.
 * If S1==0 (default), then getpid() is used as seed.
 * If S2==0 (default), then time(NULL) is used as seed.
 */
void Randombase_::reset(int S1, int S2)
{
    Seed1=!S1? getpid(): S1;
    Seed2=!S2? time(NULL): S2;
} 
// END of reset(Seed)

// ---- RNG ----

// Returns a random int value.
// This is L'Ecuyer's algorithm.
// Protected
int Randombase_::int_rand(void) const
{
    // magic constants
    static const int C1=53668, C2=40014, C3=12211, C4=2147483563,
	    	D1=52774, D2=40692, D3=3791, D4=2147483399;
    register int k, z;
    
    k = Seed1 / C1;
    Seed1 = C2 * (Seed1 - k * C1) - k * C3;
    if (Seed1<0) Seed1 += C4;

    k = Seed2 / D1;
    Seed2 = D2 * (Seed2 - k * D1) - k * D3;
    if (Seed2<0) Seed2 += D4;

    z = Seed1 - Seed2;
    if (z<1) z += 2147483562;

    return z;
}
// END of int_rand()

// dbl_rand(): returns a random double value between 0.0 and 1.0.
// Protected
double Randombase_::dbl_rand(void) const
{
    static const double FACTOR=4.6566130573e-10;    // 1.0/2147483563.0
    return FACTOR*int_rand();
}
// END of dbl_rand()

// ==== Randomnorm_ METHODS ====

// operator()(): returns a random double value
// with zero mean and unit variance (std normal distribution).
double Randomnorm_::operator()(void) const
{
    static const double EPSILON=1e-10;
    static bool Spare=false;	// these could be "mutable" members
    static double Spval=0.0;
    
    register double Fac, R, U1, U2;
    
    if (Spare)	// we had a value from a previous call
    {
	Spare=false;	// spare value consumed
	return Spval;	// use it
    }
    
    // make two values, save one, return the other
    do  // get two uniform random numbers within the unit circle
    {
	U1=2.0*dbl_rand()-1.0;
	U2=2.0*dbl_rand()-1.0;
	R=U1*U1+U2*U2;
    }
    while (R>=1.0 || R<=EPSILON);
    Fac=sqrt(-2.0*log(R)/R);

    // save spare value to be returned next time
    Spval=U1*Fac; Spare=true;

    return(U2*Fac); // return the other
}
// END of operator()()

// ==== END OF METHODS Random.cc ====
