// ==== METHODS Rangedimexc.cc ====
#include"Rangedimexc.hh"

// Exception classes for range and dimension mismatch exceptions.
// All derived from the Utilsexc_ base exception class for my utilities.

// 8-Oct-2001. Andras Aszodi

// ==== STANDARD HEADERS ====

#include <math.h>

// ==== MODULE HEADER ====

using namespace RazorBack;

// ==== METHODS ====

/**
 * Constructs an index range exception thrown in a function called Thnm.
 * Bidx is the out-of-range index, Idx is the maximal allowed index.
 * F is the fatality indicator (default false).
 */
Indexrangexc_::Indexrangexc_(const char *Thnm, unsigned int Bidx, unsigned int Idx, bool F)
    : Utilsexc_(Thnm, F), Badidx(Bidx), Maxidx(Idx)
{
    ostringstream Ost;
    print_prefix(Ost);
    Ost<<"Index "<<Badidx<<" is out of range [0.."<<Maxidx<<"]";
    Message=Ost.str();
}

// ---- Badvalexc_ ----

/**
 * Constructs a bad value exception thrown by Thnm, with fatality indicator F.
 * Bad is the "bad value" that was not accepted for some reason.
 */
Badvalexc_::Badvalexc_(const char *Thnm, double Bad, bool F)
	: Utilsexc_(Thnm, F), Badval(Bad)
{
    ostringstream Ost;
    print_prefix(Ost);
    Ost<<"Argument "<<bad_val()<<" is not allowed";
    Message=Ost.str();
}

// ---- Argexc_ ----

// Static init
const double Argexc_::NEG_INFINITY=-HUGE_VAL;
const double Argexc_::POS_INFINITY=+HUGE_VAL;

/**
 * Constructs an argument range exception thrown by Thnm.
 * Min and/or Max can be specified as NEG_INFINITY and POS_INFINITY.
 * If Min>Max, then they will be swapped silently, however, it is not checked
 * whether Bad is really outside the range [Min..Max] :-)
 */
Argexc_::Argexc_(const char *Thnm, double Bad, double Min, double Max, bool F)
	: Badvalexc_(Thnm, Bad, F)
{
    if (Min<Max)
    {
	Minval=Min; Maxval=Max;
    }
    else
    {
	Minval=Max; Maxval=Min;
    }
    
    ostringstream Ost;
    print_prefix(Ost);
    Ost<<"Argument "<<bad_val()<<" out of range [";
    if (Minval<=NEG_INFINITY) Ost<<"-oo";
    else Ost<<Minval;
    Ost<<"..";
    if (Maxval>=POS_INFINITY) Ost<<"+oo";
    else Ost<<Maxval;
    Ost<<"]";
    Message=Ost.str();
}

// ---- Dimexc_ ----

/**
 * Constructs a dimension mismatch exception thrown by a function Thnm.
 * D1, D2 are the mismatched dimensions, F is the fatality indicator.
 */
Dimexc_::Dimexc_(const char *Thnm, unsigned int D1, unsigned int D2, bool F)
    : Utilsexc_(Thnm, F), Dim1(D1), Dim2(D2)
{
    ostringstream Ost;
    print_prefix(Ost);
    Ost<<"Dimension mismatch ("
	<<Dim1<<"!="<<Dim2<<")";
    Message=Ost.str();
}

// ==== END OF METHODS Rangedimexc.cc ====
