#ifndef RANGEDIMEXC_HEADER
#define RANGEDIMEXC_HEADER

// ==== Rangedimexc.hh ====

/**
 * \file Rangedimexc.hh
 * Exception classes for range and dimension mismatch exceptions
 * in the Razorback library.
 */

// 8-Oct-2001. Andras Aszodi

// ---- STANDARD HEADERS ----

#include <stdlib.h>
#include <string.h>

#ifdef HAVE_STD
#include <iostream>
#include <iomanip>
    using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#endif

// ---- MODULE HEADERS ----

#include"Utilsexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Indexrangexc_: index range exception. Stores the bad index and the max. size
 * both of which must be specified (no empty default ctor).
 * Index should have been in the range [0..Maxsize].
 * Usage: throw Indexrangexc_("wrong_function", BadIdx, Maxsize)
 */
class Indexrangexc_: public Utilsexc_
{
    /// data
    protected:
    
    unsigned int Badidx, Maxidx; ///< bad index causing the exception, max.range
    
    // methods
    public:
    
    /**
     * Constructs an index range exception thrown in a function called Thnm.
     * Bidx is the out-of-range index, Idx is the maximal allowed index.
     * F is the fatality indicator (default false).
     */
    Indexrangexc_(const char *Thnm,
	    unsigned int Bidx, unsigned int Idx, bool F=false);

    /// bad_idx(): returns the "bad" index that caused the exception.
    int bad_idx() const { return Badidx; }
    
    /// max_idx(): returns the maximal index that was violated.
    int max_idx() const { return Maxidx; }
    
    private:
    Indexrangexc_();     // disallow default ctor
};
// END OF CLASS Indexrangexc_

/**
 * Badvalexc_: bad argument exception. Stores the bad value.
 * Usage: throw Badvalexc_("wrong_function", Badval)
 */
class Badvalexc_: public Utilsexc_
{
    // data
    protected:
    
    double Badval;  	    ///< bad argument
    
    // methods
    public:
    
    /**
     * Constructs a bad value exception thrown by Thnm, with fatality indicator F.
     * Bad is the "bad value" that was not accepted for some reason.
     */
    Badvalexc_(const char *Thnm, double Bad, bool F=false);

    /// Returns the "bad" value that caused the exception.
    double bad_val() const { return(Badval); }
    
    private:
    Badvalexc_();	    // disallow default ctor
};
// END OF CLASS Badvalexc_

/**
 * Argexc_: argument range exception. Stores the bad value
 * and the allowed range. The range specification may use the symbolic
 * constants NEG_INFINITY and POS_INFINITY.
 * All of these must be specified (no empty default ctor).
 * Usage: throw Argexc_("wrong_function", Badval, Minval, Maxval)
 */
class Argexc_: public Badvalexc_
{
    // data
    public:
    static const double NEG_INFINITY, POS_INFINITY;
    
    protected:
    double Minval, Maxval; ///< allowed range
    
    // methods
    public:
    
    /**
     * Constructs an argument range exception thrown by Thnm.
     * Min and/or Max can be specified as NEG_INFINITY and POS_INFINITY.
     * If Min>Max, then they will be swapped silently, however, it is not checked
     * whether Bad is really outside the range [Min..Max] :-)
     */
    Argexc_(const char *Thnm, double Bad, double Min, double Max, bool F=false);

    /// Returns the acceptable lower limit.
    double min_val() const { return Minval; }
    
    /// Returns the acceptable upper limit.
    double max_val() const { return Maxval; }
    
    private:
    Argexc_();	    // disallow default ctor
};
// END OF CLASS Argexc_

/**
 * Dimexc_: dimension mismatch exception. Stores the two dimensions
 * which should have been equal. Both of them must be specified
 * (no empty default ctor). 
 * Usage: throw Dimexc_("wrong_function", Dim1, Dim2).
 */
class Dimexc_: public Utilsexc_
{
    //data
    protected:
    
    unsigned int Dim1, Dim2; ///< mismatched dimensions
    
    /// methods
    public:
    
    /**
     * Constructs a dimension mismatch exception thrown by a function Thnm.
     * D1, D2 are the mismatched dimensions, F is the fatality indicator.
     */
    Dimexc_(const char *Thnm, unsigned int D1, unsigned int D2, bool F=false);
    
    /// bad_dim(): i<=1 returns Dim1, i>=2 returns Dim2.
    unsigned int bad_dim(int i) const
    {
	return(i<=1? Dim1: Dim2);
    }
    
    private:
    Dimexc_();	    // disallow default ctor
};
// END OF CLASS Dimexc_

} // RazorBack

// ==== END OF HEADER Rangedimexc.hh ====

#endif	// RANGEDIMEXC_HEADER
