// ==== METHODS Safety.c++ ====
#include"Safety.hh"

// This tiny class implements a few safety precautions
// which can be switched on and off: safe division that
// traps div-by-zero errors, provides a "small number"
// with which it is unsafe to divide a double, and an improved
// implementation of hypot() for those architectures which
// do not provide it.

// 14-Nov-2001. Andras Aszodi

// ---- STANDARD HEADERS ----

#include <stdlib.h>
#include <math.h>

#ifdef HAVE_STD
#include <iostream>
#include <iomanip>
using namespace std;
#else
#include <iostream.h>
#include <iomanip.h>
#endif


// ---- MODULE HEADER ----

using namespace RazorBack;

// ==== OS-DEPENDENT SETTINGS FOR hypot() ====

// SGI: hypot() is not strict ANSI, available only
// if the X/Open compatibility mode is selected
#if defined(__sgi)
    #if defined(_XOPEN4) && defined(_NO_ANSIMODE) \
    	    && !defined(_POSIX_C_SOURCE)
    	#define USE_BUILTIN_HYPOT
    #else
    	#undef USE_BUILTIN_HYPOT
    #endif
#endif

// Linux/GCC: hypot() available
#if defined(__linux__)
    #define USE_BUILTIN_HYPOT
#endif

// Alpha and Tru64 
#if defined(__alpha)
#if defined (_XOPEN_SOURCE)
#define USE_BUILTIN_HYPOT
#else
#undef USE_BUILTIN_HYPOT
#endif
#endif

// add any other OS here
#if defined(__any_other_OS__)
    // .........
#endif
	
// ==== Static initialisation ====

Fparch_<double> Safety_::Fp;
const double Safety_::SMALL=sqrt(Fp.smallest())/Fp.eps();	// unsafe to do 1.0/SMALL

// ==== Safety_ MEMBER FUNCTIONS ====

// safe_div(): returns Num/Denom if Denom is reliably non-0.
// If it is closer to 0 then the internal parameter SMALL, then
// it will be replaced by SMALL and the division is performed
// with a warning. The line number is printed when greater than 0 (default).
double Safety_::safe_div(double Num, double Denom, int Lineno) const
{
    if (Usesafediv && fabs(Denom)<SMALL)  // problem
    {
	cerr<<"\n! Safety_::safe_div("<<Num<<", "<<Denom<<"): Dangerous division";
	if (Lineno) cerr<<" at line "<<Lineno<<endl;
	else cerr<<endl;
	Denom=Denom>=0.0? SMALL: -SMALL;
    }
    return(Num/Denom);
}
// END of safe_div()

// pythag(): returns the value sqrt(a^2+b^2) without over-
// or underflows. On some machines the math library contains
// the function hypot(x, y) which does the same job and will be used
// whenever possible.
double Safety_::pythag(double a, double b) const
{
#ifdef USE_BUILTIN_HYPOT
    return(hypot(a, b));
#else
    register double absa=fabs(a), absb=fabs(b), ratio;

    if (absa<SMALL) return(absb);
    if (absb<SMALL) return(absa);
    if (absa>absb)
    {
	ratio=absb/absa; return(absa*sqrt(1.0+ratio*ratio));
    }
    else
    {
	ratio=absa/absb; return(absb*sqrt(1.0+ratio*ratio));
    }
#endif
}
// END of pythag(a,b)
#undef USE_BUILTIN_HYPOT

// ==== END OF METHODS Safety.cc ====
