#ifndef SAFETY_CLASS
#define SAFETY_CLASS

// ==== Safety.hh ====

/**
 * \file Safety.hh
 * Implements a few safety precautions
 * which can be switched on and off: safe division that
 * traps div-by-zero errors, provides a "small number"
 * with which it is unsafe to divide a double, and an improved
 * implementation of hypot() for those architectures which
 * do not provide it.
 * Note that this class contains architecture-dependent methods
 * and is therefore not fully portable.
 */

// 14-Nov-2001. Andras Aszodi

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

// ==== MODULE HEADERS ====

#include "Fparch.hh"

// ==== CLASSES ====

namespace RazorBack {

class Safety_
{
    // data
    private:
	    
    static Fparch_<double> Fp;	///< double-precision floating-point consts
    static const double SMALL;	///< smallest non-0 for which 1.0/SMALL is OK
    bool Usesafediv;	    	///< use safe division check if set to true
    
    // methods
    public:
    
    /**
    * ctor: set Safediv to true if you want to use the safe-division
    * feature (default).
    */
    explicit Safety_(bool Safediv=true): Usesafediv(Safediv) {}
	    
    /// return the smallest number which can still divide 1.0 w/o problems 
    static double small() { return SMALL; }
    
    /**
    * safe_div(): returns the Usesafediv variable.
    * safe_div(S): sets Usesafediv, returns old value.
    */
    bool safe_div() const { return Usesafediv; }
    bool safe_div(bool S) { bool Olds=Usesafediv; Usesafediv=S; return Olds; }
    
    /**
    * safe_div(): returns Num/Denom if Denom is reliably non-0.
    * If it is closer to 0 then the internal parameter SMALL, then
    * it will be replaced by SMALL and the division is performed
    * with a warning. The line number is printed when greater than 0 (default).
    */
    double safe_div(double Num, double Denom, int Lineno=0) const;

    /**
    * pythag(): returns the value sqrt(a^2+b^2) without over-
    * or underflows. On some machines the math library contains
    * the function hypot(x, y) which does the same job and will be used
    * whenever possible. 
    */
    double pythag(double a, double b) const;
};
// END OF CLASS Safety_

} // RazorBack

// ==== END OF HEADER Safety.hh ====

#endif	// SAFETY_CLASS
