// ==== FUNCTIONS Spl.cc ====
#include"Spl.hh"

// Third-order splines.

// 13-Dec-2000, Andras Aszodi

// ---- HEADER ----

using namespace RazorBack;

// ==== Spl_ METHODS ====

/**
 * Sets the data vector length to L>0.
 * Does nothing if L==0 or if L==len().
 * Returns the old length.
 */
unsigned int Spl_::len(unsigned int L)
{
    if (!L || L==len()) return L;   // no change
    
    clone();	// there will be a change, non-const access
    data().Ev=false;
    unsigned int Oldlen=len();
    data().len(L);  // resize data object
    return Oldlen;
}
// END of len(L)

/**
 * add(Idx, x, y):
 * adds a pair of independent (x) and dependent (y) variable points
 * to the Idx-th [0..len()-1] position of the calling object.
 * Indexrangexc_ will be thrown if Idx is illegal.
 * Use this method to "fill" up the spline object before fitting.
 */
void Spl_::add(unsigned int Idx, double x, double y)
{
    if (Idx>=len())
	throw Indexrangexc_("Spl_::add()", Idx, len()-1);
    clone(); data().Ev=false;
    data().Xvec[Idx]=x;
    data().Yvec[Idx]=y;
}
// END of add(Idx, x, y)

/**
* fit_spl(yp1, Nat1, ypn, Natn):
* fits a cubic spline to a series of data points already
* stored in the calling object. 
* Does not do anything if the fit has already been done.
* Throws Emptyexc_ if the calling object was empty.
* Throws Generalexc_ if there were duplicate data points.
* yp1: user-supplied first derivative at the 1st point.
* If Nat1 is true, then the value of yp1 is ignored and instead a zero 2nd
* derivative is assumed (natural spline).
* ypn, Natn: the same for the last point.
*/
void Spl_::fit_spl(double yp1, bool Nat1, double ypn, bool Natn)
{
    if (data().Ev) return;	// fitting already done
    if (len()<2)
	throw Emptyexc_("Spl_::fit_spl()");
    
    clone();	// now we may do the fit!
    // but check first for duplicates in the data
    // and do the sorting
    if (!check_data())
    	throw Utilsexc_("Spl_::fit_spl()", false, "Duplicate abscissa points");
    
    // Shorthands for the vector iterators
    // so that we can keep the original C array syntax :-)
    vector<double>::iterator 
	    X=data().Xvec.begin(),
	    Y=data().Yvec.begin(),
	    Y2=data().Y2vec.begin(),
	    Yin=data().Yintvec.begin();
    unsigned int i, k, n=len();
    double p, Qn, Sig, Un, dx;
    double *u;

    // init temp array (conventional)
    u=new double [n];
    
    if (Nat1) Y2[0] = u[0] = 0.0;
    else
    {
	Y2[0] = -0.5;
	u[0] = 3.0 / (X[1] - X[0]) * ((Y[1] - Y[0]) / (X[1] - X[0]) - yp1);
    }
    
    for (i = 2; i < n; i++)
    {
	Sig = (X[i - 1] - X[i - 2]) / (X[i] - X[i - 2]);
	p = Sig * Y2[i - 2] + 2.0;
	Y2[i - 1] = (Sig - 1.0) / p;
	u[i - 1] = (Y[i] - Y[i - 1]) / (X[i] - X[i - 1]) -
		   (Y[i - 1] - Y[i - 2]) / (X[i - 1] - X[i - 2]);
	u[i - 1] = (6.0 * u[i - 1] / (X[i] - X[i - 2]) - Sig * u[i - 2]) / p;
    }
    
    if (Natn) Qn = Un = 0.0;
    else
    {
	Qn = 0.5;
	Un = 3.0 / (X[n - 1] - X[n - 2]) *
	     (ypn - (Y[n - 1] - Y[n - 2]) / (X[n - 1] - X[n - 2]));
    }
    
    Y2[n - 1] = (Un - Qn * u[n - 2]) / (Qn * Y2[n - 2] + 1.0);
    for (k = n - 1; k >= 1; k--)
    Y2[k - 1] = Y2[k - 1] * Y2[k] + u[k - 1];
    
    // calculating the integral
    Yin[0] = 0.0;
    for (k = 2; k <= n; k++)
    {
	dx = X[k - 1] - X[k - 2];
	Yin[k - 1] = dx * (Y[k - 1] + Y[k - 2]) / 2 -
		     dx * dx * dx * (Y2[k - 1] + Y2[k - 2]) / 24 + Yin[k - 2];
    }
    
    delete [] u ; data().Ev=true;
}
// END of fit_spl(yp1, Nat1, ypn, Natn)

/**
 * eval_spl(xi, Der1, Der2, Der3, Integ): 
 * evaluates a spline at a point xi. If xi is out of
 * range then 0.0 is returned. Otherwise the return value is y(xi). 
 * Adapted from Numerical Recipes.
 * The derivatives and the integral are returned only if
 * they are not NULL: this way,  the user can calc any combination
 * of the first three derivatives and the integral.
 *   xi: interpolation point
 *   Der1..3: 1st..3rd derivative of y at xi
 *   Integ: definite integral of y(x) between X[0] and xi.
 * The derivative and integral ptrs are NULL by default.
 * Prints a warning and returns 0.0 if the values were modified.
 * Throws Emptyexc_ if the calling object was empty.
 * Throws Argexc_ if xi was out of range.
 */
double Spl_::eval_spl(double xi, double *Der1, double *Der2,
		double *Der3, double *Integ) const 
{
    if (!len())
    	throw Emptyexc_("Spl_::eval_spl()");
    
    if (!data().Ev)	// re-fitting necessary
	throw Utilsexc_("Spl_::eval_spl()", false, "Data modified, please call fit_spl()");
    
    // Shorthands for the vector iterators
    // so that we can keep the original C array syntax :-)
    vector<double>::const_iterator 
	    X=data().Xvec.begin(),
	    Y=data().Yvec.begin(),
	    Y2=data().Y2vec.begin(),
	    Yin=data().Yintvec.begin();
    unsigned int kl, kh, n=len();
    double a, b, h, yh, yl, y2h, y2l, dx, yi;

    if (xi > X[n - 1] || xi < X[0])
    	throw Argexc_("Spl_::eval_spl()", xi, X[0], X[n - 1]);

    // find the smallest interval containing xi
    h=smallest_interval(xi, kl, kh);
    
    a = (X[kh] - xi) / h; b = 1.0 - a;    // useful temps
    yh = Y[kh]; yl = Y[kl];
    y2h = Y2[kh]; y2l = Y2[kl];
    
    // value 
    yi = a * yl + b * yh +
	((a * a * a - a) * y2l + (b * b * b - b) * y2h) * h * h / 6.0;
	
    // derivatives
    if (Der1!=NULL)
	*Der1 = (yh - yl) / h + h * ((1.0 - 3.0 * a * a) * y2l + (3.0 * b * b - 1) * y2h) / 6.0;
    if (Der2!=NULL)
	*Der2 = a * y2l + b * y2h;
    if (Der3!=NULL)
	*Der3 = (y2h - y2l) / h;
    
    // integral 
    if (Integ!=NULL)
    {
	dx = xi - X[kl];
	*Integ = dx * (yi + yl) / 2.0 - dx * dx * dx * 
	    (a * y2l + b * y2h + y2l) / 24.0 + Yin[kl];
    }
    
    return yi;
}
// END of eval_spl(xi, Der1, Der2, Der3, Integ)

/**
 * integ_spl(Low, Up): returns the definite integral of the spline Spl
 * between Low and Up.
 * Prints a warning and returns 0.0 if spl_fit() was not invoked
 * before calling this method.
 * Throws Argexc_ if Low or Up are out of the fitted interval.
 */
double Spl_::integ_spl(double Low, double Up) const
{
    if (!data().Ev)	// re-fitting necessary
    {
	cerr<<"\n? Spl_::integ_spl(): Data modified, please call fit_spl()\n";
	return 0.0;
    }
    
    unsigned int n=len();
    vector<double>::const_iterator X=data().Xvec.begin();   // C shorthand
    
    if (Low < X[0] || Low > X[n-1])
	throw Argexc_("Spl_::integ_spl(Low,...)", Low, X[0], X[n-1]);
    if (Up < X[0] || Up > X[n-1])
	throw Argexc_("Spl_::integ_spl(...,Up)", Up, X[0], X[n-1]);
    
    return (Low <= Up)? int_0x(Up) - int_0x(Low): int_0x(Low) - int_0x(Up);
}
// END of integ_spl(Low, Up)

// int_0x(): returns the definite integral of the spline between X[0] and xi.
// xi is guaranteed to be within the legal range.
// Private
double Spl_::int_0x(double xi) const
{
    unsigned int kl, kh, n=len();
    double a, b, h, yi, yh, yl, Der2, y2h, y2l, dx;
    // Shorthands for the vector iterators
    // so that we can keep the original C array syntax :-)
    vector<double>::const_iterator 
	    X=data().Xvec.begin(),
	    Y=data().Yvec.begin(),
	    Y2=data().Y2vec.begin(),
	    Yin=data().Yintvec.begin();
    
    // find the smallest interval containing xi
    h=smallest_interval(xi, kl, kh);
    if (h == 0.0) return(0.0);

    a = (X[kh] - xi) / h; b = 1.0 - a;
    yh = Y[kh]; yl = Y[kl];
    y2h = Y2[kh]; y2l = Y2[kl];
    
    yi = a * yl + b * yh +
       ((a * a * a - a) * y2l + (b * b * b - b) * y2h) * h * h / 6.0;
    // value
    Der2 = a * y2l + b * y2h;
    dx = xi - X[kl];
    return (dx * (yi + yl) / 2.0 - dx * dx * dx * (Der2 + y2l) / 24.0 + Yin[kl]);
}
// END of int_0x()

// check_data():
// sorts the x-y data point pairs into ascending x order
// and checks if there are any duplicate points (that have 
// the same abscissa).
// Returns false if duplicate(s) have been found,
// true if OK.
// Private
bool Spl_::check_data()
{
    vector<double>::iterator X=data().Xvec.begin(),
	    	    	    Y=data().Yvec.begin();
    vector<Datapoint_> Dps(len());
    register unsigned i;
    
    // fill up the point pair array
    for (i=0; i<len(); ++i)
    {
	Dps[i].x=X[i]; Dps[i].y=Y[i];
    }
    
    // sort in ascending x order
    sort(Dps.begin(), Dps.end());
    
    // are there any duplicates?
    for (i=1; i<len(); ++i)
    {
	if (Dps[i].x==Dps[i-1].x)
	    return false;   // dupl found, failed
    }
    
    // was OK, put back sorted points
    for (i=0; i<len(); ++i)
    {
	X[i]=Dps[i].x; Y[i]=Dps[i].y;
    }
    return true;
}
// END of check_data()

// smallest_interval(xi, kl, kh):
// Identifies the smallest interval X[kl]..X[kh] that encloses xi.
// Returns the length of that interval.
// Private
double Spl_::smallest_interval(double xi,
    unsigned int& kl, unsigned int& kh) const
{
    vector<double>::const_iterator X=data().Xvec.begin();
    register unsigned int k;
    double h;
    
    kl = 0;
    kh = len()-1;
    while (kh - kl > 1)
    {
	k = (kh + kl) / 2;
	if (X[k] > xi) kh = k;
	else kl = k;
    }
    h = X[kh] - X[kl];   // length of the interval
    return h;
}

// ==== END OF METHODS Spl.cc ====
