#ifndef SPL_CLASS_HEADER
#define SPL_CLASS_HEADER

// ====	HEADER Spl.hh ====

/**
 * \file Spl.hh
 * Third-order splines.
 */

// 12-Dec-2000. Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// This module depends on the Standard Template Library.
 
// ---- STANDARD HEADERS ----

#include <stdlib.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#include <vector>
#include <algorithm>
    using namespace std;
#else
#include <vector.h>	// STL vectors
#include <algo.h>	// STL algorithms
#endif

// ---- MODULE HEADERS ----

#include "Refstorage.hh"
#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

class Spl_;  	// fwd decl

/**
 * Splstorage_: this is a helper class that stores the data
 * needed by the Spl_ class. Objects of this class are wrapped
 * by the Refstorage_ class that implements reference counting
 * and copy-on-write. The Spl_ class inherits that Refstorage_ wrapper.
 * This class has only private members and methods and Spl_ is its 
 * only friend that is allowed to tamper with them.
 */
class Splstorage_
{
    friend class Spl_;
    
    // data
    private:
    
    vector<double> Xvec;	    	///< independent variable
    vector<double> Yvec;	    	///< dependent variable
    vector<double> Y2vec;	    	///< second derivatives
    vector<double> Yintvec;	    	///< integrals
    unsigned int Len;	    	    	///< common data vector length
    bool Ev;	    	    	    	///< Evaluation needed?
    
    // methods are also private, only the friendly Spl_ can access them!
	    
    /// Inits to work on N data points (default 2)
    explicit Splstorage_(unsigned int N=2):
	    Xvec(N), Yvec(N), Y2vec(N), Yintvec(N), Len(N), Ev(false) {}
    
    /**
     * Changes the vector sizes to L.
     * Guaranteed to be invoked only if L is different from Len.
     * See Spl_::len(L).
     */
    void len(unsigned int L)
    {
	if (L>Len)
	{
	    // storage grows, pad with zeroes
	    int Increment=L-Len;
	    Xvec.insert(Xvec.end(), Increment, 0.0);
	    Yvec.insert(Yvec.end(), Increment, 0.0);
	    Y2vec.insert(Y2vec.end(), Increment, 0.0);
	    Yintvec.insert(Yintvec.end(), Increment, 0.0);
	}
	else
	{
	    // remove last items
	    Xvec.erase(Xvec.begin()+Len, Xvec.end());
	    Yvec.erase(Yvec.begin()+Len, Yvec.end());
	    Y2vec.erase(Y2vec.begin()+Len, Y2vec.end());
	    Yintvec.erase(Yintvec.begin()+Len, Yintvec.end());
	}
	Len=L;
    }
    
    /// Zeroes all data vectors and prepares for a new fit.
    void reset() 
    {
	fill(Xvec.begin(), Xvec.end(), 0.0);
	fill(Yvec.begin(), Yvec.end(), 0.0);
	fill(Y2vec.begin(), Y2vec.end(), 0.0);
	fill(Yintvec.begin(), Yintvec.end(), 0.0);
	Ev=false;
    }
    
};
// END OF CLASS Splstorage_

/**
 * Class Spl_ : implements third-order splines. 
 * Can be queried for interpolated values, 1st..3rd derivatives
 * and integrals.
 */
class Spl_: protected Refstorage_<Splstorage_>
{
    
    // methods
    public:
    
    /// Inits to work on N data points (default 2).
    explicit Spl_(unsigned int N=2): 
    	Refstorage_<Splstorage_>(new Splstorage_(N)) {}
    
    /// Returns the current length of the data vectors.
    unsigned int len() const { return data().Len; }
    
    /**
     * Sets the data vector length to L.
     * Does nothing if L==0 or if L==len().
     * Returns the old length.
     */
    unsigned int len(unsigned int L);
    
    /// Zeroes all data vectors and prepares for a new fit.
    void reset() 
    {
	clone();
	data().reset();
    }
    
    /**
     * add(Idx, x, y):
     * adds a pair of independent (x) and dependent (y) variable points
     * to the Idx-th [0..len()-1] position of the calling object.
     * Indexrangexc_ will be thrown if Idx is illegal.
     * Use this method to "fill" up the spline object before fitting.
     */
    void add(unsigned int Idx, double x, double y);

    /**
    * fit_spl(yp1, Nat1, ypn, Natn):
    * fits a cubic spline to a series of data points already
    * stored in the calling object. 
    * Does not do anything if the fit has already been done.
    * Throws Emptyexc if the calling object was empty.
    * Throws Generalexc_ if there were duplicate data points.
    * yp1: user-supplied first derivative at the 1st point.
    * If Nat1 is true, then the value of yp1 is ignored and instead a zero 2nd
    * derivative is assumed (natural spline).
    * ypn, Natn: the same for the last point.
    */
    void fit_spl(double yp1, bool Nat1, double ypn, bool Natn);

    /**
     * eval_spl(xi, Der1, Der2, Der3, Integ): 
     * evaluates a spline at a point xi. If xi is out of
     * range then 0.0 is returned. Otherwise the return value is y(xi). 
     * Adapted from Numerical Recipes.
     * The derivatives and the integral are returned only if
     * they are not NULL: this way,  the user can calc any combination
     * of the first three derivatives and the integral.
     *   xi: interpolation point
     *   Der1..3: 1st..3rd derivative of y at xi
     *   Integ: definite integral of y(x) between X[0] and xi.
     * The derivative and integral ptrs are NULL by default.
     * Prints a warning and returns 0.0 if the values were modified.
     * Throws Emptyexc_ if the calling object was empty.
     * Throws Argexc_ if xi was out of range.
     */
    double eval_spl(double xi, double *Der1=NULL, double *Der2=NULL,
		    double *Der3=NULL, double *Integ=NULL) const;
    
    /**
     * integ_spl(Low, Up): returns the definite integral of the spline Spl
     * between Low and Up.
     * Prints a warning and returns 0.0 if spl_fit() was not invoked
     * before calling this method.
     * Throws Argexc_ if Low or Up are out of the fitted interval.
     */
    double integ_spl(double Low, double Up) const;
    
    private:

    // an x-y point pair that can be sorted according to x
    struct Datapoint_
    {
	double x, y;
	bool operator<(const Datapoint_& Rhs) const { return (x<Rhs.x); }
    };
    
    bool check_data();
    double int_0x(double xi) const;
    double smallest_interval(double xi,
	unsigned int& kl, unsigned int& kh) const;
};
// END OF CLASS Spl_

} // RazorBack

// ==== END OF HEADER Spl.hh ====

#endif
