// ==== METHODS Stat2.cc ====
#include"Stat2.hh"

// Very simple 1 and 2-variable statistics (avg,SD,skewness,kurtosis,corr).

// 26-Sep-2001. Andras Aszodi

// ---- STANDARD HEADERS ----

#include <math.h>
#include <limits.h>
#include <float.h>
#include <string.h>

// ---- MODULE HEADERS ----

using namespace RazorBack;

// ==== FUNCTIONS ====

// ---- Static initialisation ----

const double Stat_::BIG_NUMBER=1.7e308;
#ifdef DBL_EPSILON
    const double Stat_::SMALL_NUMBER=DBL_EPSILON;
#else
    const double Stat_::SMALL_NUMBER=1e-15;
#endif

// ---- Stat_ METHODS ----

// Stat+=Val: adds a new data point Val to Stat. Returns calling object
Stat_& Stat_::operator+=(double Val)
{
    Sx+=Val; N++;
    if (Val<Min) Min=Val;
    if (Val>Max) Max=Val;
    
    // summing higher powers for moments
    register double Valpower=Val*Val;
    Sx2+=Valpower;
    Sx3+=Valpower*Val;	// for skewness
    Sx4+=Valpower*Valpower; 	// for kurtosis
    return *this;
}
// END of +=

// min(), max(): return the minimal/maximal value or throws Toofewexc_
// if there were no values at all.
double Stat_::min() const
{
    if (!N) throw Toofewexc_("Stat_::min()", "Number of data points");
    else return(Min);
}

double Stat_::max() const
{
    if (!N) throw Toofewexc_("Stat_::max()", "Number of data points");
    else return(Max);
}
// END of min(), max()

// avg(): returns the average. Throws Toofewexc_ if
// there were no data items at all.
double Stat_::avg() const
{
    if (!N) throw Toofewexc_("Stat_::avg()", "Number of data points");
    else return(Sx/N);
}
// END of avg()

// sd(): returns the standard deviation. Throws Toofewexc_ 
// if there were no data items at all. If N=1, 0.0 is returned automatically.
double Stat_::sd() const
{
    if (!N) throw Toofewexc_("Stat_::sd()", "Number of data points");
    if (N==1) return 0.0;
    double Xavg=Sx/N;
    double Numer=fabs(Sx2-N*Xavg*Xavg); // beware of roundoff errors :-)
    return((Numer<SMALL_NUMBER)? 0.0: sqrt(Numer/N));
}
// END of sd()

/*
 * skew(): returns the estimated skewness. Throws Toofewexc_
 * if there were no data items at all. If N==1 happens to be 0.0,
 * then 0.0 is returned. Throws Sdevexc_ if the s.d. was not larger than 0.0.
 */
double Stat_::skew() const
{
    if (!N) throw Toofewexc_("Stat_::skew()", "Number of data points");
    if (N==1) return 0.0;
    
    double Xavg=avg(), Xsd=sd(), Num, Denom;
    if (Xsd<=0.0) throw Sdevexc_("Stat_::skew()", Xsd);
    
    Num=Sx3-3*Xavg*Sx2+2*N*Xavg*Xavg*Xavg;
    Denom=N*Xsd*Xsd*Xsd;
    return Num/Denom;
}
// END of skew()

/*
 * kurt(): returns the estimated kurtosis. Note that 3.0 is NOT subtracted
 * from this value (as opposed to the Numerical Recipes convention),
 * so the kurtosis of a Gaussian distribution should be about 3.0.
 * Throws Toofewexc_ if there were no data items at all.
 * If N==1 happens to be 0.0, then 0.0 is returned.
 * Throws Sdevexc_ if the s.d. was not larger than 0.0.
 */
double Stat_::kurt() const
{
    if (!N) throw Toofewexc_("Stat_::kurt()", "Number of data points");
    if (N==1) return 0.0;
    
    double Xavg=avg(), Xsd=sd(), Num, Denom;
    if (Xsd<=0.0) throw Sdevexc_("Stat_::kurt()", Xsd);
    
    Num=Sx4-4*Xavg*Sx3+6*Xavg*Xavg*Sx2-3*N*Xavg*Xavg*Xavg*Xavg;
    Denom=N*Xsd*Xsd*Xsd*Xsd;
    return Num/Denom;
}
// END of kurt()

// ---- I/O (serialization) ----

// friend fn defs must be in scope
namespace RazorBack {

// Prints a Stat_ object to Out. The format is:-
// STAT
// N = 1000
// MIN = -4.7362
// MAX = 6.30655
// SUM = 957.413
// SUM2 = 4575.58
// SUM3 = 11291.8
// SUM4 = 57072.3
// ENDSTAT
ostream& operator<<(ostream& Out, const Stat_& S)
{
    Iovar_<unsigned int> Nio("N", S.N);
    Iovar_<double> Minio("MIN", S.Min), Maxio("MAX", S.Max),
	    Sio("SUM", S.Sx), S2io("SUM2", S.Sx2),
	    S3io("SUM3", S.Sx3), S4io("SUM4", S.Sx4);

    Out<<"STAT"<<endl;
    Out<<Nio<<endl<<Minio<<endl<<Maxio<<endl;
    Out<<Sio<<endl<<S2io<<endl<<S3io<<endl<<S4io<<endl;
    Out<<"ENDSTAT"<<endl;
    return Out;
}
// END of <<

// Reads a Stat_ object from In. The format is:-
// STAT
// N = 1000
// MIN = -4.7362
// MAX = 6.30655
// SUM = 957.413
// SUM2 = 4575.58
// SUM3 = 11291.8
// SUM4 = 57072.3
// ENDSTAT
istream& operator>>(istream& In, Stat_& S)
{
#define LINELEN 120
    char Line[LINELEN];
    bool Statseen=false, Endseen=false;
    
    // read In line-by-line, look for a line that begins with "STAT"
    while (In.getline(Line, LINELEN, '\n').good())
    {
	if (!strncmp(Line, "STAT", 4))
	{
	    Statseen=true;
	    break;
	}
    }
    if (!Statseen) return In;	// not found
    
    // read the Iovar tokens, these set the failbit if not OK
    Iovar_<unsigned int> Nio("N");
    Iovar_<double> Minio("MIN"), Maxio("MAX"),
	    Sio("SUM"), S2io("SUM2"),
	    S3io("SUM3"), S4io("SUM4");
    In>>Nio>>Minio>>Maxio>>Sio>>S2io>>S3io>>S4io;
    if (!In.good()) return In;	// something went wrong
    
    // get the closing "ENDSTAT" token
    while (In.getline(Line, LINELEN, '\n').good())
    {
	if (!strncmp(Line, "ENDSTAT", 7))
	{
	    Endseen=true;
	    break;
	}
    }
    if (!Endseen) return In;	// not found
    
    // seems to be OK, set the values in S
    S.N=Nio.value();
    S.Min=Minio.value(); S.Max=Maxio.value();
    S.Sx=Sio.value(); S.Sx2=S2io.value();
    S.Sx3=S3io.value(); S.Sx4=S4io.value();
    
    return In;
#undef LINELEN
}
// END of >>

} // RazorBack

// ---- Stat2_ METHODS ----

// corr(): returns the correlation coefficient between the X and Y
// points. Throws Toofewexc_ if there were no data points.
// Throws Sdevexc_ if the s.d. was not larger than 0.0.
double Stat2_::corr() const
{
    unsigned int N;
    if ((N=data_no())<2) throw Toofewexc_("Stat2_::corr()", "Number of data points", N);

    double Sd=Xs.sd()*Ys.sd();
    if (Sd<=0.0) throw Sdevexc_("Stat2_::corr()", Sd);
    return((Sxy-N*Xs.avg()*Ys.avg())/(N*Sd));
}
// END of corr()

// ==== END OF FUNCTIONS Stat2.cc ====
