// ==== METHODS Statprob.cc ====
#include "Statprob.hh"

/**
 * \file
 * Probability integrals for statistical tests.
 * @author Andras Aszodi
 * @date 7-Dec-2001.
 */

// ==== MODULE HEADERS ====

using namespace RazorBack;

// ==== CLASSES ====

// The auxiliary class for evaluating continued fractions
// is encapsulated in this file for the time being.
// May be promoted to a separate module later if necessary.
// Implements Lentz's algorithm with modifications.
class Contfrac_
{
    // data
    private:
	    
    static const double SMALL;    // a tiny number
    
    double Nr, Dr;  	    // temporaries
    double Approx, Factor;  // approximated value and factor from old to new
    
    // methods
    public:
	    
    // Inits with the constant term d0 (0.0 by default).
    explicit Contfrac_(double d0=0.0):
	    Nr(fudge_zero(d0)), Dr(0.0), Approx(Nr), Factor(1.0) {}
    
    // Returns the current approximated value.
    double approx() const { return Approx; }
    
    // Returns the last multiplication factor
    // used to generate the current approximation.
    double factor() const { return Factor; }
    
    /*
     * Given a continued fraction of the form
     * 
     * \verbatim
     *      n1  n2	   nk
     * d0 + --- --- ...---
     *      d1+ d2+    dk
     * \endverbatim
     * 
     * it gets evaluated here to the k:th term, given nk and dk,
     * provided the evaluation up to the k-1:th term has already been done.
     */
    void eval(double nk, double dk);

    // hidden methods
    protected:
	    
    static double fudge_zero(double x)
    {
	return (x==0.0)? SMALL: x;
    }
};
// END OF CLASS Contfrac_

// ---- Static init ----

const unsigned int Statprob_::MAXITER=100;
const double Statprob_::EPSILON=1.0e-12;

const double Contfrac_::SMALL=1.0e-30;

// ==== METHODS ====

// ---- Statprob_ methods ----

// Returns the tail of the standard normal distribution at z.
double Statprob_::stdnorm_tail(double z)
{
    static const double SQRT2=1.414213562373;
    register double Cdf;
    
    Cdf=0.5*((z<0.0)? erfc(-z/SQRT2): (1.0+erf(z/SQRT2)));
    return (1.0-Cdf);
}
// END of stdnorm_tail(z)

/**
 * Returns the significance level for the t test.
 * If the value returned by t_test() is low, then the probability is large
 * that t is high.
 */
double Statprob_::t_test(unsigned int df, double t)
{
    register double x=df/(df+t*t);
    return ibeta_regul(0.5*df, 0.5, x);
}
// END of t_test(df, t)

/**
 * Returns the significance level for the test that the
 * F ratio is larger than 1.0.
 * If the value returned by F_test() is low, then the probability is large
 * that F is high.
 */
double Statprob_::F_test(unsigned int df1, unsigned int df2, double F)
{
    register double x=df2/(df2+df1*F);
    return ibeta_regul(0.5*df2, 0.5*df1, x);
}
// END of F_test(df1, df2, F)

/**
 * Returns the significance level for the test that the Chi2 is exceeded
 * for a correct model. The less the probability, the better. :-)
 */
double Statprob_::chi2_test(unsigned int df, double Chi2)
{
    return 1.0-igamma_regul(0.5*df, 0.5*Chi2);
}
// END of chi2_test(df, Chi2)

/**
 * Returns the Q probability value for the Studentized Range.
 * df is the degrees of freedom, kno is the number of groups.
 */
double Statprob_::studrange_prob(double q, unsigned int df, unsigned int kno)
{
    // The implementation is based on Algorithm AS190:
    // Appl. Statist. (1983) Vol.32, No. 2             
    // with corrections from Appl. Statist. (1985) Vol.34 (1).
    // The code here was taken from the "mwanova" package by Antonio Santos.
    // After-edited to get rid of the GOTOs, made constants const,
    // uses stdnorm_tail().
    // Accuracy was increased by modifying a few parameters
    // and the size of the vw[] and qw[] arrays (see comments below).
    // The minimum probability returned is now about 1e-07.
    static const double CV1=0.193064705, CV2=0.293525326, CVMAX=0.39894228,
	    PCUTJ=3e-12,    // was 0.00003
	    PCUTK=1e-12,    // was 0.0001
	    STEP=0.225;     // was 0.45
    static const double CV[4]={0.318309886, -0.268132716e-2, 0.347222222e-2, 0.833333333e-1};
    static const unsigned int DOFMAX=120, 
	    JMIN=6, 	// was 3,
	    JMAX=30,	// was 15,
	    KMIN=14, 	// was 7,
	    KMAX=30;	// was 15;
    
    register int  k, j, jj, jump;
    double Retval=0.0;
    double vw[62], qw[62];  // previous size was 31
    double g, gmid, r1, c, h, v2, gstep, pk1, pk2, gk, pk;
    double w0, pz, x, hj, ehj, pj;

    // check initial values
    if (df<1)
	throw Badvalexc_("Statprob_::studrange_prob(): df<1", df);
    if (kno<2)
	throw Badvalexc_("Statprob_::studrange_prob(): kno<2", kno);
    if (q<=0.0)
	throw Badvalexc_("Statprob_::studrange_prob(): q<=0.0", q);

    // computing constants, local midpoint, adjusting steps
    g=STEP * pow(kno, -0.2);
    gmid=0.5 * log(double(kno));    // G++ 3.1 needs explicit conversion
    r1=kno-1.0;
    c=log(kno*g*CVMAX);
    if (df <= DOFMAX)
    { 
	h=STEP * pow(df, -0.5);
	v2=df * 0.5;
	if (df == 1.0) c=CV1;
	if (df == 2.0) c=CV2;
	if (!((df == 1.0) || (df == 2.0)) )
	    c=sqrt(v2) * CV[0] /(1.0 + ((CV[1]/v2 + CV[2])/ v2 + CV[3]) / v2);
	c=log(c*kno*g*h);  
    }

    // computing integral
    gstep=g;
    qw[1]=qw[JMAX+1]=(-1.0);
    pk1=pk2=1.0;

    for (k=1; k<=KMAX; k++)
    {
	gstep-=g;
	do
	{
	    gstep*=(-1.0);
	    gk=gmid + gstep;
	    pk=0.0;
	    if (!((pk2 <= PCUTK) && (k > KMIN)))
	    {
		w0=c-gk*gk*0.5;
		pz=stdnorm_tail(gk);
		x=stdnorm_tail(gk-q) - pz;
		if (x > 0.0) pk=exp(w0+r1*log(x));
		if (df <= DOFMAX)
		{
		    jump=(-1)*static_cast<int>(JMAX);	// for Compaq compiler...
		    do
		    {
			jump+=JMAX;
			for (j=1; j<=JMAX; j++)
			{
			    jj=j+jump;
			    if (qw[jj] <= 0.0)
			    {
				hj=h*j;
				if (j < JMAX) qw[jj+1]=(-1.0);
				ehj=exp(hj);
				qw[jj]=q*ehj;
				vw[jj]=df*(hj+0.5-ehj*ehj*0.5);
			    }
			    pj=0.0;
			    x=stdnorm_tail(gk-qw[jj]) - pz;
			    if (x > 0.0) pj=exp(w0+vw[jj]+r1*log(x));
			    pk+=pj;
			    if (pj > PCUTJ) continue;
			    if ((jj > JMIN) || (k> KMIN)) break;
			}

    			h*=(-1.0);
    		    }
		    while (h < 0.0);
		}
	    }

    	    Retval += pk;
	    if ((k > KMIN) && (pk <= PCUTK) && (pk1 <= PCUTK))
	    	return (1.0-Retval);
	    pk2=pk1;
	    pk1=pk;
	}
	while (gstep > 0.0);
    }

    return (1.0-Retval);
}
// END of q_prob(q, df, kno)

// ---- Special functions ----

/// The regularized incomplete beta function.
double Statprob_::ibeta_regul(double a, double b, double x)
{
    double Prefactor, Ibeta;
    
    if (x<0.0 || x>1.0) // bad value
	throw Argexc_("Statprob_::ibeta_regul()", x, 0.0, 1.0);
    if (x==0.0 || x==1.0) return x;
    
    // do the prefactor with the builtin ln(gamma)
    Prefactor=exp(lgamma(a+b)-lgamma(a)-lgamma(b)+a*log(x)+b*log(1.0-x));
    Ibeta=(x<((a+1.0)/(a+b+2.0)))? 
	    Prefactor*beta_contfrac(a, b, x)/a:
	    1.0-Prefactor*beta_contfrac(b, a, 1.0-x)/b;
    return Ibeta;
}
// END of ibeta_regul(a,b,x)

/**
 * The regularized incomplete gamma function
 * (1/Gamma(a))*INTEGRAL[exp(-t)*t^(a-1) dt, 0, x].
 */
double Statprob_::igamma_regul(double a, double x)
{
    if (x < 0.0)
	throw Badvalexc_("Statprob_::igamma_regul(): x<0.0", x);
    if (a <= 0.0)
	throw Badvalexc_("Statprob_::igamma_regul(): a<=0.0", a);
    if (x==0.0) return 0.0;

    double G=(x < (a+1.0))? gamma_series(a, x): 1.0-gamma_contfrac(a, x);
    return G;
}
// END of igamma_regul(a, x)

// ---- aux ----

// Series approximation of the incomplete gamma function.
// Hidden static
double Statprob_::gamma_series(double a, double x)
{
    register unsigned int n;
    register double Sum, Term, An, Igamma;

    Sum=1.0; Term=1.0;
    for (n=1, An=a+1.0; n<MAXITER; ++n, An+=1.0)
    {
	Term*=(x/An);
	if (fabs(Term)<EPSILON*fabs(Sum))
	    break;  	// converged
	Sum+=Term;
    }
    Igamma=exp(-x+a*log(x)-lgamma(a+1.0))*Sum;
    
    // may not have converged
    return Igamma;
}
// END of gamma_series(a,x)

// Continued fraction approximation of the incomplete gamma function.
// Hidden static
double Statprob_::gamma_contfrac(double a, double x)
{
    // setting up the continued fraction, d0==0.0 by default */
    Contfrac_ Contfrac;
    
    register unsigned int i;
    register double Frac, Ni, Di, Igamma;

    // do the first step, convergence not tested
    Contfrac.eval(1.0, x+1.0-a);

    for (i=1; i<MAXITER; ++i)
    {
	// do the recurrence
	Ni=i*(a-i);
	Di=x-a+2.0*i+1.0;
	Contfrac.eval(Ni, Di);
	Frac=Contfrac.approx();
	if (fabs(Contfrac.factor()-1.0)<EPSILON)
	    break;  	// converged 
    }
    Igamma=exp(-x+a*log(x)-lgamma(a))*Frac;
        
    // may not have converged
    return Igamma;
}
// END of gamma_contfrac(a,x)

// beta_contfrac: continued fraction approximation of the beta function.
// Hidden static
double Statprob_::beta_contfrac(double a, double b, double x)
{
    // setting up the continued fraction, d0==0.0 by default
    Contfrac_ Contfrac;
    
    register unsigned int i;
    register double Ni, A2, A2m1, A2p1, Frac;
    
    A2m1=a-1.0; A2=a; A2p1=a+1.0;
    for (i=0; i<MAXITER; ++i)
    {
	// even coeff
	Ni=!i? 1.0: i*(b-i)*x/(A2m1*A2);
	Contfrac.eval(Ni, 1.0);
	Frac=Contfrac.approx();
	if (fabs(Contfrac.factor()-1.0)<EPSILON)
	    break;  	// converged
	
	// odd coeff
	Ni=-(a+i)*(a+b+i)*x/(A2*A2p1);
	Contfrac.eval(Ni, 1.0);
	Frac=Contfrac.approx();
	if (fabs(Contfrac.factor()-1.0)<EPSILON)
	    break;  	// converged
	
	// stepping
	A2m1=A2p1;
	A2+=2.0;
	A2p1=A2+1.0;
    }
    
    // may not have converged
    return Frac;
}
// END of beta_contfrac(a,b,x)

// ---- Continued fraction evaluation methods ----

/*
 * Given a continued fraction of the form
 * 
 * \verbatim
 *      n1  n2	   nk
 * d0 + --- --- ...---
 *      d1+ d2+    dk
 * \endverbatim
 * 
 * it gets evaluated here to the k:th term, given nk and dk,
 * provided the evaluation up to the k-1:th term has already been done.
 */
void Contfrac_::eval(double nk, double dk)
{
    register double Newd;
    
    Nr=fudge_zero(dk+nk/Nr);
    
    Newd=fudge_zero(dk+nk*Dr);
    Dr=1.0/Newd;
    
    Factor=Nr*Dr;
    Approx*=Factor;
}
// END of eval(nk, dk)

// ==== END OF METHODS Statprob.cc ====
