#ifndef SVD_CLASS
#define SVD_CLASS

// ==== HEADER Svd.hh ====

/**
 * \file Svd.hh
 * Singular value decomposition based on the algorithm
 * in Golub & Van Loan.
 */

// 14-Nov-2001. Andras Aszodi

// ----	STANDARD HEADERS ----

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <math.h>

#ifdef HAVE_STD
#  include <iostream>
using namespace std;
#else
#  include <iostream.h>
#endif

// ---- INCLUDE FILES ----

#include "Basemat.hh"
#include "Rectmat.hh"
#include "Vector.hh"
#include "Safety.hh"
#include "Rangedimexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Class Svd_ : a simple class for Singular Value Decomposition.
 * SVD decomposes a rectangular matrix A into the form UWV'
 * where W is diagonal and V is square. U and V are Rmatdbl_ objects,
 * W is stored as a Vector_ object. 
 * The class supports a few methods for performing SVD and analysing
 * the results. A general linear equation solver is also provided.
 */
class Svd_
{
    // data
    private:
    
    static const Safety_ Safe;	///< safe division and hypot() classlet
    static const double SVD_EPSILON;	///< a small value
    
    Rmatdbl_ U;	///< general Row x Col matrix
    Vecdbl_ W;	///< Col-long vector
    Rmatdbl_ V;	///< Col x Col square matrix
    unsigned int R, Rorig, C;	///< row and column nos.
    
    // methods
    public:
    
    /**
    * Set up SVD for a Row x Col matrix where Row>=Col. If Row<Col,
    * then the rows of U will be padded to make it ColxCol
    * with a warning. If either Row or Col is 0, will be changed
    * to 3 (cf matrix/vector ctors) and a warning printed. 
    * The default size is also 3x3.
    * The members will be initialised to 0.0 (all elements).
    */
    explicit Svd_(unsigned int Row=3, unsigned int Col=3);
    
    /// Returns the U matrix.
    const Rmatdbl_& u() const { return(U); }
    
    /// Returns the vector of the singular values.
    const Vecdbl_& w() const { return(W); }
    
    /// Returns the V matrix.
    const Rmatdbl_& v() const { return(V); }
    
    /// rno() returns the current number of rows.
    unsigned int rno() const { return(R); }
    
    /// cno() returns the current number of columns.
    unsigned int cno() const { return(C); }
    
    /**
    * set_size(): modifies the sizes of the SVD components to
    * accommodate a Row x Col matrix. If Row < Col, then the
    * rows of U will be padded to give a Col x Col matrix.
    * If Row or Col is 0, then 3 will be used instead and a warning printed.
    * Without arguments, the size is set to 3x3.
    */
    void set_size(unsigned int Row=3, unsigned int Col=3);
    
    /**
    * make_decomp(A): the SVD done
    * on the general matrix A (which is preserved). The
    * three member objects U, W, V will be set in the calling
    * object. Throws Svdexc_ if there was a convergence problem
    * or a div-by-zero attempt in the core routine.
    */
    void make_decomp(const Rmatdbl_& A);
	
    /**
    * rank_cond(): checks the singular values after SVD.
    * The condition number in *Cond
    * (ratio of the largest and smallest singular value) is also
    * calculated if Cond!=NULL. The singular values which are smaller than
    * Eps times the largest are set to 0.0.
    * Return value: the rank of the matrix.
    */
    unsigned int rank_cond(double Eps=SVD_EPSILON, double *Cond=NULL);
    
    /**
    * lin_solve(): back-substitution routine for solving linear equations
    * AX=B. A should be SV-decomposed into U, W and V' by make_decomp()
    * and the weight vector should be "conditioned" (small entries
    * zeroed) by rank_cond() prior to the call to this routine.
    * The right-hand-side vector B must have a dim equal to the original
    * no. of rows in A otherwise a Dimexc_ exception will be thrown.
    * Return value: the X vector.
    */
    Vecdbl_ lin_solve(const Vecdbl_& B) const;
    
    /// reset_data(): zeroes the matrices and the W vector.
    void reset_data() { U.set_values(); W.set_values(); V.set_values(); }
    
    /// <<: the overloaded output operator. Prints a neat listing to Out
    friend ostream& operator<<(ostream& Out, const Svd_& Svd);
    
    private:
	    
    void svd_core();
    void utb(const Vecdbl_& B, Vecdbl_& Utb) const;
};
// END OF CLASS Svd_

/**
 * An Svdexc_ exception can be thrown if something went wrong in Svd_::make_decomp().
 * Stores an error description string only.
 */
class Svdexc_: public Utilsexc_
{
    public:
    explicit Svdexc_(const char *Thnm, const char *Msg, bool F=false):
	    Utilsexc_(Thnm, F, Msg) {}
    private:
    Svdexc_();	// no default ctor
};
// END OF CLASS Svdexc_

} // RazorBack

// ==== END OF HEADER Svd.hh ====
#endif	// SVD_CLASS
