#ifndef SYMMAT_TMPL_HEADER
#define SYMMAT_TMPL_HEADER

// ==== HEADER Symmat.hh ====

/**
 * \file symmat.h
 * Symmetric (lower triangular) matrix templates.
 */

// 8-Dec-2000. Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// Exception support is assumed.
// This module depends on the Standard Template Library.
 
// ---- COMPILER COMMAND LINE MACROS ----

// Define the macro INCLUDE_TMPL_DEFS if your compiler
// does not support automatic template source inclusion.

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

// ---- MODULE HEADERS ----

#include "Vector.hh"
#include "Basemat.hh"
#include "Rectmat.hh"
#include "Squarexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Class Symmat_: a class of symmetric square matrices (Row x Row).
 * Derived from Basemat_.
 */
template <class T_>
class Symmat_: public Basemat_<T_>
{
    // data
    private:
    
    /*
     * Layout information:-
     * The matrix elements are stored in "triangular" row major order.
     * 
     * 00
     * 10 11    --> 00 10 11 20 21 22
     * 20 21 22
     * 
     */
    size_t Size;	//< no. of rows and columns
    
    // methods
    public:
    
    /// default ctor: 3x3 is the default size
    explicit Symmat_(size_t S=3)
    {
	if (!S) S=3;
	Size=S;
	Matstorage_<T_> *Newmat=new Matstorage_<T_>(tri_size(S));
	replace_storage(Newmat);	// alloc storage
    }
    
    /// copy ctor
    Symmat_(const Symmat_<T_>& Sm):
	    Basemat_<T_>(Sm), Size(Sm.rno()) {}
    
    /**
     * init Symmat_ with a conventional array. 
     * It is supposed to have the "triangular" row-major layout. No checks
    */
    Symmat_(const T_ *Arr, size_t S):
	    Size(S)
    {
	Matstorage_<T_> *Newmat=new Matstorage_<T_>(tri_size(S));
	for (register unsigned int i=0; i<tri_size(S); i++)
	    Newmat->elems()[i]=Arr[i];	// direct copy assuming same layout
	replace_storage(Newmat);
    }
    
    /// destructor
    virtual ~Symmat_() {}
    
    	/// Conversions
    
    /// Symmat->Rectmat conversion
    operator Rectmat_<T_>() const;
    
    /**
    * triangle(Rect, Low): copies the lower triangle of Rect
    * to the calling object if Low is true (default) or the upper triangle
    * if Low is false. Rectmat must be square, otherwise Squarexc_ is thrown.
    * The calling object is resized silently.
    * Returns reference to calling object.
    */
    Symmat_<T_>& triangle(const Rectmat_<T_>& Rect, bool Low=true);
    
    	/// Access
    
    /// rno(), cno(): return the number of rows and columns.
    virtual size_t rno() const { return Size; }
    virtual size_t cno() const { return Size; }
    
    /**
    * The C-style [] operator provides unchecked access.
    * It is assumed (but NOT TESTED) that when writing Sym[i][j],
    * the condition i>=j is true. You have been warned.
    */
    virtual typename vector<T_>::const_iterator operator[](size_t Idx) const
    	{ return (elems().begin()+tri_size(Idx)); }
    virtual typename vector<T_>::iterator operator[](size_t Idx)
    	{ clone(); return (elems().begin()+tri_size(Idx)); }
    
    /**
    * at(Ridx, Cidx): does unchecked C indexing but Ridx and Cidx
    * are swapped if Ridx<Cidx.
    */
    const T_& at(size_t Ridx, size_t Cidx) const
    {
	if (Ridx<Cidx) swap(Ridx, Cidx);
	return(elems(tri_size(Ridx)+Cidx));
    }
    T_& at(size_t Ridx, size_t Cidx)
    {
	clone();
	if (Ridx<Cidx) swap(Ridx, Cidx);
	return(elems(tri_size(Ridx)+Cidx));
    }

    /**
    * The two-argument function call operator provides optional checked access
    * when Check is true, and FORTRAN-style indexing when Ftnidx is true.
    * Indices are always swapped silently so that Sym(i,j) is Sym[j][i]
    * when i<j. Throws Indexrangexc_ on error.
    */
    virtual const T_& operator()(size_t Ridx, size_t Cidx) const;
    virtual T_& operator()(size_t Ridx, size_t Cidx);
    
    /**
    * size(S): changes the size of the calling object to S x S.
    * Nothing is done if the size parameter is the same as the current
    * size or if it is 0. When the matrix "shrinks", then
    * the upper left corner is preserved, when it grows, then the new
    * rows and columns are padded with "zeroes" T_(0).
    */
    void size(size_t S);

    /**
    * row(Idx): returns the Idx-th row as a Vector_ object safely.
    * row(Vec,Idx): sets the elements in the Idx-th row to the elements
    * of the vector Vec, provided Idx is legal and Vec has the right
    * number of elements.
    * col(Idx) and col(Vec,Idx) do exactly the same (symmetry!).
    */
    virtual Vector_<T_> row(size_t Idx) const;
    virtual void row(const Vector_<T_>& Vec, size_t Idx);
    virtual Vector_<T_> col(size_t Idx) const { return row(Idx); }
    virtual void col(const Vector_<T_>& Vec, size_t Idx) { row(Vec, Idx); }

    	/// Simple scalar operations
    
    Symmat_<T_> operator*(const T_& Factor) const;
    Symmat_<T_> operator/(const T_& Div) const;
    
    	/// Linear algebra
    
    /**
    * Matrix addition and subtraction.
    * Dimexc_ is thrown on dimension mismatches.
    */
    Symmat_<T_>& operator+=(const Symmat_<T_>& Mat);
    Symmat_<T_> operator+(const Symmat_<T_>& Mat) const;
    Rectmat_<T_> operator+(const Rectmat_<T_>& Mat) const;
    Symmat_<T_>& operator-=(const Symmat_<T_>& Mat);
    Symmat_<T_> operator-(const Symmat_<T_>& Mat) const;
    /// (Symmat-Rectmat is done via Sym->Rect conversion and Rect-Rect)

    /**
    * Matrix*matrix and matrix*vector multiplication.
    * Mixed Symmat/Rectmat operators are provided to avoid conversions.
    * Dimexc_ is thrown on dimension mismatches.
    */
    Rectmat_<T_> operator*(const Symmat_<T_>& Mat) const;
    Rectmat_<T_> operator*(const Rectmat_<T_>& Mat) const;
    Vector_<T_> operator*(const Vector_<T_>& Vec) const;

    	/// Diagonals
    
    /**
    * diag(): copies the diagonal into a Vector_ object.
    * diag(Vec): sets the diagonal to the values in the Vector_ object Vec.
    * If the dimensions don't match, a Dimexc_ exception is thrown.
    */
    virtual Vector_<T_> diag() const;
    virtual void diag(const Vector_<T_>& Vec);

    /**
    * diag_matrix(Dval): turns the calling object into a diagonal matrix 
    * with the value Dval in all diagonal positions. 
    */
    virtual void diag_matrix(const T_& Dval);

    /// trace(): returns the sum of the diagonal elements.
    virtual T_ trace() const;

    /// I/O
    virtual void read_matrix(istream& In);
    
    /// hidden methods
    protected:
    
    void print_rows(ostream& Out, size_t Rw, size_t Jbeg, 
	size_t Items, size_t Width, size_t Prec) const;
    static unsigned int tri_size(unsigned int X) { return((X*(X+1))/2); }
};
// END OF CLASS Symmat_

// ---- OUT-OF-CLASS FUNCTIONS ----

// These were originally friend templates but EGCS 1.1.2
// did not like them. 12-Jul-1999.

template <class T_>
Rectmat_<T_> operator+(const Rectmat_<T_>& R, const Symmat_<T_>& S)
    { return (S+R); }
template <class T_>
Rectmat_<T_> operator-(const Rectmat_<T_>& Mat, const Symmat_<T_>& Sym);
template <class T_>
Symmat_<T_> operator*(const T_& Factor, const Symmat_<T_>& Bm);
template <class T_>
Rectmat_<T_> operator*(const Rectmat_<T_>& Rect, const Symmat_<T_>& Sym);

// ---- TYPEDEFS ----

// Float and double matrices (for convenience)
typedef Symmat_<double> Smatdbl_;
typedef Symmat_<float> Smatflt_;

} // RazorBack

#ifdef INCLUDE_TMPL_DEFS
#include "Symmat.cc"
#endif

// ==== END OF TEMPLATE HEADER Symmat.hh ====

#endif	    // SYMMAT_TMPL_HEADER
