#ifndef SYMVAR_HEADER
#define SYMVAR_HEADER

// ==== CLASS HEADER Symvar.hh ====

/**
 * \file Symvar.hh
 * Class to represent "symbolic variables" that have
 * a string value taken from a fixed list of possible strings.
 */

// 5-Oct-2001. Andras Aszodi

// ==== STANDARD HEADERS ====

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>

#ifdef HAVE_STD
#  include <list>
#  include <iostream>
#  include <iomanip>
#  include <algorithm>
  using namespace std;
#else
#  include <list.h>
#  include <iostream.h>
#  include <iomanip.h>
#  include <algo.h>
#endif

#include <string>
#ifndef NPOS
#define NPOS string::npos
#endif

// ==== MODULE HEADERS ====

#include "Utilsexc.hh"

// ==== CLASSES ====

namespace RazorBack {

/**
 * Objects of this class represent symbolic variables. Such variables
 * may store a string value from a set of prescribed tokens.
 */
class Symvar_
{
    // data
    private:
	    
    const string Separators;
    list<string> Toklist;
    list<string>::iterator Curtokptr;
    
    // methods
    public:
	    
    /**
     * Inits to contain the allowed tokens to Tokenstr. This string shall contain the tokens
     * separated by one of the allowed separator characters (given by Seps,
     * "|, \t" by default). For example, set_tokens("tok1|tok2,tok3")
     * specifies the three acceptable tokens "tok1", "tok2", "tok3".
     * The current token iterator will be set to the first one.
     */
    explicit Symvar_(const string& Tokenstr="", const string& Seps="|, \t"):
	    Separators(Seps), Toklist() { set_tokens(Tokenstr); }
    
    /// Copy constructor.
    Symvar_(const Symvar_& Sv):
	    Separators(Sv.Separators), Toklist(Sv.Toklist)
	    { curtokptr_offset(Sv.curtokptr_offset()); }
    
    /// Destructor.
    virtual ~Symvar_() {}
    
    /// Assignment.
    Symvar_& operator=(const Symvar_& Rhs)
    {
	if (this!=&Rhs)
	{
	    Toklist=Rhs.Toklist;
	    curtokptr_offset(Rhs.curtokptr_offset());
	}
	return *this;
    }
    
    /**
     * Sets the allowed tokens to Tokenstr. This string shall contain the tokens
     * separated by one of the allowed separator characters (returned by the
     * static method legal_separators()). For example, set_tokens("tok1|tok2,tok3")
     * specifies the three acceptable tokens "tok1", "tok2", "tok3".
     * The current token iterator will be set to the first one.
     * Returns the number of tokens seen.
     */
    unsigned int set_tokens(const string& Tokenstr);
    
    /// Adds a new token Newtok to the list of allowed tokens.
    void add_token(const string& Newtok) { Toklist.push_back(Newtok); }
    
    /**
     * Removes the token Tok from the list of allowed tokens.
     * If the token to be removed was the current one, then
     * the new current token will be the first in the internal list.
     */
    void remove_token(const string& Tok);
    
    // ---- Access ----
    
    /// Returns a string containing all accepted separator characters.
    const string& separators() { return Separators; }
    
    /// Returns true if Tok is one of the valid tokens. Comparison is case-sensitive.
    bool valid_token(const string& Tok) const
    {
	return (Toklist.end()!=find(Toklist.begin(),Toklist.end(), Tok));
    }
    
    /// Returns true if the calling object contains no valid tokens.
    bool empty() const { return Toklist.empty(); }
    
    /**
     * Returns a reference to the current token.
     * Throws Emptyexc_ if the calling object has no tokens defined.
     */
    const string& token() const
    {
	if (empty())
	    throw Emptyexc_("Symvar_::token()");
	return *Curtokptr;
    }
    
    /**
     * Sets the current token to Tok if it is valid.
     * Returns the old current token.
     * Throws Emptyexc_ if the calling object was empty.
     */
    const string& token(const string& Tok);
    
    // ---- I/O ----
    
    /**
     * Input: reads the next string into Sv.
     * If it is a valid token, then it will be stored in Sv,
     * otherwise Sv is not modified and In's failbit will be set.
     */
    friend istream& operator>>(istream& In, Symvar_& Sv);
    
    /// Output: writes the token string in Sv to Out. No newline at the end.
    friend ostream& operator<<(ostream& Out, const Symvar_& Sv);
    
    /**
     * Constructs and returns a string that contains all legal tokens,
     * separated by the first char in the string returned by separator().
     * If the calling object is empty, then "" is returned.
     */
    string all_tokens() const;
    
    // hidden methods
    protected:
    
    unsigned int curtokptr_offset() const;
    void curtokptr_offset(unsigned int Offs);
};
// END OF CLASS Symvar_

} // RazorBack

// ==== END OF HEADER Symvar.hh ====

#endif      // SYMVAR_HEADER
