#ifndef UTILSEXC_HEADER
#define UTILSEXC_HEADER

// ==== Utilsexc.hh ====

/**
 * \file Utilsexc.hh
 * Base class for the exceptions thrown by the Razorback C++ library.
 * Implements a consistent error reporting scheme.
 */

// 8-Oct-2001. Andras Aszodi

// ---- STANDARD HEADERS ----

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>

#ifdef HAVE_STD
#  include <iostream>
#  include <iomanip>
#  include <sstream>
#else
#  include <iostream.h>
#  include <iomanip.h>
#  include <strstream.h>
#  define istringstream istrstream 
#  define ostringstream ostrstream 
#endif

#include <string>
using namespace std;
#ifndef NPOS
#define NPOS string::npos
#endif

// ==== CLASSES ====

namespace RazorBack {
    
/**
 * Utilsexc_: the exception base class for the Razorback C++ library.
 * Can store the name of the function which has thrown it
 * and a flag indicating warnings or fatal conditions.
 */
class Utilsexc_
{
    // data
    private:
    string Throwname;    ///< for storing throw function name
    bool Fatal;     ///< true for fatal errors, false for warnings
    
    protected:
    string Message; 	///< stores the formatted message, derived classes need access
    
    // methods
    public:
    
    /**
     * Init with throwing function name Thnm, and fatality indicator F.
     * By default it is false, indicating that the exceptional condition
     * was not that bad.
     * Msg, if not NULL (the default) or "" may contain a verbal description
     * of the error. Use this facility when it is not worthwhile to derive
     * an exception class for the error you wish to signal.
     */
    explicit Utilsexc_(const char *Thnm, bool F=false, const char *Msg=NULL);
    
    /// Returns the function name causing the exception.
    const string& throw_name() const { return Throwname; }
    
    /// Returns the complete formatted error message.
    const string& error_message() const { return Message; }
    
    /// Returns the fatality indicator value.
    bool is_fatal() const { return Fatal; }
    
    // hidden methods
    protected:
    
    void print_prefix(ostringstream& Ost) const;
    
    private:
    Utilsexc_();    // disallow default ctor
};
// END OF CLASS Utilsexc_

/**
 * Emptyexc_: empty object exception. Should be thrown 
 * if objects containing no data are accessed.
 * Usage: throw Emptyexc_("bad_function")
 */
class Emptyexc_: public Utilsexc_
{
    // methods
    public:
	
    /// Constructs an exception thrown in Thnm.
    explicit Emptyexc_(const char *Thnm, bool F=false);
    
    private:
    Emptyexc_();    // disallow default ctor
};
// END OF CLASS Emptyexc_

} // RazorBack

// ==== END OF HEADER Utilsexc.hh ====

#endif	// UTILSEXC_HEADER
