#ifndef VECTOR_TMPL_HEADER
#define VECTOR_TMPL_HEADER

// ==== TEMPLATE HEADER Vector.hh ====

/**
 * \file Vector.hh
 * Linear algebra vector template class based on STL.
 */

// 12-Jul-1999. Andras Aszodi

// ---- COMPATIBILITY NOTES ----

// Exception support is assumed.
// This module depends on the Standard Template Library.
 
// ---- COMPILER COMMAND LINE MACROS ----

// Define the macro INCLUDE_TMPL_DEFS if your compiler
// does not support automatic template source inclusion.

// ---- STANDARD HEADERS ----

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <math.h>

#ifdef HAVE_STD
#  include <iostream>
#  include <iomanip>
#  include <vector>
#  include <algorithm>
#  include <numeric>
#  include <functional>
   using namespace std;
#else
#  include <iostream.h>
#  include <iomanip.h>
#  include <vector.h>
#  include <algo.h>
#  include <function.h>
#endif

// ---- LIBRARY HEADERS ----

#include "Rangedimexc.hh"	// linear algebra exceptions

// ==== CLASSES ====

namespace RazorBack {

/**
 * Vector_: this class implements a template for a linear
 * algebra vector class. The objects are N>0 -dimensional
 * vectors with all the usual arithmetic operations. The base
 * type T_ must support meaningful arithmetics and output <<.
 */
template <class T_>
class Vector_
{
    // data
    private:
    
    vector<T_> Vec;	// the STL container holding the coordinates
    size_t Dim;	// the dimension
    bool Ftnidx;    	// true for FORTRAN-style [1..N] indexing
    bool Check;     	// true when range checking is enabled.
    
    // methods
    public:
    
    // constructors and destructor
    
    /**
    * Construct an N-dimensional vector (default 3, 0 disallowed)
    * and set its elements to X (default is "0").
    */
    explicit Vector_(size_t N=3, const T_& X=T_(0)):
	Vec(!N? 3: N, X), Dim(!N? 3: N), 
	Ftnidx(false), Check(false) {}
    
    /**
    * Construct the Vector_ from an STL vector V. If V is empty,
    * then a 3-dimensional Vector_ will be created.
    */
    explicit Vector_(const vector<T_>& V): 
	    Vec(V), Dim(Vec.size()), Ftnidx(false), Check(false)
    {
	if (V.empty()) dim(3);
    }
    
    /**
    * Construct the Vector_ from a traditional C-style array.
    * There are no checks: you have been warned.
    */
    Vector_(const T_* Arr, size_t Len):
	    Vec(Len), Dim(Len), Ftnidx(false), Check(false)
    {
	for (register unsigned int i=0; i<Len; i++)
	    Vec[i]=Arr[i];
    }
    
    /// Destructor: empty virtual for eventual derived classes
    virtual ~Vector_() {}
    
    /**
    * []: unchecked C-style access. Ignores FORTRAN access status.
    * (): allows range-checked and FORTRAN-style indexing.
    * Throws Indexrangexc_ on range errors.
    */
    const T_& operator[](size_t Idx) const { return Vec[Idx]; }
    T_& operator[](size_t Idx) { return Vec[Idx]; }
    const T_& operator()(size_t Idx) const;
    T_& operator()(size_t Idx);
    
    /**
    * ftn_idx(): returns "true" if the calling vector object 
    * was switched to use FORTRAN-style [1..N]
    * indexing instead of the usual [0..N-1] C-style access. Useful
    * for porting Numerical Recipes C code which uses the FORTRAN
    * indexing convention. 
    * ftn_idx(F): switches the FORTRAN boolean variable to F.
    * Returns old state.
    */
    bool ftn_idx() const { return Ftnidx; }
    bool ftn_idx(bool F) { bool Of=Ftnidx; Ftnidx=F; return Of; }
    
    /**
    * range_check(): returns true if range checking is enabled.
    * range_check(C): sets Check to C, returns old range checking status.
    */
    bool range_check() const { return Check; }
    bool range_check(bool C) { bool Oc=Check; Check=C; return Oc; }
    
    /// set_values(): sets all coordinates to X (0 by default).
    Vector_<T_>& set_values(const T_& X=T_(0));
    
    /**
    * dim(): returns the dimension (size) of the vector.
    * dim(D): adjusts the dimension to D or to 3 if D==0.
    * Returns previous dimension.
    */
    size_t dim() const { return(Dim); }
    size_t dim(size_t D);
    
    /// begin(), end(): for STL-style manipulations
    typename vector<T_>::const_iterator begin() const { return Vec.begin(); }
    typename vector<T_>::iterator begin() { return Vec.begin(); }
    typename vector<T_>::const_iterator end() const { return Vec.end(); }
    typename vector<T_>::iterator end() { return Vec.end(); }
    
    /**
    * Scalar arithmetics: add, subtract, multiply or divide 
    * each coordinate the T_ type object X "in place". The modified
    * calling object is returned.
    */
    Vector_<T_>& operator+=(const T_& X);
    Vector_<T_>& operator-=(const T_& X);
    Vector_<T_>& operator*=(const T_& X);
    Vector_<T_> operator*(const T_& X) const;
    Vector_<T_>& operator/=(const T_& X);
    Vector_<T_> operator/(const T_& X) const;
    
    /**
    * Vector arithmetics: addition, subtraction, scalar product.
    * The operands must have equal dimensions otherwise a Dimexc_ is thrown.
    */
    Vector_<T_>& operator+=(const Vector_<T_>& V);
    Vector_<T_> operator+(const Vector_<T_>& V) const;
    Vector_<T_>& operator-=(const Vector_<T_>& V);
    Vector_<T_> operator-(const Vector_<T_>& V) const;
    T_ operator*(const Vector_<T_>& V) const;
    
    	/// Modulus
    
    /**
    * vec_len2(): calculates the square of the Euclidean norm.
    * Can be invoked either as "Vector.vec_len2()" or "vec_len2(Vector)".
    * vec_len(): returns the Euclidean norm for types T_ which
    * support a sqrt(T_) operation.
    */
    T_ vec_len2() const { return((*this) * (*this)); }
    T_ vec_len() const { return(sqrt(vec_len2())); }

    /**
    * vec_norm(): normalises the calling object to a unit vector "in place".
    * Returns the original length. The type T_ must support
    * a sqrt(T_) operation. "Null"-vectors should ideally
    * throw a "division by zero" exception later.
    */
    T_ vec_norm();
    
    	/// STL-style transformations
    
    /**
    * transform(): this method allows the modification of each
    * coordinate of the calling object "in place" using the
    * STL function object Unaryfunc. Could be used for things like
    * taking the square root of each coordinate etc. See also
    * the arithmetics methods for multiplications by a scalar etc.
    */
    void transform(unary_function<T_, T_> Unaryfunc)
    {
	std::transform(Vec.begin(), Vec.end(), Vec.begin(), Unaryfunc);
    }
    
    /// reads from In into the calling object.
    void read(istream& In);
};
// END OF CLASS Vector_

// ---- OUT-OF-CLASS FUNCTIONS ----

// These were originally friend templates but EGCS 1.1.2
// did not like them. 12-Jul-1999.

/**
 * Pre-multiplies the vector V with scalar X.
 */
template <class T_>
Vector_<T_> operator*(const T_& X, const Vector_<T_>& V);	// see Bug Notes

/**
 * cross_prod(): returns the cross-product of two 3D vectors.
 * If either of the arguments are not 3D, then a Dimexc_ exception
 * is thrown.
 */
template <class T_>
Vector_<T_> cross_prod(const Vector_<T_>& V1, const Vector_<T_>& V2); // see Bug Notes

/**
 * vec_len2(): calculates the square of the Euclidean norm.
 * This can be invoked as "vec_len2(Vector)".
 * Note that there is a member function version, too. Just syntactic sugar :-)
 */
template <class T_>
T_ vec_len2(const Vector_<T_>& V) { return(V*V); }

/**
 * vec_len(): returns the Euclidean norm for types T_ which
 * support a sqrt(T_) operation.
 */
template <class T_>
T_ vec_len(const Vector_<T_>& V) { return(sqrt(V*V)); }

/**
 * diff_len2(): returns the squared length of V1-V2. On dim
 * mismatch a Dimexc_ exception is thrown.
 */
template <class T_>
T_ diff_len2(const Vector_<T_>& V1, const Vector_<T_>& V2);	// see Bug notes

    // I/O

/**
 * Reads a vector V from the stream In.
 * This is how the vector format looks like:-
 * \verbatim
    ------------------
    0 | -1.740000e+01
    1 | -2.320000e+01
    2 | -2.900000e+01
    ==================
    \endverbatim
 */
template <class T_>
istream& operator>>(istream& In, Vector_<T_>& V)
{
    V.read(In);
    return In;
}

/// Writes the vector V to stream Out.
template <class T_>
ostream& operator<<(ostream& Out, const Vector_<T_>& V);

// ---- TYPEDEFS ----

/// Float and double vectors (for convenience)
typedef Vector_<double> Vecdbl_;
typedef Vector_<float> Vecflt_;

} // RazorBack

#ifdef INCLUDE_TMPL_DEFS
#include "Vector.cc"
#endif

// ==== END OF TEMPLATE HEADER Vector.hh ====

#endif	// VECTOR_TMPL_HEADER
