/* ==== FUNCTIONS bestrot.c ==== */

/* An implementation of the 3D point set alignment algorithm by
 * A. D. McLachlan. Reference:
 * McLachlan, A. D. (1979): J. Mol. Biol. 128: 49-79.
 */

/* 10-Nov-2000. Andras Aszodi */

/* ---- STANDARD HEADERS ---- */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

/* ---- MODULE HEADERS ---- */

#include "bestrot.h"
#include "svd.h"
#include "lu.h"

/* ---- DEFINITIONS ---- */

#define DIM 3

/* ==== FUNCTIONS ==== */

/*
 * center_vectors(X, Ctr, W):
 * Calculates the weighted centroid of the 3-dimensional row vectors in the X matrix
 * and puts the result into the Ctr vector.
 * X must have 3 columns otherwise 0 is returned.
 * The weight vector W must have the same dimension as the number of rows in X.
 * The size of Ctr will automatically set to 3 if necessary.
 * If W==NULL, then uniform weighting is assumed.
 * Return value: 0 on error, the no. of vectors (rows) in X if OK.
 */
int center_vectors(Matrect_ *X, Vect_ *Ctr, const Vect_ *W)
{
	register unsigned int i, j, Vno;
	register double Wsum, Temp1, Temp2;
	int deallocate = 0;
    
	/* dim checks */
	if (X->Cno!=DIM)
	{
		fprintf(stderr, "\n? center_vectors(): X[%u][%u]: Row vectors must be %u-dimensional\n",
						X->Rno, X->Cno, DIM);
		return 0;
	}
	Vno=X->Rno;
	
	/* allocate centroid vector if absent */
	if (Ctr==NULL)
	{
		Ctr = (Vect_*) malloc(sizeof(Vect_));
		deallocate = 1;
		init_vect(Ctr,DIM);
	}
	if (Ctr->Len!=DIM)
		resize_vect(Ctr, DIM);
    
	/* get the weighted centroid */
	if (W!=NULL)
	{
		Wsum=0.0;
		for(i=0; i<Vno; i++) Wsum+=W->Vec[i];
	}
	else Wsum=Vno;
	for (j=0; j<DIM; j++)
	{
		Temp1=0.0;
		for (i=0; i<Vno; i++)
		{
	    Temp2=X->Mat[i][j];
	    if (W!=NULL) Temp2*=W->Vec[i];
	    Temp1+=Temp2;
		}
		Ctr->Vec[j]=Temp1/Wsum;
	}
    
	/* subtract Ctr from all vectors in X */
	for (i=0; i<Vno; i++)
		for (j=0; j<DIM; j++)
	    X->Mat[i][j]-=Ctr->Vec[j];

	if (deallocate==1)
	{
		free_vect(Ctr);
		free(Ctr);
	}
    
	return Vno;
}
/* END of center_vectors(X, Ctr, W) */

/*
 * best_rot(X, Y, W, Transform):
 * finds the best rotation matrix that brings a set of
 * vectors X into another set Y. X, Y shall contain the same number of
 * 3-dim vectors as their rows.
 * W is a weight vector with as many elements as the rows in X or Y. 
 * If W==NULL, then uniform weighting is assumed.
 * Transform is a 3x3 square matrix (resized if necessary) that on
 * return contains the X->Y transformation. It is assumed that
 * X and Y were centered before the call and that Transform
 * has properly been initialised.
 * NOTE: this routine cannot handle the degenerate cases when
 * the point sets are Dim<3-dimensional. When this happens,
 * a warning is printed to stderr
 * and -1.0 (a meaningless RMS value) is returned.
 * Otherwise, the weighted least-squares error function is returned.
 */
double best_rot(const Matrect_ *X, const Matrect_ *Y, 
    	const Vect_ *W, Matrect_ *Transform)
{
	register unsigned int i, j, k, n, Vno;
	int Psign, Rank;
	Svdm_ Svd;
	Matrect_ U;
	double **K, **H;
	register double Err=0.0, Temp1, Temp2, Wsum=0.0;
	double Detu;
    
	/* checks */
	if (X->Cno!=DIM)
	{
		fprintf(stderr, "\n? best_rot(): X[%u][%u]: Row vectors must be %u-dimensional\n",
						X->Rno, X->Cno, DIM);
		return -1.0;
	}
	if (Y->Cno!=DIM)
	{
		fprintf(stderr, "\n? best_rot(): Y[%u][%u]: Row vectors must be %u-dimensional\n",
						Y->Rno, Y->Cno, DIM);
		return -1.0;
	}
	Vno=X->Rno;
	if (Vno!=Y->Rno)
	{
		fprintf(stderr, "\n? best_rot(): X:Y row dim mismatch (%u!=%u)\n",
						Vno, Y->Rno);
		return -1.0;
	}
	if (W!=NULL && W->Len!=Vno)
	{
		fprintf(stderr, "\n? best_rot(): X:W dim mismatch (%u!=%u)\n",
						Vno, W->Len);
		return -1.0;
	}
    
	/* set up the matrix to be SVD-d */
	init_matrect(&U, DIM, DIM);
	for (i=0; i<DIM; i++)
		for (j=0; j<DIM; j++)
		{
	    Temp1=0.0;
	    for (k=0; k<Vno; k++)
	    {
				Temp2=X->Mat[k][i]*Y->Mat[k][j];
				if (W!=NULL) Temp2*=W->Vec[k];
				Temp1+=Temp2;
	    }
	    U.Mat[i][j]=Temp1;
		}
    
	/* set up and perform SVD */
	svd_setup(&Svd, DIM, DIM);
	svd_copy(&U, &Svd);
    
	/* check rank: do nothing if rank was lost */
	if ((Rank=rank_cond(&Svd, SVD_EPSILON, NULL))<3)
	{
		fprintf(stderr, "\n? best_rot(): Rank deficiency %u<%u\n", Rank, DIM);
		free_svd(&Svd); free_matrect(&U);
		return(-1.0);
	}
    
	/* get the determinant of U and store its sign in Psign */
	Psign=lu_decomp(&U, NULL);
	Detu=lu_det(&U, Psign);
	Psign=(Detu>0)? 1: -1;
    
	/* generate the transform matrix: here we explicitly
	 * use DIM==3 because McLachlan does not say what to do
	 * if Psign==-1 and DIM>3 and I don't know :-)
	 */
	resize_matrect(Transform, DIM, DIM);
	K=Svd.U.Mat;   /* shorthands */
	H=Svd.V.Mat;
	for (i=0; i<DIM; i++)
		for (j=0; j<DIM; j++)
	    Transform->Mat[i][j]=
				K[i][0]*H[j][0]+K[i][1]*H[j][1]+Psign*K[i][2]*H[j][2];
    
	/* evaluate the error function */
	for (n=0; n<Vno; n++)
	{
		Temp2=0.0;
		for (i=0; i<DIM; i++)
		{
	    Temp1=0.0;
	    for (j=0; j<DIM; j++)
				Temp1+=Transform->Mat[i][j]*Y->Mat[n][j];
	    Temp1-=X->Mat[n][i];
	    Temp2+=Temp1*Temp1;
		}
		if (W!=NULL)
		{
	    Err += W->Vec[n]*Temp2;
	    Wsum += W->Vec[n];
		}
		else Err+=Temp2;
	}
	Err/=(W!=NULL)? Wsum: Vno;
    
	/* cleanup */
	free_svd(&Svd); 
	free_matrect(&U);

	return sqrt(Err);
}
/* END of best_rot(X, Y, W, Transform) */

#undef DIM

/* ==== END OF FUNCTIONS bestrot.c ==== */
