/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef BESTROT_HEADER
#define BESTROT_HEADER

/* ==== HEADER bestrot.h ==== */

/* An implementation of the point set alignment algorithm by
 * A. D. McLachlan. Reference:
 * McLachlan, A. D. (1979): J. Mol. Biol. 128: 49-79.
 */

/* 10-Nov-2000. Andras Aszodi */

/* ---- INCLUDE FILES ---- */

#include "matrect.h"
#include "vect.h"

/* ---- PROTOTYPES ---- */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * center_vectors(X, Ctr, W):
 * Calculates the weighted centroid of the 3-dimensional row vectors in the X matrix
 * and puts the result into the Ctr vector.
 * X must have 3 columns otherwise 0 is returned.
 * The weight vector W must have the same dimension as the number of rows in X.
 * The size of Ctr will automatically set to 3 if necessary.
 * If Ctr==NULL, then the vector will be allocated inside: don't forget to free up
 * when you are done with it.
 * If W==NULL, then uniform weighting is assumed.
 * Return value: 0 on error, the no. of vectors (rows) in X if OK.
 */
int center_vectors(Matrect_ *X, Vect_ *Ctr, const Vect_ *W);

/**
 * best_rot(X, Y, W, Transform):
 * finds the best rotation matrix that brings a set of
 * vectors X into another set Y. X, Y shall contain the same number of
 * 3-dim vectors as their rows.
 * W is a weight vector with as many elements as the rows in X or Y. 
 * If W==NULL, then uniform weighting is assumed.
 * Transform is a 3x3 square matrix (resized if necessary) that on
 * return contains the X->Y transformation. It is assumed that
 * X and Y were centered before the call and that Transform
 * has properly been initialised.
 * NOTE: this routine cannot handle the degenerate cases when
 * the point sets are Dim<3-dimensional. When this happens,
 * a warning is printed to stderr
 * and -1.0 (a meaningless RMS value) is returned.
 * Otherwise, the weighted least-squares error function is returned.
 */
double best_rot(const Matrect_ *X, const Matrect_ *Y, 
    	const Vect_ *W, Matrect_ *Transform);

#ifdef __cplusplus
}
#endif

/* ==== END OF HEADER bestrot.h ==== */

#endif	/* BESTROT_HEADER */
