/* ==== FUNCTIONS lu.c ==== */

/* Solves systems of linear equations by LU-decomposition. */

/* 21-Aug-2000. Andras Aszodi */

/* ==== STANDARD HEADERS ==== */

#include <stdio.h>
#include <math.h>

/* ==== MODULE HEADERS ==== */

#include "lu.h"

/* ==== DEFINITIONS ==== */

#ifdef FLT_MIN
#define LU_EPSILON (10.0*FLT_MIN)
#else
#define LU_EPSILON (1.0e-10)
#endif

/* ==== FUNCTIONS ==== */

/* lu_decomp(M, Perm): 
 * performs an LU-decomposition in place on the square matrix *M.
 * Based on partial pivoting: the row permutations are done in Perm[]
 * (assumed to be of correct size) and will be used by lu_solve().
 * If Perm==NULL, then a permutation vector will be created
 * internally and will be freed before return. This option can be
 * used when only the determinant is calculated from the LU-decomposition.
 * Note that the contents of *M will be overwritten.
 * Return value: the sign of the determinant of the permutation
 * matrix (+/-1) or 0 if *M is singular. If *M is not square,
 * a warning is printed and 0 is returned.
 */
int lu_decomp(Matrect_ *M, int *Perm)
{
    register size_t i, j, k, n, imax;
    int Psign=1;
    int *Idx=NULL;
    register double Large, Pivot, Tmp, Tmp2;
    double *Scal=NULL;
    double **A=NULL;
    
    /* check and array initialisation */
    if (M==NULL)
    {
	fputs("\n? lu_decomp(NULL,...): invalid matrix arg\n", stderr);
	return 0;
    }
    if (!M->Rno || !M->Cno || M->Rno!=M->Cno)
    {
	fprintf(stderr, "\n? lu_decomp(M,...): invalid matrix size %ux%u\n",
		M->Rno, M->Cno);
	return 0;
    }
    A=M->Mat;	/* direct access to matrix elements thru A[i][j]*/
    n=M->Rno;	/* size */
    
    /* if there is no permutation vector, provide one internally */
    if (Perm==NULL)
	Idx=(int *) calloc(n, sizeof(int));
    else Idx=Perm;  /* just use the one provided */
    
    Scal=(double *) calloc(n, sizeof(double));	/* implicit scaling array */
    
    /* get implicit scaling: if a row contains 0-s only,
     * then the matrix is singular which will be indicated
     * by setting Psign=0. Precision is controlled by
     * the constant LU_EPSILON (see Definitions above).
     */
    for (i=0; Psign && i<n; i++)
    {
	Large=0.0;
	for (j=0; j<n; j++)
	    if ((Tmp=fabs(A[i][j]))>Large) Large=Tmp;
	if (Large<LU_EPSILON)	/* (almost) singular */
	{ Psign=0; break; }
	Scal[i]=1.0/Large;
    }
    
    /* loop over columns */
    for (j=0; Psign && j<n; j++)
    {
	for (i=0; i<j; i++)
	{
	    Tmp=A[i][j];
	    for (k=0; k<i; k++) Tmp-=A[i][k]*A[k][j];
	    A[i][j]=Tmp;
	}
	
	/* find largest pivot */
	Large=0.0;
	for (i=j; i<n; i++)
	{
	    Tmp=A[i][j];
	    for (k=0; k<j; k++) Tmp-=A[i][k]*A[k][j];
	    A[i][j]=Tmp;
	    
	    if ((Tmp2=Scal[i]*fabs(Tmp))>=Large)    /* best so far */
	    { Large=Tmp2; imax=i; }
	}
	
	/* interchange rows? */
	if (j!=imax)
	{
	    for (k=0; k<n; k++)	    /* not too efficient copy */
	    {
		Tmp=A[imax][k]; A[imax][k]=A[j][k]; 
		A[j][k]=Tmp;
	    }
	    Psign*=(-1);    /* parity change */
	    Scal[imax]=Scal[j];
	}
	Idx[j]=imax;
	
	/* get the pivot */
	Pivot=A[j][j];
	if (fabs(Pivot)<LU_EPSILON)	/* singularity */
	{ Psign=0; break; }
	
	/* divide by the pivot */
	if (j<n-1)
	    for (i=j+1; i<n; i++) A[i][j]/=Pivot;
    }	/* for j */
    
    free(Scal); 
    if (Perm==NULL) free(Idx);	/* internal permutation vector was used */
    return(Psign);
}
/* END of lu_decomp(M, Perm) */

/* lu_det(Lu, Psign):
 * calculates the determinant of the LU-decomposed
 * square matrix *Lu. Psign is the permutation sign returned by
 * lu_decomp(*Lu, NULL). If you did not invoke lu_decomp(..)
 * before lu_det(..), then you can blame only yourself :-)
 */
double lu_det(const Matrect_ *Lu, int Psign)
{
    register int i;
    size_t n;
    register double Det, Aii;
    double **A=NULL;
    
    /* check and array initialisation */
    if (Lu==NULL)
    {
	fputs("\n? lu_det(NULL,...): invalid matrix arg\n", stderr);
	return 0.0;
    }
    if (!Lu->Rno || !Lu->Cno || Lu->Rno!=Lu->Cno)
    {
	fprintf(stderr, "\n? lu_det(Lu,...): invalid matrix size %ux%u\n",
		Lu->Rno, Lu->Cno);
	return 0.0;
    }
    A=Lu->Mat;	/* direct access to matrix elements thru A[i][j]*/
    n=Lu->Rno;	/* size */
    
    if (!Psign) return 0.0;	/* matrix is singular */
    
    Det=0.0;
    for (i=0; i<n; i++)
    {
	if ((Aii=A[i][i])<0.0) Psign*=(-1);   /* save sign */
	Det+=log(fabs(Aii));	/* sum logarithms to avoid overflow */
    }
    Det=Psign*exp(Det);
    return Det;
}
/* END of lu_det(Lu, Psign) */

/* lu_solve(M, Perm, b):
 * solves the linear equation A*x=b (A is n*n, x,b are n-long arrays).
 * A is supposed to have been LU-decomposed by lu_decomp() above and
 * the row permutation is stored in Perm[]. You must invoke
 * lu_decomp(M, Perm) with a pre-allocated Perm vector before
 * using this routine.
 * b[] is the "right-hand-side" vector which contains the solution on return.
 */
void lu_solve(const Matrect_ *M, const int Perm[], double b[])
{
    register size_t n, ip;
    register int i, j;
    register double Tmp;
    double **A=NULL;

    /* check and array initialisation */
    if (M==NULL || Perm==NULL || b==NULL)
    {
	fputs("\n? lu_solve(): NULL arg detected\n", stderr);
	return;
    }
    if (!M->Rno || !M->Cno || M->Rno!=M->Cno)
    {
	fprintf(stderr, "\n? lu_solve(M,...): invalid matrix size %ux%u\n",
		M->Rno, M->Cno);
	return;
    }
    A=M->Mat;	/* direct access to matrix elements thru A[i][j]*/
    n=M->Rno;	/* size */
    
    /* permute forward */
    for (i=0; i<n-1; i++)
	if ((ip=Perm[i])!=i)
	{ Tmp=b[ip]; b[ip]=b[i]; b[i]=Tmp; }
    
    /* forward substitution */
    for (i=0; i<n; i++)
    {
	Tmp=b[i];
	for (j=0; j<i; j++) Tmp-=A[i][j]*b[j];
	b[i]=Tmp;
    }
    
    /* back substitution */
    for (i=n-1; i>=0; i--)
    {
	Tmp=b[i];
	for (j=i+1; j<n; j++) Tmp-=A[i][j]*b[j];
	b[i]=Tmp/A[i][i];
    }
}
/* END of lu_solve(M, Perm, b) */

#undef LU_EPSILON

/* ==== END OF FUNCTIONS lu.c ==== */
