/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef MATRECT_HEADER
#define MATRECT_HEADER

/* ==== HEADER matrect.h ==== */

/**
 * \file matrect.h
 * Double-precision rectangular matrix module.
 * Supports safe creation, destruction, copy, resize, input/output.
 */

/* 16-Nov-2000. Andras Aszodi */

/* ==== STANDARD HEADERS ==== */

#include <stdlib.h>
#include <stdio.h>

/* ==== STRUCTS ==== */

/**
 * Matrect_: a struct storing a rectangular double matrix.
 * The matrix elements are in row major order in an array pointed to by Mat[0],
 * Mat is Rno long and contains pointers to the matrix rows.
 * The (i,j):th matrix element can be accessed as Matrect_.Mat[i][j].
 */
typedef struct
{
    double **Mat;   	/**< data storage */
    unsigned int Rno, Cno;	/**< row and column logical size */
    unsigned int Rlen, Elen;	/**< physical length of Mat and Mat[0] */
} Matrect_;

/* ==== PROTOTYPES ==== */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * init_matrect(M, Rno, Cno):
 * Sets up the matrix *M to contain Rno rows and Cno columns.
 * If Rno==0 or Cno==0 then they will be set to 1.
 * This must be the first operation performed on a Matrect_ struct.
 * Disaster strikes if *M is initialised more than once.
 */
void init_matrect(Matrect_ *M, unsigned int Rno, unsigned int Cno);

/**
 * array_matrect(M, Array):
 * Assumes that Array contains a matrix in row major order
 * and copies its contents into M.
 * It is your responsibility to make sure that Array has the right size
 * and the appropriate layout. This routine assumes no responsibility
 * and performs no dim checks. M must have been initialised and
 * set to the desired size before calling this routine.
 * The only support you can expect that nothing is done if
 * either of the parameters is NULL.
 */
void array_matrect(Matrect_ *M, const double *Array);

/**
 * resize_matrect(M, Newrno, Newcno):
 * changes the size of *M so that Newrno rows and Newcno columns can be stored.
 * Does nothing if the dimensions are the same as before.
 * If *M was empty then it will be initialised.
 * Newly added rows/cols will be initialised to 0.0.
 * If Newrno or Newcno are 0, then they will be set to 1.
 */
void resize_matrect(Matrect_ *M, unsigned int Newrno, unsigned int Newcno);

/**
 * copy_matrect(From, To):
 * Copies the contents of *From to *To. *To will be initialised
 * and/or resized if necessary.
 */
void copy_matrect(const Matrect_ *From, Matrect_ *To);

/**
 * read_matrect(In, M):
 * reads a matrix from the file In into *M.
 * Format is the "nice format" produced by write_matrect().
 * Returns 0 on error, the number of lines processed otherwise.
 */
int read_matrect(FILE *In, Matrect_ *M);

/**
 * write_matrect(Out, M, Linewidth, Width, Prec): 
 * lists M to the open stream Out with entries in scientific format,
 * Prec digits precision. If a row takes up more than 80 chars,
 * then the matrix is cut up nicely.
 */
void write_matrect(FILE *Out, const Matrect_ *M, int Prec);

/**
 * free_matrect(M):
 * frees up all storage associated with *M.
 * *M will be kept, it will be set to an invalid empty state.
 * This must be the last operation performed on a Matrect_ struct.
 */
void free_matrect(Matrect_ *M);
    
#ifdef __cplusplus
}
#endif

/* ==== END OF HEADER matrect.h ==== */

#endif	    /* MATRECT_HEADER */
