/* ==== FUNCTIONS svd.c ==== */

/* Singular Value Decomposition, based on the Numerical Recipes
 * routine 'svdcmp'. Rewritten from my 1994 implementation.
 * 
 * Reference:- Press, W.H., Flannery, B.P., Teukolsky, S.A. and Vetterling, W.T.:
 * Numerical Recipes in C: The Art of Scientific Computing.
 * Cambridge University Press, 1992.
 */

/* 23-Aug-2000. Andras Aszodi. */

/* ==== STANDARD HEADERS ==== */

#include <stdio.h>
#include <math.h>
#include <string.h>

/* ==== MODULE HEADER ==== */

#include "svd.h"

/* ==== DEFINITIONS ==== */

/* Built-in sqrt(x^2+y^2) or pythag_recipes(): check math.h */
#ifdef __EXTENSIONS__
#define PYTHAG(a, b) hypot(a, b)    /* built-in BSD4.3 ANSI C extension */
#else
#define PYTHAG(a, b) pythag_recipes(a, b)
#endif

/* ==== PROTOTYPES ==== */

static int svdcmp(double **a, int m, int n, double *w, double **v);

static void c_2_FORTRAN(Svdm_ *Svdm);
static void FORTRAN_2_c(Svdm_ *Svdm);

static double pythag_recipes(double a, double b);
static double max(double a, double b);
static double sign(double a, double b);

/* ==== FUNCTIONS ==== */

/* 
 * svd_setup(Svdm, Row, Col): 
 * allocates space for the members of *Svdm.
 * The matrix to be decomposed is (Row x Col), 
 * where Row>=Col. If Row<Col, then the rows in U will be padded
 * to get a Col x Col matrix.
 * Return value: the actual rows in Svdm->U, or -1 in case of trouble.
 */
int svd_setup(Svdm_ *Svdm, unsigned int Row, unsigned int Col)
{
    if (!Row || !Col)	/* para */
    {
	fprintf(stderr, "? svd_setup(Svdm, %u, %u): Bad dimensions\n", 
		Row, Col);
	return -1;
    }
    
    /* 
     * If Row<Col, then u is padded with Col-Row zeroed rows 
     * to make it a Col x Col matrix when later A->u is copied
     * and a warning is printed. 
     */
    if (Row<Col)
    {
	fprintf(stderr, "? svd_setup(Svdm, %u, %u): rows padded\n", 
	    Row, Col);
	Row=Col; 
    }
    
    /* alloc */
    init_matrect(&(Svdm->U), Row, Col);
    init_matrect(&(Svdm->V), Col, Col);
    Svdm->W=(double *) calloc(Col, sizeof(double));
    Svdm->Row=Row; Svdm->Col=Col;
    return Row;    /* actual no. of rows, may be padded */
}
/* END of svd_setup(Svdm, Row, Col) */

/*
 * svd_direct(Svdm):
 * Performs the SVD of a matrix that is already stored in Svdm->U.
 * Use this routine if you do not want to preserve the SVD-d matrix.
 * Return value: -1 on error, 0 if OK.
 */
int svd_direct(Svdm_ *Svdm)
{
    int Err=0;
    
    c_2_FORTRAN(Svdm);
    Err=svdcmp(Svdm->U.Mat, Svdm->Row, Svdm->Col, Svdm->W, Svdm->V.Mat);
    FORTRAN_2_c(Svdm);
    return Err;
}
/* END of svd_direct(Svdm) */
    
/*
 * svd_copy(A, Svdm):
 * copies the matrix A into *Svdm and performs the SVD.
 * *Svdm will be resized if necessary.
 * Use this routine if you wish to preserve A.
 * Return value: -1 on error, 0 if OK.
 */
int svd_copy(const Matrect_ *A, Svdm_ *Svdm)
{
    int Err=0;
    
    copy_matrect(A, &(Svdm->U)); /* copy and resize */
    if (A->Rno<A->Cno)
    {
	/* zero padding necessary */
	resize_matrect(&(Svdm->U), A->Cno, A->Cno);
    }
    if (A->Cno!=Svdm->Col)
    {
	resize_matrect(&(Svdm->V), A->Cno, A->Cno);
	Svdm->W=(double *) realloc(Svdm->W, A->Cno*sizeof(double));
    }
    Svdm->Row=Svdm->U.Rno;
    Svdm->Col=Svdm->U.Cno;
    
    Err=svd_direct(Svdm);
    return Err;
}
/* END of svd_copy(A, Svdm) */

/* 
 * rank_cond(Svdm, Eps, Cond): 
 * checks the singular values Svdm->W[] after the decomposition.
 * If Cond!=NULL, then the condition number 
 * (ratio of the largest and smallest singular value) is
 * calculated. The singular values which are smaller than
 * Eps times the largest are set to 0.0.
 * Return value: the rank of the matrix.
 */
unsigned int rank_cond(Svdm_ *Svdm, double Eps, double *Cond)
{
    double *W=Svdm->W;	/* shorthand */
    double Wmax=-HUGE_VAL, Wmin=HUGE_VAL;
    unsigned int i, N=Svdm->Col, Rank;
    
    /* get the largest and smallest singular value */
    for (i=0; i<N; i++)
    {
	if (W[i]>Wmax) Wmax=W[i];
	if (W[i]<Wmin) Wmin=W[i];
    }
    
    /* calc the condition number: set to HUGE_VAL if Wmin==0.0 */
    if (Cond!=NULL)
	*Cond=(Wmin==0.0)? HUGE_VAL: Wmax/Wmin;
    
    /* set all singular values which are smaller than Eps*Wmax
     * to zero: this is the conditioning. Calc the rank
     */
    Wmax*=fabs(Eps); Rank=N;
    for (i=0; i<N; i++)
	if (W[i]<Wmax) 
	{
	    W[i]=0.0; Rank--;
	}
    return Rank;
}
/* END of rank_cond(Svdm, Eps, Cond) */

/* 
 * svd_solve(Svdm, B, X): 
 * back-substitution routine for solving linear equations
 * AX=B. Svdm contains the SVD of the matrix A
 * and the weight vector should be "conditioned" (small entries
 * zeroed) by rank_cond(..) prior to the call to this routine.
 * Rhs[] is the right-hand-side vector assumed to be Svdm->Row long (not checked).
 * The result is returned in X[] (Col long).
 */
void svd_solve(const Svdm_ *Svdm, const double Rhs[], double X[])
{
    register double *Tmp=NULL;
    register double Sum;
    register unsigned int i, j;
    
    /* get WU'*B first */
    Tmp=(double *) calloc(Svdm->Col, sizeof(double));	/* calloc() zeroing is essential here */
    for (j=0; j<Svdm->Col; j++)
	if (Svdm->W[j]!=0.0)	/* skip zeroed */
	{
	    Sum=0.0;
	    for (i=0; i<Svdm->Row; i++) Sum+=Svdm->U.Mat[i][j]*Rhs[i];
	    Tmp[j]=Sum/Svdm->W[j];
	}
    
    /* multiply Tmp by V to get solution vector */
    for (i=0; i<Svdm->Col; i++)
    {
	Sum=0.0;
	for (j=0; j<Svdm->Col; j++) Sum+=Svdm->V.Mat[i][j]*Tmp[j];
	X[i]=Sum;
    }
    free(Tmp);
}
/* END of svd_solve(Svdm, Rhs, X) */

/* free_svd: cleans up the space allocated to the *Svdm object. */
void free_svd(Svdm_ *Svdm)
{
    free_matrect(&(Svdm->U));
    free_matrect(&(Svdm->V));
    free(Svdm->W);
}
/* END of free_svd(Svdm) */

/* ---- SVD CORE (c) Numerical Recipes ---- */

/* svdcmp: the singular value decomposition routine with the
 * slight modifications described above. A couple of comments
 * are also added. Note that this routine still has 1..N indexing
 * and is called only by svd_comp() above that takes care of
 * the pointer shifts. The matrix "a" is destroyed by the routine.
 * It is an m x n matrix, m>=n, and this condition is checked
 * by the wrapper routine before the call.
 * The routine below is protected by the copyright held by the
 * authors of Numerical Recipes even though I do not believe in
 * this. Nevertheless, thanks are due to them.
 */
static int svdcmp(double **a, int m, int n, double *w, double **v)
{
    	static const unsigned int MAX_ITER=30;
	
	/* double precision throughout */
	register int flag,i,its,j,jj,k,l,nm;
	register double c,f,h,s,x,y,z;
	register double anorm=0.0,g=0.0,scale=0.0;
	double *rv1;

	/* if (m<n) check removed: done by wrapper svd_comp() */
	/* rv1 allocation: index shifting done here */
	rv1=(double *) calloc(n, sizeof(double));
	rv1--;
	
	/* Householder reduction to bidiagonal form */
	for (i=1;i<=n;i++) {
		l=i+1;
		rv1[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m) {
			for (k=i;k<=m;k++) scale += fabs(a[k][i]);
			if (scale) {
				for (k=i;k<=m;k++) {
					a[k][i] /= scale;
					s += a[k][i]*a[k][i];
				}
				f=a[i][i];
				g = -sign(sqrt(s),f);
				h=f*g-s;
				a[i][i]=f-g;
				if (i != n) {
					for (j=l;j<=n;j++) {
						for (s=0.0,k=i;k<=m;k++) s += a[k][i]*a[k][j];
						f=s/h;
						for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
					}
				}
				for (k=i;k<=m;k++) a[k][i] *= scale;
			}
		}
		w[i]=scale*g;
		g=s=scale=0.0;
		if (i <= m && i != n) {
			for (k=l;k<=n;k++) scale += fabs(a[i][k]);
			if (scale) {
				for (k=l;k<=n;k++) {
					a[i][k] /= scale;
					s += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = -sign(sqrt(s),f);
				h=f*g-s;
				a[i][l]=f-g;
				for (k=l;k<=n;k++) rv1[k]=a[i][k]/h;
				if (i != m) {
					for (j=l;j<=m;j++) {
						for (s=0.0,k=l;k<=n;k++) s += a[j][k]*a[i][k];
						for (k=l;k<=n;k++) a[j][k] += s*rv1[k];
					}
				}
				for (k=l;k<=n;k++) a[i][k] *= scale;
			}
		}
		anorm=max(anorm,(fabs(w[i])+fabs(rv1[i])));
	}
	
	/* Accumulation of right-hand transformations */
	for (i=n;i>=1;i--) {
		if (i < n) {
			if (g) {
				for (j=l;j<=n;j++)
					v[j][i]=(a[i][j]/a[i][l])/g;
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=n;k++) s += a[i][k]*v[k][j];
					for (k=l;k<=n;k++) v[k][j] += s*v[k][i];
				}
			}
			for (j=l;j<=n;j++) v[i][j]=v[j][i]=0.0;
		}
		v[i][i]=1.0;
		g=rv1[i];
		l=i;
	}
	
	/* Accumulation of left-hand transformations */
	for (i=n;i>=1;i--) {
		l=i+1;
		g=w[i];
		if (i < n)
			for (j=l;j<=n;j++) a[i][j]=0.0;
		if (g) {
			g=1.0/g;
			if (i != n) {
				for (j=l;j<=n;j++) {
					for (s=0.0,k=l;k<=m;k++) s += a[k][i]*a[k][j];
					f=(s/a[i][i])*g;
					for (k=i;k<=m;k++) a[k][j] += f*a[k][i];
				}
			}
			for (j=i;j<=m;j++) a[j][i] *= g;
		} else {
			for (j=i;j<=m;j++) a[j][i]=0.0;
		}
		++a[i][i];
	}
	
	/* Diagonalisation of the bidiagonal form */
	for (k=n;k>=1;k--) {
		for (its=1;its<=30;its++) {
			flag=1;
			for (l=k;l>=1;l--) {
				nm=l-1;
				if (fabs(rv1[l])+anorm == anorm) {
					flag=0;
					break;
				}
				if (fabs(w[nm])+anorm == anorm) break;
			}
			if (flag) {
				c=0.0;
				s=1.0;
				for (i=l;i<=k;i++) {
					f=s*rv1[i];
					if (fabs(f)+anorm != anorm) {
						g=w[i];
						h=PYTHAG(f,g);
						w[i]=h;
						h=1.0/h;
						c=g*h;
						s=(-f*h);
						for (j=1;j<=m;j++) {
							y=a[j][nm];
							z=a[j][i];
							a[j][nm]=y*c+z*s;
							a[j][i]=z*c-y*s;
						}
					}
				}
			}
			z=w[k];
			if (l == k) {
				if (z < 0.0) {
					w[k] = -z;
					for (j=1;j<=n;j++) v[j][k]=(-v[j][k]);
				}
				break;
			}
			if (its == MAX_ITER) 
			{
			    fprintf(stderr, 
				"\n? svd_comp(): Convergence limit (%u) exceeded\n", MAX_ITER);
			    return(-1);
			}
			x=w[l];
			nm=k-1;
			y=w[nm];
			g=rv1[nm];
			h=rv1[k];
			f=((y-z)*(y+z)+(g-h)*(g+h))/(2.0*h*y);
			g=PYTHAG(f,1.0);
			f=((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x;
			c=s=1.0;
			for (j=l;j<=nm;j++) {
				i=j+1;
				g=rv1[i];
				y=w[i];
				h=s*g;
				g=c*g;
				z=PYTHAG(f,h);
				rv1[j]=z;
				c=f/z;
				s=h/z;
				f=x*c+g*s;
				g=g*c-x*s;
				h=y*s;
				y=y*c;
				for (jj=1;jj<=n;jj++) {
					x=v[jj][j];
					z=v[jj][i];
					v[jj][j]=x*c+z*s;
					v[jj][i]=z*c-x*s;
				}
				z=PYTHAG(f,h);
				w[j]=z;
				if (z) {
					z=1.0/z;
					c=f*z;
					s=h*z;
				}
				f=(c*g)+(s*y);
				x=(c*y)-(s*g);
				for (jj=1;jj<=m;jj++) {
					y=a[jj][j];
					z=a[jj][i];
					a[jj][j]=y*c+z*s;
					a[jj][i]=z*c-y*s;
				}
			}
			rv1[l]=0.0;
			rv1[k]=f;
			w[k]=x;
		}
	}
	rv1++; free(rv1);   /* instead of free_vector() */
	return(0);	/* OK */
}
/* END of svdcmp */

/* ---- Auxiliaries ---- */

/*
 * c_2_FORTRAN(Svdm), FORTRAN_2_c(Svdm):
 * These routines convert between C-style [0..N-1] and FORTRAN-style [1..N] indexing.
 * Easier than modifying the core svdcmp() Num. Recipes routine that is
 * a perfect example of a C code written in FORTRAN... :-)
 */
static void c_2_FORTRAN(Svdm_ *Svdm)
{
    register unsigned int i;
    
    for (i=0; i<Svdm->Row; ++i)
	Svdm->U.Mat[i]--;
    Svdm->U.Mat--;
    for (i=0; i<Svdm->Col; ++i)
	Svdm->V.Mat[i]--;
    Svdm->V.Mat--;
    Svdm->W--;
}

static void FORTRAN_2_c(Svdm_ *Svdm)
{
    register unsigned int i;
    
    Svdm->U.Mat++;
    for (i=0; i<Svdm->Row; ++i)
	Svdm->U.Mat[i]++;
    Svdm->V.Mat++;
    for (i=0; i<Svdm->Col; ++i)
	Svdm->V.Mat[i]++;
    Svdm->W++;
}

/* pythag_recipes(a, b): 
 * returns the value sqrt(a^2+b^2) without over-
 * or underflows. On SGI machines when the extended ANSI switch
 * -xansi is defined, then the BSD4.3-style math library contains
 * the function hypot(x, y) which does the same job and will be used
 * whenever possible. The macro PYTHAG is defined accordingly
 * (see the Definitions section above): if hypot() is not available, 
 * then the routine below will be substituted.
 */
static double pythag_recipes(double a, double b)
{
    register double at=fabs(a), bt=fabs(b), ct;
    
    if (at<DBL_EPSILON) return bt;
    if (bt<DBL_EPSILON) return at;
    if (at>bt)
    {
        ct=bt/at; return(at*sqrt(1.0+ct*ct));
    }
    else
    {
        ct=at/bt; return(bt*sqrt(1.0+ct*ct));
    }
}
/* END of pythag_recipes(a, b) */

/* The following two functions hardly deserve any comments */
static double max(double a, double b)
{
    return((a>b)? a: b);
}

static double sign(double a, double b)
{
    return((b>=0.0)? fabs(a): -fabs(a));
}

#undef PYTHAG
#undef MAX_ITER

/* ==== END OF FUNCTIONS svd.c ==== */
