/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SVD_HEADER
#define SVD_HEADER

/* ==== HEADER svd.h ==== */

/**
 * \file svd.h
 * Singular Value Decomposition, based on the Numerical Recipes
 * routine 'svdcmp'. Rewritten from my 1994 implementation.
 * Supports setup, SVD, rank conditioning, linear equation solving.
 * 
 * Reference:- Press, W.H., Flannery, B.P., Teukolsky, S.A. and Vetterling, W.T.:
 * Numerical Recipes in C: The Art of Scientific Computing.
 * Cambridge University Press, 1992.
 */

/* 23-Aug-2000. Andras Aszodi. */

/* ==== STANDARD HEADERS ==== */

#include <stdlib.h>
#include <float.h>

/* ==== MODULE HEADERS ==== */

#include "matrect.h"

/* ==== DEFINITIONS ==== */

#ifdef DBL_EPSILON
#define SVD_EPSILON (10.0*DBL_EPSILON)
#else
#define SVD_EPSILON (1.0e-10)
#endif

/* ===== TYPEDEFS ==== */

/** Svdm_: contains the SVD decomposition matrices and the dimensions. */
typedef struct
{
    Matrect_ U, V;  /**< U is Row x Col, V is Col x Col, Row>=Col */
    double *W;	    /**< W has Col entries */
    unsigned int Row, Col;  /**< the sizes of U */
} Svdm_;

/* ==== PROTOTYPES ==== */

#ifdef __cplusplus
extern "C" {
#endif

/**
 * svd_setup(Svdm, Row, Col): 
 * allocates space for the members of *Svdm.
 * The matrix to be decomposed is (Row x Col), 
 * where Row>=Col. If Row<Col, then the rows in U will be padded
 * to get a Col x Col matrix.
 * Return value: the actual rows in Svdm->U, or -1 in case of trouble.
 */
int svd_setup(Svdm_ *Svdm, unsigned int Row, unsigned int Col);

/**
 * svd_direct(Svdm):
 * Performs the SVD of a matrix that is already stored in Svdm->U.
 * Use this routine if you do not want to preserve the SVD-d matrix.
 * Return value: -1 on error, 0 if OK.
 */
int svd_direct(Svdm_ *Svdm);

/**
 * svd_copy(A, Svdm):
 * copies the matrix A into *Svdm and performs the SVD.
 * *Svdm will be resized if necessary.
 * Use this routine if you wish to preserve A.
 * Return value: -1 on error, 0 if OK.
 */
int svd_copy(const Matrect_ *A, Svdm_ *Svdm);

/**
 * rank_cond(Svdm, Eps, Cond): 
 * checks the singular values Svdm->W[] after the decomposition.
 * If Cond!=NULL, then the condition number 
 * (ratio of the largest and smallest singular value) is
 * calculated. The singular values which are smaller than
 * Eps times the largest are set to 0.0.
 * Return value: the rank of the matrix.
 */
unsigned int rank_cond(Svdm_ *Svdm, double Eps, double *Cond);

/**
 * svd_solve(Svdm, B, X): 
 * back-substitution routine for solving linear equations
 * AX=B. Svdm contains the SVD of the matrix A
 * and the weight vector should be "conditioned" (small entries
 * zeroed) by rank_cond(..) prior to the call to this routine.
 * Rhs[] is the right-hand-side vector assumed to be Svdm->Row long (not checked).
 * The result is returned in X[] (Col long).
 */
void svd_solve(const Svdm_ *Svdm, const double Rhs[], double X[]);

/* free_svd: cleans up the space allocated to the *Svdm object. */
void free_svd(Svdm_ *Svdm);

#ifdef __cplusplus
}
#endif

/* ==== END OF HEADER svd.h ==== */

#endif	/* SVD_HEADER */
