/* ==== FUNCTIONS vect.c ==== */

/* Double-precision vectors. */

/* 9-Nov-2000. Andras Aszodi */

/* ==== STANDARD HEADERS ==== */

#include <math.h>
#include <float.h>
#include <string.h>

/* ==== MODULE HEADERS ==== */

#include "vect.h"

/* ==== PROTOTYPES ==== */

static int dim_mismatch(const char *Funcname, const Vect_ *V1, const Vect_ *V2);

/* ==== FUNCTIONS ==== */

/* ---- init, resize, destroy ---- */

/*
 * init_vect(V, L):
 * inits *V to contain L items. If L==0, then L=3 is used.
 * This must be the first call on *V before any other operation is done.
 */
void init_vect(Vect_ *V, unsigned int L)
{
    if (V==NULL)
    {
	fprintf(stderr, "\n? init_vect(NULL, %u)\n", L);
	return;
    }
    if (!L) L=3;
    V->Vec=(double *) calloc(L, sizeof(double));
    V->Len=V->Physlen=L;
}
/* END OF init_vect(V, L) */

/*
 * resize_vect(V, Newsize):
 * changes *V so that its logical length is now Newsize.
 * If Newsize==0, then Newsize=3 will be used.
 * If *V was empty, then it will be re-initialised.
 * Does nothing if Newsize==V->Len.
 */
void resize_vect(Vect_ *V, unsigned int Newsize)
{
    register unsigned int i;
    unsigned int Newphyslen;
    
    if (V==NULL || Newsize==V->Len) return;    /* no change */
    if (!Newsize) Newsize=3;
    if (V->Vec==NULL)
    {
	init_vect(V, Newsize);
	return;
    }
    if (Newsize>V->Physlen) 	/* realloc necessary */
    {
	Newphyslen=(Newsize>2*V->Physlen)? Newsize: 2*V->Physlen;
	V->Vec=(double *) realloc(V->Vec, Newphyslen*sizeof(double));
	V->Physlen=Newphyslen;
    }
    
    /* zero new elements if vector grows */
    for (i=V->Len; i<Newsize; ++i)
	V->Vec[i]=0.0;
    V->Len=Newsize;
}
/* resize_vect(V, Newsize) */

/*
 * copy_vect(From, To):
 * copies the contents of *From to *To.
 * *To is resized if necessary.
 */
void copy_vect(const Vect_ *From, Vect_ *To)
{
    /* various pathological cases */
    if (From==To) return;
    if (From==NULL || To==NULL) return;
    
    if (From->Len>To->Physlen)
	resize_vect(To, From->Len); /* make room if necessary */
    memcpy(To->Vec, From->Vec, From->Len*sizeof(double));
    To->Len=From->Len;
}
/* END of copy_vect(From, To) */

/*
 * free_vect(V):
 * frees up all storage associated with *V.
 * Does not deallocate V itself.
 */
void free_vect(Vect_ *V)
{
    if (V==NULL) return;
    free(V->Vec); V->Vec=NULL;
    V->Len=V->Physlen=0;
}
/* END of free_vect(V) */

/*
 * vect_scalar(V, Opcode, Scalar):
 * Performs a simple arithmetic operation in place between the elements of *V
 * and Scalar, depending on Opcode. 
 * Supported operations:-
 * 	'=': set all elements of *V to Scalar.
 * 	'+': adds Scalar to all elements of *V.
 * 	'-': subtracts Scalar from all elements of *V.
 * 	'*': multiplies all elements of *V by Scalar.
 * 	'/': divides all elements of *V by Scalar. Warns if Scalar == 0.
 * Prints a warning if the Opcode is different from these listed above.
 * Returns an error code: 0 if div by zero or op not supported, the length otherwise.
 */
int vect_scalar(Vect_ *V, char Opcode, double Scalar)
{
    register unsigned int i, Len;
    double Invscalar;
    
    Len=V->Len;
    switch (Opcode)
    {
	case '=':
	    for (i=0; i<Len; ++i)
		V->Vec[i]=Scalar;
	    break;
	case '+':
	    for (i=0; i<Len; ++i)
		V->Vec[i]+=Scalar;
	    break;
	case '-':
	    for (i=0; i<Len; ++i)
		V->Vec[i]-=Scalar;
	    break;
	case '*':
	    for (i=0; i<Len; ++i)
		V->Vec[i]*=Scalar;
	    break;
	case '/':
	    if (Scalar==0.0)
	    {
		fputs("\n? vect_scalar(): Div by zero\n", stderr);
		return 0;
	    }
	    Invscalar=1.0/Scalar;
	    for (i=0; i<Len; ++i)
		V->Vec[i]*=Invscalar;
	    break;
	default:
	    fprintf(stderr, 
		"\n? vect_scalar(...,\'%c\',...): Operation not implemented\n",
		Opcode);
	    return 0;
	break;
    }
    return Len;
}
/* END of vect_scalar(V, Opcode, Scalar) */

/*
 * vect_inplace(V, Opcode, R):
 * Performs binary in-place operations on R and V and puts the result in R.
 * Supported operations:-
 * 	'+': performs R[i]+=V[i], for all i
 * 	'-': performs R[i]-=V[i], for all i
 * 	'*': performs R[i]*=V[i], for all i
 * V and R must have the same logical length. If not, a warning is printed
 * and 0 is returned.
 * Return value: the common length on success.
 */
int vect_inplace(const Vect_ *V, char Opcode, Vect_ *R)
{
    register unsigned int i, Len;
    
    if (dim_mismatch("vect_inplace()", V, R)) return 0;
    Len=V->Len;
    switch (Opcode)
    {
	case '+':
	    for (i=0; i<Len; ++i)
		R->Vec[i]+=V->Vec[i];
	    break;
	case '-':
	    for (i=0; i<Len; ++i)
		R->Vec[i]-=V->Vec[i];
	    break;
	case '*':
	    for (i=0; i<Len; ++i)
		R->Vec[i]*=V->Vec[i];
	    break;
	default:
	    fprintf(stderr, 
		"\n? vect_inplace(...,\'%c\',...): Operation not implemented\n",
		Opcode);
	    return 0;
	break;
    }
    return Len;
}
/* END of vect_inplace(V, Opcode, R) */

/* 
 * vect_vect(V1, Opcode, V2, R):
 * Performs binary operations on V1 and V2 and puts the result in R.
 * Supported operations:-
 * 	'+': performs R[i]=V1[i]+V2[i], for all i
 * 	'-': performs R[i]=V1[i]-V2[i], for all i
 * 	'*': performs R[i]=V1[i]*V2[i], for all i
 * V1, V2 and R must have the same logical length. If not, a warning is printed
 * and 0 is returned.
 * Return value: the common length on success.
 */
int vect_vect(const Vect_ *V1, char Opcode, const Vect_ *V2, Vect_ *R)
{
    register unsigned int i, Len;
    
    if (dim_mismatch("vect_vect()", V1, V2)) return 0;
    if (dim_mismatch("vect_vect()", V2, R)) return 0;
    Len=V1->Len;
    switch (Opcode)
    {
	case '+':
	    for (i=0; i<Len; ++i)
		R->Vec[i]=V1->Vec[i]+V2->Vec[i];
	    break;
	case '-':
	    for (i=0; i<Len; ++i)
		R->Vec[i]=V1->Vec[i]-V2->Vec[i];
	    break;
	case '*':
	    for (i=0; i<Len; ++i)
		R->Vec[i]=V1->Vec[i]*V2->Vec[i];
	    break;
	default:
	    fprintf(stderr, 
		"\n? vect_vect(...,\'%c\',...): Operation not implemented\n",
		Opcode);
	    return 0;
	break;
    }
    return Len;
}
/* END of vect_vect(V1, Opcode, V2, R) */

/*
 * scalar_product(V1, V2):
 * returns the scalar product of V1 and V2.
 * Returns 0.0 if either of the vectors is empty or the dimensions don't match.
 */
double scalar_product(const Vect_ *V1, const Vect_ *V2)
{
    register unsigned int i;
    register double Scalprod=0.0;
    
    if (dim_mismatch("scalar_product()", V1, V2)) return 0.0;
    
    for (i=0; i<V1->Len; ++i)
	Scalprod+=(V1->Vec[i]*V2->Vec[i]);
    return Scalprod;
}
/* END of scalar_product(V1, V2) */

/*
 * vect_product(V1, V2, P):
 * Calculates the vectorial (cross) product of V1 and V2
 * and puts the result in P.
 * V1 and V2 must be 3-dimensional, if not, then 
 * P will be unchanged and 0 is returned.
 * The dimension of P will be set to 3.
 * Returns 1 on success.
 */
int vect_product(const Vect_ *V1, const Vect_ *V2, Vect_ *P)
{
    if (V1->Len!=3 || V2->Len!=3)
    {
	fputs("\n? vector_product(): Operands must be 3D\n", stderr);
	return 0;
    }
    resize_vect(P, 3);	/* force output to 3D */
    
    P->Vec[0]=(V1->Vec[1])*(V2->Vec[2])-(V1->Vec[2])*(V2->Vec[1]);
    P->Vec[1]=(V1->Vec[2])*(V2->Vec[0])-(V1->Vec[0])*(V2->Vec[2]);
    P->Vec[2]=(V1->Vec[0])*(V2->Vec[1])-(V1->Vec[1])*(V2->Vec[0]);
    return 1;
}
/* END of vect_product(V1, V2, P) */

/*
 * vect_len2(V):
 * Returns the squared length of V or 0.0 if it was empty.
 */
double vect_len2(const Vect_ *V)
{
    register unsigned int i;
    register double Len2=0.0, Temp;
    
    for (i=0; i<V->Len; ++i)
    {
	Temp=V->Vec[i];
	Len2+=Temp*Temp;
    }
    return Len2;
}
/* END of vect_len2(V) */

/*
 * vect_len(V):
 * returns the length of the vector V, or 0.0 if empty.
 */
double vect_len(const Vect_ *V)
{
    return sqrt(vect_len2(V));
}
/* END of vect_len(V) */

/*
 * vect_norm(V):
 * Normalises V to unit length. Does nothing to null-vectors.
 * Returns the original length.
 */
double vect_norm(Vect_ *V)
{
    register double Len=vect_len(V);
    if (Len!=0.0)
    	vect_scalar(V, '/', Len);
    return Len;
}
/* END of vect_norm(V) */

/*
 * vect_norm_eps(V, Eps):
 * Calculates the length of V. If that is smaller than |Eps|,
 * then V is assumed to be a null-vector, its coordinates will be set to 0.0
 * and 0.0 is returned as its previous length.
 * Otherwise, V is normalised just like in vect_norm(V).
 */
double vect_norm_eps(Vect_ *V, double Eps)
{
    register double Len=vect_len(V);
    if (Len<fabs(Eps))
    {
	vect_scalar(V, '=', 0.0); /* force to nullvector */
	Len=0.0;
    }
    else
	vect_scalar(V, '/', Len);
    return Len;
}
/* END of vect_norm_eps(V, Eps) */

/* ---- Input/output ---- */

#define LINEWIDTH 80

/*
 * read_vect(In, V):
 * reads a vector from the file In into V.
 * Returns the number of errors encountered during parsing (0 if OK).
 * If there were errors, V remains unchanged.
 */
int read_vect(FILE *In, Vect_ *V)
{
    char Line[LINEWIDTH+1], *Eolnptr;
    unsigned int Len=0, Dashlen, Idx;
    Vect_ Temp;
    double Coord;
    short Topseen=0, Botseen=0;
    int Error=0;
    
    init_vect(&Temp, 0);
    while (memset(Line, 0, LINEWIDTH+1),NULL!=fgets(Line, LINEWIDTH, In))
    {
	/* skip empty lines and comments */
	if (Line[0]=='\0' || Line[0]=='\n' || Line[0]=='#')
	    continue;
	
	/* throw away terminating newline */
	Eolnptr=strrchr(Line, '\n');
	if (Eolnptr!=NULL) *Eolnptr='\0';
	
	if (!Topseen)	    /* look for top delimiter "----" */
	{
	    Dashlen=strspn(Line, "-");
	    Topseen=(Dashlen==strlen(Line));	/* found top delimiter */
	    continue;
	}
	
	/* check for bottom delimiter */
	Dashlen=strspn(Line, "=");
	if (Dashlen==strlen(Line))
	{
	    if (!Len) ++Error;	/* no coords btw "---" and "===" */
	    Botseen=1;
	    break;	/* we are done */
	}
	
	/* read a coordinate now */
	if (2!=sscanf(Line, "%u | %lf", &Idx, &Coord))
	{
	    fprintf(stderr, "\n? read_vect(): Cannot parse line \"%s\"\n",
		    Line);
	    ++Error;
	    continue;
	}
	if (Idx!=Len)
	{
	    fprintf(stderr, "\n? read_vect(): Coord [%u] expected, [%u] found\n",
		    Len, Idx);
	    ++Error;
	    continue;
	}
	
	/* make room if necessary and store */
	resize_vect(&Temp, Len+1);
	Temp.Vec[Len]=Coord;
	++Len;
    }	    /* while */
    
    /* handle errors */
    if (!Topseen || !Botseen)	/* could not parse */
	++Error;
    if (Error) return Error;
    
    /* transfer to V */
    free_vect(V);	/* removes V's storage */
    *V=Temp;     /* memberwise shallow copy, OK here */
    return 0;
}
/* END of read_vect(In, V) */

/*
 * write_vect(Out, V, Prec):
 * writes the vector V to the file Out using scientific notation,
 * with Prec figures after the decimal point.
 */
void write_vect(FILE *Out, const Vect_ *V, unsigned int Prec)
{
    char Rowformat[LINEWIDTH+1];
    unsigned int Nwidth, Width, Delimwidth, k;
    
    if (!Prec) Prec=6;
    
    Nwidth=(int)(ceil(log10((double)V->Len)));	/* index print width */
    Width=Prec+7;   /* coord print width, scientific format */
    Delimwidth=Nwidth+3+Width;

    /* for say 20 dimensions and Prec=3, a coord line would look like
     * "20 | 1.234e+00"
     * that needs a format string (Rowformat) like
     * "%2d | %.3e\n"
     * which is made by the magical sprintf here.
     */
    sprintf(Rowformat, "%%%dd | %%.%de\n", Nwidth, Prec);
    
    /* print top delimiter: "-----" */
    for (k=0; k<Delimwidth; ++k)
	putc('-', Out);
    putc('\n', Out);
    
    /* print the coordinates */
    for (k=0; k<V->Len; ++k)
	fprintf(Out, Rowformat, k, V->Vec[k]);
    
    /* print bottom delimiter: "=====" */
    for (k=0; k<Delimwidth; ++k)
	putc('=', Out);
    putc('\n', Out);
}
/* END of write_vect(Out, V, Prec) */
    
#undef LINEWIDTH
    
/* ---- Auxiliaries ---- */

/*
 * dim_mismatch(Funcname, V1, V2):
 * If V1->Len!=V2->Len, then an error message is printed to stderr
 * where Funcname points to a string containing the function name
 * where the problem occurred, and 1 is returned.
 * If there was no dimension mismatch, then 0 is returned.
 */
static int dim_mismatch(const char *Funcname, const Vect_ *V1, const Vect_ *V2)
{
    unsigned int L1=V1->Len, L2=V2->Len;
    
    if (L1!=L2)
    {
	fprintf(stderr, "\n? %s: Dim mismatch (%d!=%d)\n",
		Funcname, L1, L2);
	return 1;
    }
    else return 0;
}
/* END of dim_mismatch(Funcname, V1, V2) */

/* ==== END OF FUNCTIONS vect.c ==== */
