/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef BRON_KERBOSCH_HH
#define BRON_KERBOSCH_HH

#include <boost/graph/adjacency_list.hpp>

#include "cluster_base.hh"

/**
 * \brief this class performs a search for connected components
 *
 * By calling the operator() of the class the cliques of a graph are searched.
 * These cliques can be retrieved later, one by one, calling the get_cluster().
 * The number of found cliques is retrieved by the operator() and the 
 * num_clusters() method.
 *
 * The algorithm has been implemented according to [1] version 2. The main
 * difference is the implementation of the adjacency information: <br>
 * The algorithm in the paper uses a boolean matrix as a representation of the 
 * connectivity information, whilst this implementation uses direct access to the
 * graph's edge collection (is_connected()).
 *
 * 
 * \b Literature:
 * 
 * [1] Coen Bron and Joep Kerbosch,
 * Algorithm 457: Finding All Cliques of an Undirected Graph [H]
 * Comm. ACM 16, 9 (Sept. 1973), 575-577

 */
template<class graph_t>
class bron_kerbosch: public cluster_algorithm<graph_t>
{
private:
	
	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief checkes if two vertices are connected
	 *
	 * \return true if the vertex \p a and the vertex \p b are connected by an 
	 *         edge, false otherwise.
	 *
	 * \param a,b vertices of interest
	 * \param graph graph structures containing the vertices \p a and \p b.
	 */
	bool is_connected(vertex_t a, vertex_t b, const graph_t& graph);

	/**
	 * \brief recursive procedure of the Bron Kerbosch algorithm
	 *
	 * This method is the main part of the Bron- Kerbosch clique detection 
	 * algorithm.
	 */
	void extend(const graph_t& graph, cluster_t& complete_subgraph,
							vector<vertex_t> old_set, int ne, int ce);

public:
	
	//*****************************************************************************
	// CONSTRUCTORS & DESTRUCTORS
	//*****************************************************************************

	/**
	 * \brief standard constructor initializing a ready-to-use instance of the 
	 * bron_kerbosch class
	 */
	bron_kerbosch(): cluster_algorithm<graph_t>()	{ }
	
	/**
	 * \brief function operator performing the Bron- Kerbosch clique detection
	 *
	 * The function performs the clique detection algorithm of Bron- Kerbosch.
	 *
	 * \return the number of detected cliques.
	 */
	virtual int operator()(graph_t& graph);
	
};

#ifdef INCLUDE_TMPL_DEFS
  #include "bron_kerbosch_templ.cc"
#endif



#endif //CONNECTED_COMPS_HH
