/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef BRON_KERBOSCH_TEMPL_CC
#define BRON_KERBOSCH_TEMPL_CC

//*******************************************************************************
template<class graph_t>
bool bron_kerbosch<graph_t>::is_connected(vertex_t a, vertex_t b, 
																					const graph_t& graph)
//*******************************************************************************
{
	if (a == b) 
		return true;

	if ((edge(a,b,graph)).second) 
		return true;

	return false;
}


//*******************************************************************************
template <class graph_t>
void bron_kerbosch<graph_t>::extend(const graph_t& graph,
																		cluster_t& complete_subgraph,
																		vector<vertex_t> old_set, int ne, int ce)
//*******************************************************************************
{
	vertex_t fixp = 0;
	int s;
	int minnod = ce;
	int nod = 0;
    
	// Determine each counter value and look for minimum
	for (int i = 0; i < ce && minnod != 0; i++) 
	{
		vertex_t p = old_set[i];
		int count = 0;
		int pos;

		// Count disconnections of the vertices
		for (int j = ne; j < ce && count < minnod; j++) 
		{
			if (!is_connected(p,old_set[j],graph)) 
			{
				count++;
				pos = j;                          // Save position of potential candidate
			}
		}

		// Test new minimum
		if (count < minnod) 
		{
			minnod = count;
			fixp = p;
            
			if (i < ne) 
				s = pos;
			else 
			{
				s = i;
				// If fixed point initially chosen from 'candidates' then
				// number of disconnections will be preincreased by one
				nod = 1;
			}
		}
	}

#ifdef __DEBUGGING
	cout << "---------------------" << endl;
	cout << "s      = " << s << endl;
	cout << "fixp   = " << fixp << endl;
	cout << "nod    = " << nod << endl;
	cout << "minnod = " << minnod << endl;
	cout << "ne = " << ne << endl;
	cout << "ce = " << ce << endl;
	cout << "---------------------" << endl;
	for (vector<vertex_t>::const_iterator vi = old_set.begin(); vi != old_set.end(); ++vi) 
		cout << *vi << " ";
	cout << endl;
	cout << "---------------------" << endl;
#endif //__DEBUGGING

	// Backtrackcycle
	for (nod += minnod; nod >= 1; nod--) 
	{
		// Interchange
		vertex_t p = old_set[s];
		old_set[s] = old_set[ne];
		old_set[ne] = p;

		vertex_t sel = old_set[ne];
        
#ifdef __DEBUGGING
		for (vector<vertex_t>::const_iterator vi = old_set.begin(); vi != old_set.end(); ++vi) 
	    cout << *vi << " ";
		cout << endl;
#endif //__DEBUGGING

		// Fill new set 'not'
		vector<vertex_t> new_set; // 1..ce
		new_set.reserve(ce);
		int new_ne = 0;
		for (int i = 0; i < ne; i++) 
		{
			if (is_connected(sel,old_set[i],graph)) 
			{
				new_set.push_back(old_set[i]);
				new_ne++;
			}
		}
        
		// Fill new set 'candidates'
		int new_ce = new_ne;
		for (int i = ne + 1; i < ce; i++) 
		{
			if (is_connected(sel,old_set[i],graph)) 
			{
				new_set.push_back(old_set[i]);
				new_ce++;
			}
		}
        
		// Add to 'complete_subgraph'
		complete_subgraph.insert(sel);

#ifdef __DEBUGGING
		cout << "====================" << endl;
		cout << "Candidates: ";
		typedef vector<vertex_t>::const_iterator ITOR;
		for (ITOR vi = new_set.begin(); vi != new_set.end(); ++vi) 
	    cout << *vi << " ";

		cout << endl;
		cout << "Clique: ";
		for (clusters_cit vi = complete_subgraph.begin(); vi != complete_subgraph.end(); vi++)
	    cout << *vi << " ";

		cout << endl;
		cout << "====================" << endl;
#endif //__DEBUGGING

		if (new_ce == 0) 
		{
			// Write clique
			clusters.push_back(complete_subgraph);

#ifdef __DEBUGGING
			cout << "Clique of size " << complete_subgraph.size()
					 << " found" << endl;
			cout << "Vertices in clique:" << endl;
			for (cluster_cit vi = complete_subgraph.begin(); vi != complete_subgraph.end(); vi++) 
				cout << *vi << " ";
			cout << endl;
#endif //__DEBUGGING

		} 
		else if (new_ne < new_ce) 
			extend(graph, complete_subgraph, new_set, new_ne, new_ce);
        
		// Remove from 'complete_subgraph'
		complete_subgraph.erase(sel);
        
		// Add to 'not'
		ne++;
        
		// Select a candidate disconnected to the fixed point
		if (nod > 1) 
		{
			s = ne;
#ifdef __DEBUGGING
	    cout << "size of old_set = " << old_set.size() << endl;
#endif //__DEBUGGING
			while (is_connected(fixp,old_set[s],graph)) 
				s++;
		}
	} // end of backtrackcycle
}


//*******************************************************************************
template <class graph_t>
int bron_kerbosch<graph_t>::operator()(graph_t& graph)
//*******************************************************************************
{
	int N = num_vertices(graph);
	vector<vertex_t> all(N); // 1..N
	vertices_t vits = vertices(graph);
	copy(vits.first, vits.second, all.begin());

	cluster_t complete_subgraph; // behaves like a stack
	extend(graph, complete_subgraph, all, 0, N);

	return clusters.size();
}

#endif //BRON_KERBOSCH_TEMPL_CC
