/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CLUSTER_BASE_HH
#define CLUSTER_BASE_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
  #include <list>
#else
  #include <list.h>
#endif 



#include <boost/config.hpp>
#include <boost/graph/graph_traits.hpp>

using namespace std;
using namespace boost;

/**
 * \brief an abstract base class for graph algorithms, searching for clusters
 * in grapghs.
 */
template<class graph_t>
class cluster_algorithm
{
protected:

	//*****************************************************************************
	// TYPE DEFINITIONS (graph traits)
	//*****************************************************************************

  /**
   * \brief typedefinition of the vertex descriptor
   */
  typedef graph_traits<graph_t>::vertex_descriptor vertex_t;

  /**
   * \brief typedefinition of the edge descriptor
   */
  typedef graph_traits<graph_t>::edge_descriptor edge_t;

  /**
   * \brief iterator type for the vertices
   */
  typedef graph_traits<graph_t>::vertex_iterator vertex_it;

	/**
	 * \brief pair of vertex iterators
	 */
	typedef pair<vertex_it, vertex_it> vertices_t;

  /**
   * \brief iterator type for the edges
   */
  typedef graph_traits<graph_t>::edge_iterator edge_it;
  
  /**
   * \brief out-edge iterator type
   */
  typedef graph_traits<graph_t>::out_edge_iterator edge_oit;

  /**
   * \brief adjancency iterator
   */
  typedef graph_traits<graph_t>::adjacency_iterator vertex_ait;

	/**
	 * \brief pair of adjacency iterators
	 */
	typedef pair<vertex_ait, vertex_ait> adjacency_t;

public:
	//*****************************************************************************
	// TYPE DEFINITIONS (clusters)
	//*****************************************************************************

	//@{
	/**
	 * \brief a list of clustered vertices + iterators
	 */
	typedef set<vertex_t> cluster_t;
	typedef cluster_t::iterator cluster_it;
	typedef cluster_t::const_iterator cluster_cit;
	//@}

	//@{
	/**
	 * \brief a list of clusters + iterators
	 */
	typedef list<cluster_t> clusters_t;
	typedef clusters_t::iterator clusters_it;
	typedef clusters_t::const_iterator clusters_cit;
	//@}

protected:
	//*****************************************************************************
	// PROTECTED MEMBERS
	//*****************************************************************************
	
	/**
	 * \brief collection of cliques found in the graph
	 *
	 * This member stores the cliques of the graph passed to the operator().
	 */
	clusters_t clusters;

	/**
	 * \brief this is a speed up for the clique retrieval
	 */
	pair<int, clusters_cit> last_retrieved;

	
public:
	
	//*****************************************************************************
	// PUBLIC METHODS (abstract)
	//*****************************************************************************

	/**
	 * \brief function operator performing the clustering of \p graph
	 *
	 * The function operator should invoke the clustering and return the number of
	 * clusters found.
	 *
	 * \return number of found clusters
	 *
	 * \param graph graph structure for the clustering
	 */
	virtual int operator()(graph_t& graph) = 0;

	/**
	 * \brief should retrieve the number of detected clusters
	 */
	virtual int num_clusters() const { return clusters.size(); }
	
	/**
	 * \brief should retrieve the cluster for index \p ix into the output iterator
	 * \p ins.
	 *
	 * \return a pair of constant cluster iterators \c cluster_cit representing the 
	 *         elements of the cluster at index \p ix.
	 *
	 * \param ix index of the desired cluster (must be within 
	 *           <tt>[0 - num_clusters()-1]</tt>.
	 * \param count is a reference to a variable receiving the number of vertices in
	 *              the cluster
	 */
	virtual pair<cluster_cit, cluster_cit> get_cluster(int ix, size_t& count);
};


#ifdef INCLUDE_TMPL_DEFS
  #include "cluster_base_templ.cc"
#endif



#endif //CLUSTER_BASE_HH
