/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CONNECTED_COMPS_HH
#define CONNECTED_COMPS_HH

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/depth_first_search.hpp>

#include "cluster_base.hh"

/**
 * \brief visitor class performing the search for connected components
 *
 * This visitor class tracks the visited vertices during each single run of a 
 * depth first search and records the number of the component for each single 
 * vertex.
 *
 * A component is a subset U of the vertices V of an undirected graph G, where a
 * path exists between every pair of vertices in U.
 */
template<class vertex_t, class graph_t> 
class connected_visitor: public default_dfs_visitor
{
private:
	/**
	 * \brief internal data structure recording the component number for each 
	 * vertex
	 *
	 * The component number starts with \c 0 and goes to <tt>num_components() -
	 * 1</tt>.
	 */
	map<vertex_t, int>& visit_map;
	
	/**
	 * \brief index of the currently searched component
	 */
	int current_component;
	
public:

	/**
	 * \brief standard constructor initializing a ready-to-use visitor object
	 */
	connected_visitor(map<vertex_t, int>& compmap): visit_map(compmap), 
																									current_component(-1) { }
	
	/**
	 * \brief increments the component counter whenever the search through a new
	 * component is invoked
	 *
	 * According to the 
	 * <a href ="http://www.boost.org/libs/graph/doc/depth_first_search.html">
	 * depth first search</a> of the BOOST graph library, the depth first search
	 * is called seperately for every vertex, which has not been visited so far.
	 * The <a href ="http://www.boost.org/libs/graph/doc/DFSVisitor.html">
	 * DFS-Visitor</a> provides the start_vertex() method which is invoked,
	 * venever the search starts with a new, yet undiscovered, vertex.
	 *
	 * The method simply increments the index of the current component.
	 *
	 * \return void 
	 *
	 * \param s vertex which starts the new connected component
	 * \param g reference to the graph structure
	 */
	void start_vertex(vertex_t s, const graph_t& g) { ++current_component; }
	
	/**
	 * \brief sets the component index of the vertex \p v
	 *
	 * Whenever a vertex is discovered for the first time, the corresponding entry
	 * in the #visit_map is set to the index of the current 
	 * component.
	 *
	 * \return void
	 *
	 * \param u newly discovered vertex 
	 * \param g reference to the graph structure
	 */
	void discover_vertex(vertex_t v, const graph_t& g) 	{ visit_map[v] = current_component; }	
};


/**
 * \brief this class performs a search for connected components
 *
 * By calling the operator() of the class the search for connected components is
 * performed, which can be retrieved on by one by calling the get_cluster() 
 * method. The number of found clusters is retrieved by the operator() and the 
 * num_clusters() method.
 */
template<class graph_t>
class connected_components: public cluster_algorithm<graph_t>
{
public:

	/**
	 * \brief standard constructor initilising a ready-to-use instance of 
	 * connected components.
	 */
	connected_components(): cluster_algorithm<graph_t>() { }

	/**
	 * \brief function operator performing the clustering of \p graph
	 *
	 * The clustering is done by a search for connected components.
	 * \sa the \link connected_visitor visitor for connected components search
	 * \endlink
	 *
	 * \return number of found clusters
	 *
	 * \param graph graph structure for the clustering
	 */
	virtual int operator()(graph_t& graph);
};

#ifdef INCLUDE_TMPL_DEFS
  #include "connected_comps_templ.cc"
#endif

#endif //CONNECTED_COMPS_HH
