/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file binaryio.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 10 Feb 2004, last CVS commit 2004/02/10 14:57:21
 */

#ifndef BINARYIO_HH
#define BINARYIO_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#include <iostream>
#else
#include <iostream.h>
#endif

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 
	/**
	 * \brief binary stream class module
	 *
	 * The binary stream class module is a static class that provides method to
	 * write to and read from a binary stream in system normal or reversed byte
	 * order.
	 *
	 * The class has methods to read/write:
	 *
	 * - single values of type \c double, \c float, \c integer, \c short 
	 * - fixed sized arrays containing \c double, \c float, \c integer, \c short
	 *   values.
	 *
	 * In addition the reverseBytes() function flips a fixed sequence of bytes
	 * to obtain reverse byte order.
	 */
	class BinaryIO {
	public:

		/**
		 * \brief reverses a sequence of bytes
		 * 
		 * The sequence is manipulated in place so that the first byte is
		 * exchanged with the last, the second with the butlast and so on ...
		 * 
		 * \param seq the sequence of bytes to be reversed
		 * \param len the length of this sequence
		 */
		static void reverseBytes (char* seq, const int len);

		/**
		 * \brief reads a double value from a binary stream
		 * 
		 * \param stream  the binary stream
		 * \param value   the double value
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes read from the binary stream
		 */
		static size_t readDouble(std::istream & stream, double & value, bool reverse);
		/**
		 * \brief reads an array of double values from a binary stream
		 *
		 * \param stream  the binary stream
		 * \param value   an array of size \p len that accepts the read array
		 * \param len     the length of the array
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes read from the binary stream
		 */
		static size_t readDoubleArr (std::istream & stream, double * value, int len, 
																 bool reverse);
		/**
		 * \brief reads a float value from a binary stream
		 * 
		 * \param stream  the binary stream
		 * \param value   the float value
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes read from the binary stream
		 */
		static size_t readFloat (std::istream & stream, float & value, bool reverse);
		/**
		 * \brief reads an array of float values from a binary stream
		 *
		 * \param stream  the binary stream
		 * \param value   an array of size \p len that accepts the read array
		 * \param len     the length of the array
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes read from the binary stream
		 */
		static size_t readFloatArr (std::istream & stream, float * value, int len, 
																bool reverse);
		/**
		 * \brief reads an integer value from a binary stream
		 * 
		 * \param stream  the binary stream
		 * \param value   the integer value
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes read from the binary stream
		 */
		static size_t readInt (std::istream & stream, int & value, bool reverse);
		/**
		 * \brief reads an array of integer values from a binary stream
		 *
		 * \param stream  the binary stream
		 * \param value   an array of size \p len that accepts the read array
		 * \param len     the length of the array
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes read from the binary stream
		 */
		static size_t readIntArr (std::istream & stream, int * value, int len, 
															bool reverse);
		/**
		 * \brief reads a short integer value from a binary stream
		 * 
		 * \param stream  the binary stream
		 * \param value   the short integer value
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes read from the binary stream
		 */
		static size_t readShort (std::istream & stream, short & value, bool reverse);
		/**
		 * \brief reads an array of short integer values from a binary stream
		 *
		 * \param stream  the binary stream
		 * \param value   an array of size \p len that accepts the read array
		 * \param len     the length of the array
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes read from the binary stream
		 */
		static size_t readShortArr (std::istream & stream, short * value, 
																int len, bool reverse);
    
		/**
		 * \brief writes a double value into a binary stream
		 * 
		 * \param stream  the binary stream
		 * \param value   the double value
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes written to the binary stream
		 */
		static size_t writeDouble (std::ostream & stream, const double value, 
															 bool reverse);
		/**
		 * \brief writes an array of double values into a binary stream
		 *
		 * \param stream  the binary stream
		 * \param value   an array of size \p len that contains the values to write
		 * \param len     the length of the array
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes written to the binary stream
		 */
		static size_t writeDoubleArr (std::ostream & stream, const double * value, 
																	int len, bool reverse);
		/**
		 * \brief writes a float value into a binary stream
		 * 
		 * \param stream  the binary stream
		 * \param value   the floast value
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes written to the binary stream
		 */
		static size_t writeFloat (std::ostream & stream, const float value, 
															bool reverse);
		/**
		 * \brief writes an array of float values into a binary stream
		 *
		 * \param stream  the binary stream
		 * \param value   an array of size \p len that contains the values to write
		 * \param len     the length of the array
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes written to the binary stream
		 */
		static size_t writeFloatArr (std::ostream & stream, const float * value, 
																 int len, bool reverse);
		/**
		 * \brief writes an integer value into a binary stream
		 * 
		 * \param stream  the binary stream
		 * \param value   the integer value
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes written to the binary stream
		 */
		static size_t writeInt (std::ostream & stream, const int value, bool reverse);
		/**
		 * \brief writes an array of integer values into a binary stream
		 *
		 * \param stream  the binary stream
		 * \param value   an array of size \p len that contains the values to write
		 * \param len     the length of the array
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes written to the binary stream
		 */
		static size_t writeIntArr (std::ostream & stream, const int * value, int len, 
															 bool reverse);
		/**
		 * \brief writes a short integer value into a binary stream
		 * 
		 * \param stream  the binary stream
		 * \param value   the short integer value
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes written to the binary stream
		 */
		static size_t writeShort (std::ostream & stream, const short value, 
															bool reverse);
		/**
		 * \brief writes an array of short integer values into a binary stream
		 *
		 * \param stream  the binary stream
		 * \param value   an array of size \p len that contains the values to write
		 * \param len     the length of the array
		 * \param reverse a flag that triggers a reverse byte order if \c true
		 *
		 * \return the number of bytes written to the binary stream
		 */
		static size_t writeShortArr (std::ostream & stream, const short * value, 
																 int len, bool reverse);

	};
#ifdef HAVE_NAMESPACES
};
#endif 

#endif // BINARYIO_HH
