/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SURFACE_COMMON_HH
#define SURFACE_COMMON_HH

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#include <string>
#else
#include <string>
#endif

#include <float.h>

#include <Vector.hh>
#include <Rectmat.hh>

/**
 * \file common.hh
 * 
 * This file contains common definitions that are valid and useful for all
 * modules in this application
 */

/**
 * \brief shortcut for the razorback double vector
 */
typedef RazorBack::Vector_<double> RbDblV;

/**
 * \brief shortcut for the razorback rectangular double matrix
 */
typedef RazorBack::Rectmat_<double> RbDblM;

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif
  /**
   * \brief the mother of all surfcomp exceptions
   */
	class Surfcomp_e 
	{ 
	public:
		/**
		 * \brief standard and default constructor
		 *
		 * The constructor creates an exception class with the specified error
		 * number and error message.
		 *
		 * \param errorno error number
		 * \param msg     the error message
		 */
		Surfcomp_e(int errorno=0, std::string msg = "" ) 
				:errorno_(errorno), message_(msg) { }

		/**
		 * \brief retrieves the error number
		 *
		 * \return the number of the error
		 */
		int getErrorNo() const { return errorno_; }

		/**
		 * \brief retrieves the error message
		 *
		 * \return the message associated with that error
		 */
		std::string getMessage() const { return message_; }

	protected:
		
		/**
		 * \brief the error number
		 */
		int errorno_;
		/**
		 * \brief the error message
		 */
		std::string message_;
	};

	const int OUTOFRANGE_ERR = -1;
	/**
	 * \brief out of range error
	 *
	 * The error number is -1
	 */
	class OutOfRange_e : public Surfcomp_e
	{ 
	public:
		OutOfRange_e(std::string msg = "") :Surfcomp_e(OUTOFRANGE_ERR, msg) { }
	};

	const int INVALIDREF_ERR = -2;
	/**
	 * \brief invalid reference error
	 *
	 * The error number is -2
	 */
	class InvalidReference_e : public Surfcomp_e
	{ 
	public:
		InvalidReference_e(std::string msg = "") :Surfcomp_e(INVALIDREF_ERR, msg) {}
	};

	const int INVALIDITER_ERR = -3;
	/**
	 * \brief invalid iterator error
	 *
	 * The error number is -3
	 */
	class InvalidIter_e : public Surfcomp_e
	{ 
	public:
		InvalidIter_e(std::string msg = "") :Surfcomp_e(INVALIDITER_ERR, msg) { }
	};

	/**
	 * \brief error that occurs if the dimension of a transformation matrix is
	 * not 4x4.
	 */
	class TransformationMatrixDimension_e : public Surfcomp_e { };


	/**
	 * \brief a type for the different face types on a molecular surface
	 */
	enum facetype_t
	{
		ftUnknown,   ///< the face type is not specified
		ftToric,     ///< faces that belong to a toric reentrant surface 
		ftReentrant, ///< reentrant faces 
		ftContact    ///< contact faces
	};

	//@{
	/**
	 * \brief constants for the convenient indexing of point coordinates and
	 * triangle components.
	 */
	const int A = 0;
	const int B = 1;
	const int C = 2;

	const int X = 0;
	const int Y = 1;
	const int Z = 2;
	const int W = 3;
	//@}
	
	/**
	 * \brief successor constant array 
	 * 
	 * This array holds the indices of the successors of the three vertex
	 * triangles in counter clockwise ordering.
	 */
	const int SUCC[3] = { 1, 2, 0 };

	template<class Iter>
	struct lt_iterator
	{
		bool operator()(const Iter& a, 
										const Iter& b) const
		{
			return (&(*a)<&(*b));
		}
	};


	//@{
	/**
	 * \brief limit for the double type
	 */
	extern double double_limit;
  //@}	

	/**
	 * \brief performs a failproof copy of an arbitrary vector to a
	 * position or normal vector
	 *
	 * This method assures, that a vector of arbitrary length is assigned
	 * correctly to a destination vector . The missing components are filled with
	 * zeros. If \p src has fewer elements than \p dest, \p lastelement can be
	 * used to assign a different default value to the last element (this is
	 * useful for 4D position coordiates where the last element (W) is always 1).
	 * 
	 * \param src         the arbitrary source vector
	 * \param dest        a reference to the position or normal vector
   * \param lastelement the value of the last element if \p src has fewer
   *                    elements than \p dest (default is 1)
	 *
	 */
	void copyVector(const RbDblV& src , RbDblV& dest, double lastelement = 1);

	/** 
	 * \brief calculates the cross product
	 * 
	 * \param a,b the input vectors
	 * 
	 * \return the cross product of \p a and \p b
	 */
	RbDblV cross(const RbDblV& a, const RbDblV& b);

	//@}

#ifdef HAVE_NAMESPACES
};
#endif


#define MAX(A,B) (A>B) ? A : B
#define MIN(A,B) (A<B) ? A : B

//******************************************************************************
// ERROR CODES
//******************************************************************************

const int E_FILE_NOT_FOUND   = -100;
const int E_CANNOT_OPEN_FILE = -101;

const int E_WRONG_POINT_INDEX = -200;

#endif // SURFACE_COMMON_HH

