/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef DOUBLETTCHECKER_HH
#define DOUBLETTCHECKER_HH

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <float.h>
#include <math.h>

#ifdef HAVE_STD
#include <set>
#else
#include <multiset.h>
#endif

#include "common.hh"
#include "point.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif

	const double dbl_limit_sqr = double_limit*double_limit;

	/**
	 * \brief a doublett checker
	 *
	 * The doublett checker holds a special hash of points plus index and tests
	 * every new point, that is passed to it if that point is just a doublett
	 * of a yet existing point. The check is done by the function operator()().
	 */
	class DoublettChecker 
	{
	public:
		/**
		 * \brief default constructor
		 *
		 * The default constructor initializes the hash
		 */
		DoublettChecker () : sSpots() { }
				
		/**
		 * \brief function operator, checks for doubletts
		 *
		 * The function operator takes a new point together with its suggested
		 * index in the vector of a surface object. It checks if the point is a
		 * doublett, adds it to the hash if not and returns either the index of
		 * the unique representation. The index of the unique representation is
		 * either the new index in the multiset (size()-1) if the point is still
		 * unique or the index of the found doublett.
		 *
		 * \param point a reference to a valid point of the surface
		 *
		 * \return the index of the unique representation of this point
		 */
		int operator()(const Point& point);

		/**
		 *\brief returns the status of the last check
		 *
		 * If the last point that was checked has been a doublett this function
		 * returns \c true.
		 *
		 * \return \c true if the last check retrieved a doublett, \c false
		 *         otherwise.
		 */
		bool wasDoublett() const { return wasdoublett_; }

		/**
		 * \brief purges the internal data
		 *
		 * This method makes the DoublettChecker object ready for a new run. It
		 * clears the internal set of spots and sets the wasdoublett_ to \c false.
		 * Use this method only after a complete run.
		 */
		void clear();
			
	private:

		//**************************************************************************
		// PRIVATE TYPE DEFINITIONS
		//**************************************************************************

		/**
		 * \brief the internal element of the doublett datastructure
		 *
		 * The spot_t stores the x, y and z coordinates together with the
		 * distance r from the origin and the index of the point. The internal
		 * less then operator (operator<()) uses this distance to sort the points
		 * into bins and enable a faster detection of doubletts. The equal
		 * operator (operator==()) checks if all three coordinates are equal.
		 */
		struct spot_t
		{
				
			double r;
			double x,y,z;
			int    index;

			spot_t(int ix, const Point& point);

			bool operator<(const spot_t& b) const { 
				return ((r-b.r)<-dbl_limit_sqr); }
			bool operator==(const spot_t& b) const;
		};

		//**************************************************************************
		// PRIVATE MEMBERS
		//**************************************************************************
			
		/**
		 * \brief the geometric hash that holds the surface points
		 *
		 * for every existing radius r this multiset holds all points that fall
		 * in this distance range.
		 */
		std::multiset<spot_t> sSpots;
		
		/**
		 * \brief state that contains the result of the last check
		 *
		 * This member is either \c true or \c false depending if the last point
		 * checked was a doublett or not.
		 */
		bool wasdoublett_;

	};

	//****************************************************************************
	inline DoublettChecker::spot_t::spot_t(int ix, 
																				 const Point& point)
	//****************************************************************************
	:index(ix), 	 
	 x(point.getPosition()[X]), 
	 y(point.getPosition()[Y]), 
	 z(point.getPosition()[Z]),
	 r(0) { r = x*x + y*y + z*z; }


	//****************************************************************************
	inline bool DoublettChecker::spot_t::operator==(const spot_t& b) const
  //****************************************************************************
	{
		return (
			(fabs(x-b.x)<double_limit) &&
			(fabs(y-b.y)<double_limit) &&
			(fabs(z-b.z)<double_limit));
	}

#ifdef HAVE_NAMESPACES
};	
#endif

#endif //DOUBLETTCHECKER_HH
