/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file fileformatfactory.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 09 Jun 2004, last CVS commit 2004/06/09 05:43:51
 */

#ifndef FILEFORMATFACTORY_HH
#define FILEFORMATFACTORY_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <map>
#else
#  include <map.h>
#endif

#include "common.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

  class SurfaceIO;

  /**
   * \brief a collection of SurfaceIO factory methods
   *
   * The FileFormatFactory represents a collection of factory methods for
   * different subclasses of the SurfaceIO type. For each subtype a factory
   * method of the form <code> SurfaceIO* factory()</code> can be registered to
   * the FileFormatFactory together with a name that identifies the class
   * uniquely and a brief informative description, that can be used in file
   * dialogs and so on.  (register(const std::string&, format_t)). The
   * factories can also be unregistered (unregister(const std::string&)). A new
   * (dynamic) instance of a SurfaceIO subclass is returned by the
   * createClass(const std::string&) method. The file format description can be
   * retrieved via getDescription(const std::string&).
   *
   *
   * The following example illustrates how to register a SurfaceIO subclass:
   *
   * \code
   * class MyData : public SurfaceIO
   * {
   *   ...
   *
   *   static SurfaceIO* createInstance() { return new MyData(); }
   *   static void register();
   * };
   *
   * ...
   *
   * void MyData::register()
   * {
   *   FileFormatFactory::instance().registerClass("mydata", 
   *                                               &MyData::createInstance,
   *                                               "files from my surface generator");
   * }
   *
   * \endcode
   *
   * An alternative for the registration is the use of the
   * REGISTER_IOCLASS(format, classname, description) macro.
   *
   * The class is a Singleton(127) and the instance can be retrieved by
   * instance().
   */
  class FileFormatFactory
  {
  
  public:
    //**************************************************************************
    // PUBLIC TYPES
    //**************************************************************************

    /**
     * \brief the type of the factory method pointers.
     */
    typedef SurfaceIO* (*factory_t)();

    /// \brief the type of information stored for each file type
    struct format_t
    {
      factory_t method;
      std::string description;
      
      format_t(const factory_t& _method, const std::string& _description) 
          : method(_method), description(_description) { }
    };

    /**
     * \brief a constant iterator to the internally stored class/method pairs
     */
    typedef std::map<std::string, format_t>::const_iterator factoryIter;
    
    //**************************************************************************
    // OPERATIONS
    //**************************************************************************

    /** 
     * \brief registers a factory method
     *
     * \param name identifier of the new class
     * \param factory method pointer to the class' factory method
     * 
     * \return \c true if the class could be added to the registry or \c false
     *         if \p name is already registered.
     *
     * \sa unregisterClass(const std::string&)
     */
    bool registerClass(const std::string& name, format_t factory);
    /** 
     * \brief unregisters a class
     * 
     * \param name the id of the class factory that should be unregistered
     * 
     * \return \c true if the class factory was found and unregistered, \c
     *         false otherwise.
     * 
     * \sa registerClass(const std::string&, format_t)
     */
    bool unregisterClass(const std::string& name);

    /** 
     * \brief checks if a class can be created by the FileFormatFactory
     *
     * \param name the id of the desired class factory
     * 
     * \return \c true if the class is registered, \c false otherwise.
     *
     * \sa createClass(const std::string& name);
     */
    bool hasClass(const std::string& name) const;
    /** 
     * \brief creates a new instance of a specific class
     * 
     * \param name the id of the desired class
     * 
     * \return a new instance of the class, created by the registered class
     *         factory.
     */
    SurfaceIO* createClass(const std::string& name) const;

    /** 
     * \brief retrieves information about the file format
     * 
     * \param name the id of the desired class
     * 
     * \return a short information string about the file format
     */
    const std::string& getDescription(const std::string& name) const;

		//@{
		/** 
		 * \brief provides access to the file format sequence
		 * 
		 * 
		 * \return an factoryIter to the begin and behind the end of the registered
		 *         file format sequence.
		 */
    factoryIter begin() const { return factories_.begin(); }
    factoryIter end()   const { return factories_.end(); }
		//@}

    //**************************************************************************
    // SINGLETON
    //**************************************************************************

    /** 
     * \brief provides access to the instance of the FileFormatFactory
     * 
     * \return a reference to the instance of the FileFormatFactory
     */
    static FileFormatFactory& instance();
    
  private:
    
    //**************************************************************************
    // PRIVATE TYPES
    //**************************************************************************

    //@{
    /**
     * \brief type of the factory map plus iterators
     */
    typedef std::map<std::string, format_t> factories_t;
    typedef factories_t::const_iterator factories_cit;
    typedef factories_t::iterator factories_it;
    //@}

    //**************************************************************************
    // PRIVATE MEMBERS
    //**************************************************************************

    /**
     * \brief map of SurfaceIO factory methods
     * 
     * The factory map is used internally to store the static method pointers
     * to the various SurfaceIO factory methods.
     */
    factories_t factories_;

    //**************************************************************************
    // SINGLETON
    //**************************************************************************

    /** 
     * \brief default constructor
     * 
     * The constructor is private according to the Singleton(127) pattern.
     */
    FileFormatFactory() : factories_() { }

  };

#ifdef HAVE_NAMESPACES
};
#endif 

/** 
 * \brief MACRO that registers a data class upon intialization.
 *
 * This macro can be used in the implementation of a specific SurfaceIO class to
 * register the class at the FileFormatFactory. \p classname can be any subclass of
 * SurfaceIO that redefines the static method SurfaceIO::createInstance().
 *
 * An example class definition (header file):
 *
 * \code
 * class MyFileFormatIO : public SurfaceIO
 * {
 *   ...
 *
 *   static SurfaceIO* createInstance() { return new MyFileFormatIO(); }
 *   static void register();
 * };
 * \endcode
 *
 * and the corresponding registration call in the source (.cc, .cxx or .cpp
 * file):
 * 
 * \code 
 * REGISTER_IOCLASS(MyFileFormatIO);
 * \endcode
 * 
 * \param classname the name of the class (without quotes) that should be
 *                  registered.
 */
#define REGISTER_IOCLASS(format, classname, description) \
struct IORegistration_##format \
{ \
  IORegistration_##format() { \
    Surface::FileFormatFactory::instance(). \
      registerClass(#format, \
                    FileFormatFactory::format_t(&classname::createInstance, \
                                                description)); \
  } \
}; \
\
static IORegistration_##format registration_##format; 


#endif // FILEFORMATFACTORY_HH
