/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file graphiterator.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 25 Mar 2004, last CVS commit 2004/03/25 18:09:11
 */

#ifndef GRAPHITERATOR_HH
#define GRAPHITERATOR_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "viewiterator.hh"
#include "graphview_traits.hh"

#include "surface.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	class GraphView;

	/**
	 * \brief a template for the graph iterators that allows direct iterator
	 * access to the associated points
	 *
	 * \param Iter is the class type of the internal vertex iterator
	 * \param T    is the class type of the Point
	 * \param View is the class type of the graphview
	 */
	template<class Iter, class T, class View = GraphView>
	class GraphIter : public ViewIterator
	{
	public:
		//**************************************************************************
    // CONSTRUCTOR AND ASSIGNMENT
    //**************************************************************************

		GraphIter() : impl_(), valid_(false) { }

		/**
		 * \brief copy constructor
		 */
		GraphIter(const GraphIter<Iter,T,View>& src) 
				: ViewIterator(src), impl_(src.impl_), valid_(src.valid_) { }

		/**
		 * \brief assignment operator
		 */
		GraphIter<Iter,T,View>& operator=(const GraphIter<Iter,T,View>& src) {
      if (view_) view_->detach(*this);

			impl_  = src.impl_;
			valid_ = src.valid_;

			src.view_->attach(*this);
			return *this;
		}

		//**************************************************************************
		// OPERATOR
		//**************************************************************************

		/**
		 * \brief prefix increment
		 *
		 * increments the iterator to the next position and returns a reference.
		 *
		 * \return a reference to the \c this at the new position
		 */
		GraphIter<Iter, T, View>& operator++() {
			CHECKITER;
			++impl_; 
			return *this; 
		}
		/**
		 * \brief postfix increment
		 *
		 * returns a copy of \c this and increments the iterator to the next
		 * position.
		 *
		 * \return a copy of the actual iterator
		 */
		GraphIter<Iter, T, View> operator++(int) {
			CHECKITER;
			GraphIter<Iter, T, View> tmp = *this;
			++impl_; 
			return tmp; 
		}

		/**
		 * \brief prefix decrement
		 *
		 * decrements the iterator to the previous position and returns a
		 * reference.
		 *
		 * \return a reference to the \c this at the new position
		 */
		GraphIter<Iter, T, View>& operator--() {
			CHECKITER;
			--impl_; 
			return *this; 
		}
		/**
		 * \brief postfix decrement
		 *
		 * returns a copy of \c this and decrements the iterator to the previous
		 * position.
		 *
		 * \return a copy of the actual iterator
		 */
		GraphIter<Iter, T, View> operator--(int) {
			CHECKITER;
			GraphIter<Iter, T, View> tmp = *this;
			--impl_; 
			return tmp; 
		}

		/**
		 * \brief arrow operator (access to the real element by arrow)
		 *
		 * The arrow operator provides access to the real element of the type T*.
		 *
		 * \return a pointer to the real element
		 */
		T* operator->() {
			CHECKITER; 

			View* view = dynamic_cast<View*>(view_);
			return view->iteratormap_[*impl_].operator->(); 
		}

		/**
		 * \brief dereferencing operator (access to the real element by ref.)
		 *
		 * The dereferencing operator provides access to the real element of the
		 * type T&.
		 *
		 * \return a reference to the real element
		 */
		T& operator*() {
			CHECKITER; 

			View* view = dynamic_cast<View*>(view_);
			return *(view->iteratormap_[*impl_]); 
		}

		/** 
		 * \brief transformation into surface entity point iterators
		 * 
		 * \return the SurfaceEntity point iterator represented by that iterator
		 */
		operator SurfaceEntity::points_it () const {
			View* view = dynamic_cast<View*>(view_);
			return view->iteratormap_[*impl_];
		}
			

		//**************************************************************************
		// OBSERVER
		//**************************************************************************

		void update() { ViewIterator::update(); valid_ = false; }

		//**************************************************************************
		// COMPARISON
		//**************************************************************************

		/**
		 * \brief checks for inequality between to iterators
		 *
		 * \param test then other iterator
		 *
		 * \return \c true if the iterators are not equal.
		 */
		bool operator!=(const GraphIter<Iter, T, View>& test) {
			CHECKITER;
			return impl_ != test.impl_;
		}

		/**
		 * \brief checks for equality between to iterators
		 *
		 * \param test then other iterator
		 *
		 * \return \c true if the iterators are equal.
		 */
		bool operator==(const GraphIter<Iter, T, View>& test) {
			CHECKITER;
			return impl_ == test.impl_;
		}

		bool isValid() const { return valid_; }
		
	protected:
				
		Iter impl_;
		bool valid_;

	private:
		
		friend class GraphView;
		
		/**
		 * \brief the real constructor is private to make it accessible only from
		 * the graph view.
		 */
		GraphIter(GraphView& view, const Iter& impl)
				: ViewIterator(view)	{
			impl_  = impl;
			valid_ = true;
		}
	};

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // GRAPHITERATOR_HH
