/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "graphview.hh"

#ifdef HAVE_STD
#  include <list>
#else
#  include <list.h>
#endif
#include <iostream>

#include <boost/graph/dijkstra_shortest_paths.hpp>

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
using namespace boost;
#endif 

//#include <iostream>

//******************************************************************************
GraphView::GraphView() 
//******************************************************************************
: SurfaceView(), graph_(), pointmap_(), oftweightcalc_(true)
{
	iteratormap_ = get(vertex_data, graph_);
	elengthmap_  = get(edge_weight, graph_);
	distmap_     = get(vertex_distance, graph_);
	premap_     = get(vertex_predecessor, graph_);
}

//******************************************************************************
GraphView::GraphView(SurfaceEntity& surface)
//******************************************************************************
: SurfaceView(surface), graph_(), pointmap_(), oftweightcalc_(true)
{ 
	iteratormap_ = get(boost::vertex_data, graph_);
	elengthmap_  = get(boost::edge_weight, graph_);
	distmap_     = get(vertex_distance, graph_);
	surface_->lock();
}

//******************************************************************************
void GraphView::update()
//******************************************************************************
{
	notify();
}

//******************************************************************************
void GraphView::buildSelectionEntity(DeepCopyInformation& copyinfo) const
//******************************************************************************
{
  SurfaceEntity& entity = copyinfo.newview->getSurfaceEntity();
  //redo unlock while building
  entity.lock(false);

  //copy all the points
	graphview_traits::vertex_it v,end;		
  for(tie(v,end) = vertices(graph_); v!=end; ++v)
  {
    SurfaceEntity::points_it oldpoint = get(iteratormap_, *v);
    SurfaceEntity::points_it newpoint = entity.addPoint();

    *newpoint = *oldpoint;
    newpoint->setData(oldpoint->data());

    //build up the pointmap
    copyinfo.pointmap[oldpoint] = newpoint;
  }

  SurfaceEntity::triangles_cit t;
  for(t=surface_->beginTriangles();t!=surface_->endTriangles();++t)
    //scan all triangles of this entity if they represent 3 edges
    if (getNEdges(t))
      entity.addTriangle(copyinfo.pointmap[t->getVertex(A)],
                         copyinfo.pointmap[t->getVertex(B)],
                         copyinfo.pointmap[t->getVertex(C)])->copyFlags(*t);

  //relock after building
  entity.lock();
}

//******************************************************************************
void GraphView::buildSelectionView(DeepCopyInformation& copyinfo) const
//******************************************************************************
{
  GraphView* view = dynamic_cast<GraphView*>(copyinfo.newview);
  
  //select all the points
	graphview_traits::vertex_it v,end;		
  for(tie(v,end) = vertices(graph_); v!=end; ++v)
    view->selectPoint(copyinfo.pointmap[get(iteratormap_, *v)]);

  //select all the triangles in the new entity
  SurfaceEntity::triangles_cit t;
  for(t=view->surface_->beginTriangles();t!=view->surface_->endTriangles();++t)
    view->selectTriangle(t);
}

//******************************************************************************
GraphView::vertexIter GraphView::selectPoint(SurfaceEntity::points_it point)
//******************************************************************************
{
	pointmap_cit p = pointmap_.find(point);
	if (p==pointmap_.end())
	{
		graphview_traits::vertex_it vit,b;		

		//add the point and assign the iterator as a vertex property
		graphview_traits::vertex_t vertex = add_vertex(graph_);
		put(iteratormap_, vertex, point);
    put(premap_, vertex, vertex);

		tie(b,vit) = vertices(graph_);
		--vit;
		vertexIter v(*this, vit);

		pointmap_[point] = vit;

		return v;
	}
	else
		return end();
}

//******************************************************************************
void GraphView::deselectPoint(vertexIter vertexiter)
//******************************************************************************
{
	//retrieve the corresponding vertex and point iterator
	graphview_traits::vertex_t v = *vertexiter.impl_;
	SurfaceEntity::points_it   p = iteratormap_[v];

	clear_vertex(v,graph_);
	remove_vertex(v,graph_);
	
	//clear the entry from the pointmap
	pointmap_it pm = pointmap_.find(p);
	if (pm!=pointmap_.end())
		pointmap_.erase(pm);
}

//******************************************************************************
size_t GraphView::selectTriangle(SurfaceEntity::triangles_cit triangle,
																 size_t minimum)
//******************************************************************************
{
	typedef pair<graphview_traits::vertex_t, graphview_traits::vertex_t> vpair_t;

	size_t c_edges = 0;
	list<vpair_t> todo_edges;

	//iterator over all point iterators
	for(int i=A;i<=C;++i)
	{
		//get the target and source points
		pointmap_it source_p = pointmap_.find(triangle->getVertex(i));
		pointmap_it target_p = pointmap_.find(triangle->getVertex(SUCC[i]));		

		//check if both points are members of the graph
		if ((source_p!=pointmap_.end()) && (target_p!=pointmap_.end()))
		{
			todo_edges.push_back(vpair_t(*source_p->second,	*target_p->second));
			++c_edges;
		}
	}

	//trigger adding of edges only if the minimum number of vertices is in the
	//graph
	if (c_edges>=minimum)
	{
		list<vpair_t>::const_iterator l;
		for(l=todo_edges.begin();l!=todo_edges.end();++l)
		{
			graphview_traits::edge_t e;
			bool constructed;
			tie(e, constructed) = add_edge(l->first, l->second, graph_);


			//proceed only if the edge is constructed
			if (constructed && oftweightcalc_)
				elengthmap_[e] = calcDistance(l->first, l->second);
		}
	}

	return c_edges;
}

//******************************************************************************
size_t GraphView::deselectTriangle(SurfaceEntity::triangles_cit triangle)
//******************************************************************************
{
	size_t c = 0;
	//iterator over all point iterators
	for(int i=A;i<=C;++i)
	{
		//get the target and source points
		pointmap_it source_p = pointmap_.find(triangle->getVertex(i));
		pointmap_it target_p = pointmap_.find(triangle->getVertex(SUCC[i]));		

		//check if both points are members of the graph
		if ((source_p!=pointmap_.end()) && (target_p!=pointmap_.end()))
		{
			graphview_traits::edge_t e;
			bool present;
			tie(e, present) = edge(*source_p->second, *target_p->second, graph_);

			//proceed only if the edge is constructed
			if (present)
			{
				remove_edge(e, graph_);
				++c;
			}
		}
	}

	return c;
}

//******************************************************************************
size_t GraphView::deselectTriangle(SurfaceEntity::triangles_cit triangle, 
																	 size_t threshold)
//******************************************************************************
{
	size_t np = 0;

	//iterator over all point iterators
	for(int i=A;i<=C;++i)
		if (pointmap_.find(triangle->getVertex(i)) != pointmap_.end()) ++np;

	//remove edges only of the number of points is below threshold
	if (np<threshold)
		return deselectTriangle(triangle);
	else
		return 0;
}

//******************************************************************************
bool GraphView::addEdge(vertexIter source, vertexIter target)
//******************************************************************************
{
	graphview_traits::edge_t e;
	bool constructed;
	tie(e, constructed) = add_edge(*source.impl_, *target.impl_, graph_);

	//proceed only if the edge is constructed
	if (constructed && oftweightcalc_)
			elengthmap_[e] = calcDistance(*source.impl_, *target.impl_);
	
	return constructed;
}

//******************************************************************************
bool GraphView::removeEdge(vertexIter source, vertexIter target)
//******************************************************************************
{
	graphview_traits::edge_t e;
	bool present;
	tie(e, present) = edge(*source.impl_, *target.impl_, graph_);

	if (present)
		remove_edge(e, graph_);
	
	return present;
}

//******************************************************************************
size_t GraphView::removeSingles()
//******************************************************************************
{
	size_t c = 0;

	graphview_traits::vertex_it v,end;
	for(tie(v,end) = vertices(graph_); v!=end; )
	{
		graphview_traits::vertex_it w = v++;
		if (out_degree(*w, graph_) == 0) 
		{
			++c;
			remove_vertex(*w, graph_);
		}
	}
		
	if (c>0) notify();
	return c;
}

//******************************************************************************
size_t GraphView::getNEdges(SurfaceEntity::triangles_cit triangle) const
//******************************************************************************
{
	int c_edges = 0;

	//iterator over all point iterators
	for(int i=A;i<=C;++i)
	{
		if (connected(triangle->getVertex(i),triangle->getVertex(SUCC[i])))
			++c_edges;
		
		if (connected(triangle->getVertex(SUCC[i]),triangle->getVertex(i)))
			++c_edges;
	}

	return c_edges;
}

//******************************************************************************
GraphView::vertexIter GraphView::getPredecessor(const vertexIter& v)
//******************************************************************************
{
  return selectionOf(getPredecessorPoint(v));
}

//******************************************************************************
SurfaceEntity::points_it GraphView::getPredecessorPoint(const vertexIter& v) 
  const
//******************************************************************************
{
  return get(iteratormap_, get(premap_, *v.impl_));
}

//******************************************************************************
bool GraphView::connected(const SurfaceEntity::points_it&  source,
													const SurfaceEntity::points_it&  target) const
//******************************************************************************
{
		//get the target and source points
		pointmap_cit source_p = pointmap_.find(source);
		pointmap_cit target_p = pointmap_.find(target);			

		//check if source and target are selected
		if ((source_p != pointmap_.end()) && (target_p != pointmap_.end()))
			return edge(*source_p->second, *target_p->second, graph_).second;
		else
			return false;
}

//******************************************************************************
void GraphView::recalcWeights()
//******************************************************************************
{
	graphview_traits::edge_it e,end;
	for(tie(e,end) = edges(graph_); e!=end; ++e)
		put(elengthmap_, *e, calcDistance(source(*e, graph_),target(*e, graph_)));
}

//******************************************************************************
void GraphView::compOnSurfaceDistances(vertexIter v)
//******************************************************************************
{
	doOnSurfaceDistances(*v.impl_);
}

//******************************************************************************
void GraphView::compOnSurfaceDistances(SurfaceEntity::points_it p)
//******************************************************************************
{
	pointmap_cit pm = pointmap_.find(p);
	if (pm!=pointmap_.end())
		doOnSurfaceDistances(*pm->second);
}

//******************************************************************************
void GraphView::compOnSurfaceDistances(size_t ix)
//******************************************************************************
{
	if (ix<num_vertices(graph_))
	  doOnSurfaceDistances(vertex(ix, graph_));
	else
		throw OutOfRange_e("Vertex index out of range!");
			
}


//******************************************************************************
GraphView::vertexIter GraphView::selectionOf(SurfaceEntity::points_it element)
//******************************************************************************
{
	pointmap_cit pm = pointmap_.find(element);
	if (pm != pointmap_.end())
		return vertexIter(*this, pm->second);
	else
		return end();
}

//******************************************************************************
RbDblV GraphView::getCenter() const
//******************************************************************************
{
  RbDblV center(4,0);
	graphview_traits::vertex_it v,end;		

  for(tie(v,end) = vertices(graph_); v!=end; ++v)
    center += get(iteratormap_, *v)->getPosition();
  
  center /= getNPoints();
  center[W] = 1;

  return center;
}

//******************************************************************************
RbDblV GraphView::getExtensions() const
//******************************************************************************
{
  RbDblV maxpos(4,-DBL_MAX); maxpos[W] = 1;
  RbDblV minpos(4,DBL_MAX); minpos[W] = 1;

	graphview_traits::vertex_it v,end;		

  for(tie(v,end) = vertices(graph_); v!=end; ++v)
  {
    const RbDblV& pos = get(iteratormap_, *v)->getPosition();
    for(int i=X;i<W;++i)
    {
      if (pos[i]<minpos[i]) minpos[i] = pos[i];
      if (pos[i]>maxpos[i]) maxpos[i] = pos[i];
    }
  }

  return (maxpos - minpos);
}

//******************************************************************************
double GraphView::calcDistance(graphview_traits::vertex_t a, 
															 graphview_traits::vertex_t b) const
//******************************************************************************
{
	return *get(iteratormap_, a) - *get(iteratormap_, b);
}

//******************************************************************************
void GraphView::doOnSurfaceDistances(graphview_traits::vertex_t v)
//******************************************************************************
{
	generateIndex();
	dijkstra_shortest_paths(graph_, v, 
                          distance_map(distmap_).predecessor_map(premap_));
}

//******************************************************************************
void GraphView::generateIndex()
//******************************************************************************
{
	graphview_traits::vertex_it vit,end;
	graphview_traits::vindexmap_t indexmap = get(boost::vertex_index, graph_);
	int i = 0;
	for(tie(vit,end) = vertices(graph_); vit!=end; ++vit, ++i)
	{
		put(indexmap, *vit, i);
	}
}
