/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file graphview.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 25 Mar 2004, last CVS commit 2004/03/25 14:14:03
 */

#ifndef GRAPHVIEW_HH
#define GRAPHVIEW_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include <boost/tuple/tuple.hpp>

#include "surfaceview.hh"
#include "point.hh"
#include "graphview_traits.hh"
#include "graphiterator.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	/**
	 * \brief a graph view class
	 *
	 * Because of the complexity of a graph view on an SurfaceEntity the
	 * structural patterns (adding and removing of points or triangles) are not
	 * allowed, as long as a graph view exists for the entity. This restriction
	 * is not active for changing the properties of points or connections between
	 * triangles.
	 *
	 * Edges are always added in the counter clockwise direction. So if a
	 * triangle is used to add edges to the graph, only three edges in the
	 * directions of the triangles order are introduced. The class takes care of
	 * determining the edge length and stores them as edge weights. Edge weights
	 * are one of the most important additional information stored in the
	 * graph. The calculation of edge weights can be done automatically (the
	 * default) when oftweightcalc_ is set to true. This is controlled by
	 * oftWeightCalc(bool). If on the fly weight calculation is turned off, it
	 * can be done explicitly using the method recalcWeights(). The automatic
	 * weight calculation is turned on automatically in the constructor.
	 *
	 * Based on the edge length a on-surface distance can be calculated between a
	 * given point and all other points in the graph. This calculation is
	 * performed by the dijkstra-shortest path algorithm and is started with one
	 * of the compOnSurfaceDistances() methods, which take a vertex iterator, a
	 * surface entity point iterator or the vertex index as the center for the
	 * computation. The distances of all other vertices to that center can then
	 * be retrieved by getDistance().
	 */
	class GraphView : public SurfaceView
	{
	public:

		//****************************************************************************
		// PUBLIC TYPE DEFINITIONS
		//****************************************************************************

		/**
		 * \brief the type of the standard vertex iterator
		 */
		typedef GraphIter<graphview_traits::vertex_it, 
											Point> vertexIter;
		
		/**
		 * \brief the type of the adjacency vertex iterator
		 */
		typedef GraphIter<graphview_traits::adjacency_it, 
											Point> adjacencyIter;
		
		//****************************************************************************
		// CONSTRUCTORS AND DESTRUCTOR
		//****************************************************************************

		/**
		 * \brief default constructor
		 *
		 * Creates a vector surface view, without an observed entity object
		 *
		 * \sa GraphView(SurfaceEntity& const)
		 */
		GraphView();
		
		/**
		 * \brief initializing constructor
		 *
		 * This constructor creates a vector surface view, that observes the entity
		 * \p surface.
		 *
		 * \param surface SurfaceEntity that is observed.
		 *
		 * \sa GraphView()
		 */
		GraphView(SurfaceEntity& surface);

		~GraphView() { if (surface_) surface_->lock(false);}

		//**************************************************************************
		// OBSERVER
		//**************************************************************************

		bool setSurfaceEntity(SurfaceEntity& surface);
		
		/**
		 * \brief is called by the observed intity, when the surface has changed
		 *
		 * \note The current implementation of update does nothing.
		 */
		void update();

    //**************************************************************************
    // DEEP COPY
    //**************************************************************************

  protected:

    void buildSelectionEntity(DeepCopyInformation& copyinfo) const;
    void buildSelectionView(DeepCopyInformation& copyinfo) const;

  public:

		//**************************************************************************
		// VIEW BUILD OPERATIONS
		//**************************************************************************

		/**
		 * \brief adds a point to the vector of selected surface points
		 *
		 * The method adds a new point to the vector of selected surface points. It
		 * is checked, if the point has already been selected by that view, and if
		 * so, the method returns an iterator that is equal to end(). This
		 * operation invalidates all attached ViewIterators!
		 *
		 * \param point an iterator in the SurfaceEntity point list.
		 *
		 * \return the new vertex iterator or end() if the point has already been
		 *         added to the graph.
		 *
		 * \sa selectTriangle(), deselectPoint()
		 */
		vertexIter selectPoint(SurfaceEntity::points_it point);
		/**
		 * \brief removes a point from the selected points
		 *
		 * The point, represented by pointIter is removed from the the graph. This
		 * operation invalidates all attached ViewIterators!
		 *
		 * \param vertexIter the iterator to the point that should be deselected
		 *
		 * \sa selectPoint(), deselect(Triangle)
		 */
		void deselectPoint(vertexIter vertexiter);

		/**
		 * \brief adds an edge to the graph
		 *
		 * For every edge in the \p triangle the method adds new edges to the graph
		 * that represents the selected surface part. The edges are only added if
		 * both vertices at their ends are already present in the graph. Beside
		 * that it can be specified how many edges of the triangle must be between
		 * points selected by the graph view to add the edges at all. The default
		 * \p minimum is three, so that only triangles that are completely build by
		 * selected points are added.
		 *
		 * This operation invalidates all attached ViewIterators!
		 *
		 * \param triangle an iterator in the SurfaceEntity triangle list.
		 * \param minimum  the minimal number of triangle vertices that should be 
		 *                 selected by the graph view. (default = 3)
		 *
		 * \return the number of edges that are members of the graph view
		 *
		 * \sa selectPoint(), 
		 *     addEdge(), deselectTriangle(SurfaceEntity::triangles_cit)
		 */
		size_t selectTriangle(SurfaceEntity::triangles_cit triangle, 
													size_t minimum = 3);
		/**
		 * \brief removes the edges between the points of the \p triangle
		 *
		 * The triangle, represented by a SurfaceEntity::triangles_it specifies the
		 * edges that should be removed from the graph if present. This operation
		 * invalidates all attached ViewIterators!
		 *
		 * \param triangle an iterator in the SurfaceEntity triangle list.
		 *
		 * \return the number of edges that have been removed.
		 *
		 * \sa selectTriangle(), removeEdge(), deselectPoint(),
		 *     deselectTriangle(SurfaceEntity::triangles_cit, size_t)
		 */
		size_t deselectTriangle(SurfaceEntity::triangles_cit triangle);

		/** 
		 * \brief removes the edges between the points of the \p triangle, if less
		 * then \p threshold points are selected by the graph view.
		 *
		 * First it is checked how much points of the triangle are selected by the
		 * view. If the number is below the threshold all remaining edges of the
		 * triangle are removed. If threshold is above 3 the triangle is removed
		 * anyway but calling deselectTriangle(SurfaceEntity::triangles_cit) would
		 * have been more efficient.
		 * 
		 * \param triangle  an iterator in the SurfaceEntity triangle list;
		 * \param threshold the number of triangle edge points that must be selected
		 *                  to keep the triangle self selected
		 * 
		 * \return the number of removed edges
		 */
		size_t deselectTriangle(SurfaceEntity::triangles_cit triangle, 
														size_t threshold);

		/**
		 * \brief adds an edge between two points in the graph
		 *
		 * Additional to the triangle based setting and removal of edges single
		 * edges can be added to the graph using this method. An edge from \p
		 * source to \p target is added to the graph.
		 *
		 * \param source the beginning point of the new edge
		 * \param target the endpoint of the new edge
		 *
		 * \return \c true if the edge was added, \c false if it was already present.
		 *
		 * \sa selectTriangle(), removeEdge()
		 */
		bool addEdge(vertexIter source, vertexIter target);
		/**
		 * \brief removes a single edge between two points
		 *
		 * This method removes the edge between \p source and \p target if present,
		 * otherwise it does nothing. 
		 *
		 * \param source the beginning point of the edge that should be removed
		 * \param target the endpoint of the to be removed edge
		 *
		 * \return \c true if the edge could be removed, \c false if it was not there.
		 *
		 * \sa deselectTriangle(), addEdge()
		 */
		bool removeEdge(vertexIter source, vertexIter target);

		/**
		 * \brief removes all non-connected vertices
		 *
		 * This method removes all vertices that are not connected to any other
		 * point.
		 *
		 * \return the number of removed vertices
		 */
		size_t removeSingles();

		//**************************************************************************
		// OPERATIONS
		//**************************************************************************

		/**
		 * \brief recalculates all edge weights
		 *
		 * This program recalculates all edge weights. It can be used if the
		 * position of the atoms have changed, or if weight calculation was turned
		 * off for a while.
		 */
		void recalcWeights();

		//@{
		/**
		 * \brief calculates the on surface distance to all vertices from \p
		 * central
		 *
		 * This method calculates the distance from the \p central vertex to all
		 * other vertices in this graph view based on their edge weights. For this
		 * it uses the dijkstra shortest path algorithm. The distances are stored
		 * in the distance property of the graph.
		 *
		 *
		 * \param center    the central vertex (either by index, as 
		 *                  SurfaceEntity::points_it or  as a vertexIter
		 *
		 * \exception OutOfRange_e if the center is given as a index and this index
		 *                         is out of range
		 */
		void compOnSurfaceDistances(vertexIter v);
		void compOnSurfaceDistances(SurfaceEntity::points_it p);
		void compOnSurfaceDistances(size_t ix);
		//@}

		//**************************************************************************
		// PROPERTIES AND FLAGS
		//**************************************************************************

		/**
		 * \brief gives back the number of selected points on the surface
		 *
		 * \return the number of points in that particular surface view. (i.e. the
		 * number of vertices in the graph)
		 * 
		 * \sa getNEdges()
		 */
		size_t getNPoints() const { return boost::num_vertices(graph_); }

		/**
		 * \brief gives back the number of edges in the graph
		 *
		 * \note Each pair of vertices is connected by either one or two edges. One
		 * if the vertices are on the boundary of the selection, or two if either
		 * one of them is inside.
		 *
		 * \return the number of edges between the selected surface points.
		 * 
		 * \sa getNPoints()
		 */
		size_t getNEdges() const { return boost::num_edges(graph_); }
	
		/**
		 * \brief counts the edges that are available between the points of a
		 * triangle.
		 *
		 * This method counts the edges that are available between the points of a
		 * triangle. This is method is necessary for identifying the status of a
		 * triangle in the selected surface area:
		 * 
		 * - \b 6 edges   : the triangle fully embeded in the selection (\c interior)
		 * - \b 5 edges   : the triangle is a valid part of the boundary of the 
		 *                  selection. (\c border)
		 * - \b 4-3 edges : the triangle is dangling or stand-alone
		 * - \b 2 edges   : only two vertices of the triangle are selected.
		 * - \b 1 edge    : the triangle is adjacent to the border of the patch
		 * - \b 0 edges   : the triangle is not part of the selection
		 */
		size_t getNEdges(SurfaceEntity::triangles_cit triangle) const;

		//@{
		/** 
		 * \brief checks if a connection from \p source to \p target exists
		 * 
		 * \param source the source point
		 * \param target the target point of the queried edge
		 * 
		 * \return \c true if a directed edge between \p source and \p target
		 *         exists, \c false otherwise
		 */
		bool connected(const SurfaceEntity::points_it&  source,
									 const SurfaceEntity::points_it&  target) const;
		template<class GraphIter>	bool connected(const GraphIter& source,
																						 const GraphIter& target) const {
			return edge(*source.impl_, *target.impl_, graph_).second;
		}
    //@}

		/**
		 * \brief retrieves the point iterator associated with the \p ix'th vertex.
		 *
		 * This method is mainly for builders that use a graphview to build more
		 * sophisticated views like patches. It gives the opportunity to get a
		 * direct access to the point iterators associated with the vertex for a
		 * further selection of them in the build process.
		 *
		 * \param ix the index of the vertex
		 *
		 * \return the point iterator associated with the index
		 *
		 * \exception OutOfRange_e if the index is out of range
		 */
		SurfaceEntity::points_it operator[](size_t ix) const;

		/**
		 * \brief denotes if edge-weights are calculated on the fly
		 *
		 * \return the status of the on-the-fly weight calculation flag
		 *
		 * \sa oftWeightCalc(), oftweightcalc_
		 */
		bool isOftWeightCalc() const { return oftweightcalc_; }
		/**
		 * \brief enables or disables the on-the-fly weight calculation
		 *
		 * If the automatic weight calculation was turned off, enabling it again
		 * does not calculate all the weights immedeately. For that recalcWeights()
		 * should be called.
		 *
		 * \param enable if \c true on-the-fly weight calculation is enabled.
		 *
		 * \sa isOftWeightCalc(), oftweightcalc_
		 */
		bool oftWeightCalc(bool enable=true) { oftweightcalc_ = enable; }

		//@{
		/**
		 * \brief retrieves the distance of the specified vertex
		 *
		 * \param v a vertexIter or adjacencyIter of the desired vertex
		 *
		 * \return the distance from the central vertex
		 */
		double getDistance(const vertexIter& v)    const { return get(distmap_, *v.impl_);}
		double getDistance(const adjacencyIter& v) const { return get(distmap_, *v.impl_);}

		/** 
		 * \brief retrives the edge length between two points;
		 *
		 * This method returns the euclidean distance between the points
		 * represented by the vertices \p v and \p w. If the two vertices are not
		 * connected -1 is returned.
		 * 
		 * \param v the source point on the surface
		 * \param w the target point on the surface
		 * 
		 * \return the euclidean distance between the vertices v and w or -1 if not
		 *         connected.
		 */
		template<class IterA, class IterB>
		double getDistance(const IterA& v, const IterB& w) const;
		//@}

    /** 
     * \brief returns the predecessor of a vertex in the graph
     *
     * The predecessor is determined by an on-surface distance calculation and
     * describes that point on the surface that is next on the path to the
     * central point from which the on-surface distances are calculated.
     * 
     * \param v a vertex iterator
     * 
     * \return the predecessor of the vertex \p v. The predecessor is 
     *         given as another vertexIter.
     *
     * \sa getPredecessorPoint(const vertexIter&)
     */
    vertexIter getPredecessor(const vertexIter& v); 
    /** 
     * \brief returns the predecessor of a vertex in the graph
     *
     * This allows a quick access to the predecessor point on the surface,
     * because no lookup for vertex iterators is necessary. For further
     * information see getPredecessor(const vertexIter&).
     * 
     * \param v a vertex iterator
     * 
     * \return the predecessor of the vertex \p v. The predecessor is 
     *         given as a point iterator of the surface.
     *
     * \sa getPredecessorPoint(const vertexIter&)
     */
    SurfaceEntity::points_it getPredecessorPoint(const vertexIter& v) const; 
    //@}

    RbDblV getCenter() const;
    RbDblV getExtensions() const;

		//**************************************************************************
		// ITERATORS
		//**************************************************************************

		//@{
		/// \name vertex point iterators
		/**
		 * \brief iterator for the selected point vertex sequence
		 *
		 * The class provides begin and end iterators for the point sequence for
		 * mutable access to the single, selected points in the graph view.
		 */
		vertexIter begin();
		vertexIter end();
		//@}

		/** 
		 * \brief retrieves the selection for the \p point.
		 * 
		 * \param point a point iterator in the associated surface entity
		 * 
		 * \return a vertexIter pointing to the corresponding vertex of \p point or
		 *         end() if the point is not selected by the view
		 */
		vertexIter selectionOf(SurfaceEntity::points_it point);

		/**
		 * \brief returns two pointers to the adjacency of a specific vertex point
		 *
		 * \param vertex a vertex iterator pointing to the central vertex
		 *
		 * \return a pair of adjacencyIter objects that point to the begin and end
		 *         of the adjacency list of \p vertex
		 */
		std::pair<adjacencyIter, adjacencyIter> adjacency(const vertexIter& vertex);

  protected:
    
    //**************************************************************************
    // PROTECTED METHODS
    //**************************************************************************

    virtual SurfaceView* createView() const { return new GraphView(); }

	private:

		friend class vertexIter;
		friend class adjacencyIter;

		//****************************************************************************
		// POINT DATA
		//****************************************************************************

		//@{
		/**
		 * \brief adding and deleting columns is not allowed from a graph view
		 *
		 * \sa SurfaceView::addColumn(const std::string&, double),
		 *     SurfaceView::addColumn(const std::string&, const RbDblV&)
		 *     SurfaceView::deleteColumn(size_t)
		 */
		size_t addColumn(const std::string& name, double def=0.0) {}
		size_t addColumn(const std::string& name, const RbDblV& def) {}
		void deleteColumn(size_t col) {}
    //@}
		

		//**************************************************************************
		// PRIVATE TYPE DEFINITIONS
		//**************************************************************************

		/**
		 * \brief a maptype for a fast access to the selected points + iterators
		 */
		typedef map<SurfaceEntity::points_it, 
								graphview_traits::vertex_it, 
								lt_iterator<SurfaceEntity::points_it> > pointmap_t;
		typedef pointmap_t::iterator pointmap_it;
		typedef pointmap_t::const_iterator pointmap_cit;

		//**************************************************************************
		// PRIVATE METHODS
		//**************************************************************************

		/**
		 * \brief a wrapper for the calculation of point distances
		 *
		 * The method returns the distance between the two surface points that are
		 * represented by the given graph vertices.
		 *
		 * \param a, b two graph vertices
		 *
		 * \return the euclidean distance of the two points
		 */
		double calcDistance(graphview_traits::vertex_t a, 
												graphview_traits::vertex_t b) const;

		/**
		 * \brief internal function for on surface distance calculation
		 *
		 * This function is called by the three public calcOnSurfaceDistance()
		 * methods. It takes a simple vertex descriptor as the parameter for the \p
		 * center.
		 *
		 * \param center    the central vertex (either by index, as 
		 *                  SurfaceEntity::points_it or  as a vertexIter
		 */
		void doOnSurfaceDistances(graphview_traits::vertex_t v);

		/** 
		 * \brief assigns an index to all vertices in the graph
		 *
		 * This operation is necessary before any algorithm is invoked that needs a
		 * unique, continuous graph index.
		 */
		void generateIndex();

		//**************************************************************************
		// PRIVATE MEMEBERS
		//**************************************************************************

		/**
		 * \brief the internal graph representation
		 */
		graphview_traits::graph_t graph_;

		/**
		 * \brief the internal map for fast point->vertex access
		 */
		pointmap_t pointmap_;

		/**
		 * \brief the boost::propertymap that contains the iterators into the point
		 * entity
		 */
		graphview_traits::vdatamap_t iteratormap_;

		/**
		 * \brief the boost::propertymap for the vertex distances
		 */
		graphview_traits::vdistmap_t distmap_;

		/**
		 * \brief the boost::propertymap for the vertex predecessors
		 */
		graphview_traits::vpremap_t premap_;

		/**
		 * \brief the boost::propertymap that contains the edge weights
		 */
		graphview_traits::elenghtmap_t elengthmap_;

		//**************************************************************************
		// FLAGS
		//**************************************************************************

		/**
		 * \brief specifies if weights should be calculated for each new edge
		 * automatically
		 */
		bool oftweightcalc_;
	};

	//****************************************************************************
	inline bool GraphView::setSurfaceEntity(SurfaceEntity& surface)
  //****************************************************************************
	{
	  if (SurfaceView::setSurfaceEntity(surface))
		{
			surface_->lock();
			return true;
		}
		else
			return false;
	}

  //****************************************************************************
	inline SurfaceEntity::points_it GraphView::operator[](size_t ix) const
  //****************************************************************************
	{
		if (ix<num_vertices(graph_))
			return get(iteratormap_, vertex(ix, graph_));
		else
			throw OutOfRange_e("vertex index out of range!");
	}

  //****************************************************************************
	inline GraphView::vertexIter GraphView::begin()
  //****************************************************************************
	{
		graphview_traits::vertex_it b,e;
		boost::tie(b,e) = boost::vertices(graph_);
		
		return vertexIter(*this, b); 
	}

  //****************************************************************************
	inline GraphView::vertexIter GraphView::end()
  //****************************************************************************
	{
		graphview_traits::vertex_it b,e;
		boost::tie(b,e) = boost::vertices(graph_);
		
		return vertexIter(*this, e); 
	}

	//****************************************************************************
	inline std::pair<GraphView::adjacencyIter, GraphView::adjacencyIter> 
	GraphView::adjacency(const vertexIter& vertex)
  //****************************************************************************
	{
		graphview_traits::adjacency_it b,e;		
		boost::tie(b,e) = boost::adjacent_vertices(*vertex.impl_, graph_);
		return std::make_pair(adjacencyIter(*this, b),
													adjacencyIter(*this, e));
	}

	//@{
	/** 
	 * \brief a comparison operator between adjacency and vertex iterators
	 * 
	 * \param vit a vertex iterator of a graph view
	 * \param ait an adjacency iterator of a graph view
	 * 
	 * \return \c true if both iterators represent the same surface entity point,
	 *         \c false otherwise
	 */
	
	/*
	inline bool operator==(const GraphView::vertexIter& vit, 
												 const GraphView::adjacencyIter& ait)
  //****************************************************************************
	{
		return vit.operator->() == ait.operator->();
	}

	inline bool operator==(const GraphView::adjacencyIter& ait,
												 const GraphView::vertexIter& vit)
  //****************************************************************************
	{
		return vit.operator->() == ait.operator->();
	}
	*/
	//@}

//******************************************************************************
template<class IterA, class IterB>
double GraphView::getDistance(const IterA& v, const IterB& w) const
//******************************************************************************
{
	graphview_traits::edge_t e;
	bool present;
	tie(e, present) = edge(*v.impl_, *w.impl_, graph_);

	if (present)
		return get(elengthmap_, e);
	else 
		return -1.0;
}



#ifdef HAVE_NAMESPACES
};
#endif 

#endif // GRAPHVIEW_HH
