/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file graphview_traits.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 25 Mar 2004, last CVS commit 2004/03/25 18:10:52
 */

#ifndef GRAPHVIEW_TRAITS_HH
#define GRAPHVIEW_TRAITS_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include <boost/config.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map.hpp>
#include <boost/graph/adjacency_list.hpp>

#include "surface.hh"

#ifdef HAVE_NAMESPACES
namespace boost
{ 
#endif
	enum vertex_data_t { 
    vertex_data = 300,
  };
  BOOST_INSTALL_PROPERTY(vertex, data);
#ifdef HAVE_NAMESPACES
};

namespace Surface
{
#endif 

	/**
	 * \brief trait class for the internal representation of the GraphView object.
	 */
	struct graphview_traits
	{
		//**************************************************************************
    // VERTEX AND EDGE PROPERTIES
    //**************************************************************************

    /// \brief a color property for different algorithms
    typedef boost::property<boost::vertex_color_t, int> vcolor_t;
		
    /// \brief the distances between the vertices and a central vertex
    typedef boost::property<boost::vertex_distance_t, double, 
														 vcolor_t> vdistance_t;

    /// \brief surface entity iterators representing the predecessor of a vertex
		typedef boost::property<boost::vertex_predecessor_t, void*, vdistance_t> vpre_t;

    /// \brief surface entity iterators represent the point data
		typedef boost::property<boost::vertex_data_t, 
														SurfaceEntity::points_it, vpre_t> vdata_t;

		/// \brief index property for each vertex
		typedef boost::property<boost::vertex_index_t, int, vdata_t> vindex_t;

		/**
		 * \brief euklidean distance between each adjecent pair of vertices
		 *
		 * The length of the edges is stored as the weights (colors) of the mesh's
		 * edges.
		 */
		typedef boost::property<boost::edge_weight_t, double> elength_t;

		/// \brief the edge index property
		typedef boost::property<boost::edge_index_t, int, elength_t> eindex_t;

		//**************************************************************************
		// GRAPH, EDGE and VERTEX plus ITERATORS
		//**************************************************************************

		/**
		 * \brief the type of the internal graphview adjacency-list graph
		 *
		 * The internal graph representation of the graph view is a directed graph
		 * with edge and vertex vectors.
		 */
		typedef boost::adjacency_list<boost::multisetS, boost::listS, boost::directedS, 
																	vindex_t, eindex_t> graph_t;

		/**
		 * \brief typedefinition of the vertex descriptor
		 */
		typedef boost::graph_traits<graph_t>::vertex_descriptor vertex_t;

		/**
		 * \brief typedefinition of the edge descriptor
		 */
		typedef boost::graph_traits<graph_t>::edge_descriptor edge_t;

		/**
		 * \brief iterator type for the vertices
		 */
		typedef boost::graph_traits<graph_t>::vertex_iterator vertex_it;
  
		/**
		 * \brief graph edge iterator type
		 */
		typedef boost::graph_traits<graph_t>::edge_iterator edge_it;

		/**
		 * \brief out-edge iterator type
		 */
		typedef boost::graph_traits<graph_t>::out_edge_iterator edge_oit;

		/**
		 * \brief adjancecy iterator
		 */
		typedef boost::graph_traits<graph_t>::adjacency_iterator adjacency_it;

		//**************************************************************************
		// PROPERTY
		//**************************************************************************

		/**
		 * \brief property map for the vertex color
		 */
		typedef boost::property_map<graph_t, 
																boost::vertex_color_t>::type vcolormap_t;

		/**
		 * \brief property map for the vertex distances
		 */
		typedef boost::property_map<graph_t, 
																boost::vertex_distance_t>::type vdistmap_t;

		/**
		 * \brief property map for the predecessor points
		 */
		typedef boost::property_map<graph_t, boost::vertex_predecessor_t>::type 
    vpremap_t;

		/**
		 * \brief property map for the surface points
		 */
		typedef boost::property_map<graph_t, boost::vertex_data_t>::type vdatamap_t;

		/**
		 * \brief property map for the vertex indices
		 */
		typedef boost::property_map<graph_t, 
																boost::vertex_index_t>::type vindexmap_t;

		/**
		 * \brief property map for the edge lengths (weights)
		 */
		typedef boost::property_map<graph_t, 
																boost::edge_weight_t>::type elenghtmap_t;

		/**
		 * \brief property map for the edge lengths (weights)
		 */
		typedef boost::property_map<graph_t, boost::edge_index_t>::type eindexmap_t;
	};

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // GRAPHVIEW_TRAITS_HH
