/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "msmsio.hh"

#ifdef HAVE_STD
#include <fstream>
#include <sstream>
#include <memory>
#else
#include <map.h>
#include <fstream.h>
#include <sstream.h>
#include <memory.h>
#endif

#include "surface.hh"
#include "point.hh"
#include "triangle.hh"
#include "vectorview.hh"
#include "fileformatfactory.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace RazorBack;
using namespace Surface;
#endif 

REGISTER_IOCLASS(msms, MSMSIO, "Michael Sanner's molecular surface files");

//#include <iostream>

//******************************************************************************
SurfaceEntity* MSMSIO::read(istream& astream)
//******************************************************************************
{
	/* INITIALIZATION */
	SurfaceIO::read(astream);

  ReadingData readdata(new SurfaceEntity);

	//the line buffer
	char line[200];

	//a map to remember the point indices for the right iterators
	map<int,SurfaceEntity::points_it> indexmap;

  parseHeader(astream, readdata);
  
	/* read all the vertices line by line */
	int master = 0;
  for(int v=0;v<readdata.num_points;++v)
	{
    parsePoint(astream, readdata);
    SurfaceEntity::points_it p = readdata.surface->end();
    --p;

    //checking for doubletts
		if (doublettCheck(v,*p))
			//if the point is not unique delete it from the surface
			readdata.surface->deletePoint(p);
		else
		{
			//assign the master index and put rember in the indexmap
			p->setMasterIndex(master);
			indexmap[master] = p;
			//increase the master
			++master;
		}
	}

	/* TRIANGLE PART */
  
  parseHeader2(astream, readdata);

	/* read all the triangles line by line */
	for(int t=0;t<readdata.num_triangles;++t)
	{
		int i;

    TriangleData tdata = parseTriangle(astream);
		checkTrianglePoints(readdata.num_points, tdata);
			
		for(i=A;i<=C;++i)
			tdata.points[i] = mDoubletts[tdata.points[i]];
		
		
		if (!isDegenerated(tdata))
		{
			SurfaceEntity::points_it a = indexmap[tdata.points[0]];
			SurfaceEntity::points_it b = indexmap[tdata.points[1]];
			SurfaceEntity::points_it c = indexmap[tdata.points[2]];
			
			SurfaceEntity::triangles_it t = readdata.surface->addTriangle(a, b, c);
			t->setFaceType(facetype_t(tdata.triangle_face));
			t->setFaceNum(--tdata.face_number);
		}
	}
	
	return readdata.surface.release();
}

//******************************************************************************
bool MSMSIO::write (ostream & astream,	VectorView& view)
//******************************************************************************
{
	/* INITIALIZATION */
	char line[200];

	map<SurfaceEntity::points_it, size_t,lt_iterator<SurfaceEntity::points_it> > 
		indexmap;

  /* VERTEX PART */

	//write two dummy header lines
	astream << endl << endl;

	//write headline
  //#points #sphere probe_r density
  sprintf(line,"%7d %7d %2.3f %2.3f\n",
					view.getNPoints(),
					view.getSphereCount(),
					view.getPointDensity(),
					view.getProbeRadius());
  astream << line;

	//write each point
	int ix = 0;
	for(VectorView::pointIter p=view.begin();p!=view.end();++p, ++ix)
  {
		indexmap[p] = ix;

		//write the data for each vertex
		sprintf(line, "%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %7d %7d %2d\n",
						p->getPosition()[X],
						p->getPosition()[Y],
						p->getPosition()[Z],
						p->getNormal()[X],
						p->getNormal()[Y],
						p->getNormal()[Z],
						1,
						p->getNearestAtom()+1,
						(int) p->getFaceType());

    astream << line;
	}

	/* TRIANGLE PART */

	//write two dummy header lines
	astream << endl << endl;

	//write headline
  //#triangles #sphere probe_r density
  sprintf(line,"%7d %7d %2.3f %2.3f\n",
					view.getNTriangles(),
					view.getSphereCount(),
					view.getPointDensity(),
					view.getProbeRadius());
  astream << line;

	//write each point
	VectorView::triangleIter t;
	for(t=view.beginTriangles();t!=view.endTriangles();++t)
  {

    sprintf(line, "%6d %6d %6d %2d %6d\n",
						indexmap[t->getVertex(A)]+1,
						indexmap[t->getVertex(B)]+1,
						indexmap[t->getVertex(C)]+1,
						(int) t->getFaceType(),
						t->getFaceNum()+1);
    astream << line;
	}	

	return true;
}

//******************************************************************************
void MSMSIO::parseHeader(istream& astream, ReadingData& readdata)
//******************************************************************************
{
  /* POINT HEADER */

	//the line buffer
	char line[512];

  // elements of the header
	int nspheres;
	double probe;
	double density;
  
  //read comment header lines
  astream.getline(line,512);
  astream.getline(line,512);

  // Read third header line
  astream.getline(line,512);
  stringstream buffer(line);

  buffer >> readdata.num_points >> nspheres >> density >> probe;

	readdata.surface->setSphereCount(nspheres);
	readdata.surface->setProbeRadius(probe);
	readdata.surface->setPointDensity(density);
}

//******************************************************************************
void MSMSIO::parsePoint(istream& astream, ReadingData& readdata)
//******************************************************************************
{
  /* READ A POINT LINE */
  
 	//the line buffer
	char line[512];

  //pointdata
  Vector_<double> position(3);
  Vector_<double> normal(3);
  int face_num;
  int closest_sphere;
  int face;

  astream.getline(line,512);
  stringstream buffer(line);
  
  buffer >> position[0] >> position[1] >> position[2] 
         >> normal[0] >> normal[1] >> normal[2]
         >> face_num >> closest_sphere >> face;

  //setup the point
  SurfaceEntity::points_it p = readdata.surface->addPoint();
  
  p->setPosition(position);
  p->setNormal(normal);
  //closest sphere 1-based!
  p->setNearestAtom(--closest_sphere); 
  p->setFaceType(facetype_t(face));
}

//******************************************************************************
void MSMSIO::parseHeader2(istream& astream, ReadingData& readdata)
//******************************************************************************
{
  /* TRIANGLE HEADER */

	//the line buffer
	char line[512];

  //read comment header lines
  astream.getline(line,512);
  astream.getline(line,512);

  // Read third header line
  astream.getline(line,512);
  stringstream buffer(line);

  buffer >> readdata.num_triangles;
}

//******************************************************************************
MSMSIO::TriangleData MSMSIO::parseTriangle(std::istream& astream)
//******************************************************************************
{
  char line[512];
  TriangleData tdata;
  
  //read the next line
  astream.getline(line,512);
  stringstream buffer(line);

  buffer >> tdata.points[0] >> tdata.points[1] >> tdata.points[2]
         >> tdata.triangle_face >> tdata.face_number;

  //the MSMS triangle indices are 1-based;
  for(int i=0;i<3;++i) --tdata.points[i];

  return tdata;
}
