/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef MSMSIO_HH
#define MSMSIO_HH

/**
 * \file msmsio.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 11 Feb 2004, last CVS commit 2004/02/11 08:45:22
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <memory>
#else
#  include <memory.h>
#endif

#include "surfaceio.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	/**
	 * \brief the MSMS file reader/writer
	 *
	 * This class performs write and read operation for the files generated and
	 * used by the MSMS program of Michael Sanner.
	 *
	 * For further information about the file format see \ref msms_fileformat
	 * "The MSMS File Format".
	 */
	class MSMSIO : public SurfaceIO
	{
	public:
			
		//**************************************************************************
		// CONSTRUCTOR & DESTRUCTOR
		//**************************************************************************

		/**
		 * \brief default constructor
		 *
		 * simply calls the SurfaceIO default constructor
		 *
		 * \param checkdoubletts turns doublett checking on or off
		 */
		MSMSIO(bool checkdoubletts=false) : SurfaceIO(checkdoubletts) { }

    /** 
     * \brief factory method
     * 
     * \return a fresh instance of a MSMSIO class
     */
    static SurfaceIO* createInstance() { return new MSMSIO(); }

		//**************************************************************************
		// OPERATIONS
		//**************************************************************************

		using SurfaceIO::read;
		SurfaceEntity* read(std::istream& astream);
    
		using SurfaceIO::write;
		bool write (std::ostream & astream, VectorView& view);

  private:

    struct ReadingData
    {
      size_t num_points;
      size_t num_triangles;

      size_t first_point;
      size_t first_triangle;
      
      auto_ptr<SurfaceEntity> surface;

      ReadingData(SurfaceEntity* entity = 0)
          : surface(entity), num_points(0), num_triangles(0), 
            first_point(0), first_triangle(0) { } 
    };

    //**************************************************************************
    // OPERATIONS FOR READING
    //**************************************************************************

    void parseHeader(std::istream& astream, ReadingData& readdata);
    void parsePoint(std::istream& astream, ReadingData& readdata);
    void parseHeader2(std::istream& astream, ReadingData& readdata);
    TriangleData parseTriangle(std::istream& astream);
	};

#ifdef HAVE_NAMESPACES
};
#endif 

  /**
   * \page msms_fileformat The MSMS file formats
   *
   * The MSMS file format is created from Michael Sanners molecular surface tool.
   * It is used by his own molecular surface viewer, which is available on the
   * SGI/IRIX platform.
   *
   * Each MSMS object consists of two files:
   * \li The file containing the vertex coordinates and normals  
   *     \ref vert_file "*.vert File":
   * \li The file containing the face descriptions
   *     \ref face_file "*.face File":
   *
   * \section vert_file Format of MSMS Vert file.
   *
   * Format description of coordinates file (by Torsten Schindler):
   *
   * 3 header lines followed by one vertex per line:
   * Each vertex line consists of 
   * \li the coordinates (x,y,z),
   * \li the normals (nx,ny,nz) 
   * \li number of the face to which the vertex belongs.
   * \li (1 based) index of the closest sphere.
   * \li a flag denoting 
   * <dl>
   *   <dt>1</dt> <dd>for vertices which belong to toric reentrant faces
   *            (including vertices of the analytical surface)</dd>
   *   <dt>2</dt> <dd>for vertices inside reentrant faces and </dd>
   *   <dt>3</dt> <dd>forvertices  inside  contact faces</dd>
   * </dl>
   *
   * "%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %7d %7d %2d"
   *
   * \section face_file Format of MSMS face file.
   *
   * Format description of coordinates file (by Torsten Schindler):
   *
   * The face file contains three header lines followed by one triangle per line
   * First header  line: comment
   * Second header line: comment about the content of the third line
   * Third header  line: number of triangles, number of spheres in the set,
   *                     triangulation density and probe sphere radius
   * Triangle data lines:
   * \li first three number (1 based) vertex indices
   * \li next field: 
   *   <dl>
   *   <dt>1</dt><dd>for a triangle in a toric reentrant face</dd>
   *   <dt>2</dt><dd>for a triangle in a spheric reentrant face and</dd>
   *   <dt>3</dt><dd>for a triangle in a contact face</dd>
   *   </dl>
   * \li last number in line: (1 based) face number in the analytical description
   *     of the solvent excluded surface.
   *
   * "%6d %6d %6d %2d %6d"
   */

#endif // MSMSIO_HH
