/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "offio.hh"

#ifdef HAVE_STD
#include <fstream>
#include <sstream>
#include <memory>
#else
#include <map.h>
#include <fstream.h>
#include <sstream.h>
#include <memory.h>
#endif

#include "surface.hh"
#include "point.hh"
#include "triangle.hh"
#include "vectorview.hh"
#include "fileformatfactory.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace RazorBack;
using namespace Surface;
#endif 

REGISTER_IOCLASS(off, OffIO, "geomview polyhedra description format");

//#include <iostream>

//******************************************************************************
SurfaceEntity* OffIO::read(istream& astream)
//******************************************************************************
{
	SurfaceIO::read(astream);

  char line_buffer[200];

  // number of vertices, number of spheres, probe radius and density
  int nvertex, nfaces, nedges;

  Vector_<double> xyz(3);  // Coordinates
  Vector_<double> nxyz(3); // Normals

  bool read_normals = false; //reading the surface normals if available

	//a map to remember the point indices for the right iterators
	map<int,SurfaceEntity::points_it> indexmap;

  // Read header lines
  string format_descriptor;

  //first line
  astream.getline(line_buffer,200);
  stringstream head1_stream(line_buffer);

  head1_stream >> format_descriptor;

  int pos = format_descriptor.find("OFF");
  if (pos==format_descriptor.npos)
    throw WrongFormat_e(0, format_descriptor, "");

  auto_ptr<SurfaceEntity> surface(new SurfaceEntity());

  //testing for available surface formats
  for(;pos>=0;--pos)
    read_normals = read_normals | (format_descriptor[pos]=='N');

  astream.getline(line_buffer,200);
  stringstream head2_stream(line_buffer);

  // number of vertices, number of faces and number of edges
  head2_stream >> nvertex >> nfaces >> nedges;

  
  /** READ VERTICES **/

	//the actual master index
	int master = 0;
  for (int v=0;v<nvertex;v++)
  {
    // Read coordinates, normals, closest_sphere and face of the vertices
    astream.getline(line_buffer,200);
    stringstream string_buffer(line_buffer);

    //read the coordinates
    string_buffer >> xyz[0] >> xyz[1] >> xyz[2];
    
    //read the surface normals
    if (read_normals)
      string_buffer >> nxyz[0] >> nxyz[1] >> nxyz[2];
    else
      fill(nxyz.begin(),nxyz.end(),0);

		//setup the point
		SurfaceEntity::points_it p = surface->addPoint();

		p->setPosition(xyz);
		p->setNormal(nxyz);
		p->setFaceType(ftUnknown);
		p->setNearestAtom(0);
		
    //checking for doubletts
		if (doublettCheck(v,*p))
		{
			//if the point is not unique delete it from the surface
			surface->deletePoint(p);
		}
		else
		{
			//assign the master index and put rember in the indexmap
			p->setMasterIndex(master);
			indexmap[master] = p;
			//increase the master
			++master;
		}
	}

	/** READ TRIANGLES **/

  for (int t=0;t<nfaces;t++)
	{

		astream.getline(line_buffer,200);

		stringstream string_buffer(line_buffer);

		int triangle_face, face_number;
		TriangleData tdata;
    
		int nfaceVertices = 0;
		int i = 0;

		string_buffer >> nfaceVertices 
									>> tdata.points[0] >> tdata.points[1] >> tdata.points[2];
	
		checkTrianglePoints(nvertex, tdata);
			
		for(i=A;i<=C;++i)
			tdata.points[i] = mDoubletts[tdata.points[i]];
		
		
		if (!isDegenerated(tdata))
		{
			SurfaceEntity::points_it a = indexmap[tdata.points[0]];
			SurfaceEntity::points_it b = indexmap[tdata.points[1]];
			SurfaceEntity::points_it c = indexmap[tdata.points[2]];
			
			SurfaceEntity::triangles_it t = surface->addTriangle(a, b, c);
		}
	}
	
	return surface.release();
}

//******************************************************************************
bool OffIO::write (ostream & astream,	VectorView& view)
//******************************************************************************
{
	/* INITIALIZATION */
	char line[200];

	map<SurfaceEntity::points_it, size_t,lt_iterator<SurfaceEntity::points_it> > 
		indexmap;

	/* HEAD */

	//write format descriptor
  astream << "NOFF" << endl;
  
  //#vertex #vertices, #faces, #edges
  sprintf(line,"%7d %7d %7d\n",
					view.getNPoints(),
					view.getNTriangles(),
					getEdgeNumber(view));

  astream << line;

  /* VERTEX PART */

	//write each point
	int ix = 0;
	for(VectorView::pointIter p=view.begin();p!=view.end();++p, ++ix)
  {
		indexmap[p] = ix;

		//write the data for each vertex
    sprintf(line, "%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f\n",
						p->getPosition()[X],
						p->getPosition()[Y],
						p->getPosition()[Z],
						p->getNormal()[X],
						p->getNormal()[Y],
						p->getNormal()[Z]);

    astream << line;
	}

	/* TRIANGLE PART */

	//write each point
	VectorView::triangleIter t;
	for(t=view.beginTriangles();t!=view.endTriangles();++t)
  {
    sprintf(line, "3 %6d %6d %6d\n",
						indexmap[t->getVertex(A)],
						indexmap[t->getVertex(B)],
						indexmap[t->getVertex(C)]);
    astream << line;
	}	

	return true;
}

//******************************************************************************
size_t OffIO::getEdgeNumber(VectorView& view) const
//******************************************************************************
{
	//WILL BE REPLACED BY GRAPH VIEW AND CORRECT NUMBER OF EDGES
	return view.getNTriangles()*2;
}
