/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef OFFIO_HH
#define OFFIO_HH

/**
 * \file offio.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 11 Feb 2004, last CVS commit 2004/02/11 08:45:22
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "surfaceio.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	/**
	 * \brief the OFF file reader/writer
	 *
	 * This class performs write and read operation for a class of files generated and
	 * used by the GEOMVIEW program for the definition of triangulated surfaces.
	 *
	 * For further information about the file format see \ref off_fileformat
	 * "The OFF File Format".
	 */
	class OffIO : public SurfaceIO
	{
	public:
			
		//**************************************************************************
		// CONSTRUCTOR & DESTRUCTOR
		//**************************************************************************

		/**
		 * \brief default constructor
		 *
		 * simply calls the SurfaceIO default constructor
		 *
		 * \param checkdoubletts turns doublett checking on or off
		 */
		OffIO(bool checkdoubletts=false) : SurfaceIO(checkdoubletts) { }
			
    /** 
     * \brief factory method
     * 
     * \return a fresh instance of a OffIO class
     */
    static SurfaceIO* createInstance() { return new OffIO(); }

		//**************************************************************************
		// OPERATIONS
		//**************************************************************************

		using SurfaceIO::read;
		SurfaceEntity* read(std::istream& astream);
    
		using SurfaceIO::write;
		bool write (std::ostream & astream, VectorView& view);

	private:

		/**
		 * \brief calculates the number of edges in a given view
		 *
		 * This method generates the correct number of edges selected in the view
		 *
		 * \return the number of edges selected by the view
		 *
		 * \param view a reference to a VectorView
		 */
		size_t getEdgeNumber(VectorView& view) const;

	};

#ifdef HAVE_NAMESPACES
};
#endif 

/**
 * \page off_fileformat The OFF file format
 *
 * This file format is used by the <a href="http://www.geomview.org/">Geomview</a>
 * viusalization program. A detailed description can therefore be found under:
 * <a href="http://www.geomview.org/docs/html/geomview_41.html#SEC44">OFF Files</a>
 *
 * The following descripton is the peticular form of OFF files, which are created
 * by the writer:
 *
 * \subsection offfile_header File Header:
 * The file consists of two header lines:
 * \li The first containing the string \c CNOFF: <br>
 * \c C denotes the usage of the color index to state the kind of each point <br>
 * \c N denotes that the vertex normals are saved <br>
 * \c OFF specifies that the off file format is used <br>
 *
 * \li The second line contains the number of vertices, faces and edges:<br>
 * \em num_vertices, \em num_faces, \em num_edges
 * 
 * \subsection offfile_vertices Vertex Lines:
 * The vertex lines consist of the following data:
 * POS_X POS_Y POS_Z NORM_X NORM_Y NORM_Z VERTEX_STATE
 *
 * \param POS_[XYZ] position of the vertex in 3D coordinates
 * \param NORM_[XYZ] direction of the normal vector at the verex
 *
 * \subsection offfile_faces Face Lines:
 * The face lines consist of the following data:
 * 3 VERTEX_1 VERTEX_2 VERTEX_3
 *
 * \param 3 count of vertices in a face: since Trimesh handles triangulated
 *        meshes, this must be 3
 * \param VERTEX_[123] indices of the three vertices forming the face.
 *
 * \note The vertices must be given in the correct order. (Looking from on them
 * from the outside of the surface.)
 * 
 */

#endif // OFFIO_HH
