/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "patchbuilder.hh"

#ifdef HAVE_STD
#  include <algorithm>
#else
#  include <algo.h>
#endif

#include "graphview.hh"
#include "haspoint.hh"
#include "spherebuilder.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
#endif 

//#include <iostream>

//******************************************************************************
template<class Out>
size_t getPossibleTriangles(SurfaceEntity* const surface, 
														const GraphView* const selection,
														Out ins)
//******************************************************************************
{
	size_t c = 0;
	SurfaceEntity::triangles_it t;
	for(t = surface->beginTriangles();t!=surface->endTriangles();++t)
	{
		if (selection->getNEdges(t)>=3) 
		{
			*ins++ = t;
			++c;
		}
	}
}

//******************************************************************************
void getBorderEdge(SurfaceEntity::points_it& source, 
									 SurfaceEntity::points_it& target, 
									 const SurfaceEntity::triangles_it& triangle,
									 const GraphView* const selection)
//******************************************************************************
{
	for(int i=A;i<=C;++i)
	{
		source = triangle->getVertex(i);
		target = triangle->getVertex(SUCC[i]);
		
		//the border edge is found if no edge from target to source (resulting from
		//an adjacent triangle) can be detected
		if (!selection->connected(target, source))
			break;
	}
}

//******************************************************************************
void PatchBuilder::initialize(SurfaceEntity& surface)
//******************************************************************************
{
	ViewBuilder::initialize(surface);
	view_ = new PatchView(surface);
	
	unselected_vits.clear();
	margin_vits.clear();
}

//******************************************************************************
void PatchBuilder::deselectPoints()
//******************************************************************************
{
	//deselect the unselected points
	for(vlist_it uv = unselected_vits.begin();uv!=unselected_vits.end();++uv)
		selection->deselectPoint(*uv);
}

//******************************************************************************
GraphView* PatchBuilder::preSelection()
//******************************************************************************
{
	//the preselection sphere for the graph containing all points within radius_
	//in 3D space.
	SphereBuilder<GraphView> preselector(radius_ * (1+ margin_) + DBL_EPSILON, 
																						 center_->getPosition());
	//build the preselection graph.
	return dynamic_cast<GraphView*>(surface_->createView(preselector));
}

//******************************************************************************
void PatchBuilder::pointSelection()
//******************************************************************************
{
	//calculate the on-surface distances from the central point
	selection->compOnSurfaceDistances(center_);
	
	//identify the unselected and margin vertices
	for(GraphView::vertexIter vit=selection->begin();vit!=selection->end();++vit)
	{
	  double surfacedistance = selection->getDistance(vit);
		if ((surfacedistance<(radius_ * (1 + margin_) + DBL_EPSILON)) &&
				(surfacedistance>(radius_ + DBL_EPSILON)))
			margin_vits.push_back(vit);
		else if (surfacedistance>(radius_ + margin_ + DBL_EPSILON))
			unselected_vits.push_back(vit);
	}
}

//******************************************************************************
void PatchBuilder::checkMargin()
//******************************************************************************
{
	//check if any margin vertex has a selected neighbor
	for(vlist_cit mv = margin_vits.begin();mv!=margin_vits.end();++mv)
	{
		bool selected = false;
		//search the adjacency of mv
		GraphView::adjacencyIter begin_adj, end_adj;
		boost::tie(begin_adj, end_adj) = selection->adjacency(*mv);
		for(GraphView::adjacencyIter av = begin_adj;av!=end_adj;++av)
			//check for selected neighbors via the on-surface distance (more
			//efficient)
			if (selection->getDistance(av)<(radius_ + DBL_EPSILON))	
			{
				selected = true;
				break;
			}
		
		if (!selected) unselected_vits.push_back(*mv);
	}
}

//******************************************************************************
void PatchBuilder::select()
//******************************************************************************
{
	if ((radius_ < DBL_EPSILON) || (!surface_)) return;
	
	//the list of dangling and border triangles
	tlist_t possible_triangles, dangling_triangles, border_triangles;
	tlist_it t;

	selection = preSelection();
	pointSelection();

	getPossibleTriangles(surface_, selection, back_inserter(possible_triangles));
	
	checkMargin();

	//deselect the unselected points
	for(vlist_it uv = unselected_vits.begin();uv!=unselected_vits.end();++uv)
		selection->deselectPoint(*uv);

	//deselect the yet unselected triangles
	for(t=possible_triangles.begin();t!=possible_triangles.end();)
	{
		tlist_it titer = t++;
		//deselect only if not all points of the triangle are within the patch
		if (selection->deselectTriangle(*titer,3)>0)
			//remove the triangle from the possible triangles if it is not selected
			possible_triangles.erase(titer);
	}

	// 2. REMOVING OF DANGLING TRIANGLES

	do
	{
		//clear the dangling and border triangles
		dangling_triangles.clear();
		border_triangles.clear();

		//identify all the dangling and border triangles and remove the yet unselected
		for(t=possible_triangles.begin();t!=possible_triangles.end();)
		{
			tlist_it titer = t++;
			
			size_t nedges = selection->getNEdges(*titer);

			if (nedges == 5)
			{
				//border triangle
				border_triangles.push_back(*titer);
			}
			else if (nedges < 5)
			{
				//dangling triangle
				dangling_triangles.push_back(*titer);
				//delete from the list of possible triangles
				possible_triangles.erase(titer);
			}
		}
		
		//deselect the dangling triangles
		for(t=dangling_triangles.begin();t!=dangling_triangles.end();++t)
			selection->deselectTriangle(*t);
		
	} while (dangling_triangles.size()>0);

	// 3. IDENTIFY THE BORDER VERTICES AND THE BORDER WALK
	
	while (border_triangles.size()>0)
	{
		list<SurfaceEntity::points_it> newborder;

		SurfaceEntity::points_it source = surface_->end();
		SurfaceEntity::points_it target = surface_->end();

		//start a new border at the first triangle in the border_triangles
		t = border_triangles.begin();


		//find the border edge and add the first border point
		getBorderEdge(source, target, *t, selection);
		newborder.push_back(source);

		while (true)
		{
			//add the new border point
			newborder.push_back(target);

			//delete the actual triangle from the border list
			border_triangles.erase(t);

			//look for the other triangle with the 
			t = find_if(border_triangles.begin(), border_triangles.end(), 
									HasPoint<SurfaceEntity::triangles_it>(target));

			if (t!=border_triangles.end())
			{
				//make the target the source and find the new target vertex
				source = target;
				target = (*t)->successor(source);
			}
			else
				//the border walk is finished
				break;
      
      if (target == newborder.front())
        //break also if we have reached the beginning
        break;
		} 

		//determine if this border is larger than the actual patch border and if so
		//make it the actual one
		if (newborder.size()>view_->border_.size())
			newborder.swap(view_->border_);

	}

	// 4. CREATE THE VIEW: COPY THE POINT AND TRIANGLES 
	
	//copy the selected points
	for(int i=0;i<selection->getNPoints();++i)
		view_->selectPoint((*selection)[i]);

	for(t = possible_triangles.begin();t!=possible_triangles.end();++t)
		view_->selectTriangle(*t);

  view_->rebuildBorderMap();

	view_->setRadius(radius_);
	view_->setMargin(margin_);
  view_->setCentralPoint(center_);

	//we do not need the selection any more
	delete selection;
}
