/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file patchbuilder.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 06 Apr 2004, last CVS commit 2004/04/06 13:04:07
 */

#ifndef PATCHBUILDER_HH
#define PATCHBUILDER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "viewbuilder.hh"
#include "surface.hh"
#include "graphview.hh"
#include "patchview.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	/**
	 * \brief builds a surface patch as a PatchView object
	 *
	 * The patch selection is described in detail in \ref patch_selection
	 * "The selection of circular surface patches".
	 */
	class PatchBuilder: public ViewBuilder
	{
	public:

		//**************************************************************************
		// CONSTRUCTOR
		//**************************************************************************
		
		/**
		 * \brief default constructor.
		 *
		 * It assigns initial, but not suitable values to radius_ and center_.
		 *
		 * \sa PatchBuilder(double, SurfaceEntity::points_it)
		 */
		PatchBuilder() 
				: ViewBuilder(), radius_(0.0), margin_(0.05), center_(), view_(0),
					unselected_vits(), margin_vits(), selection(0) { }
		/**
		 * \brief initializing constructor
		 *
		 * It assigns the given value to the radius_ and center_ of the patch
		 * so that building can start instantly.
		 *
		 * \sa PatchBuilder()
		 */
		PatchBuilder(const SurfaceEntity::points_it& center, 
								 double radius, double margin = 0.05)
				: ViewBuilder(), radius_(radius), margin_(margin), 
					center_(center), view_(0),
					unselected_vits(), margin_vits(), selection(0) { }

		//**************************************************************************
    // PROPERTY METHODS
    //**************************************************************************

		/**
		 * \brief returns an iterator to the central point of the patch
		 *
		 * \return a mutable reference to the sphere center position vector
		 *
		 * \sa center_, setCenter(SurfaceEntity::points_it)
		 */
		SurfaceEntity::points_it getCenter() const { return center_; }
		/**
		 * \brief assigns a new central point
		 *
		 * \param center the new central point. The components of this vector
		 *                 will be copied to the sphere center's position vector.
		 *
		 * \sa center_, getCenter()
		 */
		void setCenter(const SurfaceEntity::points_it& center) { center_ = center; } 

		/**
		 * \brief gives back the radius of the patch
		 *
		 * \return the radius of the patch
		 *
		 * \sa radius_, setRadius(double)
		 */
		double getRadius() const { return radius_; }
		/**
		 * \brief sets the patch radius to a new value
		 *
		 * \param radius the new patch radius
		 *
		 * \sa radius_, getRadius()
		 */
		void setRadius(double radius) { if (radius>0) radius_ = radius; }

		/**
		 * \brief gives back the margin of the patch radius
		 *
		 * \return the margin of the patch radius
		 *
		 * \sa margin_, setMargin(double)
		 */
		double getMargin() const { return margin_; }
		/**
		 * \brief sets the patch radius margin to a new value
		 *
		 * \param margin the new patch radius margin
		 *
		 * \sa margin_, getMargin()
		 */
		void setMargin(double margin) { if (margin>0) margin_ = margin; }

	  //**************************************************************************
		// BUILDER
		//**************************************************************************

		/**
		 * \brief initializes the View object
		 *
		 * This method creates the PatchView object.
		 *
		 * \param surface a reference to the SurfaceEntity
		 *
		 * \sa select()
		 */
		virtual void initialize(SurfaceEntity& surface);
		/**
		 * \brief builds the View
		 *
		 * The select method builds the surface patch based on a graph view that
		 * has been preselection by a selection sphere of the size of the patch
		 * radius.
		 */
		virtual void select();
		
		/**
		 * \brief retrieves the ready View
		 *
		 * \return a pointer to the constructed and ready View as a
		 *         SurfaceView.
		 */
		virtual SurfaceView* getView() const { 	return view_; }

	private:
		
		/**
		 * \brief the patch radius
		 *
		 * The patch radius is meassured on the surface.
		 *
		 * \sa getRadius(), setRadius()
		 */
		double radius_;

		/**
		 * \brief the patch radius margin
		 *
		 * The patch radius margin is given in percentage of the radius_ (e.g. 0.05
		 * means 5% of the radius_).
		 *
		 * \sa getMargin(), setMargin()
		 */
		double margin_;

		/**
		 * \brief the central point of the patch
		 *
		 * \sa getCenter(), setCenter() 
		 */
		SurfaceEntity::points_it center_;

 	  /**
		 * \brief the PatchView object
		 * \sa getView()
		 */
		PatchView* view_;

	private:

		//**************************************************************************
		// INTERNAL DATA
		//**************************************************************************

    /**
		 * \brief typedefs for vertexIter lists  
		 */
		typedef list<GraphView::vertexIter> vlist_t;
		typedef vlist_t::iterator vlist_it;
		typedef vlist_t::const_iterator vlist_cit;
		
		/**
		 * \brief typedes for the SurfaceEntity::triangles_it lists
		 */
		typedef list<SurfaceEntity::triangles_it> tlist_t;
		typedef tlist_t::iterator tlist_it;
		typedef tlist_t::const_iterator tlist_cit;

		/**
		 * \brief the internal graph view used for the patch selection
		 */
		GraphView* selection;

		/**
		 * \brief the list of unselected vertices
		 */
		vlist_t unselected_vits;

		/**
		 * \brief the list of vertices that are within the margin of the patch
		 */
		vlist_t margin_vits;

		//**************************************************************************
		// INTERNAL METHDOS
		//**************************************************************************

		/** 
		 * \brief deselects a set of points
		 */
		void deselectPoints();

		/** 
		 * \brief selects all the candidates for the patch
		 */
		void pointSelection();

		/** 
		 * \brief performs a preselection on the point and triangles based on
		 * Euclidean distance.
		 *
		 * \return a GraphView that contains all the preselected points and triangles
		 */
		GraphView* preSelection();
		
		/** 
		 * \brief evaluates the point in the margin of the patch. If they are
		 * connected to a selected point take them.
		 */
		void checkMargin();
		
	};

#ifdef HAVE_NAMESPACES
};
#endif 

  /**
   * \page patch_selection The selection of circular surface patches
	 *
   * \li preselect all points and triangles that are placed within tha patch
   * radius but meassured in euclidean distance into a GraphView
	 *
	 * \li calculate the on-surface distance from the center of the patch to all
	 * other points
	 *
	 * \li generater two lists: the first one contains points that are inside the
	 * radius and the second consists of all the points that are outside the
	 * patch radius but within the given margin. 
	 *
	 * \li all the points within the margin are check if they are adjacent to a
	 * selected point, they are added to the selected point set.
	 *
	 * \li remove all other points from the 
	 */
#endif // PATCHBUILDER_HH
