/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file PatchView.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 06 Apr 2004, last CVS commit 2004/04/06 13:07:54
 */

#ifndef PATCHVIEW_HH
#define PATCHVIEW_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#include <list>
#else
#include <list.h>
#endif

#include "vectorviewiterator.hh"
#include "vectorview.hh"


#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	class PatchBuilder;

	/**
	 * \brief represents a view onto a surface patch
	 *
	 * The only extension to the VectorView class that is introduced by the
	 * PatchView is an ordered list of border vertices (border_). The border is
	 * represented by a sequence of point iterators at the boundary of the patch
	 * sorted in counter clockwise order. The border list can be accessed as a
	 * sequence by the beginBorder() and endBorder() method. In the same way can
	 * the borderPosOf() method retrieve an iterator into the border sequence of
	 * any point of the entity, if this point is part of the border.
	 *
	 * The patch also stores the radius and the radius margin. These values are
	 * only saved within the patch but are not checked. They can be assigned with
	 * setRadius(double) and setMargin(double) and retrieved with getRadius() and
	 * getMargin(). The central surface point of the patch can be stored under
	 * center_ (setCenter(), getCenter()).
	 */
	class PatchView : public VectorView
	{
	private:
		
		//@{
		/**
		 * \brief type for the list of border points + iterators
		 */
		typedef std::list<SurfaceEntity::points_it> border_t;
    typedef border_t::iterator border_it;
		typedef border_t::const_iterator border_cit;
		//@}

	public:

		//**************************************************************************
		// PUBLIC TYPE DEFINITIONS
		//**************************************************************************

		/**
		 * \brief the point iterator type for the border point list.
		 */
		typedef VectorViewIt<border_it, 
												 SurfaceEntity::points_it, 
												 SurfaceEntity::points_cit, Point> borderIter;
		
		//**************************************************************************
		// CONSTRUCTORS
		//**************************************************************************

		/** 
		 * \brief the default constructor
		 * 
		 * Creates a PatchView without an observed entity object
		 */
		PatchView() : VectorView(), border_(), bordermap_(), 
									radius_(), margin_(), center_() { }

		/** 
		 * \brief initializing constructor
		 *
		 * This constructor creates a patch surface view, that observes the entity
		 * \p surface.
		 * 
		 * \param surface the observed surface entity object
		 */
		PatchView(SurfaceEntity& surface) 
			:	VectorView(surface), border_(), bordermap_(),
				radius_(), margin_(), center_() { }
				
    //**************************************************************************
    // DEEP COPY
    //**************************************************************************

  protected:
    void postCopying(DeepCopyInformation& copyinfo) const;

  public:
    
		//**************************************************************************
		// ITERATOR FUNCTIONS
		//**************************************************************************

		//@{
    /** 
		 * \brief iterator functions for the list of border points
		 * 
		 * \return an iterator either to the begin or behind the end of the border
		 *         point sequence
		 */
		borderIter beginBorder() { return borderIter(*this, border_.begin()); }
		borderIter endBorder()   { return borderIter(*this, border_.end()); }
		//@}

		/** 
		 * \brief checks and returns the corresponding border iterator for a
		 *        specific entity point.
		 * 
		 * \param element a point iterator in a surface entity
		 * 
		 * \return a valid iterator into the border sequence or the end of this
		 *         sequence if the point is not included in the border
		 */
		borderIter borderPosOf(SurfaceEntity::points_it element);

    //**************************************************************************
    // PROPERTIES
    //**************************************************************************

    /**  
     * \brief retrieves the number of points at the border of the patch
     * 
     * \return the number of border points
     */
    size_t getNBorderPoints() const { return border_.size(); }

		/** 
		 * \brief retrieves the size of the patch
		 * 
		 * \return the patch radius
		 */
		double getRadius() const { return radius_; }
		/** 
		 * \brief assigns a new patch radius
		 *
		 * \param radius the new radius of the patch
		 */
		void setRadius(double radius) { radius_ = radius; }

		/** 
		 * \brief retrieves the margin of the patch radius
		 * 
		 * \return the radius margin
		 */
		double getMargin() const { return margin_; }
		/** 
		 * \brief assigns a new patch radius margin
		 *
		 * \param margin the new radius margin
		 */
		void setMargin(double margin) { margin_ = margin; }

		/** 
		 * \brief retrieves the center of the patch
		 * 
		 * \return the patch's central surface point
		 */
		SurfaceEntity::points_it getCentralPoint() const { return center_; }
		/** 
		 * \brief assigns a new patch center
		 *
		 * \param margin the new central surface point of the patch
		 */
		void setCentralPoint(const SurfaceEntity::points_it& center) { 
			center_ = center; }

    //**************************************************************************
    // OPERATIONS
    //**************************************************************************

    /** 
     * \brief rebuilds the bordermap
     */
    void rebuildBorderMap();

  protected:
    
    //**************************************************************************
    // PROTECTED METHODS
    //**************************************************************************

    virtual SurfaceView* createView() const { return new PatchView(); }

	private:

		friend class PatchBuilder;

		//**************************************************************************
		// INTERNAL PRIVATE TYPES
		//**************************************************************************

		//@{
		/**
     * \brief type for the entity points <-> border points map.
		 */
    typedef map<SurfaceEntity::points_it, border_it,
								lt_iterator<SurfaceEntity::points_it> > bordermap_t;
    typedef bordermap_t::iterator bordermap_it;
		typedef bordermap_t::const_iterator bordermap_cit;
		//@}
		 
		//****************************************************************************
		// PRIVATE MEMBERS
		//****************************************************************************

    /**
		 * \brief the list of border points in counter clockwise order
		 */
		border_t border_;

		/**
		 * \brief the map that associaties the border list iterators with their
		 * actual entity iterators.
		 *
		 * This map is necessary to allow an assignment of border points from the
		 * surface entity point iterators.
		 */
		bordermap_t bordermap_;

		/// \brief the radius of the patch
		double radius_;
		/// \brief the radius margin of the patch
		double margin_;

		/// \brief the central surface point of the patch
		SurfaceEntity::points_it center_;

	};

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // PATCHVIEW_HH
