/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef POINT_HH
#define POINT_HH

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include <float.h>
#include <math.h>

#include "common.hh"
#include "surfacedata.hh"

//#include <iostream>

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif

	class SurfaceEntity;

	/**
	 * \brief the representation of a surface point on a molecular surface
	 *
	 * This class describes a point on a molecular surface. A surface point in
	 * general has a position in 3D space (position_) and a vector that points
	 * outwards into the direction that is normal to the surface
	 * (normal_). When dealing with surface patches (continous parts of
	 * complete surfaces) it is necessary to have a reference to the original
	 * surface point (master_). Molecular surface points in particular have two
	 * additional parameters: the type of the face that contains this point
	 * (facetype_) and the index of the closest atom sphere (nearestatom_). All
	 * this properties can be set and examined by the usual
	 *
	 * By far the most important property is the position vector. The class
	 * thus provides two special methods to access the content of this vector
	 * most efficiently:
	 * 
	 * - position()     : retrieves a mutable reference to the position vector 
	 *                    for fast access
	 * - operator[](int): provides read/write access to the components of the 
	 *                    position vector.
	 *
	 * All points must be associated to a SurfaceEntity object. This object is thus
	 * a friend of this class.
	 */
	class Point
	{

	public:

		//**************************************************************************
		// CONSTRUCTORS
		//**************************************************************************

		/**
		 * \brief default constructor
		 *
		 * The default creates a point at the origin with a normal along the z
		 * axis and all other parameters assigned to plausible values.
		 *
		 * \sa Point(const RbDblV&, const RbDblV&, int, int, facetype_t),
		 *     Point(Point& src)
		 */
		Point();

		/**
		 * \brief standard constructor
		 *
		 * The standard constructor initializes the position and normal vector
		 * with the given coordinates and assigns the given or plausible values
		 * to the other necessary parameters.
		 *
		 * \param position    the position vector
		 * \param normal      the direction of the normal vector
		 * \param master      the initial master index
		 * \param nearestatom the index of the closest atom
		 * \param facetype    the type of the face containing this point
		 *
		 * \sa Point(), Point(Point& src)
		 */
		Point(const RbDblV& position, const RbDblV& normal, int master=-1,
					int nearestatom=-1, facetype_t facetype=ftUnknown);

		/**
		 * \brief copy constructor
		 *
		 * This constructor creates an exact copy of the \p src point
		 *
		 * \param src the source point
		 *
		 * \sa Point(const RbDblV&, const RbDblV&, int, int, facetype_t),
		 *     Point()
		 */
		Point(const Point& src)
				: master_(src.master_), 
					nearestatom_(src.nearestatom_), facetype_(src.facetype_), 
					position_(src.position_), normal_(src.normal_) { }

		/**
		 * \brief assignment operator
		 *
		 * This operator assigns an exact copy (except for data_) of the \p src
		 * point to this object.
		 *
		 * \param src the source point
     *
     * \return a reference to this point
		 *
		 * \sa Point(const Point& src)
		 */
		Point& operator=(const Point& src);

		//**************************************************************************
		// PROPERTY METHODS
		//**************************************************************************

		/**
		 * \brief retrieves the master index of the point
		 *
		 * \return the masterindex of the point
		 *
		 * \sa master_, setMasterIndex(int)
		 **/
		int getMasterIndex() const { return master_;} 
		/**
		 * \brief assigns a new value to the point's master index
		 *
		 * \param master the new masterindex of this point
		 *
		 * \sa master_, getMasterIndex()
		 */
		void setMasterIndex(int master) {	master_ = master; }
	
		/**
		 * \brief retrieves the nearest atom of the point
		 *
		 * \return the index of the nearest atom to this point in the molecule
		 *
		 * \sa nearestatom_, setNearestAtom(int)
		 **/
		int getNearestAtom() const { return nearestatom_; }
		/**
		 * \brief assigns a new nearest atom to the point
		 *
		 * \param nearestatom the index of the atom that is nearest to this point
		 *
		 * \sa nearestatom_, getNearestAtom()
		 */
		void setNearestAtom(int nearestatom) { nearestatom_ = nearestatom; }
	
		/**
		 * \brief retrieves the type of the face this point belongs to
		 *
		 * \return the facetype of the face that includes this point
		 *
		 * \sa facetype_, setFaceType(facetype_t)
		 **/
		facetype_t getFaceType() const { return facetype_; }
		/**
		 * \brief assigns a new facetype to the point
		 *
		 * \param facetype the type of the face that contains this point
		 *
		 * \sa facetype_, getFaceType()
		 */
		void setFaceType(facetype_t facetype) { facetype_ = facetype; }
	
		/**
		 * \brief returns a constant reference to the position of the point in space
		 *
		 * \return a constant reference to the point's position vector
		 *
		 * \sa position_, position(), setPosition(RbDblV&), operator[](int)
		 */
		const RbDblV& getPosition() const { return position_; }
		/**
		 * \brief returns a mutable reference to the position of the point in space
		 *
		 * \return a mutable reference to the point's position vector
		 *
		 * \note this method is a fast access to the points coordinates. It
		 * should be used only if a fast modification of the whole vector is
		 * necessary. For a component wise transformation the operator[](int)
		 * should be used. Unitary matrix transformations can also be applied by
		 * the transformPosition(const RbDblM&) method.
		 *
		 * \sa position_, getPosition(), setPosition(RbDblV&), operator[](int)
		 */
		RbDblV& position() { return position_; }
		/**
		 * \brief assigns a new position vector to the point
		 *
		 * \param position the new position vector. The components of this vector
		 *                 will be copied to the point's position vector.
		 *
		 * \sa position_, getPosition(), position(), operator[](int)
		 */
		void setPosition( const RbDblV& position );

		/**
		 * \brief element operator. Provides a reference to the components of the
		 * position vector
		 *
		 * \param ix the index of the position's vector component. You can use
		 *           the X, Y, Z predefined constants.
		 *
		 * \sa position_, getPosition(),  position(), setPosition(RbDblV&)
		 */
		double& operator[](int ix) { return position_[ix]; }

		/**
		 * \brief returns a constant reference to the normal of the point in space
		 *
		 * \return a constant reference to the point's normal vector
		 *
		 * \sa normal_, normal(), setNormal(RbDblV&), operator[](int)
		 */
		const RbDblV& getNormal() const { return normal_; }
		/**
		 * \brief returns a mutable reference to the normal of the point in space
		 *
		 * \return a mutable reference to the point's normal vector
		 *
		 * \note this method is a fast access to the points coordinates. It
		 * should be used only if a fast modification of the whole vector is
		 * necessary. For a component wise transformation the operator[](int)
		 * should be used. Unitary matrix transformations can also be applied by
		 * the transformNormal(const RbDblM&) method.
		 *
		 * \sa normal_, getNormal(), setNormal(RbDblV&), operator[](int)
		 */
		RbDblV& normal() { return normal_; }
		/**
		 * \brief assigns a new normal vector to the point
		 *
		 * \param normal the new normal vector. The components of this vector
		 *                 will be copied to the point's normal vector.
		 *
		 * \sa normal_, getNormal(), normal(), operator[](int)
		 */
		void setNormal(const RbDblV& normal);

		/**
		 * \brief retrieves a constant reference to the row data vector
		 *
		 * \return a constant reference to the RbDblV holding the data for this
		 * row
		 *
		 * \sa setData(size_t, double), setData(const RbDblV&), data_
		 */
		const RbDblV& data() const { return *data_; }
		/**
		 * \brief assigns a new value to the row data vector element at index \p
		 * col.
		 *
		 * \param col   the column that receives the new value
		 * \param value the new value
		 *
		 * \exception OutOfRange_e if the column index is out of bounds
		 *
		 * \sa data(), setData(const RbDblV&), data_
		 */
		void setData(size_t col, double value);
   
    /**
     * \brief assigns a complete new data row to the point
     *
     * If the new row vector \p pointdata has more elements than the data
     * matrix has columns, only the number of columns is copied. If the data
     * matrix has more columns, these are filled with zero.
     * 
     * \param pointdata the new point data
     *
     * \sa setData(size_t, double), data(), data_
     */
    void setData(const RbDblV& pointdata);
			
	private:

		//***************************************************************************
		// PRIVATE MEMBERS
		//***************************************************************************
	
		/**
		 * \brief a reference to  a surface point in a different surface object.
		 *
		 * The master index is used when a part of a surface is represented by an
		 * individual object (patch). In this case it serves as a link between
		 * the parent surface and its parts.
		 *
		 * \sa getMasterIndex(), setMasterIndex(int)
		 */
		int master_;

		/**
		 * \brief the position vector of this point.
		 *
		 * The position vector describes the location of the point in space. It
		 * has four components (X, Y, Z, W). The first three components are the
		 * usual x, y and z coordinates, the fourth component is for
		 * transformation purposes and should be 1 all the time.
		 *
		 * \sa operator[](int), getPosition(),  position(), setPosition(RbDblV&)
		 */
		RbDblV position_;
		/**
		 * \brief the direction of the normal vector of this point.
		 *
		 * The normal vector describes the direction normal to the surface at the
		 * point. For the description of the components see position_.
		 *
		 * \sa getNormal(),  normal(), setNormal(RbDblV&)
		 */
		RbDblV normal_;
			
		/**
		 * \brief an iterator to the row vector that holds the data
		 *
		 * \sa SurfaceData
		 */
		SurfaceData::rows_it data_;

		/**
		 * \brief saves the index of the nearest atom in the molecule.
		 *
		 * Each point on the surface has an atom that is closest to its
		 * position. This atom is called the nearest atom and its index can be
		 * stored for each point.
		 *
		 * \note Be carful when using this index to identify atoms. Different
		 * representations of the molecule can have different indices for their
		 * atoms. So be sure to use the same representation of the molecule for
		 * the assigning and retrieval of the nearest atoms.
		 *
		 * \sa getNearestAtom() ,setNearestAtom(int)
		 */
		int nearestatom_;
		/**
		 * \brief contains the type of the face that contains this point
		 *
		 * On molecular surfaces three different types are possible:
		 *
		 * - \c ftToric     for points which belong to toric reentrant faces 
		 * - \c ftReentrant for points inside reentrant faces 
		 * - \c ftContact   for points inside contact faces.
		 */
		facetype_t facetype_;

		//**************************************************************************
		// FRIENDS
		//**************************************************************************
			
		friend class SurfaceEntity;

		friend bool operator==(const Point& a, const Point& b);
		friend double operator-(const Point& a, const Point& b);
	};

	bool operator==(const Point& a, const Point& b);

	//****************************************************************************
	// INLINE METHODS
	//****************************************************************************
		
	//****************************************************************************
	inline Point::Point()
	//****************************************************************************
	: master_(-1), nearestatom_(-1),
		facetype_(ftUnknown), position_(4,0.0), normal_(4,0.0)
	{
		position_[W] = 1;
		normal_[Z] = 1;
		normal_[W] = 1;
	}

  /** 
	 * \brief compares two points and checks for equality
	 * 
	 * \param a,b two points that should be compared
	 * 
	 * \return \c true if the points \p a and  \p b are equal.
	 */		
	inline bool operator==(const Point& a, const Point& b)
	//****************************************************************************
	{
		int i;

		bool result = 
			((a.master_ == b.master_) &&
			 (a.facetype_ == b.facetype_) &&
			 (a.nearestatom_ == b.nearestatom_));

		//compare positions
		for (i=X;i<=W;++i)
			if (result)
				result = result && ((fabs(a.position_[i]-b.position_[i])<double_limit));
		
		//compare normals
		for (i=X;i<=W;++i)
			if (result)
				result = result && ((fabs(a.normal_[i]-b.normal_[i])<double_limit));

		return result;
	}

	/**
	 * \brief calculates the distance between two points
	 *
	 * \param a,b two surface points
	 *
	 * \return the euclidean distance betweenthe two surface points
	 */
	inline double operator-(const Point& a, const Point& b)
	//****************************************************************************
	{
		return (a.position_ - b.position_).vec_len();
	}

#ifdef HAVE_NAMESPACES
};
#endif

#endif // POINT_HH
