/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file simplebuilder.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 18 Mar 2004, last CVS commit 2004/03/18 13:52:33
 */

#ifndef SIMPLEBUILDER_HH
#define SIMPLEBUILDER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "viewbuilder.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 
	
	/**
	 * \brief constructs a simple view for a SurfaceEntity
	 *
	 * A simple view is a view onto a whole SurfaceEntity. The SimpleView object
	 * selects all points and triangles in the SurfaceEntity and adds them to the
	 * View object.
	 *
	 * \param View a subclass of SurfaceView that provides the methods
	 *             selectPoint() and selectTriangle().
	 */
	template<class View>
	class SimpleBuilder : public ViewBuilder
	{
	public:
		//**************************************************************************
		// CONSTRUCTOR
		//**************************************************************************

		/**
		 * \brief default constructor
		 */
		SimpleBuilder() : ViewBuilder() { }
		
		/**
		 * \brief initializes the View object
		 *
		 * This method creates the View object and reserves as much space as
		 * needed to select all triangles and points.
		 *
		 * \param surface a reference to the SurfaceEntity
		 *
		 * \sa select()
		 */
		virtual void initialize(SurfaceEntity& surface);
		/**
		 * \brief builds the View
		 *
		 * The select method selects every point and triangle in the SurfaceEntity.
		 */
		virtual void select();
		
		/**
		 * \brief retrieves the ready View
		 *
		 * \return a pointer to the constructed and ready View as a
		 *         SurfaceView.
		 */
		virtual SurfaceView* getView() const { 	return view_; }

	private:

		/**
		 * \brief the pointer to the View that is build.
		 */
		View* view_;
	};

  //****************************************************************************
	template<class View> 
	void SimpleBuilder<View>::initialize(SurfaceEntity& surface)
	//****************************************************************************
	{
		ViewBuilder::initialize(surface);
	 
		view_ = new View(*surface_);
		// reserve space for all points and vertices
		view_->reserve();
	} 

  //****************************************************************************
	template<class View>
	void SimpleBuilder<View>::select()
  //****************************************************************************
	{
		SurfaceEntity::points_it p;
		SurfaceEntity::triangles_it t;

		for(p = surface_->begin();p!=surface_->end();++p)
			view_->selectPoint(p);

		for(t = surface_->beginTriangles();t!=surface_->endTriangles();++t)
			view_->selectTriangle(t);
	}

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // SIMPLEBUILDER_HH
