/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "sldio.hh"

#ifdef HAVE_STD
#include <fstream>
#include <sstream>
#include <memory>
#else
#include <map.h>
#include <fstream.h>
#include <sstream.h>
#include <memory.h>
#endif

#include "surface.hh"
#include "binaryio.hh"
#include "vectorview.hh"
#include "fileformatfactory.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace RazorBack;
using namespace Surface;
#endif 

REGISTER_IOCLASS(sld, SLDIO, "Sybyl/Molcad mol. surface description format");

const long FILEID         = -4712;
const long REVERSE        = -1729232897;
const long SLD_HASNORMALS = 0x00000001;
const long SLD_HASATOMS   = 0x00000002;

//#include <iostream>

//******************************************************************************
SurfaceEntity* SLDIO::read(istream& astream)
//******************************************************************************
{
	SurfaceIO::read(astream);

	//check the file ID
	int id = 0;
	bool reverse_ = false;

	BinaryIO::readInt(astream, id, false);
	//reverse is also used in the file conversion
	reverse_ = (id == REVERSE);
	
	if ((id!=FILEID) && (!reverse_))
	{
		//if the file format is wrong throw an exception
		throw WrongFormat_e(id, "", "");
	}
	else
	{
		//format is OK start reading


		/* INITIALIZATION */

		//counter
		int i;

		//header variables: header size, size of a triangle, size of a vertex,
    //                  number of trianlges, dots, connections and props.
		//                  t-Flag
		int head_size, triangle_size, vertex_size;
		int nvertices, ntriangles, nedges, natoms, nprops;
		int t_flag;

		//a map to remember the point indices for the right iterators
		map<int,SurfaceEntity::points_it> indexmap;

		//the returning surface object
		auto_ptr<SurfaceEntity> surface(new SurfaceEntity());

		/* FILE HEAD */

		/* Read header information */
		BinaryIO::readInt(astream, head_size, reverse_);
		BinaryIO::readInt(astream, triangle_size, reverse_);
		BinaryIO::readInt(astream, vertex_size, reverse_);
		
		BinaryIO::readInt(astream, ntriangles, reverse_);
		BinaryIO::readInt(astream, nvertices, reverse_);
		BinaryIO::readInt(astream, nedges, reverse_);
		BinaryIO::readInt(astream, natoms, reverse_);
		BinaryIO::readInt(astream, nprops, reverse_);

		BinaryIO::readInt(astream, t_flag, reverse_);

		//read the property head information
		for(i=0;i<nprops;i++)
		{
			char propname[30];
      //read property name and create a new data column
			astream.read((char*) propname, 30);
			surface->addColumn(propname, 0.0);
		}

		//store the number of closest spheres (natoms)
		surface->setSphereCount(natoms);

		/* VERTEX PART */
		
		//shift to the vertex entries
		astream.seekg(head_size+triangle_size * ntriangles, ios::beg);

		//the actual master index
		int master = 0;
		for (int v=0;v<nvertices;v++)
		{
			int j;

			//vertex data
			float xyz[3];        // coordinates
			float nxyz[3];       // normals
			float* props = new float[nprops]; // properties
			short int closest_sphere;   // index of the closest VdW sphere = atom index
			short int dummy;     // some dummy value as spacer

			//normal and position vector
			Vector_<double> xyz_(3);
			Vector_<double> nxyz_(3);
			
			BinaryIO::readFloatArr(astream, xyz, 3, reverse_);
			if (t_flag & SLD_HASNORMALS) 
				BinaryIO::readFloatArr(astream, nxyz, 3, reverse_);
			BinaryIO::readShort(astream, closest_sphere, reverse_);
			
			if (nprops > 0)
				BinaryIO::readFloatArr(astream, props, nprops, reverse_);
			
			//copy coords and normals to Vector_'s
			for(j=0;j<3;++j)
			{
				xyz_[j] = xyz[j];
				nxyz_[j] = nxyz[j];
			}

			//setup the point
			SurfaceEntity::points_it p = surface->addPoint();

			p->setPosition(xyz_);
			p->setNormal(nxyz_);
			//nearest atom in SLD is 1-based
			p->setNearestAtom(--closest_sphere); 
			p->setFaceType(ftUnknown);

			//assign the data values
			for(j=0;j<nprops;++j) 
				p->setData(j, props[j]);

			//checking for doubletts
			if (doublettCheck(v,*p))
				//if the point is not unique delete it from the surface
				surface->deletePoint(p);
			else
			{
				//assign the master index and put rember in the indexmap
				p->setMasterIndex(master);
				indexmap[master] = p;
				//increase the master
				++master;
			}
		}

		/* TRIANGLE PART */
		
		/* Read triangles */
		//shift back to the triangle entries
		astream.seekg(head_size, ios::beg);
		
		/* read all the triangles record by record */
		for(int t=0;t<ntriangles;++t)
		{
			TriangleData tdata;
    
			//read the triangle indices
			BinaryIO::readIntArr(astream, tdata.points, 3, reverse_);
			
			checkTrianglePoints(nvertices, tdata);
			
			for(i=A;i<=C;++i)
				tdata.points[i] = mDoubletts[tdata.points[i]];
		
			if (!isDegenerated(tdata))
			{
				SurfaceEntity::points_it a = indexmap[tdata.points[0]];
				SurfaceEntity::points_it b = indexmap[tdata.points[1]];
				SurfaceEntity::points_it c = indexmap[tdata.points[2]];
			
				SurfaceEntity::triangles_it t = surface->addTriangle(a, b, c);
				t->setFaceType(ftUnknown);
				t->setFaceNum(0);
			}
		}
	
		return surface.release();
	}
}

//******************************************************************************
bool SLDIO::write (ostream & astream,	VectorView& view)
//******************************************************************************
{
	/* INITIALIZATION */
	map<SurfaceEntity::points_cit, size_t,lt_iterator<SurfaceEntity::points_cit> > 
		indexmap;

	int i;

	/* FILE HEAD */
	
  //counts
	int nvertex = view.getNPoints();
	int nfaces  = view.getNTriangles();
	int nedges  = view.getNTriangles()*2;
	int natoms  = view.getSphereCount();
	int nprops   = view.getNColumns();

	//sizes
	int head_size = 10 * sizeof(int) + nprops * 30;
	int triangle_size = 3* sizeof(int);
	int vertex_size = 6 * sizeof(float) + sizeof(short) + nprops * sizeof(float);

	int t_flag = SLD_HASNORMALS | SLD_HASATOMS;

	//write file ID
	int dummy = FILEID;
	BinaryIO::writeInt(astream, dummy, reverse_);

  //#vertices, #faces, #edges
	BinaryIO::writeInt(astream, head_size, reverse_);
	BinaryIO::writeInt(astream, triangle_size, reverse_);
	BinaryIO::writeInt(astream, vertex_size, reverse_);
		
	BinaryIO::writeInt(astream, nfaces, reverse_);
	BinaryIO::writeInt(astream, nvertex, reverse_);
	BinaryIO::writeInt(astream, nedges, reverse_);
	BinaryIO::writeInt(astream, natoms, reverse_);
	BinaryIO::writeInt(astream, nprops, reverse_);

	BinaryIO::writeInt(astream, t_flag, reverse_);

	//write property names
	for(i=0;i<nprops;i++)
	{
		char propname[30] = "";
		const string& colname = view.getColumnName(i);

		size_t namelen = colname.size();
		if (namelen>29) namelen = 29;

		//copy the column name and null terminate it
		strncpy(propname,colname.c_str(), namelen);
		propname[namelen] = 0;

		astream.write((char*) propname, 30);
	}
		
	//generate the vertex index map
	int ix = 0;
	for(VectorView::pointIter p=view.begin();p!=view.end();++p, ++ix)
  	indexmap[p] = ix;

	/* TRIANGLE PART */
	
	VectorView::triangleIter t;
	for(t=view.beginTriangles();t!=view.endTriangles();++t)
  {
		int vixs[3];
		for (int i=A;i<=C;++i)
			vixs[i] = indexmap[t->getVertex(i)];
		BinaryIO::writeIntArr(astream, vixs, 3, reverse_);
  }
	
  /* VERTEX PART */

	for(VectorView::pointIter p=view.begin();p!=view.end();++p, ++ix)
  {
		float xyz[3];
		float nxyz[3];

		//copy the position and normal vector
		for(i=0;i<3;++i)
		{
			xyz[i]  = p->getPosition()[i];
			nxyz[i] = p->getNormal()[i];
		}

		BinaryIO::writeFloatArr(astream, xyz, 3, reverse_);
		BinaryIO::writeFloatArr(astream, nxyz, 3, reverse_);
		//nearest atom in SLD is 1-based
		BinaryIO::writeShort(astream, p->getNearestAtom()+1, reverse_);

		//property values
		if (nprops>0)
		{
			float* props = new float[nprops];
			
			for(i=0;i<nprops;++i)
				props[i] = p->data()[i];

			BinaryIO::writeFloatArr(astream, props, nprops, reverse_);

			delete[] props;
		}
  }

	return true;
}
