/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file sldio.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 01 Apr 2004, last CVS commit 2004/04/01 08:26:50
 */

#ifndef SLDIO_HH
#define SLDIO_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "surfaceio.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	/**
	 * \brief the MOLCAD SLD file reader/writer
	 *
	 * This class performs write and read operation for the files generated and
	 * used by the MOLCAD program of J&uuml;rgen Brickmann.
	 *
	 * SLD files are used in SYBYL, which is available for several platforms. To
	 * overcome the byte order incompatibilities it may be necessary to reverse
	 * the bytes of the various data types. This is done automatically when
	 * reading a file but it has to be specified, if a file has to be written in
	 * reversed byte order. To specify this, please set the property reverse_ to
	 * true with enableReverse() before you write the file. If you want to know
	 * wether the last read hat to reverse the bytes you can check this by
	 * isReverse(), because read(std::istream&) also sets the reverse_ property. This
	 * has the advantage that files are written in the same way as they are read
	 * for the first time.
	 *
	 * For further information about the file format see \ref SLD_fileformat
	 * "The SLD File Format".
	 */
	class SLDIO : public SurfaceIO
	{
	public:

		//**************************************************************************
		// CONSTRUCTOR & DESTRUCTOR
		//**************************************************************************

		/**
		 * \brief default constructor
		 *
		 * simply calls the SurfaceIO default constructor
		 *
		 * \param checkdoubletts turns doublett checking on or off
		 */
		SLDIO(bool checkdoubletts=false) : SurfaceIO(checkdoubletts) { }
			
    /** 
     * \brief factory method
     * 
     * \return a fresh instance of a SLDIO class
     */
    static SurfaceIO* createInstance() { return new SLDIO(); }

		//**************************************************************************
		// OPERATIONS
		//**************************************************************************

		using SurfaceIO::read;
		SurfaceEntity* read(std::istream& astream);
    
		using SurfaceIO::write;
		bool write (std::ostream& astream, VectorView& view);

		//**************************************************************************
		// PROPERTIES
		//**************************************************************************
		
		/**
		 * \brief check if byte reversing is on
		 *
		 * \return \c true if IO is done in reversed byte order
		 *
		 * \sa enableReverse(), reverse_
		 */
		bool isReverse() const { return reverse_; }
		
		/**
		 * \brief enables or disables the byte reversion
		 *
		 * \param reverse if \c true (default) turns on byte reversal
		 *
		 * \return isReverse(), reverse_
		 */
		void enableReverse(bool reverse = true) { reverse_ = reverse; }

	private:

		/**
		 * \brief indicator for byte reversal
		 */
		bool reverse_;
	};

#ifdef HAVE_NAMESPACES
};
#endif 

/**
 * \page sldformat_page SLD object format
 *
 * \note I do not understand the SLD format completely. That is mainly because 
 * the only information I coudl get about the structure of this file format was
 * the source code of MOLEKEL, which was provided by Stefan Portmann.Together 
 * with the code and Emacs hexl-mode I could decode the binary structure of the
 * MOLCAD SLD files, which was my main concern.
 *
 * The SLD format is a binary file format describing triangulated meshes together
 * with vertex properties. It is used by molecular graphics and modelling
 * to describe the various kinds of molecular surfaces.
 *
 * \subsection sldfile_header File Header:
 *
 * The file header contains information about the surface structure and the
 * vertex properties. Its lengths differs with the number of attached properties.
 *
 * The following list contains the meanings of each double word in the header:
 *
 * <table>
 * <tr><td>0000-0003</td>
 *     <td>SLD file format ID code: -4712</td>
 * <tr><td>0004-0007</td>
 *     <td>size of header: 10 * sizeof(int) + number of properties * 30</td>
 * <tr><td>0008-000B</td>
 *     <td>size of triangle description (usually 3 * sizeof(int)</td>
 * <tr><td>000C-000F</td>
 *     <td>size of vertex description</td>
 * <tr><td>0010-0013</td>
 *     <td>number of triangles</td>
 * <tr><td>0014-0017</td>
 *     <td>number of vertices</td>
 * <tr><td>0018-001B</td>
 *     <td>number of connections</td>
 * <tr><td>001C-001F</td>
 *     <td>number of atoms</td>
 * <tr><td>0020-0023</td>
 *     <td>number of properties</td>
 * <tr><td>0024-0027</td>
 *     <td>t-Flag (I do not understand the definition of this element)</td>
 * 
 * The rest of the header, if any, contains 30 character long descriptions of any
 * attached property. Thus the file header size has allways the formula:
 * \f$ 40 + n \cdot 30 \f$, where n is the number of properties.
 *
 * After the triangle and vertex section is a section of readable comment lines
 * which can contain some useful information for the user.
 * 
 * \subsection sldfile_faces Triangle Section:
 * 
 * The triangle information consists of three 4-byte integers representing the
 * indices of the triangles vertices.

 * \subsection sldfile_vertices Vertex Section:
 *
 * Each vertex is described by the following sequence:
 *
 * X Y Z NX NY NZ (float) ATOMID (short int) 0e (?) VERTEX PROPERTIES (one float
 * per prop.) 
 * 
 */

#endif // SLDIO_HH
