/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file spherebuilder.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 18 Mar 2004, last CVS commit 2004/03/18 13:52:33
 */

#ifndef SPHEREBUILDER_HH
#define SPHEREBUILDER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <list>
#else
#  include <list.h>
#endif

#include "common.hh"
#include "viewbuilder.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 
	
	/**
	 * \brief constructs a sphere view for a SurfaceEntity
	 *
	 * A sphere view selects all points and triangles that are placed within a
	 * sphere of the size radius_ and with the center at center_. The center
	 * point and the radius of the selection sphere must be specified in the
	 * constructor or by the property methods setRadius() and setCenter().
	 * 
	 * The distance metrics used are Euclidean distances.
	 *
	 * \param View a subclass of SurfaceView that provides the methods
	 *             selectPoint() and selectTriangle().
	 */
	template<class View>
	class SphereBuilder : public ViewBuilder
	{
	public:
		//**************************************************************************
		// CONSTRUCTOR
		//**************************************************************************

		/**
		 * \brief default constructor.
		 *
		 * It assigns initial, but not suitable values to radius_ and center_.
		 *
		 * \sa SphereBuilder(double, const RbDblV&)
		 */
		SphereBuilder() 
				: ViewBuilder(), radius_(0.0), center_(4,0.0) {	
			center_[W] = 1.0;	
		}
		/**
		 * \brief initializing constructor
		 *
		 * It assigns the given value to the radius_ and center_ of the selection
		 * sphere so that building can start instantly.
		 */
		SphereBuilder(double radius, const RbDblV& center)
				: ViewBuilder(), radius_(radius), center_(4,0.0) { 
			setCenter(center);
		}

		//**************************************************************************
		// PROPERTY METHODS
		//**************************************************************************

		/**
		 * \brief returns a constant reference to the position of the center point
		 *
		 * \return a mutable reference to the sphere center position vector
		 *
		 * \sa center_, setCenter(RbDblV&)
		 */
		const RbDblV& getCenter() const { return center_; }
		/**
		 * \brief assigns a new position to the central point
		 *
		 * \param center the new center position. The components of this vector
		 *                 will be copied to the sphere center's position vector.
		 *
		 * \sa center_, getCenter()
		 */
		void setCenter(const RbDblV& center) { copyVector(center, center_); }

		/**
		 * \brief gives back the radius of the selection sphere
		 *
		 * \return the radius of the selection sphere
		 *
		 * \sa radius_, setRadius(double)
		 */
		double getRadius() const { return radius_; }
		/**
		 * \brief sets the radius of the probe sphere to a new value
		 *
		 * \param radius the new radius of the probe sphere
		 *
		 * \sa radius_, getRadius()
		 */
		void setRadius(double radius) { if (radius>0) radius_ = radius; }
		
		//**************************************************************************
		// BUILDER
		//**************************************************************************

		
		/**
		 * \brief initializes the View object
		 *
		 * This method creates the View object and reserves as much space as
		 * needed to select all triangles and points.
		 *
		 * \param surface a reference to the SurfaceEntity
		 *
		 * \sa select()
		 */
		virtual void initialize(SurfaceEntity& surface);
		/**
		 * \brief builds the View
		 *
		 * The select method selects every point and triangle in the SurfaceEntity
		 * that is within the selection sphere
		 */
		virtual void select();
		
		/**
		 * \brief retrieves the ready View
		 *
		 * \return a pointer to the constructed and ready View as a
		 *         SurfaceView.
		 */
		virtual SurfaceView* getView() const { 	return view_; }

	private:

		//**************************************************************************
		// PRIVATE MEMBERS
		//**************************************************************************

		/**
		 * \brief the center of the selection sphere
		 *
		 * \sa getCenter(), setCenter(const RbDblV&)
		 */
		RbDblV center_;

		/**
		 * \brief the radius of the selection sphere
		 *
		 * \sa getRadius(), setRadius(double)
		 */
		double radius_;
		
		/**
		 * \brief the pointer to the View that is build.
		 */
		View* view_;
		
	};

  //****************************************************************************
	template<class View> 
	void SphereBuilder<View>::initialize(SurfaceEntity& surface)
	//****************************************************************************
	{
		ViewBuilder::initialize(surface);
	 
		view_ = new View(*surface_);
		// reserve space for all points and vertices
		view_->reserve();
	} 

  //****************************************************************************
	template<class View>
	void SphereBuilder<View>::select()
  //****************************************************************************
	{
		SurfaceEntity::points_it p;
		SurfaceEntity::triangles_it t;
		
		//take the square of the radius for faster evaluation
		double radius_check = radius_ * radius_;

		for(p = surface_->begin();p!=surface_->end();++p)
		{
			double distance = (p->getPosition() - center_).vec_len2();
			if (distance<radius_check)
				view_->selectPoint(p);
		}

		for(t = surface_->beginTriangles();t!=surface_->endTriangles();++t)
			view_->selectTriangle(t);
	}

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // SPHEREBUILDER_HH
