/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "surface.hh"

#include <float.h>

#include "trianglelist.hh"
#include "surfaceview.hh"
#include "viewbuilder.hh"

#ifdef HAVE_STD
#  include <list>
#  include <map>
#  include <iterator>
#  include <algorithm>
#else
#  include <list.h>
#  include <map.h>
#  include <iterator.h>
#  include <algo.h>
#endif

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
using namespace RazorBack;
#endif

//#include <iostream>

//******************************************************************************
SurfaceEntity::SurfaceEntity(double density, double proberadius, 
														 size_t spherecount)
//******************************************************************************
:pointdensity_(density), 
 proberadius_(proberadius), 
 spherecount_(spherecount),
 points_(), 
 triangles_(), 
 data_(),
 observers_(),
 writelock_(0)
{

}

//******************************************************************************
SurfaceEntity::SurfaceEntity(const SurfaceEntity& src)
//******************************************************************************
:pointdensity_(src.pointdensity_), 
 proberadius_(src.proberadius_), 
 spherecount_(src.spherecount_), 
 data_(src.data_), 
 points_(),
 triangles_(),
 observers_(),
 writelock_(0) 
{
  operator=(src);
}

//******************************************************************************
SurfaceEntity& SurfaceEntity::operator=(const SurfaceEntity& src)
//******************************************************************************
{
  if (writelock_==0)
  {
    //delete old data
    points_.clear();
    triangles_.clear();

    //assign members
    pointdensity_ = src.pointdensity_; 
    proberadius_  = src.proberadius_;
    spherecount_  = src.spherecount_;
    data_         = src.data_;
    
    /*
      because of the connections between the triangles and the points we must
      create a map between the old and new point iterators to generate a valid
      copy of all triangles
    */
    map<points_cit, points_it, lt_iterator<points_cit> > oldnew;

    //copy the points
    for(points_cit p=src.points_.begin();p!=src.points_.end();++p)
    {
      points_.push_back(*p);
      //get the last iterator
      oldnew[p] = points_.end();
      --oldnew[p];
    }

    //copy the triangles
    for(triangles_cit t=src.triangles_.begin();t!=src.triangles_.end();++t)
    {
      triangles_.push_back(*t);

      //get a reference to the new triangle (for convenience)
      Triangle& triangle = triangles_.back();
	 
      //reassign the point pointers
      for(int i=A;i<=C;++i)
        triangle.vertices_[i] = oldnew[t->vertices_[i]];
    }

    /*
      The data rows in the surface data (data_) and the points in the point list
      have allways the same order -> reassign the iterator references into the
      point data element by element.
    */
	
    //reassign point->data references
    SurfaceData::rows_it r = data_.begin();
    for(points_it p=points_.begin();p!=points_.end();++p, ++r)
      p->data_ = r;
  }

  return *this;
}
    

//******************************************************************************
SurfaceEntity SurfaceEntity::cloneTemplate() const
//******************************************************************************
{
  SurfaceEntity clone(pointdensity_, proberadius_, spherecount_);
  
  //copy the empty columns
  for(int c=0;c<getNColumns();++c)
    clone.addColumn(getColumnName(c), 0);

  return clone;
}

//******************************************************************************
SurfaceEntity::points_it SurfaceEntity::addPoint()
//******************************************************************************
{
	if (writelock_==0)
	{
		points_.push_back(Point());
		//get a reference to the new point
		points_it point = points_.end();
		--point;
		
		//make a row in the surface point data array and establish the reference
		point->data_ = data_.addRow();
		
		//notify all observers
		notify();
		
		return point;
	}
	else
		return points_.end();
}

//******************************************************************************
SurfaceEntity::triangles_it SurfaceEntity::addTriangle(points_it a, points_it b, 
																											 points_it c)
//******************************************************************************
{
	if (writelock_==0)
	{
		triangles_.push_back(Triangle(a,b,c));
		//get a reference to the new triangle
		triangles_it triangle = triangles_.end();

		//notify all observers
		notify();
		
		return --triangle;
	}
	else
		return triangles_.end();
}

//******************************************************************************
void SurfaceEntity::deleteTriangle(triangles_it t)
//******************************************************************************
{
	if (writelock_==0) 
	{
		triangles_.erase(t);

		//notify all observers
		notify();
	}
}

//******************************************************************************
void SurfaceEntity::deletePoint(points_it p, bool trianglecheck)
//******************************************************************************
{
	if (writelock_==0)
	{
		//if demanded clear all the triangles of this point
		if (trianglecheck) clearPoint(p);
	
		//delete the point and the data row for that particular point
		data_.deleteRow(p->data_);
		points_.erase(p);

		//notify all observers
		notify();
	}
}

//******************************************************************************
void SurfaceEntity::attach(SurfaceView& observer)
//******************************************************************************
{
	list<SurfaceView*>::iterator pos = find(observers_.begin(), observers_.end(),
																					&observer);
	if (pos == observers_.end())
		observers_.push_back(&observer);
}

//******************************************************************************
void SurfaceEntity::detach(SurfaceView& observer)
//******************************************************************************
{
	list<SurfaceView*>::iterator pos = find(observers_.begin(), observers_.end(),
																					&observer);
	
	if (pos != observers_.end())
		observers_.erase(pos);
}			

//******************************************************************************
SurfaceView* SurfaceEntity::createView(ViewBuilder& builder)
//******************************************************************************
{
	builder.initialize(*this);
	builder.select();

	SurfaceView* view = builder.getView();

	return view;
}

//******************************************************************************
void SurfaceEntity::setPointDensity(double density)
//******************************************************************************
{ 
	if (density>=-DBL_EPSILON) 
		pointdensity_=density; 
}
		
//******************************************************************************
void SurfaceEntity::setProbeRadius(double radius)
//******************************************************************************
{ 
	if (radius>=-DBL_EPSILON) 
		proberadius_=radius; 
}

//******************************************************************************
size_t SurfaceEntity::clearPoint(points_cit p)
//******************************************************************************
{
	if (writelock_==0)
	{
		//check if any triangles are associated with this point 
		TriangleList listgenerator(*this, p);
		list<triangles_it> tlist;
		//retrieve all the triangles and store their number
		int c = listgenerator(back_inserter(tlist));
		
		//remove them
		for(list<triangles_it>::const_iterator t=tlist.begin();t!=tlist.end();++t)
			triangles_.erase(*t);
		
		return c;
	}
	else
		return 0;
}

//******************************************************************************
void SurfaceEntity::notify()
//******************************************************************************
{
	for(list<SurfaceView*>::iterator i=observers_.begin();i!=observers_.end();++i)
		(*i)->update();
}
