/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SURFACE_HH
#define SURFACE_HH

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#include <list>
#else
#include <list.h>
#endif

#include "common.hh"
#include "point.hh"
#include "triangle.hh"
#include "surfacedata.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif
	class SurfaceView;
	class ViewBuilder;

	/**
	 * \brief the surface class
	 *
	 * The surface entity is the core data structure of the surface library. It
	 * contains the real points, triangles and the surface data of a molecular
	 * surface. From the entity, different views can be generated that contain
	 * only references to the points and triangles. The relationship between the
	 * entities and the views is an observer pattern (293). Various different
	 * subclasses of SurfaceView can be used as views to the SurfaceEntity, and
	 * more than one view at a time are possible. The views are notified about
	 * structural changes (addition and deletion of points and triangles) by the
	 * notify() method. All the views are stored in the observers_ data member. A
	 * view can be generated by createView() where a ViewBuilder object has to be
	 * provided to create the desired type of view. 
	 * 
	 * A surface entity can also be write locked by calling lock() with the
	 * argument \c true. This increases the internal lock counter and guarentees
	 * that the no structural changes can be done until the writelock_ counter is
	 * zero again. If write locking is not necessary any more it has to be turned
	 * off by calling lock() with the argument \c false. This decrements the
	 * writelock_ flag per one indicating that writing is permitted by again by
	 * the calling operation.
	 *
	 * The points and triangles of the surface are stored as lists in points_ and
	 * triangles_. The entity itself provides only begin() and end() iterator
	 * access for the points and beginTriangles() and endTriangles() access for
	 * the triangles. More sophisticated access will be provided by the
	 * appropriate surface views (e.g. a VectorView provides random access to a
	 * selection (or all) points and triangles of the entity.). Points and
	 * triangles can be added and deleted (in constant time) by addPoint(),
	 * deletePoint(), addTriangle() and deleteTriangle() respectively. 
	 *
	 * \note Although the operations themself are of constant time, it may be
	 * possible that propagation of the changes to the views may cause more
	 * complex time behaviour.
	 *
	 * The surface entity also has an interface to the point data matrix that is
	 * stored as a SurfaceData object in data_. One can add new columns with
	 * addColumn(const std::string&, double) if one wants to assign the same
	 * constant value for each point or 
	 * addColumn(const std::string&, const RbDblV&) if one wants to use a 
	 * predefined data column vector. The name of a data column can be retrieved
	 * (getColumnName()) or changed (setColumnName() while the actual number of
	 * columns is provided by getNColumns(). The complete vector of an existing
	 * column can be changed by setColumn() and a column can be deleted by
	 * deleteColumn(size). Access to the data for each point is provided by the
	 * Point class itself.
	 *
	 * Beside the points and triangles a molecular surface can also be associated
	 * with a probe sphere radius (proberadius_), a point density (pointdensity_)
	 * and the number of spheres or atom that form this surface
	 * (spherecount_). All this properties can be accessed and changed by
	 * appropriate get and set methods.
	 */
	class SurfaceEntity
	{
	private:

		//**************************************************************************
		// PRIVATE TYPE DEFINITIONS
		//**************************************************************************
			
		/**
		 * \brief the type definition of the point list
		 */
		typedef std::list<Point> points_t;
		/**
		 * \brief the type definition of the triangle list
		 */
		typedef std::list<Triangle> triangles_t;

	public:
			
		//**************************************************************************
		// PUBLIC TYPE DEFINITIONS
		//**************************************************************************

		//@{
		///\name point list iterator types
			
		///\brief constant point list iterator
		typedef points_t::const_iterator points_cit;
		///\brief mutable point list iterator
		typedef points_t::iterator points_it;
		//@}

		///\brief constant triangle list iterator
		typedef triangles_t::const_iterator triangles_cit;
		///\brief mutable point list iterator
		typedef triangles_t::iterator triangles_it;

		//**************************************************************************
		// CONSTRUCTORS & DESTRUCTOR
		//**************************************************************************

		/**
		 * \brief default and standard constructor
		 *
		 * This constructor initializes the flags with the given arguments or
		 * with default values. The point and triangle lists are initialized.
		 *
		 * \param density     the point density on the surface
		 * \param proberadius the radius of the probe sphere
		 * \param spherecount the number of atomic spheres beneath the surface
		 *
		 * \sa SurfaceEntity(const SurfaceEntity&)
		 */
		SurfaceEntity(double density=0.0, double proberadius=0.0, 
									size_t spherecount = 0);
		/**
		 * \brief copy constructor
		 *
		 * The copy constructor creates a copy of the give \p src SurfaceEntity
		 * object.
		 *
		 * \param src the source object
		 *
		 * \sa SurfaceEntity(double, double, size_t)
		 */
		SurfaceEntity(const SurfaceEntity& src);

    //**************************************************************************
    // COPYING
    //**************************************************************************

		/**
		 * \brief assignment operator
		 *
		 * The assignment operator assigns a copy of the \p src SurfaceEntity
		 * to this object
		 *
		 * \param src the source object
		 *
		 * \sa SurfaceEntity(const SurfaceEntity& src)
		 */
		SurfaceEntity& operator=(const SurfaceEntity& src);
    

    /** 
     * \brief clones the empty surface entity
     *
     * This method creates a copy of the surface entity that contains
     * no points and triangles but the general parameters like point density,
     * probe radius, number of spheres.
     * 
     * \return an empty clone of the surface entity
     */
    SurfaceEntity cloneTemplate() const;

		//**************************************************************************
		// POINT AND TRIANGLES
		//**************************************************************************

		/**
		 * \brief adds a new point to the surface
		 *
		 * The new point does not have to have meaningful coordinates or normal
		 * components, but it should be initialized by the default constructor
		 * (Point::Point()). The methods stores the pointer in the point
		 * list and sets the Point::fIndex property according to the position
		 * in this list. Once a point is added to a surface the surface object
		 * is responsible for it.
		 *
		 * \param point a valid pointer to a surface point
		 *
		 * \return an iterator to the new point, or to the end() position if
		 *         there was not enough memory or the surface was write locked.
		 *
		 * \note It is not checked, if the point has already be added to the
		 * surface. Thus the user must take care that points are not added
		 * more than once.
		 *
		 * \sa points_
		 */
		points_it addPoint();
		/**
		 * \brief adds a new triangle to the surface based on pointers to surface
		 * points
		 * 
		 * The methods adds a new triangle to the surface. The vertices of the
		 * triangles are specified by pointers to points on the surface. The
		 * points must be members of the surface, otherwise the triangle cannot
		 * be added.
		 *
		 * \param a,b,c valid point iterators
		 *
		 * \return an iterator pointing to the newly created triangle or to
		 *         endTriangles() if there was not enough memory or the surface
		 *         was write locked.
		 *
		 * \note It is not checked, if the triangle has already be added to the
		 * surface. Thus the user must take care that triangles are not added
		 * more than once.
		 *
		 * \sa deleteTriangle(triangles_it)
		 *     triangles_
		 */
		triangles_it addTriangle(points_it a, points_it b, points_it c);

		/**
		 * \brief removes a triangle from the list by its index
		 *
		 * The triangle \p t is removed from the list of triangles.
		 *
		 * \param t an iterator to the to be deleted triangle
		 *
		 * \sa addTriangle(points_it, points_it, points_it)
		 *     triangles_
		 */
		void deleteTriangle(triangles_it t);
		/**
		 * \brief removes a point from the list by its index
		 *
		 * The point \p p is removed from the list of points. If \p trianglecheck
		 * is on all triangles that belong to this point are also deleted.
		 *
		 * \param p             an iterator to the point
		 * \param trianglecheck if \p true removes all triangles that have this
		 *                      point from the surface.
		 *
		 * \sa addPoint(),
		 *     points_
		 */
		void deletePoint(points_it p, bool trianglecheck=false);

		//**************************************************************************
		// OBSERVER
		//**************************************************************************

		/**
		 * \brief registers a new observer object
		 *
		 * \param observer a reference to a new observer object
		 *
		 * \sa detach(SurfaceView&)
		 */
		void attach(SurfaceView& observer);
		/**
		 * \brief unregisters an obsolete observer object
		 *
		 * \param observer a reference to an obsolete observer object
		 *
		 * \sa attach(SurfaceView&)
		 */
		void detach(SurfaceView& observer);
			
		/**
		 * \brief enables or disables the write lock of the surface
		 *
		 * \param writelock if \c true the surface is write locked, if false it
		 *                  is not.
		 *
		 * \sa writelock_
		 */
		void lock(bool writelock = true);

		/**
		 * \brief returns the writelock status
		 *
		 * \return the status of the writelock
		 *
		 * \sa lock(), writelock_
		 */
		bool isLocked() const { return (writelock_>0); }
		
		/**
		 * \brief returns a view for this surface
		 *
		 * The view will be created by an instance of a ViewBuilder object. The
		 * ViewBuilder is responsible for the instantiationn and creation of the
		 * surface view.
		 *
		 * \param builder an instance of a ViewBuilder object
		 *
		 * \return a pointer to new SurfaceView that observes this entity.
		 */
		SurfaceView* createView(ViewBuilder& builder);
			
		//**************************************************************************
		// POINT DATA
		//**************************************************************************

		/**
		 * \brief retrieves the number of columns in the point data matrix
		 *
		 * \return the number of columns in each point data vector
		 *
		 * \sa data_
		 */
		size_t getNColumns() const { return data_.getNColumns(); }

		//@{
		/**
		 * \brief adds a column at the end of the point data matrix
		 *
		 * The definition of the parameters and the effect is the same as of
		 * SurfaceData::addColumn(const std::string&, double) or
		 * SurfaceData::addColumn(const std::string&, RbDblV&).
		 *
		 * \param name the name of the column 
		 * \param def  a default value or column vector that is assigned to all 
		 *             new entries in existing rows.
		 *
		 * \return the index of the new column or -1 if the surface is write
		 *         locked.
		 *
		 *     deleteColumn(rows_it), rows_
		 */
		size_t addColumn(const std::string& name, double def=0.0) {
			if (!writelock_) return data_.addColumn(name, def); else return -1; }
		size_t addColumn(const std::string& name, const RbDblV& def) {
			if (!writelock_) return data_.addColumn(name, def); else return -1; }
		//@}

		/**
		 * \brief removes \p col from the point data matrix
		 *
		 * \param col the column to remove
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa addColumn(const std::string&, double), 
		 *     addRow(const std::string&, const RbDblV&)
		 */
		void deleteColumn(size_t col) {
			if (!writelock_) data_.deleteColumn(col); }

		/**
		 * \brief sets the elements of a column to new values
		 *
		 * The method sets new values for the specified column \p col in the
		 * surface point data matrix.
		 *
		 * \param col     the index of the desired column
		 * \param coldata the new column vector
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa setColumnTitle(), SurfaceData::setColumn()
		 */
		void setColumn(size_t col, const RbDblV& coldata) {
			data_.setColumn(col, coldata); }

		/**
		 * \brief retrieves the name of column \p ix
		 *
		 * \param col  the index of the column
		 *
		 * \return the name of the column \p ix or an empty string if the index
		 *         is out of bounds.
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa setColumnName(size_t, const std::string&), 
		 *     SurfaceData::getColumnName(size_t)   
		 */
		std::string getColumnName(size_t col) const {
			return data_.getColumnName(col); }
		/**
		 * \brief sets the name of a specific column
		 *
		 * \param col  the index of the column
		 * \param name the new name of the column
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa getColumnName(size_t),
		 *     SurfaceData::setColumnName(size_t, const std::string&)
		 */
		void setColumnName(size_t col, const std::string& name) {
			data_.setColumnName(col, name); }

    /** 
     * \brief retrieves the index of a data column by name
     * 
     * \param name the name of the desired data column
     * 
     * \return the index of the desired data column if it exists, -1 if the
     *         column is new
     */
    int indexOfColumn(const std::string& name) const {
      return data_.indexOfColumn(name); }
    
			
		/**
		 * \brief creates a nxn double matrix from the surface point data
		 *
		 * \return a RbDblM object with the data of this object.
		 *
		 * \sa SurfaceData::extractMatrix()
		 */
		RbDblM extractDataMatrix() const { return data_.extractMatrix(); }
		/**
		 * \brief creates a column vector copy of the \p col'th column of the
		 *        surface point data
		 * \param col the desired column
		 *
		 * \sa SurfaceData::extractColumn(size_t)
		 */
		RbDblV extractDataColumn(size_t col) const { 
			return data_.extractColumn(col); }

		//**************************************************************************
		// OTHER PROPERTY METHODS
		//**************************************************************************

		/**
		 * \brief gives back the number of points on the surface
		 *
		 * \return the number of surface points (entries in the point list
		 * points_)
		 * 
		 * \sa getNTriangles()
		 */
		size_t getNPoints() const { return points_.size(); }
		/**
		 * \brief gives back the number of triangles on the surface
		 *
		 * \return the number of surface triangles (entries in the triangle list
		 * points_)
		 * 
		 * \sa getNPoints()
		 */
		size_t getNTriangles() const { return triangles_.size(); }

		/**
		 * \brief retrieves the number of atomic spheres in the molecule
		 *
		 * \return the number of atomic spheres in the molecule that build the
		 *          surface
		 *
		 * \sa spherecount_, setSphereCount(size_t)
		 */
		size_t getSphereCount() const { return spherecount_; }
		/**
		 * \brief sets the number of atomic spheres to a new value
		 *
		 * \param spherecount the new number of atomic spheres that build the
		 *                    surface
		 *
		 * \sa spherecount_, getSphereCount()
		 */
		void setSphereCount(size_t spherecount) { spherecount_ = spherecount; }

		/**
		 * \brief retrieves the point density on the surface
		 *
		 * \return the number of surface points per square angstrom.
		 *
		 * \sa pointdensity_, setPointDensity(double)
		 */
		double getPointDensity() const { return pointdensity_; }
		/**
		 * \brief sets the point density to a new value
		 *
		 * \param density the new number of of surface points per square
		 *                angstrom.
		 *
		 * \sa pointdensity_, getPointDensity()
		 */
		void setPointDensity( double density );

		/**
		 * \brief gives back the radius of the probe sphere
		 *
		 * \return the radius of the probe sphere
		 *
		 * \sa proberadius_, setProbeRadius(double)
		 */
		double getProbeRadius() const { return proberadius_; }
		/**
		 * \brief sets the radius of the probe sphere to a new value
		 *
		 * \param radius the new radius of the probe sphere
		 *
		 * \sa proberadius_, getProbeRadius()
		 */
		void setProbeRadius(double radius);

		//**************************************************************************
		// ITERATORS
		//**************************************************************************

		//@{
		/// \name point iterators
		/**
		 * \brief iterator for the point sequence
		 *
		 * The class provides begin and end iterators for the point sequence for
		 * mutable as well as constant access to the single points.
		 */
		points_it begin() { return points_.begin(); }
		points_it end()   { return points_.end(); }
		points_cit begin() const { return points_.begin(); }
		points_cit end()   const { return points_.end(); }
		//@}

		//@{
		/// \name triangle iterators
		/**
		 * \brief iterator for the triangle sequence
		 *
		 * The class provides begin and end iterators for the triangle sequence for
		 * mutable as well as constant access to the single triangles.
		 */
		triangles_it beginTriangles() { return triangles_.begin(); }
		triangles_it endTriangles()   { return triangles_.end(); }
		triangles_cit beginTriangles() const { return triangles_.begin(); }
		triangles_cit endTriangles()   const { return triangles_.end(); }
		//@}

	private:

		//**************************************************************************
		// PRIVATE METHODS
		//**************************************************************************

		/**
		 * \brief removes all the triangles that belong to a point
		 *
		 * \param p an iterator to the point that should be cleared
		 *
		 * \return the number of deleted triangles.
		 *
		 * \sa deletePoint(points_it)
		 */
		size_t clearPoint(points_cit p);

		/**
		 * \brief the notivy method for the observer(293) pattern
		 *
		 * This method notifies all attached observers that the data structure
		 * and its contents has changed.
		 */
		void notify();

		//**************************************************************************
		// PRIVATE DATA MEMBERS
		//**************************************************************************

		/**
		 * \brief the number of spheres that build the molecular surface
		 *
		 * \sa getSphereCount(), setSphereCount(size_t)
		 */
		size_t spherecount_;
		/**
		 * \brief the number of points on the surface per square angstrom.
		 *
		 * \sa getPointDensity(), setPointDensity(double)
		 */
		double pointdensity_;
		/**
		 * \brief the radius of the probe that generated the molecular surface
		 *
		 * \sa getProbeRadius(), setProbeRadius(double)
		 */
		double proberadius_;

		/**
		 * \brief the array of points
		 */
		points_t points_;
		/**
		 * \brief the array of triangles
		 */
		triangles_t triangles_;
		/**
		 * \brief the surface data linked row vector matrix
		 */
		SurfaceData data_; 

		/**
		 * \brief indicates if the surface is write locked.
		 *
		 * A write-lock is necessary in some situations. I.e. if a graph view is
		 * processing the data structure or more. During write lock, no
		 * triangles, points or data columns must be added or deleted. Changes to
		 * the values of the elements are possible.
		 *
		 * To be allow multiple write locks that are not interferring with each
		 * other, the entity records how many locks have been set.
		 */
		size_t writelock_;

		/**
		 * \brief the list of observers
		 */
		list<SurfaceView*> observers_;
	};

	//****************************************************************************
	inline void SurfaceEntity::lock(bool writelock)
  //****************************************************************************
	{
		if (writelock)
			++writelock_;
		else if (writelock_>0)
			--writelock_;
	}


#ifdef HAVE_NAMESPACES
};
#endif

#endif // POINT_C_H
