/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "surfacedata.hh"

#ifdef HAVE_STD
#  include <algorithm>
#else
#  include <algo.h>
#endif

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
#endif 

//******************************************************************************
SurfaceData::SurfaceData(const RbDblM& src)
//******************************************************************************
:columns_(src.cno()), rows_()
{
	for(int i=0;i<src.rno();++i)
	{
		//add a new row
		rows_it r = addRow();
		//copy the elements column wise
		for(int j=0;j<src.cno();++j) (*r)[j] = src(i,j);
	}
}

//******************************************************************************
SurfaceData& SurfaceData::operator=(const RbDblM& src)
//******************************************************************************
{
	columns_.clear();
	rows_.clear();

	columns_.resize(src.cno());
	
	for(int i=0;i<src.rno();++i)
	{
		//add a new row
		rows_it r = addRow();
		//copy the elements column wise
		for(int j=0;j<src.cno();++j) (*r)[j] = src(i,j);
	}

	return *this;
}

//******************************************************************************
SurfaceData::rows_it SurfaceData::addRow()
//******************************************************************************
{
	return insertRow(rows_.end());
}

//******************************************************************************
SurfaceData::rows_it SurfaceData::insertRow(rows_it before)
//******************************************************************************
{
	return rows_.insert(before, RbDblV(columns_.size()));
}

//******************************************************************************
void SurfaceData::deleteRow(rows_it row)
//******************************************************************************
{
	rows_.erase(row);
}

//******************************************************************************
size_t SurfaceData::addColumn(const string& name, double def)
//******************************************************************************
{
	int colix = columns_.size();

	//add the column at the end
	columns_.push_back(name);
	
	//adjust the row sizes
	for(rows_it r=rows_.begin();r!=rows_.end();++r)
	{
		r->dim(columns_.size());
		(*r)[colix] = def;
	}

	return colix;
}

//******************************************************************************
size_t SurfaceData::addColumn(const std::string& name, const RbDblV& def)
//******************************************************************************
{
	int colix = columns_.size();

	//add the column at the end
	columns_.push_back(name);
	
	//adjust the row sizes and assign the new values
	vector<double>::const_iterator i = def.begin();
	for(rows_it r=rows_.begin();r!=rows_.end();++r)
	{
		r->dim(columns_.size());
		if (i!=def.end())
			(*r)[colix] = *i++;
		else
			(*r)[colix] = 0.0;
	}

	return colix;
}

//******************************************************************************
void SurfaceData::deleteColumn(size_t col)
//******************************************************************************
{
	if ((col<0) || (col>=columns_.size()))
		throw OutOfRange_e("The column index was out of bounds!");

	//delete the column
	vector<string>::iterator c = columns_.begin() +col;
	columns_.erase(c);
	
	//remove the column from all row vectors 
	//RAZORBACK SPECIFIC IMPLEMENTATION
	for(rows_it r=rows_.begin();r!=rows_.end();++r)
	{
		for(int i=col+1;i<r->dim();++i)
			//shift the entries left after the deleted column
			(*r)[i-1] = (*r)[i];
		
		r->dim(columns_.size());
	}
}

//******************************************************************************
RbDblM SurfaceData::extractMatrix() const
//******************************************************************************
{
	RbDblM result(rows_.size(),columns_.size());

	int i=0;
	for(rows_cit r=rows_.begin();r!=rows_.end();++r, ++i)
	{
		for(int j=0;j<columns_.size();++j)
			result(i,j) = (*r)[j];
	}

	return result;
}

//******************************************************************************
RbDblV SurfaceData::extractColumn(size_t col) const
//******************************************************************************
{
	if ((col<0) || (col>=columns_.size()))
		throw OutOfRange_e("The column index was out of bounds!");

	RbDblV result(rows_.size());

	int i=0;
	for(rows_cit r=rows_.begin();r!=rows_.end();++r, ++i)
		result[i] = (*r)[col];

	return result;
}
			
//******************************************************************************
string SurfaceData::getColumnName(size_t col) const
//******************************************************************************
{
	if ((col<0) || (col>=columns_.size()))
		throw OutOfRange_e("The column index was out of bounds!");

	return columns_[col];
}

//******************************************************************************
void SurfaceData::setColumnName(size_t col, const string& name)
//******************************************************************************
{
	if ((col<0) || (col>=columns_.size()))
		throw OutOfRange_e("The column index was out of bounds!");
	
	columns_[col] = name;
}

//******************************************************************************
int SurfaceData::indexOfColumn(const std::string& name) const
//******************************************************************************
{
  std::vector<std::string>::const_iterator 
    col = find(columns_.begin(), columns_.end(), name);

  if (col!=columns_.end())
    return col - columns_.begin();
  else
    return -1;
}

//******************************************************************************
void SurfaceData::setColumn(size_t col, const RbDblV& coldata)
//******************************************************************************
{
	if ((col<0) || (col>=columns_.size()))
		throw OutOfRange_e("The column index was out of bounds!");

	vector<double>::const_iterator i = coldata.begin();
	for(rows_it r=rows_.begin();r!=rows_.end();++r)
		if (i!=coldata.end())
			(*r)[col] = *i++;
		else
			(*r)[col] = 0.0;
}
