/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file surfacedata.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 02 Mar 2004, last CVS commit 2004/03/02 21:01:12
 */

#ifndef SURFACEDATA_HH
#define SURFACEDATA_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "common.hh"

#ifdef HAVE_STD
#  include <vector>
#  include <list>
#else
#  include <vector.h>
#  include <list.h>
#endif

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	/**
	 * \brief a datastructure for attached surface properties
	 *
	 * For various purposes it can be necessary to attach floating point
	 * properties to each surface point. The SurfaceData structure contains a
	 * linked list of row vectors (rows_) that contain an entry for each
	 * column. The columns are named and the names are stored in a std::vector
	 * (columns_) of std::strings.
	 *
	 * It is possible to add (addRow()), insert(insertRow()) and delete
	 * (deleteRow()) signle rows, and to add 
	 * (addColumn(const std::string&, double) columns at the end and delete 
	 * columns (deleteColumns()) from every position in the column vector. 
	 * The column names can be retrieved and set for every existing column 
	 * (getColumnName() and setColumnName()) and the number of columns and rows 
	 * are available by getNColumns() and getNRows().
	 *
	 * It is also possible to extract and assign complete column vectors
	 * (extractColumn(), addColumn(const std::string& , const RbDblV&),
	 * setColumn(size_t, const RbDblV&).
	 *
	 * The class provides iterator based access to the row vectors (begin(),
	 * end()) as well as constant and mutable element access
	 * (operator()(rows_it, size_t) and operator()(rows_cit, size_t)).
	 */
	class SurfaceData
	{
	public:

		//************************************************************************
		// PUBLIC TYPE DEFINITIONS
		//************************************************************************

		//@{
		/**
		 * \brief type definition for the list of row vectors
		 *
		 * The row vectors are collected in a standard list. The iterator types
		 * for that list are also defined.
		 */
		typedef std::list<RbDblV>      rows_t;
		typedef rows_t::iterator       rows_it;
		typedef rows_t::const_iterator rows_cit;

		typedef rows_t::iterator iterator;
		typedef rows_t::const_iterator const_iterator;
		//@}

		//************************************************************************
		// CONSTRUCTORS
		//************************************************************************

		/**
		 * \brief default and standard constructor
		 *
		 * The default or standard constructor initializes a surface data matrix
		 * with \p col columns.
		 *
		 * \param col the number of columns that is prepared for this data matrix.
		 *
		 * \sa SurfaceData(const SurfaceData&)
		 */
		SurfaceData(size_t col=0): columns_(col), rows_() { }

		/**
		 * \brief copy constructor
		 *
		 * \param src the source surface data matrix
		 *
		 * \sa SurfaceData(), operator=(const SurfaceData&)
		 */
		SurfaceData(const SurfaceData& src)
				: columns_(src.columns_), rows_(src.rows_) { }

		/**
		 * \brief a special constructor that copies the elements of a normal nxn
		 *        matrix.
		 *
		 * \param src the nxn matrix
		 * 
		 * \sa operator=(const RbDblM& src)
		 */
		SurfaceData(const RbDblM& src);

		/**
		 * \brief assignment operator
		 *
		 * \param src the source surface data matrix
		 *
		 * \return a reference to itself
		 *
		 * \sa SurfaceData(const SurfaceData&)
		 */
		SurfaceData& operator=(const SurfaceData& src);
			
		/**
		 * \brief a special assignment operator that copies the elements of a
		 * normal nxn matrix.
		 *
		 * \param src the nxn matrix
		 * 
		 * \sa SurfaceData(const RbDblM& src)
		 */
		SurfaceData& operator=(const RbDblM& src);

		//************************************************************************
		// PUBLIC OPERATIONS
		//************************************************************************

		/**
		 * \brief adds a row at the end of the row vector sequence list
		 *
		 * Any new row has as many elements as the columns_ header vector has
		 * entries.
		 *
		 * \return an iterator pointing to the new entry in the row vector
		 * sequence list.
		 *
		 * \sa inserRow(rows_it), deleteRow(rows_it), rows_
		 */
		rows_it addRow();
		/**
		 * \brief inserts a row before the \p row into the row vector list
		 *
		 * \param before an iterator pointing before the place of the new row
		 *
		 * \sa addRow(), deleteRow(rows_it), rows_
		 */
		rows_it insertRow(rows_it before);
		/**
		 * \brief removes \p row from the list of row vectors
		 *
		 * \param row the row to remove
		 *
		 * \sa addRow(const std::string&, double),
		 *     addRow(const std::string&, const RbDblV&),
		 *     insertRow(rows_it), rows_
		 */
		void    deleteRow(rows_it row);
			
		//@{
		/**
		 * \brief adds a column at the end of the column vector
		 *
		 * All rows in the row vector list are automatically updated by a new
		 * element with the \p def value or the corresponding element in a
		 * column vector. If the column vector has less than rows_.size()
		 * elements all additional rows got a 0.0 in the new column.
		 *
		 * \param name the name of the column 
		 * \param def  a default value or column vector that is assigned to all 
		 *             new entries in existing rows.
		 *
		 * \return the index of the new column.
		 *
		 * \sa deleteColumn(rows_it), rows_
		 */
		size_t addColumn(const std::string& name, double def=0.0);
		size_t addColumn(const std::string& name, const RbDblV& def);
		//@}

		/**
		 * \brief removes \p col from the column head vector
		 *
		 * \param col the column to remove
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa addColumn(const std::string&, double), 
		 *     addRow(const std::string&, const RbDblV&)
		 */
		void   deleteColumn(size_t col);

		/**
		 * \brief creates a nxn double matrix out of the data stored in the row
		 * vectors
		 *
		 * The method copies the entries of the surface data matrix into the rows
		 * and columns of a new RbDblM.
		 *
		 * \return a RbDblM object with the data of this object.
		 */
		RbDblM extractMatrix() const;

		/**
		 * \brief creates a column vector copy of the \p col'th column
		 *
		 * A copy of all \p col'th entries in the row vectors is assembled in a
		 * RbDblV that represents the column vector.
		 *
		 * \param col the desired column
		 */
		RbDblV extractColumn(size_t col) const;
			
		//************************************************************************
		// PUBLIC PROPERTY METHODS
		//************************************************************************

		/**
		 * \brief retrieves the number of columns in each row vector
		 *
		 * \return number of columns in each row vector by the number of column
		 *         heads.
		 *
		 * \sa addColumn(const std::string&, double)
		 */
		size_t getNColumns() const { return columns_.size(); }

		/**
		 * \brief retrieves the number of rows in each row vector
		 *
		 * \sa addRow(const std::string&, double),
		 *     insertRow(size_t, const std::string&, double)
		 */
		size_t getNRows() const  { return rows_.size(); }

		/**
		 * \brief retrieves the name of column \p ix
		 *
		 * \param col  the index of the column
		 *
		 * \return the name of the column \p ix or an empty string if the index
		 *         is out of bounds.
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa setColumnName(size_t, const std::string&), columns_
		 */
		std::string getColumnName(size_t col) const;
		/**
		 * \brief sets the name of a specific column
		 *
		 * \param col  the index of the column
		 * \param name the new name of the column
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa getColumnName(size_t), columns_
		 */
		void setColumnName(size_t col, const std::string& name);

    /** 
     * \brief retrieves the index of a column by name
     * 
     * \param name the name of the desired column
     * 
     * \return the index of the desired column if it exists, -1 if the column
     *         is new
     */
    int indexOfColumn(const std::string& name) const;

		/**
		 * \brief checked read/write access to an element in the surface data
		 * matrix.
		 *
		 * \param row an iterator into the desired row
		 * \param col the column index of the desired column
		 *
		 * \exception OutOfRange_e if \p row or \p col is out of bounds
		 *
		 * \sa operator(rows_cit, size_t)
		 */
		double& operator()(rows_it row, size_t col);

		/**
		 * \brief checked read-only access to an element in the surface data
		 * matrix.
		 *
		 * \param row an iterator into the desired row
		 * \param col the column index of the desired column
		 *
		 * \exception OutOfRange_e if \p row or \p col is out of bounds
		 *
		 * \sa operator(rows_it, size_t)
		 */
		double operator()(rows_cit row, size_t col) const;

		/**
		 * \brief assigns new values for a complete column
		 *
		 * The method sets new values for the specified column \p col in all row
		 * vectors.
		 *
		 * \param col     the index of the desired column
		 * \param coldata the new column vector
		 *
		 * \exception OutOfRange_e if \p row or \p col is out of bounds
		 *
		 * \sa addColumn(cconst std::string&, const RbDblV&)
		 */
		void setColumn(size_t col, const RbDblV& coldata);

		//************************************************************************
		// ITERATOR ACCESS METHODS
		//************************************************************************
			
		//@{
		///\name iterator access methods
		/**
		 * \brief iterator for the row vector sequence
		 *
		 * the surface data class provides constant and non-constant iterator
		 * functions for the first element and end of sequence in the rowvector
		 * list.
		 */
		rows_it begin() { return rows_.begin(); }
		rows_it end() { return rows_.end(); }
		rows_cit begin() const { return rows_.begin(); }
		rows_cit end() const { return rows_.end(); }
		//@}
						
	private:
			
		//************************************************************************
		// PRIVATE MEMBERS
		//************************************************************************

		/**
		 * \brief vector of column head names
		 *
		 * The number of columns in the surface data object is allways equal to
		 * the number of column heads.
		 */
		std::vector<std::string> columns_;

		/**
		 * \brief the list of row vectors
		 *
		 * This list contains a row vector of double values for each point in the
		 * surface.
		 */
		rows_t rows_;
	};

	//****************************************************************************
	//INLINE METHODS
	//****************************************************************************

	//****************************************************************************
	inline SurfaceData& SurfaceData::operator=(const SurfaceData& src)
	//****************************************************************************
	{
		columns_ = src.columns_;
		rows_    = src.rows_;
			
		return *this;
	}

	//****************************************************************************
	inline double& SurfaceData::operator()(rows_it row, size_t col)
	//****************************************************************************
	{
		if ((col<0) || (col>=columns_.size()))
			throw OutOfRange_e("The column index was out of bounds!");
			
		return (*row)[col];
	}

	//****************************************************************************
	inline double SurfaceData::operator()(rows_cit row, size_t col) const
	//****************************************************************************
	{
		if ((col<0) || (col>=columns_.size()))
			throw OutOfRange_e("The column index was out of bounds!");
	
		return (*row)[col];
	}


#ifdef HAVE_NAMESPACES
};
#endif 

#endif // SURFACEDATA_HH
