/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifdef HAVE_STD
#include <fstream>
#else
#include <fstream.h>
#endif

#include "surfaceio.hh"
#include "doublettchecker.hh"
#include "point.hh"
#include "vectorview.hh"
#include "simplebuilder.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
#endif 

//#include <iostream>

//******************************************************************************
WrongFormat_e::WrongFormat_e(int formatID, const string& format, 
                             const std::string& filename )
		: SurfaceIORead_e(WRONGFORMAT_ERR,"")
//******************************************************************************
{
	string location= (filename!="") ? filename : "IO stream";

	char buffer[50];
  sprintf(buffer, "(%x) %s", formatID, format.c_str());
	message_  = "The ";
	message_ += location;
	message_ += " does not have the right file format: ";
	message_ += buffer;
}

//******************************************************************************
SurfaceEntity* SurfaceIO::read(const std::string& afile)
//******************************************************************************
{
  ifstream stream(afile.c_str());
  if (!stream)
  {
		//prepare the message string
		stringstream errormsg;
		errormsg << "File: " << afile << " not found" << endl;

		throw SurfaceIORead_e(E_FILE_NOT_FOUND,errormsg.str());
    return 0;
  }

  SurfaceEntity* result;

	try
	{
    result = read(stream);  
  }
	catch (WrongFormat_e err)
	{
    stream.close();
		//catch the WrongFormat exception to add the file information
		throw WrongFormat_e(err.getFormatID(), err.getFormat(), afile);
	}

  stream.close();
  return result;
}

//******************************************************************************
SurfaceEntity* SurfaceIO::read(std::istream& astream)
//******************************************************************************
{
	if (doublettchecker_) doublettchecker_->clear();
}

//******************************************************************************
bool SurfaceIO::write (const std::string& afile, SurfaceEntity& surface)
//******************************************************************************
{
  ofstream stream(afile.c_str());
  if (!stream)
  {
		//prepare the message string
		stringstream errormsg;
		errormsg << "File: " << afile << " could not be opened" << endl;

		throw SurfaceIOWrite_e(E_CANNOT_OPEN_FILE ,errormsg.str());
    return false;
  }
  
  bool result = write(stream, surface);  
  stream.close();
  return result;	
}

//******************************************************************************
bool SurfaceIO::write (const std::string& afile, VectorView& view)
//******************************************************************************
{
  ofstream stream(afile.c_str());
  if (!stream)
  {
		//prepare the message string
		stringstream errormsg;
		errormsg << "File: " << afile << " could not be opened" << endl;

		throw SurfaceIOWrite_e(E_CANNOT_OPEN_FILE ,errormsg.str());
    return false;
  }
  
  bool result = write(stream, view);  
  stream.close();
  return result;	
}

//******************************************************************************
bool SurfaceIO::write (ostream & astream,	SurfaceEntity& surface)
//******************************************************************************
{
	SimpleBuilder<VectorView> builder;
	auto_ptr<VectorView> 
		view(dynamic_cast<VectorView*>(surface.createView(builder)));

	if (view.get())
		return write(astream, *view);
	else
		return false;
}

//******************************************************************************
SurfaceIO::SurfaceIO(bool checkdoubletts)
//******************************************************************************
:mDoubletts(), fDoublettDetection(false), doublettchecker_(0)
{
	enableDoublettDetection(checkdoubletts);
}

//******************************************************************************
SurfaceIO::~SurfaceIO()
//******************************************************************************
{
	if (doublettchecker_)
	{
		delete doublettchecker_;
		doublettchecker_ = 0;
	}
}

//******************************************************************************
void SurfaceIO::enableDoublettDetection(bool on)
//******************************************************************************
{
	if (on!=fDoublettDetection)
	{
		fDoublettDetection = on;
		if (on && (!doublettchecker_))
			doublettchecker_ = new DoublettChecker();
		else if ((!on) && (doublettchecker_))
		{
			delete doublettchecker_;
			doublettchecker_ = 0;
		}
	}
}

//******************************************************************************
bool SurfaceIO::doublettCheck (int ix, const Point& point)
//******************************************************************************
{
	if (doublettchecker_)
	{
		mDoubletts[ix] = (*doublettchecker_)(point);

		return (doublettchecker_->wasDoublett());
	}
	else
	{
		mDoubletts[ix]=ix;
		return false;
	}
}
 
//******************************************************************************
bool SurfaceIO::isDegenerated(const TriangleData& triangle) const
//******************************************************************************
{
	for(int i=0;i<3;++i)
		for(int j=i+1;j<3;++j)
			if (triangle.points[i]==triangle.points[j])
				return true;

	return false;
}

//******************************************************************************
void SurfaceIO::checkTrianglePoints(size_t pointcount, 
																		const TriangleData& triangle) const
//******************************************************************************
{
	for(int i=A;i<=C;++i)
	{
		//check correct index
		if ((triangle.points[i]<0) || (triangle.points[i]>=pointcount))
		{
			//throw an exception
			stringstream errormsg;
			errormsg << "The triangle (" 
							 << triangle.points[A] << ", " 
							 << triangle.points[B] << ", "
							 << triangle.points[C] << ") contains a wrong point index: " 
							 << triangle.points[i] << endl;
			
			//throw the exception
					throw SurfaceIORead_e(E_WRONG_POINT_INDEX,errormsg.str());
		}
	}	
}
