/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SURFACEIO_HH
#define SURFACEIO_HH

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>

#ifdef HAVE_STD
#include <map>
#include <iostream>
#else
#include <map.h>
#include <iostream.h>
#endif

#include "common.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif

	class Point;
	class SurfaceEntity;
	class VectorView;

	class DoublettChecker;

	/**
	 * \brief exception that notifies IO-read errors
	 */
	class SurfaceIORead_e : public Surfcomp_e 
	{ 
	public:
		/**
		 * \brief standard constructor
		 */
		SurfaceIORead_e(int errorno, const std::string& msg) 
				: Surfcomp_e(errorno,msg) { }
	};

	/**
	 * \brief exception that notifies IO-write errors
	 */
	class SurfaceIOWrite_e : public Surfcomp_e
	{ 
	public:

		/**
		 * \brief standard constructor
		 */
		SurfaceIOWrite_e(int errorno, const std::string& msg) 
				: Surfcomp_e(errorno,msg) { }
	};

	const int WRONGFORMAT_ERR = -102;
	/**
	 * \brief error for wrong SLD file formats#
	 *
	 * The error code of this error is -101. In addition the exception has a
	 * format_ parameter that stores the wrong file format ID.
	 */
	class WrongFormat_e : public SurfaceIORead_e
	{
	public:
		/**
		 * \brief constructor
		 *
		 * The constructor takes the wrong format ID and the filename and generates
		 * a suitable error message.
		 *
		 * \param format   the wrong format ID
		 * \param filename the name of this file or empty for a stream
		 */
		WrongFormat_e(const int formatID, const string& format, 
                  const std::string& filename );

		/**
		 * \brief retrieves the wrong format string
		 *
		 * \return the wrong format string
		 *
		 * \sa format_
		 */
		string getFormat() const { return format_; }

		/**
		 * \brief retrieves the wrong format ID code
		 *
		 * \return the wrong format ID code
		 *
		 * \sa formatID_
		 */
		int getFormatID() const { return formatID_; }

	private:
		
		/**
		 * \brief the wrong format string
		 */
		string format_;
		
		/**
		 * \brief the wrong format ID code
		 */
		int formatID_;
	};

	/**
	 * \brief the base class for all surface IO classes
	 */
	class SurfaceIO 
	{
	public:

		//**************************************************************************
		// CONSTRUCTOR AND DESTRUCTOR
		//**************************************************************************

		/**
		 * \brief standard and default constructor
		 *
		 * The constructor initializes the doublett checker if demanded and sets
		 * all other parameters to meaningfull values.
		 *
		 * \param checkdoubletts turns doublett checking on or off
		 */
		SurfaceIO(bool checkdoubletts = false);

		/**
		 * \brief destructor
		 *
		 * The destructor takes care of the doublett checker if it is still
		 * active.
		 */
		~SurfaceIO();

    /** 
     * \brief factory method
     * 
     * \return a fresh instance of a SurfaceIO class
     */
    static SurfaceIO* createInstance() { return 0; }
			
		//**************************************************************************
		// OPERATIONS
		//**************************************************************************

		//@{
		///\name reader methods
		/**
		 * \brief method to read a surface definition from a file
		 *
		 * The method reads a surface definition from a file. With this information
		 * a new SurfaceEntity object is created and returned to the user. If the
		 * surface could not be read a SurfaceIORead_e error is triggered and NULL
		 * returned.
		 *
		 * \param afile   the name of an existing filename that contains the
		 *                surface information
		 *
		 * \return a pointer to a valid Surface object that contains the new
		 *         surface.
		 *
		 * \exception SurfaceIORead_e whenever an error during the writing
		 *                            occurs. The principle problem is given by 
		 *                            the error message.
		 *
		 * \sa read(std::istream), 
		 *     writeToFile(const std::string&), 
		 *     writeToStream(std::ostream& astream)
		 */
		virtual SurfaceEntity* read(const std::string& afile);
		/**
		 * \brief method to read a surface definition from a stream
		 *
		 * The method reads a surface definition from an input stream. With this
		 * information a new SurfaceEntity object is created and returned to the
		 * user. If the surface could not be read a SurfaceIORead_e error is
		 * triggered and NULL returned.
		 *
		 * \param astream a valid input stream positioned at the beginning of a 
		 *                surface definition
		 *
		 * \return a pointer to a valid Surface object that contains the new
		 *         surface.
		 *
		 * \exception SurfaceIORead_e whenever an error during the writing
		 *                            occurs. The principle problem is given by 
		 *                            the error message.
		 *
		 * \sa write(const std::string&), write(std::ostream& astream)
		 */
		virtual SurfaceEntity* read(std::istream& astream);
		//@}
    
		//@{
		///\name entity writer methods
		/**
		 * \brief method to write a surface definition into a file/stream
		 *
		 * The method writes a surface definition of the object \p surface into a
		 * file or an output stream. If the surface could not be written a
		 * SurfaceIOWrite_e error is triggered and \c false returned.
		 *
		 * \param afile   the name of an existing filename that contains the
		 *                surface information
		 *
		 * or
		 *
		 * \param astream a valid input stream positioned at the beginning of a 
		 *                surface definition
		 * \param surface a pointer to a valid SurfaceEntity object
		 *
		 * \return \c true if the surface could be written successfully, \c false
		 *         otherwise.
		 *
		 * \exception SurfaceIOWrite_e whenever an error during the writing
		 *                             occurs. The principle problem is given by 
		 *                             the error message.
		 *
		 * \sa write(const std::string&, const SurfaceEntity&), 
		 *     write(ostream& astream, const SurfaceEntity&),
		 *     write(const std::string&, const VectorView&), 
		 *     write(ostream& astream, const VectorView&)
		 */
		virtual bool write (const std::string& afile, SurfaceEntity& surface);
		virtual bool write (std::ostream & astream, SurfaceEntity& surface);
		//@}

		//@{
		///\name view writer methods
		/**
		 * \brief method to write a surface definition into a file/stream
		 *
		 * The method writes a surface definition of the vector view object \p view
		 * into a file or an output stream. If the surface could not be written a
		 * SurfaceIOWrite_e error is triggered and \c false returned.
		 *
		 * \param afile   the name of an existing filename that contains the
		 *                surface information
		 *
		 * or
		 *
		 * \param astream a valid input stream positioned at the beginning of a 
		 *                surface definition
		 * \param view a reference to a valid VectorView object
		 *
		 * \return \c true if the surface could be written successfully, \c false
		 *         otherwise.
		 *
		 * \exception SurfaceIOWrite_e whenever an error during the writing
		 *                             occurs. The principle problem is given by 
		 *                             the error message.
		 *
		 * \sa write(const std::string&, const SurfaceEntity&), 
		 *     write(ostream& astream, const SurfaceEntity&),
		 *     write(const std::string&, const VectorView&), 
		 *     write(ostream& astream, const VectorView&)
		 */
		virtual bool write (const std::string& afile, VectorView& view);
		virtual bool write (std::ostream & astream, VectorView& view) = 0;
		//@}

		//**************************************************************************
		// PROPERTY METHODS
		//**************************************************************************

		/**
		 * \brief states if the doublett detection is on or off
		 *
		 * \return \c true if the doublett detection is on, \c false otherwise.
		 *
		 * \sa fDoublettDetection, enableDoublettDetection(bool)
		 */
		bool isDoublettDetection() { return fDoublettDetection; }
		/**
		 * \brief enables or disables the doublett detection
		 *
		 * If the doublett detection is turned on a DoublettChecker object is
		 * created and deleted if the doublett detection is turned off. So any
		 * information stored in the doublett detector is deleted when doublett
		 * detection is turned off.
		 *
		 * \param on if \c true doublett detection is turned on, off otherwise.
		 *
		 * \sa isDoublettDetection(), fDoublettDetection
		 */
		void enableDoublettDetection(bool on = true);
  
	protected:

		//**************************************************************************
		// PROTECTED DATA TYPES
		//**************************************************************************
		
		/**
		 * \brief the triangle information as it comes from the files
		 *
		 * In the files the triangle points are stored by indices. Internally they
		 * are stored as iterators. To allow a compact representation of the file
		 * information one can use this data structure.
		 *
		 * 
		 * \sa isDegenerated(const TriangleData&)
		 */
		struct TriangleData
    {
      int points[3];
      int triangle_face;
      int face_number;
    };

		//**************************************************************************
		// PROTECTED METHODS
		//**************************************************************************

		/**
		 * \brief checks for doubletts
		 *
		 * The methods calls the doublett checker and passes the \p point to get
		 * back any information if the point is a doublett and which point is the
		 * original one. This information is stored in the mDoubletts map
		 *
		 * \param ix    the original index of the point
		 * \param point a pointer to the new point that should be checked
		 *
		 * \return \c true if the point is a doublett of an already existing point,
		 *         \c false if it is unique.
		 *
		 * \sa mDoubletts, doublettchecker_
		 */
		bool doublettCheck (int ix, const Point& point);

		/** 
		 * \brief checks for the consistency of the triangle
		 *
		 * Sometimes surface representation contain duplicated points (see
		 * doublettCheck()). If they are collapsed to one point, smoe triangles may
		 * consist of at least two equal points. These triangles are bad and must
		 * not be registered in the data structre.
		 * 
		 * \param triangle the triangle that has to be checked
		 * 
		 * \return \c true if the triangle has not three unique edge points, \c
		 *         false if it is ok.
		 */
		bool isDegenerated(const TriangleData& triangle) const;
		
		/** 
		 * \brief checks the validity of the triangle point indices
		 * 
		 * All three point indices of the triangle are checked if they are within
		 * the range of surface points. If one index is out of the range a
		 * SurfaceIORead_e exception is thrown.
		 * 
		 * \exception SurfaceIORead_e if a point index is out of range
		 * 
		 * \param pointcount the number of points in the file
		 * \param triangle   the file input data of the triangle (especially the
		 *                   point indices)
		 */
		void checkTrianglePoints(size_t pointcount, 
														 const TriangleData& triangle) const;


		//**************************************************************************
		// PROTECTED MEMBERS
		//**************************************************************************

		/**
		 * \brief the doublett map
		 *
		 * The doublett map holds for every point index in the file the real
		 * index in the surface's point vector.
		 */
		std::map<size_t,size_t> mDoubletts;

	private:

		//**************************************************************************
		// PRIVATE MEMBERS
		//**************************************************************************
			
		/**
		 * \brief flag that indicates active doublett detection
		 *
		 * \sa isDoublettDetection(), enableDoublettDetection(bool), 
		 *     doublettchecker_
		 */
		bool fDoublettDetection;

		/**
		 * \brief the active doublett checker
		 *
		 * fDoubletDetection, doublettCheck(Point*)
		 */
		DoublettChecker* doublettchecker_;
	};

#ifdef HAVE_NAMESPACES
};
#endif

#endif //SURFACEIO_HH
