/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SURFACEVIEW_HH
#define SURFACEVIEW_HH

/**
 * \file surfaceview.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 14 Mar 2004, last CVS commit 2004/03/14 13:25:09
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <list>
#  include <map>
#else
#  include <list.h>
#  include <map.h>
#endif

#include "common.hh"
#include "surface.hh"


//#include <iostream>

#define CHECKREF \
  if (!surface_) throw InvalidReference_e("surface_ points to NULL");

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif

	class ViewIterator;

	/**
	 * \brief an abstract base class for surface views
	 *
	 * A SurfaceView is an observer of a SurfaceEntity. It holds an internal
	 * pointer to the viewed SurfaceEntity which is made public as a reference
	 * (setSurfaceEntity() and getSurfaceEntity()). The view can be updated via
	 * the abstract update() method. The SurfaceView detaches itself from the
	 * surface_ when it is destructed.
	 *
	 * SurfaceView provides a one-to-one propagation of the simple SurfaceEntity
	 * properties sphere-count, probe-radius and point-density.
	 *
	 * A SurfaceView can also be observed by ViewIterator objects, it therefore
	 * has also a notify(), attach() and detach() method.
	 *
	 * \todo add copy constructor for surface views
	 * \todo enable the turn-off of notifications during build process
	 */
	class SurfaceView
	{
	public:

		//**************************************************************************
		// CONSTRUCTOR AND DESTRUCTOR
		//**************************************************************************

		/**
		 * \brief default constructor
		 *
		 * Creates a surface view, that does not view any SurfaceEntity.
		 *
		 * \sa SurfaceView(SurfaceEntity& const)
		 */
		SurfaceView() : surface_(0), entityowner_(false) { }
		
		/**
		 * \brief initializing constructor
		 *
		 * This constructor creates a surface view, that observes the entity \p
		 * surface.
		 *
		 * \param surface SurfaceEntity that is observed.
		 *
		 * \sa SurfaceView()
		 */
		SurfaceView(SurfaceEntity& surface) 
        : surface_(&surface), entityowner_(false)  { surface_->attach(*this); }

		/**
		 * \brief destructor
		 *
		 * The destructor detaches itself from the ovserved SurfaceEntity.
		 */
		virtual ~SurfaceView();

		/**
		 * \brief preoccupies memery for a specified amount of selected points and
		 *        triangles.
		 *
		 * This method can be used in the subclasses to reserves space for \p
		 * pointcount selected points and \p trianglecount selected triangles. If
		 * the parameters are omitted, it should reserves enough memory to select
		 * every point and triangle of the observed SurfaceEntity.
		 *
		 * The base-class implementation does nothing.
		 *
		 * \param pointcount    expected number of selected points
		 * \param trianglecount expected number of selected triangles
		 *
		 * \sa pointits_, trianglits_
		 */
		virtual void reserve(size_t pointcount = 0, size_t trianglecount = 0) { }

		//**************************************************************************
		// OBSERVER (observed and observing)
		//**************************************************************************

		/**
		 * \brief notification method
		 *
		 * The update method is called, usually by the observed SurfaceEntity
		 * (surface_) to inform the view, that the surface's internal state has
		 * changed.
		 */
		virtual void update() = 0;

		/**
		 * \brief assigns the SurfaceEntity, that should be observed
		 *
		 * A view can be connected to a SurfaceEntity only once, thus the
		 * setSurfaceEntity method assigns \p surface only if surface_ has no value
		 * (i.e. is NULL). If the surface_ has allready been set, the method
		 * returns \c false.
		 *
		 * \param surface the observed SurfaceEntity
		 *
		 * \return \c true if the surface was set, \c false if it was not possible
		 *         any more.
		 *
		 * \sa SurfaceEntity(SurfaceEntity& const), getSurfaceEntity(), surface_
		 */
		virtual bool setSurfaceEntity(SurfaceEntity& surface);
		/**
		 * \brief retrieves the observed SurfaceEntity
     *
		 * \return the observed SurfaceEntity
		 *
		 * \exception InvalidReference_e if surface_ has no value.
		 */
		SurfaceEntity& getSurfaceEntity() const;
    /** 
     * \brief retrieves the observed entity and transfers the ownership
     *
     * The method gives back a pointer to the observed SurfaceEntity object and
     * transfers the ownership to the calling object.
     * 
     * \return a pointer to the observed surface entity
     *
     * \sa isOwner(), entityowner_
     */
    SurfaceEntity* takeSurfaceEntity() const;

		/**
		 * \brief notifies the attached observers about a change
		 *
		 * The notification is usually a broadcast of a notification message, the
		 * SurfaceEntity sends.
		 */
		virtual void notify();

		/**
		 * \brief registers a new ViewIterator as observer 
		 *
		 * \param observer a reference to a new ViewIterator, that observes this
		 *                 view
		 *
		 * \sa detach(ViewIterator&)
		 */
		void attach(ViewIterator& observer);
		/**
		 * \brief unregisters an obsolete ViewIterator
		 *
		 * \param observer a reference to an obsolete ViewIterator
		 *
		 * \sa attach(SurfaceView*)
		 */
		void detach(ViewIterator& observer);
		
    //**************************************************************************
    // DEEP COPY
    //**************************************************************************

    /** 
     * \brief a deep copy operation.
     *
     * The deep copy operation generates a new SurfaceEntity object that
     * containes the selection of the current view and builds a view of the
     * exact subclass to show the total information of the new entity. The
     * specific operations are performed by the virtual methods
     * (buildSelectionEntity(DeepCopyInformation&),
     * buildSelectionView(DeepCopyInformation&),
     * postCopying(DeepCopyInformation&)).
     *
     * The created SurfaceEntity object is owned by the new view, that means
     * that it is automatically destroyed when the view is deleted. To get the
     * ownership of the new entity, one has to call getSurfaceEntity(bool) with
     * the \c newowner flag set to \c true.
     *
     * The clone operation is a Template Method (325) pattern.
     * 
     * \return a new surface view that points to an entity (deep) copy of this
     *         view's selection
     */
    SurfaceView* clone() const;

  protected:

    /**
     * \brief a parameter class, used in the deep copy operation
     *
     * This class collects all the information that is necessary in a view deep
     * copy operation:
     *
     * - a map between the point iterators of the old and new entity
     * - a map between the triangle iterators of the old and new entity
     * - a pointer to the new surface view
     */
    struct DeepCopyInformation
    {
      std::map<SurfaceEntity::points_it, 
               SurfaceEntity::points_it, 
               lt_iterator<SurfaceEntity::points_cit> > pointmap;

      std::map<SurfaceEntity::triangles_it, 
               SurfaceEntity::triangles_it,
               lt_iterator<SurfaceEntity::triangles_cit> > trianglemap;

      SurfaceView* newview;

      DeepCopyInformation(): pointmap(), trianglemap(), newview(0) { }
    };

    /** 
     * \brief generates a new SurfaceEntity based on the current selection
     *
     * This method is an operation of the clone() method. The new entity, that
     * should receive copies of the selected points and triangles is passed by
     * the SurfaceView::surface_ member of the DeepCopyInformation::newview
     * member. 
     *
     * \note This method must be overriden by all views that support a deep
     * copy.
     * 
     * \param copyinfo a reference to the copy information of the clone
     *                 operation
     */
    virtual void buildSelectionEntity(DeepCopyInformation& copyinfo) const { }
    /** 
     * \brief builds the cloned view in a deep copy operation
     *
     * This method is an operation of the clone() method. The new view, that
     * should observe all the points and triangles of the new entity is given
     * by the DeepCopyInformation::newview member.
     *
     * \note This method must be overriden by all views that support a deep
     * copy.
     *
     * \param copyinfo a reference to the copy information of the clone
     *                 operation
     */    
    virtual void buildSelectionView(DeepCopyInformation& copyinfo) const { }
    /** 
     * \brief post copying operations in deep copy processes
     *
     * This method is an operation of the clone() method. 
     *
     * \note This method must be overriden by all views that need to perform
     * post copying stuf in a deep copy.
     *
     * \param copyinfo a reference to the copy information of the clone
     *                 operation
     */    
    virtual void postCopying(DeepCopyInformation& copyinfo) const { }

  public:

		//**************************************************************************
		// POINT DATA
		//**************************************************************************

		/**
		 * \brief retrieves the number of columns in the point data matrix
		 *
		 * \return the number of columns in each point data vector
		 *
		 * \sa data_
		 */
		size_t getNColumns() const { CHECKREF; return surface_->getNColumns(); }

		//@{
		/**
		 * \brief adds a column at the end of the point data matrix
		 *
		 * The definition of the parameters and the effect is the same as of
		 * SurfaceData::addColumn(const std::string&, double) or
		 * SurfaceData::addColumn(const std::string&, RbDblV&).
		 *
		 * \param name the name of the column 
		 * \param def  a default value or column vector that is assigned to all 
		 *             new entries in existing rows.
		 *
		 * \return the index of the new column or -1 if the surface is write
		 *         locked.
		 *
		 *     deleteColumn(rows_it), rows_
		 */
		size_t addColumn(const std::string& name, double def=0.0) {
			CHECKREF; return surface_->addColumn(name, def); }
		size_t addColumn(const std::string& name, const RbDblV& def) {
			CHECKREF; return surface_->addColumn(name, def); }
		//@}

		/**
		 * \brief removes \p col from the point data matrix
		 *
		 * \param col the column to remove
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa addColumn(const std::string&, double), 
		 *     addRow(const std::string&, const RbDblV&)
		 */
		void deleteColumn(size_t col) {
			CHECKREF; surface_->deleteColumn(col); }

		/**
		 * \brief sets the elements of a column to new values
		 *
		 * The method sets new values for the specified column \p col in the
		 * surface point data matrix.
		 *
		 * \param col     the index of the desired column
		 * \param coldata the new column vector
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa setColumnTitle(), SurfaceData::setColumn()
		 */
		void setColumn(size_t col, const RbDblV& coldata) {
			CHECKREF; surface_->setColumn(col, coldata); }

		/**
		 * \brief retrieves the name of column \p ix
		 *
		 * \param col  the index of the column
		 *
		 * \return the name of the column \p ix or an empty string if the index
		 *         is out of bounds.
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa setColumnName(size_t, const std::string&), 
		 *     SurfaceData::getColumnName(size_t)   
		 */
		std::string getColumnName(size_t col) const {
			CHECKREF; return surface_->getColumnName(col); }
		/**
		 * \brief sets the name of a specific column
		 *
		 * \param col  the index of the column
		 * \param name the new name of the column
		 *
		 * \exception OutOfRange_e if the index is out of bounds
		 *
		 * \sa getColumnName(size_t),
		 *     SurfaceData::setColumnName(size_t, const std::string&)
		 */
		void setColumnName(size_t col, const std::string& name) {
			CHECKREF; surface_->setColumnName(col, name); }
			
		/**
		 * \brief creates a nxn double matrix from the surface point data
		 *
		 * \return a RbDblM object with the data of this object.
		 *
		 * \sa SurfaceData::extractMatrix()
		 */
		RbDblM extractDataMatrix() const { 
			CHECKREF; return surface_->extractDataMatrix(); }
		/**
		 * \brief creates a column vector copy of the \p col'th column of the
		 *        surface point data
		 * \param col the desired column
		 *
		 * \sa SurfaceData::extractColumn(size_t)
		 */
		RbDblV extractDataColumn(size_t col) const { 
			CHECKREF; return surface_->extractDataColumn(col); }
    
    /** 
     * \brief retrieves the central point of the selection
     * 
     * \return the central point of the current selection
     */
    virtual RbDblV getCenter() const = 0;
    /** 
     * \brief the extensions in every direction of the view's points
     * 
     * \return the x,y,z extensions of the points selected by the view
     */
    virtual RbDblV getExtensions() const = 0;

		//**************************************************************************
		// OTHER PROPERTY METHODS
		//**************************************************************************

		/**
		 * \brief retrieves the number of atomic spheres in the molecule
		 *
		 * \return the number of atomic spheres in the molecule that build the
		 *          surface
		 *
		 * \sa spherecount_, setSphereCount(size_t)
		 */
		size_t getSphereCount() const { 
			CHECKREF; return surface_->getSphereCount(); }
		/**
		 * \brief sets the number of atomic spheres to a new value
		 *
		 * \param spherecount the new number of atomic spheres that build the
		 *                    surface
		 *
		 * \sa spherecount_, getSphereCount()
		 */
		void setSphereCount(size_t spherecount) { 
			CHECKREF; return surface_->setSphereCount(spherecount); }

		/**
		 * \brief retrieves the point density on the surface
		 *
		 * \return the number of surface points per square angstrom.
		 *
		 * \sa pointdensity_, setPointDensity(double)
		 */
		double getPointDensity() const { 
			CHECKREF; return surface_->getPointDensity(); }
		/**
		 * \brief sets the point density to a new value
		 *
		 * \param density the new number of of surface points per square
		 *                angstrom.
		 *
		 * \sa pointdensity_, getPointDensity()
		 */
		void setPointDensity( double density ) {
			CHECKREF; return surface_->setPointDensity(density); }

		/**
		 * \brief gives back the radius of the probe sphere
		 *
		 * \return the radius of the probe sphere
		 *
		 * \sa proberadius_, setProbeRadius(double)
		 */
		double getProbeRadius() const { 
			CHECKREF; return surface_->getProbeRadius(); }
		/**
		 * \brief sets the radius of the probe sphere to a new value
		 *
		 * \param radius the new radius of the probe sphere
		 *
		 * \sa proberadius_, getProbeRadius()
		 */
		void setProbeRadius(double radius) {
			CHECKREF; return surface_->setProbeRadius(radius); }

    /** 
     * \brief retrieves the status of the relation between the observed entity
     *        and the view.
     *
     * \return \c true if the view is the owner of the observed SurfacEntity object.
     *
     * \sa entityowner_, takeSurfaceEntity()
     */
    bool isOwner() const { return entityowner_; }

  protected:
    
    //**************************************************************************
    // PROTECTED METHODS
    //**************************************************************************

    /**
     * \brief internal factory method.
     *
     * The internal factory method is used to get a new instance of the current
     * SurfaceView subclass.
     *
     * \return a new object of the same SurfaceView type.
     */
    virtual SurfaceView* createView() const = 0;

		/**
		 * \brief a pointer to the observed SurfaceEntity object
		 */
		SurfaceEntity* surface_;

	private:

    //**************************************************************************
    // DATA MEMBERS
    //**************************************************************************

		/**
     * \brief indicates the ownership of the view over its observed entity.
     *
     * If the view is the owner of the observed SurfaceEntity, it will destroy
     * the entity in its destructor. A view can only be the owner of the entity
     * if it has created the entity itself (deep copy ...).
     */ 
    mutable bool entityowner_;

		/**
		 * \brief the list of observers in form of ViewIterators
		 */
		std::list<ViewIterator*> observers_;

	};

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // SURFACEVIEW_HH
