/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef TRIANGLE_HH
#define TRIANGLE_HH

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "common.hh"
#include "point.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif //HAVE_NAMESPACES

	class SurfaceEntity;
		
	/**
	 * \brief the triangle class of the molecule surfaces
	 *
	 * A surface primarly consists of points but this points are grouped into
	 * triangles. Triangles provide the neighborhood information of the surface
	 * points. They thus have references to the three Point objects that build
	 * the vertices of the triangle (vertices_). Additionally the type of the
	 * face (facetype_) and the number of the analytical surface (facenum_) are
	 * stored for each triangle. The last three flags are accessible via the
	 * usual get and set methods.
	 *
	 * The vertices can only be set in the construction of the triangle. Access
	 * to the vertices is later possible via the element operator
	 * (operator[](int)) which returns the Point object associated with the
	 * particular vertex position (use the defined constants A, B, or C). An
	 * additional method delivers the successor of a distinct vertex and with
	 * isPointVertex() the membership of a Point object to a triangle can bee
	 * checked.
	 */
	class Triangle
	{
	public:

		//************************************************************************
		// TYPE DEFINITIONS
		//************************************************************************

		///\brief constant point list iterator
		typedef list<Point>::const_iterator points_cit;
		///\brief mutable point list iterator (reference type)
		typedef list<Point>::iterator points_it;

			
	public:
			
		//************************************************************************
		// CONSTRUCTORS
		//************************************************************************

		/**
		 * \brief default constructor
		 *
		 * The default constructor sets all the parameters to the given values
		 * or assigns appropriate values.
		 *
		 * \param facetype the type of this face
		 * \param facenum  the number of the analytical surface
		 *
		 * \sa Triangle(const Triangle&)
		 */
		Triangle(facetype_t facetype = ftUnknown, int facenum = -1);

		/**
		 * \brief standard constructor
		 *
		 * The standard constructor sets all the parameters to the given values
		 * or assigns appropriate values and assigns iterators to the vertices.
		 *
		 * \param a,b,c    iterators to the surface points that form that triangle
		 * \param facetype the type of this face
		 * \param facenum  the number of the analytical surface
		 *
		 * \sa Triangle(const Triangle&)
		 */
		Triangle(points_it a, points_it b, points_it c,
						 facetype_t facetype = ftUnknown, int facenum = -1);
		/**
		 * \brief copy constructor
		 *
		 * Creates a copy of the triangle \p src.
		 *
		 * \param src the source triangle
		 *
		 * \sa Triangle(Point* const, Point* const, Point* const, facetype_t, int)
		 */
		Triangle(const Triangle& src);

		//************************************************************************
		// OPERATIONS
		//************************************************************************

    /** 
     * \brief copies the flags from a \p src triangle
     *
     * The flags are the 
     * - facetype_ and
     * - facenum_ 
     * members.
     * 
     * \param src the template triangle
     */
    void copyFlags(const Triangle& src);

		/**
		 * \brief checks if the given \p point is a vertex of the triangle
		 *
		 * \return \c true if \p point is pointing to a vertex of the triangle
		 *
		 * \sa vertices_, operator[](int)
		 */
		bool isPointVertex(points_cit point) const;
			
		//************************************************************************
		// PROPERTY METHODS
		//************************************************************************

		/**
		 * \brief gives back the type of this face
		 *
		 * \return the type of this surface
		 *
		 * \sa facetype_, setFaceType(facetype_t)
		 */
		facetype_t getFaceType() const { return facetype_; }
		/**
		 * \brief sets the type of the surface to a new value
		 *
		 * \param facetype the new triangle face type
		 *
		 * \sa facetype_, getFaceType()
		 */
		void setFaceType(facetype_t facetype) { facetype_ = facetype; }

		/**
		 * \brief retrieves the index of the analytical surface
		 *
		 * \return the number of the analytical surface
		 *
		 * \sa facenum_, setFaceNum(int)
		 */
		int getFaceNum() const { return facenum_; }
		/**
		 * \brief assigns a new index of the analytical surface
		 *
		 * \param facenum the new number of the analytical surface
		 *
		 * \sa facenum_, getFaceNum()
		 */
		void setFaceNum(int facenum) { facenum_ = facenum; }

		/**
		 * \brief retrieves a constant pointer to one of the vertices
		 *
		 * \param ix the index of the vertex (you can use the predefined constants 
		 *           A, B or C)
		 *
		 * \return a constant point associated with the desired vertex or zero if
		 * the index was out of range
		 */
		points_it getVertex(int ix) const;
			
		/**
		 * \brief the element operator returns the points associated with the
		 * vertices
		 *
		 * \param ix the index of the vertex (you can use the predefined constants 
		 *           A, B or C)
		 *
		 * \return the point associated with the desired vertex
		 *
		 * \exception OutOfRange_e if the index is less than zero or larger than
		 *                         two.
		 */
		points_it& operator[](int ix);

		/**
		 * \brief finds the successor to the specified \p vertex in counter
		 * clockwise order
		 *
		 * The rule for the successor is simple:
		 * - B follows A 
		 * - C follows B
		 * - A follows C
		 *
		 *
		 * \param vertex the predeccessor.
		 *
		 * \return a pointer to the next vertex CCW or NULL if \p vertex is not
		 *         part of this triangle.
		 */
		points_it successor(points_it vertex) const;

		/** 
		 * \brief calculates the triangle center
		 * 
		 * The triangle center is calculated by
		 * 
		 * \f[
		 *    c_i = \frac{1}{3} \cdot \sum^{3}_{k=1} p_{k,i}
		 * \f]
		 *
		 * where <em>c<sub>i</sub></em> is the i'th coordinate of the center and
		 * <em>p<sub>k,i</sub></em> is the i'th coordinate of the k'th triangle
		 * point.
		 * 
		 * \return the centroid of the three triangle points
		 */
		RbDblV getCenter() const;

		/** 
		 * \brief retrieves the triangle normal
		 * 
		 * The normal vector has always a length of 1. For the calculation of the
		 * normal see calcNormal().
		 * 
		 * \return the normal vector of the triangle
		 *
		 * \sa calcNormal()
		 */
		RbDblV getNormal() const;

		/** 
		 * \brief retrieves the area of the triangle
		 * 
		 * The triangles area is deduced from the length of the triangle normal by
		 * the following relationship:
		 *
		 * \f[
		 *     A_t = \left\Vert \textrm{\bf{n}} \right\Vert
		 * \f]
		 *
		 * where <em>A<sub>t</sub></em> is the area of the surface and \b n is the
		 * full-length normal vector.
		 * 
		 * \return the area of the triangle
		 *
		 * \sa calcNormal()
		 */
		double getArea() const;
			
	private:

		//**************************************************************************
		// PRIVATE METHODS
		//**************************************************************************

		/**
		 * \brief calculates the normal of the triangle
		 * 
		 * The triangle normal is the cross product of the vectors that connect one
		 * point (e.g. b) with the other edges:
		 * 
		 * \f[
		 *    \textrm{\bf{n}} = 
     *      \frac{(\textrm{\bf{C}}-\textrm{\bf{B}})
     *            \times
     *            (\textrm{\bf{A}}-\textrm{\bf{B}})
		 *           }
     *           {\left\Vert 
     *            (\textrm{\bf{C}}-\textrm{\bf{B}})
		 *						\times
		 *            (\textrm{\bf{A}}-\textrm{\bf{B}})\right\Vert 
		 *           }
		 * \f]
		 *
		 * where <b>A</b>, <b>B</b> and <b>C</b> are the coordinate vectors of the
		 * three edge points and <b>n</b> is the normal vector. 
		 * 
		 * \return the full length triangle normal
		 *
		 * \sa getNormal(), getArea()
		 */
		RbDblV calcNormal() const;


		//**************************************************************************
		// PRIVATE MEMBERS
		//**************************************************************************

		/**
		 * \brief contains the type of the triangle face
		 *
		 * On molecular surfaces three different types are possible:
		 *
		 * - \c ftToric     for triangles which belong to toric reentrant faces 
		 * - \c ftReentrant for triangles inside reentrant faces 
		 * - \c ftContact   for triangles inside contact faces.
		 */
		facetype_t facetype_;

		/**
		 * \brief the number of the analytical surface
		 *
		 * The solvent accessible surface of a molecule can be divided into
		 * different parts each described by a simple equation. The facenum
		 * contains the index of the corrsponding analytical surface.
		 */
		int facenum_;

		/**
		 * \brief the array of the three vertices.
		 *
		 * Each vertex is represented by a pointer to the corresponding surface point
		 */
		points_it vertices_[3];

		friend class SurfaceEntity;
	};

	//**************************************************************************
	inline Triangle::Triangle(facetype_t facetype, int facenum)
    //**************************************************************************
	:facetype_(facetype), facenum_(facenum)
	{
	}

	//**************************************************************************
	inline Triangle::Triangle(points_it a, points_it b, points_it c,
														facetype_t facetype, int facenum)
    //**************************************************************************
	:facetype_(facetype), facenum_(facenum)
	{
		vertices_[A] = a;
		vertices_[B] = b;
		vertices_[C] = c;
	}

	//**************************************************************************
	inline Triangle::Triangle(const Triangle& src)
    //**************************************************************************
	:facetype_(src.facetype_), facenum_(src.facenum_)		
	{
		vertices_[A] = src.vertices_[A];
		vertices_[B] = src.vertices_[B];
		vertices_[C] = src.vertices_[C];
	}

	//**************************************************************************
	inline Triangle::points_it Triangle::getVertex(int ix) const
    //**************************************************************************
	{
		if ((ix>=A) && (ix<=C))
			return vertices_[ix];
		else 
			return 0;
	}

	//**************************************************************************
	inline bool Triangle::isPointVertex(Triangle::points_cit point) const
    //**************************************************************************
	{
		return ((point == vertices_[A]) ||
						(point == vertices_[B]) ||
						(point == vertices_[C]));
	}

#ifdef HAVE_NAMESPACES
}; 
#endif //HAVE_NAMESPACES

#endif // TRIANGLE_HH

