/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef TRIANGLELIST_HH
#define TRIANGLELIST_HH

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "common.hh"
#include "point.hh"
#include "surface.hh"
#include "triangle.hh"

namespace Surface
{
	/**
	 * \brief creates sequences of triangles that belong to one point
	 */
	class TriangleList
	{
	public:
			
		//************************************************************************
		// CONSTRUCTOR
		//************************************************************************
		
		/**
		 * \brief the standard constructor
		 *
		 * The triangle list controller expects a reference to a surface and an
		 * iterator to a point in this surface. The search index in the triangle
		 * vector is set to zero.
		 *
		 * \param surface a constant reference to a valid SurfaceEntity object 
		 * \param point   an iterator to a valid point on the surface.
		 */
		TriangleList (SurfaceEntity& surface,
									SurfaceEntity::points_cit point)
				:surface_(&surface), point_(point), fTriangleIndex(0) { }
		
		//************************************************************************
		// OPERATIONS
		//************************************************************************
			
		/**
		 * \brief retrieves the next triangle containing the point
		 * 
		 * The method searches for the next triangle in the triangle vector of
		 * the surface that contains this point. It starts at the position of
		 * triangle index plus the value of the \p jump parameter. fTriangleIndex
		 * is pointing to the triangle after the one that was found in the
		 * triangle list.
		 *
		 * \return an iterator to the next triangle that contains the point, or
		 *         SurfaceEntity::endTriangles() if no more triangles were found.
		 */
		SurfaceEntity::triangles_it findNext(int jump=0) const;

		/**
		 * \brief retrieves the complete list of triangles that contain the point
		 *
		 * The complete sequence of triangles containing the surface point point_
		 * are written into the output iterator \p ins by their iterators into
		 * the triangle list of the surface. The value of fTriangleIndex is not
		 * changed.
		 *
		 * \return the number of triangles that have been written into the
		 *         triangle list.
		 */
		template<class Out>
		size_t operator()(Out ins) const;

		//************************************************************************
		// PROPERTY METHODS
		//************************************************************************

		/**
		 * \brief retrieves the status of the triangle search index
		 *
		 * \return the position of the triangle index in the triangle list
		 *
		 * \sa fTriangleIndex, resetIndex()
		 */
		SurfaceEntity::triangles_it getTriangleIndex() const { 
			return fTriangleIndex; }
		/**
		 * \brief resets the position of the search index
		 *
		 * The fTriangleIndex is reset to the beginning of the triangle vector
		 *
		 * \sa fTriangleIndex, getTriangleIndex()
		 */
		void resetIndex() const { fTriangleIndex = surface_->beginTriangles(); } 
			
	private:

		//************************************************************************
		// PRIVATE MEMBERS
		//************************************************************************

		/**
		 * \brief the surface point of interest
		 */
		SurfaceEntity::points_cit point_;

		/**
		 * \brief the surface object
		 */
		SurfaceEntity* const surface_;

		/**
		 * \brief the actual index in the surface's triangle vector
		 */
		mutable SurfaceEntity::triangles_it fTriangleIndex;

	};

	//**************************************************************************
	template<class Out>
	size_t TriangleList::operator()(Out ins) const
		//**************************************************************************
	{
		SurfaceEntity::triangles_it t;
		size_t c = 0;
			
		for(t=surface_->beginTriangles();t!=surface_->endTriangles();++t)
			if (t->isPointVertex(point_)) 
			{
				*ins++ = t;
				++c;
			}

		return c;
	}
};


#endif //TRIANGLELIST_HH
