/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "vectorview.hh"

//#include <iostream>

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
#endif 

//******************************************************************************
void VectorView::update()
//******************************************************************************
{
	notify();
}

//******************************************************************************
void VectorView::buildSelectionEntity(DeepCopyInformation& copyinfo) const
//******************************************************************************
{
  SurfaceEntity& entity = copyinfo.newview->getSurfaceEntity();
  
  //add each point
  for(pointits_cit p=pointits_.begin();p!=pointits_.end();++p)
  {
    SurfaceEntity::points_it oldpoint = *p;
    SurfaceEntity::points_it newpoint = entity.addPoint();

    *newpoint = *oldpoint;
    newpoint->setData(oldpoint->data());

    //build up the pointmap
    copyinfo.pointmap[oldpoint] = newpoint;
  }

  //add each triangle
  for(triangleits_cit t=triangleits_.begin();t!=triangleits_.end();++t)
  {
    SurfaceEntity::triangles_it oldtriangle = *t;
    SurfaceEntity::triangles_it  newtriangle = 
      entity.addTriangle(copyinfo.pointmap[oldtriangle->getVertex(A)],
                          copyinfo.pointmap[oldtriangle->getVertex(B)],
                          copyinfo.pointmap[oldtriangle->getVertex(C)]);

    newtriangle->copyFlags(*oldtriangle);

    //build up the trianglemap
    copyinfo.trianglemap[oldtriangle] = newtriangle;
  }
}

//******************************************************************************
void VectorView::buildSelectionView(DeepCopyInformation& copyinfo) const
//******************************************************************************
{
  VectorView* view = dynamic_cast<VectorView*>(copyinfo.newview);
  view->reserve(getNPoints(), getNTriangles());

  //select each point
  for(pointits_cit p=pointits_.begin();p!=pointits_.end();++p)
    view->selectPoint(copyinfo.pointmap[*p]);

  //select each triangle
  for(triangleits_cit t=triangleits_.begin();t!=triangleits_.end();++t)
    view->selectTriangle(copyinfo.trianglemap[*t]);
}

//******************************************************************************
VectorView::pointIter VectorView::selectPoint(SurfaceEntity::points_it point)
//******************************************************************************
{
	pointits_it p;

	//look if point is already added to the selection
	pointmap_cit pm = pointmap_.find(point);

	if (pm == pointmap_.end())
	{
		pointits_.push_back(point);
		
		p = pointits_.end();
		pointmap_[point] = --p;
		
		notify();
	}
	else
		p = pm->second;

	return pointIter(*this, p);
}

//******************************************************************************
void VectorView::deselectPoint(pointIter pointiter)
//******************************************************************************
{
	pointits_.erase(pointiter.impl_);

	rebuildPointMap();
	notify();
}

//******************************************************************************
VectorView::triangleIter 
VectorView::selectTriangle(SurfaceEntity::triangles_it triangle,
													 size_t minimum)
//******************************************************************************
{
	int c_vertices = 0;
	triangleits_it t;

	//check the minimum parameter
	if (minimum>3) minimum = 3;

	//check every vertex of the triangle
	for(int v=A;v<=C;++v)
	{
		//look if the vertex/point is already added to the selection
		pointmap_cit pm = pointmap_.find(triangle->getVertex(v));
		if (pm!=pointmap_.end()) ++c_vertices;
	}

	//add only if minimum points are selected
	if (c_vertices>=minimum)
	{
		//look if the triangle has already been selected
		trianglemap_cit tm = trianglemap_.find(triangle);
		if (tm == trianglemap_.end())
		{
			triangleits_.push_back(triangle);

			t = triangleits_.end();
			trianglemap_[triangle] = --t;

			notify();
		}
		else
			t = tm->second;
	}
	else
		t = triangleits_.end();

	return triangleIter(*this,t);
}

//******************************************************************************
void VectorView::deselectTriangle(triangleIter triangleiter)
//******************************************************************************
{
	triangleits_.erase(triangleiter.impl_);

	rebuildTriangleMap();
	notify();
}

//******************************************************************************
void VectorView::reserve(size_t pointcount, size_t trianglecount)
//******************************************************************************
{
	if (pointcount>0)
		pointits_.reserve(pointcount);
	else
		pointits_.reserve(surface_->getNPoints());
		
	if (trianglecount>0)
		triangleits_.reserve(trianglecount);
	else
		triangleits_.reserve(surface_->getNTriangles());
}

//******************************************************************************
VectorView::pointIter VectorView::addPoint()
//******************************************************************************
{
	CHECKREF;
	if (!surface_->isLocked())
	{
		SurfaceEntity::points_it p = surface_->addPoint();
		return selectPoint(p);
	}
}

//******************************************************************************
VectorView::triangleIter VectorView::addTriangle(pointIter a, pointIter b, 
																								 pointIter c)
//******************************************************************************
{
	CHECKREF;
	
	if (!surface_->isLocked())
	{
		SurfaceEntity::triangles_it t = surface_->addTriangle(*a.impl_, 
																													*b.impl_, 
																													*c.impl_);
		return selectTriangle(t);
	}
}

//******************************************************************************
void VectorView::deletePoint(pointIter p)
//******************************************************************************
{
	CHECKREF;
	if (!surface_->isLocked() && p.isValid())
	{
		SurfaceEntity::points_it point = *p.impl_;
		deselectPoint(p);
		
		surface_->deletePoint(point);
	}
}

//******************************************************************************
void VectorView::deleteTriangle(triangleIter t)
//******************************************************************************
{
	CHECKREF;
	if (!surface_->isLocked() && t.isValid())
	{
		SurfaceEntity::triangles_it triangle = *t.impl_;
		deselectTriangle(t);

		surface_->deleteTriangle(triangle);
	}	
}

//******************************************************************************
VectorView::pointIter
VectorView::selectionOf(SurfaceEntity::points_it element)
//******************************************************************************
{
	pointmap_cit pm = pointmap_.find(element);
	if (pm != pointmap_.end())
		return pointIter(*this, pm->second);
	else
		return pointIter(*this, pointits_.end());
			
}

//******************************************************************************
VectorView::triangleIter
VectorView::selectionOf(SurfaceEntity::triangles_it element)
//******************************************************************************
{
	trianglemap_cit tm = trianglemap_.find(element);
	if (tm != trianglemap_.end())
		return triangleIter(*this, tm->second);
	else
		return triangleIter(*this, triangleits_.end());
}

//******************************************************************************
RbDblV VectorView::getCenter() const
//******************************************************************************
{
  RbDblV center(4,0);
  for(pointits_cit p=pointits_.begin();p!=pointits_.end();++p)
    center += (*p)->getPosition();
  
  center /= pointits_.size();
  center[W] = 1;

  return center;
}

//******************************************************************************
RbDblV VectorView::getExtensions() const
//******************************************************************************
{
  RbDblV maxpos(4,-DBL_MAX); maxpos[W] = 1;
  RbDblV minpos(4,DBL_MAX); minpos[W] = 1;

  for(pointits_cit p=pointits_.begin();p!=pointits_.end();++p)
  {
    const RbDblV& pos = (*p)->getPosition();
    for(int i=X;i<W;++i)
    {
      if (pos[i]<minpos[i]) minpos[i] = pos[i];
      if (pos[i]>maxpos[i]) maxpos[i] = pos[i];
    }
  }

  return (maxpos - minpos);
}

//******************************************************************************
void VectorView::rebuildPointMap()
//******************************************************************************
{
	pointmap_.clear();

	for(pointits_it p=pointits_.begin();p!=pointits_.end();++p)
		pointmap_[*p] = p;
}

//******************************************************************************
void VectorView::rebuildTriangleMap()
//******************************************************************************
{
	trianglemap_.clear();

	for(triangleits_it t=triangleits_.begin();t!=triangleits_.end();++t)
		trianglemap_[*t] = t;
}
