/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file vectorview.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 16 Mar 2004, last CVS commit 2004/03/16 18:38:29
 */

#ifndef VECTORVIEW_HH
#define VECTORVIEW_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <vector>
#  include <map>
#else
#  include <vector.h>
#  include <map.h>
#endif

#include "surfaceview.hh"
#include "vectorviewiterator.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	/**
	 * \brief a view for random access to the points and triangles
	 *
	 * The vector view provides random access to a selection of points and
	 * triangles on a SurfaceEntity. Points can be selected and deselected by
	 * selectPoint() and deselectPoint() while selectTriangle() and
	 * deselectTriangle() perform the same operations for triangles. The view
	 * provides the possibility to add and delete real points and triangles in
	 * the entity it observes. The methods addPoint(), deletePoint(),
	 * addTriangle() and deleteTriangle() work in the same way as their
	 * SurfaceEntity counterparts except that they deal with VectorViewIt
	 * iterators instead of the iterator types of the entity.
	 *
	 * Access to the selected elements is provided vie pointIter a triangleIter
	 * iterator classes. These are specializations of the VectorViewIt class and
	 * they act as observers on the VectorView while at the same time provide
	 * direct access to the point and triangle interface via the
	 * VectorViewIt::operator->() operation. For the selected points the view
	 * provides sequence access via begin() and end() operations and random
	 * access via the pointAt() method. For triangles the beginTriangles() and
	 * endTriangles() return the first and behind-the-last selected triangle
	 * while triangleAt() provides random access.
	 *
	 * The view also propagates the surface data interface of the SurfaceEntity
	 * class by the same methodnames and parameters.
	 *
	 * When constructing a VectorView it is advisable to reserve() the
	 * appropriate memory for points and triangles to avoid frequent reallocation
	 * in the buildup process.
	 * 
	 * A VectorView contains a map that allows association of plain SurfaceEntity
	 * iterators to selected points and triangles. This allows checked retrieval
	 * of specific entity elements or the transformation of a entity iterator
	 * into a view iterator (selectionOf(SurfaceEntity::points_it element),
	 * selectionOf(SurfaceEntity::triangles_it element).
	 */
	class VectorView : public SurfaceView
	{
		
	private:

		//**************************************************************************
		// PRIVATE TYPE DEFINITIONS
		//**************************************************************************

		//@{
		/**
		 * \brief type for the vector of selected points + iterators
		 */
		typedef vector<SurfaceEntity::points_it> pointits_t;
		typedef pointits_t::iterator pointits_it;
		typedef pointits_t::const_iterator pointits_cit;
		//@}

		//@{
		/**
		 * \brief type for the vector of selected triangles + iterators
		 */
		typedef vector<SurfaceEntity::triangles_it> triangleits_t;
		typedef triangleits_t::iterator triangleits_it;
		typedef triangleits_t::const_iterator triangleits_cit;
		//@}

	public:

		/**
		 * \brief public iterator for the sequence of selected points
		 */
		typedef VectorViewIt<pointits_it, 
												 SurfaceEntity::points_it, 
												 SurfaceEntity::points_cit, Point> pointIter;

		/**
		 * \brief public iterator for the sequence of selected triangles
		 */
		typedef VectorViewIt<triangleits_it, 
												 SurfaceEntity::triangles_it,
												 SurfaceEntity::triangles_cit, Triangle> triangleIter;

		//**************************************************************************
		// CONSTRUCTOR
		//**************************************************************************
		
		/**
		 * \brief default constructor
		 *
		 * Creates a vector surface view, without an observed entity object
		 *
		 * \sa VectorView(SurfaceEntity& const)
		 */
		VectorView() : SurfaceView(), 
									 pointits_(), triangleits_(),
									 pointmap_(), trianglemap_() { }
		
		/**
		 * \brief initializing constructor
		 *
		 * This constructor creates a vector surface view, that observes the entity
		 * \p surface.
		 *
		 * \param surface SurfaceEntity that is observed.
		 *
		 * \sa VectorView()
		 */
		VectorView(SurfaceEntity& surface) : SurfaceView(surface),
																				 pointits_(), triangleits_(),
																				 pointmap_(), trianglemap_() { }

		//**************************************************************************
		// OBSERVER
		//**************************************************************************

		/**
		 * \brief is called by the observed intity, when the surface has changed
		 *
		 * \note The current implementation of update does nothing.
		 */
		void update();

    //****************************************************************************
    // COPYING
    //****************************************************************************

  protected:
    
    virtual void buildSelectionEntity(DeepCopyInformation& copyinfo) const;
    virtual void buildSelectionView(DeepCopyInformation& copyinfo) const;

  public:

		//****************************************************************************
		// VIEW BUILD OPERATIONS
		//****************************************************************************

		/**
		 * \brief adds a point to the vector of selected surface points
		 *
		 * The method adds a new point to the vector of selected surface points. If
		 * the point has not been added it returns the new iterator otherwise the
		 * iterator of old selection. To provide a fast addition of new points it
		 * is advisable to reserve enough space for the vector of selected points
		 * (reserve()). This operation invalidates all attached ViewIterators!
		 *
		 * \param point an iterator in the SurfaceEntity point list.
		 *
		 * \return the pointIter of the new selected point
		 *
		 * \sa selectTriangle(), deselectPoint()
		 */
		pointIter selectPoint(SurfaceEntity::points_it point);
		/**
		 * \brief removes a point from the selected points
		 *
		 * The point, represented by pointIter is removed from the vector of
		 * selected points. This operation invalidates all attached ViewIterators!
		 *
		 * \param pointiter the iterator to the point that should be deselected
		 *
		 * \sa selectPoint(), deselectTriangle()
		 */
		void deselectPoint(pointIter pointiter);

		/**
		 * \brief adds a triangle to the vector of selected surface triangles
		 *
		 * The method adds a new triangle to the vector of selected surface
		 * triangles. If the triangle has not been added it returns the new
		 * iterator (if allowed, see below) otherwise the iterator of old
		 * selection. To provide a fast addition of new triangles it is advisable
		 * to reserve enough space for the vector of selected triangles
		 * (reserve()). This operation invalidates all attached ViewIterators!
		 *
		 * The parameter minimal specifies how many edge-points of the triangle
		 * must already be selected by the VectorView. The default is three, so
		 * that only those triangles are selected that are fully selected on the
		 * point basis. If the selection was not allowed the method returns the
		 * endTriangles() iterator.
		 *
		 * \param triangle an iterator in the SurfaceEntity triangle list.  
		 * \param minimum  the number of triangle edge-points that must already be
		 *                 selected.
		 *
		 * \return the triangleIter of the newly selected, the already selected or
		 *         the endTriangle() triangle
		 *
		 * \sa selectPoint(), deselectTriangle()
		 */
		triangleIter selectTriangle(SurfaceEntity::triangles_it triangle,
																size_t minimum = 3);
		/**
		 * \brief removes a triangle from the selected triangles
		 *
		 * The triangle, represented by triangleIter is removed from the vector of
		 * selected triangles. This operation invalidates all attached ViewIterators!
		 *
		 * \param triangleiter the iterator to the triangle that should be
		 * deselected
		 *
		 * \sa selectTriangle(), deselectPoint()
		 */
		void deselectTriangle(triangleIter triangleiter);

		void reserve(size_t pointcount = 0, size_t trianglecount = 0);

		//**************************************************************************
		// POINT AND TRIANGLE OPERATIONS
		//**************************************************************************

		/**
		 * \brief adds and selects a new point to the SurfaceEntity.
		 *
		 * The method adds a new point to the SurfaceEntity, adds it to the
		 * selection of this view object. This operation invalidates all attached
		 * ViewIterators!
		 *
		 * \return the pointIter that represents the selected new point.
		 *
		 * \sa addTriangle(), deletePoint();
		 */
		pointIter addPoint();
		/**
		 * \brief removes the point \p p from the entity and the view
		 *
		 * This method removes the point from the entity and the view. Be careful
		 * with this operation, because it will remove also triangles attached to
		 * that point, and this triangles will most likely be part of the
		 * view. This operation invalidates all attached ViewIterators!
		 * 
		 *
		 * \param p a point, represented by its pointIter, that should be removed
		 *          from the view and from the entity
		 *
		 * \sa addPoint(), deleteTriangle()
		 */
		void deletePoint(pointIter p);

		/**
		 * \brief adds and selects a new triangle to the SurfaceEntity.
		 *
		 * The method adds a new triangle to the SurfaceEntity, adds it to the
		 * selection of this view object. This operation invalidates all attached
		 * ViewIterators! Only points in this selection can be used for the
		 * formation of the new triangle.
		 *
		 * \param a,b,c pointIters that represent selected points of the entity
		 *              that should form the triangle.
		 *
		 * \return the triangleIter that represents the selected new triangle.
		 *
		 * \sa addPoint(), deleteTriangle();
		 */
		triangleIter addTriangle(pointIter a, pointIter b, pointIter c);
		/**
		 * \brief removes the triangle \p p from the entity and the view
		 *
		 * This method removes the triangle from the entity and the view. This
		 * operation invalidates all attached ViewIterators!
		 * 
		 *
		 * \param t a triangle, represented by its triangleIter, that should be
		 *          removed from the view and from the entity
		 *
		 * \sa addTriangle(), deleteTriangle()
		 */
		void deleteTriangle(triangleIter t);

		//**************************************************************************
		// OTHER PROPERTIES AND CHECKS
		//**************************************************************************

		/**
		 * \brief gives back the number of selected points on the surface
		 *
		 * \return the number of points in that particular surface view. (i.e. the
		 * number of selected points)
		 * 
		 * \sa getNTriangles()
		 */
		size_t getNPoints() const { return pointits_.size(); }
		/**
		 * \brief gives back the number of selected triangles on the surface
		 *
		 * \return the number of selected surface triangles in this view
		 * 
		 * \sa getNPoints()
		 */
		size_t getNTriangles() const { return triangleits_.size(); }

		//@{
		/**
		 * \brief retrieves the selection of a point or triangle or \c end.
		 *
		 * This method takes an SurfaceEntity point or triangle iterator and does a
		 * lookup in the pointmap_ or trianglemap_ if the element has been selected
		 * it returns the corresponding pointIter or triangleIter otherwise it
		 * gives back the end of the corresponding selection sequence.
		 *
		 * \param element the entity iterator of the point or triangle
		 * 
		 * \return a point iterator that points to the seletion of the entity
		 *         element or to the end of the selection sequences if the element 
		 *         is not selected by this view.
		 */
		pointIter selectionOf(SurfaceEntity::points_it element);
		triangleIter selectionOf(SurfaceEntity::triangles_it element);
		//@}

    RbDblV getCenter() const;
    RbDblV getExtensions() const;
	
		//**************************************************************************
		// ITERATORS
		//**************************************************************************

		//@{
		/// \name point iterators
		/**
		 * \brief iterator for the selected point sequence
		 *
		 * The class provides begin and end iterators for the point sequence for
		 * mutable access to the single, selected points in the surface view.
		 */
		pointIter begin() { return pointIter(*this, pointits_.begin()); }
		pointIter end()   { return pointIter(*this, pointits_.end()); }
		//@}

		//@{
		/// \name triangle iterators
		/**
		 * \brief iterator for the selected triangle sequence
		 *
		 * The class provides begin and end iterators for the triangle sequence for
		 * mutable access to the single, selected triangles in the surface view.
		 */
		triangleIter beginTriangles() { 
			return triangleIter(*this,triangleits_.begin()); }
		triangleIter endTriangles()   { 
			return triangleIter(*this,triangleits_.end()); }
		//@}

		/**
		 * \brief gives the pointIter or pointConstIter for a given index \p ix
		 *
		 * The method returns a pointIter for selected point with index \p ix. If
		 * the index is out of bounds the end() iterator is returned.
		 *
		 * \param ix the index of the desired point
		 *
		 * \return a pointIter or pointConstIter to the desired point, or end of
		 *         the sequence if the index was out of bounds.
		 */
		pointIter pointAt(size_t ix);

		/**
		 * \brief gives the triangleIter or triangleConstIter for a given index \p
		 *        ix
		 *
		 * The method returns a triangleIter for selected triangle with index \p
		 * ix. If the index is out of bounds the end() iterator is returned.
		 *
		 * \param ix the index of the desired triangle
		 *
		 * \return a triangleIter or triangleConstIter to the desired triangle, or
		 *         end of the sequence if the index was out of bounds.
		 */
		triangleIter triangleAt(size_t ix);

  protected:
    
    //**************************************************************************
    // PROTECTED METHODS
    //**************************************************************************

    virtual SurfaceView* createView() const { return new VectorView(); }

	private:

		//**************************************************************************
		// INTERNAL PRIVATE TYPES
		//**************************************************************************
		
		//@{
		/**
		 * \brief type for the entity points <-> selected points map.
		 */
		typedef std::map<SurfaceEntity::points_it, pointits_it,
								lt_iterator<SurfaceEntity::points_it> > pointmap_t;
		typedef pointmap_t::iterator pointmap_it;
		typedef pointmap_t::const_iterator pointmap_cit;
		//@}

		//@{
		/**
		 * \brief type for the entity triangles <-> selected triangles map.
		 */
		typedef std::map<SurfaceEntity::triangles_it, triangleits_it,
                     lt_iterator<SurfaceEntity::triangles_it> > trianglemap_t;
		typedef trianglemap_t::iterator trianglemap_it;
		typedef trianglemap_t::const_iterator trianglemap_cit;
		//@}

		//**************************************************************************
		// PRIVATE METHODS
		//**************************************************************************

		//@{
		/**
		 * \brief rebuilds either the point or triangle entity-selection map.
		 *
		 * The methods rebuild either the structure that maps the entity elements
		 * (point or triangles) to the selected elements. This gets necessary
		 * whenever a point or triangle is deselected or deleted.
		 */
		void rebuildPointMap();
		void rebuildTriangleMap();
		//@}
	
		//**************************************************************************
		// PRIVATE MEMBERS
		//**************************************************************************

		/**
		 * \brief the vector of selected points
		 *
		 * This vector holds a sequence of iterators into the point list of the
		 * SurfaceEntity.
		 *
		 * \sa SurfaceEntity::points_
		 */
		pointits_t pointits_;
		/**
		 * \brief the vector of selected triangles
		 *
		 * This vector holds a sequence of iterators into the triangle list of the
		 * SurfaceEntity.
		 *
		 * \sa SurfaceEntity::triangles_
		 */
		triangleits_t triangleits_;

		/**
		 * \brief the map of entity points to selected points
		 */
		pointmap_t pointmap_;

		/**
		 * \brief the map of entity triangles to selected triangles
		 */
		trianglemap_t trianglemap_;
	};

	//**************************************************************************
	inline VectorView::pointIter VectorView::pointAt(size_t ix)
		//**************************************************************************
	{
		if (ix<pointits_.size())
			return pointIter(*this, pointits_.begin() + ix);
	}

	//**************************************************************************
	inline VectorView::triangleIter VectorView::triangleAt(size_t ix)
	//**************************************************************************
	{
		if (ix<triangleits_.size())
			return triangleIter(*this, triangleits_.begin() + ix);
	}

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // VECTORVIEW_HH
