/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file vectorviewiterator.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 28 Mar 2004, last CVS commit 2004/03/28 09:36:07
 */

#ifndef VECTORVIEWITERATOR_HH
#define VECTORVIEWITERATOR_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "viewiterator.hh"

#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 

	class VectorView;
	class PatchView;

	/**
	 * \brief iterator template for the vector views point and triangle iterator
	 *
	 * This iterator template exists for convinience, because the point and
	 * triangle iterator are excactly the same except for the Point and Triangle
	 * type. The class combines the IterTempl, which is a template for indirected
	 * iterators and the ViewIterator base class that handles the oberver stuff.
	 *
	 * Note that the registration of the VectorViewIt as an observer to the
	 * associated view is done automatically in the copy constructor and the
	 * assignment operator.
	 */
	template<class Iter, class EntityIt, class EntityConstIt, class T>
	class VectorViewIt: public IterTempl<Iter, T>, public ViewIterator
	{
	public:

		/**
		 * \brief default constructor
		 * 
		 * \sa VectorViewIt(const VectorViewIt&), 
		 *     VectorViewIt(SurfaceView&, const Iter&)
		 */
		VectorViewIt() : ViewIterator(), IterTempl<Iter,T>() { }

		/**
		 * \brief copy constructor
		 *
		 * The copy constructor creates a new copy of the VectorViewIt \p src and
		 * registers itself as an observer to the corresponding surface view.
		 *
		 * \param src the source VectorViewIt object.
		 */
		VectorViewIt(const VectorViewIt& src) 
				: ViewIterator(src), IterTempl<Iter,T>()	{
			impl_  = src.impl_;
			valid_ = src.valid_;
		}

		/** 
		 * \brief assignment operator
		 *
		 * The assignment operator is probably the most frequent way of
		 * initialization. It duplicates a copy of \p source to itself and attaches
		 * itself to the new view as an observer. If the iterator has been attached
		 * to another view this connection is removed.
		 *
		 * \param src the source VectorViewIt
		 * 
		 * \return a reference to itself after the \p source has been duplicated
		 */
		VectorViewIt<Iter, EntityIt, EntityConstIt, T>& 
		operator=(const VectorViewIt& src) {
			if(view_) view_->detach(*this);
			src.view_->attach(*this);
			
			impl_  = src.impl_;
			valid_ = src.valid_;

			return *this;
		}
		
		//@{
   /** 
		* \brief conversion operator for constant and non constant entity iterators
		* 
		* \return an entity iterator based on the current position of itself in the
		*         vector view selection
		*/
		operator EntityIt ()      const { return *impl_;	}
		operator EntityConstIt () const { return *impl_;	}
    //@}

		void update() { ViewIterator::update(); valid_ = false; }
		
	private:
		
		friend class VectorView;
		friend class PatchView;
		
		/** 
		 * \brief the internal constructor
		 *
		 * The internal constructor is used by the surface view to generate a valid
		 * view iterator based on a position (\p impl) in its selection sequence.
		 * 
		 * \param view a reference to the corresponding surface view
		 * \param impl the iterator in the vector view selection sequence
		 *
		 * \sa VectorViewIt(), VectorViewIt(const VectorViewIt&)
		 */
		explicit VectorViewIt(SurfaceView& view, const Iter& impl)
				: ViewIterator(view), IterTempl<Iter,T>()	{
			impl_  = impl;
			valid_ = true;
		}
	};

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // VECTORVIEWITERATOR_HH
