/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file viewiterator.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 17 Mar 2004, last CVS commit 2004/03/17 16:13:55
 */

#ifndef VIEWITERATOR_HH
#define VIEWITERATOR_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include "common.hh"

//#include <iostream>

#define CHECKITER \
  if (!valid_) throw InvalidIter_e("view iterator is invalid!");


#ifdef HAVE_NAMESPACES
namespace Surface
{
#endif 
	
	class SurfaceView;
	
	class ViewIterator
	{
	public:
		virtual void update();
		
		ViewIterator() : view_(0) { }
		ViewIterator(SurfaceView& view); 
		ViewIterator(const ViewIterator& src);
		
		~ViewIterator();
		
		SurfaceView& getSurfaceView() const;
		bool setSurfaceView(SurfaceView& view);
	protected:
		
		SurfaceView* view_;
	};

  //****************************************************************************
	inline SurfaceView& ViewIterator::getSurfaceView() const
  //****************************************************************************
	{
		if (!view_)
			throw InvalidReference_e("view_ points to NULL!");
		
		return *view_;
	}

  //****************************************************************************
	inline bool ViewIterator::setSurfaceView(SurfaceView& view)
  //****************************************************************************
	{
		if (!view_)
		{
			view_ = &view;
			return true;
		}
		else
			return false;
	}

	template<class ContIter, class T>
	class IterTempl 
	{
	public:

		//**************************************************************************
		// CONSTRUCTOR
		//**************************************************************************

		IterTempl() : impl_(), valid_(false) { }

		IterTempl(const ContIter& impl) 
				: impl_(impl), valid_(true) { }

		//**************************************************************************
		// OPERATOR
		//**************************************************************************

		/**
		 * \brief prefix increment
		 *
		 * increments the iterator to the next position and returns a reference.
		 *
		 * \return a reference to the \c this at the new position
		 */
		IterTempl<ContIter, T>& operator++() {
			CHECKITER;
			++impl_; 
			return *this; 
		}
		/**
		 * \brief postfix increment
		 *
		 * returns a copy of \c this and increments the iterator to the next
		 * position.
		 *
		 * \return a copy of the actual iterator
		 */
		IterTempl<ContIter, T> operator++(int) {
			CHECKITER;
			IterTempl<ContIter, T> tmp = *this;
			++impl_; 
			return tmp; 
		}

		/**
		 * \brief prefix decrement
		 *
		 * decrements the iterator to the previous position and returns a
		 * reference.
		 *
		 * \return a reference to the \c this at the new position
		 */
		IterTempl<ContIter, T>& operator--() {
			CHECKITER;
			--impl_; 
			return *this; 
		}
		/**
		 * \brief postfix decrement
		 *
		 * returns a copy of \c this and decrements the iterator to the previous
		 * position.
		 *
		 * \return a copy of the actual iterator
		 */
		IterTempl<ContIter, T> operator--(int) {
			CHECKITER;
			IterTempl<ContIter, T> tmp = *this;
			--impl_; 
			return tmp; 
		}

		/**
		 * \brief arrow operator (access to the real element by arrow)
		 *
		 * The arrow operator provides access to the real element of the type T*.
		 *
		 * \return a pointer to the real element
		 */
		T* operator->() const { CHECKITER; return impl_->operator->(); }

		/**
		 * \brief dereferencing operator (access to the real element by ref.)
		 *
		 * The dereferencing operator provides access to the real element of the
		 * type T&.
		 *
		 * \return a reference to the real element
		 */
		T& operator*() const { CHECKITER; return *(*impl_); }

		/**
		 * \brief checks for inequality between to iterators
		 *
		 * \param test then other iterator
		 *
		 * \return \c true if the iterators are not equal.
		 */
		bool operator!=(const IterTempl<ContIter,T>& test) const {
			CHECKITER;
			return impl_ != test.impl_;
		}

		/**
		 * \brief checks for equality between to iterators
		 *
		 * \param test then other iterator
		 *
		 * \return \c true if the iterators are equal.
		 */
		bool operator==(const IterTempl<ContIter,T>& test) const {
			CHECKITER;
			return impl_ == test.impl_;
		}



		bool isValid() const { return valid_; }
		
	protected:
				
		ContIter impl_;
		bool     valid_;
	};

#ifdef HAVE_NAMESPACES
};
#endif 

#endif // VIEWITERATOR_HH
