/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "graphview_test.hh"

#include <float.h>
#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include "templates.hh"
#include "utils.hh"

#include <graphview.hh>
#include <vectorview.hh>
#include <simplebuilder.hh>
#include <spherebuilder.hh>

#ifdef HAVE_STD
#  include <iostream>
#else
#  include <iostream.h>
#endif

#ifdef HAVE_NAMESPACES
using namespace boost::unit_test_framework;
using namespace Surface;
using namespace std;
#endif 

//******************************************************************************
graphview_test::graphview_test()
//******************************************************************************
: surface(3.0, 1.4, 0), view(0)
{
	makeDodekaeder(surface);
}

//******************************************************************************
graphview_test::~graphview_test()
//******************************************************************************
{
	if (view) delete view;
}

//******************************************************************************
void graphview_test::test_simplebuilder()
//******************************************************************************
{
  BOOST_MESSAGE("\n" << DELIMITER << "\n GraphView simplebuilder test\n");

	int i,j;
	GraphView::vertexIter p;
	SurfaceEntity::triangles_it t;
				
	BOOST_MESSAGE(" !! create a new graph view");
	SimpleBuilder<GraphView> builder;
	view = dynamic_cast<GraphView*>(surface.createView(builder));

	BOOST_MESSAGE(" ** test the number of vertices and edges");
	BOOST_CHECK(view->getNPoints() == surface.getNPoints());
	BOOST_CHECK(view->getNEdges()  == dodek_edges);

	BOOST_MESSAGE(" ** test for the correct return of point density, sphere count and "
			 << "probe radius");
	BOOST_CHECK_CLOSE(view->getProbeRadius(),surface.getProbeRadius(), 
										DBL_EPSILON);
	BOOST_CHECK_CLOSE(view->getPointDensity(),surface.getPointDensity(), 
										DBL_EPSILON);
	BOOST_CHECK(view->getSphereCount() == surface.getSphereCount());
				
	BOOST_MESSAGE(" ** check positions from begin to end");
	for(p=view->begin();p!=view->end();++p)
		for(int j=0;j<3;++j)
		{
			int master = p->getMasterIndex();
			BOOST_CHECK_CLOSE((*p)[j], dodek_points[master][j], DBL_EPSILON);
		}

	BOOST_MESSAGE(" ** check all the connections");

	for(t=surface.beginTriangles();t!=surface.endTriangles();++t)
	{
		SurfaceEntity::points_it a = t->getVertex(A);
		SurfaceEntity::points_it b = t->getVertex(B);
		SurfaceEntity::points_it c = t->getVertex(C);

		GraphView::vertexIter va = view->selectionOf(a);
		GraphView::vertexIter vb = view->selectionOf(b);
		GraphView::vertexIter vc = view->selectionOf(c);

		BOOST_CHECK(view->connected(a,b));
		BOOST_CHECK(view->connected(b,c));
		BOOST_CHECK(view->connected(c,a));
		BOOST_CHECK(view->connected(vb,va));
		BOOST_CHECK(view->connected(vc,vb));
		BOOST_CHECK(view->connected(va,vc));
	}
}

//******************************************************************************
void graphview_test::test_deepcopy()
//******************************************************************************
{
  BOOST_MESSAGE("\n" << DELIMITER << "\n GraphView deep copy test\n");

  BOOST_MESSAGE(" !! create a subselection");
  SphereBuilder<GraphView> subselector(3.0, view->begin()->getPosition());
  GraphView* subselection = 
    dynamic_cast<GraphView*>(surface.createView(subselector));

  BOOST_MESSAGE(" !! make a deep copy of the subselection view \n");
  GraphView* deepcopy = dynamic_cast<GraphView*>(subselection->clone());

  BOOST_MESSAGE(" ** check if the two views are not identical\n");
  BOOST_CHECK(deepcopy != view);
  
  BOOST_MESSAGE(" ** check if the two entities are not identical\n");
  BOOST_CHECK(&deepcopy->getSurfaceEntity() != &view->getSurfaceEntity());

  compare_views(*subselection, *deepcopy);

  delete deepcopy;
  delete subselection;

}

//******************************************************************************
void graphview_test::test_destruction()
//******************************************************************************
{
  BOOST_MESSAGE("\n" << DELIMITER << "\n testing correct GraphView destruction\n");

  SimpleBuilder<VectorView> vectorbuilder;
  SimpleBuilder<GraphView>  graphbuilder;

  BOOST_MESSAGE(" !! create and clone a vector view");
  VectorView* vview = 
    dynamic_cast<VectorView*>(surface.createView(vectorbuilder));
  VectorView* view = dynamic_cast<VectorView*>(vview->clone());

  delete vview;

  BOOST_MESSAGE(" !! create a graph view from the clone");
  GraphView* graph = 
    dynamic_cast<GraphView*>(view->getSurfaceEntity().createView(graphbuilder));

  BOOST_MESSAGE(" !! delete the graph view");
  delete graph;
  
  BOOST_MESSAGE(" ** check if the clone deletes without an exception");
  delete view;
}

//******************************************************************************
void graphview_test::test_weights()
//******************************************************************************
{
  BOOST_MESSAGE("\n" << DELIMITER << "\n GraphView edge weight test\n");

	SurfaceEntity::triangles_it t;
		
	BOOST_MESSAGE(" ** compare all edge weights with the corresponding "
	<< "euclidean distance");
	for(t=surface.beginTriangles();t!=surface.endTriangles();++t)
	{
		for(int j=0;j<3;++j)
		{
			int a = j;
			int b = (a==2) ? 0 : a + 1;
				
			//get the points and vertices
			SurfaceEntity::points_it pa = t->getVertex(a);
			SurfaceEntity::points_it pb = t->getVertex(b);
			GraphView::vertexIter va = view->selectionOf(pa);
			GraphView::vertexIter vb = view->selectionOf(pb);
								
			double gdist = view->getDistance(va, vb);
			double eucdist = (pa->getPosition() - pb->getPosition()).vec_len();

			BOOST_CHECK_CLOSE(gdist, eucdist, DBL_EPSILON);
		}
	}
}

//******************************************************************************
void graphview_test::test_onsurfacedistance()
//******************************************************************************
{
	SurfaceEntity::points_it center = surface.begin();
	view->compOnSurfaceDistances(center);

	for(GraphView::vertexIter v = view->begin(); v!= view->end();++v)
		BOOST_CHECK_CLOSE(view->getDistance(v), 
											dodek_onsurfacedists[v->getMasterIndex()], 0.001);
}
			
//******************************************************************************
graphview_test_suite::graphview_test_suite()
//******************************************************************************
: test_suite("graphview_test_suite") 
{
	// add member function test cases to a test suite
	boost::shared_ptr<graphview_test> instance(new graphview_test);
		
	test_case* simplebuilder_test_case = 
		BOOST_CLASS_TEST_CASE(&graphview_test::test_simplebuilder,	instance); 
	test_case* deepcopy_test_case = 
		BOOST_CLASS_TEST_CASE(&graphview_test::test_deepcopy,	instance); 
  test_case* destruction_test_case =
		BOOST_CLASS_TEST_CASE(&graphview_test::test_destruction,	instance); 
	test_case* weights_test_case = 
		BOOST_CLASS_TEST_CASE(&graphview_test::test_weights,	instance); 
	test_case* onsurfacedistance_test_case = 
		BOOST_CLASS_TEST_CASE(&graphview_test::test_onsurfacedistance,	instance); 
		
	//test case, number of tests that should fail
	add(simplebuilder_test_case, 0);
	add(deepcopy_test_case, 0);
	add(destruction_test_case, 0);
	add(weights_test_case, 0);
	add(onsurfacedistance_test_case, 0);
}
