/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "patchview_test.hh"

#include <float.h>
#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include "templates.hh"
#include "utils.hh"

#include <surface.hh>
#include <simplebuilder.hh>
#include <vectorview.hh>
#include <patchview.hh>
#include <patchbuilder.hh>
#include <offio.hh>

#ifdef HAVE_STD
#  include <iostream>
#else
#  include <iostream.h>
#endif

#ifdef HAVE_NAMESPACES
using namespace boost::unit_test_framework;
using namespace Surface;
using namespace std;
#endif 

int BORDER_1THL_0196[33] = 
{ 
	235, 171, 170, 173, 183, 182, 244, 249, 253, 261, 335,
	341, 406, 405, 471, 617, 544, 543, 539, 535, 534, 536,
	531, 526, 521, 457, 456, 453, 389, 386, 385, 320, 319
};

//******************************************************************************
patchview_test::patchview_test()
//******************************************************************************
: surface(3.0, 1.4, 0), patch(0), view(0)
{
	//Surface::double_limit = 0.001;

	OffIO offio;
	surface = offio.read("1THL.off");
	
	if (!surface)
		BOOST_ERROR(" EE Could not open 1THL.off\n");

	SimpleBuilder<VectorView> builder;
	view = dynamic_cast<VectorView*>(surface->createView(builder));
}

//******************************************************************************
patchview_test::~patchview_test()
//******************************************************************************
{
	if (patch) delete patch;
	if (view)  delete view;
	if (surface) delete surface;
}

//******************************************************************************
void patchview_test::test_patchbuilder()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n PatchView simplebuilder test \n");

	BOOST_MESSAGE(" !! generate a patch over point 196");
	PatchBuilder patchbuilder(view->pointAt(196), 4);
	patch = dynamic_cast<PatchView*>(surface->createView(patchbuilder));

	BOOST_MESSAGE(" ** check if the patchradius is stored with 4.0");
	BOOST_CHECK_CLOSE(patch->getRadius(), 4.0, DBL_EPSILON);
	BOOST_MESSAGE(" ** check if the radius margin is stored with 0.05");
	BOOST_CHECK_CLOSE(patch->getMargin(), 0.05, DBL_EPSILON);
	BOOST_MESSAGE(" ** check if the center is stored with point 196");
	BOOST_CHECK(patch->getCentralPoint()==
              (SurfaceEntity::points_it) view->pointAt(196));
	
	//stream test
	SurfaceEntity* streampatch;
	SurfaceEntity* reference;
	stringstream stream;
	OffIO offio;

	reference = offio.read("1THL_196.off");
	
	BOOST_MESSAGE(" !! write the patch into a memory stream");
	offio.write(stream, *patch);

	BOOST_MESSAGE(" !! read the patch back from memory stream");
	stream.seekp(0,ios_base::beg);
	streampatch = offio.read(stream);

	BOOST_MESSAGE(" !! read the reference surface (1THL_196.off)");
	reference = offio.read("1THL_196.off");


	check_surfaces(*streampatch, *reference);
	
	delete streampatch;
	delete reference;
}

//******************************************************************************
void patchview_test::test_border()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n PatchView simplebuilder test \n");

	BOOST_MESSAGE(" ** compare the border of the patch with the reference border");
	int* ref = BORDER_1THL_0196;
	PatchView::borderIter b;
	for (b=patch->beginBorder();b!=patch->endBorder();++b, ++ref)
		BOOST_CHECK( b->getMasterIndex() == *ref);
}

			
//******************************************************************************
void patchview_test::test_savepatch()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n save PatchView\n");

	OffIO offio;
	offio.write("testpatch.off", *patch);
}

//******************************************************************************
void patchview_test::test_deepcopy()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << 
								"\n VectorView deep copy test \n");

  BOOST_MESSAGE(" !! make a deep copy of the patch \n");
  PatchView* deepcopy = dynamic_cast<PatchView*>(patch->clone());
  compare_views(*patch, *deepcopy);

  BOOST_MESSAGE(" ** check if the two views are not identical\n");
  BOOST_CHECK(deepcopy != view);
  
  BOOST_MESSAGE(" ** check if the two entities are not identical\n");
  BOOST_CHECK(&deepcopy->getSurfaceEntity() != &view->getSurfaceEntity());

	BOOST_MESSAGE(" ** check if the patchradius is stored with 4.0");
	BOOST_CHECK_CLOSE(deepcopy->getRadius(), 4.0, DBL_EPSILON);
	BOOST_MESSAGE(" ** check if the radius margin is stored with 0.05");
	BOOST_CHECK_CLOSE(deepcopy->getMargin(), 0.05, DBL_EPSILON);
	BOOST_MESSAGE(" ** check if the center is stored with point 196");
	BOOST_CHECK(deepcopy->getCentralPoint()->getMasterIndex()==196);
	
  
  delete deepcopy;
}

//******************************************************************************
patchview_test_suite::patchview_test_suite() 
//******************************************************************************
: test_suite("patchview_test_suite")
{
	// add member function test cases to a test suite
	boost::shared_ptr<patchview_test> instance(new patchview_test);
		
	test_case* patchbuilder_test_case = 
		BOOST_CLASS_TEST_CASE(&patchview_test::test_patchbuilder,	instance); 
	test_case* border_test_case = 
		BOOST_CLASS_TEST_CASE(&patchview_test::test_border,	instance); 
	test_case* deepcopy_test_case = 
		BOOST_CLASS_TEST_CASE(&patchview_test::test_deepcopy,	instance); 
	test_case* savepatch_test_case = 
		BOOST_CLASS_TEST_CASE(&patchview_test::test_savepatch,	instance); 

	//test case, number of tests that should fail
	add(patchbuilder_test_case, 0);
	add(border_test_case, 0);
	add(deepcopy_test_case, 0);
	//add(savepatch_test_case, 0);
}
