/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "sld_test.hh"

#include <float.h>
#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include "templates.hh"
#include "utils.hh"

#include <point.hh>
#include <triangle.hh>
#include <simplebuilder.hh>
#include <vectorview.hh>

#ifdef HAVE_STD
#  include <iostream>
#else
#  include <iostream.h>
#endif

#ifdef HAVE_NAMESPACES
using namespace boost::unit_test_framework;
using namespace Surface;
using namespace std;
#endif 

//******************************************************************************
sld_test::sld_test()
//******************************************************************************
: surface(0), dodecaeder(), sldio()
{ 
	makeDodekaeder(dodecaeder, ftUnknown);			 
	Surface::double_limit = 0.001;
}

//******************************************************************************
sld_test::~sld_test()
//******************************************************************************
{ 
	if(surface) delete surface; 
}

		 
													 
//******************************************************************************
void sld_test::test_input()
//******************************************************************************
{
	BOOST_MESSAGE(DELIMITER << "\n" << " SLD input test " << "\n");
	
	BOOST_MESSAGE("!! read the sld file dodecaeder.sld");
	surface = sldio.read("dodecaeder.sld");
	
	surface->setPointDensity(1.0);
	surface->setProbeRadius(1.5);
	
	check_surfaces(*surface, dodecaeder);
}
		 
//******************************************************************************
void sld_test::test_entity_output()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " SLD output test for surface entity " << "\n");
	//stream test
	stringstream stream;
	
	BOOST_MESSAGE(" !! write the surface into a stream");
	sldio.write(stream, *surface);
	
	BOOST_MESSAGE(" !! read the surface back from the stream");
	stream.seekp(0,ios_base::beg);
	auto_ptr<SurfaceEntity> querystream(sldio.read(stream));
	
	querystream->setPointDensity(1.0);
	querystream->setProbeRadius(1.5);
	check_surfaces(*querystream, dodecaeder);
	
	//file test
	BOOST_MESSAGE("\n"
								<< " !! write the surface into a 'test.sld'");
	sldio.write("test.sld", *surface);
	
	BOOST_MESSAGE(" !! read the surface back from 'test.sld'");
	auto_ptr<SurfaceEntity> queryfile(sldio.read("test.sld"));
	
	queryfile->setPointDensity(1.0);
	queryfile->setProbeRadius(1.5);
	check_surfaces(*queryfile, dodecaeder);

}

//******************************************************************************
void sld_test::test_view_output()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " SLD output test for vector view " << "\n");

	BOOST_MESSAGE(" !! generate a vector view of the total surface");
	SimpleBuilder<VectorView> builder;
	auto_ptr<VectorView>
		view(dynamic_cast<VectorView*>(surface->createView(builder)));

	//stream test
	stringstream stream;

	BOOST_MESSAGE(" !! write the view into a stream");
	sldio.write(stream, *view);
			 
	BOOST_MESSAGE(" !! read the surface back from the stream");
	stream.seekp(0,ios_base::beg);
	auto_ptr<SurfaceEntity> querystream(sldio.read(stream));

	querystream->setPointDensity(1.0);
	querystream->setProbeRadius(1.5);
	check_surfaces(*querystream, dodecaeder);

	//file test
	BOOST_MESSAGE("\n"
								<< " !! write the view into a 'test.sld'");
	sldio.write("test.sld", *view);
			 
	BOOST_MESSAGE(" !! read the surface back from 'test.sld'");
	auto_ptr<SurfaceEntity> queryfile(sldio.read("test.sld"));

	queryfile->setPointDensity(1.0);
	queryfile->setProbeRadius(1.5);
	check_surfaces(*queryfile, dodecaeder);

}

//******************************************************************************
void sld_test::test_wrongformat()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n SLD wrong format test\n");
	try
	{
		BOOST_MESSAGE(" !! try to read dodecaeder.msms as SLD file");
		auto_ptr<SurfaceEntity> queryfile(sldio.read("dodecaeder.msms"));

		BOOST_ERROR(" EE did not throw an error!");
	}
	catch (WrongFormat_e err)
	{
		BOOST_MESSAGE(" ** exception thrown: \n    " << err.getMessage() 
                  << " -> passed");
	}
}
	 
//******************************************************************************
sld_test_suite::sld_test_suite()
//******************************************************************************
: test_suite("surface_test_suite") 
{
	// add member function test cases to a test suite
	boost::shared_ptr<sld_test> instance(new sld_test);
			 
	test_case* input_test_case = BOOST_CLASS_TEST_CASE(&sld_test::test_input, 
																										 instance); 
  test_case* entity_output_test_case = 
		BOOST_CLASS_TEST_CASE(&sld_test::test_entity_output, instance); 
	test_case* view_output_test_case = 
		BOOST_CLASS_TEST_CASE(&sld_test::test_view_output, instance); 
	test_case* wrongformat_test_case = 
		BOOST_CLASS_TEST_CASE(&sld_test::test_wrongformat, instance); 
	
	//test case, number of tests that should fail
	add(input_test_case, 0);
	add(entity_output_test_case, 0);
	add(view_output_test_case, 0);
	add(wrongformat_test_case, 0);
}
