/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "surface_test.hh"

#include <float.h>
#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include "templates.hh"
#include "utils.hh"

#include <point.hh>
#include <triangle.hh>

#ifdef HAVE_STD
#  include <iostream>
#else
#  include <iostream.h>
#endif

#ifdef HAVE_NAMESPACES
using namespace boost::unit_test_framework;
using namespace Surface;
using namespace std;
#endif 

//******************************************************************************
surface_test::surface_test()
//******************************************************************************
: pa(3,0), pb(3,0), pc(3,0), na(3,0), nb(3,0), nc(3,0),
	pointdata(3,0),
	surface(3.0, 1.4, 20)
{
	pa[0] = 1; pb[0] = 0; pc[0] = 0;
	pa[1] = 0; pb[1] = 1; pc[1] = 0;
	pa[2] = 0; pb[2] = 0; pc[2] = 1;
			 
	na[0] = 1; nb[0] = 1; nc[0] = 1;
	na[1] = 1; nb[1] = 1; nc[1] = 1;
	na[2] = 1; nb[2] = 1; nc[2] = 1;
			 
	pointdata[0] = 0.1;
	pointdata[1] = 0.2;
	pointdata[2] = 0.3;
}
		 
//******************************************************************************
void surface_test::test_init()
//******************************************************************************
{
	BOOST_MESSAGE( "\n" << DELIMITER << "\n SurfaceEntity initialization test\n");

	SurfaceEntity::points_cit p,q;
			 
	BOOST_MESSAGE(" ** check if the density, probe radius and sphere count is assigned "
			 << "correct" );
	BOOST_CHECK_CLOSE(surface.getPointDensity(), 3.0, DBL_EPSILON);
	BOOST_CHECK_CLOSE(surface.getProbeRadius(), 1.4, DBL_EPSILON);
	BOOST_CHECK(surface.getSphereCount()== 20);
			 
	SurfaceEntity::points_it a = surface.addPoint();
	a->setPosition(pa);
	a->setNormal(na);
			 
	BOOST_MESSAGE(" ** check if the returned iterator is equal to the beginning of the "
			 << "triangle sequence" );
	BOOST_CHECK(surface.begin()==a);
			 
	BOOST_MESSAGE(" ** check if the coordinates are correct" );
	BOOST_CHECK_CLOSE((*a)[0], pa[0], DBL_EPSILON);
	BOOST_CHECK_CLOSE((*a)[1], pa[1], DBL_EPSILON);
	BOOST_CHECK_CLOSE((*a)[2], pa[2], DBL_EPSILON);
			 
	BOOST_MESSAGE(" ** check if the normals are correct" );
	BOOST_CHECK_CLOSE(a->getNormal()[0], na[0], DBL_EPSILON);
	BOOST_CHECK_CLOSE(a->getNormal()[1], na[1], DBL_EPSILON);
	BOOST_CHECK_CLOSE(a->getNormal()[2], na[2], DBL_EPSILON);
			 
			 
	SurfaceEntity::points_it b = surface.addPoint();
	b->setPosition(pb);
	b->setNormal(nb);
			 
	SurfaceEntity::points_it c = surface.addPoint();
	c->setPosition(pc);
	c->setNormal(nc);
			 
	SurfaceEntity::triangles_it t = surface.addTriangle(a, b, c);
	BOOST_MESSAGE(" ** check if the returned iterator is equal to the beginning of the "
			 << "triangle sequence" );
	BOOST_CHECK(surface.beginTriangles()==t);
			 
	BOOST_MESSAGE(" ** check if the points have been created" );
	BOOST_CHECK(surface.getNPoints()== 3);
	BOOST_MESSAGE(" ** check if the triangle has been created" );
	BOOST_CHECK(surface.getNTriangles()== 1);
			 
	BOOST_MESSAGE(" ** check if the three edgepoints of the triangle are the three "
			 << "surface points" );
	p=surface.begin();
	BOOST_CHECK(*p++ == *(*t)[0]);
	BOOST_CHECK(*p++ == *(*t)[1]);
	BOOST_CHECK(*p++ == *(*t)[2]);
}
		 
//******************************************************************************
void surface_test::test_adddelete()
//******************************************************************************
{
	BOOST_MESSAGE( "\n" << DELIMITER << "\n SurfaceEntity add-delete test\n");

	int i,j;
	int size;
	SurfaceEntity::points_it p, pnew;
	SurfaceEntity::triangles_it t;

	RbDblV column;

	size = surface.getNPoints();

	BOOST_MESSAGE(" !! add a new point" );
	pnew = surface.addPoint();
	pnew->setMasterIndex(100);
			

	BOOST_MESSAGE(" ** check if the surface has now " << size + 1
			 << " points" );
	BOOST_CHECK(surface.getNPoints() == size + 1);

	BOOST_MESSAGE(" ** check if the new point is accessible" );
	BOOST_CHECK(pnew->getMasterIndex() == 100 );

	BOOST_MESSAGE(" ** check if a point data column has now " << size + 1 
			 << " elements." );
	column = surface.extractDataColumn(0);
	BOOST_CHECK(column.dim() == size + 1 );

	BOOST_MESSAGE(" !! delete the new point" );
	surface.deletePoint(pnew);
				
	BOOST_MESSAGE(" ** check if the surface has again " << size 
			 << " points" );
	BOOST_CHECK(surface.getNPoints() == size);

	BOOST_MESSAGE(" ** check if a point data column has again " << size 
			 << " elements." );
	column = surface.extractDataColumn(0);
	BOOST_CHECK(column.dim() == size );

}

//******************************************************************************
void surface_test::test_data()
//******************************************************************************
{
	BOOST_MESSAGE( "\n" << DELIMITER << "\n SurfaceEntity point data test\n");

	SurfaceEntity::points_it p;
	int i=0;

	BOOST_MESSAGE(" !! add a column \"test\"" );
	size_t col = surface.addColumn("test", -1);

	BOOST_MESSAGE(" ** check for the right number of columns" );
	BOOST_CHECK(surface.getNColumns() == 1);
	BOOST_MESSAGE(" ** check for the right column name" );
	BOOST_CHECK(surface.getColumnName(col) == "test");
			 
	BOOST_MESSAGE(" ** check the size of each point data vector assign point data"
								<< "values\n"
								<< " ** check if the data was set correct." << "\n"
								<< " !! assign the pointdata[index] value for each point." );
	i = 0;
	for(p=surface.begin();p!=surface.end();++p, ++i)
	{
		BOOST_CHECK(p->data().dim() == 1);
		p->setData(col, pointdata[i]);
		BOOST_CHECK_CLOSE(p->data()[col], pointdata[i], DBL_EPSILON);
	}

	BOOST_MESSAGE(" ** check the size of extracted column and compare it with the " 
			 << "original data" );
	RbDblV testcol = surface.extractDataColumn(col);
	BOOST_CHECK(testcol.dim() == surface.getNPoints());
	for(i=0;i<testcol.dim();++i)
		BOOST_CHECK_CLOSE(testcol[i], pointdata[i], DBL_EPSILON);

	BOOST_MESSAGE(" !! add another column name = check, with pointdata as default"
			 );
	col = surface.addColumn("check", pointdata);
	BOOST_MESSAGE(" ** check for the right number of columns" );
	BOOST_CHECK(surface.getNColumns() == 2);
	BOOST_MESSAGE(" ** check for the right column name" );
	BOOST_CHECK(surface.getColumnName(col) == "check");
			 
	BOOST_MESSAGE(" ** check the size of each point data vector + check if the data "
			 << "was set correct." );
	//increment the value for later check
	i = 0;
	for(p=surface.begin();p!=surface.end();++p, ++i)
	{
		BOOST_CHECK(p->data().dim() == 2);
		BOOST_CHECK_CLOSE(p->data()[col], pointdata[i], DBL_EPSILON);
		p->setData(col, p->data()[col] + 1);
	}

	BOOST_MESSAGE(" ** check the extraction of the data matrix and compare each column " 
			 << "\n ** with the original point data (+1 in the second)" );
	RbDblM testmat = surface.extractDataMatrix();
	BOOST_CHECK(testmat.rno() == surface.getNPoints());
	BOOST_CHECK(testmat.cno() == 2);
	for(i=0;i<testmat.rno();++i)
	{
		BOOST_CHECK_CLOSE(testmat(i,0), pointdata[i]    , DBL_EPSILON);
		BOOST_CHECK_CLOSE(testmat(i,1), pointdata[i] +1 , DBL_EPSILON);
	}

	BOOST_MESSAGE(" !! delete the first column." );
	surface.deleteColumn(0);
	BOOST_MESSAGE(" ** check if one column is left" );
	BOOST_CHECK(surface.getNColumns() == 1);
			 
	i = 0;
	for(p=surface.begin();p!=surface.end();++p, ++i)
	{
		BOOST_MESSAGE(" ** check if the row vectors have only one column" );
		BOOST_CHECK(p->data().dim() == 1);
		BOOST_MESSAGE(" ** check if the column left is the second column from before" 
				 );
		BOOST_CHECK_CLOSE(p->data()[0], pointdata[i] + 1, DBL_EPSILON);
	}
			 
}

//******************************************************************************
void surface_test::test_copy()
//******************************************************************************
{
	BOOST_MESSAGE( "\n" << DELIMITER << "\n SurfaceEntity copy test\n");

	SurfaceEntity::points_cit p,q;

	BOOST_MESSAGE(" !! copy the surface." );
	SurfaceEntity copy(surface);
			 
	BOOST_MESSAGE(" ** check if the density, probe radius and sphere count is copied "
			 << "correct" );
	BOOST_CHECK_CLOSE(copy.getPointDensity(), 3.0, DBL_EPSILON);
	BOOST_CHECK_CLOSE(copy.getProbeRadius(), 1.4, DBL_EPSILON);
	BOOST_CHECK(copy.getSphereCount()== 20);
			 
	BOOST_MESSAGE(" ** check if the points are copied" );
	BOOST_CHECK(copy.getNPoints()== 3);
	BOOST_MESSAGE(" ** check if the triangles have been copied" );
	BOOST_CHECK(copy.getNTriangles()== 1);
			 
	BOOST_MESSAGE(" ** check if all points of the surface are equal" );
	p=surface.begin();
	for(q=copy.begin();q!=copy.end();++p, ++q)
	{
		BOOST_CHECK( p != q);
		BOOST_CHECK(*p == *q);
	}
			 
	BOOST_MESSAGE(" ** check if the three edgepoints of the new triangle are the three "
			 << "surface points" );
	Triangle& tref = *(copy.beginTriangles());
	q=copy.begin();
	BOOST_CHECK(q++ == tref[0]);
	BOOST_CHECK(q++ == tref[1]);
	BOOST_CHECK(q++ == tref[2]);
}

//******************************************************************************
void surface_test::test_assign()
//******************************************************************************
{
	BOOST_MESSAGE( "\n" << DELIMITER << "\n SurfaceEntity assignment test\n");
  SurfaceEntity assigned = surface;
  check_surfaces(assigned, surface);
}
	 
//******************************************************************************
surface_test_suite::surface_test_suite()
//******************************************************************************
: test_suite("surface_test_suite") 
{
	// add member function test cases to a test suite
	boost::shared_ptr<surface_test> instance(new surface_test);
		 
	test_case* init_test_case = BOOST_CLASS_TEST_CASE(&surface_test::test_init, 
																										instance); 
  test_case* data_test_case = BOOST_CLASS_TEST_CASE(&surface_test::test_data, 
																									  instance);
	test_case* adddelete_test_case = 
		BOOST_CLASS_TEST_CASE(&surface_test::test_adddelete, instance);
	test_case* copy_test_case = BOOST_CLASS_TEST_CASE(&surface_test::test_copy, 
																	instance);
	test_case* assign_test_case = 
    BOOST_CLASS_TEST_CASE(&surface_test::test_assign,	instance);

  //test case, number of tests that should fail
  add(init_test_case, 0);
	add(data_test_case, 0);
	add(adddelete_test_case, 0);
	add(copy_test_case, 0);
	add(assign_test_case, 0);
}
	  
