/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "utils.hh"

#include <float.h>
#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <surface.hh>
#include <vectorview.hh>
#include <patchview.hh>
#include <graphview.hh>

#ifdef HAVE_NAMESPACES
using namespace std;

using boost::unit_test_framework::test_suite;
using boost::unit_test_framework::test_case;
using namespace Surface;
#endif


//****************************************************************************
void check_surfaces(const SurfaceEntity& query, const SurfaceEntity& templ)
//****************************************************************************
{
	SurfaceEntity::points_cit p,dp;
	SurfaceEntity::triangles_cit t,dt;

	BOOST_MESSAGE(" ** check if the surface has " 
								<< templ.getNPoints() << " vertices and " 
								<< templ.getNTriangles() << " triangles.");
	BOOST_CHECK(query.getNPoints()     == templ.getNPoints());
	BOOST_CHECK(query.getNTriangles() == templ.getNTriangles());

	BOOST_MESSAGE(" ** check the point density (should be "
								<< templ.getPointDensity() << ")");
	BOOST_CHECK_CLOSE(query.getPointDensity(), templ.getPointDensity(), 
										DBL_EPSILON);

	BOOST_MESSAGE(" ** check the probe radius (should be "
								<< templ.getProbeRadius() << ")");
	BOOST_CHECK_CLOSE(query.getProbeRadius(), templ.getProbeRadius(), 
										DBL_EPSILON);

	BOOST_MESSAGE(" ** check the number of spheres (should be " 
								<< templ.getSphereCount() << ")");
	BOOST_CHECK(query.getSphereCount() == templ.getSphereCount());
			 
	BOOST_MESSAGE(" ** compare the point with those of the template");
	dp = templ.begin();
	for(p=query.begin();p!=query.end();++p, ++dp)
	{
		BOOST_CHECK(*p == *dp);

		/*
		cout << p->getPosition()[0] << ", "
		     << p->getPosition()[1] << ", "
		     << p->getPosition()[2] << endl;
		cout << dp->getPosition()[0] << ", "
				 << dp->getPosition()[1] << ", "
				 << dp->getPosition()[2] << endl;
		*/

		/*
		cout << p->getNearestAtom() << " == " << dp->getNearestAtom() << endl;
		cout << p->getFaceType() << " == " << dp->getFaceType() << endl;
		cout << p->getMasterIndex() << " == " << dp->getMasterIndex() << endl;
		*/
	}
			 
	BOOST_MESSAGE(" ** compare the triangles with those of the template");
	dt = templ.beginTriangles();
	for(t=query.beginTriangles();t!=query.endTriangles();++t, ++dt)
	{
		BOOST_CHECK(t->getFaceType() == dt->getFaceType());
		BOOST_CHECK(t->getFaceNum() == dt->getFaceNum());
				 
		for(int v=A;v<=C;++v)
		{
			BOOST_CHECK(t->getVertex(v)->getMasterIndex() == 
									dt->getVertex(v)->getMasterIndex());

		}
	}
}

//******************************************************************************
void compare_views(SurfaceView& templ, SurfaceView& query)
//******************************************************************************
{
  BOOST_MESSAGE("");

	BOOST_MESSAGE(" ** check the point density (should be "
								<< templ.getPointDensity() << ")");
	BOOST_CHECK_CLOSE(query.getPointDensity(), templ.getPointDensity(), 
										DBL_EPSILON);

	BOOST_MESSAGE(" ** check the probe radius (should be "
								<< templ.getProbeRadius() << ")");
	BOOST_CHECK_CLOSE(query.getProbeRadius(), templ.getProbeRadius(), 
										DBL_EPSILON);

	BOOST_MESSAGE(" ** check the number of spheres (should be " 
								<< templ.getSphereCount() << ")");
	BOOST_CHECK(query.getSphereCount() == templ.getSphereCount());

  BOOST_MESSAGE(" ** check the column count (should be " 
                << templ.getNColumns() << ")");
  BOOST_CHECK(query.getNColumns() == templ.getNColumns());

  BOOST_MESSAGE(" ** check the names of the columns");
  for (int c=0;c<templ.getNColumns();++c)
    BOOST_CHECK(query.getColumnName(c) == templ.getColumnName(c));
}

//******************************************************************************
void compare_views(VectorView& templ, VectorView& query)
//******************************************************************************
{
  compare_views((SurfaceView&) templ, (SurfaceView&) query);

	BOOST_MESSAGE(" ** check if the view has " 
								<< templ.getNPoints() << " vertices and " 
								<< templ.getNTriangles() << " triangles selected.");
	BOOST_CHECK(query.getNPoints()     == templ.getNPoints());
	BOOST_CHECK(query.getNTriangles() == templ.getNTriangles());

	VectorView::pointIter p,dp;
	VectorView::triangleIter t,dt;

	BOOST_MESSAGE(" ** compare the point with those of the template");
	dp = templ.begin();
	for(p=query.begin();p!=query.end();++p, ++dp)
	{
		BOOST_CHECK(*p == *dp);
    
    //check the point data
    for(int c=0;c<templ.getNColumns();++c)
      BOOST_CHECK_CLOSE(p->data()[c], dp->data()[c], DBL_EPSILON);
	}

	BOOST_MESSAGE(" ** compare the triangles with those of the template");
	dt = templ.beginTriangles();
	for(t=query.beginTriangles();t!=query.endTriangles();++t, ++dt)
	{
		BOOST_CHECK(t->getFaceType() == dt->getFaceType());
		BOOST_CHECK(t->getFaceNum() == dt->getFaceNum());
				 
		for(int v=A;v<=C;++v)
		{
			BOOST_CHECK(t->getVertex(v)->getMasterIndex() == 
									dt->getVertex(v)->getMasterIndex());

		}
	}
}

//****************************************************************************
void compare_views(PatchView& templ,  PatchView& query)
//****************************************************************************
{
  compare_views((VectorView&) templ, (VectorView&) query);

	BOOST_MESSAGE(" ** check if the view has " 
								<< templ.getNPoints() << " vertices and " 
								<< templ.getNTriangles() << " triangles selected.");
	BOOST_CHECK(query.getNPoints()     == templ.getNPoints());
	BOOST_CHECK(query.getNTriangles() == templ.getNTriangles());

	BOOST_MESSAGE(" ** check if both patches have both " 
                << templ.getNBorderPoints() << " border points");
  BOOST_CHECK(templ.getNBorderPoints() == query.getNBorderPoints());

  BOOST_MESSAGE(" ** compare the border vertices");
  PatchView::borderIter b, db;
  db = query.beginBorder();
  for(b=templ.beginBorder();b!=templ.endBorder();++b, ++db)
    BOOST_CHECK(*b == *db);
}

//****************************************************************************
void compare_views(GraphView& templ,  GraphView& query)
//****************************************************************************
{
  compare_views((SurfaceView&) templ, (SurfaceView&) query);

  GraphView::vertexIter tv, qv;

	BOOST_MESSAGE(" ** compare the point with those of the template");
	tv = templ.begin();
	for(qv=query.begin();qv!=query.end();++qv, ++tv)
	{
		BOOST_CHECK(*qv == *tv);
    
    //check the point data
    for(int c=0;c<templ.getNColumns();++c)
      BOOST_CHECK_CLOSE(qv->data()[c], tv->data()[c], DBL_EPSILON);
	}

	BOOST_MESSAGE(" ** compare all connections and non-connections");
  GraphView::vertexIter tvs, tvt, qvs, qvt;
	tvs = templ.begin();
	for(qvs=query.begin();qvs!=query.end();++qvs, ++tvs)
  {
    tvt = templ.begin();
    for(qvt=query.begin();qv!=query.end();++qvt, ++tvt)
      BOOST_CHECK(templ.connected(tvs,tvt) == templ.connected(qvs,qvt));
  }
}
