/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "vectorview_test.hh"

#include <float.h>
#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include "templates.hh"
#include "utils.hh"

#include <vectorview.hh>
#include <simplebuilder.hh>
#include <spherebuilder.hh>

#ifdef HAVE_STD
#  include <iostream>
#else
#  include <iostream.h>
#endif

#ifdef HAVE_NAMESPACES
using namespace boost::unit_test_framework;
using namespace Surface;
using namespace std;
#endif 

//******************************************************************************
vectorview_test::vectorview_test()
//******************************************************************************
: surface(3.0, 1.4, 0), view(0)
{
	makeDodekaeder(surface);
}

//******************************************************************************
vectorview_test::~vectorview_test()
//******************************************************************************
{
	if (view) delete view;
}

//******************************************************************************
void vectorview_test::test_simplebuilder()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n VectorView simplebuilder test \n");

	int i,j;
	VectorView::pointIter p;
	VectorView::triangleIter t;

				
	BOOST_MESSAGE(" !! create a new vector view" );
	SimpleBuilder<VectorView> builder;
	view = dynamic_cast<VectorView*>(surface.createView(builder));

	BOOST_MESSAGE(" ** test the number of vertices and triangles selected" );
	BOOST_CHECK(view->getNPoints() == surface.getNPoints());
	BOOST_CHECK(view->getNTriangles() == surface.getNTriangles());

	BOOST_MESSAGE(" ** test for the correct return of point density, sphere count and "
			 << "probe radius" );
	BOOST_CHECK_CLOSE(view->getProbeRadius(),surface.getProbeRadius(), 
										DBL_EPSILON);
	BOOST_CHECK_CLOSE(view->getPointDensity(),surface.getPointDensity(), 
										DBL_EPSILON);
	BOOST_CHECK(view->getSphereCount() == surface.getSphereCount());
				
	BOOST_MESSAGE(" ** check the 6th points position (for random access)" );
	p = view->pointAt(5);
	for(j=0;j<3;++j)
		BOOST_CHECK_CLOSE((*p)[j], dodek_points[5][j], DBL_EPSILON);

	BOOST_MESSAGE(" ** check positions from begin to end" );
	i = 0;
	for(p=view->begin();p!=view->end();++p, ++i)
		for(int j=0;j<3;++j)
			BOOST_CHECK_CLOSE((*p)[j], dodek_points[i][j], DBL_EPSILON);
		
	BOOST_MESSAGE(" ** check the triangles" );
	t = view->triangleAt(7);
	for(j=0;j<3;++j)
		BOOST_CHECK((*t)[j]->getMasterIndex() == dodek_connections[7][j]);

	BOOST_MESSAGE(" ** check connections from begin to end" );
	i = 0;
	for(t=view->beginTriangles();t!=view->endTriangles();++t, ++i)
		for(int j=0;j<3;++j)
			BOOST_CHECK((*t)[j]->getMasterIndex() == dodek_connections[i][j]);
}
			
//******************************************************************************
void vectorview_test::test_adddelete()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n VectorView add-delete test \n");

	int i,j;
	int size;
	VectorView::pointIter p, pnew;
	VectorView::triangleIter t;

	size = view->getNPoints();

	p = view->begin();
	t = view->beginTriangles();

	BOOST_MESSAGE(" !! add a new point" );
	pnew = view->addPoint();
	pnew->setMasterIndex(100);
			
	BOOST_MESSAGE(" ** check if the iterators are invalid" );
	BOOST_CHECK(!p.isValid());
	BOOST_CHECK(!t.isValid());

	BOOST_MESSAGE(" ** check if the surface and the view have now " << size + 1
			 << " more points" );
	BOOST_CHECK(view->getNPoints() == size + 1);
	BOOST_CHECK(surface.getNPoints() == size + 1);

	BOOST_MESSAGE(" ** check if the new point is accessible" );
	BOOST_CHECK(view->pointAt(size)->getMasterIndex() == 100 );

	BOOST_MESSAGE(" !! reassign" );
	p = view->begin();
	t = view->beginTriangles();

	BOOST_MESSAGE(" ** check if the iterators are valid again" );
	BOOST_CHECK(p.isValid());
	BOOST_CHECK(t.isValid());
				
	BOOST_MESSAGE(" !! delete the new point" );
	view->deletePoint(pnew);
				
	BOOST_MESSAGE(" ** check if the iterators are invalid" );
	BOOST_CHECK(!p.isValid());
	BOOST_CHECK(!t.isValid());
				
	BOOST_MESSAGE(" ** check if the surface and the view have again " << size 
			 << " points" );
	BOOST_CHECK(view->getNPoints() == size);
	BOOST_CHECK(surface.getNPoints() == size);
}

//******************************************************************************
void vectorview_test::test_deepcopy()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << 
								"\n VectorView deep copy test \n");

  BOOST_MESSAGE(" !! make a deep copy of the view \n");
  VectorView* deepcopy = dynamic_cast<VectorView*>(view->clone());
  compare_views(*view, *deepcopy);

  SurfaceEntity& newentity = deepcopy->getSurfaceEntity();

  BOOST_MESSAGE(" ** check if the two views are not identical\n");
  BOOST_CHECK(deepcopy != view);
  
  BOOST_MESSAGE(" ** check if the two entities are not identical\n");
  BOOST_CHECK(&deepcopy->getSurfaceEntity() != &view->getSurfaceEntity());
  
  BOOST_MESSAGE(" ** compare the two surface entities\n");
  check_surfaces(newentity, surface);

  BOOST_MESSAGE(" !! create a subselection\n");
  SphereBuilder<VectorView> subselector(3.0, view->begin()->getPosition());
  VectorView* subselection = 
    dynamic_cast<VectorView*>(surface.createView(subselector));

  BOOST_MESSAGE(" !! make a deep copy of the subselection view \n");
  deepcopy = dynamic_cast<VectorView*>(subselection->clone());

  compare_views(*subselection, *deepcopy);

  delete deepcopy;
  delete subselection;
}

//******************************************************************************
void vectorview_test::test_entityowner()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << 
								"\n SurfaceView entity ownership test \n");

  BOOST_MESSAGE(" !! make a deep copy of the view \n");
  SurfaceView* owningview = view->clone();
  
  BOOST_MESSAGE(" ** check if the view is the owner of the observed entity");
  BOOST_CHECK(owningview->isOwner());

  BOOST_MESSAGE(" !! get a reference to the new surface entity \n");
  SurfaceEntity& entityref = owningview->getSurfaceEntity();
  BOOST_MESSAGE(" ** check if the view is still owner of the observed entity");
  BOOST_CHECK(owningview->isOwner());

  BOOST_MESSAGE(" !! take a pointer to the new surface entity \n");
  SurfaceEntity* entityptr = owningview->takeSurfaceEntity();
  BOOST_MESSAGE(" ** check if the view is not the owner of the observed entity");
  BOOST_CHECK(!owningview->isOwner());

  BOOST_MESSAGE(" ** check if the surface is deleted together with the view");
  delete owningview;
  entityptr->getNPoints();
  delete entityptr;
}

//******************************************************************************
void vectorview_test::test_iterator()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << 
								"\n VectorView iterator test \n");

	SurfaceEntity::triangles_it t   = view->beginTriangles();
	SurfaceEntity::triangles_cit tc = view->beginTriangles();
	SurfaceEntity::points_it p      = view->begin();
	SurfaceEntity::points_cit pc    = view->begin();

  BOOST_MESSAGE(" !! create a temporary view");
  SimpleBuilder<VectorView> builder;
  VectorView* tempview = dynamic_cast<VectorView*>(surface.createView(builder));

  BOOST_MESSAGE(" !! create a view iterator");  
  VectorView::pointIter pv = tempview->begin();

  BOOST_MESSAGE(" ** check if the iterator is valid!");
  BOOST_CHECK(pv.isValid());

  BOOST_MESSAGE(" !! delete the view");
  delete tempview;
  
  BOOST_MESSAGE(" ** check if the iterator is not valid!");
  BOOST_CHECK(!pv.isValid());
}

//******************************************************************************
void vectorview_test::test_data()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n VectorView data test \n");

	size_t size  = view->getNColumns();

	BOOST_MESSAGE(" !! add a data column to the vector view object." );
	size_t index = view->addColumn("test", 0.25);
	BOOST_MESSAGE(" ** check if the column count is now " << size + 1 << "! " 
			 );
	BOOST_CHECK(view->getNColumns() == (size + 1));
				
	BOOST_MESSAGE(" ** check if all the point data values are equal to 0.25 " 
			 << "\n !! and set the value to the index + 0.3 " );
	int ix = 0;
	for(VectorView::pointIter p=view->begin();p!=view->end();++p, ++ix)
	{
		BOOST_CHECK_CLOSE(p->data()[index], 0.25, DBL_EPSILON);
		p->setData(index, ix + 0.3);
	}

	BOOST_MESSAGE(" !! extract the test data column from the view " );
	RbDblV column = view->extractDataColumn(index);
				
	BOOST_MESSAGE(" ** check if the entries of the vector all have ix + 0.3" 
			 << "\n !! give them the new value ix * ix (for later test)" );
	for(ix = 0; ix<column.dim();++ix)
	{
		BOOST_CHECK_CLOSE(column[ix], ix + 0.3, DBL_EPSILON);
		column[ix] = ix * ix;
	}

	BOOST_MESSAGE(" !! add a new column with the new values (\"test2\")" );
	size_t index_2 = view->addColumn("test2", column);

	BOOST_MESSAGE(" ** check if the column count is now " << size + 2 << "! " 
			 );
	BOOST_CHECK(view->getNColumns() == (size + 2));

	BOOST_MESSAGE(" ** check if all the point data values of the new column" 
			 << " are correct" );
	ix = 0;
	for(VectorView::pointIter p=view->begin();p!=view->end();++p, ++ix)
		BOOST_CHECK_CLOSE(p->data()[index_2], column[ix], DBL_EPSILON);

	size = view->getNColumns();

	BOOST_MESSAGE(" !! extract the complete data matrix" );
	RbDblM matrix = view->extractDataMatrix();

	BOOST_MESSAGE(" ** check if the matrix has " << size << " columns and "
			 << view->getNPoints() << " rows." );
	BOOST_CHECK(matrix.cno() == size);
	BOOST_CHECK(matrix.rno() == view->getNPoints());

	BOOST_MESSAGE(" ** check if the data values are correct." );
	for(int r=0;r<matrix.rno();++r)
	{
		BOOST_CHECK_CLOSE(matrix(r,0), r+0.3, DBL_EPSILON);
		BOOST_CHECK_CLOSE(matrix(r,1), double(r * r), DBL_EPSILON);
	}

	BOOST_MESSAGE(" !! remove the first column" );
	view->deleteColumn(index);

	BOOST_MESSAGE(" ** check if the column count is now " << size - 1 << "! " 
			 );
	BOOST_CHECK(view->getNColumns() == (size - 1));

	BOOST_MESSAGE(" ** check if the column name is \"test2\"!" );
	BOOST_CHECK(view->getColumnName(index) == "test2");

	BOOST_MESSAGE(" ** check if all the point data values of the new column" 
			 << " are correct" );
	ix = 0;
	for(VectorView::pointIter p=view->begin();p!=view->end();++p, ++ix)
		BOOST_CHECK_CLOSE(p->data()[index], column[ix], DBL_EPSILON);
				
	BOOST_MESSAGE(" !! set the name of the first column to \"_test_\"" );
	view->setColumnName(index, "_test_");

	BOOST_MESSAGE(" ** check if the column name is now \"_test_\"!" );
	BOOST_CHECK(view->getColumnName(index) == "_test_");
				
}

//******************************************************************************
vectorview_test_suite::vectorview_test_suite() 
//******************************************************************************
: test_suite("vectorview_test_suite")
{
	// add member function test cases to a test suite
	boost::shared_ptr<vectorview_test> instance(new vectorview_test);
		
	test_case* simplebuilder_test_case = 
		BOOST_CLASS_TEST_CASE(&vectorview_test::test_simplebuilder,	instance); 
	test_case* adddelete_test_case = 
		BOOST_CLASS_TEST_CASE(&vectorview_test::test_adddelete,	instance); 
	test_case* deepcopy_test_case = 
		BOOST_CLASS_TEST_CASE(&vectorview_test::test_deepcopy,	instance); 
	test_case* owning_test_case = 
		BOOST_CLASS_TEST_CASE(&vectorview_test::test_entityowner,	instance); 
	test_case* iterator_test_case = 
		BOOST_CLASS_TEST_CASE(&vectorview_test::test_iterator,	instance); 
	test_case* data_test_case = 
		BOOST_CLASS_TEST_CASE(&vectorview_test::test_data,	instance); 
		
	//test case, number of tests that should fail
	add(simplebuilder_test_case, 0);
	add(adddelete_test_case, 0);
	add(deepcopy_test_case, 0);
	add(owning_test_case, 0);
	add(iterator_test_case, 0);
	add(data_test_case, 0);
}
