/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "curvaturecalculator.hh"

#include <Lu.hh>
#include <limits>

#include <graphview.hh>
#include <surface.hh>
#include <spherebuilder.hh>
#include <simplebuilder.hh>

#include "globalcurvselector.hh"
#include "localcurvselector.hh"
#include "propcalcfactory.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace RazorBack;
SURFACENS_USE
#endif 

#include <iostream>

REGISTER_PROPCALCCLASS("curvature", CurvatureCalculator);

//******************************************************************************
double det(const RbDblM& mat)
//******************************************************************************
{
	return 
		mat(0,0)*(mat(1,1)*mat(2,2) - mat(2,1)*mat(1,2)) -
		mat(1,0)*(mat(0,1)*mat(2,2) - mat(2,1)*mat(0,2)) +
		mat(2,0)*(mat(0,1)*mat(1,2) - mat(1,1)*mat(0,2));		
		
}

//******************************************************************************
RbDblV solve(const RbDblM& A, const RbDblV& b)
//******************************************************************************
{
	double D = det(A);

// 	cout << " DD D = " << D << endl;

// 	if (fabs(D)<DBL_EPSILON)
// 	{
// 		cout << " DD " << A(0,0) << " " << A(0,1) << " " << A(0,2) << endl;
// 		cout << " DD " << A(1,0) << " " << A(1,1) << " " << A(1,2) << endl;
// 		cout << " DD " << A(2,0) << " " << A(2,1) << " " << A(2,2) << endl;
// 	}

	RbDblM d1(3,3);
	d1.col(A.col(1), 0);
	d1.col(A.col(2), 1);
	d1.col(b, 2);

	RbDblM d2(3,3);
	d2.col(A.col(0), 0);
	d2.col(A.col(2), 1);
	d2.col(b, 2);

	RbDblM d3(3,3);
	d3.col(A.col(0), 0);
	d3.col(A.col(1), 1);
	d3.col(b, 2);
	

	RbDblV result(3);
	result[0] = det(d1)/D;
	result[1] = det(d2)/D;
	result[2] = det(d3)/D;

	return result;
}

//******************************************************************************
CurvatureCalculator::CurvatureCalculator()
//******************************************************************************
: PropCalculator(), cutoff_(-1.0)
{
  colnames_[0] = "CC1";
  colnames_[1] = "CC2";
  colnames_[2] = "STI";
}

//******************************************************************************
bool CurvatureCalculator::setParameter(const std::string& name, 
																			 const std::string& value)
//******************************************************************************
{
	if (uppercase(name) == "CUTOFF")
	{
		if (!sscanf(value.c_str(), "%lf", &cutoff_))
			throw WrongParameterValue(name, value);

		return true;
	}
	else
		return false;
}


//******************************************************************************
bool CurvatureCalculator::operator()()
//******************************************************************************
{
  if (surface_->isLocked())
    return false;

  vector<int> colindices = setupColumns();

	CurvSelector* selector = 0;
	if (cutoff_<0)
		selector = new LocalCurvSelector(surface_);
	else
	  selector = new GlobalCurvSelector(surface_, cutoff_);

  for(SurfaceEntity::points_it p=surface_->begin();p!=surface_->end();++p)
  {
    int size = (*selector)(p);
    
    RbDblM coordinates(size, 3);
    RbDblM A(3,3);
    RbDblV b(3);

    //generate the equation system
    transformCoordinates(p, selector->begin(), selector->end(), coordinates);
    makeEquations(coordinates, A, b);

    //solve for the Hessian matrix by LU decomposition
    Lu_ solver(A);
    RbDblV hessian = solver.solve(b);

// 		cout << " DD p = " << p->getMasterIndex() << endl;
// 		RbDblV test = solve(A, b);
// 		cout << " DD test = " 
// 				 << test[0] << " " << test[1] << " " << test[2] << endl;
// 		cout << " DD hess = " 
// 				 << hessian[0] << " " << hessian[1] << " " << hessian[2] << endl;



    //calculate the curvatures
    p->setData(colindices[0], eigenvalue(hessian, 1));
    p->setData(colindices[1], eigenvalue(hessian, -1));
    p->setData(colindices[2], calcSTI(p->data()[colindices[0]],
                                      p->data()[colindices[1]]));
  }

	delete selector;
}


//******************************************************************************
std::string CurvatureCalculator::getColName(curvature_t curvature) const
//******************************************************************************
{
  return colnames_[curvature2ix(curvature)];
}

//******************************************************************************
void CurvatureCalculator::setColName(curvature_t curvature, 
                                     const std::string& name)
//******************************************************************************
{
  colnames_[curvature2ix(curvature)] = name;
}
    
//******************************************************************************
int CurvatureCalculator::curvature2ix(curvature_t curvature) const
//******************************************************************************
{
  int ix=(int) curvature;
  if (ix<0) ix=0;
  if (ix>2) ix=2;

  return ix;
}

//******************************************************************************
vector<int> CurvatureCalculator::setupColumns()
//******************************************************************************
{
  //setup the columns and column indices
  vector<int> indices(3);
  for(int c=0;c<3;++c)
  {
    indices[c] = surface_->indexOfColumn(colnames_[c]);
    if (indices[c]<0)
      indices[c] = surface_->addColumn(colnames_[c]);
  }

  return indices;
}

//******************************************************************************
void CurvatureCalculator::transformCoordinates(
	const SurfaceEntity::points_it& p,
  CurvSelector::const_iterator begin, 
	const CurvSelector::const_iterator end, RbDblM& coordinates)
//******************************************************************************
{
  const RbDblV& center = p->getPosition();
  RbDblV en = p->getNormal();
	en[W] = 0;
  en.vec_norm();

  RbDblV eu = cross(en, (*begin)->getPosition() - center);
	eu[W] = 0;
  eu.vec_norm();

  RbDblV ev = cross(en, eu);
	ev[W] = 0;

	
	int r=0;
	while (begin!=end)
	{
		RbDblV pos = (*begin++)->getPosition() - center;
		pos[W] = 0;
    coordinates(r, 0) = eu * pos;
		coordinates(r, 1) = ev * pos;
		coordinates(r, 2) = en * pos;

		++r;
	}
}

//******************************************************************************
void CurvatureCalculator::makeEquations(const RbDblM& coordinates, 
                                        RbDblM& A, RbDblV& b)
//******************************************************************************
{
  //necessary values
  double u4   = 0;
  double u3v  = 0;
  double u2v2 = 0;
  double uv3  = 0;
  double v4   = 0;
  double nu2  = 0;
  double nuv  = 0;
  double nv2  = 0;

  for(int r=0;r<coordinates.rno();++r)
  {
    double u = coordinates(r, 0);
    double v = coordinates(r, 1);
    double n = coordinates(r, 2);

    double u2 = u * u;
    double v2 = v * v;

    u4   += u2 * u2;
    u3v  += u2 * u * v;
    u2v2 += u2 * v2;
    uv3  += u * v2 * v;
    v4   += v2 * v2;
    nu2  += n * u2;
    nuv  += n * v * u;
    nv2  += n * v2;
  }

  A(0,0) = u4;
  A(0,1) = 2*u3v;
  A(0,2) = u2v2;
  A(1,0) = u3v;
  A(1,1) = 2*u2v2;
  A(1,2) = uv3;
  A(2,0) = u2v2;
  A(2,1) = 2*uv3;
  A(2,2) = v4;

  b[0] = nu2;
  b[1] = nuv;
  b[2] = nv2;
}

//******************************************************************************
double CurvatureCalculator::eigenvalue(const RbDblV& hessian, double sign)
//******************************************************************************
{
  double diff02 = hessian[0] - hessian[2];

	if ((hessian[0]==numeric_limits<double>::quiet_NaN()) ||
			(hessian[1]==numeric_limits<double>::quiet_NaN()) ||
			(hessian[2]==numeric_limits<double>::quiet_NaN()))
	{
// 		cout <<  " DD nan" << endl;
		return 0;
	}
	
  return ((hessian[0] + hessian[2]) +
					sign * sqrt(4*hessian[1]*hessian[1] + diff02*diff02));
}

//******************************************************************************
double CurvatureCalculator::calcSTI(double cc1, double cc2)
//******************************************************************************
{
	if (((cc1>0) && (cc2>0)) || 
			((cc1>0) && (cc2<DBL_EPSILON) && (fabs(cc1)>fabs(cc2))))
		return (cc1-cc2)/cc1;
	else if (((cc1<0) && (cc2<0)) || 
					 ((cc1>0) && (cc2<=DBL_EPSILON) && (fabs(cc1)<=fabs(cc2))))
		return (cc1 + 3*cc2)/cc2;
  else
		return -1.0;
}

