/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file curvature.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 18 Jul 2004, last CVS commit 2004/07/18 09:05:21
 */

#ifndef CURVATURE_HH
#define CURVATURE_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string>
#include <surface.hh>

#include <surfacens.hh>
SURFACE_FORWARD(GraphView);

#include "curvselector.hh"
#include "propcalculator.hh"

/**
 * \brief a curvature calculator
 * 
 * The curvature calculation is based on a paper by Zachmann et. al. in
 * 1992. It calculates approximated canonical curvatures for each point of the
 * surface based on the Hessian matrix of an elliptical or hyperbolical
 * paraboloid which are fitted to the neighborhood of the point in a least
 * squares sence. 
 */
class CurvatureCalculator : public PropCalculator
{
public:

  //****************************************************************************
  // PUBLIC TYPE DEFINITION
  //****************************************************************************

  /**
   * \brief the different types of curvatures
   * 
   * The calculator calculates three different curvature values: 
   */
  enum curvature_t 
  { 
    ctCC1, ///< canonical curvature 1
    ctCC2, ///< canonical curvature 2
    ctSTI, ///< shape topology index
  };

	//****************************************************************************
	// INSTANTIATION AND PARAMETER SETUP
	//****************************************************************************

	static PropCalculator* createInstance() { return new CurvatureCalculator(); }

	bool setParameter(const std::string& name, const std::string& value);

  //****************************************************************************
  // OPERATIONS
  //****************************************************************************

  /** 
   * \brief performs the curvature calculation for all surface points
   * 
   * \return \c true if the calculation was successfull, \c false otherwise
   */
  bool operator()();

  //****************************************************************************
  // PROPERTY METHODS
  //****************************************************************************

  double getCutoff() const { return cutoff_; }
  void setCutoff(double cutoff=-1.0) { cutoff_ = cutoff; }

  std::string getColName(curvature_t curvature) const; 
  void setColName(curvature_t curvature, const std::string& name);

	std::string getShortName() const { return "cannonical curvatures "; }
	std::string getDescription() const { 
		return "cannonical and global curvatures after Zachmann et. al. 1992"; }
    
private:

	/** 
	 * \brief default destructor
	 */
  CurvatureCalculator();

	//****************************************************************************
	// PRIVATE MEMBERS
	//****************************************************************************

  /// \brief the cutoff range for the point selection
  double cutoff_;

  /// \brief the names of the three columns that are added to the surface
  std::string colnames_[3];

private:

  //****************************************************************************
  // INTERNAL TYPES
  //****************************************************************************

  //****************************************************************************
  // INTERNAL PRIVATE METHODS
  //****************************************************************************

  //@{
  /// \brief internal method used by the algorithm
  int curvature2ix(curvature_t curvature) const;
  vector<int> setupColumns();

  void transformCoordinates(const SURFACENS SurfaceEntity::points_it& p,
														CurvSelector::const_iterator begin, 
														const CurvSelector::const_iterator end, 
														RbDblM& coordinates);
  void makeEquations(const RbDblM& coordinates, RbDblM& A, RbDblV& b);

	double eigenvalue(const RbDblV& hessian, double sign);
	double calcSTI(double cc1, double cc2);
  //@}

  //****************************************************************************
  // INTERNAL DATA MEMBERS
  //****************************************************************************

};

#endif // CURVATURE_HH
