/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file globalcurvselector.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 19 Jul 2004, last CVS commit 2004/07/19 07:12:15
 */

#ifndef CURVSELECTOR_HH
#define CURVSELECTOR_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <set>
#else
#  include <set.h>
#endif

#include <surface.hh>
#include <surfacens.hh>

/**
 * \brief base class for point selection in the curvature calculation
 *
 * This class is the base class for different types of point selectors in the
 * curvature calculation. Concrete selection classes have to specify the
 * operator()() method. The points are stored in a std::set of
 * SurfaceEntity::point_it objects and the access to the selection set is
 * handled by the base class via the begin() and end() functions.
 */

class CurvSelector
{
protected:

	//****************************************************************************
	// PRIVATE TYPE DEFNIITIONS
	//****************************************************************************

  /// \brief point selection set
  typedef set<SURFACENS SurfaceEntity::points_it, 
              SURFACENS lt_iterator<SURFACENS SurfaceEntity::points_it> > 
  selection_t;
	typedef selection_t::iterator selection_it;

public:

	//****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//****************************************************************************

  //@{
	/// \brief constant iterator type for the point selection
	typedef selection_t::const_iterator const_iterator;
  //@}

	//****************************************************************************
	// PUBLIC METHODS
	//****************************************************************************

	/** 
	 * \brief standard constructor
	 * 
	 * \param surface a pointer to the handled surface
	 */
	CurvSelector(SURFACENS SurfaceEntity* surface) 
		: surface_(surface), points_() {}

	/** 
	 * \brief virtual destructor
	 *
	 * The virtual destructor is here for correct destruction of subclasses.
	 */
	virtual ~CurvSelector() {}

	/** 
	 * \brief selection operator
	 *
	 * The selection operator must be implemented by the concrete selection
	 * classes. It must return the number of selected points.
	 *
	 * \param p the current reference point that serves as the center of the
	 *          calculation.
	 * 
	 * \return the number of selected points
	 */
  virtual size_t operator()(const SURFACENS SurfaceEntity::points_it& p) = 0;

	//@{
	/** 
	 * \brief begin and end iterators for the selected sequence.
	 * 
	 * \return a constant iterator to the beginning and behind the end of the
	 * selected point sequence.
	 */
	const_iterator begin() const { return points_.begin(); }
	const_iterator end()   const { return points_.end(); }
	//@}

protected:
	
	/// \brief pointer to the handled surface
	SURFACENS SurfaceEntity* surface_;
	
	/// \brief surface point selection set
	selection_t points_;
};


#endif // CURVSELECTOR_HH
