/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file curvature.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 18 Jul 2004, last CVS commit 2004/07/18 09:05:21
 */

#ifndef CURVATURE_HH
#define CURVATURE_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string>
#include <surface.hh>

#include <surfacens.hh>
SURFACE_FORWARD(GraphView);

#include "curvselector.hh"
#include "fieldcalculator.hh"

/**
 * \brief an ESP calculator
 * 
 * The ESP is calculated based on the atomic charges of the atom using coloumbs law:
 *
 * \f[
 *  ESP(p_i) = \sum\limits_{j=1}\limits^{N} \frac{q_j}{\Vert r_i - r_j \Vert}
 * \f]
 *
 * where p<sub>i</sub> is the i'th surface point and r<sub>i</sub> is its
 * position in 3D space. q<sub>j</sub> and r<sub>j</sub> are the partial charge
 * and position of the j'th atom respectively.
 */
class ESPCalculator : public FieldCalculator
{
public:

	//****************************************************************************
	// INSTANTIATION AND PARAMETER SETUP
	//****************************************************************************

	static PropCalculator* createInstance() { return new ESPCalculator(); }

  //****************************************************************************
  // OPERATIONS
  //****************************************************************************

  /** 
   * \brief performs the esp calculation for all surface points
   * 
   * \return \c true if the calculation was successfull, \c false otherwise
   */
  bool operator()();

  //****************************************************************************
  // PROPERTY METHODS
  //****************************************************************************

  std::string getColName() const { return colname_; }
  void setColName(const std::string& name) { colname_ = name; }

	std::string getShortName() const { return "electrostatic potential"; }
	std::string getDescription() const { 
		return 
	"Electrostatic potential calculated from atomic point charges by Coloumbs law"; }
    
private:

	/** 
	 * \brief default destructor
	 */
  ESPCalculator() : FieldCalculator(), colname_("ESP") { }

	/// \brief the name of the calculated data column
	std::string colname_;

private:
	
  //@{
  /// \brief internal method used by the algorithm
  int setupColumn();
	//@}
};

#endif // CURVATURE_HH
