/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file fieldcalculator.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 21 Jul 2004, last CVS commit 2004/07/21 08:13:07
 */

#ifndef FIELDCALCULATOR_HH
#define FIELDCALCULATOR_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "atoms.hh"
#include "propcalculator.hh"

/**
 * \brief an exception for read errors in PDB files
 */

class PDBReadError : public SURFACENS Surfcomp_e
{
public:

	PDBReadError(const std::string& filename) :
		Surfcomp_e(-4202,"Error parsing the information in " + filename +"!") { }
};


/**
 * \brief an abstract base class for molecular field calculators
 * 
 * Molecular fields like the electrostatic potential can be mapped onto the
 * surface in a very easy way. They usually depend only on the positions of the
 * atoms and a field source property assigned to every field. Thus all field
 * calculators need a list of the atoms beneath a surface plus a field value
 * for that. This has to be provided by a pdb file where the occupancy
 * information is taken as the field property for each atom. The name of the
 * pdb file is passed to the calculator by the named parameter \p PDB.
 * 
 */
class FieldCalculator : public PropCalculator
{
public:

	/// \brief virtual destructor
	virtual ~FieldCalculator() { }
							
	virtual bool setParameter(const std::string& name, const std::string& value);

protected:
	
	/** 
	 * \brief default constructor
	 * 
	 * The default constructor is protected to prohibit direct creation.
	 */
	FieldCalculator(): PropCalculator(), atoms_() { }

	/// \brief the list of the relevant atom information
	atoms_t atoms_;

private:

	/** 
	 * \brief reads the molecular information from a file
	 * 
	 * \param filename a c-string that contains the location of the pdb file
	 *
	 * \exception PDBReadError if the information in the given pdb file could not
	 *                         be parsed correctly.
	 * 
	 * \return \c true if the molecule could be read successfully
	 */
	bool readMolecule(const char* filename);
};

#endif // FIELDCALCULATOR_HH
