/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "primitives.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
#endif 

#ifdef HAVE_STD
#  include <fstream>
#else
#  include <fstream.h>
#endif

#include <surface.hh>
#include <vectorview.hh>
#include <simplebuilder.hh>

//******************************************************************************
int determineAtom(const Triangle& triangle)
//******************************************************************************
{
	int a = triangle.getVertex(A)->getNearestAtom();
	int b = triangle.getVertex(B)->getNearestAtom();
	int c = triangle.getVertex(C)->getNearestAtom();

	if ((a == b) && (b == c) && (c == a))
		return a;
	else if ((a == b) || (a == c))
		return a;
	else if (b == c)
		return b;
	else
		return c;
}

//******************************************************************************
Primitive::Primitive(const Triangle& triangle, 
										 double innerdielectric, double outerdielectric)
//******************************************************************************
: center_(4,0), normal_(4,0), area_(0.0), 
	innerdielectric_(innerdielectric), outerdielectric_(outerdielectric), 
	atom_(-1)
{
	center_ = triangle.getCenter();
	normal_ = triangle.getNormal();
	area_   = triangle.getArea();

	atom_   = determineAtom(triangle);
}

//******************************************************************************
Primitive::Primitive(const Primitive& src)
//******************************************************************************
: center_(4,0), normal_(4,0), area_(0.0), 
	innerdielectric_(1.0), outerdielectric_(78.390), 
	atom_(-1)
{
	operator=(src);
}

//******************************************************************************
Primitive& Primitive::operator=(const Primitive& src)
//******************************************************************************
{
	center_ = src.center_;
	normal_ = src.normal_;
	area_ = src.area_;
	innerdielectric_ = src.innerdielectric_;
	outerdielectric_ = src.outerdielectric_;
	atom_ = src.atom_;

	return *this;
}

//******************************************************************************
Primitives::Primitives(SurfaceEntity& surface, 
											 double innerdielectric, double outerdielectric)
//******************************************************************************
: view_(0), primitives_(), innerdielectric_(), outerdielectric_()
{
	init(surface, innerdielectric, outerdielectric);
}

//******************************************************************************
Primitives::~Primitives()
//******************************************************************************
{
	if (view_) delete view_;
}

//******************************************************************************
void Primitives::init(SurfaceEntity& surface, 
											double innerdielectric, double outerdielectric)
//******************************************************************************
{
	innerdielectric_ = innerdielectric;
	outerdielectric_ = outerdielectric;
	
	SimpleBuilder<VectorView> builder;
	view_ = dynamic_cast<VectorView*>(surface.createView(builder));	
}


//******************************************************************************
void Primitives::operator()()
//******************************************************************************
{
	VectorView::triangleIter t;
	for(t=view_->beginTriangles();t!=view_->endTriangles();++t)
		primitives_.push_back(Primitive(*t, innerdielectric_, outerdielectric_));
}


//******************************************************************************
bool Primitives::write(const std::string& filename)
//******************************************************************************
{
  ofstream stream(filename.c_str());
  if (!stream)
  {
		//prepare the message string
		stringstream errormsg;
		cerr << "File: " << filename << " could not be opened" << endl;
    return false;
  }
  
  bool result = write(stream);  
  stream.close();
  return result;	
}


//******************************************************************************
bool Primitives::write(std::ostream& stream)
//******************************************************************************
{
	char buffer[128];
	
	for(primitives_cit p=primitives_.begin();p!=primitives_.end();++p)
	{
		sprintf(buffer, 
						"%12.6f%12.6f%12.6f%12.6f%12.6f%12.6f%12.8f%8.3f%8.3f%6d%4d%4d\n",
						p->getCenter()[0],
						p->getCenter()[1],
						p->getCenter()[2],
						p->getNormal()[0],
						p->getNormal()[1],
						p->getNormal()[2],
						p->getArea(),
						p->getInnerDielectric(),
						p->getOuterDielectric(),
						p->getAtom()+1,
						1,1);

		stream << buffer;					 
	}

	return true;
}


