/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file propcalcfactory.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 09 Jun 2004, last CVS commit 2004/06/09 05:43:51
 */

#ifndef PROPCALCFACTORY_HH
#define PROPCALCFACTORY_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <map>
#else
#  include <map.h>
#endif

#include "common.hh"


  class PropCalculator;

 /**
  * \brief a collection of PropCalculator factory methods
  *
  * The PropCalcFactory represents a collection of factory methods for different
  * subclasses of the PropCalculator type. For each subtype a factory method of
  * the form <code> PropCalculator* factory()</code> can be registered to the
  * PropcalcFactory together with a name that identifies the class uniquely
  * (register(const std::string&, factory_t)). The factories can also be
  * unregistered (unregister(const std::string&)). A new (dynamic) instance of
  * a PropCalculator subclass is returned by the createClass(const std::string&)
  * method.
  *
  * The following example illustrates how to register a PropCalculator subclass:
  *
  * \code
  * class MyPropCalculator : public PropCalculator
  * {
  *   ...
  *
  *   static PropCalculator* createInstance() { return new MyPropCalculator(); }
  *   static void register();
  * };
  *
  * ...
  *
  * void MyPropCalculator::register()
  * {
  *   PropCalcFactory::instance().registerClass("MyPropCalculator", 
  *                                             &MyPropCalculator::createInstance);
  * }
  *
  * \endcode
  *
  * An alternative for the registration is the use of the
  * REGISTER_PROPCALCCLASS(id, classname) macro.
  *
  * The class is a Singleton(127) and the instance can be retrieved by
  * instance().
  */
class PropCalcFactory
{

public:
	//**************************************************************************
	// TYPES
	//**************************************************************************

	/**
	 * \brief the type of the factory method pointers.
	 */
	typedef PropCalculator* (*factory_t)();


private:

	//@{
	/**
	 * \brief type of the factory map plus iterators
	 */
	typedef std::map<std::string, factory_t> factories_t;
	typedef factories_t::const_iterator factories_cit;
	typedef factories_t::iterator factories_it;
	//@}
  
public:
	/**
	 * \brief a constant iterator to the internally stored class/method pairs
	 */
	typedef factories_cit factoryIter;
    
	//**************************************************************************
  // OPERATIONS
  //**************************************************************************

	/** 
	 * \brief registers a factory method
	 *
	 * \param name identifier of the new class
	 * \param factory method pointer to the class' factory method
	 * 
	 * \return \c true if the class could be added to the registry or \c false
	 *         if \p name is already registered.
	 *
	 * \sa unregisterClass(const std::string&)
	 */
    bool registerClass(const std::string& name, factory_t factory);
	/** 
	 * \brief unregisters a class
	 * 
	 * \param name the id of the class factory that should be unregistered
	 * 
	 * \return \c true if the class factory was found and unregistered, \c
	 *         false otherwise.
	 * 
	 * \sa registerClass(const std::string&, factory_t)
	 */
	bool unregisterClass(const std::string& name);

	/** 
	 * \brief checks if a class can be created by the PropCalcFactory
	 *
	 * \param name the id of the desired class factory
	 * 
	 * \return \c true if the class is registered, \c false otherwise.
	 *
	 * \sa createClass(const std::string& name);
	 */
	bool hasClass(const std::string& name) const;
	/** 
	 * \brief creates a new instance of a specific class
	 * 
	 * \param name the id of the desired class
	 * 
	 * \return a new instance of the class, created by the registered class
	 *         factory.
	 */
	PropCalculator* createClass(const std::string& name) const;

	//@{
	/** 
	 * \brief provides access to the property calculator sequence
	 * 
	 * 
	 * \return an factoryIter to the begin and behind the end of the registered
	 *         property calculator sequence.
	 */
	factoryIter begin() const { return factories_.begin(); }
	factoryIter end()   const { return factories_.end(); }
	//@}

	//**************************************************************************
  // SINGLETON
  //**************************************************************************

	/** 
	 * \brief provides access to the instance of the PropCalcFactory
	 * 
	 * \return a reference to the instance of the PropCalcFactory
	 */
	static PropCalcFactory& instance();
    
private:
    
	//**************************************************************************
	// PRIVATE MEMBERS
	//**************************************************************************

	/**
	 * \brief map of PropCalculator factory methods
	 * 
	 * The factory map is used internally to store the static method pointers
	 * to the various PropCalculator factory methods.
	 */
	factories_t factories_;

	//**************************************************************************
  // SINGLETON
  //**************************************************************************

	/** 
	 * \brief default constructor
	 * 
	 * The constructor is private according to the Singleton(127) pattern.
	 */
	PropCalcFactory() : factories_() { }

	/**
	 * \brief the singleton instance
	 */
	static PropCalcFactory* instance_;
};

/** 
 * \brief MACRO that registers a data class upon intialization.
 *
 * This macro can be used in the implementation of a specific data class to
 * register the class at the PropCalcFactory. \p classname can be any subclass of
 * PropCalculator that redefines the static method PropCalculator::createInstance().
 *
 * An example class definition (header file):
 *
 * \code
 * class MyPropCalculator : public PropCalculator
 * {
 *   ...
 *
 *   static PropCalculator* createInstance() { return new MyPropCalculator(); }
 *   static void register();
 * };
 * \endcode
 *
 * and the corresponding registration call in the source (.cc, .cxx or .cpp
 * file):
 * 
 * \code 
 * REGISTER_DATACLASS("myprop", MyPropCalculator);
 * \endcode
 * 
 * \param id        the registration name of the class
 * \param classname the name of the class (without quotes) that should be
 *                  registered.
 */
#define REGISTER_PROPCALCCLASS(id, classname) \
struct PropCalcRegistration \
{ \
  PropCalcRegistration() { \
    PropCalcFactory::instance(). \
      registerClass(id, &classname::createInstance);	\
  } \
}; \
\
static PropCalcRegistration registration; 


#endif // DATAFACTORY_HH
