/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file propcalculator.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 20 Jul 2004, last CVS commit 2004/07/20 14:09:57
 */

#ifndef PROPCALCULATOR_HH
#define PROPCALCULATOR_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include <string>

#include <surfacens.hh>
#include <common.hh>
SURFACE_FORWARD(SurfaceEntity);

/**
 * \brief an exception for maleformed values in named parameters.
 *
 * The maleformed named parameter can be identified by its name and value pair
 * via getName() and getValue().
 */

class WrongParameterValue : public SURFACENS Surfcomp_e
{
public:

	/** 
	 * \brief standard constructor
	 * 
	 * \param name the name of the maleformed parameter
	 * \param value the value of the parameter
	 */
	WrongParameterValue(const std::string name, const std::string value);

	/** 
	 * \brief gives the name of the bad parameter
	 * 
	 * \return the name of the wrong parameter.
	 *
	 * \sa name_, getValue()
	 */
	const std::string& getName() const { return name_; }
	/** 
	 * \brief gives the value of the bad parameter
	 * 
	 * \return the value of the wrong parameter.
	 *
	 * \sa value_, getName()
	 */
	const std::string& getValue() const { return value_; }

private:

	/// \brief the name of the bad parameter
	std::string name_;
	/// \brief the value of the bad parameter
	std::string value_;
};

/**
 * \brief the base class for property calculators
 *
 * It takes care of the pointer to the handled surface. The calculation is
 * triggered by a call to the function operator operator()().  The data is
 * initialized by the init() function and an instance can be created with the
 * static createInstance() method.
 *
 * The single calculators can read different named parameters which can be
 * given by setParameter(const std::string&, const std::string&).
 */

class PropCalculator
{
public:

	//****************************************************************************
	// DESTRUCTION, INITIALIZATION and INSTANTIATION
	//****************************************************************************

	/// \brief virtual destructor
	virtual ~PropCalculator() { }
		
	/** 
	 * \brief initializes the calculator
	 * 
	 * \param surface a pointer to the handled surface
	 */
	void init(SURFACENS SurfaceEntity* surface);

	/** 
	 * \brief creates an instance of a PropCalculator subclass
	 *  
	 * Subclasses must override this static method to allow the automatic
	 * generation of property calculators.
	 *
	 * \return a pointer to a fresh instance of a PropCalculator
	 */
	static PropCalculator* createInstance() { return 0; }

	//****************************************************************************
	// OPERATIONS
	//****************************************************************************

		/** 
		 * \brief parses the value of a named parameter.
		 * 
		 * \param name the name of the parameter
		 * \param value the value of the parameter
		 *
		 * \exception WrongParameterValue if the value of the named parameter could
		 *                                not be used.
		 * 
		 * \return \c true if the parameter was accepted \c false if it was not important
		 */
	virtual bool setParameter(const std::string& name, const std::string& value) {
		return false; }
	virtual bool operator()() = 0;

	//****************************************************************************
	// PROPERTY METHODS
	//****************************************************************************
	
	/** 
	 * \brief the short name of the property
	 * 
	 * \return a string containing the short name of the property.
	 *
	 * \sa getDescription()
	 */
	virtual std::string getShortName() const { return ""; }
	/** 
	 * \brief a description of the property
	 * 
	 * \return a string containing a more detailed description of the property
	 *
	 * \sa getDescription()
	 */
	virtual std::string getDescription() const { return ""; }

	
protected:
	
	/** 
	 * \brief default constructor
	 * 
	 * The default constructor is protected to prohibit direct creation.
	 */
	PropCalculator(): surface_(0) { }

	/** 
	 * \brief transforms into all uppercase
	 * 
	 * \param str the string that has to be converted into all uppercase letters
	 * 
	 * \return the all-uppercase version of the string
	 */
	std::string uppercase(const std::string& str);

	/// \brief a pointer to the handled surface
	SURFACENS SurfaceEntity* surface_;
};

#endif // PROPCALCULATOR_HH
