/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "propgenerator.hh"

#include <stdio.h>
#include <unistd.h>
#include <string.h>

#ifdef HAVE_STD
#  include <fstream>
#else
#  include <fstream.h>
#endif

#include <fileformatfactory.hh>
#include <surface.hh>
#include <surfaceio.hh>

#include "propcalcfactory.hh"
#include "propcalculator.hh"
#include "dlloader.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
SURFACENS_USE
#endif 

//******************************************************************************
PropGenerator::PropGenerator()
//******************************************************************************
: surface_(0), 	surfaceio_(), properties_(), parameters_(), 
	surfacefile_(), outputfile_(), help_(false), rewrite_(false), status_(0)
{

}

//******************************************************************************
PropGenerator::~PropGenerator()
//******************************************************************************
{
	for(props_it p=properties_.begin(); p!=properties_.end(); ++p)
		if (*p)
		{
			delete *p;
			*p = 0;
		}

	if (surface_) delete surface_;
	if (surfaceio_) delete surfaceio_;
}

//******************************************************************************
bool PropGenerator::init(int argc, char** argv)
//******************************************************************************
{
	if (!parseOptions(argc, argv)) 
		return false;
	else if (help_)
		return true;
	else
		return readSurface();
}

//******************************************************************************
bool PropGenerator::operator()()
//******************************************************************************
{
	for(props_it p=properties_.begin();p!=properties_.end();++p)
	{
		PropCalculator& property = *(*p);
		property.init(surface_);

		if (!property())
			cerr << "Could not calculate the " << property.getShortName() 
					 << " property." << endl;

	}

	if (rewrite_)	writeSurface();

	if ((outputfile_!="") || (surfacefile_!="") || (!rewrite_))
	{
		if (outputfile_!="")
		{
			ofstream out(outputfile_.c_str());

			if (!out)
			{
				cerr << "Could not write surface data into " << outputfile_ << "!" 
						 << endl;
				status_ = 6;
			}
			else
			{
				writeData(out);
				out.close();
			}
		}
		else
			writeData(cout);
	}
}

//******************************************************************************
void PropGenerator::usage(ostream& out)
//******************************************************************************
{
  out << "Usage:" << endl
      << "  curvature [-d -h -r -i <INFILE> -o <OUTFILE> -f <FILE_FORMAT> ]"
      << " -D<PARAMETER>=<VALUE> <PROPERTIES>" << endl
      << "    -h     print this help message" << endl
      << "    -d     remove doublett points" << endl
			<< "    -r     rewrite the data augmented surface into the <INFILE>\n"
      << "    -i     the name of the input file or stdin if not specified\n"
      << "    -o     the name of the output file or stdout if not specified\n" 
      << "    -f     the format of the input file or (default = sld)\n" 
			<< "    -D     defines a named parameter\n"
      << endl;
	
	out << "  possible values for <FILE_FORMAT>: " << endl;
  for(FileFormatFactory::factoryIter f=FileFormatFactory::instance().begin();
      f!=FileFormatFactory::instance().end();++f)
    out << "    - " << f->first << endl;	

	out << "  properties that can be calculated" << endl;
  for(PropCalcFactory::factoryIter f=PropCalcFactory::instance().begin();
      f!=PropCalcFactory::instance().end();++f)
	{
		PropCalculator* property = 
			PropCalcFactory::instance().createClass(f->first);
    out << "    - " << f->first << " " << property->getDescription() << endl;
		
		delete property;
	}
}
	
//******************************************************************************
bool PropGenerator::addPropCalculator(const string& id)
//******************************************************************************
{
	PropCalcFactory& factory = PropCalcFactory::instance();

  if (!factory.hasClass(id))
  {
    cerr << id << " not known as a surface property!" << endl;

		status_ = 3;
    return false;
	}

	PropCalculator* prop = factory.createClass(id);
	for(parameters_cit p=parameters_.begin();p!=parameters_.end();++p)
		prop->setParameter(p->first, p->second);

	properties_.push_back(prop);
	
	return true;
}

//******************************************************************************
bool PropGenerator::parseOptions(int argc, char** argv)
//******************************************************************************
{
	extern char* optarg;
	extern int optopt;
  extern int optind;

	std::string format = "sld";
	bool checkdoubletts = false;

  bool error = false;
	
  int c;
  while (((c = getopt(argc, argv, "hdro:i:f:D:")) != -1) && (!error))
  {
    char buffer[200];
    switch(c)
    {
      case 'i' : error=!(sscanf(optarg, "%s", buffer)); surfacefile_ = buffer;
				break;
			case 'o' : error=!(sscanf(optarg, "%s", buffer)); outputfile_ = buffer;
				break;
			case 'f' : error=!(sscanf(optarg, "%s", buffer)); format = buffer;
				break;
			case 'D' : error=!readParameter(optarg);
				break;
			case 'h' : help_ = true;
				break;
			case 'd' : checkdoubletts = true;
				break;
			case 'r' : rewrite_ = true;
				break;
			case '?' : 
				cerr << "unknown option '" << optopt << "' continuing anyway!" << endl;
				break;
			case ':' : 
				cerr << "missing argument for " << optopt << "!" << endl;	error=true;
				break;
		}
	}

	//setup the file format
	error=!loadFormat(format, checkdoubletts); 

	//read the properties
	while ((optind<argc) && (!error)) 
		error =!addPropCalculator(argv[optind++]);

	if (error) status_ = 1;
	return !error;
}

//******************************************************************************
bool PropGenerator::loadFormat(const string& format, bool checkdoubletts)
//******************************************************************************
{
	FileFormatFactory& factory = FileFormatFactory::instance();

  if (!factory.hasClass(format))
  {
    cerr << format << " not known as surface file format!" << endl;

		status_ = 2;
    return false;
  }

  //retrieve the input and output classes
  surfaceio_  = factory.createClass(format);
	surfaceio_->enableDoublettDetection(checkdoubletts);

	return true;
}

//******************************************************************************
bool PropGenerator::readParameter(const char* specifier)
//******************************************************************************
{
	char* str = new char[strlen(specifier)];
	strcpy(str, specifier);

	char* p = strstr(str, "=");
	if (p)
	{
		*p=0;
		char* name  = str;
		char* value = p+1;
		parameters_[name] = value;

		return true;
	}
	else
		return false;

//	bool result = sscanf(specifier, "%s=%s", name, value);
}

//******************************************************************************
bool PropGenerator::readSurface()
//******************************************************************************
{
	string infile = surfacefile_;

  if (infile =="")
  {
    surface_ = surfaceio_->read(cin);
    infile = "stdin";
  }
  else
    surface_ = surfaceio_->read(infile);

  if (!surface_)
  {
    cerr << infile << " is not a valid input source !" << endl;
    status_ = 4;
		return false;
  }
	else
		return true;
}

//******************************************************************************
bool PropGenerator::writeSurface()
	//******************************************************************************
{
	string outfile = surfacefile_;
	bool result = true;

  if (outfile=="")
  {
    result = surfaceio_->write(cout, *surface_);
    outfile = "stdout";
  }
  else
    result = surfaceio_->write(outfile, *surface_);

	if (!result)
	{
    cerr << "The surface could not be written to " << outfile << "!" << endl;
    status_ = 5;
  }
	
	return result;
}
	
//******************************************************************************
void PropGenerator::writeData(ostream& out)
//******************************************************************************
{
	int c;
	char buffer[64];

	//headline
	for(c=0;c<surface_->getNColumns();++c)
	{
		sprintf(buffer, "%-12s ", surface_->getColumnName(c).c_str());
		out << buffer;
	}
	out << endl;

	//data
	for(SurfaceEntity::points_it p=surface_->begin();p!=surface_->end();++p)
	{
		for(c=0;c<surface_->getNColumns();++c)
		{
			sprintf(buffer, "%12.6f ", p->data()[c]);
			out << buffer;
		}
		out << endl;
	}
	out << endl;
}


//******************************************************************************
int main(int argc, char** argv)
//******************************************************************************
{
  DLLoader format_loader("SURFACEFORMATS");
  DLLoader prop_loader("SURFACEPROPERTIES");

	PropGenerator generator;

	if (generator.init(argc, argv))
		if (generator.wantHelp()) 
			generator.usage(cout);
		else
			generator();

	if ((generator.getStatus()>0) && (generator.getStatus()>=3))
		generator.usage(cerr);
	
	return generator.getStatus();	
}
