/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "primitives.hh"

#include <fileformatfactory.hh>
#include <surfaceio.hh>
#include <surface.hh>

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
#endif 

//******************************************************************************
void usage(ostream& out)
//******************************************************************************
{
  out << "Usage:" << endl
      << "  surf2dt [-d -h] -i <INFILE> -o <OUTFILE> -f <FILE_FORMAT>"
      << " <DIELECTRIC_1> <DIELECTRIC_2> " << endl
      << "    -h     print this help message" << endl
      << "    -d     remove doublett points" << endl
      << "    -i the name of the input file or stdin if not specified\n"
      << "    -o the name of the output file or stdout if not specified\n" 
      << "    -f the format of the input file or (default = sld)\n" 
      << endl
      << "  possible values for <FILE_FORMAT>: " << endl;

  for(FileFormatFactory::factoryIter f=FileFormatFactory::instance().begin();
      f!=FileFormatFactory::instance().end();++f)
    out << "    - " << f->first << endl;

	out << endl
			<< "  the two dielectric constants are 1.00 and 78.390 by default" << endl;
}

//******************************************************************************
bool getOptions(int argc, char** argv,
                bool& help, bool& doublettchecking,
                string& input_file, string& output_file, string& input_format,
								double& dielectric_1, double& dielectric_2)
//******************************************************************************
{
  extern char *optarg;
  extern int optind;
  
  bool error_flag = false;

  int c;
  while ((c = getopt(argc, argv, "o:i:f:dh")) != -1)
  {
    char buffer[200];
    switch(c)
    {
      case 'o' :
      {
				if (!sscanf(optarg, "%s", buffer))
				{
					error_flag = true;
					break;
				}
	
				output_file = buffer;
				break;
      }

      case 'i' :
      {
				if (!sscanf(optarg, "%s", buffer))
				{
					error_flag = true;
					break;
				}
	
				input_file = buffer;
				break;
      }
      
      case 'f' :
      {
				if (!sscanf(optarg, "%s", buffer))
				{
					error_flag = true;
					break;
				}
	
				input_format = buffer;
				break;
      }
      
      case 'd':
      {
				doublettchecking = true;
				break;
      }

      case 'h':
      {
				help = true;
				break;
      }

      case '?':
				error_flag = true;
    }
  }

//read format arguments

  if (optind<argc)
    error_flag = !sscanf(argv[optind++], "%lf", &dielectric_1);
  
  if (optind<argc)
    error_flag = !sscanf(argv[optind++], "%lf", &dielectric_2);

  return !error_flag;
}


//******************************************************************************
int main (int argc, char** argv)
//******************************************************************************
{
  FileFormatFactory& factory = FileFormatFactory::instance();

  string infile = "";
  string outfile = "";
  string informat = "sld";
  bool help = false;
  bool doublettchecking = false;

	double dielectric_1 = 1.0;
	double dielectric_2 = 78.390;

  
  if (!getOptions(argc, argv, help, doublettchecking, 
                  infile, outfile, informat, dielectric_1, dielectric_2))
  {
    usage(cerr);
    return 1;
  }
  
  
  if (!factory.hasClass(informat))
  {
    cerr << informat << " not known as surface file format!" << endl;
    return 2;
  }

  if (help)
  {
    usage(cout);
    return 0;
  }

  //retrieve the input and output classes
  SurfaceIO* input  = factory.createClass(informat);
  input->enableDoublettDetection(doublettchecking);

  SurfaceEntity* surface = 0;

  //input
  if (infile=="")
  {
    surface = input->read(cin);
    infile = "stdin";
  }
  else
    surface = input->read(infile);
	
	delete input;

  if (!surface)
  {
    cerr << infile << " is not a valid input source for a surface in the " 
         << informat << " file format!" << endl;

    return 3;
  }

	Primitives* primitives = new Primitives(*surface, dielectric_1, dielectric_2);
	(*primitives)();

  //output 
  bool result = false;
  if (outfile=="")
  {
    result = primitives->write(cout);
    outfile = "stdout";
  }
  else
    result = primitives->write(outfile);

	delete primitives;
	delete surface;

  if (!result)
	{
    cerr << "The surface could not be written to " << outfile << "!" << endl;
    return 4;
  }
	else
		return 0;
}
