/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <iostream>
#  include <iomanip>
#else
#  include <iostream.h>
#endif
#include <string>

#include <fileformatfactory.hh>
#include <surfaceio.hh>
#include <surface.hh>

#include "dlloader.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace Surface;
#endif 

//******************************************************************************
void usage(ostream& out)
//******************************************************************************
{
  out << "Usage:" << endl
      << "  surfconvert [-d -h] -i <INFILE> -o <OUTFILE> "
      << "<INFORMAT> <OUTFORMAT>" << endl
      << "    -h     print this help message" << endl
      << "    -d     remove doublett points" << endl
      << "    -i the name of the input file or stdin if not specified\n"
      << "    -o the name of the output file or stdout if not specified\n" 
      << endl
      << "  possible values for <INFILE> and <OUTFILE>: " << endl;

  for(FileFormatFactory::factoryIter f=FileFormatFactory::instance().begin();
      f!=FileFormatFactory::instance().end();++f)
    out << "    - " << left << setw(10) << f->first << f->second.description << endl;
}

//******************************************************************************
bool getOptions(int argc, char** argv,
                bool help, bool& doublettchecking,
                string& input_file, string& output_file, 
                string& input_format, string& output_format)
//******************************************************************************
{
  extern char *optarg;
  extern int optind;
  
  bool error_flag = false;

  int c;
  while ((c = getopt(argc, argv, "o:i:dh")) != -1)
  {
    char buffer[200];
    switch(c)
    {
      case 'o' :
      {
				if (!sscanf(optarg, "%s", buffer))
				{
					error_flag = true;
					break;
				}
	
				output_file = buffer;
				break;
      }

      case 'i' :
      {
				if (!sscanf(optarg, "%s", buffer))
				{
					error_flag = true;
					break;
				}
	
				input_file = buffer;
				break;
      }
      
      case 'd':
      {
				doublettchecking = true;
				break;
      }

      case 'h':
      {
				help = true;
				break;
      }

      case '?':
				error_flag = true;
    }
  }

//read format arguments

  if (optind<argc)
    input_format = argv[optind++];
  
  if (optind<argc)
    output_format = argv[optind++];

  return !error_flag;
}

//******************************************************************************
int main (int argc, char** argv)
//******************************************************************************
{
  DLLoader loader("SURFACEFORMATS");

  FileFormatFactory& factory = FileFormatFactory::instance();

  string infile = "";
  string outfile = "";
  string informat = "";
  string outformat = "";

  bool help = false;
  bool doublettchecking = false;

  
  if (!getOptions(argc, argv, help, doublettchecking, 
                  infile, outfile, informat, outformat))
  {
    usage(cerr);
    return 1;
  }
  
  if ((informat=="") || (outformat==""))
  {
    cerr << "You  must specify an input and output file format!" << endl;
    usage(cerr);
    return 2;
  }
  
  if (!factory.hasClass(informat))
  {
    cerr << informat << " not known as surface file format!" << endl;
    return 3;
  }

  if (!factory.hasClass(outformat))
  {
    cerr << informat << " not known as surface file format!" << endl;
    return 4;
  }

  if (help)
  {
    usage(cout);
    return 0;
  }

	if (doublettchecking)
		cout << " DD doublett checking" << endl;

  //retrieve the input and output classes
  SurfaceIO* input  = factory.createClass(informat);
  SurfaceIO* output = factory.createClass(outformat);
  input->enableDoublettDetection(doublettchecking);

  SurfaceEntity* surface = 0;

  //input
  if (infile=="")
  {
    surface = input->read(cin);
    infile = "stdin";
  }
  else
    surface = input->read(infile);

  if (!surface)
  {
    cerr << infile << " is not a valid input source for a surface in the " 
         << informat << " file format!" << endl;

    delete input;
    delete output;
    return 5;
  }

  //output 
  bool result = false;
  if (outfile=="")
  {
    result = output->write(cout, *surface);
    outfile = "stdout";
  }
  else
    result = output->write(outfile, *surface);
  
  if (!result)
  {
    cerr << "The surface could not be written to " << outfile << "!" << endl;

    delete input;
    delete output;
    delete surface;
    return 6;
  }

  delete input;
  delete output;
  delete surface;
  return 0;
}
