/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef ASSOCGRAPH_HH
#define ASSOCGRAPH_HH

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map.hpp>

#include <trimesh_traits.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <utility>
#    include <iostream>
#  else
#    include <pair.h>
#    include <iostream.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <utility>
#  include <iostream>
#endif //HAVE_CONFIG_H

#ifdef HAVE_NAMESPACES
namespace boost
{
#endif
  enum vertex_pair_t     { vertex_pair    = 901 };
  enum vertex_KKS_t      { vertex_KKS     = 902 };
  enum vertex_Rfunc_t    { vertex_Rfunc   = 903 };
	enum vertex_angle_t    { vertex_angle   = 904 };
	enum vertex_zero_t     { vertex_zero    = 905 };
	enum edge_distances_t  { edge_distances = 906 };
	enum edge_axisA_t      { edge_axisA     = 907 };
	enum edge_axisB_t      { edge_axisB     = 908 };
	enum edge_angles_t     { edge_angles    = 909 };
	enum edge_Rs_t         { edge_Rs        = 910 };

  BOOST_INSTALL_PROPERTY(vertex, pair);
  BOOST_INSTALL_PROPERTY(vertex, KKS);
  BOOST_INSTALL_PROPERTY(vertex, Rfunc);
	BOOST_INSTALL_PROPERTY(vertex, angle);
  BOOST_INSTALL_PROPERTY(vertex, zero);
	BOOST_INSTALL_PROPERTY(edge,   distances);
	BOOST_INSTALL_PROPERTY(edge,   axisA);
	BOOST_INSTALL_PROPERTY(edge,   axisB);
	BOOST_INSTALL_PROPERTY(edge,   angles);
	BOOST_INSTALL_PROPERTY(edge,   Rs);
#ifdef HAVE_NAMESPACES
};
#endif
using namespace boost;

typedef pair<double, double> dblpair_t;

class assocgraph_traits
{
public:
	
	//*****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//*****************************************************************************

	/**
	 * \brief shortcut to the trimesh_traits::Vertex definition
	 */
	typedef trimesh_traits::vertex_t svertex_t;

	/**
	 * \brief pair of trimesh vertices
	 */
	typedef pair<svertex_t, svertex_t> meshvertexpair_t;

	//*****************************************************************************
	// VERTEX PROPERTY DEFINITIONS
	//*****************************************************************************
	
	/**
   * \brief the pair of matching extremes on A and B
   *
	 * The pair of matching extremes is represented by a std::pair of trimesh
	 * vertex descriptors. The first representing the vertex on surface A and the
	 * second, representing the vertex on surface B
   */
  typedef property<vertex_pair_t, meshvertexpair_t> pair_prop;

	/**
	 * \brief the similarity value from Kuiper-Kolmogorov-Smirnov's statistics
	 */
	typedef property<vertex_KKS_t, double, pair_prop> kks_prop;
	
	/**
	 * \brief the angular function of correlation between A and B
	 *
	 * The angular correlation function is obtained in a complete HI scan between
	 * both patches around vertex A and B and is stored as a vector of R-values
	 * associated with distinct rotations.
	 */
	typedef property<vertex_Rfunc_t, double*, kks_prop> rfunc_prop;

	/**
	 * \brief stores the angle of maximum correlation between the patches of the
	 * matching point pair
	 */
	typedef property<vertex_angle_t, double, rfunc_prop> angle_prop;

	/**
	 * \brief the border points of the patches that got mapped to r=1 and phi=0
	 */
	typedef property<vertex_zero_t, meshvertexpair_t, angle_prop> zero_prop;
	
	/**
	 * \brief a vertex color for each association vertex
	 */
	typedef property<vertex_color_t, int, zero_prop> color_prop;
	
	/**
	 * \brief a vertex index for each association vertex
	 */
	typedef property<vertex_index_t, int, color_prop> index_prop;

	//*****************************************************************************
	// EDGE PROPERTIES
	//*****************************************************************************
	
	/**
	 * \brief a pair of euclidian distances between the vertices on A and B
	 */
	typedef property<edge_distances_t, dblpair_t> dist_prop;

	/**
	 * \brief the points defining the axis between A1 and A2.
	 *
	 * The points are defined as follows: p(A1) is the point on patch A1 on the
	 * axis A1-A2 between the center of A1 and A2. p(A2) is that same point on A2.
	 */
	typedef property<edge_axisA_t, meshvertexpair_t, dist_prop> axisA_prop;

	/**
	 * \brief the points defining the axis between B1 and B2.
	 *
	 * The points are defined as follows: p(B1) is the point on patch B1 on the
	 * axis B1-B2 between the center of B1 and B2. p(B2) is that same point on B2.
	 */
	typedef property<edge_axisB_t, meshvertexpair_t, axisA_prop> axisB_prop;
	
	/**
	 * \brief the angles of optimal overlap on pair 1 and 2
	 *
	 * The angles of optimal overlap for pair 1 and 2 represent the rotation
	 * angles of B1 and B2 with optimal overlap (correlation coefficient) against
	 * A1 and A2 around the axis 1-2.
	 */
	typedef property<edge_angles_t, dblpair_t, axisB_prop> angles_prop;
	
	/**
	 * \brief the correlation coefficients corresponding to the angles of optimal
	 * overlap.
	 */
	typedef property<edge_Rs_t, dblpair_t, angles_prop> rs_prop;

	/**
	 * \brief an edge-weight property for the quartetts
	 */
	typedef property<edge_weight_t, int, rs_prop> weight_prop;
	
	//*****************************************************************************
	// GRAPH TYPE DEFINITION
	//*****************************************************************************

	/**
	 * \brief the association graph type
	 *
	 * The association graph stores the vertices (matching point pairs) and the
	 * edges (pairs of matching point pairs) in linked lists. Thus it is fast to
	 * add and remove vertices and edges. The graph is undirected and contains the 
	 * above specified set of vertex and edge properties.
	 */
	typedef adjacency_list<listS, listS, undirectedS, index_prop, weight_prop> 
	  graph_t;	
	
	//*****************************************************************************
	// PROPERTY MAP TYPE DEFINITION
	//*****************************************************************************

	/**
	 * \brief property map for matching point pairs
	 */
	typedef property_map<graph_t, vertex_pair_t>::type pairmap_t;
	
	/**
	 * \brief property map for KKS similarity meassures
	 */
	typedef property_map<graph_t, vertex_KKS_t>::type kksmap_t;
	
	/**
	 * \brief property map for the correlation functions
	 */
	typedef property_map<graph_t, vertex_Rfunc_t>::type rfuncmap_t;

	/**
	 * \brief property map for the correlation functions
	 */
	typedef property_map<graph_t, vertex_angle_t>::type anglemap_t;

	/**
	 * \brief property map for the zero vertex pair
	 */
	typedef property_map<graph_t, vertex_zero_t>::type zeromap_t;

	/**
	 * \brief property map for the vertex colors
	 */
	typedef property_map<graph_t, vertex_color_t>::type colormap_t;
	
	/**
	 * \brief property map for the vertex indices
	 */
	typedef property_map<graph_t, vertex_index_t>::type indexmap_t;

	/**
	 * \brief property map for the pair distances
	 */
	typedef property_map<graph_t, edge_distances_t>::type distancesmap_t;
	
	/**
	 * \brief property map for the axis points on A1, A2
	 */
	typedef property_map<graph_t, edge_axisA_t>::type axisAmap_t;

	/**
	 * \brief property map for the axis points on B1, B2
	 */
	typedef property_map<graph_t, edge_axisB_t>::type axisBmap_t;

	/**
	 * \brief property map for the optimal angles
	 */
	typedef property_map<graph_t, edge_angles_t>::type anglesmap_t;
	
	/**
	 * \brief property map for the optimal correlation coefficients
	 */
	typedef property_map<graph_t, edge_Rs_t>::type rsmap_t;

	//*****************************************************************************
	// VERTEX, EDGE descriptors and iterators
	//*****************************************************************************

  /**
   * \brief typedefinition of the vertex descriptor
   */
  typedef graph_traits<graph_t>::vertex_descriptor vertex_t;

  /**
   * \brief typedefinition of the edge descriptor
   */
  typedef graph_traits<graph_t>::edge_descriptor edge_t;

  /**
   * \brief iterator type for the vertices
   */
  typedef graph_traits<graph_t>::vertex_iterator vertex_it;

	/**
	 * \brief pair of vertex iterators
	 */
	typedef pair<vertex_it, vertex_it> vertices_t;

  /**
   * \brief iterator type for the edges
   */
  typedef graph_traits<graph_t>::edge_iterator edge_it;
  
	/**
	 * \brief pair of edge iterators
	 */
	typedef pair<edge_it, edge_it> edges_t;

  /**
   * \brief out-edge iterator type
   */
  typedef graph_traits<graph_t>::out_edge_iterator edge_oit;

  /**
   * \brief adjancency iterator
   */
  typedef graph_traits<graph_t>::adjacency_iterator vertex_ait;

	/**
	 * \brief pair of adjacency iterators
	 */
	typedef pair<vertex_ait, vertex_ait> adjacency_t;
	
};

typedef assocgraph_traits::graph_t assocgraph_t;

void writegraph(ostream& stream, assocgraph_t& graph, const char* title);

assocgraph_traits::vertex_t createVertex(assocgraph_t& graph);

void removeVertex(assocgraph_traits::vertex_t v, assocgraph_t& graph);
void clearGraph(assocgraph_t& graph);

#endif ASSOCGRAPH_HH
