/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "cluster.hh"

#include <Vector.hh>

#include <bestrot.h>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLStringTokenizer.hpp>

#include <float.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
# endif

#ifdef HAVE_STD
#  include <string>
#  include <sstream>
#  include <iterator>
#else
#  include <string.h>
#  include <sstream.h>
#  include <iterator.h>
#endif

#include "utils.hh"
#include <xmlutils.hh>

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace RazorBack;
XERCESNS_USE
#endif

//*******************************************************************************
template <class It>
size_t clamp_distance(It begin, It end)
//*******************************************************************************
{
	size_t c = 0;

  double max_value = DBL_MIN;
	double min_value = DBL_MAX;
	
	for(It i=begin;i!=end;++i)
	{
	  if (*i > max_value) max_value = *i;
		if (*i < min_value) min_value = *i;
	}

	for(It i=begin;i!=end;++i, ++c)
		*i = 1 - (*i - min_value) / (max_value - min_value);

	return c;
}

//*******************************************************************************
void calcTransformation(Rectmat_<double>& transform, const Matrect_& rotation, 
												const Vect_& centerA, Vect_ centerB)
//*******************************************************************************
{
	/*
		The complete transformation matrix has the form:
		
		r11 r12 r13 (r1 * vB0) + xBA
		r21 r22 r23 (r2 * vB0) + yBA
		r31 r32 r33 (r3 * vB0) + zBA
		 0   0   0         1

		vB0 = -centerB
    vBA = centerA
		r   = rotation
	 */	
	
	//inverse the centerB
	vect_scalar(&centerB, '*', -1.0);
	
	//set the transformation matrix 4x4
	for(int i=0; i<3; ++i)
	{
		double scalar = 0.0;
		for(int j=0;j<3;++j)
		{
			transform(i,j) = rotation.Mat[i][j];
			//scalar product  r[i] * vB0
			scalar += rotation.Mat[i][j] * centerB.Vec[j];
		}

		transform(i,3) = scalar + centerA.Vec[i];
	}
	
	//forth row
	for(int j=0;j<3;++j) transform(3,j) = 0;
	transform(3,3) = 1;	
}

//*******************************************************************************
Cluster::Cluster()
//*******************************************************************************
: vPatchPairs(), lCorresponds(), dTransformation(4,4), dRMSD(-1), dNPoints(0)
{
	pSurfaces[0] = pSurfaces[1] = 0;
}
	
//*******************************************************************************
Cluster::Cluster(Trimesh** surfaces,const size_t patchpairs)
//*******************************************************************************
: vPatchPairs(), lCorresponds(), dTransformation(4,4), dRMSD(-1), dNPoints(0)
{
	init(surfaces,patchpairs);
}

//*******************************************************************************
void Cluster::init(Trimesh** surfaces, const size_t patchpairs)
//*******************************************************************************
{
	pSurfaces[0] = surfaces[0];
	pSurfaces[1] = surfaces[1];
	if (patchpairs>0)
		vPatchPairs.reserve(patchpairs);
}


//*******************************************************************************
void Cluster::addPair(const Trimesh::vertex_t vA, const Trimesh::vertex_t vB, 
											Trimesh& mapA, Trimesh& mapB, const double R, 
											const double angle)
//*******************************************************************************
{
	//initialize the new patchpair
	vPatchPairs.push_back(patchpair_t(vA, vB, R, angle));
  int ix = vPatchPairs.size() - 1;

	addCorrespondencies(ix, mapA, mapB, angle);
}

//*******************************************************************************
void Cluster::addCorrespondencies(int ix, Trimesh& mapA, Trimesh& mapB, 
																	const double angle)
//*******************************************************************************
{
	if ((ix>=0) && (ix<vPatchPairs.size()))
	{
		//initialize the correspondent object
		Correspondent correspondent;
		correspondent.addCorrespondencies(mapA, mapB, angle);
		
		vPatchPairs[ix].corresponds.clear();
		correspondent(back_inserter(vPatchPairs[ix].corresponds));
	}
}

//*******************************************************************************
void Cluster::addCorrespondencies(pairs_it p, Trimesh& mapA, Trimesh& mapB, 
																	const double angle)
//*******************************************************************************
{
	if (p!=vPatchPairs.end())
	{
		//initialize the correspondent object
		Correspondent correspondent;
		correspondent.addCorrespondencies(mapA, mapB, angle);
		
		p->corresponds.clear();
		correspondent(back_inserter(p->corresponds));
	}
}

//*******************************************************************************
size_t Cluster::merge(const Cluster& A, const Cluster& B)
//*******************************************************************************
{
	typedef pair<Trimesh::vertex_t, Trimesh::vertex_t> simple_t;
	set<simple_t> checkset;

	//reset the patchpairs vector
	vPatchPairs.clear();
	vPatchPairs.reserve(A.vPatchPairs.size() + B.vPatchPairs.size());

	for(pairs_cit p=A.beginPairs();p!=A.endPairs();++p)
	{
		//insert the pair into the check set and add the pair to the patch pairs
		checkset.insert(simple_t(p->A, p->B));
		vPatchPairs.push_back(*p);
	}

	for(pairs_cit p=B.beginPairs();p!=B.endPairs();++p)
	{
		set<simple_t>::iterator pos = checkset.find(simple_t(p->A, p->B));
		
		//insert the pair into the patch pairs only if it is new
		if (pos==checkset.end())
			vPatchPairs.push_back(*p);
	}

	return vPatchPairs.size();
}
	
//*******************************************************************************
void Cluster::compCorresponds() const
//*******************************************************************************
{
	Correspondent correspondent;
	vector<patchpair_t>::const_iterator i;
	
	//add all correspondence pairs to the correspondent object
	for(i=vPatchPairs.begin();i!=vPatchPairs.end();++i)
		correspondent.addCorrespondencies(i->corresponds.begin(),
																			i->corresponds.end());
	
	lCorresponds.clear();
	correspondent(back_inserter(lCorresponds));

	dNPoints = lCorresponds.size();
}

//*******************************************************************************
double Cluster::doRMSFit(const bool weight)
//*******************************************************************************
{
	/* INITIALIZATION OF ROTATION RELEVANT MATRICES*/
	int i,j;

	int points = lCorresponds.size();

	//point matrices;
	Matrect_ rmscoords[2];
	init_matrect(&rmscoords[0], points, 3);
	init_matrect(&rmscoords[1], points, 3);
	
	//rotation matrix
	Matrect_ rotation;
	init_matrect(&rotation, 3, 3);	
	
	//centers
	Vect_ center[2];
	for(i=0;i<2;++i)
		init_vect(&center[i], 3);

	//BESTROT Weights
	Vect_ weights;
	init_vect(&weights, points);

	//assign the points from the different correspondent point pairs
	i=0;
	for(corresponds_cit c=lCorresponds.begin();c!=lCorresponds.end();++c, ++i)
	{
		const Vector_<double>& posA = pSurfaces[0]->vertCoordinates(c->first);
		const Vector_<double>& posB = pSurfaces[1]->vertCoordinates(c->second);
		
		for(j=0;j<3;++j)
		{
			rmscoords[0].Mat[i][j] = posA[j];
			rmscoords[1].Mat[i][j] = posB[j];
		}
		
		//the weights are the distances on the harmonic maps between the
		//correspondent vertices
		weights.Vec[i] = c->distance;
	}

	//calculate the centers
	center_vectors(&rmscoords[0], &center[0], 0);
	center_vectors(&rmscoords[1], &center[1], 0);

	//rescale the weights to get 1 for the shortest and 0 for the most distant pair
	clamp_distance(weights.Vec, weights.Vec + weights.Len);

	//assign the rmsWeights to the correspondent point pairs;
	i=0;
	for(corresponds_it c=lCorresponds.begin();c!=lCorresponds.end();++c, ++i)
		c->rmsWeight = weights.Vec[i];
	
	/* COMPUTATION */

	//transformation
	dRMSD = best_rot(&rmscoords[0], &rmscoords[1], 
									 (weight) ? &weights : 0, &rotation);
	
	//on error RMSD = -1
	if ((dRMSD + 1.0)>DBL_EPSILON)
		calcTransformation(dTransformation, rotation, center[0], center[1]);

	free_vect(&weights);
	free_matrect(&rmscoords[0]);
	free_matrect(&rmscoords[1]);
	free_matrect(&rotation);	
	
	return dRMSD;
}	

//*******************************************************************************
double Cluster::calcRMSD(const Rectmat_<double>& trans, const bool weight) const
//*******************************************************************************
{
	double msd = 0.0;
	double weightsum = 0.0;
	
	for(corresponds_cit c=lCorresponds.begin();c!=lCorresponds.end();++c)
	{
		//retrieve the reference to the point on surface A

		//point on B will be transformed through a 4x4 matrix -> must be copied into
		//a 4 element vector; A for calculations
		Vector_<double> posA(4,1.0);
		Vector_<double> posB(4,1.0);
		
		const Vector_<double>& vertexA = pSurfaces[0]->vertCoordinates(c->first);
		const Vector_<double>& vertexB = pSurfaces[1]->vertCoordinates(c->second);
		
		for(int i=0;i<3;++i)
		{
			posA[i] = vertexA[i];
			posB[i] = vertexB[i];
		}
		
		//transformation and calculation of difference vector in A
	  posA -= trans * posB;

		msd += (weight) ? c->rmsWeight * posA.vec_len2() : posA.vec_len2();
		weightsum += (weight) ? c->rmsWeight : 1.0;
	}
	
	msd /= weightsum;
	
	return sqrt(msd);
}

//*******************************************************************************
double Cluster::getScore(const string name) const
//*******************************************************************************
{
	scoringmap_cit it = mScores.find(name);
	if (it!=mScores.end())
		return it->second;
	else
		return -DBL_MAX;
}

//*******************************************************************************
double Cluster::setScore(const string name, double score)
//*******************************************************************************
{
	double old = -DBL_MAX;

	scoringmap_it it = mScores.find(name);
	if (it!=mScores.end())	
	{
		old = it->second;
		it->second = score;
	}
	else
		mScores[name] = score;
	
	return old;
}

//*******************************************************************************
// XML WRITING ROUTINES
//*******************************************************************************

//*******************************************************************************
bool writePairToDOM(DOMElement* node, const patchpair_t& p)
//*******************************************************************************
{
	DOMElement*   patchpair = XMLUtils::createElement(node, "patchpair");
	//A attribute
  patchpair->setAttribute(XMLStr("vertex_a"), XMLStr().fromValue((int)p.A));
  patchpair->setAttribute(XMLStr("vertex_b"), XMLStr().fromValue((int)p.B));
  patchpair->setAttribute(XMLStr("r"), XMLStr().fromValue(p.R));
  patchpair->setAttribute(XMLStr("angle"), XMLStr().fromValue(p.angle));
}

//*******************************************************************************
bool writeScoreToDOM(DOMElement* node, Cluster::scoringmap_cit& score)
//*******************************************************************************
{
	DOMElement* scorenode = XMLUtils::createTextNode(node, "score", 
																									 ftoa(score->second));
	scorenode->setAttribute(XMLStr("name"), XMLStr(score->first));
}

//*******************************************************************************
DOMElement* Cluster::writeToDOM(DOMElement* node, const int experiment) const
//*******************************************************************************
{
	/* INITIALIZATION */
	string transformation;
	int i, j;

	/* CLUSTER */

	//create the cluster node + attributes
	DOMElement*   clusternode = XMLUtils::createElement(node, "cluster");

	//experiment attribute
	if (experiment>0)
		clusternode->setAttribute(XMLStr("experiment"),	
															XMLStr().fromValue(experiment));
	//RMSD Attribute
	clusternode->setAttribute(XMLStr("rmsd"),	XMLStr().fromValue(dRMSD));
	//Correspondent counts
	clusternode->setAttribute(XMLStr("points"),	
														XMLStr().fromValue((int)dNPoints));
	//Patchpair counts
	clusternode->setAttribute(XMLStr("pairs"), 
														XMLStr().fromValue((int)vPatchPairs.size()));

	/* TRANSFORMATION NODE */
	
	//create the transformation node string
	for(i=0;i<4;++i)
		for(j=0;j<4;++j)
		{
			//add a space
			transformation += " ";
			//add the transformation matrix entry (i,j)
			transformation += ftoa(dTransformation(i,j));
		}
	XMLUtils::createTextNode(clusternode, "transformation", transformation);

	/* SCORES */

	//create the list node + size attribute
	DOMElement* scores = XMLUtils::createElement(clusternode, "scores");

	//size attribute
	scores->setAttribute(XMLStr("size"), XMLStr().fromValue((int)mScores.size()));

	//add single scores
	for(scoringmap_cit s=mScores.begin();s!=mScores.end();++s)
		writeScoreToDOM(scores, s);
	
  /* PATCHPAIRS */

	//create the list node + size attribute
	DOMElement* patchpairs = XMLUtils::createElement(clusternode, "patchpairs");
	//clusternode->appendChild(patchpairs);

	//size attribute
	patchpairs->setAttribute(XMLStr("size"),
													 XMLStr().fromValue((int)vPatchPairs.size()));

	//add the single patchpairs
	for(vector<patchpair_t>::const_iterator p = vPatchPairs.begin();
			p!=vPatchPairs.end();++p)
		writePairToDOM(patchpairs, *p);

	return clusternode;
}

//*******************************************************************************
// XML READING ROUTINES
//*******************************************************************************

//*******************************************************************************
bool readPairFromDOM(const DOMElement* node, patchpair_t& p)
//*******************************************************************************
{
	string value;
	
	if (!(XMLStr(node->getTagName()) == XMLStr("patchpair")))
		return false;
	
	/* PATCHPAIR ATTRIBUTES */

	//A attribute
	p.A     = (int)XMLUtils::getAttribute(node, "vertex_a");
	p.B     = (int)XMLUtils::getAttribute(node, "vertex_b");
	p.R     = XMLUtils::getAttribute(node, "r");
	p.angle = XMLUtils::getAttribute(node, "angle");
}

//*******************************************************************************
pair<string, double> readScoreFromDOM(const DOMElement* node)
//*******************************************************************************
{
	if (XMLStr(node->getTagName()) == XMLStr("score"))
		return make_pair((string)XMLUtils::getAttribute(node, "name"), 
										 (double)XMLUtils::getText(node));
	else
		return make_pair("", -DBL_MAX);
}

//*******************************************************************************
bool Cluster::readFromDOM(const DOMElement* node)
//*******************************************************************************
{
	if (!(XMLStr(node->getTagName()) == XMLStr("cluster")))
		return false;

	/* CLUSTER ATTRIBUTES */
	dRMSD = XMLUtils::getAttribute(node, "rmsd");
	dNPoints = (int)XMLUtils::getAttribute(node, "points");

	//parse the child nodes
	const DOMNode* child = node->getFirstChild();
	while (child)
	{
		/* CHECK FOR TRANSFORMATION ************************************************/

		if (const DOMElement* elem=XMLUtils::elementByName(child, "transformation"))
		{
			stringstream matrix(XMLUtils::getText(elem));
			int i = 0;
			while ((!matrix.eof()) && (i<16))
			{
				matrix >> dTransformation(i/4, i%4);
				++i;
			}
		}

		/* CHECK FOR SCORES  ***************************************************/
		
		if  (const DOMElement* scores=XMLUtils::elementByName(child, "scores"))
		{
			const DOMNode* grandchild = scores->getFirstChild();
			while (grandchild)
			{
				if (const DOMElement* score=XMLUtils::elementByName(grandchild,"score"))
					mScores.insert(readScoreFromDOM(score));

				grandchild = grandchild->getNextSibling();
			}
		}

		/* CHECK FOR PATCHPAIRS  ***************************************************/
		
		if  (const DOMElement* pairs=XMLUtils::elementByName(child, "patchpairs"))
		{
			int npairs = XMLUtils::getAttribute(pairs, "size");

      //resize the vPatchPairs vector
			vPatchPairs.clear();
			vPatchPairs.resize(npairs);

			const DOMNode* grandchild = pairs->getFirstChild();
			vector<patchpair_t>::iterator p= vPatchPairs.begin();
			while (grandchild)
			{
				if (const DOMElement* pair=
						XMLUtils::elementByName(grandchild, "patchpair"))
				{
					readPairFromDOM(pair,*p);
					++p;
				}
				grandchild = grandchild->getNextSibling();
			}
		}

		//ITERATION
		child = child->getNextSibling();
	}
	
	return true;
}
