/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CLUSTER_HH
#define CLUSTER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif 

#ifdef HAVE_STD
#  include <vector>
#  include <map>
#  include <string>
#else
#  include <vector.h>
#  include <map.h>
#  include <string.h>
#endif

#include <Rectmat.hh>
#include <trimesh.hh>

#include "correspondent.hh"
#include "xercesns.hh"

//forward declaration of the DOMElement class
XERCES_FORWARD(DOMElement);
XERCES_FORWARD(DOMDocument);

/**
 * \brief the representation of a patch pair in the cluster
 */
struct patchpair_t
{
	Trimesh::vertex_t A, B;
	double R;
	double angle;
	corresponds_t corresponds;

	patchpair_t() : A(0), B(0), R(-1), angle(0), corresponds() { }
	patchpair_t(Trimesh::vertex_t _A, Trimesh::vertex_t _B, double _R, double _angle) 
			: A(_A), B(_B), R(_R), angle(_angle), corresponds() { }

	patchpair_t(const patchpair_t& src)
			: A(src.A), B(src.B) , R(src.R), angle(src.angle), 
				corresponds(src.corresponds) { }

	bool operator<(const patchpair_t& other)
	{
		return (A==other.A) ? (B<other.B) : (A<other.A);
	}
};

class Cluster
{

public:

	//*****************************************************************************
	// PUBLIC TYPEDEFS
	//*****************************************************************************

	/**
	 * \brief iterator type for the patchpair sequence
	 */
	typedef vector<patchpair_t>::const_iterator pairs_cit;
	typedef vector<patchpair_t>::iterator pairs_it;

	/**
	 * \brief the type of the scoring map plus iterators
	 */
	typedef std::map<std::string, double> scoringmap_t;
	typedef scoringmap_t::iterator scoringmap_it;
	typedef scoringmap_t::const_iterator scoringmap_cit;

	//*****************************************************************************
	// CONSTRUCTOR AND INITIALIZATION
	//*****************************************************************************
	
	/**
	 * \brief standard constructor
	 *
	 * The standard constructor initializes vPatchPairs and lCorresponds with
	 * standard values, sets dTransformation to a 4x4 matrix and assigns NULL
	 * to the surface pointers.
	 *
	 * \sa Cluster(Trimesh**)
	 */
	Cluster();
	
	/**
	 * \brief initializing constructor
	 *
	 * The initializing constructor performs the same initializiation as Cluster()
	 * and calls init(Trimesh**) to set the surface pointers.
	 *
	 * \param surfaces   a two membered field containing pointers to both surfaces
	 * \param patchpairs number of pairs to be inserted into the cluster
	 *
	 */
	Cluster(Trimesh** surfaces, const size_t patchpairs = 0);

	/**
	 * \brief initiaiziation
	 *
	 * The init routine sets the surface pointers to the passed values
	 *
	 * \param surfaces a two membered field containing pointers to both surfaces
	 */
	void init(Trimesh** surfaces, const size_t patchpairs = 0);

	/**
	 * \brief adds a pair of corresponding patches to the cluster
	 *
	 * This method adds a matching point pair to the cluster. To add a matching
	 * point pair, you need to pass the vertex numbers, the harmonic maps as well
	 * as the best correlation coefficient and the best overlapping angle to the
	 * method. It calculates the list of corresponding point pairs and adds an
	 * appropriate patchpair_t entry to the cluster.
	 *
	 * \param vA,vB     the identifiers of the central vertices on A and B
	 * \param mapA,mapB references to the harmonic maps on A and B respectively
	 * \param R         the best correlation coefficient for this patch pair
	 * \param angle     the angle of maximal overlap between A and B
	 */
	void addPair(const Trimesh::vertex_t vA, const Trimesh::vertex_t vB, 
							 Trimesh& mapA, Trimesh& mapB, const double R, const double angle);

	/**
	 * \brief adds the correspondent points to patchpair \p ix according to the
	 *        maps and angle
	 *
	 * This method adds the correspondencies that are the result of the overlap of
	 * \p mapA and \p mapB at \p angle. This method can be used to include
	 * correspondencies for existing patch pairs \p ix.
	 *
	 * \param ix        the index of the point pair
	 * \param mapA,mapB references to the harmonic maps on A and B respectively
	 * \param angle     the angle of maximal overlap between A and B
	 */
	void addCorrespondencies(int ix, Trimesh& mapA, Trimesh& mapB, 
													 const double angle);

	/**
	 * \brief adds the correspondent points to patchpair \p p according to the
	 *        maps and angle
	 *
	 * This method adds the correspondencies that are the result of the overlap of
	 * \p mapA and \p mapB at \p angle. This method can be used to include
	 * correspondencies for existing patch pairs \p ix.
	 *
	 * \param p         an iterator pointing to a valid point pair
	 * \param mapA,mapB references to the harmonic maps on A and B respectively
	 * \param angle     the angle of maximal overlap between A and B
	 */
	void addCorrespondencies(pairs_it p, Trimesh& mapA, Trimesh& mapB, 
													 const double angle);
	
	/**
	 * \brief uses the union of the patchpairs of the given clusters \p A, \p B to
	 * form a new merged cluster.
	 *
	 * The method takes the union (A + B) of the patchpairs and creates a new
	 * cluster. 
	 *
	 * \note No conflicts are checked within this method. So be sure to provide
	 * clusters with non-ambigous point pairs.
	 *
	 * \return number of patchpairs in the new merged cluster
	 */
	size_t merge(const Cluster& A, const Cluster& B);
	
	//*****************************************************************************
	// OPERATIONS ON THE DATA
	//*****************************************************************************
	
	/**
	 * \brief computes the correspondence table of the cluster 
	 *
	 * The correspondence table of the cluster is stored in lCorresponds
	 */
	void compCorresponds() const;
	
	/**
	 * \brief performs the rigid body transformation between A and B
	 *
	 * A transformation matrix is computed and stored in dTransformation so that
	 * the identified correspond. If you want to weight the correspondent pairs by
	 * their distance on the harmonic map overlap, set the flag \p weight to
	 * \c true
	 *
	 * \param weight specifies if the points are weighted by their map distance;
	 *               default is \c true
	 *
	 * \return the root mean square distance of the rigid body transformation.
	 */
	double doRMSFit(const bool weight = true );

	/**
	 * \brief calculates the RMSD for the given transformation matrix \p trans
	 *
	 * The method transforms each point (in the list) of B by \p trans and
	 * calculates the root mean square distances to its corresponding point
	 * on A. 
	 *
	 * \param trans a 4x4 transformation matrix
	 *
	 * \return the root mean square distance for the given transformation
	 */
	double calcRMSD(const Rectmat_<double>& trans, const bool weight = true) const;

	/**
	 * \brief retrieves the surface \p ix of the cluster
	 *
	 * \parem ix index of the relevant surace (0 = A, 1 = B)
	 *
	 * \return a pointer to a Trimesh* structure
	 */
	Trimesh* getSurface(const int ix) 
	{ return (ix==0) ? pSurfaces[0] : pSurfaces[1]; }

	/**
	 * \brief removes the list of corresponding point pairs
	 *
	 * This is a memory save method, that can be used in clustering algorithm to
	 * delete the corresponding point pairs if they are not needed any more.
	 */
	void clearCorresponds() { lCorresponds.clear(); } const

	//*****************************************************************************
	// PROPERTY METHODS
	//*****************************************************************************

	/**
	 * \brief states if the cluster has a filled list of correspondencies
	 *
	 * \return \c true if the cluster has correspondencies available
	 */
	bool hasCorrespondencies() const { return lCorresponds.size()>0; }

	corresponds_cit beginCorr() const { return lCorresponds.begin(); }
	corresponds_cit endCorr()   const { return lCorresponds.end(); }
	/**< 
	 * \brief provide constant access to the sequence of correspondence point
	 * pairs in the cluster
	 *
	 * \return a constant reference to the begin or end of the correspondent point
	 * pairs sequence respectively
	 */

	corresponds_it beginCorr() { return lCorresponds.begin(); }
	corresponds_it endCorr()   { return lCorresponds.end(); }
	/**< 
	 * \brief provide constant access to the sequence of correspondence point
	 * pairs in the cluster
	 *
	 * \return a reference to the begin or end of the correspondent point pairs
	 * sequence respectively
	 */

	/**
	 * \brief retrieves the number of corresponding point pairs in the cluster
	 *
	 * \return the size of the correspondence list lCorresponds
	 */
	size_t getNPoints() const { return dNPoints; }

	/**
	 * \brief retrieves the number of patch pairs building the cluster
	 *
	 * \return the number of patchpairs in the cluster
	 */
	size_t getNPairs() const { return vPatchPairs.size(); }

	/**
	 * \brief retrieves the number of scoring values stored for that cluster
	 *
	 * \return the number of scoring values stored in the mScores map
	 */
	size_t getNScores() const { return mScores.size(); }

	/**
	 * \name Pair iterators
	 */
	//@{
	/**<
	 * \brief provide access to the sequence of patchpairs 
	 *
	 * The methods return the begin or end of the sequence of patch pairs. The
	 * access is either constant or variable.
	 */
	pairs_cit	beginPairs() const { return vPatchPairs.begin(); }
	pairs_cit	endPairs()   const { return vPatchPairs.end(); }

	pairs_it	beginPairs() { return vPatchPairs.begin(); }
	pairs_it	endPairs()   { return vPatchPairs.end(); }
	//@}

	/**
	 * \brief retrieves the root mean square deviation of the cluster transformation
	 *
	 * The root mean square distance is meassured for all point pairs in the
	 * cluster's correspondence list after the rigid body transformation. It is
	 * also returned directly from the doRMSFit() method.
	 *
	 * \return the root mean square distance of the rigid body transformation.
	 */
	double getRMSD() const { return dRMSD; }

	/**
	 * \brief retrieves the transformation matrix of the rigid body transformation
	 *
	 * The transformation matrix is returned as a constant reference to the
	 * dTransformation Rectmat_.
	 *
	 * \return constant reference to the rigid body transformation matrix
	 */
	const Rectmat_<double>& getTransformation() const { return dTransformation; }

	/**
	 * \brief retrieves the scoring value named  \p name
	 *
	 * \param name the name of the scoring value
	 *
	 * \return the scoring value assigned to that name
	 *
	 * \sa mScores, setScore(const std::string, double)
	 */
	double getScore(const std::string name) const;
	/**
	 * \brief assigns a (new) value to the score \p name
	 *
	 * \param name  the name of the scoring value
	 * \param score the scoring value
	 *
	 * \return the old scoring value
	 *
	 * \sa mScores, getScore(const std::string)
	 */
	double setScore(const std::string name, double score);

	/**
	 * \name  Scoring iterators
	 */
  //@{
	/**
	 * \brief retrieves the beginning or end of the stored scoring (name/value)
	 * pairs
	 *
	 * There is either constant or non constant access.
	 */
	scoringmap_cit beginScores() const { return mScores.begin(); }
	scoringmap_it  beginScores()       { return mScores.begin(); }
	scoringmap_cit endScores()   const { return mScores.end(); }
	scoringmap_it  endScores()         { return mScores.end(); }
  //@}


	//*****************************************************************************
	// OUTPUT METHODS
	//*****************************************************************************

	/**
	 * \brief writes the cluster information as a new child node of the given DOM
	 * element
	 *
	 * \param node       pointer to the parent DOMElement
	 * \param experiment experiment ID in the database
	 *
	 * \return \c true if successful
	 */
	XERCESNS DOMElement* writeToDOM(XERCESNS DOMElement* node, 
																	const int experiment = 0) const;
 

	/**
	 * \brief reads the cluster information from the given node of the given DOM
	 * element
	 *
	 * \param node       pointer to the parent DOMElement
	 */
	bool readFromDOM(const XERCESNS DOMElement* node);

private:

	/**
	 * \brief a vector of patch pairs that represent the cluster
	 */
	vector<patchpair_t> vPatchPairs;

	/**
	 * \brief the correspondence table of the complete cluster
	 *
	 * This correspondence table is used to generate the rigid body transformation
	 * between both surfaces.
	 */
	mutable corresponds_t lCorresponds;
	
	/**
	 * \brief pointer to the surfaces of the comparison
	 */
	Trimesh* pSurfaces[2];

	/**
	 * \brief root mean square deviation of the rigid body transformation
	 */
	double dRMSD;

	/**
	 * \brief the number of corresponding point pairs
	 */
	mutable size_t dNPoints;
	
	/**
	 * \brief the transformation matrix for surface A to achieve the rigid body
	 * transformation onto B.
	 */
	Rectmat_<double> dTransformation;

	/**
	 * \brief the map of scoring values for that cluster
	 */
	 scoringmap_t mScores;
};

#endif //CLUSTER_HH
