/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "consensor.hh"

#include "math.h"

#ifdef HAVE_STD
#include <functional>
#include <algorithm>
#include <set>
#else
#include <function.h>
#include <algo.h>
#include <set.h>
#endif

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace RazorBack;
#endif


//*******************************************************************************
void Consensor::init(const Rectmat_<double>& data, 
										 const Vector_<double>& weights, bool autoscale)
//*******************************************************************************
{
	dData = data;

	vWeights = weights;
	vWeights.dim(dData.cno());

	fAutoScale = autoscale;
	
	vRanks.clear();
	vRanks.resize(dData.rno());
}

//*******************************************************************************
void Consensor::setData(const Rectmat_<double>& data)
//*******************************************************************************
{
	dData = data;
	vWeights.dim(dData.cno());
	
}

//*******************************************************************************
bool Consensor::setData(size_t row, size_t col, double value)
//*******************************************************************************
{
	if ((row>=0) && (row<dData.rno()) && (col>=0) && (col<dData.cno()))
	{
		dData(row,col) = value;
		return true;
	}
	else
		return false;
}	

//*******************************************************************************
void Consensor::redim(size_t rows, size_t cols)
//*******************************************************************************
{
	dData.size(rows,cols);
	vWeights.dim(cols);
}

//*******************************************************************************
void Consensor::setWeights(const Vector_<double>& weights)
//*******************************************************************************
{
	vWeights = weights;
	vWeights.dim(dData.cno());	
}	

//*******************************************************************************
bool Consensor::setWeight(size_t col, double weight)
//*******************************************************************************
{
	if ((col>=0) && (col<vWeights.dim()))
	{
		vWeights[col] = weight;
		return true;
	}
	else 
		return false;
}	

//*******************************************************************************
void Consensor::enableAutoScaling(bool enable)
//*******************************************************************************
{
	fAutoScale = enable;
}	

//*******************************************************************************
score_t Consensor::rowByRank(unsigned int rank) const
//*******************************************************************************
{
	if ((rank>0) & (rank<=vRanks.size()))
		return vRanks[rank-1];
	else
		return make_pair(0.0,-1);
}	
	
//*******************************************************************************
void Consensor::compConsensusScore()
//*******************************************************************************
{
	size_t n = dData.rno();	              //the number of rows is needed very often

	//calculate the autoscaled values (if demanded)
	if (fAutoScale)
	{

		for(int c=0;c<dData.cno();++c)
		{
			//calculate mean and standard deviation for each column
			double mean = 0.0;
			double sdev = 0.0;

			for(int r=0;r<n;++r)
			{
				double value = dData(r,c);
				mean += value/n;
				sdev += (value*value)/n;
			}

			sdev -= mean*mean;
			sdev  = sqrt(sdev);

			//autoscale the column
			for(int r=0;r<n;++r)
			{
				dData(r,c) -= mean;
				dData(r,c) /= sdev;
			}
		}
	}

	//perform the column weighting
	for(int c=0;c<dData.cno();++c)
		for(int r=0;r<n;++r)
			dData(r,c) *= vWeights[c];

	//clear the consensus ranking map and resize it
	vRanks.clear();
	vRanks.resize(dData.rno());

}	
	
//*******************************************************************************
// C L A S S :  R A N K B Y R A N K
//*******************************************************************************

//*******************************************************************************
void RankByRank::compConsensusScore()
//*******************************************************************************
{
	//this method needs no autoscaling 
	fAutoScale = false;
	//call the base class virtual function
	Consensor::compConsensusScore();
	
	/* RANKING:
	 *
	 * - Each column is ranked according to a descending order of its values 
	 *   -> the ranks are written back into the data matrix, destroying the 
   *      original values
	 * 
	 * - The average rank for each row is calculated and stored together with the
	 *   row indices in a list
	 *
	 * - The list is sorted and the rank/rowindex pairs are stored in the database
	 */

	typedef set<score_t, greater<score_t> > scoreset_t;
	typedef scoreset_t::const_iterator scoreset_cit;

	int r;

	//calculate the weight sum
	double wsum = 0.0;
	for(int v=0;v<vWeights.dim();++v)	wsum += fabs(vWeights[v]);
		

	/* RANK EACH COLUMN */
	for(int c=0;c<dData.cno();++c)
	{
		scoreset_t scores;

		//fill the set of value pairs (will be automatically sorted)
		for(r=0;r<dData.rno();++r)
			scores.insert(make_pair(dData(r,c), r));
		
		//assign the rank of each row to the corresponding cell
		//multiplied with the absolut weight for that column
		r=1;
		for(scoreset_cit s=scores.begin();s!=scores.end();++s, ++r)
			dData(s->second, c) = r * fabs(vWeights[c]);
	}

	/* CALCULATE THE CONSENSUS SCORING */
	scoreset_t consense;
	
	//build the average rank over all columns
	for(r=0;r<dData.rno();++r)
	{
		double meanrank = 0.0;
		for(int c=0;c<dData.cno();++c) meanrank += dData(r,c);
		meanrank /= wsum;

		consense.insert(make_pair(meanrank, r));
	}

	/* STORE THE CONSENSUS RANKING IN THE MAP */
		r=0;
		for(scoreset_cit s=consense.begin();s!=consense.end();++s, ++r)
			vRanks[r] = *s;
}	


//*******************************************************************************
// C L A S S :  R A N K B Y N U M B E R
//*******************************************************************************

//*******************************************************************************
void RankByNumber::compConsensusScore()
//*******************************************************************************
{
	//call the base class virtual function
	Consensor::compConsensusScore();
	
	typedef set<score_t, greater<score_t> > scoreset_t;
	typedef scoreset_t::const_iterator scoreset_cit;

	int r;

	/* CALCULATE THE CONSENSUS SCORING */
	scoreset_t consense;
	
	//build the average rank over all columns
	for(r=0;r<dData.rno();++r)
	{
		double meanrank = 0.0;
		for(int c=0;c<dData.cno();++c) meanrank += dData(r,c);
		meanrank /= dData.rno();

		consense.insert(make_pair(meanrank, r));
	}

	/* STORE THE CONSENSUS RANKING IN THE MAP */
		r=0;
		for(scoreset_cit s=consense.begin();s!=consense.end();++s, ++r)
			vRanks[r] = *s;
}	

//******************************************************************************
// FACTORY FUNCTIONS
//******************************************************************************

//*******************************************************************************
Consensor* RankByRankFactory(const Rectmat_<double>& data, 
														 const Vector_<double>& weights)
//*******************************************************************************
{
	return new RankByRank(data, weights);
}	

//*******************************************************************************
Consensor* RankByNumberFactory(const Rectmat_<double>& data, 
															 const Vector_<double>& weights, 
															 bool autoscale)
//*******************************************************************************
{
	return new RankByNumber(data, weights, autoscale);
}	
