/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CONSENSOR_H
#define CONSENSOR_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <Rectmat.hh>
#include <Vector.hh>

#ifdef HAVE_STD
#include <vector>
#include <functional>
#else
#include <vector.h>
#include <function.h>
#endif

/**
 * \brief a scoring type
 *
 * A scoring type consists of a scoring value and the corresponding row index.
 */
typedef pair<double, int> score_t;

/**
 * \brief the base class of all consensor functors
 *
 * The consensor base class provides basic functionality for all consensor
 * classes. A consensor is a functional class that performs a consensus scoring
 * of a given data matrix. Each column of the matrix is considered in a
 * specific way to return a rank based on all data columns for the rows of the
 * matrix. The result of a consensus scoring is a map between the ranks and the
 * corresponding row indices.
 *
 * A consensor needs the data matrix (dData), a weight vector that defines the
 * contribution of each row to the consensus scoring (vWeights) and
 * instructions, if the data matrix should be autoscaled or not
 * (fAutoScale). The result can be checked by the mScoring map between ranks
 * and row indices.
 *
 * The weights in the vWeights vector can have a negative sign, which is used
 * to reverse the ordering of that column (if you want to have ascending order
 * (0 is best)). However the sign of the weight is discarded when calculating
 * the contributions, thus a score of -5 contributes more than a score of +2.
 */
class Consensor
{
public:

	//****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//****************************************************************************

	///\name rank vector type
	//@{
	/**
	 * \brief a type for the rank map + iterator definitions
	 */
	typedef std::vector<score_t> rankvec_t;
	typedef rankvec_t::const_iterator rankvec_cit;
	typedef rankvec_t::iterator       rankvec_it;
  //@}

	//****************************************************************************
	// CONSTRUCTORS AND INITIALIZATION
	//****************************************************************************

	/**
	 * \brief default constructor, initializes an empty consensor object
	 *
	 * The datamatrix (dData) and the weight vector (vweights) are initialized
	 * with zero dimensions and the fAutoScale property is set to false.
	 *
	 * \sa Consensor(const RazorBack::Rectmat_<double>&, 
	 *               const RazorBack::Vector_<double>&, bool)
	 *     Consensor(const Consensor&)
	 */
	Consensor() : vRanks(0), 
								dData(0,0), vWeights(0), fAutoScale(false) { }
	/**
	 * \brief initializing constructor
	 *
	 * The initializing constructor copies the contents of \p data, \p weights to
	 * the dData datamatrix and vWeights weight vector. The autoscale property
	 * fAutoScale is initialized with \p autoscale.
	 *
	 * \param data      a reference to the scoring data matrix
	 * \param weights   a reference to the weight vector for the scoring columns
	 * \param autoscale flag indicating if the data matrix should be autscaled 
	 *                  (\c true) or not (\c false = default).
	 *
	 * \sa Consensor(), Consensor(const Consensor&),
	 *     init(const RazorBack::Rectmat_<double>&, 
	 *          const RazorBack::Vector_<double>&, bool)
	 */
	Consensor(const RazorBack::Rectmat_<double>& data, 
						const RazorBack::Vector_<double>& weights,	bool autoscale = false)
			: vRanks(data.rno()),
				dData(data), vWeights(weights), fAutoScale(autoscale) { }
	/**
	 * \brief copy constructor
	 *
	 * The copy constructor takes another Consensor object and copies its data
	 * matrix, weighting vector and autoscaling flag. The constructor also copies
	 * the mRanks map, representing the results of the \p src Consensor.
	 *
	 * \param src the source Consensor object.
	 *
	 * \sa Consensor(const RazorBack::Rectmat_<double>&, 
	 *              const RazorBack::Vector_<double>&, bool)
	 *     Consensor()
	 */
	Consensor(const Consensor& src)
			: vRanks(src.vRanks), 
				dData(src.dData), vWeights(src.vWeights), fAutoScale(src.fAutoScale) { }

	/**
	 * \brief initialization method
	 *
	 * The initializiation method simply performs the same operations as the
	 * Consensor(const RazorBack::Rectmat_<double>&, 
	 *           const RazorBack::Vector_<double>&, bool)
	 * constructor, except that it can (re)initialize existing objects.
	 *
	 *
	 * \param data      a reference to the scoring data matrix
	 * \param weights   a reference to the weight vector for the scoring columns
	 * \param autoscale flag indicating if the data matrix should be autscaled 
	 *                  (\c true) or not (\c false = default).
	 *
	 * \sa Consensor(const RazorBack::Rectmat_<double>&, 
	 *               const RazorBack::Vector_<double>&, bool)
	 */
	void init(const RazorBack::Rectmat_<double>& data, 
						const RazorBack::Vector_<double>& weights,	bool autoscale = false);

	//****************************************************************************
	// PROPERTY METHODS
	//****************************************************************************

	/**
	 * \brief copies the content of \p data to the scoring matrix
	 *
	 * \param data a reference to a data matrix
	 */
	void setData(const RazorBack::Rectmat_<double>& data);
	/**
	 * \brief sets the value of the cell (\p row, \p col)
	 *
	 * \param row   the row index of the cell
	 * \param col   the column index of the cell
	 * \param value the new value of the cell
	 *
	 * \return \c true if the cell could be changed, or \c false if the indices
	 * were out of range.
	 */
	bool setData(size_t row, size_t col, double value);
	/**
	 * \brief retrieves the score at in \c row and \c col.
	 *
	 * \param row   the row index of the cell
	 * \param col   the column index of the cell
	 *
	 * \return the value of the cell or 0 if the indices were out of range
	 */
	double getData(size_t row, size_t col) const { return dData(row, col); }
	
	/**
	 * \brief redimensionates the scoring problem
	 *
	 * The method resizes the dData matrix to \p rows and \p cols and the vWeights
	 * vector to \p cols.
	 *
	 * \param rows the number of rows in the scoring matrix
	 * \param cols the number of columns in the scoring matrix
	 */
	void redim(size_t rows, size_t cols);

	/**
	 * \brief copies the elements of \p weights to the weight vector
	 *
	 * If the weight vector has not enough dimensions, it is resized and the
	 * missing elements are set to zero.
	 *
	 * \param weights a reference to the weight vector
	 */
	void setWeights(const RazorBack::Vector_<double>& weights);
	/**
	 * \brief change the weight of a single column
	 *
	 * \param col    the required column
	 * \param weight the new weight of the column
	 *
	 * \return \c true if the weight could be set, or false otherwise
	 */
	bool setWeight(size_t col, double weight);
	/**
	 * \brief retrieves the weight of a particular column
	 *
	 * \param col the column index
	 * 
	 * \return the weight of the column or zero if the column index is out of
	 * bounds.
	 */
	double getWeight(size_t col) const 
	{ return ((col>=0) && (col<vWeights.dim())) ? vWeights[col] : 0; }

	/**
	 * \brief enables or disables the autoscaling of the data matrix
	 *
	 * \param enable if \c true enables the autoscaling, otherwise autoscaling is
	 * disabled.
	 */
	void enableAutoScaling(bool enable = true);
	/**
	 * \brief retrieves the status of the autoscaling flag
	 *
	 * \return \c true if the autoscaling is enabled, or \c false otherwise.
	 */
	bool isAutoScaling() const { return fAutoScale; }

	//****************************************************************************
	// RESULTING DATA METHODS
	//****************************************************************************

	/**
	 * \brief returns the row at a particular \p rank
	 *
	 * \return a pair that contains the consensus score and the row index of the
	 * specified rank.
	 */
	score_t rowByRank(unsigned int rank) const;
	
	//@{
	/**
	 * \brief readonly access to the sequence of rank/row pairs
	 *
	 * The methods return the begin or end of the sequence of rank/row pairs. The
	 * sequence will start with the best row.
	 */
	rankvec_cit begin() const { return vRanks.begin(); }
	rankvec_cit end() const   { return vRanks.end(); } 
	//@}


	//*****************************************************************************
	// CALCULATIONS
	//*****************************************************************************

	/**
	 * \brief performs the consensus scoring
	 *
	 * This class performs the consensus scoring. It must be reimplemented by the
	 * derived classes. The method of the baseclass provides basic data
	 * manipulation, namely the autoscaling (if demanded), followed by the
	 * weighting. The actual calculations must be done by the child classes.
	 */
	virtual void compConsensusScore();
	
protected:

	//****************************************************************************
	// PROTECTED MEMBERS
	//****************************************************************************

	/**
	 * \brief this maps contains the consensus ranking
	 *
	 * The consensus ranking is stored in an ordered sequence of rank/row-indices
	 * pairs. This vector is populated only after a call  of compCensensusScore().
	 */
	rankvec_t vRanks;

	/**
	 * \brief the scoring data matrix
	 *
	 * This matrix contains the values of the different scoring functions for
	 * each entry.
	 *
	 * \note this data is maybe manipulated during the calculation of the
	 * consensus scoring.
	 */
	RazorBack::Rectmat_<double> dData;
	/**
	 * \brief the weighting vector for the scoring functions
	 *
	 * This vector contains a weighting factor for each column in the data
	 * matrix. The weighting factor is applied after the autoscaling.
	 */
	RazorBack::Vector_<double> vWeights;
	
	/**
	 * \brief a flag indicating if an autoscaling of the data matrix is requested
	 */
	bool fAutoScale;

};

//*******************************************************************************
// C L A S S :  R A N K B Y R A N K
//*******************************************************************************

/**
 * \brief performs a consensus scoring by average ranks
 *
 * The RankByRank calculates the ranking for each column in the data scoring
 * matrix and generates a consensus scoring value from the weighted average of
 * the single ranks of each row. For the weighting of the ranks, the values of
 * the data weighting vector are used, but their sign is discarded.
 */
class RankByRank: public Consensor
{
public:

	//****************************************************************************
	// CONSTRUCTORS AND INITIALIZATION
	//****************************************************************************

	/**
	 * \brief default constructor, initializes an empty RankByRank object
	 *
	 * The datamatrix (dData) and the weight vector (vweights) are initialized
	 * with zero dimensions and the fAutoScale property is set to false.
	 *
	 * \sa Consensor::Consensor()
	 *     RankByRank(const RazorBack::Rectmat_<double>&, 
	 *                const RazorBack::Vector_<double>&, bool)
	 *     RankByRank(const Consensor&)
	 */
	RankByRank() : Consensor() { }
	/**
	 * \brief initializing constructor
	 *
	 * The initializing constructor copies the contents of \p data, \p weights to
	 * the dData datamatrix and vWeights weight vector. The autoscale property
	 * fAutoScale is initialized with \p autoscale.
	 *
	 * \param data      a reference to the scoring data matrix
	 * \param weights   a reference to the weight vector for the scoring columns
	 *
	 * \sa Consensor::Consensor(const RazorBack::Rectmat_<double>&, 
	 *                          const RazorBack::Vector_<double>&, bool)
	 *     RankByRank()
	 *     RankByRank(const Consensor&)
	 */
	RankByRank(const RazorBack::Rectmat_<double>& data, 
						 const RazorBack::Vector_<double>& weights) 
			: Consensor(data, weights, false) { }
	/**
	 * \brief copy constructor
	 *
	 * The copy constructor takes another Consensor object and copies its data
	 * matrix, weighting vector and autoscaling flag. The constructor also copies
	 * the mRanks map, representing the results of the \p src Consensor.
	 *
	 * \param src the source Consensor object.
	 *
	 * \sa Consensor::Consensor(const Consensor&)
	 *     RankByRank(const RazorBack::Rectmat_<double>&, 
	 *                const RazorBack::Vector_<double>&, bool)
	 *     RankByRank()
	 */
	RankByRank(const Consensor& src) : Consensor(src) { fAutoScale = false; }
	
	//*****************************************************************************
	// CALCULATIONS
	//*****************************************************************************

	/**
	 * \brief performs the consensus scoring
	 *
	 * This class performs the consensus scoring rank by rank. That is it creates
	 * a consensus scoring that is determined by the average rank over all data
	 * columns.
	 */
	void compConsensusScore();
};

//*******************************************************************************
// C L A S S :  R A N K B Y N U M B E R
//*******************************************************************************

/**
 * \brief performs a consensus scoring by average scores
 *
 * The RankByNumber generates a consensus scoring value from the weighted average of
 * the single scoring values of each row.
 */
class RankByNumber: public Consensor
{
public:

	//****************************************************************************
	// CONSTRUCTORS AND INITIALIZATION
	//****************************************************************************

	/**
	 * \brief default constructor, initializes an empty RankByNumber object
	 *
	 * The datamatrix (dData) and the weight vector (vweights) are initialized
	 * with zero dimensions and the fAutoScale property is set to false.
	 *
	 * \sa Consensor::Consensor()
	 *     RankByNumber(const RazorBack::Rectmat_<double>&, 
	 *                  const RazorBack::Vector_<double>&, bool)
	 *     RankByNumber(const Consensor&)
	 */
	RankByNumber() : Consensor() { }
	/**
	 * \brief initializing constructor
	 *
	 * The initializing constructor copies the contents of \p data, \p weights to
	 * the dData datamatrix and vWeights weight vector. The autoscale property
	 * fAutoScale is initialized with \p autoscale.
	 *
	 * \param data      a reference to the scoring data matrix
	 * \param weights   a reference to the weight vector for the scoring columns
	 * \param autoscale flag indicating if the data matrix should be autscaled 
	 *                  (\c true) or not (\c false = default).
	 *
	 * \sa Consensor::Consensor(const RazorBack::Rectmat_<double>&, 
	 *                          const RazorBack::Vector_<double>&, bool)
	 *     RankByNumber()
	 *     RankByNumber(const Consensor&)
	 */
	RankByNumber(const RazorBack::Rectmat_<double>& data, 
							 const RazorBack::Vector_<double>& weights, bool autoscale=false)
			: Consensor(data, weights, autoscale) { }
	/**
	 * \brief copy constructor
	 *
	 * The copy constructor takes another Consensor object and copies its data
	 * matrix, weighting vector and autoscaling flag. The constructor also copies
	 * the mRanks map, representing the results of the \p src Consensor.
	 *
	 * \param src the source Consensor object.
	 *
	 * \sa Consensor::Consensor(const Consensor&)
	 *     RankByNumber(const RazorBack::Rectmat_<double>&, 
	 *                  const RazorBack::Vector_<double>&, bool)
	 *     RankByNumber()
	 */
	RankByNumber(const Consensor& src) : Consensor(src) { }
	
	//*****************************************************************************
	// CALCULATIONS
	//*****************************************************************************

	/**
	 * \brief performs the consensus scoring
	 *
	 * This class performs the consensus scoring rank by rank. That is it creates
	 * a consensus scoring that is determined by the average rank over all data
	 * columns.
	 */
	void compConsensusScore();
};


//******************************************************************************
// FACTORY FUNCTIONS
//******************************************************************************

Consensor* RankByRankFactory(const RazorBack::Rectmat_<double>& data, 
														 const RazorBack::Vector_<double>& weights);

Consensor* RankByNumberFactory(const RazorBack::Rectmat_<double>& data, 
															 const RazorBack::Vector_<double>& weights, 
															 bool autoscale = false);


#endif // CONSENSOR_H

