/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CORRESPONDENCES_HH
#define CORRESPONDENCES_HH


#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <slist>
#    include <set>
#  else
#    include <slist.h>
#    include <set.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <slist>
#  include <set>
#endif //HAVE_CONFIG_H

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map.hpp>

#include <trimesh.hh>

#ifdef HAVE_NAMESPACES 
using namespace std;

namespace boost
{
#endif
  enum vertex_surface_t      { vertex_surface = 1001 };
  enum vertex_vertex_t       { vertex_vertex  = 1002 };
	enum edge_correspondence_t { edge_correspondence = 1003 };

  BOOST_INSTALL_PROPERTY(vertex, surface);
  BOOST_INSTALL_PROPERTY(vertex, vertex);
	BOOST_INSTALL_PROPERTY(edge,   correspondence);
#ifdef HAVE_NAMESPACES
};
#endif

struct correspond_t
{
	size_t first;
	size_t second;
	
	double distance;

	correspond_t(const Trimesh::vertex_t A, const Trimesh::vertex_t B, const double d)
			: first(A), second(B), distance(d) { }

	bool operator<(const correspond& other) { return (distance<B.distance); }
	
	/**
	 * \brief fake equality operator
	 *
	 * This fake equality operator is used to remove all entries regarding either
	 * \c A or \c B.
	 *
	 * \param other the comparing correspondence
	 *
	 * \return true if either \c A or \c B are equal
	 */
	bool operator==(const correspond& other) 
	{ return ((A==other.A) || (A==other.A)); }

};

typedef slist<correspond_t> corresponds_t;
typedef corresponds_t::iterator corresponds_it;
typedef corresponds_t::const_iterator corresponds_cit;


class corrgraph_traits
{
public:

	//*****************************************************************************
	// GRAPH PROPERTIES
	//*****************************************************************************

	/**
	 * \brief stores the surface of the correspondence vertex
	 */
	typedef property<vertex_surface_t, int> surface_prop;
	
	/**
	 * \brief stores the point number of the correspondence vertex
	 */
	typedef property<vertex_vertex_t, Trimesh::vertex_t, surface_prop> vertex_prop;
	
	/**
	 * \brief the vertex index property
	 */
	typedef proeprty<vertex_index_t, size_t, vertex_prop> ix_prop;
	
	/**
	 * \brief reference from an edge to an element in the correspondence list
	 */
	typedef property<edge_correspondence_t, corresponds_it > edge_prop;

	//*****************************************************************************
	// GRAPH TYPE DEFINITION
	//*****************************************************************************

	/**
	 * \brief the correspondence graph
	 */
	typedef adjacency_list<listS, listS, undirectedS, ix_prop, edge_prop> graph_t;	
	
	
	//*****************************************************************************
	// PROPERTY MAPS
	//*****************************************************************************

	/**
	 * \brief surface property map
	 */
	typedef property_map<graph_t, vertex_surface_t>::type surfacemap_t;
	
	/**
	 * \brief vertex property map
	 */
	typedef property_map<graph_t, vertex_vertex_t>::type vertexmap_t;

	/**
	 * \brief the correspondence property map
	 */
	typedef property_map<graph_t, edge_correspondence_t>::type corrmap_t;

	//*****************************************************************************
	// VERTEX, EDGE descriptors and iterators
	//*****************************************************************************

  /**
   * \brief typedefinition of the vertex descriptor
   */
  typedef graph_traits<graph_t>::vertex_descriptor vertex_t;

  /**
   * \brief typedefinition of the edge descriptor
   */
  typedef graph_traits<graph_t>::edge_descriptor edge_t;

  /**
   * \brief iterator type for the vertices
   */
  typedef graph_traits<graph_t>::vertex_iterator vertex_it;

	/**
	 * \brief pair of vertex iterators
	 */
	typedef pair<vertex_it, vertex_it> vertices_t;

  /**
   * \brief iterator type for the edges
   */
  typedef graph_traits<graph_t>::edge_iterator edge_it;
  
	/**
	 * \brief pair of edge iterators
	 */
	typedef pair<edge_it, edge_it> edges_t;

  /**
   * \brief out-edge iterator type
   */
  typedef graph_traits<graph_t>::out_edge_iterator edge_oit;

  /**
   * \brief adjancency iterator
   */
  typedef graph_traits<graph_t>::adjacency_iterator vertex_ait;

	/**
	 * \brief pair of adjacency iterators
	 */
	typedef pair<vertex_ait, vertex_ait> adjacency_t;
};

typedef corrgraph_traits::graph_t corrgraph_t;

class Correspondent
{

public:
	//*****************************************************************************
	// CONSTRUCTOR AND INITIALIZING METHODS
	//*****************************************************************************


	/**
	 * \brief standard constructor
	 */
	Correspondent() : lCorresponds() { }

	/**
	 * \brief adds an ordered list of correspondencies to the list of possible
	 * correspondencies.
	 *
	 * The \p corrlist must be a single linked list ordered by ascending
	 * distances. The elements of \p corrlist are added to lCorresponds in a merge
	 * like fashion.
	 *
	 * \param corrlist ordered list of correspondencies
	 *
	 * \return the number of inserted correspondencies
	 */
	size_t addCorrespondencies(const corresponds_t& corrlist);
	
	/**
	 * \brief constructs all possible correspondencies regarding two overlapped
	 * harmonic maps.
	 *
	 * The method takes both harmonic maps, rotates the first according to \p
	 * angle and adds all possible pairs to lCorresponds. Finally the whole
	 * lCorresponds is sorted. It is recommended to use this method only if
	 * establishing the correspondencies between a single pair of patches.
	 *
	 * \param mapA,mapB two Trimesh structures allready mapped onto the unit disk
	 * \param angle     rotation of A against B (usually the best overlap position)
	 *
	 * \return the number of inserted correspondencies
	 */
	size_t addCorrespondencies(Trimesh& mapA, Trimesh& mapB, const double angle);

	/**
	 * \brief operator function that establishes all correspondencies
	 *
	 * The operator subsequently extracts the possible correspondence with minimum
	 * distance into the output sequence \p ins and then removes all entries in
	 * lCorresponds regarding the selected pair Va, Vb
	 *
	 * \param ins an output sequence (e.g. back inserter of an slist)
	 */
	template<class Out>
	operator()(Out ins);

private:

	/**
	 * \brief the sorted list of possible correspondencies
	 *
	 * The possible correspondencies are stored ordered by ascending distances (in
	 * the harmonic map overlap). Before the correspondence is checked all
	 * relevant correspondence lists (e.g. the correspondencies of all involved
	 * patches) must be added by subsequent calls of 
	 * addCorrespondencies(const corresponds_t&) or 
	 * addCorrespondencies(Trimesh&, Trimesh&, const double).
	 */
	corresponds_t lCorresponds;

};

#ifdef INCLUDE_TMPL_DEFS
  #include "correspondent_temp.cc"
#endif


#endif //CORRESPONDENCES_HH
