/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CORRESPONDENT_HH
#define CORRESPONDENT_HH

#ifdef _STANDARD_C_PLUS_PLUS
#include <list>
#include <set>
#else //_STANDARD_C_PLUS_PLUS
#include <list.h>
#include <set.h>
#endif //_STANDARD_C_PLUS_PLUS

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map.hpp>

#include <trimesh.hh>
#include <math.h>

namespace boost
{
  enum vertex_surface_t      { vertex_surface = 1001 };
  enum vertex_vertex_t       { vertex_vertex  = 1002 };
	enum edge_correspondence_t { edge_correspondence = 1003 };

  BOOST_INSTALL_PROPERTY(vertex, surface);
  BOOST_INSTALL_PROPERTY(vertex, vertex);
	BOOST_INSTALL_PROPERTY(edge,   correspondence);
};

struct correspond_t
{
	size_t first;
	size_t second;
	
	double distance;
	double rmsWeight;

	correspond_t(): first(0), second(0), distance(-1.0), rmsWeight(0) { }
	correspond_t(const Trimesh::vertex_t A, const Trimesh::vertex_t B, const double d)
			: first(A), second(B), distance(d), rmsWeight(0) { }

	bool operator<(const correspond_t& other) const { return (distance<other.distance); }
	bool operator==(const correspond_t& other) const
	{
		return ((first==other.first) && (second==other.second) && 
						(fabs(distance - other.distance)<DBL_EPSILON));
	}
	
};

typedef list<correspond_t> corresponds_t;
typedef corresponds_t::iterator corresponds_it;
typedef corresponds_t::const_iterator corresponds_cit;


class Correspondent
{

public:
	//*****************************************************************************
	// CONSTRUCTOR AND INITIALIZING METHODS
	//*****************************************************************************


	/**
	 * \brief standard constructor
	 */
	Correspondent() : lCorresponds() { }

	/**
	 * \brief adds an ordered list of correspondencies to the list of possible
	 * correspondencies.
	 *
	 * The \p corrlist must be a single linked list ordered by ascending
	 * distances. The elements of \p corrlist are added to lCorresponds in a merge
	 * like fashion.
	 *
	 * \param begin,end output iterators pointing to the start and end of an
	 *                  ordered sequence of correspondencies
	 *
	 * \return the number of inserted correspondencies
	 */
	template<class In> size_t addCorrespondencies(In begin, In end);
	
	/**
	 * \brief constructs all possible correspondencies regarding two overlapped
	 * harmonic maps.
	 *
	 * The method takes both harmonic maps, rotates the first according to \p
	 * angle and adds all possible pairs to lCorresponds. Finally the whole
	 * lCorresponds is sorted. It is recommended to use this method only if
	 * establishing the correspondencies between a single pair of patches.
	 *
	 * \param mapA,mapB two Trimesh structures allready mapped onto the unit disk
	 * \param angle     rotation of A against B (usually the best overlap position)
	 *
	 * \return the number of inserted correspondencies
	 */
	size_t addCorrespondencies(Trimesh& mapA, Trimesh& mapB, const double angle);

	/**
	 * \brief operator function that establishes all correspondencies
	 *
	 * The operator subsequently extracts the possible correspondence with minimum
	 * distance into the output sequence \p ins and then removes all entries in
	 * lCorresponds regarding the selected pair Va, Vb
	 *
	 * \param ins an output sequence (e.g. back inserter of an list)
	 *
	 * \return number of correspondents written to \p ins
	 */
	template<class Out>	size_t operator()(Out ins);

private:

	/**
	 * \brief the sorted list of possible correspondencies
	 *
	 * The possible correspondencies are stored ordered by ascending distances (in
	 * the harmonic map overlap). Before the correspondence is checked all
	 * relevant correspondence lists (e.g. the correspondencies of all involved
	 * patches) must be added by subsequent calls of 
	 * addCorrespondencies(const corresponds_t&) or 
	 * addCorrespondencies(Trimesh&, Trimesh&, const double).
	 */
	corresponds_t lCorresponds;

};

#ifdef INCLUDE_TMPL_DEFS
  #include "correspondent_templ.cc"
#endif


#endif //CORRESPONDENT_HH
