/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "rigidbody.hh"

#include <matrect.h>
#include <vect.h>
#include <bestrot.h>
#include <string.h>


//*******************************************************************************
void rbTransformation::rbFit()
//*******************************************************************************
{
	int i = 0;
	corresponds_cit c;

	//rotation and translation data
	Matrect_ rotation;
	Vect_ center[2];

	init_matrect(&rotation, 3, 3);
	for(i=0;i<2;++i) init_vect(&center[i], 3);
	
	//intialize point sets & weights
	Matrect_ rms_coords[2];
	Vect_ weights;

	for(i=0;i<2;++i)
		init_matrect(&rms_coords[i], correspondencies.size(), 3);
	init_vect(&weights, correspondencies.size());

	//build the piont set data	
	for(c=correspondencies.begin(), i=0;i<correspondencies.size();++i, ++c)
	{
		//assign the coordinates of the point sets
		for(int j=0;j<3;++j)
		{
			rms_coords[0].Mat[i][j] = fCoordinates[0][c->first][j];
			rms_coords[1].Mat[i][j] = fCoordinates[1][c->second][j];
		}

		//assign the weight if weighting is enabled
		if (fWeighted)
			weights.Vec[i] = c->distance;
		else
			weights.Vec[i] = 1.0;
	}

	//center the matrices and obtain the centroids
	for(i=0;i<2;++i)
		center_vectors(&rms_coords[i], &center[i], 0);


	/* RIGID BODY FIT */
	fRMSD = best_rot(&rms_coords[0], &rms_coords[1], &weights, &rotation);

	/* CONSTRUCTION OF THE TRANSFORMATION MATRIX */
	
	/*
		The complete transformation matrix has the form:
		
		r11 r12 r13 (r1 * vB0) + xBA
		r21 r22 r23 (r2 * vB0) + yBA
		r31 r32 r33 (r3 * vB0) + zBA
		 0   0   0         1
	 */
	
	for(i=0; i<3; ++i)
	{
		fTransformation[i*4+3] = 0.0;
		for(int j=0;j<3;++j)
		{
			fTransformation[i*4+3] -= rotation.Mat[i][j] * center[1].Vec[j];
			fTransformation[i*4+j] =  rotation.Mat[i][j];
		}
		
		fTransformation[i*4+3] += center[0].Vec[i] - center[1].Vec[i];
	}
	
	//forth row
	for(i=12;i<16;++i)
		fTransformation[i] = (i!=15) ? 0 : 1.0;
	
	

	free_matrect(&rotation);
	free_vect(&weights);
	
	for(i=0;i<2;++i)
	{
		free_matrect(&rms_coords[i]);
		free_vect(&center[i]);
	}
}

//*******************************************************************************
void rbTransformation::getCorrespondencies()
//*******************************************************************************
{
  typedef list<Trimesh::vertex_t> vertices_t;
	typedef vertices_t::iterator vertices_it;
	typedef vertices_t::const_iterator vertices_cit;

	int C,B;  //B = base C = check object

	//determine which map has more vertices and assign the A/B indices appropriate
	if (fMaps[1]->nVertices() > fMaps[0]->nVertices())
	{
		C = 0;
		B = 1;
	}
	else
	{
		C = 1;
		B = 0;
	}

	//erase all point pair correspondencies
	correspondencies.clear();

	//build the list of base vertices
	vertices_t base_vertices(fMaps[B]->begin(),fMaps[B]->end());
	
	for(Trimesh::vertex_it v=fMaps[C]->begin();v!=fMaps[C]->end();++v)
	{
		double x = fMaps[C]->vertCoordinates(*v)[0];
		double y = fMaps[C]->vertCoordinates(*v)[1];
		
		vertices_it best;
		double best_dist = DBL_MAX;
		double best_dsquare = 0.0;

		for(vertices_it w=base_vertices.begin();w!=base_vertices.end();++w)
		{
			double x_dist = x - fMaps[B]->vertCoordinates(*w)[0];
			double y_dist = y - fMaps[B]->vertCoordinates(*w)[1];
	
			double dsquare = (x_dist*x_dist) + (y_dist*y_dist);
			double dist = sqrt((x_dist*x_dist) + (y_dist*y_dist));
			
			if (dist<best_dist)
			{
				best = w;
				best_dist = dist;
				best_dsquare = dsquare;
			}
		}
		
		//create the new correspondence
		if (C==0)
			correspondencies.push_back(correspond_t(*v, *best, best_dist));
		else
			correspondencies.push_back(correspond_t(*best, *v, best_dist));


		//remove the selected vertex in the base patch
		base_vertices.erase(best);
	}
}

//*******************************************************************************
void rbTransformation::clampWeighting()
//*******************************************************************************
{
  double max_value = DBL_MIN;
	double min_value = DBL_MAX;
	
	for(corresponds_it i=correspondencies.begin();i!=correspondencies.end();++i)
	{
	  if ((i->distance) > max_value) max_value = i->distance;
		if ((i->distance) < min_value) min_value = i->distance;
	}

	for(corresponds_it i=correspondencies.begin();i!=correspondencies.end();++i)
		i->distance = 1 - (i->distance - min_value) / (max_value - min_value);
}
	
//*******************************************************************************
rbTransformation::rbTransformation()
//*******************************************************************************
: fRMSD(0), fWeighted(false)
{
	//set the transformation matrix to the identity matrix
	for(int i=0;i<4;++i)
		for(int j=0;j<4;++j) 
			fTransformation[i*4+j] = (i==j) ? 1.0 : 0.0;

	fMaps[0] = 0;
	fMaps[1] = 0;
}

//*******************************************************************************
rbTransformation::rbTransformation(Trimesh& surfaceA, Trimesh& surfaceB)
//*******************************************************************************
: fRMSD(0), fWeighted(false)
{
	//set the transformation matrix to the identity matrix
	for(int i=0;i<4;++i)
		for(int j=0;j<4;++j) 
			fTransformation[i*4+j] = (i==j) ? 1.0 : 0.0;	

	fMaps[0] = 0;
	fMaps[1] = 0;

	setCoordinates(0, surfaceA);
	setCoordinates(1, surfaceB);

	initialize();
}

//*******************************************************************************
void rbTransformation::setCoordinates(const size_t ix, Trimesh& surface)
//*******************************************************************************
{
	if ((ix==0) || (ix==1))
	{
		fCoordinates[ix].clear();
		for(Trimesh::vertex_it v = surface.begin();v!=surface.end();++v)
			fCoordinates[ix][*v] = surface.vertCoordinates(*v);
	}
}

//*******************************************************************************
void rbTransformation::setMaps(Trimesh* mapA, Trimesh* mapB)
//*******************************************************************************
{
	fMaps[0] = mapA;
	fMaps[1] = mapB;
}



//*******************************************************************************
void rbTransformation::getTransformation(double* transform_) const
//*******************************************************************************
{
	memcpy(transform_, fTransformation, 16 * sizeof(double));
}

//*******************************************************************************
double rbTransformation::operator()()
//*******************************************************************************
{
	rbFit();
	return fRMSD;
}

//*******************************************************************************
void rbTransformation::initialize()
//*******************************************************************************
{
	getCorrespondencies();
	if (fWeighted) clampWeighting();
}
