/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef RIGIDBODY_HH
#define RIGIDBODY_HH

#include <float.h>

#ifdef _STANDARD_C_PLUS_PLUS
  #include <list>
  #include <map>
#else
  #include <list.h>
  #include <map.h>
#endif

#include <Vector.hh>
#include <Rectmat.hh>
#include <harmonic_image.hh>
#include <trimesh.hh>

class rbTransformation
{

public:

	//*****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//*****************************************************************************

	/**
	 * \brief type of a coordinate map plus iterator
	 */
	typedef map<Trimesh::vertex_t, Vector_<double> > coordinatemap_t;

private:
	//*****************************************************************************
	// PRIVATE TYPE DEFINITIONS
	//*****************************************************************************

	typedef Trimesh::vertex_t vertex_t;

	/**
	 * \brief description of a corresponding vertex pair
	 */
	struct correspond_t
	{
		vertex_t first;
		vertex_t second;
	
		double distance;

		correspond_t(const vertex_t A, const vertex_t B, const double d)
				: first(A), second(B), distance(d) { }
	};

	/**
	 * \brief type of a correspondence list
	 */
	typedef list<correspond_t> corresponds_t;
	typedef corresponds_t::iterator corresponds_it;
	typedef corresponds_t::const_iterator corresponds_cit;

	//*****************************************************************************
	// PRIVATE MEMBERS
	//*****************************************************************************


	/**
	 * \brief Correspondence list
	 */
	corresponds_t correspondencies;
	
	/**
	 * \brief surface objects
	 */
	coordinatemap_t fCoordinates[2];

	/**
	 * \brief harmonic maps of the surface objects
	 */
	Trimesh* fMaps[2];

	/**
	 * \brief transformation matrix
	 *
	 * The transformation matrix contains the full rigid-body transformation to fit
	 * B into A: 
	 *  - translation of B into zero, 
	 *  - rotation to the orientation of A and
	 *  - translation of the center of B into the center of A
	 */
	 double fTransformation[16];

	/**
	 * \brief contains the RMSD value
	 */
	double fRMSD;

	//*****************************************************************************
	// FLAGS
	//*****************************************************************************

	/**
	 * \brief specifies if the fit should be weighted by the point pair distance on
	 *        the maps.
	 *
	 * If this flag is true, the distance between a point pair on the overlaid maps
	 * is taken as a weight for this pair.
	 */
	bool fWeighted;


	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief performs the rigid-body fit
	 */
	void rbFit();

	/**
	 * \brief builds the correspondence list between points in both objects
	 */
	void getCorrespondencies();

	/**
	 * \brief clamp weighting values between 0 and 1
	 */
	void clampWeighting();
	
public:

	//*****************************************************************************
	// CONSTRUCTORS
	//*****************************************************************************

	/**
	 * \brief standard constructor, does not initialize anything
	 */
	rbTransformation();

	/**
	 * \brief constructor performing a initialization of the coordinates
	 *
	 * \param surfaceA,surfaceB reference to a 3D surface object
	 */
	rbTransformation(Trimesh& surfaceA, Trimesh& surfaceB);

	//*****************************************************************************
	// PROPERTY METHODS
	//*****************************************************************************

	/**
	 * \brief (re)initialzes the coordinates of the surface object with index \p ix
	 *
	 * The method assigns \p surface_ as the new surface object of \p ix.
	 *
	 * \param ix       index of the surface object (must be 0 or 1 otherwise there will 
	 *                 be no assignment).
	 * \param surface  referrence to a 3D surface object
	 */
	void setCoordinates(const size_t ix, Trimesh& surface);

	/**
	 * \brief assigns the maps of both surface objects
	 *
	 * \param mapA,mapB 3D surface object files of the harmonic maps
	 */
	void setMaps(Trimesh* mapA, Trimesh* mapB);

	/**
	 * \brief turns weighting of point pairs ON or OFF
	 *
	 * \param weight if \c true the point pairs are weighted by their 2D distance
	 *
	 * \sa fWeighted
	 */
	void enableWeighting(const bool weight = true) { fWeighted = weight; }
	
	/**
	 * \brief returns the status of the weighted flag
	 *
	 * \return status of the weighted flag
	 *
	 * \sa fWeighted
	 */
	bool isWeighted() const { return fWeighted; }

	/**
	 * \brief copies the actual transformation matrix to the specified array of
	 *        doubles
	 *
	 * \param transform_ array of doubles receiving the transformation matrix
	 */
	void getTransformation(double* transform_) const;

	/**
	 * \brief retrieves the RMSD value of the rigid body transformation
	 *
	 * \return the RMSD value
	 */
	double RMSD() const { return fRMSD; }

	/**
	 * \brief retrieves the number of corresponding points used for the rigid-body
	 *        fit.
	 * 
	 * \return number of corresponding point pairs.
	 */
	size_t NPoints() const { return correspondencies.size(); }

	//*****************************************************************************
	// PUBLIC METHODS
	//*****************************************************************************

	/**
	 * \brief the call operator performs the rigid-body fit and returns the RMSD
	 *        value
	 *
	 * The rigid body transformation is performed for the whole set of established
	 * correspondence points.
	 *
	 * \return RMSD value of the fit.
	 */
	double operator()();

	/**
	 * \brief initializes the correspondencies
	 */
	void initialize();
};

#endif //RIGIDBODY_HH
