/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SCORECOLLECTION_H
#define SCORECOLLECTION_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#  include <list>
#else
#  include <list.h>
#endif

class Scorer;
class Cluster;

/**
 * \brief a collection of scorer functions
 *
 * The scorer function collection can be used to pass a set of scoring
 * calculators to an object that manipulates Cluster objects in an irreversible
 * way (like the SingleLink class). The class collects scoring functors of
 * different Scorer types and for different data (addScorer()) and calls one by one in
 * the function operator()(Cluster*).
 *
 * The class takes care about the Scorer objects and frees any allocated memory
 * at the end of its life cycle, or when the method clear() is called. If any
 * Scorer objects have to be removed during the lifetime of the object the
 * deleteScorer(scorerlist_it) and deleteScorer(Scorer*) methods can be used.
 *
 * The single entries of the ScoreCollection can be accessed by the begin(),
 * begin() const, end() and end() const methods.
 */
class ScoreCollection
{

public:

	///\name scorer list
	//@{
	/**
	 * \brief the Scorer list type and a constant and non-constant iterator
	 */
	typedef std::list<Scorer*> scorerlist_t;
	typedef std::list<Scorer*>::const_iterator scorerlist_cit;
	typedef std::list<Scorer*>::iterator scorerlist_it;
	//@}
	
	/**
	 * \brief default and standard constructor
	 *
	 * This constructor initializes the Scorer list
	 */
	ScoreCollection() : lScorer() { }
	/**
	 * \brief destructor
	 *
	 * The destructor frees all the Scorer* objects in the lScorer list. So any
	 * reference to the elements of the Scorer objects is invalid after the
	 * destruction of the ScoreCollection.
	 */
	~ScoreCollection();

	/**
	 * \brief removes all Scorer objects from the list and memory
	 */
	void clear();
	
	/**
	 * \brief the function operator invokes the scoring by all Scorer objects
	 *
	 * The operator calls each Scorer::operator()(Cluster*) one by one to perform
	 * the complete scoring of \p cluster.
	 *
	 * \param cluster the cluster to be scored
	 */
	void operator()(Cluster* cluster) const;

	/**
	 * \brief adds a Scorer object to the collection
	 *
	 * You can use any of the scorer-factory functions as a parameter for this
	 * method.
	 *
	 * \param scorer the new Scorer object
	 *
	 * \return an iterator pointing to the list entry of the Scorer object.
	 */
	scorerlist_it addScorer(Scorer* const scorer);
	/**
	 * \brief removes the list entry referenced by \p it
	 *
	 * \param it an iterator pointing into the scorerlist sequence.
	 * 
	 * \return \c true if the list entry could be removed successfully or \c false
	 * if it did not exist.
	 */
	bool deleteScorer(scorerlist_it it);
	/**
	 * \brief removes the list entry of the Scorer \p scorer
	 *
	 * \note This method hast to search (maybe) the whole list for the right object.
	 *
	 * \param scorer a pointer to a scoring object that should be removed from the list
	 *
	 * \return \c true if the list entry could be removed successfully or \c false
	 * if it did not exist.
	 */
	bool deleteScorer(Scorer* const scorer);

	///\name iterator functions
	//@{
	/**
	 * \brief constant or non-constant iterators pointing to the begin or end of the list sequence.
	 *
	 * \return A constant or non-constant iterator that points to the begin, or
	 * end of the lScorer sequence.
	 */
	scorerlist_it  begin()       { return lScorer.begin(); }
	scorerlist_cit begin() const { return lScorer.begin(); }
	scorerlist_it  end()         { return lScorer.end(); }
	scorerlist_cit end()   const { return lScorer.end(); }
	//@}

	/**
	 * \brief returns the number of scoring functions in the collection
	 *
	 * \return the number of list entries in the lScorer list.
	 */
	size_t size() const { return lScorer.size(); }
	
private:

	/**
	 * \brief the list of scoring functions
	 */
	scorerlist_t lScorer;

};


#endif // SCORECOLLECTION_H

