/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "scorer.hh"

#include "correspondent.hh"
#include "cluster.hh"
#include "utils.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
#endif


//*******************************************************************************
Scorer::Scorer(const RazorBack::Rectmat_<double>** data, const std::string name)
//*******************************************************************************
: fName(name)
{
	pData[0] = data[0];
	pData[1] = data[1];
}

//*******************************************************************************
int Scorer::operator()(Cluster* cluster) const
//*******************************************************************************
{
	//determine if one matrix has less columns than the other
  int colcount = 
		(pData[0]->cno()<pData[1]->cno()) ? pData[0]->cno() : pData[1]->cno();

	//loop over all columns
	for(int i=0;i<colcount;++i)
	{
		double score = calcScore(i, cluster);

		//determine the name of the score (consider more than one column
		string scorename = fName;
		if (colcount>1)
		{
			scorename += "_";
			scorename += itoa(i);
		}
		
		//write the score to the cluster
		cluster->setScore(scorename, score);
	}
	
	return colcount;
}

//*******************************************************************************
double RMSDScorer::calcScore(size_t column, Cluster* cluster) const
//*******************************************************************************
{
	double msd  = 0.0;
	double wsum = 0.0;
	
	//loop over all corresponding point pairs
	corresponds_cit c;
	for(c=cluster->beginCorr();c!=cluster->endCorr();++c)
	{
		double pA = (*pData[0])(c->first, column);
		double pB = (*pData[1])(c->second, column);
		double diff = pA - pB;

		if (fWeighting)
		{
			msd  += c->rmsWeight * diff * diff;
			wsum += c->rmsWeight;
		}
		else
		{
			msd  += diff * diff;
			wsum += 1.0;
		}
	}

  return sqrt(msd/wsum);
}

//*******************************************************************************
double MeanScorer::calcScore(size_t column, Cluster* cluster) const
//*******************************************************************************
{
	double mean = 0.0;
	double wsum = 0.0;
	
	//loop over all corresponding point pairs
	corresponds_cit c;
	for(c=cluster->beginCorr();c!=cluster->endCorr();++c)
	{
		double pA = (*pData[0])(c->first, column);
		double pB = (*pData[1])(c->second, column);
		double diff = pA - pB;

		if (fWeighting)
		{
			mean += c->rmsWeight * fabs(diff);
			wsum += c->rmsWeight;
		}
		else
		{
			mean += fabs(diff);
			wsum += 1.0;
		}
	}

  return sqrt(mean/wsum);
}

//*******************************************************************************
double PearsonScorer::calcScore(size_t column, Cluster* cluster) const
//*******************************************************************************
{
	double r = 0.0;
	int N = cluster->getNPoints();

	//required sums with p and q
	double sum_pB = 0;
	double sum_square_pB = 0;
	double sum_pAB = 0;
	double sum_pA = 0;
	double sum_square_pA = 0;

	//loop over all corresponding point pairs
	corresponds_cit c;
	for(c=cluster->beginCorr();c!=cluster->endCorr();++c)
	{
		double pA = (*pData[0])(c->first, column);
		double pB = (*pData[1])(c->second, column);

		sum_pB += pB;
		sum_square_pB += pB * pB;

		sum_pA += pA;
		sum_square_pA += pA * pA;
					
		sum_pAB += pA * pB;

	}
		
	r = N*sum_pAB - sum_pA*sum_pB;  //calculating first part of r
	return 	r/sqrt(fabs(N*sum_square_pA - sum_pA*sum_pA) * 
								 fabs(N*sum_square_pB - sum_pB*sum_pB));
}

//*******************************************************************************
Scorer* RMSDScorerFactory(const RazorBack::Rectmat_<double>** data, 
													const string name, bool weighting)
//*******************************************************************************
{
	RMSDScorer* scorer = new RMSDScorer(data, name);
	scorer->enableWeighting(weighting);
	
	return scorer;
}

//*******************************************************************************
Scorer* MeanScorerFactory(const RazorBack::Rectmat_<double>** data, 
													const string name, bool weighting)
//*******************************************************************************
{
	MeanScorer* scorer = new MeanScorer(data, name);
	scorer->enableWeighting(weighting);
	
	return scorer;
}

//*******************************************************************************
Scorer* PearsonScorerFactory(const RazorBack::Rectmat_<double>** data, 
														 const string name)
//*******************************************************************************
{
	return new PearsonScorer(data, name);
}
