/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SCORER_H
#define SCORER_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#include <string>
#else
#include <string>
#endif

#include <Rectmat.hh>

class Cluster;

//*******************************************************************************
// C L A S S :  S C O R E R
//*******************************************************************************

/**
 * \brief the base class for all scoring functors
 *
 * The scorer functors are function classes that take a Cluster object and
 * perform a specific type of scoring based on a set of properties (pData) given
 * in the constructor of the scorer. The results of the scoring functions are
 * written back to the Cluster object, one scoring value for each column in the
 * data matrices.
 *
 * The scoring function must have a unique name, otherwise it would be
 * overwritten by or overwrite other scoring values in the clusters' scoring
 * maps. If the data matrices have more than one column a _#COL identifier is
 * appended to the scoring names.
 *
 * The class has one abstract method: calcScore(size_t). This method must be
 * reimplemented in the different Scorer types to calculate the actual scoring
 * functions for any specified column in the data matrices. 
 *
 * \note If the two matrices do not have the same number of columns scores are
 * only generated for the first n columns that are present in both matrices. The
 * number of scores generated is returned by the operatort()(Cluster*) method.
 */
class Scorer
{
public:
	
	//*****************************************************************************
	// CONSTRUCTORS
	//*****************************************************************************

	/**
	 * \brief generates a scoring function object
	 *
	 * The constructor of the scoring function objects stores the pointers to the
	 * data matrices and the name.
	 *
	 * \param data a two membered array containing constant pointers to the data 
	 *             matrices
	 *
	 * \param name the name of the scoring function
	 */
	Scorer(const RazorBack::Rectmat_<double>** data, const std::string name);

	//*****************************************************************************
	// OPERATORS
	//*****************************************************************************

	/**
	 * \brief starts the calculation of all scoring values and writes them to the cluster
	 *
	 * The function operator takes the \p cluster and calculates the score for
	 * each column in the data matrices. It finally adds the scores to the scoring
	 * map of the cluster. If there are more than one column in the matrices, the
	 * score for each column is specified by a "_#COL" suffix.
	 *
	 * \param cluster a pointer to a cluster object, that is subject to the scoring
	 *
	 * \return the number of scoring values written to the cluster. This value
	 * should be equal to the number of columns in the data matrices. If it is
	 * different one of the matrices has a different number of columns.
	 */
	int operator()(Cluster* cluster) const;

protected:

	//*****************************************************************************
	// PROTECTED METHODS
	//*****************************************************************************

	/**
	 * \brief calculates the score for a \p column
	 *
	 * This abstract method should return a score of the \p cluster regarding the
	 * \p column . It must be implemented in all subclasses of Scorer.
	 *
	 * \param column the number of the data matrix column
	 * \param cluster a pointer to a cluster object, that is subject to the scoring
	 *
	 * \return the scoring value for \p column.
	 */
	virtual double calcScore(size_t column, Cluster* cluster) const = 0;
	
	//*****************************************************************************
	// PROTECTED MEMEBERS
	//*****************************************************************************
	
	/**
	 * \brief the pointers to the two data matrices.
	 *
	 * The data matrices are the basis for the calculation of the scoring values
	 */
	const RazorBack::Rectmat_<double>* pData[2];

	/**
	 * \brief the name of that particular score
	 */
	std::string fName;
};

//*******************************************************************************
// C L A S S :  R M S D S C O R E R
//*******************************************************************************

/**
 * \brief RMSD scorer
 *
 * This scorer calculates the root mean square differences between the property
 * value at the corresponding points of the cluster. The scores are by default
 * weighted by the distance of the corresponding points in the harmonic map
 * overlap.
 *
 * \f[
 *     RMSD = \sqrt{ \frac{\sum\limits_{i=0}\limits^{N} w_i \cdot 
                           (p_{Ai} - p_{Bi})^2}
 *                        {\sum\limits_{i=0}\limits^{N} w_i}}
 * \f]
 *
 * \sa Scorer
 */
class RMSDScorer: public Scorer
{
public:

	//*****************************************************************************
	// CONSTRUCTOR
	//*****************************************************************************

	/**
	 * \brief generates a RMSD scoring function object
	 *
	 * The constructor of the RMSD scoring function objects stores the pointers to the
	 * data matrices and the name.
	 *
	 * \param data a two membered array containing constant pointers to the data 
	 *             matrices
	 * \param name the name of the scoring function
	 * 
	 * \sa Scorer::Scorer(const RazorBack::Rectmat_<double>**, const std::string)
	 */
	RMSDScorer(const RazorBack::Rectmat_<double>** data, const std::string name)
			:Scorer(data,name), fWeighting(true) { }

	//*****************************************************************************
	// PROPERTY METHODS
	//*****************************************************************************
	
	/**
	 * \brief retrieves the status of the weighting flag
	 *
	 * \return \c true if the RMSD is weighted by the overlap distances, \c false
	 * otherwise
	 *
	 * \sa fWeighting, enableWeighting(bool)
	 */
	bool isWeighting() const { return fWeighting; }
	/**
	 * \brief toggles the weighting flag
	 *
	 * \param weighting enables the overlap distance weighting if \c true,
	 *                  disables it if \p false.
	 *
	 * \sa fWeighting, isWeighting()
	 */
	bool enableWeighting(bool weighting = true) { fWeighting = weighting; }

protected:

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief calculates the RMSD for a \p column
	 *
	 * For the formula of the calculation look into the detailed description of
	 * the class RMSDScorer.
	 *
	 * \param column the number of the data matrix column
	 * \param cluster a pointer to a cluster object, that is subject to the scoring
	 *
	 * \return the RMSD value for \p column.
	 */
	virtual double calcScore(size_t column, Cluster* cluster) const;

	//*****************************************************************************
	// PRIVATE MEMBERS
	//*****************************************************************************

	/**
	 * \brief weighting
	 *
	 * If \c true (default) enables the weighting of the corresponding pairs by
	 * their distance in the harmonic map overlap.
	 *
	 * \sa isWeighting(), enableWeighting(bool)
	 */
	bool fWeighting;
};

//*******************************************************************************
// C L A S S :  M E A N S C O R E R
//*******************************************************************************

/**
 * \brief Mean scorer
 *
 * This scorer calculates the mean differences between the property value at the
 * corresponding points of the cluster. The scores are by default weighted by
 * the distance of the corresponding points in the harmonic map overlap.
 *
 * \f[
 *     Mean = \frac{\sum\limits_{i=0}\limits^{N} \left| p_{Ai} - p_{Bi} \right|}
 *                        {N}
 * \f]
 *
 * \sa Scorer
 */
class MeanScorer: public Scorer
{
public:

	//*****************************************************************************
	// CONSTRUCTOR
	//*****************************************************************************

	/**
	 * \brief generates a mean scoring function object
	 *
	 * The constructor of the mean scoring function objects stores the pointers to the
	 * data matrices and the name.
	 *
	 * \param data a two membered array containing constant pointers to the data 
	 *             matrices
	 * \param name the name of the scoring function
	 * 
	 * \sa Scorer::Scorer(const RazorBack::Rectmat_<double>**, const std::string)
	 */
	MeanScorer(const RazorBack::Rectmat_<double>** data, const std::string name)
			:Scorer(data,name), fWeighting(true) { }

	//*****************************************************************************
	// PROPERTY METHODS
	//*****************************************************************************
	
	/**
	 * \brief retrieves the status of the weighting flag
	 *
	 * \return \c true if the mean is weighted by the overlap distances, \c false
	 * otherwise
	 *
	 * \sa fWeighting, enableWeighting(bool)
	 */
	bool isWeighting() const { return fWeighting; }
	/**
	 * \brief toggles the weighting flag
	 *
	 * \param weighting enables the overlap distance weighting if \c true,
	 *                  disables it if \p false.
	 *
	 * \sa fWeighting, isWeighting()
	 */
	bool enableWeighting(bool weighting = true) { fWeighting = weighting; }

protected:

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief calculates the mean for a \p column
	 *
	 * For the formula of the calculation look into the detailed description of
	 * the class MeanScorer.
	 *
	 * \param column the number of the data matrix column
	 * \param cluster a pointer to a cluster object, that is subject to the scoring
	 *
	 * \return the mean value for \p column.
	 */
	virtual double calcScore(size_t column, Cluster* cluster) const;

	//*****************************************************************************
	// PRIVATE MEMBERS
	//*****************************************************************************

	/**
	 * \brief weighting
	 *
	 * If \c true (default) enables the weighting of the corresponding pairs by
	 * their distance in the harmonic map overlap.
	 *
	 * \sa isWeighting(), enableWeighting(bool)
	 */
	bool fWeighting;
};


//*******************************************************************************
// C L A S S :  P E A R S O N S C O R E R
//*******************************************************************************

/**
 * \brief Pearson correlation scorer
 *
 * This scorer calculates the pearson correlation coefficient \em R between the
 * property values at the corresponding points in the cluster.
 *
 * \f[
 *     R = \frac{ N \, \sum\limits_{i=1}\limits^{N}p_{Ai}p_{Bi} -
 *               \sum\limits_{i=1}\limits^{N}p_{Ai} \sum\limits_{i=1}
                 \limits^{N}p_{Bi}}
 *              { \sqrt{\left| N \, \sum\limits_{i=1}\limits^{N}p_{Ai}^2 -
 *                 \left(\sum\limits_{i=1}\limits^{N}p_{Ai}\right)^2\right| \,
 *             \left| N \, \sum\limits_{i=1}\limits^{N}p_{Bi}^2 -
 *       \left(\sum\limits_{i=1}\limits^{N}p_{Bi}\right)^2\right|}}.
 * \f]
 *
 * \sa Scorer
 */
class PearsonScorer: public Scorer
{
public:

	//*****************************************************************************
	// CONSTRUCTOR
	//*****************************************************************************

	/**
	 * \brief generates a pearson scoring function object
	 *
	 * The constructor of the pearson scoring function objects stores the pointers
	 * to the data matrices and the name.
	 *
	 * \param data a two membered array containing constant pointers to the data 
	 *             matrices
	 * \param name the name of the scoring function
	 * 
	 * \sa Scorer::Scorer(const RazorBack::Rectmat_<double>**, const std::string)
	 */
	PearsonScorer(const RazorBack::Rectmat_<double>** data, const std::string name)
			:Scorer(data,name) { }

protected:

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief calculates the pearson correlation coefficient for a \p column
	 *
	 * For the formula of the calculation look into the detailed description of
	 * the class PearsonScorer.
	 *
	 * \param column the number of the data matrix column
	 * \param cluster a pointer to a cluster object, that is subject to the scoring
	 *
	 * \return the pearson correlation coefficient value for \p column.
	 */
	virtual double calcScore(size_t column, Cluster* cluster) const;

};

//*******************************************************************************
// SCORER FACTORY FUNCTIONS
//*******************************************************************************


/**
 * \brief Factory function for RMSD scorers
 * 
 * The function returns a pointer to a Scorer class that represents an RMSD
 * scorer.
 *
 * \param data      a two membered array containing constant pointers to the data 
 *                  matrices
 * \param name the  name of the scoring function
 * \param weighting specifies if overlap distance weighting should be enabled 
 *                  (default)
 * 
 * \return a pointer to a Scorer object
 *
 * \sa RMSDScorer, Scorer
 */
Scorer* RMSDScorerFactory(const RazorBack::Rectmat_<double>** data, 
													const string name, bool weighting = true);

/**
 * \brief Factory function for Mean scorers
 * 
 * The function returns a pointer to a Scorer class that represents an Mean
 * scorer.
 *
 * \param data      a two membered array containing constant pointers to the data 
 *                  matrices
 * \param name the  name of the scoring function
 * \param weighting specifies if overlap distance weighting should be enabled 
 *                  (default)
 * 
 * \return a pointer to a Scorer object
 *
 * \sa MeanScorer, Scorer
 */
Scorer* MeanScorerFactory(const RazorBack::Rectmat_<double>** data, 
													const string name, bool weighting=true);

/**
 * \brief Factory function for Mean scorers
 * 
 * The function returns a pointer to a Scorer class that represents an pearson 
 * scorer.
 *
 * \param data      a two membered array containing constant pointers to the data 
 *                  matrices
 * \param name the  name of the scoring function
 * 
 * \return a pointer to a Scorer object
 *
 * \sa PearsonScorer, Scorer
 */
Scorer* PearsonScorerFactory(const RazorBack::Rectmat_<double>** data, 
														 const string name);

#endif // SCORER_H

