/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "singlelink.hh"

#include "utils.hh"
#include <xmlutils.hh>
#include "scorecollection.hh"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <string>
#  else
#    include <map.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <map>
#  include <string>
#endif //HAVE_CONFIG_H

#include <xercesc/dom/DOM.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLStringTokenizer.hpp>

#include <trimesh.hh>
#include <float.h>

#ifdef HAVE_NAMESPACES
using namespace std;
XERCESNS_USE
#endif

typedef cpair_t<const MergeNode<SingleLink::clusters_it>*> distpair_t;

//*******************************************************************************
SingleLink::SingleLink(list<Cluster>* clusterlist, linkage_t linkage)
//*******************************************************************************
:lClusters(0), lNodes(), mDistances(), fLinkage(linkage)
{
	init(clusterlist);
}

//*******************************************************************************
SingleLink::~SingleLink()
//*******************************************************************************
{
	for(nodes_it n=lNodes.begin();n!=lNodes.end();)
	{
		nodes_it rm = n++;
		MergeNode<clusters_it>* node = *rm;
		
		//remove the node from the list and delete it
		lNodes.erase(rm);
		delete node;
	}
}
	
//*******************************************************************************
void SingleLink::init(list<Cluster>* clusterlist)
//*******************************************************************************
{
	lClusters = clusterlist;

	for(clusters_it c=lClusters->begin();c!=lClusters->end();++c)
	{
		//create a new MergeNode for the new cluster
		MergeNode<clusters_it>* node = new MergeNode<clusters_it>(0,0, c);

		//calculate the distances between the new cluster and all old clusters
		for(nodes_cit n=lNodes.begin();n!=lNodes.end();++n)
			calcCrossOverlay(*n, node);

		lNodes.push_back(node);
	}
}
	
//*******************************************************************************
size_t SingleLink::doSingleLinkage(const ScoreCollection* scorings)
//*******************************************************************************
{
	if (lNodes.size()>0)
	{
		/* INITIALIZATION */
		pair<nodes_it, nodes_it> merge;

		//retrieve the surfaces 
		Trimesh* pSurfaces[2];
		pSurfaces[0] = lNodes.front()->getCluster()->getSurface(0);
		pSurfaces[1] = lNodes.front()->getCluster()->getSurface(1);


		/* LOOP: Perform as long as there are two mergable clusters */
		while ((merge = findClosestPair()).first!=lNodes.end())
		{
			//retrieve the nodes
			pair<MergeNode<clusters_it>*, MergeNode<clusters_it>*>
				pNodes = make_pair(*merge.first, *merge.second);

			//retrieve the pairs of clusters
			pair<clusters_cit, clusters_cit> mergeCs((*merge.first)->getCluster(),
																							 (*merge.second)->getCluster());
		
			/* CLUSTER */
			//add a new cluster as the union of both selected clusters
			clusters_it merge_cluster = lClusters->insert(lClusters->end(),
																										Cluster(pSurfaces));
			merge_cluster->merge(*(mergeCs.first),*(mergeCs.second));
			merge_cluster->compCorresponds();
			merge_cluster->doRMSFit(true);
			//scoring
			if (scorings)	(*scorings)(&(*merge_cluster));

			/* NODE */
		
			//create a new merge node
			MergeNode<clusters_it>* node = 
				new MergeNode<clusters_it>(pNodes.first, pNodes.second, merge_cluster);

			//purge the correspondent point pairs from the merged clusters
			(*merge.first)->getCluster()->clearCorresponds();
			(*merge.second)->getCluster()->clearCorresponds();
		
			//remove the merged node from the nodelist
			lNodes.erase(merge.first);
			lNodes.erase(merge.second);

			/* REMOVE AND BUILD DISTANCES */

			//1. remove all distances between the merged clusters and the valid 
      //   clusters
			//2. add the distance to the new cluster to the distance map
			for(nodes_cit n=lNodes.begin();n!=lNodes.end();++n)
			{
				//remove
				distpair_t left(pNodes.first, *n);
				distpair_t right(pNodes.second,*n);
			
				mDistances.erase(left);
				mDistances.erase(right);

				//add
				calcCrossOverlay(node, *n);
			}
			//remove the pairwise distance 
			mDistances.erase(distpair_t(pNodes.first, pNodes.second));

			lNodes.push_back(node);
		}
	}
}

//*******************************************************************************
bool SingleLink::writeNode(const MergeNode<clusters_it>* node, 
													 DOMElement* parent) const
//*******************************************************************************
{
	DOMElement* cluster = node->getCluster()->writeToDOM(parent);

	if (node->getLeft()) writeNode(node->getLeft(),  cluster);
	if (node->getRight()) writeNode(node->getRight(), cluster);
}

//*******************************************************************************
bool SingleLink::writeDendrogramm(DOMElement* parent) const
//*******************************************************************************
{
	/* INITIALIZATION */
	bool success = true;
	

	DOMDocument* doc = parent->getOwnerDocument();

	/* DENDROGRAMM */
	DOMElement* dendronode = XMLUtils::createElement(parent, "dendrogramm");
	dendronode->setAttribute(XMLStr("roots"), 
													 XMLStr().fromValue((int)lNodes.size()));

	for(nodes_cit n=lNodes.begin();n!=lNodes.end();++n)
		success = success & writeNode(*n, dendronode);

	return success;
}

//*******************************************************************************
MergeNode<SingleLink::clusters_it>* 
SingleLink::readNode(const DOMElement* node, Trimesh** surfaces)
//*******************************************************************************
{
	MergeNode<clusters_it>* mnode = new MergeNode<clusters_it>;
	MergeNode<clusters_it>* childnodes[2];
	childnodes[0] = childnodes[1] = 0;
	int i;

	clusters_it cluster = lClusters->insert(lClusters->end(), Cluster(surfaces));
	cluster->readFromDOM(node);
	
	//iterate through all the child elements list
	DOMNode* child = node->getFirstChild();
	i = 0;
	while ((child) && (i<2))
	{
		if (const DOMElement* cluster = XMLUtils::elementByName(child, "cluster"))
			childnodes[i++] = readNode(cluster, surfaces);
		child = child->getNextSibling();
	}
		
	mnode->init(childnodes[0], childnodes[1], cluster);

	return mnode;
}

//*******************************************************************************
int SingleLink::readDendrogramm(const DOMElement* node, Trimesh** surfaces)
//*******************************************************************************
{
	if (!(XMLStr("dendrogramm") == XMLStr(node->getNodeName())))
		return false;

	DOMNode* child = node->getFirstChild();
	while (child)
	{
		if (const DOMElement* cluster = XMLUtils::elementByName(child, "cluster"))
			lNodes.push_back(readNode(cluster, surfaces));
		child = child->getNextSibling();
	}
}


//*******************************************************************************
bool SingleLink::checkConflicts(const MergeNode<clusters_it>* A, 
																const MergeNode<clusters_it>* B)
//*******************************************************************************
{
	//intializing
	typedef map<Trimesh::vertex_t, Trimesh::vertex_t>::const_iterator lookup_cit;

	map<Trimesh::vertex_t, Trimesh::vertex_t> lookup[2];
	bool conflict = false;
	Cluster::pairs_cit check;

	//setup the lookup table with the pairs of cluster A on surface I
	for(Cluster::pairs_cit p=A->getCluster()->beginPairs();
			p!=A->getCluster()->endPairs();++p)	
		lookup[0][p->A] = p->B;

	//setup the lookup table with the pairs of cluster A on surface II
	for(Cluster::pairs_cit p=A->getCluster()->beginPairs();
			p!=A->getCluster()->endPairs();++p)	
		lookup[1][p->B] = p->A;
	
	//check the pair sequence of cluster B
	check = B->getCluster()->beginPairs();
	while (!conflict && (check!=B->getCluster()->endPairs()))
	{
		lookup_cit l = lookup[0].find(check->A);
		if (l!=lookup[0].end()) conflict = (l->second!=check->B);

		if (conflict) break;
			
		l = lookup[1].find(check->B);
		if (l!=lookup[1].end()) conflict = (l->second!=check->A);

		++check;
	}
	
	return !conflict;
}

//*******************************************************************************
double SingleLink::calcCrossOverlay(const MergeNode<clusters_it>* A, 
																		const MergeNode<clusters_it>* B)
//*******************************************************************************
{
	double distance = 0.0;

	//determine the distance between both clusters
	if (checkConflicts(A,B))
	{
		pair<double, double> rmsd;
		rmsd.first =A->getCluster()->calcRMSD(B->getCluster()->getTransformation());
		rmsd.second=B->getCluster()->calcRMSD(A->getCluster()->getTransformation());

		switch (fLinkage)
		{
			case slSingle:    distance=(rmsd.first<rmsd.second)?rmsd.first:rmsd.second; 
				break;
			case slComplete : distance=(rmsd.first>rmsd.second)?rmsd.first:rmsd.second;
				break;
			case slAverage :  distance=(rmsd.first+rmsd.second)/2;
				break;
		}
	}
	else
		distance = -1.0;

	//add the distance to the distance map
	distpair_t cpair(A,B);
	mDistances.insert(make_pair(cpair,distance));;

	return distance;
}

//*******************************************************************************
pair<SingleLink::nodes_it, SingleLink::nodes_it> 
SingleLink::findClosestPair()
//*******************************************************************************
{
	/* INITIALIZATION */
	pair<nodes_it, nodes_it> result = make_pair(lNodes.end(),lNodes.end());
	nodes_it l,r;
	double mindist = DBL_MAX;
		 
	//check all distances for the smallest one
	for(l=lNodes.begin();l!=lNodes.end();++l)
	{
		nodes_it help = l;

		for(r=(++help);r!=lNodes.end();++r)
		{
			double distance = mDistances[distpair_t(*l,*r)];
			
			if ((distance>0) && (distance<mindist))
			{
				mindist = distance;
				result.first = l;
				result.second = r;
			}
		}
	}
	
	return result;
}
