/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SINGLELINK_HH
#define SINGLELINK_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <list>
#  else
#    include <map.h>
#    include <list.h>
#  endif
#  include <map>
#  include <list>
#endif 

#include "cluster.hh"
#include "xercesns.hh"

XERCES_FORWARD(DOMElement);
class ScoreCollection;

/**
 * \brief the type of linkage
 */
enum linkage_t 
{
	slSingle,   ///< single linkage clustering : distance is the minimum RMSD
	slComplete, ///< complete linkage clustering: distance is the maximum RMSD
	slAverage   ///< average linkage clustering: distance is the average between
							///the maximum and the minimum RMSD
};

/**
 * \brief a single entry in the distance matrix of the single linkage clustering
 * algorithm.
 *
 * The structure represents a commutative pair of indices. Its 
 * operator==(const cpair_t&) returns true if either A == A1 and B == B1 or if 
 * A == B1 and B == A1.
 *
 * The template parameter Ix must be equal-to comparable.
 */
template<class Ix> class cpair_t
{
public:

	Ix A, B;
	
	cpair_t(Ix A_, Ix B_)
	{
		A = (A_<B_) ? A_ : B_;
		B = (A_<B_) ? B_ : A_;
	}

	cpair_t(const cpair_t& src): A(src.A), B(src.B) { }

	bool operator<(const cpair_t other) const
	{
		return ((A<other.A) || ((A==other.A) && (B<other.B)));
	}
		

};

// //*******************************************************************************
// template<class Ix> bool operator<(const cpair_t<Ix>& l, const cpair_t<Ix>& r)
// //*******************************************************************************
// {
// 	return ((l.A<r.A) || ((l.A==r.A) && (l.B<r.B)));
// }

/**
 * \brief a merged cluster, representing a node in the dentrogram
 *
 *
 * The real merged clusters are Cluster objects but they are grouped together in
 * a dendrogramm - a binary tree - by the single linkage clustering. Each merged
 * cluster is a node in this tree and has a pointer to the real alignment
 * cluster object and to the two (merged) cluster nodes that formed the
 * alignment cluster.
 *
 * To score the merged clusters their RMSD is used. Thus each merged cluster
 * node contains the RMSD value of the produced alignment (getRMSD()).
 *
 * If the MergeNode is destroyed its left and right child nodes are also
 * destroyed. Therefore it is recommended not to use any other pointers to the
 * MergeNode outside the tree.
 *
 * The Cluster objects are referenced by an iterator like type that is passed as
 * the
 */
template<class Ix> class MergeNode
{
public:

	//*****************************************************************************
	// CONSTRUCTOR, INIT & DESTRUCTOR
	//*****************************************************************************

	/**
	 * \brief standard constructor
	 *
	 * The standard constructor initializes all data members with invalid values.
	 */
	MergeNode() : pLeft(0), pRight(0) { }

	/**
	 * \brief initializing constructor
	 *
	 * The initializing constructor sets the pointers to the left and right
	 * constructing alignment clusters and to the merged alignment cluster.
	 *
	 * The method simply calls the 
	 * init(MergeNode*, MergeNode*, Ix) method.
	 *
	 * \param left,right pointers to the Mergenode objects that build the merged
	 *                   cluster.
	 * \param merged     pointer to the merged Cluster object
	 */
	MergeNode(MergeNode* left, MergeNode* right, Ix merged);

	/**
	 * \brief copy constructor
	 */
	MergeNode(const MergeNode& src)
			: pLeft(src.pLeft), pRight(src.pRight), pMerged(src.pMerged) { }
	
	/**
	 * \brief destructor
	 *
	 * The destructor destroys the left and right mergenode
	 */
	~MergeNode();

  /**
	 * \brief initialization routine
	 *
	 * see MergeNode(MergeNode*, MergeNode*, Ix).
	 *
	 * \param left,right pointers to the Cluster object that build the merged
	 *                   cluster.
	 * \param merged     pointer to the merged Cluster object
	 */
	void init(MergeNode* left, MergeNode* right, Ix merged);

	//*****************************************************************************
	// PROPERTY METHODS
	//*****************************************************************************

	/**
	 * \name traversion methods 
	 */
	//@{
	/**
	 * \brief return a constant or non constant pointer to the left or right child
	 * MergeNode object
	 *
	 * \return left/right building mergenode, or NULL if this node is a
	 * leaf. There is a constant and non-constant implementation
	 *
	 * \sa getRight(), getLeft() const, getRight() const
	 */
	const MergeNode*  getLeft() const { return pLeft; }
	const MergeNode*  getRight() const { return pRight; }

	MergeNode*  getLeft() { return pLeft; }
	MergeNode*  getRight() { return pRight; }
	//@}

	/**
	 * \brief returns a const pointer to the merged Cluster object
	 *
	 * \return a constant pointer to the merged Cluster object
	 */
	const Ix getCluster() const { return pMerged; }

	/**
	 * \brief returns a pointer to the merged Cluster object
	 *
	 * \return a pointer to the merged Cluster object
	 */
	Ix getCluster() { return pMerged; }

	/**
	 * \brief returns the RMSD value of the merged cluster
	 *
	 * \return the RMSD value of the merged cluster
	 */
	double getRMSD() const { return (pMerged) ? pMerged->getRMSD() : -1.0; }

private:

	MergeNode* pLeft;
	MergeNode* pRight;
	/**<
	 * \brief pointer to the left and right merged cluster node
	 */

	/**
	 * \brief pointer to the merged Cluster object
	 */
	Ix pMerged;
};


/**
 * \brief a hirarchical linkage clustering for alignment clusters on the surface
 *
 * The surface comparison method produces a set of alignment clusters this class
 * groups them together and produces new, larger classes with a hirarchical linkage
 * algorithm: 
 *
 * - For each pair of clusters, that do not have ambigous correspondencies, the
 * two RMSDs for both cross overlays are calculated and the smaller value is
 * stored as the distance between the two alignment clusters. A cross overlay is
 * defined as the attampt to lay the correspondent point pairs of one cluster
 * over each other by the transformation matrix of the other cluster. \n
 * (distances for cluster pairs that have ambigous correspondencies are set to
 * -1)
 *
 * - The closest two clusters are merged to a new alignment cluster which
 * contains the union of both patchpair sets. For this new alignment cluster the
 * correspondent point pairs and the transformation matrix are calculated, and
 * the RMSD of the overlay is stored as a score.
 *
 * - All entries in the distance matrix for the two collapsed clusters are
 * removed and distances are calculated between the new and the remaining
 * clusters.
 *
 * - The algorithm is continued until either only one merged cluster remains or
 * no more merges can be performed because of ambigous correspondencies.
 */
class SingleLink
{
public:


	//*****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//*****************************************************************************

	typedef list<Cluster> clusters_t;
	typedef list<Cluster>::iterator clusters_it;
	typedef list<Cluster>::const_iterator clusters_cit;
	/**<
	 * \brief a type for the cluster list
	 */

	typedef list<MergeNode<clusters_it>*> nodes_t;
	typedef list<MergeNode<clusters_it>*>::iterator nodes_it;
	typedef list<MergeNode<clusters_it>*>::const_iterator nodes_cit;
	/**<
	 * \brief nodelist type plus iterators
	 */

  //*****************************************************************************
	// CONSTRUCTOR, INIT AND DESTRUCTOR
	//*****************************************************************************

	/**
	 * \brief standard, non-initializing constructor
	 */
	SingleLink() : lClusters(0), lNodes(), mDistances(), fLinkage(slSingle) {}

	/**
	 * \brief initializing constructor
	 *
	 * This constructor initializes the hirarchical linkage class completely by calling
	 * init(list<Cluster>*).
	 *
	 * \param clusterlist a pointer to the list of all alignment clusters that
	 *                    form the basis of the hirarchical linkage.
	 * \param linkage     the type of linkage (single, complete or average)
	 */
	SingleLink(list<Cluster>* clusterlist, linkage_t linkage = slSingle);

	/**
	 * \brief destructor
	 *
	 * the destructor destroys all the MergeNode elements left in the lNodes list.
	 */
	~SingleLink();
	
	/**
	 * \brief initializer
	 *
	 * The initializer creates a MergeNode for each cluster in \p clusterlist and
	 * starts the calculation of the distance map.
	 *
	 * \param clusterlist a pointer to the list of all alignment clusters that
	 *                    form the basis of the hirarchical linkage.
	 */
	void init(list<Cluster>* clusterlist);
	
	//*****************************************************************************
	// OPERATION METHODS
	//*****************************************************************************

	/**
	 * \brief starts the hirarchical linkage clustering
	 *
	 * This method starts the hirarchical linkage clustering producing a series of
	 * merged alignment clusters as described in the description of this class.
	 *
	 * \param scorings a pointer to a ScoreCollection that is used to score the
	 *                 merged clusters. If NULL (default) no scoring is performed.
	 *
	 * \return the number of found top-level clusters
	 */
	size_t doSingleLinkage(const ScoreCollection* scorings = 0);

	//*****************************************************************************
	// XML ROUTINES
	//*****************************************************************************
	
	/**
	 * \brief writes the dendrogramm built from the hirarchical linkage under the given
	 * XML node.
	 *
	 * The method stores the cluster hirarchy build by the hirarchical linkage in a
	 * tree structure under the passed XML node \p parent.
	 *
	 * The clusters are stored in a tree, where the two clusters that form a new
	 * cluster are stored as child nodes of the merged cluster. This is an
	 * extention to the XML structure of the single nodes as introduced in
	 * Cluster.
	 *
	 * \param parent the parent XML node for the dendrogramm.
	 *
	 * \sa readDendrogramm(DOMElement*)
	 */
	bool writeDendrogramm(XERCESNS DOMElement* parent) const;

	/**
	 * \brief reads a dendrogramm built from the hirarchical linkage out of the given
	 * XML node.
	 *
	 * The dendrogramm must be present in the child nodes of the given parent node
	 * \p parent in the form specified by writeDendrogramm(DOMElement*).
	 *
	 * \param node   the parent XML node for the dendrogramm.
	 * \param surfaces a two membered field of pointers to the Trimesh structures
	 *
	 * \return the experiment id or 0 if an error occured
	 *
	 * \sa writeDendrogramm(DOMElement*)
	 */
	int readDendrogramm(const XERCESNS DOMElement* node, Trimesh** surfaces);

  //*****************************************************************************
	// PROPERTY METHODS
	//*****************************************************************************

	nodes_it  begin()       { return lNodes.begin(); }
	nodes_cit begin() const { return lNodes.begin(); }
	/**< constant and mutable iterators pointing to the first merged cluster
	 *        in the nodelist
	 *
	 * \return Before the algorithm has been started with the doSingleLinkage()
	 * this will give you the first base alignment cluster. After the algorithm
	 * this will point to the first top-level cluster.
	 */

	nodes_it  end()       { return lNodes.end(); }
	nodes_cit end() const { return lNodes.end(); }
	/**<constant and mutable iterators pointing behind the last merged 
	 *        cluster in the nodelist
	 *
	 * \return Before the algorithm has been started with the operator()() this
	 * will give you the end of the base alignment cluster sequence. After the
	 * algorithm this will point to the end top-level cluster sequence.
	 */

	/**
	 * \brief retrieves the number of nodes in the node list
	 *
	 * \return the number of base alignment clusters before the algorithm has been
	 * performed, the number of top-level clusters after the hirarchical linkage.
	 */
	size_t nClusters() const { return lNodes.size(); }

private:
	
	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief writes a single node and recursively its childs into a XML structure
	 *
	 * The function is used to write the dendrogramm tree into a XML structure.
	 *
	 * \param node       the node of the dendrogram
	 * \param parent     parent XML node
	 */
	bool writeNode(const MergeNode<clusters_it>* node, 
								 XERCESNS DOMElement* parent) const;

	/**
	 * \brief reads a single node and recursively its childs from a XML structure
	 *
	 * The function is used to read the dendrogramm tree from a XML structure.
	 *
	 * \param node parent XML node
	 * \param surfaces a two membered field of pointers to the Trimesh structures
	 * 
	 * \return a pointer to the MergeNode created from the XML element
	 */
	MergeNode<clusters_it>* readNode(const XERCESNS DOMElement* node, 
																	 Trimesh** surfaces);


	/**
	 * \brief checks if a conflict exists between two clusters
	 *
	 * The function checks if the two clusters (\p A and \p B) contain any
	 * ambigous combination of patch pairs. E.g. cluster A assigns patch 209 on
	 * one to patch 286 on the other surface but cluster B assigns the same patch
	 * 209 to another patch 567.
	 *
	 * \param A,B pointers to a merge cluster node
	 *
	 * \return \c true if no conflicts have been found, \c false otherwise.
	 */
	bool checkConflicts(const MergeNode<clusters_it>* A, 
											const MergeNode<clusters_it>* B);
  /**
	 * \brief calcualtes the distance between two clusters and stores it in
	 *        the distance map.
	 *
	 * The distance between two clusters is calculated as the RMSD of the a cross
	 * overlay of one alignment cluster with the transformation matrix of the
	 * other cluster and v.v.. The minimum value is stored as the distance between
	 * both clusters.
	 *
	 * \param A,B pointers to a merge cluster node
	 *
	 * \return the minimum RMSD value of the cross overlay
	 */
	double calcCrossOverlay(const MergeNode<clusters_it>* A, 
													const MergeNode<clusters_it>* B);

	/**
	 * \brief detects the closest pair of clusters
	 *
	 * The method returns the closest pair of clusters based on the distance map
	 * mDistances. If all remaining pairs conflict with each other, twice the end
	 * of sequence is returned.
	 * 
	 */
	pair<nodes_it, nodes_it> findClosestPair();

	//*****************************************************************************
	// PRIVATE TYPE DEFINITIONS
	//*****************************************************************************

	typedef std::map<cpair_t<const MergeNode<clusters_it>*>, double> distmap_t;
	typedef distmap_t::iterator distmap_it;
	typedef distmap_t::const_iterator distmap_cit;
	/**< 
	 * \brief type of the distance map plus iterators
	 */

  //*****************************************************************************
	// DATA MEMBERS
	//*****************************************************************************

	/**
	 * \brief a pointer to a list of Cluster objects
	 *
	 * This list is the storage for the single and merged alignment clusters. The
	 * dendrogramm that is constructed by this algorithm uses only iterators into
	 * that list.
	 */
	list<Cluster>* lClusters;

	/**
	 * \brief the node list
	 *
	 * The list of merged clusters contains that root of the single merged cluster
	 * trees as they are evolving. At the end all nodes in this list are roots
	 * of alternative alignments.
	 */
	nodes_t lNodes;

	/**
	 * \brief the distance map for the hirarchical linkage decission
	 */
	distmap_t mDistances;

	/**
	 * \brief specifies the type of linkage
	 */
	linkage_t fLinkage ;
	
};

#ifdef INCLUDE_TMPL_DEFS
#  include "singlelink_templ.cc"
#endif

#endif //SINGLELINK_HH
