/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "utils.hh"
#include <fileformats.hh>
#include <stdlib.h>
#include <stdio.h>


#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <fstream>
#    include <strstream>
#    include <vector>
#  else
#    include <fstream.h>
#    include <strstream.h>
#    include <vector.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <fstream>
#  include <strstream>
#  include <vector>
#endif //HAVE_CONFIG_H

//*******************************************************************************
Trimesh* readSurf_DB(MYSQL* con, const int surface_id, 
										 double& probe, double& density)
//*******************************************************************************
{
	probe = 0.0;
	density = 0.0;

	Trimesh* mesh = 0;
	
	//parameter conversion
	char convert[50];
	sprintf(convert,"%d", surface_id);
	
  //execute the query
	string query = "SELECT surface, probe, density FROM ";
	query += SURFS_TABLE;
	query += " WHERE id = ";
	query += convert; 
	query += ";";
	mysql_query(con, query.c_str());

	//examine the result
	MYSQL_RES* res = mysql_use_result(con);
	if (res)
	{
		//retrieve the field lengths
		unsigned long *ll;
		
		MYSQL_ROW row = mysql_fetch_row(res);
		ll = mysql_fetch_lengths(res);

		//retrieve the surface BLOB data
		if ((ll[0]>0) && (row))
		{
			
			//fill the stringstream with the BLOB data and set the seek to the front
			stringbuf streambuffer;
//			dynamic_cast<streambuf*>(&streambuffer)->setbuf(row[0],ll[0]);
			streambuffer.pubsetbuf(row[0],ll[0]);
			istream blobstream(&streambuffer);
			
			probe = atof(row[1]);
			density = atof(row[2]);

			mesh=readMesh(blobstream, bin_format, false);
		}

		mysql_free_result(res);
	}
	else
		cerr << "DB Error(" << mysql_errno(con) << "): " << mysql_error(con);

	return mesh;
}

//*******************************************************************************
Vector_<double> readCenter_DB(MYSQL* con, const int center_id)
//*******************************************************************************
{
	Vector_<double> result(1,-1);

	//parameter conversion
	char id_[50];
	sprintf(id_,"%d", center_id);

	string query = "SELECT x, y, z, r FROM  ";
	query+= CENTERS_TABLE;
	query+= " WHERE id= ";
	query+= id_;
	query+= ";";
	
	mysql_query(con, query.c_str());
	
	MYSQL_RES* res = mysql_store_result(con);
	if (res)
	{
		if (mysql_num_rows(res)>0)
		{
			MYSQL_ROW row = mysql_fetch_row(res);

			result.dim(4);
			for(int i=0;i<4;++i)
				result[i] = atof(row[i]);
		}

		mysql_free_result(res);
	}
	else
		cerr << "DB Error(" << mysql_errno(con) << "): " << mysql_error(con);
	
	return result;
}

//*******************************************************************************
int writeSurf_DB(MYSQL* con, Trimesh* trimesh,
								 const double probe, const double density, const int level)
//*******************************************************************************
{
	string query = "";
	
	//parameter conversion
	char probe_[50];
	sprintf(probe_,"%f", probe);
	char density_[50];
	sprintf(density_,"%f", density);
	char level_[50];
	sprintf(level_,"%d", level);
	char points_[50];
	sprintf(points_,"%d", trimesh->nVertices());

	//fill the stream with the surfacePatch and reset it to the beginning
	stringstream blobstream(stringstream::in | stringstream::out | 
													stringstream::binary);
	writeMesh(trimesh, blobstream, bin_format);
	size_t len = blobstream.tellp();
	blobstream.seekg(0,ios::beg);

	//get out the blob data and create an escaped \0 terminated string
	string blob = blobstream.str();
	
	char* escaped_blob = new char[blob.size()*2+1];
	mysql_real_escape_string(con, escaped_blob,blob.c_str(),blob.size());	

	//build the query string
	query += "INSERT INTO ";
	query += SURFS_TABLE;
	query += " ";
	query += "(probe, density, level, points, surface) VALUES(";
	query += "(";
	query += probe_;
	query += ", ";
	query += density_;
	query += ", ";
	query += level_;
	query += ", ";
	query += points_;
	query	+= ", ";
	query += "'";
	query += escaped_blob;
	query += "');";

	//remove the escaped blob string
	delete[] escaped_blob;

	//execute the query;
	mysql_query(con, query.c_str());
	
	if (mysql_errno(con))
	{
		cerr << "DB Error(" << mysql_errno(con) << "): " << mysql_error(con);
		return 0;
	}
	else
		return mysql_insert_id(con);
	
}

//*******************************************************************************
Trimesh* readPatch_DB(MYSQL* con, 
											const int surface_id, const int center, const double r)
//*******************************************************************************
{
	Trimesh* mesh = 0;
	
	//parameter conversion
	char surface_[50];
	sprintf(surface_,"%d", surface_id);
	char center_[50];
	sprintf(center_,"%d", center);
	char radius_[50];
	sprintf(radius_,"%f", r);
	
  //execute the query
	string query = "SELECT patch FROM ";
	query += PATCH_TABLE;
	query += " WHERE surface = ";
	query += surface_;
	query += " and center = ";
	query += center_;
	query += " and radius = ";
	query += r;
	query += ";";

	mysql_query(con, query.c_str());

	//examine the result
	MYSQL_RES* res = mysql_store_result(con);
	if (res)
	{
		//retrieve the field lengths
		unsigned long *ll;
		MYSQL_ROW row = mysql_fetch_row(res);
		ll = mysql_fetch_lengths(res);

		//retrieve the surface BLOB data
		if ((ll[0]>0) && (row))
		{
			//fill the stringstream with the BLOB data and set the seek to the front
			stringbuf streambuffer;
// 			dynamic_cast<streambuf*>(&streambuffer)->setbuf(row[0],ll[0]);
			streambuffer.pubsetbuf(row[0],ll[0]);
			istream blobstream(&streambuffer);
		
			mesh=readMesh(blobstream, bin_format, false);
		}
		

		mysql_free_result(res);
	}
	else
	{
		cerr << "DB Error(" << mysql_errno(con) << "): " << mysql_error(con);
	}

	return mesh;
}

//*******************************************************************************
Trimesh* readPatch_DB(MYSQL* con, const int patch_id)
//*******************************************************************************
{
	Trimesh* mesh = 0;
	
	//parameter conversion
	char patch_[50];
	sprintf(patch_,"%d", patch_id);
	
  //execute the query
	string query = "SELECT patch FROM ";
	query += PATCH_TABLE;
	query += " WHERE id = ";
	query += patch_;
	query += ";";

	mysql_query(con, query.c_str());

	//examine the result
	MYSQL_RES* res = mysql_store_result(con);
	if (res)
	{
		//retrieve the field lengths
		unsigned long *ll;
		MYSQL_ROW row = mysql_fetch_row(res);
		ll = mysql_fetch_lengths(res);

		//retrieve the surface BLOB data
		if ((ll[0]>0) && (row))
		{
			//fill the stringstream with the BLOB data and set the seek to the front
			stringbuf streambuffer;
// 			dynamic_cast<streambuf*>(&streambuffer)->setbuf(row[0],ll[0]);
			streambuffer.pubsetbuf(row[0],ll[0]);
			istream blobstream(&streambuffer);
		
			mesh=readMesh(blobstream, bin_format, false);
		}
		

		mysql_free_result(res);
	}
	else
	{
		cerr << "DB Error(" << mysql_errno(con) << "): " << mysql_error(con);
	}
	
	return mesh;	
}

//*******************************************************************************
int writePatch_DB(MYSQL* con, surfacePatch* patch, 
									const int parent, const double radius, const int center)
//*******************************************************************************
{
	string query = "";
	MYSQL_RES* res;
	MYSQL_ROW row;

	int id = 0;

	
	//parameter conversion
	char parent_[50];
	sprintf(parent_,"%d", parent);
	char center_[50];
	sprintf(center_,"%d", center);
	char radius_[50];
	sprintf(radius_,"%f", radius);
	char points_[50];
	sprintf(points_,"%d", patch->nVertices());

	//perform the check if the patch is allready there
	query = "SELECT id FROM ";
	query+= PATCH_TABLE;
	query+= " ";
	query+= "WHERE surface = ";
	query+= parent_;
	query+= " and center = ";
	query+= center_;
	query+= " and radius = ";
	query+= radius_;
	query+= ";";

	mysql_query(con, query.c_str());
	res = mysql_store_result(con);
	
	if (mysql_num_rows(res)>0)
	{
		//if there is allready a patch with the same parameters return its id
		row = mysql_fetch_row(res);
		id = atoi(row[0]);
	}
	else
	{
		//otherwise insert the new patch into the database

		//fill the stream with the surfacePatch and reset it to the beginning
		stringstream blobstream(stringstream::in | stringstream::out | 
														stringstream::binary);
		writeMesh(patch, blobstream, bin_format);
		blobstream.seekg(0,ios::beg);

		//get out the blob data and create an escaped \0 terminated string
		string blob = blobstream.str();
	
		char* escaped_blob = new char[blob.size()*2+1];
		mysql_real_escape_string(con, escaped_blob,blob.c_str(),blob.size());	

		//build the query string
		query = "INSERT INTO ";
		query+= PATCH_TABLE;
		query+= " ";
		query+= "(surface, center, radius, points, patch) VALUES(";
		query+= parent_;
		query+= ", ";
		query+= center_;
		query+= ", ";
		query+= radius_;
		query+= ", ";
		query+= points_;
		query+= ", '";
		query+= escaped_blob;
		query+= "');";

		//remove the escaped blob string
		delete[] escaped_blob;

		//execute the query;
		mysql_query(con, query.c_str());
	
		if (mysql_errno(con))
		{
			cerr << "DB Error(" << mysql_errno(con) << "): " << mysql_error(con);
		}
		else
			id = mysql_insert_id(con);
	}
								
	mysql_free_result(res);
		
	return id;
}

//*******************************************************************************
bool readExperiment_DB(MYSQL* con, experiment_t & exp, const int exp_id, 
											 const bool allcols)
//*******************************************************************************
{
	bool ok = false;
	
	//parameter conversion
	char id_[50];
	sprintf(id_,"%d", exp_id);

	string query = "SELECT name, surface_a, surface_b, radius, resolution, step, ";
	query+= "datafile_a, datafile_b, cols, center_a, center_b FROM ";
	query+= EXPERIMENTS_TABLE;
	query+= " WHERE id= ";
	query+= id_;
	query+= ";";
	
	mysql_query(con, query.c_str());
	
	MYSQL_RES* res = mysql_store_result(con);
	if (res)
	{
	
		if (mysql_num_rows(res)>0)
		{
			MYSQL_ROW row = mysql_fetch_row(res);

			exp.name = row[0];

			exp.surface_a = atoi(row[1]);
			exp.surface_b = atoi(row[2]);

			exp.radius = atof(row[3]);
			exp.img_resolution = atoi(row[4]);
			exp.rotation_step = atof(row[5]);

			string datafile_a = row[6];
			string datafile_b = row[7];

			int cols = atoi(row[8]);
			if (exp.cols != 0)
				cols = exp.cols;
			else
				exp.cols = cols;
			if (allcols) cols=0;

			exp.center_a = atoi(row[9]);
			exp.center_b = atoi(row[10]);

			ok = (read_data(datafile_a, cols, exp.key1_data, exp.colnames[0])) 
				&& (read_data(datafile_b, cols, exp.key2_data, exp.colnames[1]));
		}

		mysql_free_result(res);
	}
	else
		cerr << "DB Error(" << mysql_errno(con) << "): " << mysql_error(con);

	return ok;
}

//*******************************************************************************
int read_header(istream& stream, vector<string>& columns)
//*******************************************************************************
{
	char line_buffer[200];
	
	stream.getline(line_buffer,200);
  stringstream header(line_buffer);
	
	copy(istream_iterator<string>(header),istream_iterator<string>(),
			 back_inserter(columns));
	
	return columns.size();
}

//*******************************************************************************
int read_data(string file_name, int cols, Rectmat_<double>& data, 
							vector<string>& columns)
//*******************************************************************************
{
	vector< Vector_<double> > rows;
	typedef vector< Vector_<double> >::const_iterator row_cit;

	int colcount = 0;
	int i=0;
	
	ifstream stream(file_name.c_str());
	
	if (!stream)
	{
		cerr << "Could not open " << file_name << " as a vertex data file!";
		return -1;
	}
	
	read_header(stream, columns);
	
	//determine the number of selected columns
  for(i=0;i<columns.size();++i)
		if ((cols==0) || ((cols >> i) & 1)) colcount++;

	//read the datalines into the row vectors
	while(!stream.eof())
	{
		Vector_<double> row(colcount);
		double datapoint = 0;
		char line_buffer[200];
		
		int col_ix  = 0;
		int act_col = 0;

		stream.getline(line_buffer,200);
		stringstream line(line_buffer);
		
		while ((!line.eof()) && (act_col<colcount))
		{
			line >> datapoint;
			if ((cols==0) || ((cols >> col_ix++) & 1)) row[act_col++] = datapoint; 
		}
					
		rows.push_back(row);
	}

	//copy the read lines into the data matrix

	//resize data matrix
	data.size(rows.size(),colcount);

	i=0;
	for(row_cit r=rows.begin();r!=rows.end();++r)
		data.row(*r,i++);

	return rows.size();
}

//*******************************************************************************
string ftoa(const double value, const int precision)
//*******************************************************************************
{
	char convert[255];
	if (precision>0)
		sprintf(convert, "%.*f", precision, value);
	else
		sprintf(convert, "%f", value);
	return string(convert);
}

//*******************************************************************************
string itoa(const int value)
//*******************************************************************************
{
	char convert[255];
	sprintf(convert, "%d", value);
	return string(convert);
}

