/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "cpfilter.hh"

#include <math.h>
#include <float.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <vector>
#    include <iterator>
#  else
#    include <vector.h>
#    include <iterator.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <vector>
#  include <iterator>
#endif //HAVE_CONFIG_H

#ifdef HAVE_NAMESPACES
using namespace std;
#endif

//*******************************************************************************
double max(double x, double y)
//*******************************************************************************
{
	return (x>=y) ? x : y;
}

//*******************************************************************************
template<class Out>
size_t getRegion(const Trimesh::vertex_t v, Trimesh* surface, 
								 double cutoff, Out ins)
//*******************************************************************************
{
	vector<double> dist(surface->nVertices());
	surface->calcOnSurfaceDistance(v,dist);
	
	int c=0;

	for(unsigned int w=0;w<dist.size();++w)
		if ((dist[w]<cutoff) && (v!=w))
		{
			*ins++ = w;
			++c;
 	  }

	return c;
}

//*******************************************************************************
void CPFilter::calcOnSurfaceDistances()
//*******************************************************************************
{
// 	//for each surface
// 	for(int s=0;s<2;++s)
// 	{
// 	  int nverts = pSurfaces[s]->nVertices();

// 		dDistances[s] = new double*[nverts];
// 		for (int i=0;i<nverts;++i) dDistances[s][i] = new double[nverts];

// 		johnson_all_pairs_shortest_paths(pSurfaces[s]->Mesh, dDistances[s]);
// 	}
}

//*******************************************************************************
size_t CPFilter::scanSurface(int six)
//*******************************************************************************
{
	int c=0;

	/* LOOP: twice, once for the tops and the second time for the bottoms */

	for(int stage=-1;stage<=1;stage+=2)
	{

		//a bitvector indicating which verices should be checked
		vector<bool> toCheck(pSurfaces[six]->nVertices(), true);

		for(Trimesh::vertex_it v=pSurfaces[six]->begin();v!=pSurfaces[six]->end();++v)
		{

			//CHECK THE POSITION
			if ((toCheck[*v]) && (fRadius[0]>0) && (fRadius[1]>0))
			{
				const Vector_<double>& pos = pSurfaces[six]->vertCoordinates(*v);
				toCheck[*v] = ((pos - fCenter[six]).vec_len()<fRadius[six]);
			}

			//CHECK THE CURVATURE
			if (toCheck[*v])
			{
				double lc1 = (*pSurfData[six])(*v, 0) * stage;
				double lc2 = (*pSurfData[six])(*v, 1) * stage;

				//check if both curvatures are below zero or above zero (depends on stage)

				if ((lc1>0) && (lc2>0))
				{
					/*proceed only if this point is worth to check*/

					bool most_critical = true;
					list<Trimesh::vertex_t> region;
					list<Trimesh::vertex_t>::const_iterator w;

 					//the important cannonical curvature
 					double lc = (stage == -1) ? lc1 : lc2;

					//receive all vertices within the cutoff distance
					getRegion(*v, pSurfaces[six], fCutOff, back_inserter(region));
			
					//iterate over all selected neighbors for lc1
					for(w=region.begin();w!=region.end();++w)
					{
						double c1 = (*pSurfData[six])(*w, 0) * stage;
						double c2 = (*pSurfData[six])(*w, 1) * stage;

						//the point must be a member of the same class
						if ((c1>0) && (c2>0))
						{
							double check_val = (stage == -1) ? c1 : c2;
							
							//check if the abs value of the important curvature is larger than the
							//check_value
							if (lc>check_val) toCheck[*w] = false;	else most_critical = false;
            }
					}

					//add the point to the list of extremes if it is the most critical in
					//the neighborhood
					if (most_critical)
					{
						if (stage<0)
							lCPLists[six].first.push_back(*v);
						else 
							lCPLists[six].second.push_back(*v);
							
						++c;
					}
				}
			}
		}
	}

	return c;
}

//*******************************************************************************
CPFilter::CPFilter(Trimesh** surfaces, distancemap_t** distances,
									 const RazorBack::Rectmat_<double>** data, double cutoff,
									 const Vector_<double>* sphereA,const Vector_<double>* sphereB)
//*******************************************************************************
{
	init(surfaces, distances, data, cutoff, sphereA, sphereB);	
}	

//*******************************************************************************
CPFilter::~CPFilter()
//*******************************************************************************
{
// 	for(int s=0;s<2;++s)
// 	{
// 	  int nverts = pSurfaces[s]->nVertices();

// 		for (int i=0;i<nverts;++i) delete[] dDistances[s][i];
// 		delete[] dDistances[s];
// 	}
}

//*******************************************************************************
void CPFilter::init(Trimesh** surfaces, distancemap_t** distances,
										const RazorBack::Rectmat_<double>** data, double cutoff,
										const Vector_<double>* sphereA,const Vector_<double>* sphereB)
//*******************************************************************************
{
	Filter::init(surfaces, distances);
	
	fCutOff = fabs(cutoff);
	for(int i=0;i<2;++i) pSurfData[i] = data[i];

	//initialize the distance matrix for the surfaces
	//calcOnSurfaceDistances();

	fCenter[0].dim(3);
	fCenter[1].dim(3);
	for(int i=0;i<3;++i)
	{
		if (sphereA) fCenter[0][i] = (*sphereA)[i];
		if (sphereB) fCenter[1][i] = (*sphereB)[i];
	}

	fRadius[0] = (sphereA) ? (*sphereA)[3] : 0;
	fRadius[1] = (sphereB) ? (*sphereB)[3] : 0;
}


//*******************************************************************************
size_t CPFilter::operator()(assocgraph_t& assocgraph)
//*******************************************************************************
{
	/* INITIALIZE */
	assocgraph_traits::pairmap_t pairs = get(vertex_pair, assocgraph);
	
	int c=0;

	//scan both surfaces for critical points;
	scanSurface(0);
	scanSurface(1);

	/* CREATE VERTICES */
	
	//create top pairs
	for(cplist_cit a=lCPLists[0].first.begin();a!=lCPLists[0].first.end();++a)
		for(cplist_cit b=lCPLists[1].first.begin();b!=lCPLists[1].first.end();++b)
		{
			assocgraph_traits::vertex_t v = createVertex(assocgraph);
			put(pairs, v, make_pair(*a, *b));

			++c;
		}

	//create bottom pairs
	for(cplist_cit a=lCPLists[0].second.begin();a!=lCPLists[0].second.end();++a)
		for(cplist_cit b=lCPLists[1].second.begin();b!=lCPLists[1].second.end();++b)
		{
			assocgraph_traits::vertex_t v = createVertex(assocgraph);
			put(pairs, v, make_pair(*a, *b));

			++c;
		}

//   /* BUILD THE DISTANCE MAPS */
// 	for(int s=0;s<2;++s)
// 	{
// 		int nverts = pSurfaces[s]->nVertices();

// 		//iterate over both lists
// 		for(cplist_cit v=lCPLists[s].first.begin();v!=lCPLists[s].second.end();++v)
// 		{
// 			//smooth jump from first.end to second.begin
// 			if (v==lCPLists[s].first.end()) v = lCPLists[s].second.begin();

// 			if (v!=lCPLists[s].second.end())
// 			{
// 				//copy the row vector for vertex v into the distance map
// 				double* dist = new double[nverts];
// 				memcpy(dist, dDistances[s][*v], nverts*sizeof(double));
				
// 				(*pDistances[s])[*v] = dist;
// 			}
// 		}
// 	}

	return c;
}
