/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CPFILTER_HH
#define CPFILTER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <list>
#  else
#    include <list.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <list>
#endif //HAVE_CONFIG_H

#include <Vector.hh>
#include <Rectmat.hh>

#include "filter.hh"

enum cptype_t { cpTop, cpBottom };

/**
 * \brief a filter to extract critical points from both surfaces
 *
 * The critical point filter examines both surfaces and searches for critical
 * points according to the local curvature. It identifies tops of convex and
 * bottoms of concave surface regions. Thus it requires data for both cannonical
 * curvatures to be present in the first and second column of the pSurfData
 * matrices.
 * 
 * After the critical points are identified the filter adds all possible pairs
 * of tops and bottoms of both surfaces as patchpairs to the associative graph.
 *
 * The search for critical points is performed in the follwing way:
 *
 * - look for all local minimas of LC1 and check every minima if LC2 at this
 * point is below zero ( = \b convex critical points (tops)) 
 * 
 * - look for all local maximas of LC2 and check every maxima if LC1 at this
 * point is above zero ( = \b concave critical points (bottoms))
 *
 * The filter does not need any additional threshold, and works a little bit
 * different then all the other filters, because he expects an empty association
 * graph and the filter action is actually a generation of vertices. However a
 * cutoff radius must be defined, that specifies the region that is scanned
 * around each vertex for a more critical point (fCutOff). This cutoff radius
 * should be equal to the cutoff radius used in the generation of the local
 * curvature property.
 */
class CPFilter : public Filter
{
public:
	
	//*****************************************************************************
	// CONSTRUCTORS AND INITIALIZATION
	//*****************************************************************************

	/**
	 * \brief default constructor
	 */
	CPFilter() : Filter() {}
	
	/**
	 * \brief initializing constructor
	 * 
	 * The initializing constructor creates a running object of the CPFilter
	 * class. It initializes all the necessary parameters by calling init()
	 * method.
	 *
	 * \param surfaces    a two membered field of pointers to trimesh structures
	 * \param distances   a two membered field of constant pointers to distance maps
	 *                    for surface 0 and 1 (or A and B).
	 * \param data        two membered field of Rectmat_<double> pointers pointing
	 *                    to the vertex data for both surfaces
	 *
	 * \sa Filter::Filter(Trimesh**, distancemap_t**)
	 */
	CPFilter(Trimesh** surfaces, distancemap_t** distances,
					 const RazorBack::Rectmat_<double>** data, double cutoff,
					 const Vector_<double>* sphereA=0, const Vector_<double>* sphereB=0);

	/**
	 * \brief destructor
	 *
	 * The destructor destroys the on-surface distance maps
	 */
	~CPFilter();

	/**
	 * \brief initializing method
	 *
	 * The method connects all the necessary data to the internal members.
	 *
	 * \param surfaces    a two membered field of pointers to trimesh structures
	 * \param distances   a two membered field of constant pointers to distance maps
	 *                    for surface 0 and 1 (or A and B).
	 * \param data        two membered field of Rectmat_<double> pointers pointing
	 *                    to the vertex data for both surfaces
	 *
	 * \sa Filter::init()
	 */
	void init(Trimesh** surfaces, distancemap_t** distances,
						const RazorBack::Rectmat_<double>** data, double cutoff,
						const Vector_<double>* sphereA=0, const Vector_<double>* sphereB=0);

	//*****************************************************************************
	// OPERATORS
	//*****************************************************************************

	/**
	 * \brief the filter operator
	 *
	 * For further details see Filter::operator().
	 *
	 * \param assocgraph a reference to an association graph, that should be
	 *                   processed.
	 * 
	 * \return the number of modified graph elements.
	 */
	size_t operator()(assocgraph_t& assocgraph);
	//*****************************************************************************
	// OUTPUT
	//*****************************************************************************

	/**
	 * \brief copies the specified set of critical points to the output iterator
	 * \p ins
	 *
	 * The method writes the sequence of critical points of type \p cptype on the
	 * surface \p surface to the output iterator \p ins.
	 *
	 * \param surface the index of the desired surface (0 or 1) 
	 * \param cptype  the type of critical points (\c cpTop for tops or \c 
	 *                cpBottom for bottoms).
	 * \param ins     the output iterator that receives the sequence
	 *
	 */
	template<class Out>
	size_t getCPoints(size_t surface, cptype_t cptype, Out ins) const;

private:

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************
	
	/**
	 * \brief initializes and calculates the distance matrices for both surfaces
	 *
	 * The method uses the johnson's all pairs shortest path algorithm implemented
	 * in the boost graph library to detect the on-surface distances between all
	 * pairs.
	 */
	void calcOnSurfaceDistances();

	/**
	 * \brief detects the critical points on one of the surfaces
	 *
	 * The method scans one surface (specified by the parameter \p six) for
	 * critical points and stores them in the appropriate lists.
	 *
	 * \param six index of the chosen surface
	 *
	 * \return number of critical points
	 */
	size_t scanSurface(int six);

	//*****************************************************************************
	// PRIVATE TYPE DEFINITIONS
	//*****************************************************************************

	typedef list<Trimesh::vertex_t> cplist_t;
	typedef list<Trimesh::vertex_t>::iterator cplist_it;
	typedef list<Trimesh::vertex_t>::const_iterator cplist_cit;
	/**<
	 * \brief type of a top or bottom list of a single surface + iterator types
	 */

	//*****************************************************************************
	// PRIVATA DATA MEMBERS
	//*****************************************************************************

	/**
	 * \brief two pointers to the surface data matrices of this experiment
	 *
	 * The data matrices must present the first and second local cannonical
	 * curvatures in their first and second column. The user has to take care to
	 * pass the right properties.
	 */
	const RazorBack::Rectmat_<double>* pSurfData[2];

	/**
	 * \brief specified the cut off distance in the extreme search
	 */
	double fCutOff;

	/**
	 * \brief the lists containing the tops and bottoms of both surfaces
	 *
	 * The data structure is a two membered field (one entry for each surface) of
	 * a pair of vertex lists representing the list of tops (first element) and
	 * the list of bottoms (second element) of the surface.
	 */
	pair<cplist_t, cplist_t> lCPLists[2];

	/**
	 * \brief complete distance matrices for both surfaces
	 */
	double** dDistances[2];

	Vector_<double> fCenter[2];
	double fRadius[2];

};

#ifdef INCLUDE_TMPL_DEFS
#  include "cpfilter_templ.cc"
#endif


#endif //CPFILTER_HH
