/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include <math.h>
#include <float.h>

#include <Vector.hh>

#include "distancefilter.hh"

#ifndef SQR
#define SQR(a) a*a
#endif

#ifndef A 
#define A 0
#endif

#ifndef B
#define B 1
#endif
	
//*******************************************************************************
DistanceFilter::DistanceFilter(Trimesh** surfaces, distancemap_t** distances,
															 const double mindist, const double tolerance, 
															 const int action)
//*******************************************************************************
:Filter(), fMinDist(0.0), fTolerance(0.0), fAction(DFA_CREATE)
{
	init(surfaces, distances, mindist, tolerance, action);
}	

//*******************************************************************************
void DistanceFilter::init(Trimesh** surfaces, distancemap_t** distances,
													const double mindist, const double tolerance,
													const int action)
//*******************************************************************************
{
	Filter::init(surfaces, distances);
	fMinDist = mindist;
	fTolerance = tolerance;
	fAction = action;
}

//*******************************************************************************
size_t DistanceFilter::operator()(assocgraph_t& assocgraph)
//*******************************************************************************
{
	/* INITIALIZATION */
	int c=0;
	bool checkmin = (fMinDist>DBL_EPSILON);

	//necessary property maps
	assocgraph_traits::pairmap_t pairmap = get(vertex_pair, assocgraph);
	assocgraph_traits::distancesmap_t distmap = get(edge_distances, assocgraph);
	
	assocgraph_traits::vertices_t vits = vertices(assocgraph);

	/* FIRST LOOP*/
	for(assocgraph_traits::vertex_it v=vits.first;v!=vits.second;++v)
	{
		Vector_<double> firstpos[2];
		
		//extract the first positions
		firstpos[A] = pSurfaces[A]->vertCoordinates(pairmap[*v].first);
		firstpos[B] = pSurfaces[B]->vertCoordinates(pairmap[*v].second);
		
		
		/* SECOND LOOP */
		for(assocgraph_traits::vertex_it w=(w=v)++;w!=vits.second;++w)
		{
			bool ok = true;
			double dA, dB;
				
			//calculate the euclidean distance only if A1 and A2 as well as B1 and B2
			//are different otherwise the quartett is invalid
			if ((pairmap[*w].first!=pairmap[*v].first) &&
					(pairmap[*w].second!=pairmap[*v].second))
			{
				Vector_<double> secondpos[2];

				secondpos[A] = pSurfaces[A]->vertCoordinates(pairmap[*w].first);
				secondpos[B] = pSurfaces[B]->vertCoordinates(pairmap[*w].second);

				dA = (firstpos[A] - secondpos[A]).vec_len();
				dB = (firstpos[B] - secondpos[B]).vec_len();
				
				//apply the filter criteria
				if (checkmin) ok = ((dA>=fMinDist) && (dB>fMinDist));
				if (ok) ok = (fabs(dA-dB)<fTolerance);
			}
			else
				//A1==A2 or B1==B2 is an invalid quartet per definition
				ok = false;

			/* FILTERING */
			if ((ok) && (fAction==DFA_CREATE))
			{
				//add the edge (v,w) and store the distances 
				pair<assocgraph_traits::edge_t, bool> e = add_edge(*v, *w, assocgraph);
				put(distmap, e.first, make_pair(dA,dB));
				++c;
			}	
			else if ((!ok) && (fAction==DFA_DELETE))
			{
				//remove any existing edge (v,w)
				pair<assocgraph_traits::edge_t, bool> e = edge(*v, *w, assocgraph);
				if (e.second)
				{
					remove_edge(e.first, assocgraph);
					++c;
				}
			}
		}
	}

	//remove the isolated vertices
	removeIsolated(assocgraph,1);
	return c;
}
