/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef DISTANCEFILTER_HH
#define DISTANCEFILTER_HH

#include "filter.hh"

const int DFA_CREATE = 0;
const int DFA_DELETE = 1;

/**
 * \brief a special filter scanning the distances between matching point
 * quartetts (pairs of pairs)
 *
 * The filter examines all edges between two matching point pairs and is thus
 * acting on quartetts of matching surface points.
 *
 * Despite the data expected by the Filter base class the distance filter needs
 * a minimum distance (fMinDist) and a distance tolerance value
 * (fTolerance). The filter criterion is as follows:
 *
 * For each quartett of matching points (A1, A2 on the first surface and
 * correspodning points B1, B2 on the second surface) the filter checks the
 * euclidean distance between A1-A2 (dA) and B1-B2 (dB). If either dA and dB are
 * larger than the minimum distance and their difference is within the tolerance
 * value. 
 *
 * The behaviour of the filter depends on the flag fAction. If fAction is set to
 * DFA_CREATE the filter builds an edge for all quartetts that pass the test, if
 * fAction is DFA_DELETE all edges that don't pass the test are removed.
 *
 * \todo The first implementation is a simple (but slow) two folded loop to
 * check all possible quartets. A second stage implementation should contain a
 * more sophisticated and faster hashing algorithm.
 */
class DistanceFilter: public Filter
{
public:
	
	/**
	 * \brief standard constructor creating an unitialized instance
	 */
	DistanceFilter() 
			:Filter(), fMinDist(0.0), fTolerance(0.0), fAction(DFA_CREATE) { }

	
	/**
	 * \brief full initializing constructor
	 *
	 * The constructor internally calls the init method and passes all the
	 * parameters to create a fully initialized instance. The mechanism is similar
	 * to the initializing constructor of Filter.
	 *
	 * \param surfaces   a two membered field of pointers to trimesh structures
	 * \param distances  a two membered field of constant pointers to distance maps
	 *                   for surface 0 and 1 (or A and B).
	 * \param mindist    the minimum quartett distance
	 * \param tolerance  the tolerance for the distances A1A2 and B1B2
	 * \param action     determines if positive edges should be created or negative 
	 *                   removed
	 *
	 * \sa Filter::Filter(Trimesh**, const distancemap_t*)
	 */
	DistanceFilter(Trimesh** surfaces, distancemap_t** distances,
								 const double mindist, const double tolerance, 
								 const int action = DFA_CREATE);
	
	/**
	 * \brief initializing method
	 *
	 * This method performs a full initialization of a DistanceFilter instance.
	 *
	 * \param surfaces   a two membered field of pointers to trimesh structures
	 * \param distances  a two membered field of constant pointers to distance maps
	 *                   for surface 0 and 1 (or A and B).
	 * \param mindist    the minimum quartett distance
	 * \param tolerance  the tolerance for the distances A1A2 and B1B2
	 * \param action     determines if positive edges should be created or negative 
	 *                   removed
	 *
	 * \sa Filter::init(Trimesh**, const distancemap_t*)
	 */
	void init(Trimesh** surfaces, distancemap_t** distances,
						const double mindist, const double tolerance,
						const int action = DFA_CREATE);

	//*****************************************************************************
	// OPERATOR
	//*****************************************************************************

	/**
	 * \brief the filter operator
	 *
	 * For further details see Filter::operator().
	 *
	 * \param assocgraph a reference to an association graph, that should be
	 *                   processed.
	 * 
	 * \return the number of modified graph elements.
	 */
	size_t operator()(assocgraph_t& assocgraph);

private:

	/**
	 * \brief the minimal distance between two matching point pairs on one surface
	 */
	double fMinDist;
	
	/**
	 * \brief the tolerance between distances A1A2 and B1B2 in a quartett
	 */
	double fTolerance;
	
	/**
	 * \brief specifies the action to perform in the filtering
	 */
	int fAction;

};

#endif //DISTANCEFILTER_HH  
