/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "duckfilter.hh"
#include <math.h>
#include <fileformats.hh>

#include <mapping.hh>
#include <normcorr_matching.hh>

#ifndef PI
#define PI 3.14159265358979
#endif

using namespace RazorBack;


//*******************************************************************************
DuckFilter::DuckFilter(Trimesh** surfaces, distancemap_t** distances,
											 const double radius, const int resolution, 
											 const double step, const double rlevel,
											 const RazorBack::Rectmat_<double>** expData,
											 patchMaker** patches, hmaps_t** hmaps)
//*******************************************************************************
:Filter(), fRadius(0), fResolution(0), fRotationStep(0), fRLevel(0)
{
	init(surfaces, distances, radius, resolution, step, rlevel, expData, patches, hmaps);
}
	
//*******************************************************************************
void DuckFilter::init(Trimesh** surfaces, distancemap_t** distances,
											const double radius, const int resolution, 
											const double step, const double rlevel,
											const RazorBack::Rectmat_<double>** expData,
											patchMaker** patches, hmaps_t** hmaps)
//*******************************************************************************
{
	Filter::init(surfaces, distances);
	
	//patch radius and resolution
	fRadius = (radius>0) ? radius : fabs(radius);
	fResolution = (resolution>0) ? resolution : abs(resolution);
	fRotationStep = step/180.0*PI;
	//check if the resolution is odd and increase it if it's even
	if (!(fResolution % 2)) ++fResolution;

	//filter criterion
	fRLevel = rlevel;

	for(int i=0;i<2;++i)
	{
		pSurfData[i] = expData[i];
		pPatches[i] = patches[i];
		pHMaps[i] = hmaps[i];
	}

}

//*******************************************************************************
size_t DuckFilter::operator()(assocgraph_t& assocgraph)
//*******************************************************************************
{
	/* INITIALIZATION */
	//harmonic image map
	typedef map<Trimesh::vertex_t, harmonicImage*> himgmap_t;
	typedef himgmap_t::iterator himgmap_it;
	himgmap_t himgmap[2];

	int c =0;
	int i;

	//necessary propertymaps
	assocgraph_traits::pairmap_t pairmap = get(vertex_pair, assocgraph);
	assocgraph_traits::kksmap_t kksmap  = get(vertex_KKS, assocgraph);
	assocgraph_traits::rfuncmap_t rfuncmap = get(vertex_Rfunc, assocgraph);
	assocgraph_traits::anglemap_t anglemap = get(vertex_angle, assocgraph);
	assocgraph_traits::zeromap_t zeromap = get(vertex_zero, assocgraph);
	
	//iterate over all vertices savely
	assocgraph_traits::vertices_t vits = vertices(assocgraph);
	for(assocgraph_traits::vertex_it vit=vits.first;vit!=vits.second;)
	{
		assocgraph_traits::vertex_it v=vit++;

		himgmap_it img[2];

		bool hasholes = false;

		/* RETRIEVE THE IMAGES FOR BOTH POINTS */
		for(i=0;i<2;++i)
		{
			//retrieve the point on surface i
			Trimesh::vertex_t point = (i==0) ? pairmap[*v].first : pairmap[*v].second;

			//check if the image for point already exists
			img[i] = himgmap[i].find(point);
			if (img[i]==himgmap[i].end())
			{
				/* CREATE THE PATCH, MAP AND IMAGE */

				//add the point as a patch center (method checks if allready added)
				pPatches[i]->addVertex(point);
				surfacePatch* patch = (*pPatches[i])[point];

				//check for holes
				hasholes = (hasholes || (patch->getStatus()==surfacePatch::ps_faultborder));
								
				if (!hasholes)
				{

					//copy the patch to the corresponding place in the harmonic map map
					hmaps_it map = pHMaps[i]->insert(
						make_pair(point,*dynamic_cast<Trimesh*>(patch))).first;
					//make the map			
					Trimesh::vertex_t zero = make_map(map->second, fRadius);
				
					//store the zero vertex in the property map
					((i==0) ? zeromap[*v].first : zeromap[*v].second) = zero;
					
					//create the image, init the data and the screening
					harmonicImage* hI = new harmonicImage(&map->second,
																								fRotationStep,fResolution);
					hI->setGridtype(harmonicImage::gtCircular);
					hI->initfromMaster("", *pSurfData[i]);
					if (i==0) hI->initSearch();
					//add the image to the image map
					img[i] = himgmap[i].insert(make_pair(point, hI)).first;
				}
			}
		}
		
		double R;
		double maxAngle;

		if (!hasholes)
		{
			//screen both harmonic images
			normcorr_matching matcher(img[0]->second);
			double* rfunc = new double[img[0]->second->getScanSteps()];

			R = img[0]->second->compare_by(*img[1]->second, matcher, false, false);
			maxAngle = -1 * img[0]->second->getMaxAngle();

			//assign the angular correlation function to the matching point pair
			matcher.copy_correlations(rfunc);
			rfuncmap[*v] = rfunc;
		}

		if ((R<fRLevel) || (hasholes))
		{
			//delete the vertex pair
			removeVertex(*v, assocgraph);
			++c;
		}
		else
		{
			//assign the R value to the KKS property
			put(kksmap, *v, R);
			put(anglemap, *v, maxAngle);
		}

		/* END VERTEX ITERATION */
	}
	
	//remove all the harmonic images from memory
	for(i=0;i<2;++i)
		for(himgmap_it j=himgmap[i].begin();j!=himgmap[i].end();++j)
			delete j->second;

	return c;
}
