/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef DUCKFILTER_HH
#define DUCKFILTER_HH

#include "filter.hh"

#include <patchmaker.hh>
#include <harmonic_image.hh>

#include <Rectmat.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#  else
#    include <map.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <map>
#endif //HAVE_CONFIG_H

/**
 * \brief performs a fast harmonic image filtering of matching pairs
 *
 * The harmonic imageing is a fast screening of the harmonic images around
 * matching point pairs. Compared to the full HI matching only the symmetry axes
 * of a circular grid image are checked. The screening is much faster but cannot
 * be used to find the best overlaps between two surface patches.
 *
 * The filter needs some additional data and data structures: 
 * - the size of the patches
 * - the resolution of the images (mus be odd)
 * - the full datamatrix for this experiment
 * - a two pointers to a patchMaker object (where the necessary patches are stored)
 * - a two pointers to a map of harmonic maps 
 */
class DuckFilter : public Filter
{
public: 
	//*****************************************************************************
	// CONSTRUCTORS AND INITIALIZATION
	//*****************************************************************************
	
	/**
	 * \brief standard constructor creating an uninitialized instance
	 */
	DuckFilter() : Filter() { }
	
	/**
	 * \brief full initializing constructor
	 *
	 * The constructor internally calls the init method and passes all the
	 * parameters to create a fully initialized instance. The mechanism is similar
	 * to the initializing constructor of Filter.
	 *
	 * \param surfaces   a two membered field of pointers to trimesh structures
	 * \param distances  a two membered field of constant pointers to distance maps
	 *                   for surface 0 and 1 (or A and B).
	 * \param radius     radius of the surface patches
	 * \param resolution resolution of the harmonic image grid (must be odd)
	 * \param step       the rotation step for the harmonic images (in degrees)
	 * \param rlevel     the threshold for the correlation coefficient.
	 * \param expData    two membered field of Rectmat_<double> pointers pointing
	 *                   to the vertex data for both surfaces
	 * \param patches    two membered field of patchMaker pointers containing the 
	 *                   experiment's patches
	 * \param hmaps      two membered field of 
	 *
	 * \sa Filter::Filter(Trimesh**, const distancemap_t*)
	 */
	DuckFilter(Trimesh** surfaces, distancemap_t** distances,
						 const double radius, const int resolution, const double step,
						 const double rlevel, const RazorBack::Rectmat_<double>** expData,
						 patchMaker** patches, hmaps_t** hmaps);
	
	/**
	 * \brief initializing method
	 *
	 * This method performs a full initialization of a DistanceFilter instance.
	 *
	 * \param surfaces   a two membered field of pointers to trimesh structures
	 * \param distances  a two membered field of constant pointers to distance maps
	 *                   for surface 0 and 1 (or A and B).
	 * \param resolution resolution of the harmonic image grid (must be odd)
	 * \param step       the rotation step for the harmonic images (in degrees)
	 * \param rlevel     the threshold for the correlation coefficient.
	 * \param expData    two membered field of Rectmat_<double> pointers pointing
	 *                   to the vertex data for both surfaces
	 * \param patches    two membered field of patchMaker pointers containing the 
	 *                   experiment's patches
	 * \param hmaps      two membered field of 
	 *
	 * \sa Filter::init(Trimesh**, const distancemap_t*)
	 */
	void init(Trimesh** surfaces, distancemap_t** distances,
						const double radius, const int resolution, const double step,
						const double rlevel, const RazorBack::Rectmat_<double>** expData,
						patchMaker** patches, hmaps_t** hmaps);
	

	//*****************************************************************************
	// OPERATOR
	//*****************************************************************************

	/**
	 * \brief the filter operator
	 *
	 * For further details see Filter::operator().
	 *
	 * \param assocgraph a reference to an association graph, that should be
	 *                   processed.
	 * 
	 * \return the number of modified graph elements.
	 */
	size_t operator()(assocgraph_t& assocgraph);

private:
	
	//*****************************************************************************
	// PRIVATE DATA MEMBERS
	//*****************************************************************************
	
	/**
	 * \brief filter criterion: the correlation coefficient threshold
	 *
	 * All matching point pairs that show a correlation coefficient below this
	 * threshold are discarded by the filter.
	 */
	double fRLevel;

	/**
	 * \brief radius of the surface patches
	 */
	double fRadius;
	
	/**
	 * \brief resolution of the harmonic images
	 *
	 * Due to the symmetry of the circular grid the resolution must be an odd 
	 * integer.
	 */
	int fResolution;

	/**
	 * \brief the rotation step of the harmonic images
	 */
	double fRotationStep;
	
	/**
	 * \brief two pointers to the surface data matrices of this experiment
	 */
	const RazorBack::Rectmat_<double>* pSurfData[2];

	/**
	 * \brief two pointers to the patchMaker objects for the experiment's surfaces.
	 *
	 * The patches need not to be generated in advance, but any patch that is
	 * generated during the filtering is added to the patchMaker objects and
	 * available for further use.
	 */
	patchMaker* pPatches[2];

	/**
	 * \brief two pointers to a map of harmonic maps.
	 *
	 * The maps of harmonic maps work like the pointers to the patchmaker objects:
	 * No maps are necessary when calling the filter, but any map generated will
	 * be placed in the maps.
	 */
	hmaps_t* pHMaps[2];

};

#endif //DUCKFILTER_HH
