/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include <getopt.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <trimesh.hh>
#include <fileformats.hh>

#include <float.h>
#include <math.h>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/util/XMLString.hpp>

#include <utils.hh>
#include <xmlutils.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif 

#ifdef HAVE_STD
#  include <iostream>
#  include <fstream>
#  include <cstdio>
#  include <string>
#  include <list>
#  include <set>
#  include <algorithm>
#else
#  include <iostream.h>
#  include <fstream.h>
#  include <cstdio.h>
#  include <string.h>
#  include <list.h>
#  include <set.h>
#  include <algo.h>
#endif

#include <boost/graph/dijkstra_shortest_paths.hpp>

#include <bron_kerbosch.hh>

#include <assocgraph.hh>
#include <cluster.hh>
#include <singlelink.hh>
#include <scorecollection.hh>
#include <scorer.hh>

#include "duckfilter.hh"
#include "distancefilter.hh"
#include "overlapfilter.hh"
#include "cpfilter.hh"
#include "fuzzyfilter.hh"
#include "positionfilter.hh"

#ifdef HAVE_NAMESPACES
XERCESNS_USE
#endif

typedef pair<double, double> result_t;

typedef list<Trimesh::vertex_t> vertexlist_t;
typedef vertexlist_t::iterator vertexlist_it;
typedef vertexlist_t::const_iterator vertexlist_cit;

typedef set<assocgraph_traits::vertex_t> cluster_t;
typedef cluster_t::iterator cluster_it;
typedef cluster_t::const_iterator cluster_cit;

typedef list<Cluster> clusters_t;
typedef clusters_t::iterator clusters_it;
typedef clusters_t::const_iterator clusters_cit;

const char* SCOREMETHOD[] = 
{
	"_rmsd",
	"_mean",
	"_corr",
	0
};

const char* SCORENAME[] =
{
	"pcurv",
	"chem",
	"mcurv",
	0
};


//*******************************************************************************
Rectmat_<double>* getFilterData(const Rectmat_<double>& data, const int notcol)
//*******************************************************************************
{
	Rectmat_<double>* result = new Rectmat_<double>(data.rno(),data.cno()-1);
	
	vector<double>::iterator res = (*result)[0];
	vector<double>::const_iterator dat = data[0];
	for(int i=0;i<data.rno();++i)
		for (int j=0;j<data.cno();++j) if (j!=notcol)	*res++ = *dat++; else	++dat;
	return result;
}

//*******************************************************************************
void printDuck(ostream& stream, assocgraph_t& assocgraph)
//*******************************************************************************
{
	assocgraph_traits::vertices_t vits = vertices(assocgraph);
	assocgraph_traits::pairmap_t pairs = get(vertex_pair, assocgraph);
	assocgraph_traits::kksmap_t kks = get(vertex_KKS, assocgraph);

	for(assocgraph_traits::vertex_it v=vits.first;v!=vits.second;++v)
		stream << pairs[*v].first << "\t" << pairs[*v].second << "\t"
					 << kks[*v] << endl;
}

//*******************************************************************************
double* makeExpData(Rectmat_<double>& data)
//*******************************************************************************
{
	const Vector_<double>& col = data.col(0);

	double* array = new double[col.dim()];
	copy(col.begin(), col.end(), array);
	return array;
}

//*******************************************************************************
void printGraph(const string& output, const string& suffix, 
								const experiment_t exp, assocgraph_t& assocgraph)
//*******************************************************************************
{
		/* GRAPH OUTPUT*/
	string graphfile = output + "_";
	graphfile += suffix + ".dot";

	ofstream graphout(graphfile.c_str());
	writegraph(graphout, assocgraph, exp.name.c_str());
	graphout.close();
}

//*******************************************************************************
void Clustering(Trimesh** surfaces, hmaps_t** maps,
								assocgraph_t& assocgraph, clusters_t& Clusters, 
								const bool verbose, const ScoreCollection* scorings = 0)
//*******************************************************************************
{
	typedef pair<cluster_algorithm<assocgraph_t>::cluster_cit,
		cluster_algorithm<assocgraph_t>::cluster_cit>	clusterrange_t;

	//assocgraph property maps
	assocgraph_traits::pairmap_t pairmap = get(vertex_pair, assocgraph);
	assocgraph_traits::kksmap_t kksmap  = get(vertex_KKS, assocgraph);
	assocgraph_traits::anglemap_t anglemap = get(vertex_angle, assocgraph);

	if (verbose)
		cout << "# Starting Bron-Kerbosch clique detection ... " << flush;

	//perform a clique detection with bron kerbosch
	bron_kerbosch<assocgraph_t> clique_detect;
	int num_c = clique_detect(assocgraph);
	
	if (verbose)
		cout << "done\n" << "# " << num_c << " clusters found !" << endl;

	//iterate through all cliques and register them as Cluster objects
	for(int i=0;i<clique_detect.num_clusters();++i)
	{
		size_t dummy;
		clusterrange_t range = clique_detect.get_cluster(i,dummy);
		cluster_t clusterpairs(range.first, range.second);
		//create a new cluster object
		clusters_it c = Clusters.insert(Clusters.end(),
																		Cluster(surfaces, clusterpairs.size()));

		if (verbose)
			cout << "#   Registering cluster " << i+1 << " ... " << flush;

		//insert every point pair into the Cluster object
	  for (cluster_cit cp=clusterpairs.begin();cp!=clusterpairs.end();++cp)
		{
			Trimesh::vertex_t A = pairmap[*cp].first;
			Trimesh::vertex_t B = pairmap[*cp].second;

			c->addPair(A, B, (*maps[0])[A], (*maps[1])[B], 
											 kksmap[*cp], anglemap[*cp]);
		}
			
		if (verbose)
			cout << "done\n#     Assembling corresponding point pairs ... " << flush;
		//calculate the correspondencies
		c->compCorresponds();
		
		if (verbose)
			cout << "done (" << c->getNPoints()<< " pairs)\n"
					 << "#     Perform rigid body transformation ... " << flush;
		//perform the best rot RMS fit
		c->doRMSFit(true);
		//scoring
		if (scorings) (*scorings)(&(*c));

		if (verbose)
			cout << "done ( RMSD = " << c->getRMSD()<< " Angs.)" << endl;
		
		if (c->getRMSD()>4.0) Clusters.erase(c);
	}
}

//*******************************************************************************
size_t extractData(Rectmat_<double>& dest, const Rectmat_<double>& src, int cols,
									 const vector<string>& colnames, int verbose = 0)
//*******************************************************************************
{
	int cix, col, colcount = 0;

	//determine the numer selected columns
	for(col=0;col!=src.cno();++col)
		if ((cols >> col) & 1) 
		{
			colcount++;
			if (verbose) cout << " " << colnames[col];
		}

	//resize the dest matrix
	dest.size(src.rno(),colcount);

	cix = 0;
	for(col=0;col<src.cno();++col)
		if ((cols >> col) & 1)
		{
			dest.col(src.col(col), cix);
			++cix;
		}

	return colcount;
}

//*******************************************************************************
void scoreSubTree(const Scorer* scorer, MergeNode<clusters_it>* node)
//*******************************************************************************
{
	if (node)
	{
		(*scorer)(&(*(node->getCluster())));
		
		scoreSubTree(scorer, node->getLeft());
		scoreSubTree(scorer, node->getRight());
	}
}

//*******************************************************************************
void writeResults(DOMElement* node, 
									const string& dbhost, const string& database, 
									const int experiment, const double patchradius, 
									const double rlevel, const double mindist, 
									const double tolerance, const double angletol,
									const double fuzzytol, const double cpcutoff,
									const int cpcol, const int fuzzycol, const int hsicol,
									const SingleLink* dendrogramm)
//*******************************************************************************
{
	XMLCh* xmlstr;
	XMLCh* attrib;

	DOMElement* parameters;
	DOMElement* subnode;

	DOMDocument* doc = node->getOwnerDocument();

	//add the parameter section
	parameters = XMLUtils::createElement(node, "parameters");
	
	//insert the parameters into the parameter section
	XMLUtils::createTextNode(parameters, "host",        dbhost);
	XMLUtils::createTextNode(parameters, "database",    database);
	XMLUtils::createTextNode(parameters, "experiment",  itoa(experiment));
	
	subnode = XMLUtils::createTextNode(parameters, "patchradius", 
																		 ftoa(patchradius));
	subnode->setAttribute(XMLStr("units"), XMLStr("Ang"));

	subnode = XMLUtils::createTextNode(parameters, "maplevel", 
																		 ftoa(rlevel));
	subnode->setAttribute(XMLStr("units"), XMLStr("corr. coefficient"));

	subnode = XMLUtils::createTextNode(parameters, "mindist", 
																		 ftoa(mindist));
	subnode->setAttribute(XMLStr("units"), XMLStr("Ang"));

	subnode = XMLUtils::createTextNode(parameters, "tolerance", 
																		 ftoa(tolerance));
	subnode->setAttribute(XMLStr("units"), XMLStr("Ang"));

	subnode = XMLUtils::createTextNode(parameters, "angletol", 
																		 ftoa(angletol));
	subnode->setAttribute(XMLStr("units"), XMLStr(""));

	subnode = XMLUtils::createTextNode(parameters, "fuzzytol", 
																		 ftoa(fuzzytol));
	subnode->setAttribute(XMLStr("units"), XMLStr("1"));

	subnode = XMLUtils::createTextNode(parameters, "cpcutoff", 
																		 ftoa(cpcutoff));
	subnode->setAttribute(XMLStr("units"), XMLStr("Ang"));

	XMLUtils::createTextNode(parameters, "cpcol",    itoa(cpcol));
	XMLUtils::createTextNode(parameters, "fuzzycol", itoa(fuzzycol));
	XMLUtils::createTextNode(parameters, "hsicol",  itoa(hsicol));
	
	dendrogramm->writeDendrogramm(node);
}

#ifdef HAVE_LONG_OPTIONS

//*******************************************************************************
bool getOptions(int argc, char* argv[], 
								int& help, int& verbose, string& dbhost, string& database, 
								string& output, int& experiment, double& rlevel,
								double& mindist, double& tolerance, double& angletol, 
								double& fuzzytol, double& cpcutoff, 
								int& cpcol, int& fuzzycol, int& hsicol, linkage_t& linkage)
//*******************************************************************************
{
	extern char *optarg;
  extern int optind;
  int c;

	static struct option long_options[] =
		{
			{"help",       no_argument,       &help,       1 },
			{"verbose",    no_argument,       &verbose,   'v'},
			{"host",       required_argument, 0,          'h'},
			{"database",   required_argument, 0,          'd'},
			{"output",     required_argument, 0,          'o'},
			{"experiment", required_argument, 0,          'e'},
			{"rlevel",     required_argument, 0,         1000},
			{"mindist",    required_argument, 0,         1001},
			{"tolerance",  required_argument, 0,         1002},
			{"angle",      required_argument, 0,         1003},
			{"fuzzytol",   required_argument, 0,         1004},
			{"cpcutoff",   required_argument, 0,         1005},
			{"cpcol",      required_argument, 0,         1006},
			{"fuzzycol",   required_argument, 0,         1007},
			{"hsicol",    required_argument, 0,         1008},
			{"linkage",    required_argument, 0,         1009},
			{0, 0, 0, 0}
		};
  int option_index = 0;       
	
  bool error = false;

  while (1)
	{
		char buffer[200];
		float number = 0;
				
		c = getopt_long(argc, argv, "vh:d:t:e:o:",
										long_options, &option_index);
		
		if (c==-1) break;
		// read all options
    switch(c)
    {
			  /* VERBOSE */
			case 'v' : verbose += 1;	break;
				/* DATABASE HOST */
			case 'h' : if (!(error = !sscanf(optarg, "%s", buffer))) dbhost = buffer;
        break;
				/* DATABASE */
			case 'd' : if (!(error = !sscanf(optarg, "%s", buffer))) database = buffer;
        break;
				/* OUTPUT */
			case 'o' : if (!(error = !sscanf(optarg, "%s", buffer))) output = buffer;
        break;
				/* EXPERIMENT */
			case 'e' : error = !sscanf(optarg, "%d", &experiment); break;
			case 1000: error = !sscanf(optarg, "%lf", &rlevel);  break;
			case 1001: error = !sscanf(optarg, "%lf", &mindist);   break;
			case 1002: error = !sscanf(optarg, "%lf", &tolerance); break;
			case 1003: error = !sscanf(optarg, "%lf", &angletol);  break;
			case 1004: error = !sscanf(optarg, "%lf", &fuzzytol);  break;
			case 1005: error = !sscanf(optarg, "%lf", &cpcutoff);  break;
			case 1006: error = !sscanf(optarg, "%d", &cpcol);      break;
			case 1007: error = !sscanf(optarg, "%d", &fuzzycol);   break;
			case 1008: error = !sscanf(optarg, "%d", &hsicol);    break;
			case 1009: 
			{
				error = !sscanf(optarg, "%s", buffer);    
				
				if (strcmp(buffer,"single")==0)
					linkage = slSingle;
				else if (strcmp(buffer,"complete")==0)
					linkage = slComplete;
				else if (strcmp(buffer,"average")==0)
					linkage = slAverage;

				break;
			}

				/* UNDEFINED */
			case '?' : error = true;	break;
		}
	}

	return !error;
}

#else 

//*******************************************************************************
bool getOptions(int argc, char* argv[], 
								int& help, int& verbose, string& dbhost, string& database, 
								string& output, int& experiment, double& rlevel,
								double& mindist, double& tolerance, double& angletol, 
								double& fuzzytol, double& cpcutoff, 
								int& cpcol, int& fuzzycol, int& hsicol, linkage_t& linkage)
//*******************************************************************************
{
	extern char *optarg;
  extern int optind;
  int c;

  int option_index = 0;       
	
  bool error = false;

  while (1)
	{
		float number = 0;
		char buffer[200];
				
		c = getopt(argc, argv, "vh:d:t:e:o:");
		
		if (c==-1) break;
		// read all options
    switch(c)
    {
			/* VERBOSE */
			case 'v' : verbose += 1;	break;
				/* DATABASE HOST */
			case 'h' : if (!(error = !sscanf(optarg, "%s", buffer))) dbhost = buffer;
        break;
				/* DATABASE */
			case 'd' : if (!(error = !sscanf(optarg, "%s", buffer))) database = buffer;
        break;
				/* OUTPUT */
			case 'o' : if (!(error = !sscanf(optarg, "%s", buffer))) output = buffer;
        break;
				/* EXPERIMENT */
			case 'e' : error = !sscanf(optarg, "%d", &experiment); break;
				/* UNDEFINED */
			case '?' : error = true;	break;
		}

		//read the rest from the enviromental variables
		
		char* var;
		var = getenv("SC_RLEVEL");
		if (strlen(var)>0)
			error = !sscanf(var, "%lf", &rlevel);

		var = getenv("SC_MINDIST");
		if (strlen(var)>0)
			error = !sscanf(var, "%lf", &mindist);

		var = getenv("SC_TOLERANCE");
		if (strlen(var)>0)
			error = !sscanf(var, "%lf", &tolerance);
		
		var = getenv("SC_ANGLETOL");
		if (strlen(var)>0)
			error = !sscanf(var, "%lf", &angletol);

		var = getenv("SC_FUZZYTOL");
		if (strlen(var)>0)
			error = !sscanf(var, "%lf", &fuzzytol);

		var = getenv("SC_CPCUTOFF");
		if (strlen(var)>0)
			error = !sscanf(var, "%lf", &cpcutoff);

		var = getenv("SC_CPCOL");
		if (strlen(var)>0)
			error = !sscanf(var, "%d", &cpcol);

		var = getenv("SC_FUZZYCOL");
		if (strlen(var)>0)
			error = !sscanf(var, "%d", &fuzzycol);

		var = getenv("SC_HSICOL");
		if (strlen(var)>0)
			error = !sscanf(var, "%d", &hsicol);

		var = getenv("SC_LINKAGE");
		if (strlen(var)>0)
		{
			error = !sscanf(var, "%s", buffer);
			
			if (!error)
				if (strcmp(buffer,"single"))
					linkage = slSingle;
				else if (strcmp(buffer,"complete"))
					linkage = slComplete;
				else if (strcmp(buffer,"average"))
					linkage = slAverage;
		}
	}

	return !error;
}

#endif

//*******************************************************************************
void usage(ostream& str = cout)
//*******************************************************************************
{
	str << "!Usage: duckcompare -h HOST -d DATABASE -e EXPERIMENT_ID " 
			<< "                    -t RESULT_TABLE | -o OUTPUTFILE" << endl;
}

//*******************************************************************************
int main(int argc, char* argv[])
//*******************************************************************************
{
	/* INITIALIZATION */

	//command line parameters
	string database = "";
	string dbhost  = DB_HOST;
	string output = "";
	int help = 0;
	int verbose = 0;
	int experiment = 0;

	double rlevel =0.0;
	double mindist  =0.5;
	double tolerance = 0.5;
	double angletol = 15;
	double fuzzytol = 0.3;
	double cpcutoff = 2.0;
	int cpcol = 0;
	int fuzzycol =0; 
	int hsicol=0;
	linkage_t linkage = slSingle;

	experiment_t exp;

	//mesh structures
	Trimesh* mesh[2];
	mesh[0] = mesh[1] = 0;

	//the three filter data matrices and their pointers
	Rectmat_<double> filterdata[6];

	const Rectmat_<double>* pCPData[2];
	const Rectmat_<double>* pFuzzyData[2];
	const Rectmat_<double>* pDuckData[2];

	pCPData[0]    = &filterdata[0];
	pCPData[1]    = &filterdata[1];
	pFuzzyData[0] = &filterdata[2];
	pFuzzyData[1] = &filterdata[3];
	pDuckData[0]  = &filterdata[4];
	pDuckData[1]  = &filterdata[5];

	/* GENERATION of SCORE COLLECTION */

	ScoreCollection scorings;
	
	//loop over all methods
	for(int m=0;m<3;++m)
	{
		//loop over all different property types (pcurve, chem, mcurve)
		for(int n=0;n<3;++n)
		{
			const Rectmat_<double>** scoredata; 

			//determine the scoring data
			switch(n)
			{
				case 0 : scoredata = pCPData;    break;
				case 1 : scoredata = pFuzzyData; break;
				case 2 : scoredata = pDuckData;  break;
			}

			string name = SCORENAME[n];
			name       += SCOREMETHOD[m];
			Scorer* scorer;
			switch(m)
			{
				case 0 : scorings.addScorer(RMSDScorerFactory(scoredata, name));	break;
				case 1 : scorings.addScorer(MeanScorerFactory(scoredata, name));	break;
				case 2 : scorings.addScorer(PearsonScorerFactory(scoredata, name));	
					break;
			}
		}
	}

	//maxima minima
	vertexlist_t minima[2];
	vertexlist_t maxima[2];

	//association graph
	assocgraph_t assocgraph;

	//distance maps
	Filter::distancemap_t* distancemaps[2];
	distancemaps[0] = new Filter::distancemap_t;
	distancemaps[1] = new Filter::distancemap_t;
	patchMaker* patches[2];
	hmaps_t* hmaps[2];

	int i;
	int npairs  = 0;

	if (!getOptions(argc,argv, help, verbose, dbhost, database, output, experiment,
									rlevel, mindist, tolerance, angletol, fuzzytol, cpcutoff, 
									cpcol, fuzzycol, hsicol, linkage))
	{
		usage(cerr);
		return 1;
	}

	if (database=="")
	{
		usage(cerr);
		return 1;
	}

	if (help)
	{
		usage();
		return 0;
	}

	cout << linkage << endl;

	/* OPEN DATABASE */
	//open the database
	MYSQL dbh;
	if (!mysql_init(&dbh))
	{
		cerr << "Could not get a database handle!" << endl;
		return 3;
	}

	if (!	mysql_real_connect(&dbh, dbhost.c_str(), DB_USER, 0, database.c_str(),
													 0, 0, 0))
	{
		cerr << "DB Error(" << mysql_errno(&dbh) << ") : " << endl
				 << mysql_error(&dbh) << endl;

		mysql_close(&dbh);
		return 4;
	}

	//reading the experiment
	if (verbose)
	{
		cout << "# Reading experiment information ... ";
		cout.flush();
	}
	if (!readExperiment_DB(&dbh, exp, experiment, true))
	{
		cerr << "Error reading experimental parameters of experiment " << experiment
				 << "!" << endl;

		mysql_close(&dbh);
		return 5;
	}

	

	if (verbose)
		cout << "done" << endl;

	//output rename
	if (output=="") output = exp.name;

	if (verbose) cout << "# Extracting the filter data matrices ... " << flush;

	//generate the filter data matrices
	for(int i=0;i<2;++i)
	{
		const Rectmat_<double>& fulldata = (i==0) ? exp.key1_data : exp.key2_data;

		cout << "# Data assignments for surface " << i << ":";
		cout << "\n#   critical points : ";
		extractData(filterdata[i], fulldata, cpcol, exp.colnames[i], verbose);
		cout << "\n#   fuzzy filter    : ";
		extractData(filterdata[2+i], fulldata, fuzzycol, exp.colnames[i], verbose);
		cout << "\n#   hsi filter     : ";
		extractData(filterdata[4+i], fulldata, hsicol, exp.colnames[i], verbose);
		cout << endl;
	}

	if (verbose) cout << "done" << endl;
	

	//read surface a, b
	for(i=0;i<2;++i)
	{
		int six = (i==0) ? exp.surface_a : exp.surface_b;
		double probe;
		double density;

		mesh[i] = readSurf_DB(&dbh, six, probe, density);
		
		if (!mesh[i])
			cerr << "Could not read data for surface " << six << "!" << endl;

		//initialize the patchmaker object and the hmap map
		patches[i] = new patchMaker(1, exp.radius, mesh[i]);
		hmaps[i] = new hmaps_t;
	}

	/* CP filtering */
	if (verbose)
		cout << "# Search for critical points on both surfaces ... " << flush;

	CPFilter* cpfilter;
	if ((exp.center_a > 0 ) && (exp.center_b > 0))
	{
		Vector_<double> sphere[2];

		sphere[0] = readCenter_DB(&dbh, exp.center_a);
		if (sphere[0].dim()!=4)
		{
			cerr << "Have problems reading center " << exp.center_a << " of surface "
					 << exp.surface_a << "!" << endl;

			return 6;
		}

		sphere[1] = readCenter_DB(&dbh, exp.center_b);
		if (sphere[1].dim()!=4)
		{
			cerr << "Have problems reading center " << exp.center_b << " of surface "
					 << exp.surface_b << "!" << endl;

			return 6;
		}

		cpfilter = new CPFilter(mesh, distancemaps, pCPData, cpcutoff, 
														&sphere[0], &sphere[1]);
	}
	else
	  cpfilter = new CPFilter(mesh, distancemaps, pCPData, cpcutoff);

	npairs = (*cpfilter)(assocgraph);

	if (verbose>=2)
	{
		string filename = output + "_cp.txt";
		ofstream cpfile(filename.c_str());

		if (cpfile)
		{
			ostream_iterator<int> out(cpfile, " ");

			for(i=0;i<2;++i)
			{
				cpfile << i << " - tops :\n";
				cpfilter->getCPoints(i, cpTop, out);
				cpfile << endl;

				cpfile << i << " - bottoms :\n";
				cpfilter->getCPoints(i, cpBottom, out);
				cpfile << endl;
			}

			cpfile.close();
		}
		
		printGraph(output, "cp", exp, assocgraph);
	}
		
	delete cpfilter;

	if (verbose)
		cout << "done\n"
				 << "# Matching point pairs after CP filter: " << npairs << endl;

	/* Position filtering */
	//filter only, if a center is set for both surfaces
	if ((exp.center_a > 0 ) && (exp.center_b > 0))
	{
		Vector_<double> sphere[2];

		sphere[0] = readCenter_DB(&dbh, exp.center_a);
		if (sphere[0].dim()!=4)
		{
			cerr << "Have problems reading center " << exp.center_a << " of surface "
					 << exp.surface_a << "!" << endl;

			return 6;
		}

		sphere[1] = readCenter_DB(&dbh, exp.center_b);
		if (sphere[1].dim()!=4)
		{
			cerr << "Have problems reading center " << exp.center_b << " of surface "
					 << exp.surface_b << "!" << endl;

			return 6;
		}

		PositionFilter* positionfilter = 
			new PositionFilter(mesh, distancemaps, sphere[0], sphere[1]);
		
		if (verbose)
			cout << "# Selecting critical pairs around centers: \n" 
					 << "# A : (" 
					 << sphere[0][0] << ", " << sphere[0][1] << ", " << sphere[0][2] << ")"
					 << " radius = " << sphere[0][3] << " Angs." << endl
					 << "# B : (" 
					 << sphere[1][0] << ", " << sphere[1][1] << ", " << sphere[1][2] << ")"
					 << " radius = " << sphere[1][3] << " Angs." << " ... " << flush;

		//performing the filter
		npairs -= (*positionfilter)(assocgraph);

		if (verbose)
			cout << "done\n"
					 << "# Critical point pairs after site selection: " << npairs << endl;

		if (verbose>=2)
			printGraph(output, "position", exp, assocgraph);

		delete positionfilter;
	}


	/* Fuzzy filtering */
	if (verbose)
		cout << "# Extract similar pairs of critical points ... " << flush;

	FuzzyFilter* fuzzyfilter = 
		new FuzzyFilter(mesh, distancemaps, pFuzzyData, fuzzytol);
	npairs -= (*fuzzyfilter)(assocgraph);

	if (verbose>=2)
		printGraph(output, "fuzzy", exp, assocgraph);
	
	delete fuzzyfilter;
	if (verbose)
		cout << "done\n"
				 << "# Matching point pairs after fuzzy filter: " << npairs << endl;

	/* HSI filtering */
	if (verbose) cout << "# Starting hsi filtering ... " << flush;

	DuckFilter duckfilter(mesh, distancemaps, exp.radius, exp.img_resolution, 
												exp.rotation_step, rlevel, pDuckData, patches, hmaps);
	npairs -=duckfilter(assocgraph);

	if (verbose>=2)
	{
		string filename = output + "_hsi.txt";
		ofstream duckout(filename.c_str());
		if (duckout)
		{
			printDuck(duckout, assocgraph);
			duckout.close();
		}

		printGraph(output, "hsi", exp, assocgraph);
	}

	if (verbose)
		cout << "done\n"
				 << "# Matching point pairs after HSI filter: " << npairs << endl;

	/* distance filtering */
	DistanceFilter distancefilter(mesh, distancemaps, mindist, tolerance);
	int edges = distancefilter(assocgraph);
	npairs = num_vertices(assocgraph);

	if (verbose>=2)
		printGraph(output, "distance", exp, assocgraph);

	if (verbose)
		cout << "# " << edges << " connections where drawn between the " << npairs 
				 << " remaining pairs." << endl;
	/* Overlap filtering */
	if (verbose)
	{
		cout << "# Overlap filtering ... ";
		cout.flush();
	}

	OverlapFilter overlapfilter(mesh, distancemaps, exp.rotation_step, rlevel, 
 															angletol,	patches, hmaps);
 	overlapfilter(assocgraph);
 	npairs = num_vertices(assocgraph);
	edges = num_edges(assocgraph);

	if (verbose>=2)
		printGraph(output, "overlap", exp, assocgraph);

	if (verbose)
		cout << "done\n"
				 << "# " << edges << " connections remain for " << npairs 
				 << " remaining pairs." << endl;

	/* PERFORMING BRON KERBOSCH */
	clusters_t clusters;
	Clustering(mesh, hmaps, assocgraph, clusters, verbose, &scorings);


		/* SINGLE LINKAGE */

		if (verbose)
			cout << "Perform single linkage cluster analysis ... " << flush;
	
		SingleLink dendrogramm(&clusters, linkage);
		dendrogramm.doSingleLinkage(&scorings);

		if (verbose) 
			cout << "done" << endl;
	
	/* XML OUTPUT */

	DOMDocument* doc = XMLUtils::createDocument("ducking");
	DOMElement* root = doc->getDocumentElement();

	writeResults(root, dbhost, database, experiment, exp.radius, rlevel, mindist, 
							 tolerance, angletol, fuzzytol, cpcutoff, cpcol, fuzzycol, hsicol,
							 &dendrogramm);
	
	string xmlfile = output+".xml";
	XMLConvenientWriter* writer = new XMLConvenientWriter();
	writer->setNode(doc);
	writer->writeFile(xmlfile);
	delete writer;

	doc->release();

	/* REMOVING DATE ENTRIES */

	clearGraph(assocgraph);
	
	for(i=0;i<2;++i)
	{
		//delete meshes
		if (mesh[i]) delete mesh[i];

		//delete distance maps
		for(Filter::distancemap_it d=distancemaps[i]->begin();
				d!=distancemaps[i]->end();++d)
			delete[] d->second;
		delete distancemaps[i];

		//delete patchmaker and hmap map
		delete patches[i];
		delete hmaps[i];
	}

	//close database
	mysql_close(&dbh);

	return 0;
}
