/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef FILTER_HH
#define FILTER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#  else
#    include <map.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <map>
#endif //HAVE_CONFIG_H

#include <trimesh.hh>
#include <assocgraph.hh>

/**
 * \brief a new exception class for errors in the filter operation
 */

class FilterError
{
private:
	int fCode;
	char* fMessage;
	
public:
	FilterError(const int code) : fCode(code), fMessage(0) { }
	FilterError(const int code, const char* msg);
	
	~FilterError();
	
	const char* Message() const { return fMessage; }
	int Code() const { return fCode; }
};

const int FE_DISTANCES_MISSING = 1000;

/**
 * \brief a base class for association graph filters
 *
 * The purpose of the filter base class is to provide a common basis for all the
 * distinctive filters. The base class manages the initialization with surface
 * meshes, distance maps as well as some commonly used operations (e.g. removal
 * of isolated matching point pairs).
 *
 * Filter is an abstract base class.
 */
class Filter 
{
public:

	//*****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//*****************************************************************************
	
	typedef map<Trimesh::vertex_t, double*>::const_iterator distancemap_cit;
	typedef map<Trimesh::vertex_t, double*>::iterator distancemap_it;
	typedef map<Trimesh::vertex_t, double*> distancemap_t;
	/**< 
	 * \brief definition of the on-surface distance maps for the selected points
	 * on surface A and B + iterators.
	 *
	 * Because the association graph only contains a subset of vertices on A and B
	 * the distances need not be calculated for all vertices on both
	 * surfaces. Thus a map is used to store all the necessary distance vectors.
	 */
	
	//*****************************************************************************
	// CONSTRUCTOR AND INITIALIZER
	//*****************************************************************************

	/**
	 * \brief standard constructor creating an uninitialized instance of Filter
	 */
	Filter();
	
	/**
	 * \brief initializing constructor producing a basically initialized instance
	 * of Filter
	 *
	 * The constructor internally calls init(Trimesh**, const distancemap_t**) to
	 * perform the filter initialization.
	 *
	 * \param surfaces  a two membered field of pointers to trimesh structures
	 * \param distances a two membered field of constant pointers to distance maps
	 *                  for surface 0 and 1 (or A and B).
	 */
	Filter(Trimesh** surfaces, distancemap_t** distances);
	
	/**
	 * \brief initialization method for a basic initialization
	 *
	 * This init-method corresponds to Filter(Trimesh**, const distancemap_t**)
	 * and assigns the given parameters to the array of surfaces and distance
	 * maps.
	 *
	 * \param surfaces  a two membered field of pointers to trimesh structures
	 * \param distances a two membered field of constant pointers to distance maps
	 *                  for surface 0 and 1 (or A and B).
	 */
	void init(Trimesh** surfaces, distancemap_t** distances);

	//*****************************************************************************
	// OPERATOR
	//*****************************************************************************

	/**
	 * \brief the filter operator
	 *
	 * The filter operator must be defined by the child filter classes and applies
	 * the filter conditions to the given association graph.
	 *
	 * Any other important parameters for a specific filtering must be set either
	 * in the initialization or by property methods.
	 *
	 * \param assocgraph a reference to an association graph, that should be
	 *                   processed.
	 * 
	 * \return the number of modified graph elements.
	 */
	virtual size_t operator()(assocgraph_t& assocgraph) = 0;

protected:
	
	//*****************************************************************************
	// PROTECTED DATA MEMBERS
	//*****************************************************************************
	
	/**
	 * \brief a two membered field pointing to the compared surfaces
	 */
	Trimesh* pSurfaces[2];
	
	/**
	 * \brief a two membered field containing constant pointers to the distance
	 * maps of both surfaces
	 * 
	 * The distance maps must contain a vector of on-surface-distances for any
	 * vertex on A or B that is present in the association graph.
	 */
	distancemap_t* pDistances[2];

	//*****************************************************************************
	// PROTECTED METHODS
	//*****************************************************************************

	/**
	 * \brief the deletor for isolated matching point pairs.
	 *
	 * For the sake of efficiency it is important to keep the association graphs
	 * as simple as possible. Thus any isolated matching point pairs should be
	 * removed after the a filtering step.
	 *
	 * \param assocgraph a reference to an association graph, that should be
	 *                   processed.
	 * \param valence    vertices with a maximum of \p valence edges are discarded.
	 * 
	 * \return the number of removed matching point pairs.
	 */
	size_t removeIsolated(assocgraph_t& assocgraph, const int valence = 0);


};


typedef map<Trimesh::vertex_t, Trimesh> hmaps_t;
typedef hmaps_t::iterator hmaps_it;

#endif //FILTER_HH
