/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include <float.h>
#include <math.h>

#include <fileformats.hh>
#include "overlapfilter.hh"
#include <Vector.hh>

#ifndef PI
#undefine PI
#endif

#define PI 3.14159265358979

using namespace RazorBack;

//*******************************************************************************
Vector_<double> cross(const Vector_<double>& a, const Vector_<double>& b)
//*******************************************************************************
{
	if ((a.dim()==3) &&(b.dim()==3))
	{
		Vector_<double> prod(3,0);
		prod[0] = a[1] * b[2] - a[2] * b[1];
		prod[1] = a[2] * b[0] - a[0] * b[2];
		prod[2] = a[0] * b[1] - a[1] * b[0];

		return prod;
	}
	else
		return Vector_<double>(1,0);
}

//*******************************************************************************
double lineDistance(const Vector_<double>& cA, const Vector_<double>& cB,
										const Vector_<double>& bV)
//*******************************************************************************
{
	if (cA.dim()==cB.dim()==bV.dim()==3)
	{
		Vector_<double> line = cB-cA;
		
		return cross(line, (cA-bV)).vec_len() / line.vec_len();
	}
	else 
		return DBL_MAX;
}

//*******************************************************************************
OverlapFilter::OverlapFilter(Trimesh** surfaces, distancemap_t** distances,
														 const double step,
														 const double rlevel, const double angletolerance,
														 patchMaker** patches, hmaps_t** hmaps)
//*******************************************************************************
:Filter(), fAngleTolerance(0), fRLevel(0), fRotationStep(0.0)
{
	init(surfaces, distances, step, rlevel, angletolerance, patches, hmaps);
}
	
//*******************************************************************************
void OverlapFilter::init(Trimesh** surfaces, distancemap_t** distances,
												 const double step,
												 const double rlevel, const double angletolerance,
												 patchMaker** patches, hmaps_t** hmaps)
//*******************************************************************************
{
	Filter::init(surfaces, distances);

	//filter criterion
	fRotationStep = step/180.0*PI;
	fRLevel = rlevel;
	fAngleTolerance = angletolerance/180.0*PI;

	for(int i=0;i<2;++i)
	{
		pPatches[i] = patches[i];
		pHMaps[i] = hmaps[i];
	}
}

//*******************************************************************************
size_t OverlapFilter::operator()(assocgraph_t& assocgraph)
//*******************************************************************************
{
	int c = 0;
	
	/* INITIALIZE */
	//necessary property maps
	assocgraph_traits::pairmap_t pairmap = get(vertex_pair, assocgraph);
	assocgraph_traits::rfuncmap_t rfuncmap = get(vertex_Rfunc, assocgraph);
	
	assocgraph_traits::axisAmap_t axisAmap = get(edge_axisA, assocgraph);
	assocgraph_traits::axisBmap_t axisBmap = get(edge_axisB, assocgraph);
	assocgraph_traits::anglemap_t anglemap = get(vertex_angle, assocgraph);
	assocgraph_traits::anglesmap_t anglesmap = get(edge_angles, assocgraph);
	assocgraph_traits::distancesmap_t distancesmap = get(edge_distances, assocgraph);
	assocgraph_traits::rsmap_t rsmap = get(edge_Rs, assocgraph);

	/* LOOP OVER ALL EDGES */
	assocgraph_traits::edges_t eits = edges(assocgraph);
	for(assocgraph_traits::edge_it edge=eits.first;edge!=eits.second;)
	{
		//maybe the edge will be removed so save the iterator
		assocgraph_traits::edge_t e = *edge++;

		//retrieve the source and target vertex
		assocgraph_traits::vertex_t v = source(e, assocgraph);
		assocgraph_traits::vertex_t w = target(e, assocgraph);

		//determine the nearest point on surface A and B
		put(axisAmap, e, getNearestPoint(0,pairmap[v].first, pairmap[w].first));
		put(axisBmap, e, getNearestPoint(1,pairmap[v].second, pairmap[w].second));

		//2D array of nearest point to zero point angles
		double nz_angle[2][2];

		//extract the map position of the zero vertex and the nearest point of all
		//patches
		for (int s=0;s<2;++s)
		{
			Trimesh::vertex_t center[2];
			Trimesh::vertex_t nv[2];
			Trimesh::vertex_t zv[2];

			center[0] = (s==0) ? pairmap[v].first : pairmap[v].second;
			center[1] = (s==0) ? pairmap[w].first : pairmap[w].second;

// 			cout << "(" << center[0] << ", " << center[1] << ")" << " ";
			
			nv[0] = (s==0) ? axisAmap[e].first : axisBmap[e].first;
			nv[1] = (s==0) ? axisAmap[e].second : axisBmap[e].second;


			//extract the coordinates
			for(int p=0;p<2;++p)
			{
				hmaps_it map;
				map = pHMaps[s]->find(center[p]);

				Vector_<double> np = map->second.vertCoordinates(nv[p]);

				/* IMPORTANT: Only valid if RADIUS of MAP = 1 */
 				if ((np[0]-1.0)>-DBL_EPSILON) np[0]=1.0;
 				if ((-1.0-np[0])>-DBL_EPSILON) np[0]=-1.0;

				if (np[1]<-DBL_EPSILON)
					nz_angle[s][p] = 2*PI - acos(np[0]) ;
				else
					nz_angle[s][p] = acos(np[0]);

			}
		}

		int nrotations = rint(2*PI/fRotationStep);
		double maxR[2];
		double maxAngle[2];

		bool angle_ok = false;

		//loop over both patchpairs
		for (int p=0;p<2;++p)
		{
			maxR[p] = -DBL_MAX;

			//retrieve the correct correlation function
			double* rfunc = rfuncmap[(p==0) ? v : w];
			
			//calculate the angle difference
// 			double alpha = nz_angle[0][p] - anglemap[(p==0) ? v : w];
// 			if (alpha<0) alpha += 2*PI;
// 			if (alpha>2*PI) alpha -= 2*PI;
// 			double beta  = nz_angle[1][p];

			double delta = fabs(nz_angle[1][p] - nz_angle[0][p] - anglemap[(p==0) ? v : w]);
			if (delta>2*PI) delta -= 2*PI;
			if (delta>PI) delta = 2*PI - delta;

// 			cout << delta << " ";
			
			angle_ok = angle_ok || (delta<fAngleTolerance); 
//      if (anglediff<0) anglediff = 2 * PI + anglediff;
			//store the angle 
			((p==0) ? anglesmap[e].first : anglesmap[e].second) = 
				nz_angle[1][p] - nz_angle[0][p];

// 			double highangle = anglediff + fAngleTolerance;
// 			double lowangle  = anglediff - fAngleTolerance;
			
// 			if (highangle > 2 * PI) highangle -= 2 * PI;
// 			if (lowangle < 0) lowangle = 2 * PI + lowangle;

// 			int start = rint(lowangle/fRotationStep);
// 			int end   = rint(highangle/fRotationStep) + 1;

// 		  for(int i=start;i<end;++i)
// 				if (rfunc[i]>maxR[p]) 
// 				{
// 					maxR[p] = rfunc[i];
// 					maxAngle[p] = i*fRotationStep;
// 				}
		}
		
		/* FILTERING */

// 		if ((maxR[0]<(fRLevel)) || (maxR[1]<(fRLevel)))

		if (!angle_ok)
		{
			remove_edge(e, assocgraph);
			++c;
		}

// 		cout << endl;
	}

	//remove isolated vertices
	while (removeIsolated(assocgraph,0)>0) ;

	return c;
}


//*******************************************************************************
pair<Trimesh::vertex_t,Trimesh::vertex_t>
OverlapFilter::getNearestPoint(const int six, 
															 Trimesh::vertex_t cA, Trimesh::vertex_t cB)
//*******************************************************************************
{
	/* INITIALIZATION */
	Trimesh::vertex_t center[2];
	center[0] = cA;
	center[1] = cB;


  surfacePatch* patch[2];
	patch[0] = (*pPatches[six])[cA];
	patch[1] = (*pPatches[six])[cB];

	//position vectors
	Vector_<double> cpos[2];
	cpos[0] = pSurfaces[six]->vertCoordinates(center[0]);
	cpos[1] = pSurfaces[six]->vertCoordinates(center[1]);

	pair<Trimesh::vertex_t,Trimesh::vertex_t> result;

	//length of line between cA and cB
	Vector_<double> line = cpos[1] - cpos[0];
	double linelength = line.vec_len();

	//loop over all patch sites
	for(int s=0;s<2;++s)
	{
		int snot = (s==0) ? 1 : 0;
		Trimesh::vertex_t bV;

		double mindist = DBL_MAX;

		list<Trimesh::vertex_t>::const_iterator v;
		for(v = patch[s]->borderBegin();v!=patch[s]->borderEnd();++v)
		{
			if ((mindist-DBL_MAX)>-DBL_EPSILON) bV = *v;
			
			const Vector_<double>& vpos = patch[s]->vertCoordinates(*v);
			//calculate the surface distance
			double distance = cross(line, (cpos[0]-vpos)).vec_len()/linelength;
			//check with minimum if it is between cA and cB
			if (((vpos-cpos[snot]).vec_len()<linelength) &&	(mindist>distance))
			{
				mindist = distance;
				bV = *v;
			}
		}

		((s==0) ? result.first : result.second) = bV;
	}

	return result;
}
