/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef OVERLAPFILTER_HH
#define OVERLAPFILTER_HH

#include <mapping.hh>
#include <patchmaker.hh>
#include "filter.hh"

/**
 * \brief a filter checking if two matching point pairs can overlap
 * simultanously
 *
 * The overlap filter checks if two matching point pairs connected by an edge in
 * the association graph can overlap at the same time. This check is done by
 *
 * - calculating the axes between A1A2 and B1B2 and the border vertices on A and
 * B that lie along this axis.
 *
 * - computing the angles (on A and B) between the zero vertex and the vertex on
 * the axis.
 *
 * - calculating an angle difference between alpha(B1-A1) and alpha(B2-A2).
 *
 * - examine the angular correlation function of A1 and A2 for the difference
 * angles and in their near vicinity (filter parameter)
 *
 * - Delete all edges that do not show a significant correlation (same threshold
 * as in the DuckFilter) in the angular range around the difference angles
 *
 * The filter has two parameters: (a) the significance level which should be the
 * same as in the DuckFilter (fThreshold), and (b) the angular tolreance
 * (fAngleTolerance).
 */
class OverlapFilter : public Filter
{
public:
	
	//*****************************************************************************
	// CONSTRUCTORS AND INITIALIZATION
	//*****************************************************************************
	
	/**
	 * \brief standard constructor creating an uninitialized instance
	 *
	 * \sa Filter::Filter()
	 */
	OverlapFilter() : Filter() { }
	
	/**
	 * \brief full initializing constructor
	 *
	 * The constructor internally calls the init method and passes all the
	 * parameters to create a fully initialized instance. The mechanism is similar
	 * to the initializing constructor of Filter.
	 *
	 * \param surfaces       a two membered field of pointers to trimesh structures
	 * \param distances      a two membered field of constant pointers to distance maps
	 *                       for surface 0 and 1 (or A and B).
	 * \param step           the rotation step of the harmonic images (in degrees)
	 * \param rlevel         the threshold for the correlation coefficient.
	 * \param angletolerance angular tolerance
	 * \param patches        two membered field of patchMaker pointers containing the 
	 *                       experiment's patches
	 * \param hmaps          two membered field of 
	 *
	 * \sa Filter::Filter(Trimesh**, const distancemap_t*)
	 */
	OverlapFilter(Trimesh** surfaces, distancemap_t** distances, const double step,
								const double rlevel, const double angletolerance,
								patchMaker** patches, hmaps_t** hmaps);
	
	/**
	 * \brief initializing method
	 *
	 * This method performs a full initialization of a DistanceFilter instance.
	 *
	 * \param surfaces       a two membered field of pointers to trimesh structures
	 * \param distances      a two membered field of constant pointers to distance maps
	 *                       for surface 0 and 1 (or A and B).
	 * \param step           the rotation step of the harmonic images (in degrees)
	 * \param rlevel         the threshold for the correlation coefficient.
	 * \param angletolerance angular tolerance
	 * \param patches        two membered field of patchMaker pointers containing the 
	 *                       experiment's patches
	 * \param hmaps          two membered field of 
	 *
	 * \sa Filter::init()
	 */
	void init(Trimesh** surfaces, distancemap_t** distances, const double step,
						const double rlevel, const double angletolerance,
						patchMaker** patches, hmaps_t** hmaps);
	

	//*****************************************************************************
	// OPERATOR
	//*****************************************************************************

	/**
	 * \brief the filter operator
	 *
	 * For further details see Filter::operator().
	 *
	 * \param assocgraph a reference to an association graph, that should be
	 *                   processed.
	 * 
	 * \return the number of modified graph elements.
	 */
	size_t operator()(assocgraph_t& assocgraph);

private:

	//*****************************************************************************
	// PRIVATE DATA MEMBERS
	//*****************************************************************************

	/**
	 * \brief the rotation step of the harmonic images
	 */
	double fRotationStep;
	
	/**
	 * \brief filter criterion: the correlation coefficient threshold
	 *
	 * All matching point pairs that show a correlation coefficient below this
	 * threshold are discarded by the filter.
	 */
	double fRLevel;

	/**
	 * \brief the angular tolerance
	 *
	 * The filter scans the difference angles +/- the angular tolerance for an
	 * R-value that lies above the threshold
	 */
	double fAngleTolerance;
	
	/**
	 * \brief two pointers to the patchMaker objects for the experiment's surfaces.
	 *
	 * The patches need not to be generated in advance, but any patch that is
	 * generated during the filtering is added to the patchMaker objects and
	 * available for further use.
	 */
	patchMaker* pPatches[2];

	/**
	 * \brief two pointers to a map of harmonic maps.
	 *
	 * The maps of harmonic maps work like the pointers to the patchmaker objects:
	 * No maps are necessary when calling the filter, but any map generated will
	 * be placed in the maps.
	 */
	hmaps_t* pHMaps[2];

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************
	
	/**
	 * \brief finds the nearest poinst on the patch around cA to the center cB
	 *
	 * The methods finds the point on the border of the patches A and B that have
	 * minimum distance to the line between cA and cB.
	 *
	 * \param six index of the surface where both patches are placed
	 * \param cA,cB a pair of patch centers
	 *
	 * \return a pair of surface vertices \c first contains the nearest point of A
	 * and \c second the same point of B.
	 */
	pair<Trimesh::vertex_t,Trimesh::vertex_t>
	getNearestPoint(const int six, Trimesh::vertex_t cA, Trimesh::vertex_t cB);
};

#endif //OVERLAPFILTER_HH
