/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef POSITIONFILTER_HH
#define POSITIONFILTER_HH

#include <Vector.hh>

#include "filter.hh"

/**
 * \brief a filter selecting the critical points of a specific spherical region
 *
 * This filter selects all pairs of critical points which are placed within a
 * sphere around a specific point in space. The filter needs two points \p
 * center and a cutoff \p radius in Ang and removes any pair of critical points
 * where either one of them is placed outside the sphere (\p center, \p radius )
 * of the specific surface.
 */

class PositionFilter: public Filter
{
public:

	//*****************************************************************************
	// CONSTRUCTOR AND DESTRUCTOR
	//*****************************************************************************

	/**
	 * \brief default constructor
	 */
	PositionFilter() : Filter() {}

	/**
	 * \brief initializing constructor creating a working instance of PositionFilter
	 *
	 * The constructor calls the init() to setup the filter parameters.
	 *
	 * \param surfaces        a two membered field of pointers to trimesh 
	 *                        structures
	 * \param distances       a two membered field of constant pointers to distance
	 *                        maps for surface 0 and 1 (or A and B).
	 * \param sphereA,sphereB a 4D position vector (x,y,z,r) defining the sphere on 
	 *                        A or B respectively
	 * \sa PositionFilter(), Filter::Filter(Trimesh**, distancemap_t*)
	 */
	PositionFilter(Trimesh** surfaces, distancemap_t** distances, 
								 const Vector_<double>& sphereA, const Vector_<double>& sphereB);

	/**
	 * \brief initializing routine
	 *
	 * \param surfaces        a two membered field of pointers to trimesh 
	 *                        structures
	 * \param distances       a two membered field of constant pointers to distance
	 *                        maps for surface 0 and 1 (or A and B).
	 * \param sphereA,sphereB a 4D position vector (x,y,z,r) defining the sphere on 
	 *                        A or B respectively
	 * \param radius          the radius of the selection sphere
	 *
	 * \sa Filter::init(Trimesh**, const distancemap_t*)
	 */
	void init(Trimesh** surfaces, distancemap_t** distances, 
						const Vector_<double>& sphereA, const Vector_<double>& sphereB);

	//*****************************************************************************
	// OPERATORS
	//*****************************************************************************

	/**
	 * \brief the filter operator
	 *
	 * For further details see Filter::operator().
	 *
	 * \param assocgraph a reference to an association graph, that should be
	 *                   processed.
	 * 
	 * \return the number of modified graph elements.
	 */
	size_t operator()(assocgraph_t& assocgraph);

private:

	//*****************************************************************************
	// PRIVATA DATA MEMBERS
	//*****************************************************************************
	
	/**
	 * \brief the sphere radius at both surfaces
	 *
	 *
	 * The radius defines how far a critical point can be placed away from the
	 * selection center on each surface.
	 */
	double fRadius[2];

	Vector_<double> fCenter[2];
	/**<
	 * \brief the centers of the selection spheres on both surfaces.
	 *
	 * The center of the selection sphere can be placed anywhere in space, it does
	 * not have to be on the surface.
	 */

};


#endif // POSITIONFILTER_HH

