/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef BINSTRIMESH_CC
#define BINMSTRIMESH_CC

#include "string.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <fstream>
#    include <sstream>
#    include <cstdio>
#    include <iostream>
#    include <iomanip>
#    include <algorithm>
#    include <vector>
#  else
#    include <fstream.h>
#    include <sstream.h>
#    include <cstdio.h>
#    include <iostream.h>
#    include <iomanip.h>
#    include <algo.h>
#    include <map.h>
#    include <vector.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <fstream>
#  include <sstream>
#  include <cstdio>
#  include <iostream>
#  include <iomanip>
#  include <algorithm>
#  include <vector>
#endif //HAVE_CONFIG_H

#include "bintrimesh.hh"

#define FILEID "BINMESH"

//*****************************************************************************
bool binTrimesh::read(const string& aFile)
//******************************************************************************
{
  //open the File
  ifstream stream(aFile.c_str(), ios::in|ios::binary);
  if (!stream)
  {
    cerr << "Coordinate file: " << aFile << " not found" << endl;
    return false;
  }

  bool result = read(stream);  

  stream.close();
  return result;
}

//******************************************************************************
bool binTrimesh::read(istream& aStream)
//******************************************************************************
{
	//check the file ID
	char id[7];
	aStream.read(id, 7);
	
	if (strncmp(id,FILEID,7)==0)
	{
	
		mesh_t& Mesh = getMesh();
		vector<Triangle>& Triangles = getTriangles();
		map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();
  
		// number of vertices, number of spheres, probe radius and density
		size_t nvertex, nfaces, nedges;

		double xyz[3];     // Coordinates
		double nxyz[3];    // Normals
		int vertex_kind;   // vertex kind
		int master_ix;          // index on the master surface
		int csphere;            // index of the closest VdW sphere

		Trimesh::vertex_t v;

		// Read header information
		aStream.read((char*) &nvertex, sizeof(size_t));
		aStream.read((char*) &nfaces , sizeof(size_t));
		aStream.read((char*) &nedges , sizeof(size_t));

		//read vertices
		for (int i=0;i<nvertex;i++)
		{
			aStream.read((char*) xyz,  sizeof(double) * 3);
			aStream.read((char*) nxyz, sizeof(double) * 3);
			aStream.read((char*) &vertex_kind, sizeof(int));
			aStream.read((char*) &master_ix, sizeof(int));
			aStream.read((char*) &csphere, sizeof(int));
			
			v = add_vertex(Mesh);
			
			//copy coords and normals to Vector_'s
			Vector_<double> xyz_(xyz, 3);
			Vector_<double> nxyz_(nxyz, 3);

			// Store vertex coordinates and normals and the other flags
			vertexdata_t data;
			data.masterIndex = master_ix;
			data.ownIndex    = i;
			data.kind        = vertex_kind;
			data.coordinates = xyz_;
			data.normal      = nxyz_;
			data.cSphere     = csphere;

			put(*data_map, v, data);
			
		}

		Triangle atriangle;  // a triangle for the triangle list

		// Read triangles
		for (int i=0;i<nfaces;i++)
		{
			size_t nvertices = num_vertices(Mesh);
			size_t triangle[3];
			double eweight[3];
    
			int nfaceVertices = 0;
			
			aStream.read((char*) &nfaceVertices, sizeof(int));
			aStream.read((char*) triangle, sizeof(size_t) * 3);
			aStream.read((char*) eweight, sizeof(double) * 3);
			
			if (nfaceVertices!=3)
			{
				cerr << "Wrong number of vertices in face Nr. " << i+1 << "!" << endl;

				Mesh.clear();
				Triangles.clear();
				return false;
			}

			//triangle index starts from zero
			//for(int j=0;j<3;++j) triangle[j]--;

			// Check if vertices of the triangle exists, than check if edges are not
			// already added to the graph than add edge to the graph

			if ((triangle[0]<0) || (triangle[0]>nvertex))
			{
				cerr << "Error: Vertex " << triangle[0] << " in triangle " << i+1
						 << " has no coordinates!" << endl;

				Mesh.clear();
				Triangles.clear();
				return false;
			}
			if ((triangle[1]<0) || (triangle[1]>nvertex))
			{
				cerr << "Error: Vertex " << triangle[1] << " in triangle " << i+1
						 << " has no coordinates!" << endl;

				Mesh.clear();
				Triangles.clear();
				return false;
			}

			if ((triangle[2]<0) || (triangle[2]>nvertex))
			{
				cerr << "Error: Vertex " << triangle[2] << " in triangle " << i+1
						 << " has no coordinates!" << endl;

				Mesh.clear();
				Triangles.clear();
				return false;
			}

			// Get the nth vertex in the graph's vertex list.
			Trimesh::vertex_t v1 = vertex(triangle[0],Mesh);
			Trimesh::vertex_t v2 = vertex(triangle[1],Mesh);
			Trimesh::vertex_t v3 = vertex(triangle[2],Mesh);

			//create the triangle object and add it to the trianglelist
			Triangle atriangle(edgepoint(triangle[0],&((*data_map)[v1].coordinates)),
												 edgepoint(triangle[1],&((*data_map)[v2].coordinates)),
												 edgepoint(triangle[2],&((*data_map)[v3].coordinates)));

			Triangles.push_back(atriangle);
			//adding the index of the face to the vertfaceIndex entries of all three
			//vertices
			for(int j=0;j<=2;++j)
				vertfaceIndex[triangle[j]].push_back(i);


			//check if any edges are allready there
			bool has_edge1 = edge(v1,v2, Mesh).second;
			bool has_edge2 = edge(v2,v3, Mesh).second;
			bool has_edge3 = edge(v3,v1, Mesh).second;

			//add every edge, that does not exist yet, and calculate its length
			pair<Trimesh::edge_t, bool> e_b;   //edge, boolean pair (returned from add_edge

			if (!has_edge1)
			{
				e_b = add_edge(v1,v2, Mesh);
				put(*edgelengths_map, e_b.first, eweight[0]);
			}
			if (!has_edge2)
			{
				e_b = add_edge(v2,v3, Mesh);
				put(*edgelengths_map, e_b.first, eweight[1]);
			}
			if (!has_edge3)
			{
				e_b = add_edge(v3,v1, Mesh);
				put(*edgelengths_map, e_b.first, eweight[2]);
			}
	
		}
		return true;
	}
	else
	{
		cerr << "Wrong binary file format!" << endl;
		return false;
	}
}

//******************************************************************************
bool binTrimesh::write(const string& aFile)
//******************************************************************************
{
  //open the File
  ofstream stream(aFile.c_str(), ios::out|ios::binary);
  if (!stream)
  {
    cerr << "File: " << aFile << " could not be created" << endl;
    return false;
  }

  bool result = write(stream);  

  stream.close();
  return result;
}


//******************************************************************************
bool binTrimesh::write(ostream& aStream)
//******************************************************************************
{
  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

	size_t n_vertices = num_vertices(Mesh);
	size_t n_faces    = Triangles.size();
	size_t n_edges    = num_edges(Mesh);

	//calculate size of buffer and alloc the memory
	size_t vertex_size = 6 * sizeof(double) + 3 * sizeof(int);
	size_t face_size = 4 * sizeof(int);

	//head = ID(7) + num_vertices(size_t) + num_faces(size_t) + num_edges(size_t)
	size_t head_size = 7 + 3 * sizeof(size_t);

	//write file ID
	aStream.write(FILEID, 7);

  //#vertices, #faces, #edges
	aStream.write((const char*) &n_vertices, sizeof(size_t));
	aStream.write((const char*) &n_faces   , sizeof(size_t));
	aStream.write((const char*) &n_edges   , sizeof(size_t));
	
  //write the vertices
  for(Trimesh::vertex_it v=mesh_structure->begin();v!=mesh_structure->end();++v)
  { 
		double coords[3];
		double normal[3];
		const int& kind = (*data_map)[*v].kind;
		const size_t& mixs = (*data_map)[*v].masterIndex;
		const int& csph = (*data_map)[*v].cSphere;

		copy((*data_map)[*v].coordinates.begin(), (*data_map)[*v].coordinates.end(), 
				 coords);
		copy((*data_map)[*v].normal.begin(), (*data_map)[*v].normal.end(),
				 normal);

		aStream.write((const char*) coords, sizeof(double) * 3);
		aStream.write((const char*) normal, sizeof(double) * 3);
		aStream.write((const char*) &kind,   sizeof(int));
		aStream.write((const char*) &mixs,   sizeof(int));
		aStream.write((const char*) &csph,   sizeof(int));
  }

  //write the triangles
  for(Trimesh::FaceIter f=mesh_structure->beginFace();f!=mesh_structure->endFace();++f)
  {
		int points = 3;
		int i      = 0;

		size_t* vixs = new size_t[points];
		double* eweight= new double[points];

		for(i=0;i<points;++i)
			vixs[i] = f->getVertexIndex(i);

		for(i=0;i<points;++i)
		{
		  int j = ((i+1)<points) ? i+1 : 0;
			Trimesh::vertex_t v = mesh_structure->getVertex(vixs[i]);
			Trimesh::vertex_t u = mesh_structure->getVertex(vixs[j]);

			pair<Trimesh::edge_t,bool> e = edge(u, v, Mesh);
			eweight[i] = ((*edgelengths_map)[e.first]);
		}
			

		aStream.write((const char*) &points, sizeof(int));
		aStream.write((const char*) vixs,   sizeof(size_t) * points);
		aStream.write((const char*) eweight, sizeof(double) * points);

		delete[] vixs;
		delete[] eweight;
  }
	
  return true;
}

#endif
