/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef BINTRIMESH_HH
#define BINTRIMESH_HH

#include "trimeshio.hh"

/**
 * \brief trimesh class, able to read and write triangulated mesh objects in
 * the BIN object format
 *
 * The \ref binformat_page "BIN object format" is used by 
 * <a href="http://www.geomview.org/">Geomview</a>, a powerfull mesh display 
 * program. It is very closely related to the \ref ownformat_page "own library format" 
 * of this library.
 */

class binTrimesh : public trimeshIO
{
public:

  /**
   * \brief constructor initializing msmsTrimesh with a Trimesh object
   */
  binTrimesh(Trimesh* aTrimesh): trimeshIO(aTrimesh) { }

  /** 
   * \brief there is a stream handling for library own format files.
   */
  bool handleStreams() const { return true; }


  /**
   * \brief reads a surface mesh object in library own format from a file 
   * specified in \p aFile
   *
   */
  bool read(const string& aFile);
  
  /**
   * \brief reads a surface mesh object in library own format from the stream
   * specified in \p aStream
   */
  bool read(istream& aStream);

  /**
   * \brief writes a surface mesh object in library own format into a file 
   * specified in \p aFile
   *
   */
  bool write(const string& aFile);

  /**
   * \brief writes a surface mesh object in library own format into a stream 
   * specified in \p aStream
   *
   */
  bool write(ostream& aStream);
};

#endif 


/**
 * \page binformat_page BIN object format
 *
 * This is the internal file format for BLOB database and file storage. This is 
 * a compact binary format.
 *
 * \section binfile_header File Header:
 * The first seven characters of the file are used as a file identification:
 * "BINMESH" indicates that the file has the BIN object format.
 * The id-string is followed by three unsigned 4byte ints:
 *  - number of vertices
 *  - number of faces (triangles)
 *  - number of edges
 *
 * 
 * \section binfile_vertices Vertex section:
 * The vertex data follows directly after the file header. For each vertex the
 * following information is available:
 * 
 * - vertex coordinates   ( 3 * double)
 * - vertex normal vector ( 3 * double)
 * - kind of the vertex   ( 4 byte int)
 * - vertex master index  ( 4 byte int)
 * - idx. of chem. sphere ( 4 byte int)
 *
 * There is one entry for each vertex (1 to number of vertices).
 * 
 * \section binfile_faces Face section:
 * The faces follow directly after the vertex data containing the following data
 * for each face:
 *
 * - number of vertices (points) in this face [usually 3] (4 byte int)
 * - indices of the face vertices ( points * 4 byte unsigned int)
 * - edgweights for each edge on the face's border (points * double). The first
 *   edge is the one between the first and the second vertex of the face and the
 *   last one is that between the last and the first vertex of the face.
 */
