/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include <conntab.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <set>
#  else
#    include <set.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <set>
#endif //HAVE_CONFIG_H

//*******************************************************************************
conntab::conntab()
//*******************************************************************************
:ctab_graph(), cvertexmap()
{
  vertexvert_map = get(vertex_vert, ctab_graph);
  vertexborder_map = get(vertex_border, ctab_graph);
  edgeface_map = get(edge_face, ctab_graph);
  edgelost_map = get(edge_lost, ctab_graph);
  edgeborder_map = get(edge_border, ctab_graph);
}

//*******************************************************************************
conntab::cVertex conntab::getBorderFace(const cVertex& cvertex)
//*******************************************************************************
{
  //look for cvertex in the cvertexmap
  cvertexmap_it vertex = cvertexmap.find(cvertex);

  //check if the cvertex is available
  if (vertex==cvertexmap.end()) throw unknown_cvertex(cvertex);
  //check if the vertex is on a border
  if (!vertexborder_map[vertex->second]) throw no_border_vertex(cvertex);

  pair<outedges_it, outedges_it> outits = out_edges(vertex->second, ctab_graph);
  //find the edge leading to another vertex on the border
  outedges_it e = find_if(outits.first, outits.second, check_edge_border(*this));

  //if no further edge could be found the structure is inconsistent
  if (e==outits.second) throw inconsistent_structure();

  return edgeface_map[*e];
  
}


//*******************************************************************************
void conntab::resetBorderFlags()
//*******************************************************************************
{
  pair<edge_it, edge_it> edge_its = edges(ctab_graph);
  //search for any border edges
  edge_it wrong_edge = find_if(edge_its.first, edge_its.second,
			       check_edge_border(*this));
				   
  //if a further edge could be found the structure is inconsistent
  if (wrong_edge!=edge_its.second) throw inconsistent_structure();
  
  pair<vertex_it, vertex_it> vits = vertices(ctab_graph);
  for(vertex_it v=vits.first;v!=vits.second;++v)
    vertexborder_map[*v] = false;
}

//*******************************************************************************
bool conntab::addVertex(const cVertex& cvertex)
//*******************************************************************************
{
  if (cvertexmap.find(cvertex)==cvertexmap.end())
  {
    Vertex v = add_vertex(ctab_graph);
    
    put(vertexvert_map, v, cvertex);
    put(vertexborder_map, v, false);

    cvertexmap[cvertex] = v;

    return true;
  }
  else 
    return false;
}

//*******************************************************************************
bool conntab::addEdge(const cVertex& source, const cVertex& target, 
		      const cVertex& face)
//*******************************************************************************
{
  //look for the source and target vertex in the cvertexmap
  cvertexmap_it v_source = cvertexmap.find(source);
  cvertexmap_it v_target = cvertexmap.find(target);

  //check if the vertices are available
  if (v_source==cvertexmap.end()) throw unknown_cvertex(source);
  if (v_target==cvertexmap.end())  throw unknown_cvertex(target);
  
  if (!edge(v_source->second, v_target->second, ctab_graph).second)
  {
    pair<Edge,bool> new_edge = add_edge(v_source->second, v_target->second, 
					ctab_graph);

    put(edgeface_map, new_edge.first, face);
    put(edgeborder_map, new_edge.first, false);
    
    //check for counter edges
    pair<Edge, bool> 
      counter = edge(v_target->second, v_source->second, ctab_graph);
    
    //if there is a counter edge add to loosefaces otherwise remove own entry
    if (counter.second)
    {
      put(edgelost_map, counter.first, (void *)0);
      put(edgeborder_map, counter.first, false);
    }
    else 
      put(edgeborder_map, new_edge.first, true);

    return true;
  }
  else
    return false;
}

//*******************************************************************************
void conntab::removeVertex(const cVertex& cvertex)
//*******************************************************************************
{
  //look for cvertex in the cvertexmap
  cvertexmap_it vertex = cvertexmap.find(cvertex);

  //check if the cvertex is available
  if (vertex==cvertexmap.end()) throw unknown_cvertex(cvertex);
  
  cvertexmap.erase(cvertex);
      
  remove_vertex(vertex->second,ctab_graph);

}


//*******************************************************************************
bool conntab::removeEdge(const cVertex& source, const cVertex& target)
//*******************************************************************************
{
  //look for the source and target vertex in the cvertexmap
  cvertexmap_it v_source = cvertexmap.find(source);
  cvertexmap_it v_target = cvertexmap.find(target);

  //check if the vertices are available
  if (v_source==cvertexmap.end()) throw unknown_cvertex(source);
  if (v_target==cvertexmap.end())  throw unknown_cvertex(target);

  //remove the edge if there is one
  pair<Edge, bool> remove = edge(v_source->second, v_target->second, ctab_graph);
  if (remove.second)
  {
    pair<Edge, bool> 
      counter = edge(v_target->second, v_source->second, ctab_graph);
    
    //if there is a counter edge add to loosefaces otherwise remove own entry
    if (counter.second)
    {
      put(edgelost_map, counter.first, edgeface_map[remove.first]);
      put(edgeborder_map, counter.first, true);
    }
    
    remove_edge(remove.first, ctab_graph);
    
    //mark the two vertices as border vertices
    put(vertexborder_map, v_source->second, true);
    put(vertexborder_map, v_target->second, true);
    
    return true;
  }
  else
    return false;
}

//*******************************************************************************
int conntab::removeLonely()
//*******************************************************************************
{
  int c=0;
  pair<vertex_it,vertex_it> vits = vertices(ctab_graph);
  
  for(vertex_it v=vits.first;v!=vits.second;vits.first=v)
  {
    ++v;

    //remove the vertex if no edge is leaving the vertex
    if (out_degree(*vits.first, ctab_graph)==0)
    {
      cVertex cv = vertexvert_map[*vits.first];
      cvertexmap.erase(cv);
      
      remove_vertex(*vits.first,ctab_graph);
      ++c;
    }
  }
  
  return c;
}
