/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CONNTAB_HH
#define CONNTAB_HH

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map.hpp>

#include "face.hh"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <algorithm>
#  else
#    include <map.h>
#    include <algo.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <map>
#  include <algorithm>
#endif //HAVE_CONFIG_H


namespace boost
{
  enum vertex_vert_t     { vertex_vert     = 500 };
  enum vertex_border_t   { vertex_border   = 501 };
  enum edge_face_t       { edge_face       = 502 };
  enum edge_lost_t       { edge_lost       = 503 };
  enum edge_border_t     { edge_border     = 504 };

  BOOST_INSTALL_PROPERTY(vertex,vert);
  BOOST_INSTALL_PROPERTY(vertex,border);
  BOOST_INSTALL_PROPERTY(edge,face);
  BOOST_INSTALL_PROPERTY(edge,lost);
  BOOST_INSTALL_PROPERTY(edge,border);
}

class conntab_traits
{
protected:
  typedef adjacency_list_traits<listS,listS,directedS>::vertex_descriptor cVertex;

  /**
   * \brief a vertex iterator for the 
   */
  typedef graph_traits<adjacency_list<listS,listS,directedS> >::vertex_iterator cvertex_it;

  /**
   * \brief specifies the conflictgraph Vertex of the vertex
   */
  typedef property<vertex_vert_t, cVertex> VertexVert;

  /**
   * \brief specifies if the vertex is at the border
   */
  typedef property<vertex_border_t, bool, VertexVert> VertexBorder;

  /**
   * \brief specifies the face to which the (half)edge belongs
   */
  typedef property<edge_face_t, cVertex> EdgeFace;

  /**
   * \brief specifies a lost face along the edge
   */
  typedef property<edge_lost_t, cVertex, EdgeFace> EdgeLost;

  /**
   * \brief specifies if the edge is at the border
   */
  typedef property<edge_border_t, bool, EdgeLost> EdgeBorder;

  /**
   * \brief typdefinition for the conflict graph.
   *
   * The conflict graph is an essential part of the convex-hull algorithm. for more detail
   * see the description of the algorithm.
   */
  typedef adjacency_list<setS, listS, directedS, VertexBorder, EdgeBorder> conntab_t;

public:
  
  /**
   * \brief typdefinition for the vertex descriptor
   */
  typedef graph_traits<conntab_t>::vertex_descriptor Vertex;

  /**
   * \brief vertex iterator for the vertices
   */
  typedef graph_traits<conntab_t>::vertex_iterator vertex_it;

  /**
   * \brief edge iterator for the edges
   */
  typedef graph_traits<conntab_t>::edge_iterator edge_it;

  /**
   * \brief adjacency iterator for the vertices
   */
  typedef graph_traits<conntab_t>::adjacency_iterator adjacency_it;

  /**
   * \brief typdefinition for the vertex descriptor
   */
  typedef graph_traits<conntab_t>::edge_descriptor Edge;

  /**
   * \brief vertex iterator for the vertices
   */
  typedef graph_traits<conntab_t>::out_edge_iterator outedges_it;

  /**
   * \brief propertymap for the conflict graph vertex descriptor
   */
  typedef property_map<conntab_t, vertex_vert_t>::type VertexVertMap;

  /**
   * \brief propertymap for the border flags of all vertices
   */
  typedef property_map<conntab_t, vertex_border_t>::type VertexBorderMap;

  /**
   * \brief propertymap for the half edge's face
   */
  typedef property_map<conntab_t, edge_face_t>::type EdgeFaceMap;

  /**
   * \brief propertymap for the lost faces at the edges
   */
  typedef property_map<conntab_t, edge_lost_t>::type EdgeLostMap;

  /**
   * \brief propertymap for the border flags of all edges
   */
  typedef property_map<conntab_t, edge_border_t>::type EdgeBorderMap;

  /**
   * \brief map type for the mapping of the conflict graph vertices to the 
   * vertices
   */
  typedef map<cVertex, Vertex> cvertexmap_t;
  typedef cvertexmap_t::iterator cvertexmap_it;

  /**
   * \brief thrown as an exception if a conflict graph vertex not known by the
   * connection table is passed to any method except conntab::addVertex.
   *
   * The unknown conflict graph vertex can be retrieved by the 
   * unknown_vertex::vertex method.
   */
  class unknown_cvertex 
  { 
  private:
    cVertex fault_vertex;

  public:
    unknown_cvertex(const cVertex& cvertex): fault_vertex(cvertex) { }
    
    cVertex vertex() const { return fault_vertex; }
  };
  
  /**
   * \brief thrown if a non border vertex is passed as a parameter to a method
   * which needs a border vertex.
   */
  class no_border_vertex 
  { 
  private:
    cVertex fault_vertex;

  public:
    no_border_vertex(const cVertex& cvertex): fault_vertex(cvertex) { }
    
    cVertex vertex() const { return fault_vertex; }
  };

  /**
   * \brief is thrown if the structure of the connection table is inconsistent
   * 
   * The structure of a connection table is inconsistent if 
   * \li there is an open border line (a border vertex has no outgoing edge to
   *     another border vertex).
   */
  class inconsistent_structure { };


};

class conntab : public conntab_traits
{
private:

  /**
   * \brief checks if a vertex is on the border
   */
  class check_vertex_border : public unary_function<conntab::Vertex,bool>
  {
  private:
    const conntab& myconntab;
    
  public:
    check_vertex_border(const conntab& ctab):myconntab(ctab) { }
    bool operator() (const conntab::Vertex v) const 
    { 
      return myconntab.vertexborder_map[v]; 
    }
  };

  friend class check_vertex_border;

  /**
   * \brief checks if a vertex is on the border
   */
  class check_edge_border : public unary_function<conntab::Edge,bool>
  {
  private:
    const conntab& myconntab;
    
  public:
    check_edge_border(const conntab& ctab):myconntab(ctab) { }
    bool operator() (const conntab::Edge e) const 
    { 
      return myconntab.edgeborder_map[e];
    }
  };

  friend class check_edge_border;

  /**
   * \brief graph structure holding the connection table
   */
  conntab_t ctab_graph;

  /**
   * \brief container mapping all the conflictgraph vertices to connection table
   * vertices
   */
  cvertexmap_t cvertexmap;

  /**
   * \brief property map containing the conflict graph vertex descriptors for
   * each vertex.
   */
  VertexVertMap vertexvert_map;
  
  /**
   * \brief property map denoting if any vertex is part of a border in the hull
   */
  VertexBorderMap vertexborder_map;
  
  /**
   * \brief property map containing the faces corresponding to each edge
   */
  EdgeFaceMap edgeface_map;

  /**
   * \brief property map containing the lost faces
   */
  EdgeLostMap edgelost_map;

  /**
   * \brief property map denoting if any edge is part of a border in the hull
   */
  EdgeBorderMap edgeborder_map;

  /**
   * \brief checks if the target of an edge is on the border
   */
    
public:

  /**
   * \brief standard constructor
   */
  conntab();

  /**
   * \brief retrieves the border face of the border vertex \p cvertex
   *
   * The border face is the face on the left side of a border vertex (along the
   * border edge leaving this vertex)
   *
   * If the vertex is not a border vertex the function throws an exception,
   * otherwise the face descriptor of the border vertex is returned
   *
   * \param cvertex vertex on a borderline of the connection table
   *
   * \exception no_border_vertex thrown
   * \exception unknown_vertex thrown if \p cvertex is not known by 
   *            the connection table.
   * \exception inconsistent_structure thrown if the border line is not
   *            closed or it is broken.
   */
  cVertex getBorderFace(const cVertex& cvertex);

  /**
   * \brief sets all vertex border flags to false.
   *
   * If any edges are added, the vertex border flags are not checked, because it
   * could cost to much computer time. This method resets the vertex border flags
   * to false if no border edge could be found
   *
   * \exception inconsistent_structure thrown, if there are any border edges left.
   */
  void resetBorderFlags();
  
  /**
   * \brief adds a vertex to the connection table
   *
   * A vertex is created for the conflict graph vertex \p cvertex in the
   * connection table. If there is allready a vertex for \p cvertex no new vertex
   * is added.
   *
   * The return value of the function is \c true if a vertex was added to the graph
   * \c false otherwise.
   *
   * \param cvertex vertex in the conflict graph laying on a border line
   */
  bool addVertex(const cVertex& cvertex);
  
  /**
   * \brief adds an edge from vertex to the connection table\p source to \p target
   * with a reference to the face \p face.
   *
   * The edge added correspondes to a side of the face \p face. If there is
   * allready an edge from \p source to \p target no edge is added.
   *
   * The return value of the function is \c true if a vertex was added to the graph
   * \c false otherwise.
   *
   * \param source vertex at the beginning of the edge
   * \param target vertex at the end of the edge
   * \param face pointer to a face in the conflict graph
   *
   * \exception unknown_vertex thrown if \p source or \p target is not known by 
   *            the connection table (if \p source is not known \p target is not
   *            checked).
   */
  bool addEdge(const cVertex& source, const cVertex& target, const cVertex& face);


  /**
   * \brief removes the vertex related to \p cvertex from the connection table
   * 
   * \param cvertex vertex in the conflict graph
   *
   * \exception unknown_vertex thrown if \p cvertex is not known by 
   *            the connection table.
   */
  void removeVertex(const cVertex& cvertex);
  
  /**
   * \brief removes an edge between \p source and \p target from the connection 
   * table.
   *
   * If an edge between \p source and \p target exists, the edge is removed, and
   * its face is added as a loose face to the counter edge from \p target to
   * \p source. If there is no counter edge, any loose face entry is removed from
   * the loose face map.
   * 
   * The vertices at both ends of the edge are marked as border vertices.
   * 
   * If an edge was removed the method returns \c true, \c false otherwise.
   *
   * \param source vertex at the beginning of the edge
   * \param target vertex at the end of the edge
   *
   * \exception unknown_vertex thrown if \p source or \p target is not known by 
   *            the connection table (if \p source is not known \p target is not
   *            checked).
   */
  bool removeEdge(const cVertex& source, const cVertex& target);

  /**
   * \brief removes all vertices which do not have any connections to other 
   * vertices in the connection table
   *
   * The method returns the number of removed vertices
   */
  int removeLonely();

  /**
   * \brief walks along a borderline in the connection table
   *
   * If there is a border line in the connection table this method fills a list
   * by the given list inserter \p ins with conflict graph vertices. The starting
   * vertex is denoted by \p startpoint which points to a valid conflict graph 
   * vertex descriptor.
   *
   * The direction of the walk is controlled by the given insert iterator:
   * \li a \c back_inserter generates a clockwise walks
   * \li a \c front_nserter generates a counterclockwise walk
   *
   * The method returns the number of vertices on the walk.
   *
   * \param ins insert iterator of a list of conflict graph vertices
   * \param startpoint pointer to a valid conflict graph vertex which will serve
   *        as a starting point
   *
   * \exception unknown_vertex thrown if \p startpoint is not known by 
   *            the connection table.
   * \exception inconsistent_structure thrown if the border line is not
   *            closed or it is broken.
   */
  template<class Out> int walkonBorder(Out ins, cVertex* startpoint = 0);
  /**
   * \brief retrieves all the faces along the border line
   */
  template<class Out> int getBorderFaces(Out ins);

  template<class Out> int getBorderVertices(Out ins);

};

#ifdef INCLUDE_TMPL_DEFS
  #include "conntab_tmpl.cc"
#endif

#endif
