/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include <convexhull.hh>
#include <stdlib.h>
#include <cstdio>


#include <offtrimesh.hh>

//******************************************************************************
int Convexhull::visible(const Face* face, const pointpos_t* point) const
//******************************************************************************
{
  Vector_<double> n = face->getNormal();
  pointpos_t* edgepoint = face->getPointPos(0);

  double scalar_prod=0;
  
  for(int i=0;i<3;++i)
   scalar_prod += ((*point)[i]-(*edgepoint)[i]) * n[i];
  
  //if scalar product is larger than zero the face is visible from point
  if (scalar_prod>1e-6) 
    return 1;
  else if ((scalar_prod<1e-6) && (scalar_prod>-1e-6))
    return 0;
  else
    return -1;
}

//******************************************************************************
void Convexhull::initializeGraph()
//******************************************************************************
{
  kind_map = get(vertex_kind, conflictgraph);
  pointptr_map = get(vertex_pointptr, conflictgraph);
  faceptr_map = get(vertex_faceptr, conflictgraph);
  meshvert_map = get(vertex_meshvert, conflictgraph);
}

//******************************************************************************
Convexhull::Vertex Convexhull::addVertex(Trimesh& trimesh, 
					 const Trimesh::vertex_t mv,
					 kind_t kind, bool add_structure)
//******************************************************************************
{
    //copy the vertex position
    pointpos_t vertex_pos = trimesh.vertCoordinates(mv);
    pointpos_t* point = new pointpos_t(vertex_pos);
    
    Vertex v = add_vertex(conflictgraph);
    
    //setting the vertex properties
    put(kind_map, v, kind);
    put(pointptr_map, v, point);
    put(faceptr_map, v, (Face*) 0);
    put(meshvert_map, v, mv);

    //add the vertex to the connection state graph
    if (add_structure) ctab_hull.addVertex(v);
    
    return v;
}

//******************************************************************************
Convexhull::Vertex Convexhull::addFace(Face* face)
//******************************************************************************
{
  Vertex v = add_vertex(conflictgraph);
  
  //set the vertex properties
  put(kind_map, v, vk_face);
  put(pointptr_map, v, (pointpos_t*) 0);
  put(faceptr_map, v, face);
  
  //create the edges representing the face-vertex connections
  for(int i=0; i<face->nPoints();++i)
  {
    Vertex pv= face->getPointIx(i);
    pair<Edge, bool> result = add_edge(pv,v,conflictgraph);    
    
    Vertex successor = face->getSuccessor(pv);
    //add edge to the structure graph
    ctab_hull.addEdge(pv,successor,v);
  }

  return v;
}

//******************************************************************************
Face* Convexhull::createFace(const Vertex v1, const Vertex v2, const Vertex v3)
//******************************************************************************
{
  Face* face = new Face();
  
  face->addPoint(v1,pointptr_map[v1]);
  face->addPoint(v2,pointptr_map[v2]);
  face->addPoint(v3,pointptr_map[v3]);
  
  return face;
}



//******************************************************************************
Convexhull::Convexhull(Trimesh& trimesh)
//******************************************************************************
: conflictgraph(), convexhull(), resulthull()
{
  initializeGraph();
	try
	{
		generateHull(trimesh);
	}
	catch(...)
	{
		throw;
		//remove all points and faces
// 		pair<VertexIter,VertexIter> vits = vertices(conflictgraph);
// 		for(VertexIter v = vits.first;v!=vits.second;++v)
// 		{
// 			if (pointptr_map[*v]) delete pointptr_map[*v];
// 			if (faceptr_map[*v]) delete faceptr_map[*v];
// 		}
	}
}

//******************************************************************************
Convexhull::~Convexhull()
//******************************************************************************
{
  //set of all faces
  
  
  //remove all points and faces
  pair<VertexIter,VertexIter> vits = vertices(conflictgraph);
  for(VertexIter v = vits.first;v!=vits.second;++v)
  {
    if (pointptr_map[*v]) delete pointptr_map[*v];
    if (faceptr_map[*v]) delete faceptr_map[*v];
  }
}

//******************************************************************************
void Convexhull::generateHull(Trimesh& trimesh)
//******************************************************************************
{
  /**
   *  INITIALIZATION
   */
  
  //temp vector for the faces, before the vertex list is constructed
  vector<Face*> face_tmp;
  
  /* 1. get all the vertices from the trimesh */
  vector<Trimesh::vertex_t> mesh_vertices(trimesh.begin(),trimesh.end());
  typedef vector<Trimesh::vertex_t>::iterator meshvert_it;

  /* 2. find the first four points */
 
  //select a face in the trimesh arbitrarily

  int faceix = (int)((double)trimesh.nFaces() *  rand()/(RAND_MAX+1.0));
  
  //get the first 3 points
  Triangle t = trimesh.getFace(faceix);
  
  //move the 3 triangle points in the conflict graph
  Vertex firstv[3];
  for(int i=0;i<3;++i)
  {
    meshvert_it vi = find(mesh_vertices.begin(),mesh_vertices.end(),
			  t.getVertexIndex(i));
    //add the vertex
    firstv[i] = addVertex(trimesh,*vi,vk_hullvertex);
    //add it to the actual convex hull
    //remove the vertex from the mesh_vertices list
    mesh_vertices.erase(vi);
  }
  
  //add the first face to the temp facelist
  face_tmp.push_back(createFace(firstv[0],firstv[1],firstv[2]));

  //select a forth point, so that a tetraeder is formed
  while (num_vertices(conflictgraph)<4)
  {
    int vi = (Trimesh::vertex_t)((double)mesh_vertices.size() * rand()/(RAND_MAX+1.0));

    pointpos_t vertex_pos = trimesh.vertCoordinates(mesh_vertices[vi]);
    pointpos_t* point = new pointpos_t(vertex_pos);

    if (visible(face_tmp.front(),point)<0)
    {
      //add the new vertex
      Vertex v = addVertex(trimesh,mesh_vertices[vi], vk_hullvertex);
      //add it to the actual convex hull

      //create the different three faces for the starting tetraheder
      face_tmp.push_back(createFace(firstv[1],firstv[0],v));
      face_tmp.push_back(createFace(firstv[2],firstv[1],v));
      face_tmp.push_back(createFace(firstv[0],firstv[2],v));
    }
  }
  
  /* 3. add all the other vertices to the conflict graph in a random order */
  Vertex new_v;

  while (mesh_vertices.size()>0)
  {
    //find the next vertex randomly
    int vi = (Trimesh::vertex_t)(((double) mesh_vertices.size()) * rand()/(RAND_MAX+1.0));
    meshvert_it vit = mesh_vertices.begin()+vi;
   
    new_v = addVertex(trimesh, *vit, vk_vertex, false);
    
    //remove Vertex from the mesh_vertices list
    mesh_vertices.erase(vit);
  }
  //mark the first face index and the next_point index
  first_face = num_vertices(conflictgraph);
  next_point = 4;

  //put the new faces in the conflict graph
  //add the faces from face_tmp
  for(vector<Face*>::iterator f=face_tmp.begin();f!=face_tmp.end();++f)
    addFace(*f);
  
  /* 4. initialize the edges in the conflict graph with the actual visibility */
  for(int vix = next_point;vix<first_face;++vix)
  {
    Vertex v = vertex(vix,conflictgraph);
    for(int wix = first_face; wix<num_vertices(conflictgraph);++wix)
    {
      Vertex w = vertex(wix,conflictgraph);

      if (visible(faceptr_map[w],pointptr_map[v])>0)
      {
	add_edge(v,w,conflictgraph);
	add_edge(w,v,conflictgraph);
      }
    }
  }

  /**
   *  ALGORITHM: ITERATION OVER ALL POINTS (5-nPoints);
   */
  Vertex next_vertex=vertex(next_point,conflictgraph);
  
  while(kind_map[next_vertex]==vk_vertex)
  {
    //if the next point has no conflicts with any face, it is inside the convex hull
    if (out_degree(next_vertex,conflictgraph)>0)
    {
      kind_map[next_vertex] = vk_hullvertex;

      //points who could be visible by new faces:
      //   points visible from faces adjacent to the horizon
      set<Vertex> vis_checklist;

      //type definition for an adjacency itertor in the conflict graph
      typedef graph_traits<conflictgraph_t>::adjacency_iterator adjacency_it;
      
      /* 1. determine the conflict region and update the connection graph */
      
      //make a list of all conflict faces (because adjacency iters become invalid
      //if edges are cleared)
      pair<adjacency_it,adjacency_it> adj = adjacent_vertices(next_vertex,
							      conflictgraph);
      //list of all conflict faces
      list<Vertex> conflict_faces;

      for(adjacency_it a=adj.first;a!=adj.second;++a)
      {
	Face* act_face = faceptr_map[*a];

	//remove all edges from the connection graph
	for(int i=0;i<act_face->nPoints();++i)
	{
	  Vertex v1 = act_face->getPointIx(i);
	  Vertex v2 = act_face->getSuccessor(v1);
	  
	  ctab_hull.removeEdge(v1,v2);
	}
	
	conflict_faces.push_back(*a);
      }

      //generate the visibility checklist
      list<Vertex> face_interest;
      ctab_hull.getBorderFaces(back_inserter(face_interest));

#ifdef LT_DEBUGGING
      dumpVertices(face_interest.begin(),face_interest.end());
#endif
      
      for(list<Vertex>::iterator f=face_interest.begin();f!=face_interest.end();++f)
      {
	pair<adjacency_it,adjacency_it> adj = adjacent_vertices(*f,conflictgraph);
	copy(adj.first, adj.second, inserter(vis_checklist,vis_checklist.begin()));
      }
       
      //remove the visible faces
      for(list<Vertex>::iterator a=conflict_faces.begin();a!=conflict_faces.end();)
      {
	Face* conflict_face = faceptr_map[*a];

	//delete the conflict_face vertex
	clear_vertex(*a,conflictgraph);
	list<Vertex>::iterator remove_vertexit = a++;
	remove_vertex(*remove_vertexit,conflictgraph);

	//delete the face pointer
	delete conflict_face;
      }

      //remove all the conflict edges from next_vertex
      clear_vertex(next_vertex,conflictgraph);
      
      //remove all the unconnected vertices from the hull connection graph
      ctab_hull.removeLonely();

      //add the vertex to the hull connection table
      ctab_hull.addVertex(next_vertex);

      /* 2. retrieve the horizon of the new point */
      
      //type of an horizon element
      typedef list<Vertex>::iterator horizon_it;
      list<Vertex> horizon;      //ordered horizon
      
      //fill the horizon
      ctab_hull.walkonBorder(back_inserter(horizon));

      /* 3. create the new faces for point next_vertex */
      
      for(horizon_it hv=horizon.begin();hv!=horizon.end();0)
      {
	//get extra iterators for the first and second point
	Vertex first_point = *hv;
	Vertex second_point = *(++hv);

	Vertex face_vertex = ctab_hull.getBorderFace(first_point);
	Face* border_face = faceptr_map[face_vertex];
	
	if (hv==horizon.end()) 
	  second_point = *horizon.begin();
	
	if (visible(border_face,pointptr_map[next_vertex])!=0)
	{
	  //create the new face and the new face vertex
	  Face* newface = createFace(next_vertex,
				     second_point,
				     first_point);
	  Vertex newface_v = addFace(newface);
	  
	  //check the visibility
	  for(set<Vertex>::iterator v=vis_checklist.begin();v!=vis_checklist.end();++v)
	    if (visible(newface,pointptr_map[*v])>0)
	    {
	      add_edge(*v,newface_v,conflictgraph);
	      add_edge(newface_v,*v,conflictgraph);
	    }
	}
	else
	{
	  border_face->insertPoint(second_point,next_vertex,
				   pointptr_map[next_vertex]);
	  //add the edges in the conflict graph
	  add_edge(next_vertex,face_vertex,conflictgraph);	
	  
	  //add edges in the ctab_hull
	  ctab_hull.addEdge(first_point,next_vertex,face_vertex);
	  ctab_hull.addEdge(next_vertex,second_point,face_vertex);

	  //remove edge between the first and the second point
	  ctab_hull.removeEdge(first_point,second_point);
	}

      }
    }

    //reset the vertex border flags in the connection graph
		ctab_hull.resetBorderFlags();

    ++next_point;
    next_vertex = vertex(next_point,conflictgraph);
    
#ifdef LT_DEBUGGING

    Trimesh* dbgMesh = getMesh();
    offTrimesh offmesh(dbgMesh);
    char filename[200];
    sprintf(filename,"debug/debug_holl%.3d.off",next_point-4);
    offmesh.write(filename);
    delete dbgMesh;

    list<Vertex> bordervertices;
    
    if (ctab_hull.getBorderVertices(back_inserter(bordervertices)))
    {
      cout << "still border vertices : ";

      for(list<Vertex>::iterator i=bordervertices.begin();i!=bordervertices.end();++i)
	cout << meshvert_map[*i] << " ";
      cout << endl;
    }
    
#endif

    
  }

  pair<VertexIter,VertexIter> vits = vertices(conflictgraph);
  //write the resulthull vector
  for(VertexIter v=vits.first;v!=vits.second;++v)
    if (kind_map[*v]==vk_hullvertex)
    {
      resulthull.push_back(meshvert_map[*v]);
      convexhull.push_back(*v);
    }

}

//******************************************************************************
Trimesh* Convexhull::getMesh() const
//******************************************************************************
{
//   Trimesh* newMesh = new Trimesh();

//   //mapping of conflict graph vertices an the new vertex indices
//   map<Vertex,Trimesh::vertex_t> vertex_indices;
  
//   //add all vertices
//   int c=0;
//   for(int i=0;i<first_face;++i)
//   {
//     Vertex v = vertex(i,conflictgraph);
//     if ((kind_map[v]==vk_hullvertex) && (out_degree(v,conflictgraph)>0))
//     {
    
//       Trimesh::vertex_t v_mesh = add_vertex(newMesh->Mesh);
    
//       put(newMesh->coordinates_map, v_mesh, *(pointptr_map[v]));
//       put(newMesh->ownindices_map, v_mesh, c++);
//       put(newMesh->vertexkinds_map, v_mesh, 0);
//       vertex_indices[v] = v_mesh;
//     }
//   }
  
//   //add all faces
//   for(int i=first_face;i<num_vertices(conflictgraph);++i)
//   {
//     Vertex v = vertex(i,conflictgraph);
//     Face* face = faceptr_map[v];
    
//     //retrieve the vertices
//     Trimesh::vertex_t v1 = vertex_indices[face->getPointIx(0)];
//     Trimesh::vertex_t v2 = vertex_indices[face->getPointIx(1)];
//     Trimesh::vertex_t v3 = vertex_indices[face->getPointIx(2)];
    
//     //get coordinates of the triangles vertices
//     Vector_<double> c1(3);
//     Vector_<double> c2(3);
//     Vector_<double> c3(3);

// 		//create the triangle object and add it to the trianglelist
// 		Triangle atriangle(edgepoint(v1,&((*coordinates_map)[v1])),
																 

//     newMesh->Triangles.push_back(atriangle);
//     //adding the index of the face to the vertfaceIndex entries of all three
//     //vertices
//     newMesh->vertfaceIndex[v1].push_back(i-first_face);
//     newMesh->vertfaceIndex[v2].push_back(i-first_face);
//     newMesh->vertfaceIndex[v3].push_back(i-first_face);


//     //check if any edges are allready there
//     bool has_edge1 = edge(v1,v2, newMesh->Mesh).second;
//     bool has_edge2 = edge(v2,v3, newMesh->Mesh).second;
//     bool has_edge3 = edge(v3,v1, newMesh->Mesh).second;

//     //add every edge, that does not exist yet, and calculate its length
//     pair<Trimesh::edge_t, bool> e_b;   //edge, boolean pair (returned from add_edge

// 		if (!has_edge1)
// 		{
// 			e_b = add_edge(v1,v2, Mesh);
// 			put(*edgelengths_map, e_b.first, calculate_length((*coordinates_map)[v1],
// 																												(*coordinates_map)[v2]));

// 		}
// 		if (!has_edge2)
// 		{
// 			e_b = add_edge(v2,v3, Mesh);
// 			put(*edgelengths_map, e_b.first, calculate_length((*coordinates_map)[v2],
// 																												(*coordinates_map)[v3]));
// 		}
// 		if (!has_edge3)
// 		{
// 			e_b = add_edge(v3,v1, Mesh);
// 			put(*edgelengths_map, e_b.first, calculate_length((*coordinates_map)[v3],
// 																												(*coordinates_map)[v1]));
// 		}
//   }

//   newMesh->calcSurfaceNormals();

//   return newMesh;

	return (Trimesh*) 0;
}

#ifdef LT_DEBUGGING

//******************************************************************************
void Convexhull::dumpConflictgraph()
//******************************************************************************
{
  typedef graph_traits<conflictgraph_t>::out_edge_iterator outedges_it;

  cout << "####################################################################" << endl;
  cout << "Step " << next_point-4 <<  endl;
  cout << "####################################################################" << endl;
  cout << "next_point : " << next_point << endl;
  cout << "first_face : " << first_face << endl << endl;

  map<Vertex,int> vertex_index;

  int i=0;
  pair<VertexIter,VertexIter> vits = vertices(conflictgraph);
  for(VertexIter v=vits.first;v!=vits.second;++v)
    vertex_index[*v]=i++;

  for(VertexIter v=vits.first;v!=vits.second;++v)
  {
    switch (kind_map[*v])
    {
      case vk_hullvertex :
      {
	cout << "HV " << vertex_index[*v] << "(" << meshvert_map[*v] << ")" << endl;
	break;
      }

      case vk_vertex :
      {
	cout << "V  " << vertex_index[*v] << "(" << meshvert_map[*v] << ")" << endl;
	break;
      }

      case vk_face :
      {
	cout << "F  " << vertex_index[*v] << endl;
	break;
      }
    }

    cout << "\t";
    pair<outedges_it,outedges_it> adj = out_edges(*v,conflictgraph);
    int j=0;
    for(outedges_it e=adj.first;e!=adj.second;++e)
    {
      cout << vertex_index[target(*e,conflictgraph)] << " ";
      if ((++j % 15)==0) cout <<endl << "\t";
    }
    cout << endl;
  }
  
  cout << endl; endl; 
}

//******************************************************************************
void Convexhull::dumpVertices(list<Vertex>::iterator beg, 
			      list<Vertex>::iterator end)
//******************************************************************************
{
  map<Vertex,int> vertex_index;

  int i=0;
  pair<VertexIter,VertexIter> vits = vertices(conflictgraph);
  for(VertexIter v=vits.first;v!=vits.second;++v)
    vertex_index[*v]=i++;
  
  for(list<Vertex>::iterator s=beg;s!=end;++s)
    cout << vertex_index[*s] << " ";
  cout << endl;
}

#endif
