/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef CONVEXHULL_HH
#define CONVEXHULL_HH

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map.hpp>

#include "face.hh"
#include "conntab.hh"
#include "trimesh.hh"
#include "triangle.hh"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <list>
#    include <set>
#    include <algorithm>
#  else
#    include <list.h>
#    include <set.h>
#    include <algo.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <list>
#  include <set>
#  include <algorithm>
#endif //HAVE_CONFIG_H

namespace boost
{
  enum vertex_kind_t     { vertex_kind     = 400 };
  enum vertex_pointptr_t { vertex_pointptr = 401 };
  enum vertex_faceptr_t  { vertex_faceptr  = 402 };
  enum vertex_meshvert_t { vertex_meshvert = 403 };

  BOOST_INSTALL_PROPERTY(vertex, kind);
  BOOST_INSTALL_PROPERTY(vertex, pointptr);
  BOOST_INSTALL_PROPERTY(vertex, faceptr);
  BOOST_INSTALL_PROPERTY(vertex, meshvert);
};

class conflictgraph_traits
{
protected:

  enum kind_t {vk_hullvertex, vk_vertex, vk_face};
  
  /**
   * \brief specifies the kind of the vertex
   *
   * look at convexhull::kind_t for description of the vertex kinds
   */
  typedef property<vertex_kind_t, kind_t > VertexKind;
  
  /**
   * \brief pointer to a surface point
   *
   * this property contains the surface point, if the vertex is representing a
   * point. If the vertex is a face, this property will be NULL.
   */
  typedef property<vertex_pointptr_t, pointpos_t*, VertexKind > VertexPointPtr;
  
  /**
   * \brief pointer to a face in the convex hull
   *
   * this property contains a pointer to a face on the convex hull, if the vertex
   * is representing a face. Otherwise the pointer will be NULL.
   */
  typedef property<vertex_faceptr_t, Face*, VertexPointPtr > VertexFacePtr;
  
  /**
   * \brief vertex in the original mesh
   * 
   * used to connect the vertices in the convex hull to the original vertices
   */
  typedef property<vertex_meshvert_t, Trimesh::vertex_t, VertexFacePtr > VertexMeshVert;

  /**
   * \brief typdefinition for the conflict graph.
   *
   * The conflict graph is an essential part of the convex-hull algorithm. for more detail
   * see the description of the algorithm.
   */
  typedef adjacency_list<listS, listS, directedS, VertexMeshVert> conflictgraph_t;

  /**
   * \brief property map for the vertex kind
   */
  typedef property_map<conflictgraph_t, vertex_kind_t>::type VertexKindMap;

  /**
   * \brief property map for the vertex pointptrs
   */
  typedef property_map<conflictgraph_t, vertex_pointptr_t>::type VertexPointPtrMap;

  /**
   * \brief property map for the vertex faceptrs
   */
  typedef property_map<conflictgraph_t, vertex_faceptr_t>::type VertexFacePtrMap;

  /**
   * \brief property map for the vertex faceptrs
   */
  typedef property_map<conflictgraph_t, vertex_meshvert_t>::type VertexMeshVertMap;
  
  void initializeGraph();

public:
  
  /**
   * \brief vertex descriptor for the conflict graph
   */
  typedef conflictgraph_t::vertex_descriptor Vertex;

  /**
   * \brief edge descriptor for the conflict graph
   */
  typedef conflictgraph_t::edge_descriptor Edge;

  /**
   * \brief vertex iterator
   */
  typedef conflictgraph_t::vertex_iterator VertexIter;

};


class Convexhull : public conflictgraph_traits
{
private:
  /**
   * \brief referrence to the list of vertex kinds
   */
  VertexKindMap kind_map;
  
  /**
   * \brief referrence to the list of points
   */ 
  VertexPointPtrMap pointptr_map;
  
  /**
   * \brief referrence to the list of faces
   */
  VertexFacePtrMap faceptr_map;
  
  /**
   * \brief referrence to the list of original mesh vertices
   */
  VertexMeshVertMap meshvert_map;

  /**
   * \brief convlict graph
   */
  conflictgraph_t conflictgraph;

  /**
   * \brief graph holding the connection state of all the vertices in the hull
   */
  conntab ctab_hull;
  
  /**
   * \brief points in the convex hull
   */
  list<Vertex> convexhull;
  typedef list<Vertex>::iterator convexhull_it;

  /**
   * \brief vertices of the original mesh in the convex hull
   */
  list<Trimesh::vertex_t> resulthull;

  /**
   * \brief first vertex pointing to yet not processed point
   */
  int next_point;
  
  /**
   * \brief first vertex pointing to a face
   */
  int first_face;
  
  /**
   * \brief gives information about the visibility of the face \p face from the
   * point \p point
   *
   * The visibility is checked by the scalar product of the normal vector of the
   * face and the vector from the first edge point to the point. If the product
   * is larger than zero, the face is visible from the point, else the face is not
   * visible.
   *
   * \param face pointer to a face on the convex hull
   * \param point pointer to a point of the surface
   */
  int visible(const Face* face, const pointpos_t* point) const;

  /**
   * \brief initialize the property maps
   */
  void initializeGraph();

  /**
   * \brief creates a conflict graph vertex out of a mesh vertex
   */
  Vertex addVertex(Trimesh& trimesh, const Trimesh::vertex_t mv, kind_t kind,
		   bool add_structure=true);
  
  /**
   * \brief creates a conflict graph vertex out of a face
   */
  Vertex addFace(Face* face);

  /**
   * \brief creates a face based on three vertices
   */
  Face* createFace(const Vertex v1, const Vertex v2, const Vertex v3);

public:

  typedef list<Trimesh::vertex_t>::iterator hullvertex_it;
  typedef list<Trimesh::vertex_t>::const_iterator hullvertex_cit;

  /**
   * \brief standard constructor
   */
  Convexhull(): conflictgraph(), convexhull(), resulthull() { initializeGraph(); };

  /**
   * \brief computes the convex hull of the triangulated mesh \p trimesh
   *
   * This constructor should be used to create a Convexhull object, because it performs
   * all the computation at once and the result can be read 
   * 
   * \note for more details please look at Convexhull::generateHull.
   *
   * \param trimesh triangulated mesh for which the convex hull should be calculated.
   */
  Convexhull(Trimesh& trimesh);
  
  /**
   * \brief standard destructor
   */
  ~Convexhull();

  /**
   * \brief reinitilizes the hull and generates a new convex hull out of \p trimesh
   *
   * A new convex hull is calculated using the triangulated mesh \p trimesh. The 
   * mesh's own indices have to be set to the vertex indices, to avoid comput.
   * effort.
   *
   * \param trimesh triangulated mesh for which the convex hull should be calculated.
   */
  void generateHull(Trimesh& trimesh);

  /**
   * \brief retrieving the first element of the convex hull
   */
  hullvertex_it begin() { return resulthull.begin(); }

  /**
   * \brief retrieving the first element of the convex hull (constant)
   */
  hullvertex_cit const_begin() const { return resulthull.begin(); }

  /**
   * \brief retrieving one after the last element of the convex hull
   */
  hullvertex_it end() { return resulthull.end(); }

  /**
   * \brief retrieving one after the last element of the convex hull (constant)
   */
  hullvertex_cit const_end() const { return resulthull.end(); }

  /**
   * \brief generates a surface mesh of the convex hull
   */
  Trimesh* getMesh() const;

#ifdef LT_DEBUGGING

  void dumpConflictgraph();

  void dumpVertices(list<Vertex>::iterator beg,list<Vertex>::iterator end);

#endif

};

#endif
