/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef DISTANCE_MAP_HH
#define DISTANCE_MAP_HH

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map.hpp>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <set>
#    include <vector>
#    include <algorithm>
#  else
#    include <map.h>
#    include <set.h>
#    include <vector.h>
#    include <algo.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <map>
#  include <set>
#  include <vector>
#  include <algorithm>
#endif //HAVE_CONFIG_H

#ifdef HAVE_NAMESPACES
using namespace boost;
#endif

/**
 * \brief representation of a distance map organized in layers around vertices
 *
 * Despite the usual matrix based representation of distance maps this 
 * implementation provides a layer based access to the edge distances from an
 * arbitrary vertex \c v to other vertex in the graph. It provides a convenient
 * and fast access to the \em {n- distant} layers around the vertex.
 *
 * 
 * 
 * This form of the distance map is faster to compute, if you are interested only
 * in a restricted part of the vicinity of the vertices. It is faster to access,
 * if you are interested in whole n-distant layers.
 * 
 * The internal representation is a distance table, a datastructure which holds 
 * the vertices in a certain distance of a central vertex \em v (a layer) in a
 * seperate container leading to the following table:
 *
 * <table>
 * <tr><td><b> distance</b></td>
 *     <td><b>vertices in the layer</b></td></tr>
 * <tr><td>1</td>
 *   <td>v<sub>1,1</sub>, v<sub>1,2</sub>, v<sub>1,3</sub>, ... v<sub>1,n</sub></td></tr>
 * <tr><td>2</td>
 *   <td>v<sub>2,1</sub>, v<sub>2,2</sub>, v<sub>2,3</sub>, ... v<sub>2,n</sub></td></tr>
 * <tr><td>...</td>
 *   <td>...</td></tr>
 * <tr><td>n</td>
 *   <td>v<sub>n,1</sub>, v<sub>n,2</sub>, v<sub>n,3</sub>, ... v<sub>n,n</sub></td></tr>
 * </table>
 *
 */
template<class graph_t> class distancemap_t
{
private:

	//*****************************************************************************
	// type definitions for the graph_t class
	//*****************************************************************************

  /**
	 * \brief type identifier for the graph's vertex type
	 */
  typedef graph_traits<graph_t>::vertex_descriptor vertex_t;

	/**
	 * \brief type identifier for the graph's vertex iterator type
	 */
  typedef graph_traits<graph_t>::vertex_iterator vertex_it;

	/**
	 * \brief type identifier for the graph's vertex size type
	 */
  typedef graph_traits<graph_t>::vertices_size_type vsize_t;

	/**
	 * \brief type identifier for the graph's adjacency iterator type
	 */
	typedef graph_traits<graph_t>::adjacency_iterator adj_it;

  /**
	 * \brief type definition for an arbitrary layer around a single vertex
	 */
	
public:

	//*****************************************************************************
	// PUBLIC Type Definitions
	//*****************************************************************************

	/**
	 * \brief type definition of a single n-distant layer
	 */
  typedef set<vertex_t> layer_t;

	/**
	 * \brief iterator of a single n-distant layer
	 */
	typedef layer_t::iterator layer_it;

	/**
	 * \brief constant iterator of a single n-distant layer
	 */
	typedef layer_t::const_iterator layer_cit;
	
	/**
	 * \brief type definition of a single vertex's distance table
	 */
	typedef vector<layer_t> distancetab_t;

  /**
	 * \brief iterator of a single vertex's distance table
	 */
	typedef distancetab_t::iterator distancetab_it;

	/**
	 * \brief constant iterator of a single vertex distance table
	 */
	typedef distancetab_t::const_iterator distancetab_cit;

	//*****************************************************************************
	// EXCEPTIONS
	//*****************************************************************************
	
	/**
	 * \brief this exception is thrown by the calcVertex method.
	 *
	 * If an uninitialized distance table is passed to the method, this exception
	 * is thrown.
	 */
	class distancetable_uninitialized { };

private:
	
	//*****************************************************************************
	// PRIVATE Type definitions depending on PUBLIC TYPES
	//*****************************************************************************

	/**
	 * \brief vector type of the graph's distance table
	 */
	typedef map<vertex_t,distancetab_t> dmap_t;
	
	/**
	 * \brief iterator of the graph's distance tab
	 */
	typedef dmap_t::iterator dmap_it;
	
	/**
	 * \brief constant iterator of the graph's distance tab
	 */
	typedef dmap_t::const_iterator dmap_cit;

	//*****************************************************************************
	// PRIVATE MEMBERS
	//*****************************************************************************

	/**
	 * \brief pointer to a graph for the distance map calculation.
	 *
	 * The graph (which must be of the type graph_t) must fullfill the following 
	 * concept:
	 * 
	 * <li> It must be a graph of the BOOST Graph Library.
	 * <li> \c{num_vertices(graph)} must be able to return the number of vertices
	 * <li> There must be an \c{vertex_index} property map
	 * <li> The function \c{vertex(index, graph)} must return the vertex associated
	 *      with the vertexindex \c index
	 * <li> \c{adjacent_vertices(vertex, graph)} must return a iterator pair for the
	 *      adjacent vertices around \c vertex.
	 */
	const graph_t* graph;


	/**
	 * \brief representation of the graph's distance table
	 *
	 * This is the internal representation of the graphs distance table. It can be
	 * directely used for the layer based access methods, but must be converted for
	 * the matrix based methods, which is a little bit time consuming, and should 
	 * therefore be done only in single cases.
	 */
	dmap_t dmap;
	
	/**
	 * \brief number of layers included in the distance map
	 *
	 * \c n_layers represents the maximal number of layers calculated for each
	 * vertex.
	 * \note The number of effectively stored layers can be smaller.
	 */
	int n_layers;
	
	/**
	 * \brief number of vertices in the graph
	 *
	 * The number of vertices is calculated by a call \c{num_vertices(graph)}.
	 */
	vsize_t n_vertices;

	/**
	 * \brief dummy layer, used to signal an empty sequence
	 */
	layer_t empty_layer;
	
	/**
	 * \brief represents the width of a layer bin
	 *
	 * The bin width is used to build layers from a on-surface distance matrix in
	 * distancemap_t::initVertex.
	 */
	double bin_width;

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief calculates the distance table of a single vertex of the graph
	 *
	 * The distance table of a single vertex is calculated (up to \c n_layers) and
	 * written into the correct \c dmap entry. The calculation is started on the 
	 * layer \p start_layer, if \p start_layer - 1 layers are allready present.
	 * 
	 * The method returns the number of layers, around the vertex.
	 *
	 * \param v identifer of the calculated vertex 
	 * \param dtable receives the distance information
	 * \param start_layer starting layer of calculation. The start_layer and all 
	 *        layers above are calculated.
	 */
	int calcVertex(const vertex_t v, distancetab_t& dtable,
								 const int start_layer = 0);

public:

	//*****************************************************************************
	// CONSTRUCTORS & INITIALIZATION
	//*****************************************************************************

	/**
	 * \brief standard constructor creating an empty instance of a distancemap.
	 *
	 * instances created with the standard constructor could not be used, unless
	 * they receive a reference to a graph.
	 */
	distancemap_t(): graph(0), dmap(), n_layers(0), n_vertices(0), empty_layer(),
									 bin_width(1) { };
	
	/**
	 * \brief constructor with full initialization.
	 *
	 * \param agraph pointer to the graph structure
	 * \param layers number of layers, which should be considered in the distance
	 *        calculation.
	 */
	distancemap_t(const graph_t* agraph, const int layers, const double binwidth = 1);

	/**
	 * \brief clears and resets the distance tables of all vertices
	 */
	void clear() { dmap.clear(); }

	/**
	 * \brief invokes the calculation of the complete distance map
	 *
	 * If you need the complete map, this is a convenient way to calculate the 
	 * whole thing at once. Allready calculated distances are preserved.
	 *
	 * If you specify a \p layer, the calculation is done for \p layers, the
	 * distancemap_t::n_layers parameter is updated, and all existing distance
	 * tables are adjusted.
	 *
	 * The method returns the number of processed vertices. A number lower than 
	 * n_vertices is not necessarely related to a problem
	 *
	 * \param layers number of layers, which should be considered in the distance
	 *        calculation.
	 */
	int calcComplete(const int layers = -1);

	/**
	 * \brief initializes the distancemap of vertex \c v by a sequence of distances
	 * to every other vertex.
	 *
	 * The distance sequence must follow the concept of an Forward Iterator, and 
	 * must contain integer values. The sequence must be ordered by the indices  of
	 * the vertices in the graph.
	 * 
	 * The function returns the number of layers around vertex v
	 *
	 * \param v vertex of the base graph to which the distance data is related
	 * \param begin of a distance sequence
	 * \param end of the distance sequence
	 */
	template<class In> int initVertexTab(const vertex_t v, In begin, In end);

	//*****************************************************************************
	// PROPERTY & MEMBER METHODS
	//*****************************************************************************

	/**
	 * \brief retrieves the number of layers calculated around vertex \c v
	 *
	 * If the vertex has not been calculated, the function returns 0, otherwise
	 * the number of layers in the distance table of vertex \c v are returned. The
	 * distances of these layers are 0,1,2, ..., n-1 edges respectively 
	 *
	 * \param v vertex of interest.
	 */
	int vertex_layers(const vertex_t v) const;
	
	/**
	 * \brief retrieves the number of layers, that are considered during the 
	 * calculation.
	 */
	int get_nlayers() const {return n_layers; }
	
	/**
	 * \brief sets the number of layers considered for the distance map to
	 * \c layers.
	 *
	 * If \c layers differs from the old value, all the distance maps should be
	 * recalculated, to obtain consistency. This is not done automatically, because
	 * the data is still valid, but to get all the new layers (if the new parameter
	 * is larger than the old one) you should recalculate or simply clear the 
	 * distance maps of all vertices.
	 *
	 * The method returns the old value.
	 *
	 * \param layers new number of considerable layers
	 */
	int set_nlayers(const int layers);

	/**
	 * \brief retrieves the number of vertices in the base graph
	 */
	vsize_t get_nvertices() const { return n_vertices; }

	/**
	 * \brief retrieves the bin width
	 */
	double get_binwidth() const { return bin_width; }
	
	/**
	 * \brief sets the bin width to a new value
	 *
	 * The function returns the old value for the bin width
	 *
	 * \param bin_width new value for the bin width
	 */
	double set_binwidth(const double bin_width);

	/**
	 * \brief returns the pointer to the base graph structure
	 */
	const graph_t* get_graph() const { return graph; }
	
	/**
	 * \brief resets the pointer to the base graph
	 *
	 * When you reset the pointer to the base graph, the distance data cleared
	 * automatically.
	 * 
	 * The method returns a pointer to the old base graph.
	 *
	 * \param new_graph pointer to the new base graph.
	 */
	const graph_t* set_graph(const graph_t* new_graph);

	//*****************************************************************************
	// DATA METHODS
	//*****************************************************************************

	/**
	 * \brief retrieves a pair of constant layer iterators for the vertex \c v and
	 * the layer \c l.
	 *
	 * The first constant iterator of the returned pair points to the first
	 * vertex in the \c l <sup>th</sup> layer, and the second points behind the
	 * last vertex of the same layer.
	 * 
	 * If the layer is beyond the maximal number of layers available for the vertex
	 * \c v, than an empty sequence is returned (first and second member of the 
	 * pair are equal)
	 *
	 * \param v vertex of interest
	 * \param l layer of interest
	 */
	pair<layer_cit,layer_cit> get_layer(const vertex_t v, const int l);

	/**
	 * \brief operator function retrieving the distance matrix row, as a 
	 * vector of \c{vertex_t}'s for the vertex \p v
	 *
	 * The method creates a vector of length \c n_vertices (which is actually the
	 * same as a row vector in the graphs distance matrix) out of the vertex's 
	 * distance table. If the distance to a distinct vertex is not 
	 *
	 * \param v vertex of interest
	 */
	vector<int> operator[](const vertex_t v) const;

	/**
	 * \brief operator function retrieving the distance between the vertices
	 * \p v and \p w.
	 *
	 * \note To find the distance from vertex \p v to vertex \p w in the worst case,
	 * all the layers in the distance table of vertex \p v have to be searched.
	 *
	 * \param v starting vertex (its distance table is used)
	 * \param w endpoint of the distance calculation
	 */
	int operator()(const vertex_t v, const vertex_t w) const;
	
};

#ifdef INCLUDE_TMPL_DEFS
  #include "distance_map_templ.cc"
#endif


#endif //DISTANCE_MAP_HH
