/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef DISTANCE_MAP_TEMPL_CC
#define DISTANCE_MAP_TEMPL_CC

#include <float.h>

//*******************************************************************************
template<class graph_t>
int distancemap_t<graph_t>::calcVertex(const vertex_t v, distancetab_t& dtable,
															const int start_layer)
//*******************************************************************************
{
	//check if the number of entries in dtable is equal to the n_layers
	if (dtable.size()!= n_layers) throw distancetable_uninitialized();
	
	//color map for the visitors
	map<vertex_t,bool> visited;

  //initialize the first layer with the vertex itself
	dtable[0].insert(v);
	
	
	/* INITIALIZE THE COLOR MAP AND CHECK THE start_layer VALIDITY */

	int l = 0; //layer counter
	
	for(distancetab_cit layer=dtable.begin();layer!=dtable.end();++layer)
	{
		//break the loop because all filled layers are reached
		if (layer->size()==0) break;

		for(layer_cit entry=layer->begin();entry!=layer->end();++entry)
			visited[*entry] = true;
		
		++l;

		//break because start_layer is reached
		if (l>start_layer) break;
	}

	/* GENERATE ONE LAYER ABOVE THE OTHER */

	int c=0; //count of calculated layers

	for(int i=l;i<n_layers;++i)
	{
		for(layer_it r=dtable[i-1].begin();r!=dtable[i-1].end();++r)
		{
			//check all adjacent vertices of the previous layer, if they are in the next layer
			pair<adj_it,adj_it> adjacency = adjacent_vertices(*r,*graph);
			for(adj_it a=adjacency.first;a!=adjacency.second;++a)
				if (!visited[*a]) 
				{
					dtable[i].insert(*a);
					visited[*a] = true;
				}
			
			//check if new layer is empty
			if (dtable[i].size()>0) ++c; else break;
 		}
	}
	
	return c;
}

//*******************************************************************************
template<class graph_t>
distancemap_t<graph_t>::distancemap_t(const graph_t* agraph, const int layers,
																			const double binwidth)
//*******************************************************************************
: graph(agraph), n_layers(layers), n_vertices(0), dmap(), empty_layer(), 
	bin_width(binwidth)
{
	n_vertices = num_vertices(*graph);
}

//*******************************************************************************
template<class graph_t>
int distancemap_t<graph_t>::calcComplete(const int layers)
//*******************************************************************************
{
	int c=0; //counter of calculated vertices
	
	//set n_layers to layers if layers > -1
	if (layers>-1) n_layers = layers; 

	//iterate over all vertices to recalculate the distance tables
	pair<vertex_it, vertex_it> vits = vertices(*graph);
  for(vertex_it v=vits.first;v!=vits.second;++v)
	{
		int start_layer = dmap[*v].size();
		dmap[*v].resize(n_layers);
		
		//if a layer has been calculated (return value of calcVertex>0) increase 
		//the counter
		if (calcVertex(*v,dmap[*v],start_layer)) ++c;
	}
	
	return c;
}

//*******************************************************************************
template<class graph_t> template<class In>
int distancemap_t<graph_t>::initVertexTab(const vertex_t v, In begin, In end)
//*******************************************************************************
{
	int c = 0;
	
	//reinitialize the entry vor vertex v
	dmap[v].clear();

	/* ITERATE THROUGH THE SEQUENCE */

	int ix = 0; //vertex index
	while (begin!=end)
	{
		if (ix>=n_vertices) break;
		vertex_t w = vertex(ix++, *graph);  //get the next vertex in the index

		//put the vertex in the right bin
		int bin;
		if (*begin<DBL_EPSILON)
			bin = 0;
		else
			bin = (int)floor(*begin/bin_width)+1;
		
		//check if a new bin has to be opened
		if (bin>=(dmap[v].size())) 
		{
			dmap[v].resize(bin+1);
			//check if the maximum number of layers has to be increased
			if (dmap[v].size()> n_layers) n_layers = dmap[v].size();
		}
		
		//insert the vertex v into the right layer
		dmap[v][bin].insert(w);
		++c;

		++begin;
	}
	
	return c;
}

//*******************************************************************************
template<class graph_t>
int distancemap_t<graph_t>::vertex_layers(const vertex_t v) const
//*******************************************************************************
{
	int c=0;

  for(layer_cit l=dmap[v].begin();l!=dmap[v].end();++l)
		if (l->size()>0) ++c;
	
	return c;
}

//*******************************************************************************
template<class graph_t>
int distancemap_t<graph_t>::set_nlayers(const int layers)
//*******************************************************************************
{
	int oldvalue = n_layers;
	n_layers = layers;
	
	return oldvalue;
}

//*******************************************************************************
template<class graph_t>
double distancemap_t<graph_t>::set_binwidth(const double new_binwidth)
//*******************************************************************************
{
	double oldvalue = bin_width;
	if (new_binwidth>DBL_EPSILON)
		bin_width = new_binwidth;
	
	return oldvalue;
}

//*******************************************************************************
template<class graph_t>
const graph_t* distancemap_t<graph_t>::set_graph(const graph_t* new_graph)
//*******************************************************************************
{
	const graph_t* oldvalue = graph;
	if (graph!=new_graph)
	{
		graph = new_graph;
		n_vertices = num_vertices(*graph);
		clear();
	}
	
	return oldvalue;
}

//*******************************************************************************
template<class graph_t>
pair<distancemap_t<graph_t>::layer_cit,distancemap_t<graph_t>::layer_cit> 
distancemap_t<graph_t>::get_layer(const vertex_t v, const int l)
//*******************************************************************************
{
	if (l<dmap[v].size())
		return pair<layer_cit,layer_cit>((dmap[v])[l].begin(),dmap[v][l].end());
	else
		return pair<layer_cit,layer_cit>(empty_layer.end(),empty_layer.end());
}

//*******************************************************************************
template<class graph_t>
vector<int> distancemap_t<graph_t>::operator[](const vertex_t v) const
//*******************************************************************************
{
	//the resulting vector initialized with -1 for not calculated distances
	vector<int> result(n_vertices,-1);
	
	//index property map of the graph
	property_map<graph_t, vertex_index_t>::type index = get(vertex_index, *graph);

	/* ITERATE THROW ALL LAYERS OF VERTEX v */
	
	int l=0; //actual layer number
	
	for (distancetab_cit layer=dmap[v].begin();layer!=dmap[v].end;++layer)
	{
		for(layer_cit entry=layer->begin();entry!=layer->end();++entry)
		{
			int ix = get(index, *entry);
			result[ix] = l;
		}
		
		++l;  //increase distance to the next layer
	}
	
	return result;
}

//*******************************************************************************
template<class graph_t>
int distancemap_t<graph_t>::operator()(const vertex_t v, const vertex_t w) const
//*******************************************************************************
{
	int result = -1;
	
	int l=0; //actual layer number
	
	for (distancetab_cit layer=dmap[v].begin();layer!=dmap[v].end;++layer)
	{
		for(layer_cit entry=layer->begin();entry!=layer->end();++entry)
			if (w==*entry) return l;
		
		++l;
	}
	
	return result;
	
}

#endif //DISTANCE_MAP_TEMPL_CC
