/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef FACETABLE_HH
#define FACETABLE_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <algorithm>
#  else
#    include <map>
#    include <algo.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <map>
#  include <algorithm>
#endif //HAVE_CONFIG_H

/**
 * \brief template class sorting a set of triangles into a rectangular grid
 *
 * The template concept for the triangles must be as follows:
 * triangle[ix] must retrieve a point description retrieving the x coordinate
 * with point[0] and the y coordinate with point[1]
 */

template<class triangle_t, class point_t>
class faceTable
{
private:
	/**
	 * \brief storage for the face table
	 *
	 * The key is: col + n * row
	 */
	multimap<int,const triangle_t*> face_table;

	/**
	 * \brief number of triangles to be stored in the table
	 */
  int n_triangles;
	/** 
	 * \brief constant determining the size of the boxes
	 * This constant is used to adjust the size of the gridcells. The larger it is
	 * the larger the cells are. The literature proposes T as 10.
	 */
	int T;

	/**
	 * \brief x size of a cell
	 */
	double step_x;

	/**
	 * \brief y size of a cell
	 */
	double step_y;

	/**
	 * \brief number of cells in a row or column
	 */
	int n_cells;

public:
	 typedef multimap<int,const triangle_t*>::const_iterator map_cit;
	 typedef pair<map_cit, map_cit> map_cits;
	
	/**
	 * \brief constructor creating the face map
	 *
	 * \param triangles specifies the number of triangles
	 * \param T_factor is a constant factor for the width and height of the map boxes.
	 *        The larger it is, the larger are the boxes.
	 * \param width horizontal range of the harmonic map (default = 2.0)
	 * \param height vertical range of the harmonic map (default = 2.0)
	 */
	faceTable(const int triangles, const int T_factor = 10,
						const double width = 2.0, const double height = 2.0);

	/**
	 * \brief standard destructor
	 */
	~faceTable();

	/**
	 * \brief the += operator for additions of triangles
	 *
	 * Triangles are added to the map with  this overloaded operator. It performs the
	 * range checking and adds the pointer of the triangle to all boxes in this range
	 */
	faceTable& operator+=(const triangle_t* triangle);

	/**
	 * \brief retrieves an iterator range of all triangles at the position \p x and 
	 * \p y
	 *
	 * The method returns a iterator pair describing a sequence of triangles which 
	 * are probably in the same box as x and y.
	 *
	 * If x and y are invalid (not in the specified range, or outside the unit disk)
	 * the program returns an [end,end] iterator sequence.
	 *
	 * \param x,y position of the point on the map. The position must be within the
	 *        range [0,width] and [0,height]
	 */
	map_cits operator()(const double x, const double y) const;
};

#ifdef INCLUDE_TMPL_DEFS
  #include "facetable_tmpl.cc"

#endif //INCLUDE_TMPL_DEFS

#endif //FACETABLE_HH
