/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef GEOMTRIMESH_HH
#define GEOMTRIMESH_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <vector>
#  else
#    include <vector.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <vector>
#endif //HAVE_CONFIG_H

#include "trimeshio.hh"


#ifdef HAVE_NAMESPACES
namespace RazorBack
{
#endif

	class Distrapprox_;

#ifdef HAVE_NAMESPACES
};
#endif

/**
 * \brief color type
 */
struct color_t
{
	double r;
	double g;
	double b;
		
	color_t(): r(0), g(0), b(0) { }
	color_t(const double r_, const double g_, const double b_):
			r(r_), g(g_), b(b_) { }

	void operator()(const double r_, const double g_, const double b_)
	{ r = r_; g = g_; b = b_; }
};

/**
 * \brief trimesh class, able to read and write triangulated mesh objects in
 * the GEOM object format used by the centerstage tool
 *
 * The \ref geomformat_page "GEOM object format" is used by 
 * <a href="http://www.math.union.edu/software/StageTools/">Centerstage</a>.
 */

class geomTrimesh : public trimeshIO
{
private:

	//*****************************************************************************
	// TYPE DEFINITIONS
	//*****************************************************************************

	typedef vector<color_t> colors_t;
	typedef colors_t::iterator colors_it;
	typedef colors_t::const_iterator colors_cit;

	//*****************************************************************************
	// PRIVATE DATA
	//*****************************************************************************

	/**
	 * \brief denotes the relevant data column for the vertex coloring
	 */
	int fColorColumn ;

	/**
	 * \brief specifies if vertex coloring should be enabled or not
	 */
	bool fVertexColoring;

	/**
	 * \brief specifies if the vertex colors should contain a dim parameter
	 *
	 * The dimming parameter can be controled by a slider or typein control in
	 * centerStage(tm). It dims the color of the vertices against white if the set
	 * to one 
	 */
	bool fDimParameter;

	/**
	 * \brief specifies if border vertices should be outlined by a 3 dot line
	 */
	bool fOutline;

	/**
	 * \brief the RGB color value of the lowest value in the range
	 *
	 * The standard value for the low color is set to blue (0,0,1)
	 */
	color_t fLowColor;
	
	/**
	 * \brief the RGB color value for the medium value in the range
	 *
	 * If any of the color components is set below zero the medium color is
	 * not considered.
	 *
	 * The standard value for the medium color is set to green (0,1,0)
	 */
	color_t fMedColor;
	
	/**
	 * \brief the RGB color value of the highest value in the range
	 *
	 * The standard value for the high color is set to red (1,0,0)
	 */
	color_t fHighColor;

	/**
	 * \brief contains the color components for each vertex in a std::vector
	 */
	colors_t fColors;

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief generates a 'vcolors' entry out of the color entry of an vertex with
	 * index \p v
	 *
	 * \return a std::string containing the color description in the form:
	 *         "{ red green blue }" 
	 *
	 * \param v index of the vertex
	 */
	string makeColorEntry(const int v) const;

	/**
	 * \brief transforms a value between 0 and 1 into a RGB color value according
	 * to the low, medium and high colors
	 *
	 * A value of zero is mapped exactly to the low color, 0.5 is mapped to the
	 * medium color if available and 1 is mapped to the high color.
	 *
	 * \param value a floating point number between 0 and 1
	 * \param color receives the RGB color for \p value
	 */
	void calcRangeColor(double value, color_t& color);
	
public:

  /**
   * \brief constructor initializing geomTrimesh with a Trimesh object
   */
  geomTrimesh(Trimesh* aTrimesh): trimeshIO(aTrimesh), 
																	fColorColumn(0), 
																	fVertexColoring(false),
																	fDimParameter(false),
																	fOutline(false),
		                              fLowColor(0,0,1),
		                              fMedColor(0,1,0),
		                              fHighColor(1,0,0),
																	fColors(0) { }

  /** 
   * \brief there is a stream handling for library own format files.
   */
  bool handleStreams() const { return true; }


  /**
   * \brief reads a surface mesh object in library own format from a file 
   * specified in \p aFile
   *
   */
  bool read(const string& aFile);
  
  /**
   * \brief reads a surface mesh object in library own format from the stream
   * specified in \p aStream
   */
  bool read(istream& aStream);

  /**
   * \brief writes a surface mesh object in library own format into a file 
   * specified in \p aFile
   *
   */
  bool write(const string& aFile);

  /**
   * \brief writes a surface mesh object in library own format into a stream 
   * specified in \p aStream
   *
   */
  bool write(ostream& aStream);

  //*****************************************************************************
  // PUBLIC PROPERTY METHODS
  //*****************************************************************************

	/**
	 * \brief retrieves the relevant column for the vertex coloring
	 *
	 * \return the index of the matrix column used for vertex coloring
	 */
	int getColorColumn() const { return fColorColumn; }

  /**
	 * \brief sets the relevant color for the vertex coloring
	 *
	 * \return the index of the old relevant column
	 *
	 * \param col new relevant color column.
	 */
	int setColorColumn(const int col);
	
	/**
	 * \brief returns \c true if vertex coloring has been enabled
	 *
	 * \return \c true if vertex coloring has been enabled
	 */
	bool isVertexColoring() const { return fVertexColoring; }

	/**
	 * \brief enables or disables the vertex coloring
	 *
	 * \return old status of the vertex coloring
	 * 
	 * \param coloring if \c true enables the coloring, disables it if \c false
	 */
	bool enableColoring(const bool coloring);

	/**
	 * \brief returns if a dimming parameter is set or not.
	 *
	 * \return \c true if the vertex colores should be equipped with a dimming
	 *         parameter
	 */
	bool isDimParameter() const { return fVertexColoring; }

	/**
	 * \brief enables or disables the use of a dimming parameter
	 *
	 * \return old status of the dimming parameter
	 * 
	 * \param dimming if \c true enables dimming, disables it otherwise
	 */
	bool enableDimParameter(const bool dimming);

	/**
	 * \brief returns if border edges shell be outlined
	 *
	 * \return \c true if the border edges are outlined by a black line
	 */
	bool isOutline() const { return fOutline; }

	/**
	 * \brief enables or disables black borderlines
	 *
	 * \return the old status of the outline flag
	 * 
	 * \param outline if \c true enables outlining of border edges
	 */
	bool enableOutline(const bool outline);

	/**
	 * \brief retrieves the RGB value of one of the range colors in the
	 * arguments \p r, \p g and \p b
	 *
	 * For valid values of the \p color parameter see 
	 * setRangeColor(const int, const double, const double, const double).
	 *
	 * \param color index of the desired range color
	 * \param r,g,b receiving RGB values of the specified colors
	 */
	void getRangeColor(const int color, double& r, double& g, double& b);

	/**
	 * \brief assigns a new RGB color value to one of the range colors
	 *
	 * Wether the high, medium or low color value should be altered is specified
	 * with the \p color parameter. The \p r, \p g, \p b values are expected to be
	 * in a range between 0 and 1.
	 *
	 * Valid values for the \p color parameter are:
	 * - 0 = low color
	 * - 1 = medium color
	 * - 2 = high color
	 *
	 * \param color index of the range color, that is going to be changed
	 * \param r,g,b new RGB values for the specified colors
	 */
	void setRangeColor(const int color, 
										 const double r, const double g, const double b);
	/**
	 * \brief calculates the colors for each vertex based on the data and color
	 * column specified in the different parameters
	 *
	 * The color calculation is done by a density probability function and is set 
	 * to the boundaries given by the parameters \p min and \p max. If both 
	 * parameters are set to zero (or omitted), the boundaries are calculated by 
	 * the range of the data in the selected column.
	 *
	 * This function calls calcColors(const Distapprox_&)
	 *
	 * \param max maximum data member in the selected column
	 * \param min minimum data member in the selected column
	 */
	void calcColors(double max = 0.0, double min = 0.0);

	/**
	 * \brief calculates the colors for each vertices with the given commulative
	 *  probability function \p dist
	 *  
	 * \param dist approximate distribution function.
	 */
	void calcColors(RazorBack::Distrapprox_* dist);

};

#endif 


/**
 * \page geomformat_page GEOM object format
 *
 * This file format is used by the 
 * <a href="http://www.math.union.edu/software/StageTools/">Centerstage</a>
 * viusalization program. 
 *
 * The following descripton is the peticular form of GEOM files, which are created
 * by the writer:
 *
 * \section geomfile_vertices Vertices:
 * Vertices are listed in an embraced section after the keyword \c Vertices:
 * 
 * The form of one vertex descriptor is:
 * V_IX: (POSX, POSY, POSZ)
 * 
 * \param POS_[XYZ] position of the vertex in 3D coordinates
 * \param V_IX index of the vertex
 *
 * \section geomfile_faces Faces:
 * The faces are listed in an embraced section after the keyword \c Faces 
 * The the face descriptors consist of the following data:
 * { VERTEX_1 VERTEX_2 VERTEX_3 }
 *
 * \param VERTEX_[123] indices of the three vertices forming the face.
 *
 * \note The vertices must be given in the correct order. (Looking from on them
 * from the outside of the surface.)
 * \note It is very important to use the right form of brackets and commas or spaces!
 * 
 */
