/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "harmonic_image.hh"
#include "facetable.hh"

#include "img_matching.hh"
#include "normcorr_matching.hh"

#include <float.h>

#ifndef PI
#define PI 3.14159265358979
#endif 

typedef normcorr_matching std_matching;

// PRIVATE METHODS

//*******************************************************************************
void harmonicImage::calcRectPoint(const size_t ix, double& x, double& y)
//*******************************************************************************
{
	double cellwidth = (2*radius)/n_resolution;

	x = -radius + ((double) (ix % n_resolution))*cellwidth + cellwidth*0.5;
	y = -radius + floor((double) (ix/n_resolution))*cellwidth + cellwidth*0.5;	
}

	
//*******************************************************************************
void harmonicImage::calcCircularPoint(const size_t ix, double& x, double& y)
//*******************************************************************************
{
	double c = -n_resolution/2.0 + 0.5 + (ix % n_resolution);
	double r = -n_resolution/2.0 + 0.5 + floor((double)(ix/n_resolution));

  double abs_c = fabs(c);
	double abs_r = fabs(r);
	
	double gamma = 0.0;
	double epsilon = 0.0;

	if(abs_c>abs_r)
	{
		//calculate the special sign for r/abs_r
		double sr = (abs_r>DBL_EPSILON) ? r/abs_r : 1.0;
		
		gamma = abs_c;
		epsilon = 2 * (gamma - c) * sr + (c*r)/gamma;
	}
	else
	{
		gamma = abs_r;
		epsilon = 2 * r - (c*r)/gamma;
	}
	
	double rad = (2*gamma)/n_resolution;
	double theta = (gamma>DBL_EPSILON) ? ( PI * epsilon)/(4*gamma) : 0;
	
	x = radius * rad * cos(theta);
	y = radius * rad * sin(theta);
}

//*******************************************************************************
bool harmonicImage::sampleTriangle(const Triangle* t, 
																	 const double x, const double y, 
																	 vector<double>::iterator result)
//*******************************************************************************
{
	//extracting the variables
	double x0 = (*t)[0][0];  double x1 = (*t)[1][0];  double x2 = (*t)[2][0];
	double y0 = (*t)[0][1];  double y1 = (*t)[1][1];  double y2 = (*t)[2][1];
	
	double det = (x1-x0)*(y2-y0)-(y1-y0)*(x2-x0);
	
	double gamma = ((x1-x0)*(y-y0)-(y1-y0)*(x-x0)) / det;
	double beta  = ((x-x0)*(y2-y0)-(y-y0)*(x2-x0)) / det;
	double alpha = 1 - gamma - beta;

	if ((alpha>=0) && (beta>=0) && (gamma>=0))
	{
		for(int i=0;i<vertex_data.cno();++i)
			*result++ = alpha * vertex_data(t->getVertexIndex(0),i) +
				          beta  * vertex_data(t->getVertexIndex(1),i) +
			            gamma * vertex_data(t->getVertexIndex(2),i);
		return true;
	}
	else 
		return false;
}


//*******************************************************************************
int harmonicImage::resampleData(RazorBack::Rectmat_<double>& data,
																bool normalize)
//*******************************************************************************
{
	normalize = normalize || fNormalize;

	typedef faceTable<Triangle,Vector_<double> >::map_cit face_cit;
	typedef faceTable<Triangle,Vector_<double> >::map_cits face_cits;

	int i;
	int c=0;

	//building the face table for fast face access during the resampling process
	faceTable<Triangle,Vector_<double> > face_table(harmonic_map->nFaces(),5);
	
	for(Trimesh::FaceIter f=harmonic_map->beginFace();f!=harmonic_map->endFace();++f)
		face_table += static_cast<const Triangle*>(&(*f));
	
	int img_size = data.rno();

	//"length" of image
	Vector_<double> lengths(data.cno(),0);

	for(i=0;i<img_size;++i)
	{
		//calculating the centers of the actual image cell
		double x = 0.0;
		double y = 0.0;
		switch (fGrid)
		{
			case gtRect :     calcRectPoint(i,x,y);	    break;
			case gtCircular : calcCircularPoint(i,x,y); break;
		}

		//get all triangles, that are in the same box
		face_cits possible_faces = face_table(x,y);
		
		//initialize the pixel as invalid
		data(i,0) = invalid_pixel;

		//search for the triangle enclosing the point x,y and calculate the pixel
		//value
		for(face_cit f=possible_faces.first;f!=possible_faces.second;++f)
		{
		  if (sampleTriangle(f->second, x, y, data[i])) 
			{
        //calculatin image "length" for normalization
				if (normalize) 
					for(int j=0;j<data.cno();++j)	lengths[j] += data[i][j] * data[i][j];

				break;
			}
						
		}
		++c;
	}


	//normalize
	if (normalize)
	{
		for(i=0;i<data.cno();++i)	lengths[i] = sqrt(lengths[i]);

		for(i=0;i<img_size;++i)
			if (abs(data(i,0)-invalid_pixel)>DBL_EPSILON)
				for(int j=0;j<data.cno();++j) 
					data[i][j]/=lengths[j];
	}

	//the image is now valid
	img_valid = true;

	return c;
}

//*******************************************************************************
int harmonicImage::resampleAxes(vector<Rectmat_<double> >& axes)
//*******************************************************************************
{
	typedef faceTable<Triangle,Vector_<double> >::map_cit face_cit;
	typedef faceTable<Triangle,Vector_<double> >::map_cits face_cits;

	int a,i,j;
	int c=0;

	//building the face table for fast face access during the resampling process
	faceTable<Triangle,Vector_<double> > face_table(harmonic_map->nFaces(),5);
	
	for(Trimesh::FaceIter f=harmonic_map->beginFace();f!=harmonic_map->endFace();++f)
		face_table += static_cast<const Triangle*>(&(*f));
	
	int img_size = n_resolution * 8;

	//the real axes 0, 1, 2, 3,
	for(a=0;a<4;++a)
	{
		//dimensionate
		axes[a].size(n_resolution, vertex_data.cno());

		//setup the start index and step for the four real axes
		int start, step;
		switch(a)
		{
			case 0 : start = n_resolution*(n_resolution-1)/2; step = 1; break;
			case 1 : start = 0; step = n_resolution + 1; break;
			case 2 : start = (n_resolution-1)/2; step = n_resolution; break;
			case 3 : start = n_resolution -1; step = n_resolution -1; break;
		}

		for(i=0;i<n_resolution;++i)
		{
			//calculating the centers of the actual image cell
			double x = 0.0;
			double y = 0.0;
			calcCircularPoint(start+i*step,x,y);

			face_cits possible_faces = face_table(x,y);
			//initialize the pixel as invalid
			axes[a](i,0) = invalid_pixel;

			//search for the triangle enclosing the point x,y and calculate the pixel
			//value
			for(face_cit f=possible_faces.first;f!=possible_faces.second;++f)
				if (sampleTriangle(f->second, x, y, axes[a][i])) 
					break;
			++c;
		}
	}

	//create the 4 inverse axes
	for(a=4;a<8;++a)
	{
		//dimensionate
		axes[a].size(n_resolution, vertex_data.cno());

		for(i=0;i<n_resolution;++i)
		{
			//select the ith row of reverse from front
			vector<double>::iterator reverse = axes[a][i];
			//select the ith row of real from back
			vector<double>::iterator real = axes[a-4][(n_resolution-1)-i];
			//copy the data
			for(j=0;j<axes[a].cno();++j) *reverse++ = *real++;
			++c;
		}
	}
	
	
	return c;	
}	

// INITIALIZATION METHODS AND CONSTRUCTORS

//*******************************************************************************
harmonicImage::harmonicImage(Trimesh* Mesh, 
														 const double rotation_step, const int resolution,
														 const double invalid_value, const double disk_radius,
														 const bool normalize, const gridtype_t grid)
//*******************************************************************************
: harmonic_map(Mesh), 
	delta_phi(rotation_step), 
	n_resolution(resolution),
	vertex_data(), 
	image_data(resolution*resolution), 
	search_data(),
	data_id(), 
	img_valid(false),
	searchdata_valid(false),
	invalid_pixel(invalid_value),
	radius(disk_radius),
	phi(0),
	fNormalize(normalize),
	fGrid(grid), 
	fMaxAngle(0)
{
	//resize the vertex data array
	vertex_data.size(harmonic_map->nVertices(),1);

	//if resolution is zero, calculate the default resolution
	if (n_resolution==0)
	{
	  n_resolution = (int)ceil(sqrt((double)harmonic_map->nVertices()));
		image_data.size(n_resolution*n_resolution,1);
	}
}

//*******************************************************************************
int harmonicImage::initfromMaster(const string& data_id,
																	const RazorBack::Rectmat_<double>& master_data, 
																	const double def)
//*******************************************************************************
{
	int c=0;
	vertex_data.size(harmonic_map->nVertices(),master_data.cno());

	//get the corresponding data points for each vertex on the patch
  for(Trimesh::vertex_it v=harmonic_map->begin();v!=harmonic_map->end();++v)
	{
		int master_ix = harmonic_map->vertMasterIx(*v);
		if (master_ix<master_data.rno())
			vertex_data.row(master_data.row(master_ix),c++);
		else 
		{
			for(int i=0;i<vertex_data.cno();++i) vertex_data(c,i) = def;
			++c;
		}
	}
	
	img_valid = false;
	searchdata_valid = false;

	image_data.size(image_data.rno(),master_data.cno());

	return c;
}

//*******************************************************************************
int harmonicImage::initfromPatch(const string& data_id,
																 const RazorBack::Rectmat_<double>& patch_data, 
																 const double def)
//*******************************************************************************
{
	if (patch_data.rno() == harmonic_map->nVertices())
	{
		vertex_data = patch_data;
		image_data.size(image_data.rno(),patch_data.cno());

		return patch_data.rno();
	}
	else
		return 0;
}

// PROPERTY FUNCTIONS

//*******************************************************************************
double harmonicImage::setInvalidValue(const double invalid_value)
//*******************************************************************************
{
	double old_value = invalid_pixel;
	
	if (fabs(invalid_value-invalid_pixel)>1e-15)
	{
		invalid_pixel = invalid_value;
		img_valid = false;
	}
	else
		old_value = invalid_value;

	return old_value;
}


//*******************************************************************************
double harmonicImage::setRotation(double angle, bool resample)
//*******************************************************************************
{
	if (fabs(angle-phi)>DBL_EPSILON)
	{
		double old_value = phi;
		rotate(angle-phi, resample);
		
		return old_value;
	}
	else 
		return phi;
}


//*******************************************************************************
double harmonicImage::setDefaultStep(const double default_step)
//*******************************************************************************
{
	double old_value = delta_phi;
	
	if ((fabs(default_step)>1e-15) && (fabs(default_step)<2*PI))
	{
		delta_phi = default_step;
		searchdata_valid = false;
	}
	else
		old_value = default_step;
	
	return old_value;
}

//*******************************************************************************
int harmonicImage::setResolution(const int n_pix)
//*******************************************************************************
{
	int old_value = n_resolution;
	
	if ((n_pix>0) && (n_pix!=n_resolution))
	{
		n_resolution = n_pix;

		image_data.size(n_resolution*n_resolution,image_data.cno());
		img_valid = false;
		searchdata_valid = false;
	}
	else
		old_value = n_pix;
	
	return old_value;
}

//*******************************************************************************
bool harmonicImage::enableNormalize(const bool normalize)
//*******************************************************************************
{
	bool oldvalue = fNormalize;
	if (fNormalize != normalize)
	{
		fNormalize = normalize;
		img_valid = false;
	}
	return oldvalue;
}

//*******************************************************************************
harmonicImage::gridtype_t harmonicImage::setGridtype(const gridtype_t grid)
//*******************************************************************************
{
	gridtype_t oldvalue = fGrid;
	if (fGrid != grid)
	{
		fGrid = grid;
		img_valid = false;
	}
	return oldvalue;
}

//*******************************************************************************
const double* harmonicImage::getImageData()
//*******************************************************************************
{
	if (!img_valid) resampleData(image_data);
	return static_cast<const double*>(&(*image_data[0]));
}


// OPERATOR FUNCTIONS

//*******************************************************************************
double harmonicImage::operator()(const int x, const int y, const int prop)
//*******************************************************************************
{
	if ((x<n_resolution) && (y<n_resolution))
	{
		if (!img_valid) resampleData(image_data);
		return image_data(x+n_resolution*y, prop);
	}	
	else 
		throw out_of_range("x or y");

	return 0;
}

//*******************************************************************************
RazorBack::Vector_<double> harmonicImage::operator[](const int ix)
//*******************************************************************************
{
	if ((ix<image_data.cno()) && (ix>=0))
	{
		if (!img_valid) resampleData(image_data);
		return image_data.col(ix);
	}	
	else 
		throw out_of_range("ix");
}

// PUBLIC METHODS

//*******************************************************************************
bool harmonicImage::rotate(double angle, const bool resample, 
													 const bool normalize)
//*******************************************************************************
{
	if (fabs(angle)<DBL_EPSILON) angle = delta_phi;

	double sin_a = sin(angle);
	double cos_a = cos(angle);
	
	//rotate the vertices
	for(Trimesh::vertex_it v=harmonic_map->begin();v!=harmonic_map->end();++v)
	{
		Vector_<double> coords = harmonic_map->vertCoordinates(*v);
		double x = coords[0];
		double y = coords[1];

		double new_x = x * cos_a - y * sin_a;
		double new_y = x * sin_a + y * cos_a;
		
		harmonic_map->setCoordinates(*v,new_x,new_y,coords[2]);
	}

	//resample the vertex data to obtain the new harmonic image
	if (resample)	resampleData(image_data);

	phi += angle;
	
	if ((phi-2*PI)>=DBL_EPSILON) 
	{
		phi -= 2*PI;

		return true;
	}
	else 
		return false;
	
}

//*******************************************************************************
double harmonicImage::compare(harmonicImage& key, const int cols,
															const bool reset, const bool bestrot)
//*******************************************************************************
{
	//check if the image data is valid
	std_matching std_matcher(this, cols);

	return compare_by(key, std_matcher, reset, bestrot);
}

//*******************************************************************************
double harmonicImage::compare_by(harmonicImage& key, img_matching& matcher, 
																 const bool reset, const bool bestrot)
//*******************************************************************************
{
	
	//the resolution of both images must be equal
	if (key.n_resolution!=n_resolution)
		if (key.getResolution() > n_resolution)
			setResolution(key.getResolution());
		else
			key.setResolution(n_resolution);

	//if the searchdata is invalid initilize the precompiled pixel arrays
	if (!searchdata_valid) initSearch();
	//check if the key data is valid
	key.doResampling();

	//set the first key of the matcher to this image if needed
  if (reset) matcher.setImage(this);
	
	matcher.resetsearch(search_data.begin());
	
	double max_corr = -1;
	fMaxAngle = 0.0;
	
	for(int i=0;i<=search_data.size()-1;++i)
	{
		double corr = matcher(key);

		if (max_corr<corr)
		{
			max_corr = corr;
			fMaxAngle = -i*delta_phi;
		}
		
	}


	//leave the key image in the position of maximum correlation
	if (bestrot)
		key.setRotation(fMaxAngle, false);

	return max_corr;
}


//*******************************************************************************
bool harmonicImage::doResampling(const bool normalize)
//*******************************************************************************
{
	if (!img_valid) 
	{
		resampleData(image_data, normalize);
		return true;
	}
	else
		return false;
}

//*******************************************************************************
bool harmonicImage::initSearch()
//*******************************************************************************
{
	if (!searchdata_valid)
	{
		//if the image is not valid, copy the first pixel array in the image_data
		bool write_imgdata = (!img_valid);
	
		//set the rotation state to zero and store the old rotation value
		double oldphi = setRotation();
	
		//reset the searchdata structure
		search_data.resize(rint(2*PI/delta_phi));
	
		//resize the single pixel arrays and resample the data
		for(searchdata_it i=search_data.begin();i!=search_data.end();++i)
		{
			i->size(n_resolution*n_resolution,image_data.cno());
			resampleData(*i);

#ifdef LT_DEBUGGING
			for(int k = 0; k < n_resolution*n_resolution;++k)
			{
				if (k>0) cout << "\t";
				cout << (*i)(k,0);
			}
			cout << endl;
#endif
			rotate(0, false);
		}
	
		//rerotate the image in the old position
		rotate(oldphi-phi, false);
	
		if (write_imgdata)
			image_data = search_data.front();
	
		img_valid = false;
		searchdata_valid = true;
	
		return true;
	}
	else 
		return false;
}

//*******************************************************************************
bool harmonicImage::initScreening()
//*******************************************************************************
{
	
	//structure to receive the eight symmetry axes
	vector<RazorBack::Rectmat_<double> > axes(8);
	
	resampleAxes(axes);
	
	//reset the searchdata structure to eight variants
	search_data.resize(8);
	
	//resize the single pixel arrays and resample the data
	int sequence = 0;
	for(searchdata_it i=search_data.begin();i!=search_data.end();++i, ++sequence)
	{
		i->size(n_resolution * 8,image_data.cno());
		vector<double>::iterator ipos = (*i)[0];
		//shift copy the sequence into the search data (0-7, 1-70, 2-701 ...)
		for(int j=sequence;j<(sequence+8);++j)
		{
			int ix = (j<8) ? j : j-8;
			//copy the rows of the specific axis
			vector<double>::iterator apos = axes[ix][0];
			for(int k=0;k<axes[ix].rno()*axes[ix].cno();++k)	
				*ipos++ = *apos++;
		}
	}

	//set the image data to the first search data entry
	image_data = search_data.front();
	
	img_valid = true;
	searchdata_valid = true;
	
	return true;
}
