/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef HARMONIC_IMAGE_HH
#define HARMONIC_IMAGE_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <string>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <string>
#endif //HAVE_CONFIG_H

#include <Rectmat.hh>
#include <Vector.hh>

#include "trimesh.hh"
#include "img_matching.hh"


#define ROTSTEP 0.0349065850398866
#define INVALID -1.0e108

/**
 * \brief representation of an harmonic image.
 * 
 * The harmonicImage class consists of a harmonic map in a triangulated mesh 
 * representation and of a vector containing some data which should be resampled
 * on the image.
 *
 */

class harmonicImage
{
public:
	//*****************************************************************************
	// PUBLIC TYPES
	//*****************************************************************************
	
	/**
	 * \brief used to specify a certain gridtype
	 */
	enum gridtype_t {gtRect, gtCircular};

private:
	/**
	 * \brief triangulated mesh containing the harmonic map
	 */
  Trimesh* harmonic_map;

	/**
	 * \brief vertex data for the imaging process
	 */
	RazorBack::Rectmat_<double> vertex_data;

	/**
	 * \brief radius of the unit disk
	 */
	double radius;

	/**
	 * \brief this value is set at pixels, which lay outside the unit disk
	 */
	double invalid_pixel;
	
	/**
	 * \brief flag indicating, if the image data is valid or not
	 *
	 * The image data becomes invalid, if the vertex data or the resampling 
	 * resolution has changed. If the image data is invalid, the image is recreated
	 * automatically if any method or operation referring to the resampled data is 
	 * called.
	 */
	bool img_valid;

	/**
	 * \brief flag indicating, if the search data is valid or not
	 *
	 * The search data becomes invalid, if the vertex data or the resampling 
	 * resolution or the rotation step has changed. If the search data is invalid, 
	 * the search data is recreated automatically at the beginning of a 
	 * harmonicImage::compare or harmonicImage::compare_by process.
	 */
	bool searchdata_valid;

	/**
	 * \brief actual rotation angle
	 *
	 * Each rotation changes the actual rotation angle and the next rotation is
	 * started from this position and not from the initial state.
	 */
	double phi;
	
	/**
	 * \brief default rotation step
	 */
	double delta_phi;

	/**
	 * \brief number of columns and rows in the resampled images
	 */
	int n_resolution;

	/**
	 * \brief denotes if the resampled image should be normalized
	 *
	 * If normalization is active the values of the harmonic image's pixel are
	 * normalized to form a unit-length vector:
	 *
	 * \f[
	 *     img^{0}_{i} = \frac{img_{i}}{\left||img\right||}
	 * \f]
	 */
	bool fNormalize;

	/**
	 * \brief specifies the type of grid that is used for resampling
	 *
	 * If fGrid is set to \c gtRect the map is resampled with a rectangular grif.
	 * In the case of \c gtCircular the grid has a circular toplogy and all points
	 * are located within the unit circle.
	 *
	 * For the calculation procedure of the gridpoints please see 
	 * calcRectPoint(const size_t, double&, double&) and 
	 * calcCircularPoint(const size_t, double&, double&).
	 */
	gridtype_t fGrid;

	/**
	 * \brief stores the rotation angle of maximum correlation of the last image
	 * compare
	 *
	 * This angle is set according to the angle of maximum correlation in the last
	 * compare process. If you are interested in the angle of maximum correlation
	 * use the method getMaxAngle().
	 */
	double fMaxAngle;

	/**
	 * \brief identifier for the vertex data
	 */
	string data_id;

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	/**
	 * \brief for a given index \p ix calculates the x and y position for a
	 * rectangular grid
	 *
	 * The index is mapped to a grid with n_resolution * n_resolution points and
	 * the index 0 points at the point -1|-1 proceeding first left then up to the
	 * point 1|1
	 *
	 * \param ix index in the range [0, n_resolution * n_resolution) for which the
	 *                 coordinates have to be calculated
	 * \param x,y double coordinates of the point \p ix
	 */
	void calcRectPoint(const size_t ix, double& x, double& y);
	
	/**
	 * \brief for a given index \p ix calculates the x and y position for a
	 * circular grid
	 *
	 * The index \p ix pointing on a gridpoint (r)ow|(c)olumn is mapped on a point
	 * inside the unit circle. The coordinates \p x and \p y are calculated with
	 * the following computations:
	 * \f[
	 *    \gamma = max(\left| c \right|, \left| r \right|)
	 * \f]
	 * If \f$\gamma = \left| c \right| \f$
	 * \f[
	 *    \xi = \frac{c \cdot r}{\gamma } + \frac{2 \cdot r \cdot
   *          \left(\gamma - c \right) }{\left| r \right|}
	 * \f]
	 * else 
	 * \f[
	 *    \xi = 2 \cdot r - \frac{c \cdot r}{\gamma}
	 * \f]
	 * 
	 * The coordinates of the point are calculated via the polar coordinates r, 
	 * \f$\theta\f$ :
	 *
	 * \f[
	 *     r = \frac{2 \cdot \gamma}{N}
	 * \f]
	 *
	 * \f[
	 *     \theta = \frac{\pi \cdot \xi}{4\gamma}
	 * \f]
	 *
	 * \param ix index in the range [0, n_resolution * n_resolution) for which the
	 *                 coordinates have to be calculated
	 * \param x,y double coordinates of the point \p ix
	 */
	void calcCircularPoint(const size_t ix, double& x, double& y);

	/**
	 * \brief calculates the barycenter coordinates and the sample value if the 
	 * triangle is valid
	 */
	bool sampleTriangle(const Triangle* t, 
											const double x, const double y, vector<double>::iterator result);
	/**
	 * \brief does the resampling of the vertex data
	 *
	 * returns the number of resampled pixels. If an error occured, the returnvalue
	 * is not equal the square of the resolution
	 */
	int resampleData(RazorBack::Rectmat_<double>& data, bool normalize = false);

	
	/**
	 * \brief does the resampling of the symmetry axes
	 *
	 * returns the number of resampled pixels. If an error occured, the returnvalue
	 * is not equal the square of the resolution
	 */
	int resampleAxes(vector<Rectmat_<double> >& axes);

protected:

	/**
	 * \brief precompiled pixel arrays for all rotation steps
	 *
	 * This set of pixel-vectors is used to compare this harmonic image to a set
	 * of other harmonic images. It contains an array of image pixels for each
	 * rotationstep between 0 and 360, defined by harmonicImage::delta_phi.
	 *
	 * The precompiled data is generated in the harmonicImage::initSearch method.
	 */
	vector< RazorBack::Rectmat_<double> > search_data;
	
	/**
	 * \brief writeable iterator for the searchdata array of pixel arrays.
	 */
	typedef vector< RazorBack::Rectmat_<double> >::iterator searchdata_it;
	
public:
	/**
	 * \brief pixel array containing the last resampled harmonic image
	 *
	 * The image_data structure in contrast to the harmonicImage::search_data
	 * structure contains the image as it was resampled the last time. It is used
	 * to create a graphical representation or to compare it against another image
	 * as the second key.
	 */
	RazorBack::Rectmat_<double> image_data;
	
	/**
	 * \brief iterator for the searchdata array of pixel arrays.
	 *
	 * search data pixel array iterators are returned by the 
	 * harmonicImage::search_begin() and harmonicImage::search_end() method.
	 */
	typedef vector< RazorBack::Rectmat_<double> >::const_iterator searchdata_cit;

	/**
	 * \brief default constructor initializing the harmonic_image
	 *
	 * The constructor initializes the Mesh and sets the size vertex_data of the 
	 * vertex data to the number of vertices in the mesh. If not specified the
	 * rotation step is set to an initial value of 2 degrees (0.03491 rad), the
	 * n_resolution is set to the sqrt(n_vertices) and the value of invalid pixels
	 * is set to -1.0 .
	 *
	 * \param Mesh a pointer to a triangulated mesh containing the harmonic map. If
	 *        the mesh does not represent a harmonic map, the results of all 
	 *        operations are undefined, but the code should not crash (not tested!).
	 * \param rotation_step specifies the default rotation step.
	 * \param resolution a different resolution than sqrt(n_vertices)
	 * \param invalid_value value of invalid pixels, outside the unit disk
	 * \param disk_radius radius of the unit disk
	 * \param normalize indicates if the data values for the image should be normalized
	 * \param grid      specifies a rectangular or circular grid
	 */
	harmonicImage(Trimesh* Mesh, 
								const double rotation_step = ROTSTEP, const int resolution = 0,
								const double invalid_value = INVALID, 
								const double disk_radius = 1.0,
								const bool normalze = false,
								const gridtype_t grid = gtRect);
	
	/**
	 * \brief initialization of the vertex data out of a dataset for the whole 
	 * surface
	 *
	 * Harmonic maps are representing patches of the master surface, but properties
	 * of vertices are often calculated on the whole surface. Therefore a dataset
	 * contains information for all the vertices not only for vertices on the patch.
	 * The masterindex of the Trimesh structure is used to pick up the right data
	 * points for the vertices on the map.
	 * 
	 * The function returns the number of copied data points. If the number is lower
	 * than the number of vertices, the data for some vertices could not be found.
	 * Their data is set to the \p default value.
	 *
	 * \param data_id descriptor of the new vertex data.
	 * \param master_data vector containing the data for all surface vertices
	 * \param default value that is assigned to vertices whith no corresponding
	 *        entry in the master_data array.
	 */
	int initfromMaster(const string& data_id,
										 const RazorBack::Rectmat_<double>& master_data, const double def=0);
	
	/**
	 * \brief initialization of the vertex data out of a dataset for the patch
	 * vertices only
	 *
	 * If the data is available only for the patch vertices use this method to 
	 * initialize the vertex data.
	 *
	 * The function returns the number of copied data points. If the number is lower
	 * than the number of vertices, the data for some vertices could not be found.
	 * Their data is set to the \p default value. These vertices are usually at the 
	 * end of the Mesh's vertex list.
	 *
	 * \param data_id    descriptor of the new vertex data.
	 * \param patch_data vector containing the data for all surface vertices
	 * \param def        value that is assigned to vertices whith no corresponding
	 *                   entry in the patch_data array.
	 */
	int initfromPatch(const string& data_id,
										const RazorBack::Rectmat_<double>& patch_data, 
										const double def=0);

	//****************************************************************************
	// PROPERTY FUNCTIONS
	//****************************************************************************
	
	/**
	 * \brief retrieves a pointer to the associated harmonic map
	 *
	 * \return a constant pointer to the currently associated harmonic map
	 */
	const Trimesh* getMap() const { return harmonic_map; }
	
	/**
	 * \brief retrieves the radius of the unit disk
	 */
	double diskRadius() const { return radius; }

	/**
	 * \brief denotes, if the image data is valid or not.
	 * 
	 * The image data becomes invalid, if the vertex data or the resampling 
	 * resolution has changed.
	 */
	bool isValid() const { return img_valid; }

	/**
	 * \brief retrieves the actual rotation angle in radiants
	 */
	double rot_angle() const { return phi; }

	/**
	 * \brief retrieves the value of invalid pixels
	 *
	 * Invalid pixels are those laying outside the unit disk
	 */
	double getInvalidValue() const { return invalid_pixel; }
	
	/**
	 * \brief sets the value of invalid pixels to \p invalid_value.
	 *
	 * See harmonicImage::getInvalidValue for further information.
	 * 
	 * Returns the old invalid pixel value.
	 * 
	 * \param invalid_value new value for invalid pixels
	 */
	double setInvalidValue(const double invalid_value);

	/**
	 * \brief retrieves the actual rotation angle in radiants
	 */
	double rotation_angle() const { return phi; }

	/**
	 * \brief resets the rotation angle to zero
	 */
	double setRotation(double angle = 0, bool resample = true);

	/**
	 * \brief retrieves the default rotation step in radiants
	 *
	 * The default rotation step is used if no \p rot_step parameter is specified
	 * in the harmonicImage::rotate method.
	 */
	double getDefaultStep() const { return delta_phi; }
	
	/**
	 * \brief sets the default rotation step to \p default_step.
	 *
	 * See harmonicImage::getDefaultStep for further information.
	 * 
	 * Returns the old rotation step if \p default_step was within ]0,2*pi[,
	 * otherwise it returns the value of \p default_step 
	 * 
	 * \param default_step new value for the default rotation step
	 */
	double setDefaultStep(const double default_step);

	/**
	 * \brief retrieves the number of pixels in a row or column
	 */
	int getResolution() const { return n_resolution; }

	/**
	 * \brief sets the resolution of the resampled images to \p n_pix rows and
	 * columns
	 *
	 * If the resolution is changed, the resampled image becomes invalid
	 * 
	 * The method returns the old resolution value if \p n_pix is larger than
	 * zero otherwise it returns n_pix
	 *
	 * \param n_pix new number of pixels in each row and column
	 */
	int setResolution(const int n_pix);

	/**
	 * \brief returns \c true if the image data is going to be normalized.
	 */
	bool isNormalized() const { return fNormalize; }

	/**
	 * \brief turns normalization on or off
	 *
	 * \return old status of the normalization flag
	 *
	 * \param normalize turns normalization of image data on if \c true and off
	 *                  if \c false.
	 */
	bool enableNormalize(const bool normalize = true);

	/**
	 * \brief retrieves the gridtype used for the sampling
	 *
	 * \return \c gtRect for rectangular and \c gtCircular for circular sampling
	 */
	gridtype_t getGridtype() const { return fGrid; }

	/**
	 * \brief sets the gridtype for the resampling
	 *
	 * \return old gridtype
	 *
	 * \param grid \c gtRect for rectangular and \c gtCircular for circular 
	 *             sampling
	 */
	gridtype_t setGridtype(const gridtype_t grid);
	
	/**
	 * \brief retrieves the data identifier of the vertex data
	 */
	string dataID() const { return data_id; }

	/**
	 * \brief returns an iterator to the first search_data pixel array
	 *
	 * \note If the search data is invalid, the iterator returned is invalid too,
	 * so please check the image validity first.
	 */
	searchdata_cit search_begin() const { return search_data.begin(); }

	/**
	 * \brief returns an iterator to the position after the last  pixel in the 
	 * image data array
	 *
	 * \note If the search data is invalid, the iterator returned is invalid too,
	 * so please check the image validity first.
	 */
	searchdata_cit search_end() const { return search_data.end(); }

	/**
	 * \brief retrieves the last angle of maximum correlation
	 *
	 * This method retrieves the value of fMaxAngle, which will be set during each
	 * compare process. To be sure to receive the correct maximum correlation
	 * angle it is recommended to check this value right after the call to
	 * compare(harmonicImage&, const int, const bool, const bool) or 
	 * compare_by(harmonicImage&, img_matching&, const bool, const bool).
	 *
	 * \return the angle of maximum correlation.
	 *
	 * \sa fMaxAngle
	 */
	double getMaxAngle() const { return fMaxAngle; }

	/**
	 * \brief retrieves a pointer to the elements of the image or vertex data 
	 * matrix
	 *
	 * This method provides direct access to the data of the image or vertex data 
	 * matrix. Before reading and writing to this data, please verify that you are
	 * in the right range.
	 *
	 * \return pointer to the head of the image data matrix.
	 *
	 * \warning Be careful when using this pointer. It gives you an unchecked and
	 * direct access to the Rectmat_<double> internal data structure.
	 */
	const double* getImageData();
//const double* getVertexData()	const { return static_cast<const double*>(vertex_data[0]); }

	//*****************************************************************************
	// OPERATOR FUNCTIONS
	//*****************************************************************************
	
	/**
	 * \brief checked index operator retrieving pixel values
	 *
	 * If the resampled image is invalid the image is generated.
	 *
	 * \return the value of the pixel at position (\p x,\p y), if
	 * \p x and \p y are valid indices ([0,resolution[). 
	 *
	 * \param x,y column and row index of the desired pixel
	 * \param prop number of the property column (starting with 0  for the first
	 *             property used in the image)
	 *
	 * \exception out_of_range thrown, if x and y are not valid indices.
	 */
	double operator()(const int x, const int y, const int prop);

	/**
	 * \brief returns the pixel vector regarding to the \p ix. property in the 
	 * image
	 *
	 * If the resampled image is invalid the image is generated.
	 *
	 * \return a RazorBack::Vector_<double> containing the pixel data for the 
	 * \p ix. property.
	 *
	 * \param ix number of the desired property (starting with 0  for the first
	 *           property used in the image)
	 *
	 * \exception out_of_range thrown, if x and y are not valid indices.
	 */
	RazorBack::Vector_<double> operator[](const int ix);

	//*****************************************************************************
	// PUBLIC METHODS
	//*****************************************************************************
	
	/**
	 * \brief regenerates a rotated resampled image.
	 *
	 * The coordinates of the harmonic map are rotated around the center of the unit
	 * disk and the resampled image is generated with these new coordinates.
	 *
	 * Returns true, if the angle passes 2*pi
	 *
	 * \param angle rotation angle.
	 * \param resample automatic resampling of the image data is invoked if the
	 *        parameter is true
	 * \param normalize triggers normalization
	 */
	bool rotate(double angle = 0, const bool resample = true, 
							const bool normalize = false);

	/**
	 * \brief compares the harmonic images of \p key with itself and returns a 
	 * correlation coefficient.
	 *
	 * The matching process is performed by rotating the key image around its axis
	 * and performing a correlation check between this image and the key image.
	 * The best correlation coefficient is returned.
	 *
	 * \note The key harmonic image will be rotated around itself, therefore it is 
	 * possible, that the key is not exactly in the same rotation state after the 
	 * method has finished the comparison. 
	 *
	 * \param key key harmonic image
	 * \param cols the data column that should be used
	 * \param reset resets the first key image
	 * \param bestrot leave the key image in the position of maximum correlation
	 */
	double compare(harmonicImage& key, const int cols = 0, 
								 const bool reset = false, const bool bestrot = true);

	/**
	 * \brief compares the harmonic images of \p key with itself, using the matching
	 * method \p img_match and returns a correlation coefficient.
	 *
	 * See also: harmonicImage::compare
	 *
	 * \param key key harmonic image
	 * \param matcher binary function object accepting two harmonicImage objects and
	 *        returning a correlation coeffitient between 0 and 1.
	 * \param reset resets the first key image
	 * \param bestrot leave the key image in the position of maximum correlation
	 */
	double compare_by(harmonicImage& key, img_matching& matcher, const bool reset = false,
										const bool bestrot = true);

	/**
	 * \brief starts the resampling process if the image is invalid
	 *
	 * The method returns true, if a resampling process was necassery.
	 */
	bool doResampling(const bool normalize = false);

	/**
	 * \brief starts the precompilation of the rotation step image data
	 *
	 * The method returns true, if a precompilation was necassery.
	 * \note No searching and patch comparison is invoked.
	 */
	bool initSearch();

	/**
	 * \brief starts the precompilation of screening data
	 *
	 * Harmonic image comparison can either be performed in a full mode
	 * (generating an image for every rotation step and compare) or by screening
	 * the 8 symmetry axes of the circular grid.
	 *
	 * The screening process is much faster than the full compare, because the
	 * image is constructed only once, and the eight different rotations of axes
	 * are generated by permutation.
	 */
	bool initScreening();
	
	/**
	 * \brief returns the number of different rotations that are scanned during a comparison
	 *
	 * \return the number of rotations scanned during a comparison with another
	 * harmonic image.
	 */
	size_t getScanSteps() const { return search_data.size(); }

	/*
		Planned: correlation calculation is done via a function class family

		   operator(vector<double> a,b) returns the correlation coefficient of the
			 two image data arrays.	
	 */
		
};

#endif //HARMONIC_IMAGE_HH
