/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef IMG_MATCHING_HH
#define IMG_MATCHING_HH

#include <Rectmat.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <functional>
#    include <vector>
#    include <list>
#  else
#    include <function.h>
#    include <vector.h>
#    include <list.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <functional>
#  include <vector>
#  include <list>
#endif //HAVE_CONFIG_H

#define MAX_COLS 32;

class harmonicImage;

/**
 *  \brief exception, thrown, if two keys do not have the same resolution
 */ 
class match_diffresolution { };

/**
 * \brief base class for all harmonic image mapping functors
 *
 * They must obey the following concept:
 *
 * The first key is specified in the constructor , \c operator(key2) compares two
 * harmonic images \c copy_corrfunction(OUT) fills the sequence OUT with the 
 * single correlation values
 */
class img_matching: public unary_function<const harmonicImage*, double>
{
protected:
	
	/**
	 * \brief container storing every calculated correlation coefficient in the 
	 * objects live
	 */
	list<double> correlations;
	
	/**
	 * \brief an integer value denoting the selected columns.
	 *
	 * For more information about the interpretation of \p cols see: 
	 * img_matching::img_matching()
	 */
	int cols;
	
	/**
	 * \brief iterator pointing to the last searchdata array
	 */
	vector< RazorBack::Rectmat_<double> >::const_iterator search_array;
	

	/**
	 * \brief referrence to the first key
	 */
	const harmonicImage* key1;

public:
	/**
	 * \brief standard constructor initializing the correlations list
	 *
   * The meaning of the \p cols parameter is as follows: <br>
	 * each bit in the integer corresponds to a column in the data matrix
	 * (i.e. 1 is the first column 2 the second, 4 the third, 8 the forth ...).
	 * 0 (\b default) indicates, that all columns should be considered. 
	 *
	 * \param key     first harmonicImage to be compared
	 * \param selcols integer denoting the columns of the image data matrix, which
	 *                should be considered in the matching computation
	 */
	img_matching(const harmonicImage* key = 0, const int selcols = 0)
			: correlations(), key1(key), search_array(), cols(selcols) { }

	/**
	 *  \brief changes the referrence of \c key1 to \p key
	 */
	virtual void setImage(const harmonicImage* key)	{ key1 = key; }
	
	/**
	 * \brief resets the search_array iterator to the first search array of key 1
	 */
  void resetsearch(vector< RazorBack::Rectmat_<double> >::const_iterator sit)
	{  
		search_array = sit;
	}

	/**
	 * \brief function operator comparing two harmonic images
	 *
	 * returns a correlation coefficient between zero and one
	 *
	 * \param key2 second harmonic image to be compared
	 */
	virtual double operator()(const harmonicImage& key2) = 0;

  /**
	 * \brief copies all the correlation values into an insert iterator \p ins
	 *
	 * The method returns the number of copied values
	 * 
	 * \param ins input iterator receiving the correlation values
	 */
	template<class Out> int copy_correlations(Out ins) const;

	/**
	 * \brief clears the correlation values list
	 */
	void clear_correlations() { correlations.clear(); }
	
};

//*******************************************************************************
template<class Out> int img_matching::copy_correlations(Out ins) const
//*******************************************************************************
{
	int c = 0;

	list<double>::const_iterator i = correlations.begin();

	while (i!=correlations.end())
	{
		*ins++ = *i++;
		++c;
	}
	
	return c;
};

#endif //IMG_MATCHING_HH
