/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "mapping.hh"
#include <Lu.hh>
#include <unistd.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <algorithm>
#    include <set>
#  else
#    include <algo.h>
#    include <set.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <algorithm>
#  include <set>
#endif //HAVE_CONFIG_H

//*******************************************************************************
void harmonic_mapping::transformCoordinates(map_t& harmonic)
//*******************************************************************************
{
	for(map_it i=harmonic.begin();i!=harmonic.end();++i)
	{
		double r = i->second.first;
		double phi = i->second.second;

		i->second.first = r * cos(phi);
		i->second.second = r * sin(phi);
	}
}

//*******************************************************************************
void harmonic_mapping::calcSprings(Trimesh& mesh)
//*******************************************************************************
{
	for(Trimesh::edge_it e = mesh.beginEdge();e!=mesh.endEdge();++e)
	{
		double weight = mesh.edgeWeight(*e);
		if (weight>1e-8)
			weight = 1/weight;
		else
			weight = 0;

		mesh.setEdgeWeight(*e,weight);
	}
}

//*******************************************************************************
void harmonic_mapping::fillMatrices(RazorBack::Rmatdbl_& A_matrix, 
																		RazorBack::Rmatdbl_& B_matrix,
																		map_t& harmonic, 
																		Trimesh& mesh, intixmap_t& ixmap)
//*******************************************************************************
{
	for(Trimesh::vertex_it v = mesh.begin();v!=mesh.end();++v)
		if (mesh.vertKind(*v)==PATCH_INSIDE)
		{
		  int mix = ixmap[*v];

			double sum_weights = 0;

			for(Trimesh::adjacency_it a=mesh.beginAdjacency(*v);a!=mesh.endAdjacency(*v);++a)
			{
				//retrieve the weight of the edge between *a and *v
				double weight = mesh.edgeWeight(*v,*a);
				
				sum_weights += weight;

				if (mesh.vertKind(*a)==PATCH_INSIDE)
					A_matrix(mix,ixmap[*a]) = -1 * weight;
				else
				{
					B_matrix(mix,0) += weight * harmonic[*a].first;
					B_matrix(mix,1) += weight * harmonic[*a].second;
				}
			}
			
			A_matrix(mix,mix) = sum_weights;
		}
}

//*******************************************************************************
int harmonic_mapping::makeixMap(Trimesh& mesh, intixmap_t& ixmap)
//*******************************************************************************
{
	int ix=0;
	for(Trimesh::vertex_it v=mesh.begin();v!=mesh.end();++v)
		if (mesh.vertKind(*v)==PATCH_INSIDE) ixmap[*v]=ix++;
	
	return ix;
}

//*******************************************************************************
void harmonic_mapping::makedistMap(Trimesh& mesh, Trimesh::vertex_t center, 
																	 distmap_t& distance)
//*******************************************************************************
{
	vector<double> dists(mesh.nVertices());
	
	mesh.calcOnSurfaceDistance(center,dists);
	for(Trimesh::vertex_it v=mesh.begin();v!=mesh.end();++v)
		if (mesh.vertKind(*v)==PATCH_CONCARVEBORDER) distance[*v]=dists[*v];
}

//*******************************************************************************
void harmonic_mapping::resetCoords(Trimesh& mesh, map_t& harmonic)
//*******************************************************************************
{
	for(Trimesh::vertex_it v=mesh.begin();v!=mesh.end();++v)
		mesh.setCoordinates(*v,harmonic[*v].first,harmonic[*v].second,0);
} 

//*******************************************************************************
void harmonic_mapping::mapInterior(RazorBack::Rmatdbl_& A_matrix, 
																	 RazorBack::Rmatdbl_& B_matrix,
																	 map_t& harmonic, const intixmap_t& ixmap)
//*******************************************************************************
{
	//create the lu decomposited form of the A_matrix
	RazorBack::Lu_ lu(A_matrix);

	RazorBack::Vecdbl_ X = lu.solve(B_matrix.col(0));
	RazorBack::Vecdbl_ Y = lu.solve(B_matrix.col(1));
	
	//write the harmonic map coordinates in the harmonic map
	for(intixmap_cit vix=ixmap.begin();vix!=ixmap.end();++vix)
	{
	  harmonic[vix->first].first = X[vix->second];
	  harmonic[vix->first].second = Y[vix->second];
	}
}

//*******************************************************************************
void harmonic_mapping::operator()(Trimesh& Mesh)
//*******************************************************************************
{
	/* Initialization section */

	typedef list<Trimesh::vertex_t> border_t;
	typedef border_t::iterator border_it;

	typedef list<border_t> borderregions_t;
	typedef borderregions_t::iterator borderregions_it;

	double radius_factor = 1/patch_radius;

	Trimesh::vertex_t center = 0;
	
	//border sequences
	border_t border;
	border_t convex_border;
	borderregions_t concarve_borders;

	//index map: mapping the interior vertices to line numbers in the matrices
	intixmap_t ixmap;

	//distance map holding the distances between the center and the other vertices
	distmap_t distances;

	//data receiving the harmonic map coordinates
	map_t harmonic_map;

	//find the central vertex and store it in center (set it then to PATCH_INSIDE
	for(Trimesh::vertex_it v=Mesh.begin();v!=Mesh.end();++v)
		if (Mesh.vertKind(*v) == PATCH_CENTER)
		{
			center = *v;
			Mesh.setvertKind(*v,PATCH_INSIDE);
			break;
		}

	//build the distance map
	makedistMap(Mesh,center,distances);

	//compute the spring constants
	calcSprings(Mesh);

	//generate the interior index map
	int interior_num = makeixMap(Mesh,ixmap);
	
	//create the matrices for the lu decomposition and solution of the equations
	RazorBack::Rmatdbl_ A(interior_num,interior_num);
	RazorBack::Rmatdbl_ B(interior_num,2);

	//build the correct sequence of the border
	int border_num = sortBorder(Mesh,back_inserter(border));
	
#ifdef LT_DEBUGGING
  
  for(border_it b=border.begin();b!=border.end();++b)
		cout << *b << "\t";
	cout << endl << endl;

#endif //LT_DEBUGGING

	//build the sequence of the convex border
	getConvexBorder(border.begin(),border.end(),back_inserter(convex_border),Mesh);
	
	//build the sequences of the concarve border pieces
	border_it start = border.begin();
	while(start!=border.end())
	{
		border_t concarve;
		if (!getNextConcarveRegion(start,border.end(),back_inserter(concarve),Mesh))
		{
			if (concarve.size()>0)
				concarve.push_back(*border.begin());

			start = border.end();
		}

#ifdef LT_DEBUGGING
		
		for(border_it b=concarve.begin();b!=concarve.end();++b)
			cout << *b << "\t";
		cout << endl;

#endif //LT_DEBUGGING
	
		//add the new concarve region to the concarve border regions
		if (concarve.size()>0)
			concarve_borders.push_back(concarve);
	}

  /* MAPPING */

	//mapping of the convex border

	/**
	 * use the whole border as a convex border
	 */

	mapConvexBorder(border.begin(), border.end(),
									Mesh,center,
									harmonic_map);
	
	//mapping all concarve border regions

// 	for(borderregions_it b=concarve_borders.begin();b!=concarve_borders.end();++b)
// 		mapConcarveBorder(b->begin(), b->end(), Mesh, center, distances,
// 											harmonic_map, radius_factor);
	
	//convert the polar coordinates into cartesians before interior mapping
	transformCoordinates(harmonic_map);
	
	//build the matrices A and B
	fillMatrices(A, B, harmonic_map, Mesh, ixmap);
	
	//mapping of the interior region of the patch
	mapInterior(A, B, harmonic_map, ixmap);
	
	
	/* generating output */
	
	//rewrite the coordinates
	resetCoords(Mesh, harmonic_map);
}

//*******************************************************************************
Trimesh::vertex_t make_map(Trimesh& Mesh, const double radius)
//*******************************************************************************
{
	harmonic_mapping mapper(radius);
	mapper(Mesh);
	return mapper.getZeroVertex();
}
