/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef MAPPING_HH
#define MAPPING_HH

#include "trimesh.hh"
#include <Rectmat.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <functional>
#    include <list>
#  else
#    include <function.h>
#    include <list.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <functional>
#  include <list>
#endif //HAVE_CONFIG_H

#ifndef PI
#define PI 3.14159265358979
#endif

/**
 *  \brief This is a functor which performs the harmonic mapping
 */
class harmonic_mapping: public unary_function<Trimesh&, void>
{
private:

	typedef pair<double,double> mappoint_t;
	typedef map<Trimesh::vertex_t,mappoint_t> map_t;
	typedef map_t::iterator map_it;

	typedef map<Trimesh::vertex_t,double> distmap_t;
	typedef distmap_t::iterator distmap_it;
	
	typedef map<Trimesh::vertex_t,int> intixmap_t;
	typedef intixmap_t::iterator intixmap_it;
	typedef intixmap_t::const_iterator intixmap_cit;

	double patch_radius;

	/**
	 * \brief stores the vertex that is posed at R=1 and phi=0
	 */
	Trimesh::vertex_t fZeroVertex;

  /**
	 * \brief converting polar in cartesian coordinates
	 * 
	 * \param harmonic map of the 2D unit disc coordinates.
	 */
	void transformCoordinates(map_t& harmonic);

  /**
	 * \brief sorts the border in the right way
	 * This method extracts all the border vertices into a back_inserter
	 * The list allways starts with a convex border vertex!
	 *
	 * \param mesh reference to the triangulated mesh
	 * \param ins insert iterator into a standard container
	 */
	template<class Out> size_t sortBorder(Trimesh& mesh, Out ins);

  /**
	 * \brief calculated the spring constants for every edge: 
	 *
	 * At the moment 1/edge_length is taken as a spring constant.
	 * The value is written into the edge_weight property
	 *
	 * \param mesh reference to the triangulated mesh
	 */
	void calcSprings(Trimesh& mesh);

  /**
	 * \brief retrieves the convex border vertices out of a list of border vertices
	 *
	 * \param begin iterator pointing to the first vertex on the border
	 * \param end iterator pointing to the last vertex on the border 
	 * \param ins insert iterator receiving the concarve vertices
	 * \param mesh reference to the triangulated mesh
	 */
	template<class In, class Out>
	int getConvexBorder(In begin, In end, Out ins, Trimesh& mesh);

  /**
	 * \brief retrieves a concarve border region
	 *
	 * retrieves a sequence of concarve border vertices out of a list of
	 * border vertices. The first and the last vertex of the sequence are
	 * the adjacent convex border vertices.
	 * 
	 * After the call begin contains the position after the last added vertex
	 * If the function returns false the last vertex in the border sequence was
	 * reached, before the tailing convex vertex could be added. In this case the
	 * first vertex of the whole border sequence has to be added at the end of the
	 * sequence.
	 *
	 * \param begin iterator pointing to the first unchecked vertex on the border
	 * \param end iterator pointing to the last vertex on the border 
	 * \param ins insert iterator receiving the concarve vertices
	 * \param mesh reference to the triangulated mesh
	 */
	template<class In, class Out> 
	bool getNextConcarveRegion(In& begin, In end, Out ins, Trimesh& mesh);

  /** 
	 * \brief mapping of the convex boundary vertices.
	 * 
	 * The convex boundary vertices are mapped on the radius at different angles
	 * 
	 * \param begin bi- iterator pointing to the begin of the convex border seq.
	 * \param end  bi- iterator pointing to the end of the convex border seq.
	 * \param harmonic map containing the 2D coordinates of every vertex v. 
	 * \param radius radius of the unit disk (default = 1)
	 * \param mesh reference to the triangulated mesh
	 * \param center central vertex of the patch
	 */
	template<class In> void mapConvexBorder(In begin, In end, 
																					Trimesh& mesh, Trimesh::vertex_t center, 
																					map_t& harmonic, const double radius=1);

  /** 
	 * \brief mapping of a concarve border region
	 * 
	 * The concarve vertices of a concarve border region are mapped on the interior
	 * of the unit disc.
	 * 
	 * \param begin bi- iterator pointing to the begin of the concarve border seq.
	 * \param end  bi- iterator pointing to the end of the concarve border seq.
	 * \param mesh reference to the triangulated mesh
	 * \param center central vertex of the patch
	 * \param distance the distance map that contains the distance information 
	 *                 between the vertices
	 * \param harmonic map containing the 2D coordinates of every vertex v. 
	 * \param radius_factor the radius factor
	 */
	template<class In>void mapConcarveBorder(In begin, In end, 
																					 Trimesh& mesh, Trimesh::vertex_t center,
																					 distmap_t& distance,
																					 map_t& harmonic, 
																					 const double radius_factor);

  /**
	 * \brief computation of matrices A and b
	 *
	 * The matrices A and b are computed according to the lab-book No 11394 p. 9-11
	 *
	 * \param A_matrix reference to matrix A
	 * \param B_matrix reference to matrix b
	 * \param harmonic map containing the 2D coordinates of every vertex v. The
	 *        coordinates are written in polar form and have to be transformed
	 *        before the interior mapping.
	 * \param mesh reference to the triangulated mesh
	 * \param ixmap container mapping the vertex indices to matrix indices
	 */
	void fillMatrices(RazorBack::Rmatdbl_& A_matrix, RazorBack::Rmatdbl_& B_matrix,	
										map_t& harmonic, Trimesh& mesh, intixmap_t& ixmap);

	/**
	 * \brief generates the index map
	 *
	 * Indices in the matrices A and b refer to internal vertices only. makeixMap
	 * maps the vertex indices of the internal part to the matrix indices
	 *
	 * \param mesh reference to the triangulated mesh
	 * \param ixmap container mapping the vertex indices to matrix indices
	 */
	int makeixMap(Trimesh& mesh, intixmap_t& ixmap);

	/**
	 * \brief generates a distance map from the central vertex to any other vertex
	 *
	 * \param mesh reference to the triangulated mesh
	 * \param center central vertex of the patch
	 * \param distance container receiving the distances for each patch vertex
	 */
	void makedistMap(Trimesh& mesh, Trimesh::vertex_t center, distmap_t& distance);

		
	/**
	 * \brief performs the interior mapping
	 *
	 * \param A_matrix reference to matrix A
	 * \param B_matrix reference to matrix b
	 * \param harmonic map containing the 2D coordinates of every vertex v. 
	 * \param ixmap container mapping the vertex indices to matrix indices
	 */
	void mapInterior(RazorBack::Rmatdbl_& A_matrix, RazorBack::Rmatdbl_& B_matrix,
									 map_t& harmonic, const intixmap_t& ixmap);

	/**
	 * \brief Sets the coordinates of the vertex to the coordinates of the harmonic
	 * map
	 *
	 * \param harmonic map containing the 2D coordinates of every vertex v. 
	 * \param mesh reference to the triangulated mesh
	 */
	void resetCoords(Trimesh& mesh, map_t& harmonic);
	

public:
	
	harmonic_mapping(double radius): patch_radius(radius) { }
	
	void operator()(Trimesh& Mesh);

	/**
	 * \brief returns the zero border vertex
	 *
	 * The zero vertex is the border vertex at r=1.0 and phi=0.0
	 *
	 * \return the surface point (vertex) which is mapped to r=2.0 and phi=0.0
	 */
  Trimesh::vertex_t getZeroVertex() const { return fZeroVertex; }
	
};

#ifdef INCLUDE_TMPL_DEFS
  #include "mapping_tmpl.cc"
#endif

/**
 * \brief shortcut function to perform a harmonic mapping
 *
 * \param Mesh reference to a triangulated mesh, which shoudl be transformed
 *        into a harmonic map.
 * \param radius radius of the surface patch, represented by the harmonic map
 *
 * \return the zero border vertex
 */
extern Trimesh::vertex_t make_map(Trimesh& Mesh, const double radius);

#endif //MAPPING_HH
