/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef MAPPING_TMPL_CC
#define MAPPING_TMPL_CC

//*******************************************************************************
template<class Out> 
size_t harmonic_mapping::sortBorder(Trimesh& mesh, Out ins)
//*******************************************************************************
{
  Trimesh::vertex_t first_vertex;
	Trimesh::vertex_t act_vertex;
	
	int c = 0;

	//find first border vertex
	for(Trimesh::vertex_it v=mesh.begin();v!=mesh.end();++v)
	  if (mesh.vertKind(*v)==PATCH_CONVEXBORDER) 
		{
			first_vertex = *v;
			act_vertex = *v;
			break;
		}

	//add the first border vertex
	do
	{
		*ins++ = act_vertex;
		++c;
		
		//look for the next border vertex in all faces of next to act_vertex
		for(Trimesh::FaceIxIter fix=mesh.beginFaceIx(act_vertex);fix!=mesh.endFaceIx(act_vertex);++fix)
		{
			const Triangle& f = mesh.getFace(*fix);
			
			Trimesh::vertex_t suc=f.successor(act_vertex);
			if ((mesh.vertKind(suc)==PATCH_CONVEXBORDER) || (mesh.vertKind(suc)==PATCH_CONCARVEBORDER))
			{
				list<int> dummy;
				mesh.commonTriangles(suc,act_vertex, back_inserter(dummy));
				if (dummy.size()==1)
				{
					act_vertex = suc;
					break;
				}
			}
		}
	}
	while (act_vertex!=first_vertex);

	return c;

}

//*******************************************************************************
template<class In, class Out>
int harmonic_mapping::getConvexBorder(In begin, In end, Out ins, Trimesh& mesh)
//*******************************************************************************
{
	int c =0;
	
	while (begin!=end)
	{
		if (mesh.vertKind(*begin)==PATCH_CONVEXBORDER) 
		{
			*ins++ = *begin;
			++c;
		}
		
		++begin;
	}
	
	return c;
}

//*******************************************************************************
template<class In, class Out>
bool harmonic_mapping::getNextConcarveRegion(In& begin, In end, 
																						 Out ins, Trimesh& mesh)
//*******************************************************************************
{
	enum { searching, adding, finishing, finished} status = searching;
	
	while ((begin!=end) && (status!=finished))
		switch (status)
		{
			case searching:
			{
				if (mesh.vertKind(*begin)==PATCH_CONCARVEBORDER)
				{
				  *ins++ = *(--begin);
					status = adding;
				}
				
				++begin;
				break;
			}
			
			case adding:
			{
				if (mesh.vertKind(*begin)==PATCH_CONCARVEBORDER)
				{
					*ins++ = *begin;
					++begin;
				}
				else
					status = finishing;

				break;
			}

			case finishing:
			{
				*ins++ = *begin++;
				
			  status = finished;
				break;
			}
		}
	
	return (status==finished);
}

//*******************************************************************************
template<class In>
void harmonic_mapping::mapConvexBorder(In begin, In end, 
																			 Trimesh& mesh, Trimesh::vertex_t center, 
																			 map_t& harmonic, const double radius)
//*******************************************************************************
{
	In real_end = end;
	--real_end;

	In real_begin = begin;
	real_begin++;

	//vector of angles between two convex boundary vertices and the central vertex
	list<double> angles;
	//sum over all angles
	double angle_sum = 0;
	
	//position of the central vertex
	Vector_<double> center_pos = mesh.vertCoordinates(center);

	//calculating the angles on the surface
	for(In i=begin;i!=end;)
		{
			Trimesh::vertex_t v1 = *i;
			Trimesh::vertex_t v2;
			if (++i==end)	v2 = *(begin); else v2 = *i;

	
			Vector_<double> v1_pos = mesh.vertCoordinates(v1);
			Vector_<double> v2_pos = mesh.vertCoordinates(v2);
		
			angle_sum += calculate_angle(v2_pos,center_pos,v1_pos);
			angles.push_back(angle_sum);
		}
	
	//add the first angle = 0
	angles.push_front(0);
	//the first border vertex is the zero border
	fZeroVertex = *begin;
	
	//calculating the angles on the map
	list<double>::iterator a = angles.begin();
	for(In i=begin;i!=end;++i)
	{
		harmonic[*i].first = radius;
		harmonic[*i].second = *(a++)/angle_sum * 2*PI;
	}
}

//*******************************************************************************
template<class In>
void harmonic_mapping::mapConcarveBorder(In begin, In end, 
																				 Trimesh& mesh, Trimesh::vertex_t center, 
																				 distmap_t& distance,
																				 map_t& harmonic, 
																				 const double radius_factor)
//*******************************************************************************
{
	In real_end = end;
	--real_end;

	In real_begin = begin;
	real_begin++;

	int count  = 0;
	for(In i=begin;i!=end;++i) ++count;

	//vector of angles between two convex boundary vertices and the central vertex
	vector<double> angles(count-1);
	//sum over all angles
	double angle_sum = 0;

	//determine the first angle of the convex sequence as an offset
	double offset = harmonic[*begin].second;
	
	//calculate the vector between the first and the last vertex (which are actually
	//vertices on the convex border)
	double convex_angle = harmonic[*(real_end)].second -	harmonic[*begin].second;
	
	if (convex_angle<0) convex_angle = 2*PI + convex_angle;
	
	//position of the central vertex
	Vector_<double> center_pos = mesh.vertCoordinates(center);

	//calculating the angles on the surface
	vector<double>::iterator a = angles.begin();
	for(In i=begin;i!=(real_end);)
	{
		Trimesh::vertex_t v1 = *i;
		Trimesh::vertex_t v2 = *(++i);
	
		Vector_<double> v1_pos = mesh.vertCoordinates(v1);
		Vector_<double> v2_pos = mesh.vertCoordinates(v2);
		
		angle_sum += calculate_angle(v2_pos,center_pos,v1_pos);
		*a++ = angle_sum;
	}
	
	//calculating the angles on the map for all the concarve vertices
	a = angles.begin();
	for(In i=real_begin;i!=(real_end);++i)
	{
		harmonic[*i].first = distance[*i]*radius_factor;
		harmonic[*i].second = offset + (*(a++)/angle_sum * convex_angle);
	}
}

#endif //MAPPING_TMPL_CC
