/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef MSMSTRIMESH_CC
#define MSMSTRIMESH_CC

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <fstream>
#    include <sstream>
#    include <cstdio>
#    include <iomanip>
#    include <algorithm>
#    include <map>
#  else
#    include <fstream.h>
#    include <sstream.h>
#    include <cstdio.h>
#    include <iomanip.h>
#    include <algo.h>
#    include <map.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <fstream>
#  include <sstream>
#  include <cstdio>
#  include <iomanip>
#  include <algorithm>
#  include <map>
#endif //HAVE_CONFIG_H

#include "msmstrimesh.hh"

//******************************************************************************
bool msmsTrimesh::read(const string& aFile)
//******************************************************************************
{
  ifstream stream(aFile.c_str());
  if (!stream)
  {
    cerr << "File: " << aFile << " not found" << endl;
    return false;
  }
  
  bool result = read(stream);  
  stream.close();

  return result;
}

//******************************************************************************
bool msmsTrimesh::read(const string& aVertFile,
			  const string& aFaceFile)
//******************************************************************************
{
  ifstream vertfile(aVertFile.c_str());
  if (!vertfile)
  {
    cerr << "Vertex File: " << aVertFile << " not found" << endl;
    return false;
  }
  
  bool result = readVertex(vertfile);
  vertfile.close();

  if (result)
  {
    ifstream facefile(aFaceFile.c_str());
    if (!facefile)
    {
      cerr << "Face File: " << aFaceFile << " not found" << endl;
      return false;
    }
  
    result = readFace(facefile);
    facefile.close();
  }
  
  return result;
}

//******************************************************************************
bool msmsTrimesh::read(istream& aStream)
//******************************************************************************
{
  bool result = readVertex(aStream);
  if (result) result = readFace(aStream);
  
  return result;
}


//******************************************************************************
bool msmsTrimesh::readVertex(istream& aStream)
//******************************************************************************
{
  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

  char line_buffer[200];

  // number of vertices, number of spheres, probe radius and density
  int nvertex;

  Vector_<double> xyz(3);  // Coordinates
  Vector_<double> nxyz(3); // Normals

  // closest_sphere and face of the vertices
  int face_num;
  int closest_sphere;
  int face;
  
  Trimesh::vertex_t v;

  //read comment header lines
  aStream.getline(line_buffer,200);
  aStream.getline(line_buffer,200);

  // Read third header line
  aStream.getline(line_buffer,200);
  stringstream string_buffer(line_buffer);

  // number of vertices, number of spheres, probe radius and density
  string_buffer >> nvertex;  
  int ix = 0;
  for (int i=0;i<nvertex;i++)
  {
    // Read coordinates, normals, closest_sphere and face of the vertices
    aStream.getline(line_buffer,200);
    stringstream string_buffer(line_buffer);

    string_buffer >> xyz[0] >> xyz[1] >> xyz[2] >> nxyz[0] >> nxyz[1] >> nxyz[2]
		  >> face_num >> closest_sphere >> face;

    //checking for doubletts
    int doublett;
    if (removeDoubletts)
      doublett = mesh_structure->getDoublett(xyz);
    else
      doublett = -1;

    // if a doublett is found set a reference to it if not set a selfreference
    if (doublett==-1) 
      doublett = ix;
    else
      cout << "Found doublett : " << i << " " << doublett << endl;
    doublettix[i] = doublett;

    //if the vertex is unique add him to the list
    if (doublett == ix)
    {
      v = add_vertex(Mesh);

      // Store vertex coordinates and normals and the other flags

			// Store vertex coordinates and normals and the other flags
			vertexdata_t data;
			data.masterIndex = ix;
			data.ownIndex    = ix;
			data.kind        = PATCH_INSIDE;
			data.coordinates = xyz;
			data.normal      = nxyz;
			data.cSphere     = closest_sphere;
			data.face_num    = face_num;
			data.face_flag   = face;

			put(*data_map, v, data);

      ix++;
    }
  }
  
  return true;
}


//******************************************************************************
bool msmsTrimesh::readFace(istream& aStream)
//******************************************************************************
{
  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

  int ntriangle;
  int nsphere2;
  double tri_density;
  double probe_sphere;
  Triangle atriangle;  // a triangle for the triangle list

  char line_buffer[200];


  aStream.getline(line_buffer,200);
  aStream.getline(line_buffer,200);

  // Read number of triangles, number of spheres, triangulation density and
  // probe sphere
  aStream.getline(line_buffer,200);

  stringstream string_buffer_face(line_buffer);

  string_buffer_face >> ntriangle;

  //Resize triangle list

  // Read triangles
  for (int i=0;i<ntriangle;i++)
  {

    aStream.getline(line_buffer,200);

    stringstream string_buffer(line_buffer);

    int triangle_face, face_number;
    int nvertices = num_vertices(Mesh);
    int triangle[3];

    string_buffer >> triangle[0] >> triangle[1] >> triangle[2]
									>> triangle_face >> face_number;

    //triangle index starts from zero
    for(int j=0;j<3;++j) triangle[j]--;

    // Check if vertices of the triangle exists, than check if edges are not
    // already added to the graph than add edge to the graph

    if (doublettix.find(triangle[0]) == doublettix.end())
    {
      cerr << "Error: Vertex " << triangle[0] << " in triangle " << i+1
					 << " has no coordinates!" << endl;
      return false;
    }
    if (doublettix.find(triangle[1]) == doublettix.end())
    {
      cerr << "Error: Vertex " << triangle[1] << " in triangle " << i+1
					 << " has no coordinates!" << endl;
      return false;
    }
    if (doublettix.find(triangle[2]) == doublettix.end())
    {
      cerr << "Error: Vertex " << triangle[2] << " in triangle " << i+1
					 << " has no coordinates!" << endl;
      return false;
    }

    //seting the triangle indices to the unique vertices
    for(int j=0;j<3;++j)
      triangle[j] = doublettix[triangle[j]];

    bool degenerated = false;
    for(int j=0;j<3;++j)
    {
      for(int k=j+1;k<3;++k)
				if (triangle[j]==triangle[k])
				{
					degenerated = true;
					break;
				}
      if (degenerated) break;
    }
      
    if (!degenerated)
    {
			// Get the nth vertex in the graph's vertex list.
			Trimesh::vertex_t v1 = vertex(triangle[0],Mesh);
			Trimesh::vertex_t v2 = vertex(triangle[1],Mesh);
			Trimesh::vertex_t v3 = vertex(triangle[2],Mesh);

			//create the triangle object and add it to the trianglelist
			Triangle atriangle(edgepoint(triangle[0],&((*data_map)[v1].coordinates)),
												 edgepoint(triangle[1],&((*data_map)[v2].coordinates)),
												 edgepoint(triangle[2],&((*data_map)[v3].coordinates)));

      //adding the face flag and the face number
      atriangle.setFaceFlag(triangle_face);
      atriangle.setFaceNum(face_number);

      Triangles.push_back(atriangle);
      //adding the index of the face to the vertfaceIndex entries of all three
      //vertices
      for(int j=0;j<=2;++j)
				vertfaceIndex[triangle[j]].push_back(i);


      //check if any edges are allready there
      bool has_edge1 = edge(v1,v2, Mesh).second;
      bool has_edge2 = edge(v2,v3, Mesh).second;
      bool has_edge3 = edge(v3,v1, Mesh).second;

			//add every edge, that does not exist yet, and calculate its length
			pair<Trimesh::edge_t, bool> e_b;   //edge, boolean pair (returned from add_edge

			if (!has_edge1)
			{
				e_b = add_edge(v1,v2, Mesh);
				put(*edgelengths_map, e_b.first, 
						calculate_length((*data_map)[v1].coordinates,
														 (*data_map)[v2].coordinates));

			}
			if (!has_edge2)
			{
				e_b = add_edge(v2,v3, Mesh);
				put(*edgelengths_map, e_b.first, 
						calculate_length((*data_map)[v2].coordinates,
														 (*data_map)[v3].coordinates));
			}
			if (!has_edge3)
			{
				e_b = add_edge(v3,v1, Mesh);
				put(*edgelengths_map, e_b.first, 
						calculate_length((*data_map)[v3].coordinates,
														 (*data_map)[v1].coordinates));
			}
	
    }
  }

  return true;  
}

//******************************************************************************
bool msmsTrimesh::write(const string& aFile)
//******************************************************************************
{
  ofstream stream(aFile.c_str());
  if (!stream)
  {
    cerr << "Could not open file: " << aFile << " for output!" << endl;
    return false;
  }
  
  bool result = writeVertex(stream);
  if (result) result = writeFace(stream);
  stream.close();
  
  return result;
}

//******************************************************************************
bool msmsTrimesh::write(ostream& aStream)
//******************************************************************************
{
  bool result = writeVertex(aStream);
  if (result) result = writeFace(aStream);  

  return result;
}


//******************************************************************************
bool msmsTrimesh::write(const string& aVertFile,
			const string& aFaceFile)
//******************************************************************************
{
  ofstream vertfile(aVertFile.c_str());
  if (!vertfile)
  {
    cerr << "Could not open vertex file: " << aVertFile << " for output" << endl;
    return false;
  }
  
  bool result = writeVertex(vertfile);
  vertfile.close();
  
  if (result)
  {
    ofstream facefile(aFaceFile.c_str());
    if (!facefile)
    {
      cerr << "Could not open face file: " << aFaceFile << " for output" << endl;
      return false;
    }
    
    result = writeFace(facefile);
    facefile.close();  
  }
  
  return result;
}


//******************************************************************************
bool msmsTrimesh::writeVertex(ostream& aStream)
//******************************************************************************
{
  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

  char headline[200];

  aStream << endl << endl;
  //#vertex #sphere probe_r density
  sprintf(headline,"%7d %7d %2.3f %2.3f\n",
					num_vertices(Mesh),
					0,0,0);
  aStream << headline;

  for(Trimesh::vertex_it v=mesh_structure->begin();v!=mesh_structure->end();++v)
  { 
    char linebuffer[200];
    
    sprintf(linebuffer, "%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %7d %7d %2d\n",
						(*data_map)[*v].coordinates[0],
						(*data_map)[*v].coordinates[1],
						(*data_map)[*v].coordinates[2],
						(*data_map)[*v].normal[0],
						(*data_map)[*v].normal[1],
						(*data_map)[*v].normal[2],
						(*data_map)[*v].face_num,
						(*data_map)[*v].cSphere,
						(*data_map)[*v].face_flag);
    aStream << linebuffer;
  }
  
  return true;
}  

//******************************************************************************
bool  msmsTrimesh::writeFace(ostream& aStream)
//******************************************************************************
{
  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

  char headline[200];

  aStream << endl << endl;
  // #triangles #spheres probe_radius density
  sprintf(headline, "%7d %7d %2.3f %2.3f\n",
	  Triangles.size(),
	  0,0,0);
  aStream << headline;

  for(Trimesh::FaceIter f=mesh_structure->beginFace();f!=mesh_structure->endFace();++f)
  {
    char linebuffer[200];

    sprintf(linebuffer, "%6d %6d %6d %2d %6d\n",
	    f->getVertexIndex(0)+1,f->getVertexIndex(1)+1,f->getVertexIndex(2)+1,
	    f->getFaceFlag(),
	    f->getFaceNum());
    aStream << linebuffer;
  }

  return true;
}

#endif
