/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef MSMSTRIMESH_HH
#define MSMSTRIMESH_HH

#include "trimesh.hh"
#include "trimeshio.hh"

/**
 * \brief trimesh class, able to save MSMS objects
 *
 * The file format is described on the object format pages under 
 * \ref msms_fileformat "The MSMS File Format".
 */
class msmsTrimesh : public trimeshIO
{
private:
  /**
   * \brief reads an MSMS vertex file into a stream
   */
  bool readVertex(istream& aStream);
  
  /**
   * \brief reads an MSMS face file into a stream
   */
  bool readFace(istream& aStream);

  /**
   * \brief writes an MSMS vertex file into a stream
   */
  bool writeVertex(ostream& aStream);
  
  /**
   * \brief writes an MSMS face file into a stream
   */
  bool writeFace(ostream& aStream);

public:

  /**
   * \brief constructor initializing msmsTrimesh with a Trimesh object
   */
  msmsTrimesh(Trimesh* aTrimesh): trimeshIO(aTrimesh) { }

  /**
   * \brief no stream handling for msms files.
   */
  bool handleStreams() const { return true; }

  /**
   * \brief read a MSMS object from a single stream
   */
  bool read(istream & aStream);

  /**
   * \brief write a MSMS object to a single stream
   */
  bool write(ostream & aStream);

  /**
   * \brief reads a MSMS Object from a single file.
   *
   * The file must contain the contents of the vertex file and the face file (in
   * this order) with one space line between.
   *
   * See also: msmsTrimesh::read(const string&, const string&, const bool)
   *
   */
  bool read(const string& aFile);

  /**
   * \brief reads a MSMS Object from the coordinates file (%.vert) and the
   * face file (%.face).
   *
   * The coordinates and surface normals are stored as internal
   * properties (see documentation of BOOST Graph library). The parameter
   * distances tells the method to generate a distance matrix during the
   * loading of the mesh.
   */
  bool read(const string& aVertFile,
	    const string& aFaceFile);

  /**
   * \brief writes a MSMS Object to a single file starting.
   *
   * The vertex and face file are stored in a single file with one spacing line 
   * between.
   *
   * See also: msmsTrimesh::write(const string&, const string&)
   */
  bool write(const string& aFile);

  /**
   * Writes the represented triangulated mesh into an MSMS coordinate and face
   * file.
   */
  bool write(const string& aVertFile,
	     const string& aFaceFile);
};

  /**
   * \page msms_fileformat The MSMS file formats
   *
   * The MSMS file format is created from Michael Sanners molecular surface tool.
   * It is used by his own molecular surface viewer, which is available on the
   * SGI/IRIX platform.
   *
   * Each MSMS object consists of two files:
   * \li The file containing the vertex coordinates and normals  
   *     \ref vert_file "*.vert File":
   * \li The file containing the face descriptions
   *     \ref face_file "*.face File":
   *
   * \subsection vert_file Format of MSMS Vert file.
   *
   * Format description of coordinates file (by Torsten Schindler):
   *
   * 3 header lines followed by one vertex per line:
   * Each vertex line consists of 
   * \li the coordinates (x,y,z),
   * \li the normals (nx,ny,nz) 
   * \li number of the face to which the vertex belongs.
   * \li (1 based) index of the closest sphere.
   * \li a flag denoting 
   * <dl>
   *   <dt>1</dt> <dd>for vertices which belong to toric reentrant faces
   *            (including vertices of the analytical surface)</dd>
   *   <dt>2</dt> <dd>for vertices inside reentrant faces and </dd>
   *   <dt>3</dt> <dd>forvertices  inside  contact faces</dd>
   * </dl>
   *
   * "%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f %7d %7d %2d"
   *
   * \subsection face_file Format of MSMS face file.
   *
   * Format description of coordinates file (by Torsten Schindler):
   *
   * The face file contains three header lines followed by one triangle per line
   * First header  line: comment
   * Second header line: comment about the content of the third line
   * Third header  line: number of triangles, number of spheres in the set,
   *                     triangulation density and probe sphere radius
   * Triangle data lines:
   * \li first three number (1 based) vertex indices
   * \li next field: 
   *   <dl>
   *   <dt>1</dt><dd>for a triangle in a toric reentrant face</dd>
   *   <dt>2</dt><dd>for a triangle in a spheric reentrant face and</dd>
   *   <dt>3</dt><dd>for a triangle in a contact face</dd>
   *   </dl>
   * \li last number in line: (1 based) face number in the analytical description
   *     of the solvent excluded surface.
   *
   * "%6d %6d %6d %2d %6d"
   */

#endif 
