/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "offtrimesh.hh"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <fstream>
#    include <sstream>
#    include <cstdio>
#    include <iostream>
#    include <iomanip>
#    include <algorithm>
#    include <map>
#  else
#    include <fstream.h>
#    include <sstream.h>
#    include <cstdio.h>
#    include <iostream.h>
#    include <iomanip.h>
#    include <algo.h>
#    include <map.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <fstream>
#  include <sstream>
#  include <cstdio>
#  include <iostream>
#  include <iomanip>
#  include <algorithm>
#  include <map>
#endif //HAVE_CONFIG_H

//******************************************************************************
bool offTrimesh::read(const string& aFile)
//******************************************************************************
{
  //open the File
  ifstream stream(aFile.c_str());
  if (!stream)
  {
    cerr << "Coordinate file: " << aFile << " not found" << endl;
    return false;
  }

  bool result = read(stream);  

  stream.close();
  return result;
}


//******************************************************************************
bool offTrimesh::read(istream& aStream)
//******************************************************************************
{
  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();
  
  char line_buffer[200];

  // number of vertices, number of spheres, probe radius and density
  int nvertex, nfaces, nedges;

  Vector_<double> xyz(3);  // Coordinates
  Vector_<double> nxyz(3); // Normals
  int vertex_kind;        // vertex kind

  bool read_normals = false; //reading the surface normals if available

  Trimesh::vertex_t v;

  // Read header lines
  string fmt_descriptor;

  //first line
  aStream.getline(line_buffer,200);
  stringstream head1_stream(line_buffer);

  head1_stream >> fmt_descriptor;

  int pos = fmt_descriptor.find("OFF");
  if (pos==fmt_descriptor.npos)
  {
    cerr << "Wrong file format! Descriptor is: " << fmt_descriptor << endl;
    return false;    
  }
  //testing for available surface formats
  for(;pos>=0;--pos)
    read_normals = read_normals | (fmt_descriptor[pos]=='N');

  aStream.getline(line_buffer,200);
  stringstream head2_stream(line_buffer);

  // number of vertices, number of faces and number of edges
  head2_stream >> nvertex >> nfaces >> nedges;
  
    
  //read vertices
  for (int i=0;i<nvertex;i++)
  {
    // Read coordinates, normals, closest_sphere and face of the vertices
    aStream.getline(line_buffer,200);
    stringstream string_buffer(line_buffer);

    //read the coordinates
    string_buffer >> xyz[0] >> xyz[1] >> xyz[2];
    
    //read the surface normals
    if (read_normals)
      string_buffer >> nxyz[0] >> nxyz[1] >> nxyz[2];
    else
      fill(nxyz.begin(),nxyz.end(),0);

    //set vertex_kind to 0

    vertex_kind = 0;

    v = add_vertex(Mesh);
    
		// Store vertex coordinates and normals and the other flags
		vertexdata_t data;
		data.masterIndex = i;
		data.ownIndex    = i;
		data.kind        = 0;
		data.coordinates = xyz;
		data.normal      = nxyz;
		data.cSphere    = 0;

    put(*data_map, v, data);
  }

  Triangle atriangle;  // a triangle for the triangle list

  // Read triangles
  for (int i=0;i<nfaces;i++)
	{

		aStream.getline(line_buffer,200);

		stringstream string_buffer(line_buffer);

		int triangle_face, face_number;
		int nvertices = num_vertices(Mesh);
		int triangle[3];
    
		int nfaceVertices = 0;

		string_buffer >> nfaceVertices >> triangle[0] >> triangle[1] >> triangle[2];

		if (nfaceVertices!=3)
		{
			cerr << "Wrong number of vertices in face Nr. " << i+1 << "!" << endl;

			Mesh.clear();
			Triangles.clear();
			return false;
		}

		//triangle index starts from zero
		//for(int j=0;j<3;++j) triangle[j]--;

		// Check if vertices of the triangle exists, than check if edges are not
		// already added to the graph than add edge to the graph

		if ((triangle[0]<0) || (triangle[0]>nvertex))
		{
			cerr << "Error: Vertex " << triangle[0] << " in triangle " << i+1
					 << " has no coordinates!" << endl;

			Mesh.clear();
			Triangles.clear();
			return false;
		}
		if ((triangle[1]<0) || (triangle[1]>nvertex))
		{
			cerr << "Error: Vertex " << triangle[1] << " in triangle " << i+1
					 << " has no coordinates!" << endl;

			Mesh.clear();
			Triangles.clear();
			return false;
		}

		if ((triangle[2]<0) || (triangle[2]>nvertex))
		{
			cerr << "Error: Vertex " << triangle[2] << " in triangle " << i+1
					 << " has no coordinates!" << endl;

			Mesh.clear();
			Triangles.clear();
			return false;
		}

		// Get the nth vertex in the graph's vertex list.
		Trimesh::vertex_t v1 = vertex(triangle[0],Mesh);
		Trimesh::vertex_t v2 = vertex(triangle[1],Mesh);
		Trimesh::vertex_t v3 = vertex(triangle[2],Mesh);

		//create the triangle object and add it to the trianglelist
		Triangle atriangle(edgepoint(triangle[0],&((*data_map)[v1].coordinates)),
											 edgepoint(triangle[1],&((*data_map)[v2].coordinates)),
											 edgepoint(triangle[2],&((*data_map)[v3].coordinates)));

		//adding the face flag and the face number
		atriangle.setFaceFlag(0);
		atriangle.setFaceNum(0);

		Triangles.push_back(atriangle);
		//adding the index of the face to the vertfaceIndex entries of all three
		//vertices
		for(int j=0;j<=2;++j)
			vertfaceIndex[triangle[j]].push_back(i);


		//check if any edges are allready there
		bool has_edge1 = edge(v1,v2, Mesh).second;
		bool has_edge2 = edge(v2,v3, Mesh).second;
		bool has_edge3 = edge(v3,v1, Mesh).second;

		//add every edge, that does not exist yet, and calculate its length
		pair<Trimesh::edge_t, bool> e_b;   //edge, boolean pair (returned from add_edge

		if (!has_edge1)
		{
			e_b = add_edge(v1,v2, Mesh);
			put(*edgelengths_map, e_b.first, 
					calculate_length((*data_map)[v1].coordinates,
													 (*data_map)[v2].coordinates));

		}
		if (!has_edge2)
		{
			e_b = add_edge(v2,v3, Mesh);
			put(*edgelengths_map, e_b.first, 
					calculate_length((*data_map)[v2].coordinates,
													 (*data_map)[v3].coordinates));
		}
		if (!has_edge3)
		{
			e_b = add_edge(v3,v1, Mesh);
			put(*edgelengths_map, e_b.first, 
					calculate_length((*data_map)[v3].coordinates,
													 (*data_map)[v1].coordinates));
		}
	
	}

	//if the surface normals are not included in the off file
	//calculate them
	mesh_structure->calcSurfaceNormals();
	
  return true;
}

//******************************************************************************
bool offTrimesh::write(const string& aFile)
//******************************************************************************
{
  //open the File
  ofstream stream(aFile.c_str());
  if (!stream)
  {
    cerr << "File: " << aFile << " could not be created" << endl;
    return false;
  }

  bool result = write(stream);  

  stream.close();
  return result;
}


//******************************************************************************
bool offTrimesh::write(ostream& aStream)
//******************************************************************************
{
  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

  char headline[200];

  //write format descriptor
  aStream << "NOFF" << endl;
  
  //#vertex #vertices, #faces, #edges
  sprintf(headline,"%7d %7d %7d\n",
					num_vertices(Mesh),
					Triangles.size(),
					num_edges(Mesh));

  aStream << headline;

  //write the vertices
  for(Trimesh::vertex_it v=mesh_structure->begin();v!=mesh_structure->end();++v)
  { 
    char linebuffer[200];
    
    sprintf(linebuffer, "%9.3f %9.3f %9.3f %9.3f %9.3f %9.3f\n",
						(*data_map)[*v].coordinates[0],
						(*data_map)[*v].coordinates[1],
						(*data_map)[*v].coordinates[2],
						(*data_map)[*v].normal[0],
						(*data_map)[*v].normal[1],
						(*data_map)[*v].normal[2]);

    aStream << linebuffer;
  }

  //write the triangles
  for(Trimesh::FaceIter f=mesh_structure->beginFace();f!=mesh_structure->endFace();++f)
  {
    char linebuffer[200];

    sprintf(linebuffer, "3 %6d %6d %6d\n",
						f->getVertexIndex(0),f->getVertexIndex(1),f->getVertexIndex(2),
						f->getFaceFlag(),
						f->getFaceNum());

    aStream << linebuffer;
  }

  return true;
}
