/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef OFFTRIMESH_HH
#define OFFTRIMESH_HH

#include "trimeshio.hh"

/**
 * \brief trimesh class, able to read and write triangulated mesh objects in
 * the OFF object format
 *
 * The \ref offformat_page "OFF object format" is used by 
 * <a href="http://www.geomview.org/">Geomview</a>, a powerfull mesh display 
 * program. It is very closely related to the \ref ownformat_page "own library format" 
 * of this library.
 */

class offTrimesh : public trimeshIO
{
public:

  /**
   * \brief constructor initializing msmsTrimesh with a Trimesh object
   */
  offTrimesh(Trimesh* aTrimesh): trimeshIO(aTrimesh) { }

  /** 
   * \brief there is a stream handling for library own format files.
   */
  bool handleStreams() const { return true; }


  /**
   * \brief reads a surface mesh object in library own format from a file 
   * specified in \p aFile
   *
   */
  bool read(const string& aFile);
  
  /**
   * \brief reads a surface mesh object in library own format from the stream
   * specified in \p aStream
   */
  bool read(istream& aStream);

  /**
   * \brief writes a surface mesh object in library own format into a file 
   * specified in \p aFile
   *
   */
  bool write(const string& aFile);

  /**
   * \brief writes a surface mesh object in library own format into a stream 
   * specified in \p aStream
   *
   */
  bool write(ostream& aStream);
};

#endif 


/**
 * \page offformat_page OFF object format
 *
 * This file format is used by the <a href="http://www.geomview.org/">Geomview</a>
 * viusalization program. A detailed description can therefore be found under:
 * <a href="http://www.geomview.org/docs/html/geomview_41.html#SEC44">OFF Files</a>
 *
 * The following descripton is the peticular form of OFF files, which are created
 * by the writer:
 *
 * \section offfile_header File Header:
 * The file consists of two header lines:
 * \li The first containing the string \c CNOFF: <br>
 * \c C denotes the usage of the color index to state the kind of each point <br>
 * \c N denotes that the vertex normals are saved <br>
 * \c OFF specifies that the off file format is used <br>
 *
 * \li The second line contains the number of vertices, faces and edges:<br>
 * \em num_vertices, \em num_faces, \em num_edges
 * 
 * \section offfile_vertices Vertex Lines:
 * The vertex lines consist of the following data:
 * POS_X POS_Y POS_Z NORM_X NORM_Y NORM_Z VERTEX_STATE
 *
 * \param POS_[XYZ] position of the vertex in 3D coordinates
 * \param NORM_[XYZ] direction of the normal vector at the verex
 *
 * \section offfile_faces Face Lines:
 * The face lines consist of the following data:
 * 3 VERTEX_1 VERTEX_2 VERTEX_3
 *
 * \param 3 count of vertices in a face: since Trimesh handles triangulated
 *        meshes, this must be 3
 * \param VERTEX_[123] indices of the three vertices forming the face.
 *
 * \note The vertices must be given in the correct order. (Looking from on them
 * from the outside of the surface.)
 * 
 */
