/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "patch.hh"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <algorithm>
#    include <vector>
#  else
#    include <map.h>
#    include <algo.h>
#    include <vector.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#    include <map>
#    include <algorithm>
#    include <vector>
#endif //HAVE_CONFIG_H

#include <boost/graph/dijkstra_shortest_paths.hpp>
//#include "../include/convexhull.hh"

//******************************************************************************
surfacePatch::surfacePatch(const Trimesh& trimesh, const vertex_t center,
			   const double radius, const double margin)
//******************************************************************************
:Trimesh(), patchflags(), fstatus(ps_ok), surface_vertices(),
 dist(), fRadius(radius), fCenter(center), lBorder()
{
  generatePatch(trimesh, center, radius, margin);
}

//*******************************************************************************
surfacePatch::surfacePatch(const surfacePatch& patch)
//*******************************************************************************
:Trimesh(patch), patchflags(patch.patchflags), 
 fstatus(patch.fstatus), surface_vertices(patch.surface_vertices),
 dist(patch.dist), fRadius(patch.fRadius), fCenter(patch.fCenter), 
 lBorder(lBorder)
{
	
}

//******************************************************************************
void surfacePatch::generatePatch(const Trimesh& trimesh, const vertex_t center,
				 const double radius, const double margin)
//******************************************************************************
{
	fRadius = radius;
	fCenter = center;

  typedef vector<patchflag>::const_iterator patch_it;
  vector<patchflag> patch(trimesh.nVertices(),patch_none);

  //prepare the batch
  preparePatch(patch, trimesh, center, radius, margin);

  //removing dangling triangles
  removeDanglings(trimesh, patch);

  //fill all holes
  closeHoles(trimesh, patch);

  //fill the triangle list
  getTriangles(trimesh, patch);
 
  ownindexmap_t ownindices;
  for(vertex_it v=trimesh.begin();v!=trimesh.end();++v)
    if (patch[*v]!=patch_none)
    {
      //copy the vertex
      ownindexpair_t ixpair = copyVertex(trimesh,*v);
      //add the vertex to the ownindex map
      ownindices[ixpair.first] = ixpair.second;
      //copy the edges
      copyEdges(trimesh,*v,ixpair.second,ownindices);
      patchflags[ixpair.second] = patch[*v];

			//add the vertex to the surface_vertices
			surface_vertices.push_back(*v);
    }

  //renumbering the face indices
  for(vector<Triangle>::iterator f=Triangles.begin();f!=Triangles.end();++f)
    for(int i=0;i<3;++i)
    {
      ownindexmap_t::iterator ix = ownindices.find(f->getVertexIndex(i));
      if (ix==ownindices.end())
	cerr << f->getVertexIndex(i) << " not available \n";
      else
	f->setVertexIndex(i,ix->second);
    }

  //set the ownindex to the vertex index and build the patchflags and patchBorder
  //map
  for(vertex_it v=begin();v!=end();++v)
  {
    if ((patchflags[*v]!=patch_inside) && (patchflags[*v]!=patch_center))
			//generate the border list
			lBorder.push_back(*v);

		data_map[*v].kind = patchflags[*v];
		data_map[*v].ownIndex = *v;
  }

  //rebuilding Vertex-Face index
  rebuildVertexFaceIx();
  
  //copy simple properties
  copyProperties(trimesh);
}

//******************************************************************************
void surfacePatch::preparePatch(vector<patchflag>& patch, 
				const Trimesh& trimesh, const vertex_t center,
				const double radius, const double margin)
//******************************************************************************
{
	int c = 0;

  //vertices for the selection of vertices and faces in the mother mesh.
  vertexlist_t border(0);
 
  //calculating the distances from the central vertex to all the other vertices
  trimesh.calcOnSurfaceDistance(center,dist);
 
  //mark all vertices inside the patch
  for(int i=0;i<trimesh.nVertices();++i)
    if (dist[i]<=radius) 
		{
			patch[i]=patch_inside;
			++c;
		}

	patch[center] = patch_center;

  //look for boundary vertices
  for(vertex_it v=trimesh.begin();v!=trimesh.end();++v)
    //check the adjacency of each patch vertex, if there are non patch vertices
    if (patch[*v]==patch_inside)
    {
      for(adjacency_it a = trimesh.beginAdjacency(*v);a!=trimesh.endAdjacency(*v);++a)
				if (patch[*a]==patch_none)
				{
					patch[*v] = patch_convexborder;
					border.push_back(*v);
					break;
				}
    }

  /**
   * if the user has specified a margin larger than zero, the non-patch neighbors
   * of the boundary vertices are searched for vertices within the margin. These
   * vertices are then added to the patch.
   *
   * Then all the boundary vertices are checked again.
   */ 
  
  if (margin>0.00001)
  {
    int num_new = 0;
    do
    {
      num_new = 0;
      vertexlist_t new_vertices;

      //check for new vertices
      for(vl_cit v=border.begin();v!=border.end();++v)
				for(adjacency_it a = trimesh.beginAdjacency(*v);a!=trimesh.endAdjacency(*v);++a)
					if ((patch[*a]==patch_none) && (dist[*a]<(radius*(1+margin))))
					{
            //new vertices are always at the border of the patch
						patch[*a] = patch_convexborder;
						new_vertices.push_back(*a);
						++num_new;
					}
      
      //add the new vertices to the border list
      copy(new_vertices.begin(),new_vertices.end(),back_inserter(border));
      
      //check all border vertices, if they are still at the border
      size_t i=0;
      while (i<border.size())
      {
				vertex_t v = border[i];
				patch[v] = patch_inside;
 
				for(adjacency_it a = trimesh.beginAdjacency(v);a!=trimesh.endAdjacency(v);++a)
					if (patch[*a]==patch_none)
					{
						patch[v] = patch_convexborder;
						break;
					}
				if (patch[v] == patch_inside)
					border.erase(find(border.begin(),border.end(),v));
				else 
					++i;
      }

    } while (num_new);
  }
}

//******************************************************************************
void surfacePatch::closeHoles(const Trimesh& trimesh, vector<patchflag>& patch)
//******************************************************************************
/**
 * THIS PROCEDURE HAS TO BE REWRITTEN !
 */
{
  typedef list<vertex_t> border_t;
  typedef border_t::iterator border_it;
  typedef list<border_t*> borderlist_t;
  typedef borderlist_t::iterator borderlist_it;

  //list of all border lines
  borderlist_t borderlist;

  //create a list of all border vertices
  border_t whole_border;
  
  for(vertex_it v=trimesh.begin();v!=trimesh.end();++v)
    if (patch[*v]==patch_convexborder)
    {
      whole_border.push_back(*v);
    }

  //for each border line (a connected line of border vertices create a new border
  //and remove the vertex from the whole border
  
  while (whole_border.size()>0)
  {

    border_t* border = new border_t();
    border_it next_v = whole_border.begin(); //iterator to the next vertex on the border
    vertex_t act_v = *(whole_border.begin());  //copy of the active  vertex

    //walk along the actual border line
    do
    {
      //put the new active vertex on the 
			vertex_t prev_v = (border->size()>0) ? border->back() : trimesh.nVertices()+1;
      border->push_back(act_v);
      
      //remove the actual vertex from the whole border list
      whole_border.erase(next_v);

      for(adjacency_it a = trimesh.beginAdjacency(act_v);a!=trimesh.endAdjacency(act_v);++a)
      {
				//find the next vertex on the border line
				next_v = find(whole_border.begin(),whole_border.end(),*a);
				if (next_v!=whole_border.end())
				{
					//initialization of common face search
					list<int> common_faces;
					trimesh.commonTriangles(*next_v, act_v, back_inserter(common_faces));

					list<int> vertices;
					
					for(list<int>::const_iterator i=common_faces.begin();i!=common_faces.end();++i)
					{
						const Triangle& face = trimesh.getFace(*i);
						for(int j=0;j<3;++j)
							if ((face.getVertexIndex(j)!=act_v) && (face.getVertexIndex(j)!=*next_v))
								if ((patch[face.getVertexIndex(j)]!=patch_none) && 
										(face.getVertexIndex(j)!=prev_v))
									vertices.push_back(face.getVertexIndex(j));
					}

					if (vertices.size()==1)
					{
						//set the active vertex
						act_v = *next_v;
						break;
					}
				}
      }
			//check if the border is complete or the border is fault
    } while ((next_v!=whole_border.end()) && (border->size()<=whole_border.size()));
		
    borderlist.push_back(border);

		if (border->size()>whole_border.size()) fstatus = ps_faultborder;
  }

  //if the amount of borderlines is larger than 1 there are holes
  if ((borderlist.size()>1) && (fstatus = ps_ok))
  {
    /* 
       CLOSE HOLES ALGORITHM
    */

    //detect largest borderline -> whould be real patch border
    int maxv = 0;
    borderlist_it largest_b = borderlist.end();
    for(borderlist_it b=borderlist.begin();b!=borderlist.end();++b)
      if ((*b)->size()==maxv) 
				fstatus = ps_largeholes;
      else if ((*b)->size()>maxv) 
      {
				maxv = (*b)->size();
				largest_b = b;
				fstatus = ps_ok;
      }

		if (fstatus!=ps_largeholes)
		{
    
			//delete the largest border from the borderlist -> becomes a holelist
			delete *largest_b;
			borderlist.erase(largest_b);

			//the patched vertices are counted to warn the user, if more vertices are
			//added to the patch than the actual number of vertices
    
			int old_vertices = count(patch.begin(),patch.end(),patch_none);
			old_vertices = trimesh.nVertices() - old_vertices;

			//close holes
			for(borderlist_it i=borderlist.begin();i!=borderlist.end();++i)
			{
				int num_newvertices = 0;
      
				border_t* b = *i;

				set<vertex_t> newverts;

				for(border_it j=b->begin();j!=b->end();++j)
				{
					vertex_t border_v=*j;

					//reset the status of the border vertex to inside
					patch[border_v] = patch_inside;

					//search for a vertex which is not contained in the patch and adjacent
					//to *border_v
					for(adjacency_it a =trimesh.beginAdjacency(border_v);a!=trimesh.endAdjacency(border_v);++a)
					{
						if (patch[*a] == patch_none)
						{
							b->push_back(*a);
							newverts.insert(*a);
							++num_newvertices;
						}

						if (num_newvertices>old_vertices) 
						{
							fstatus = ps_largeholes;
							break;
						}
					}

					if (fstatus!=ps_largeholes)
					{
						set<vertex_t>::const_iterator s;
						for(s=newverts.begin();s!=newverts.end();++s)	
							patch[*s] = patch_inside;
					}
				}
			}
		}
  }

  //delete the borders in the border list
  while (borderlist.size()>0)
  {
    borderlist_it b = borderlist.begin();
    border_t* ptr_b = *b;
    borderlist.erase(b);
		ptr_b->clear();
    delete ptr_b;
  }
  
}

//******************************************************************************
void surfacePatch::removeDanglings(const Trimesh& trimesh, vector<patchflag>& patch)
//******************************************************************************
{
  int nremoved;

  do 
  {
    nremoved = 0;
   
    for(vertex_it v=trimesh.begin();v!=trimesh.end();++v)
      if (patch[*v]!=patch_none)
      {
	int count_patch = 0;
	for(adjacency_it a = trimesh.beginAdjacency(*v);a!=trimesh.endAdjacency(*v);++a)
	  if (patch[*a]!=patch_none) ++count_patch;
      
	if (count_patch<3)
	{
	  patch[*v]=patch_none;
	  ++nremoved;
	}
      }
  } while (nremoved);

  for(vertex_it v=trimesh.begin();v!=trimesh.end();++v)
    //check the adjacency of each patch vertex, if there are non patch vertices
    if (patch[*v]==patch_inside)
    {
      for(adjacency_it a = trimesh.beginAdjacency(*v);a!=trimesh.endAdjacency(*v);++a)
	if (patch[*a]==patch_none)
	{
	  patch[*v] = patch_convexborder;
	  break;
	}
    }

}

//******************************************************************************
void surfacePatch::getTriangles(const Trimesh& trimesh, 
				const vector<patchflag>& patch)
//******************************************************************************
{
  //check the triangles
  int faceix = 0;
  for (FaceIter f=trimesh.beginFace();f!=trimesh.endFace();++f)
    if ((patch[f->getVertexIndex(0)]!=patch_none) &&
				(patch[f->getVertexIndex(1)]!=patch_none) &&
				(patch[f->getVertexIndex(2)]!=patch_none))
      Triangles.push_back(*f);
}

// //******************************************************************************
// Trimesh* surfacePatch::investigateBorder(const bool getHull)
// //******************************************************************************
// {
// 	try
// 	{
// 		Convexhull hull(*this);
		
// 		for(vertex_it v=begin();v!=end();++v)
// 			if (patchflags[*v]==patch_convexborder)	patchflags[*v] = patch_concarveborder;

// 		for(Convexhull::hullvertex_cit v=hull.const_begin();v!=hull.const_end();++v)
// 			if (patchflags[*v]==patch_concarveborder)
// 				patchflags[*v] = patch_convexborder;

// 		for(vertex_it v=begin();v!=end();++v)
// 			data_map[*v].kind = patchflags[*v];

// 		if (getHull) 
// 			return hull.getMesh();
// 		else
// 			return (Trimesh*) 0;
// 	}
// 	catch (conntab::inconsistent_structure)
// 	{
// 		for(vertex_it v=begin();v!=end();++v)
// 			patchflags[*v] = patch_convexborder;
		
// 		return (Trimesh*) 0;
// 	}
  
// }
