/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef PATCH_HH
#define PATCH_HH

#include "trimesh.hh"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <list>
#  else
#    include <list.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <list>
#endif //HAVE_CONFIG_H

/**
 * \brief A representation of surface patches
 *
 * The surfacePatch class is a c++ class for surface patches on triangulated 
 * meshes. It is based on the Trimesh class with additional features for patch
 * creation and management
 */

class surfacePatch : public Trimesh
{
private:

	//*****************************************************************************
	// PRIVATE TYPE DEFINITIONS
	//*****************************************************************************

  /**
   * \brief vertexlist typedefinitions
   */
  typedef vector<vertex_t> vertexlist_t;

public:

	//*****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//*****************************************************************************

  /**
   * \brief enumeration type for the patch flag
   *
   * The patch flag indicates if a vertex is the member of a surface patch or
   * not. If the vertex is a member, it contains also information of the vertex's
   * status in the patch (if it is at the convex or concarve boundary or in the 
   * inner regionsof the patch
   */
  enum patchflag {patch_none = 0,
		  patch_inside = 1,
		  patch_convexborder = 2,
 		  patch_concarveborder = 3,
	    patch_center = 4};

  enum status    {ps_ok, ps_largeholes, ps_faultborder};

	/**
	 * \brief surface vertex list iterator, representing the selected vertices on
	 * the surface.
	 */
  typedef vertexlist_t::const_iterator vl_cit  ;


private:

	//*****************************************************************************
	// PRIVATE MEMBERS
	//*****************************************************************************

  /**
   * \brief property map for the patchflags of the vertices
   */
  map<vertex_t,patchflag> patchflags;

  /**
   * \brief the overall condition of the patch
   */
  status fstatus;

	/**
	 * \brief list of selected vertices from the surface
	 */
	vertexlist_t surface_vertices;

	/**
	 * \brief distance vector containing the on-surface distances between all
	 * vertices on the mother surface and the central vertex
	 */
  vector<double> dist;

	/**
	 * \brief stores the patch radius of the surface patch
	 */
	double fRadius;
	
	/**
	 * \brief stores the central vertex of the surface patch
	 */
	vertex_t fCenter;

	/**
	 * \brief stores the border vertices
	 */
	list<Trimesh::vertex_t> lBorder;

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

  /**
   * \brief determines the patch-vertices
   *
   * The vertices, which are included in the surface patch are selected from the
   * set of vertices in the parent mesh \c trimesh. The parameter \c vertexlist
   * receives a list of the status of all vertices in \c trimesh relative to the
   * surface patch (not included, included, at the border)
   *
   * \param trimesh referrence to a Trimesh object, that contains the entire 
   *        surface
   * \param center  central vertex of the patch on the entire surface
   * \param radius on surface radius of the patch
   * \param margin tolerance margin: vertices who are lying within the radius
   *        plus the margin and are adjacent to the boundary of the patch are
   *        included.
   * \param patch the correct flags for each point on the surface
   */
  void preparePatch(vector<patchflag>& patch, 
		    const Trimesh& trimesh, const vertex_t center,
		    const double radius, const double margin = 0.05 );

  /**
   * \brief detects holes in the patch and closes them
   *
   * \param trimesh referrence to a Trimesh object, that contains the entire 
   *        surface
   * \param patch marked list of surface vertices (see also
   */
  void closeHoles(const Trimesh& trimesh, vector<patchflag>& patch);

  /**
   * \brief copies all triangles from \c trimesh, that are members of the patch
   * into the Triangle list.
   *
   * To copy the right triangles from the mother surface \c trimesh, the \c patch
   * must be generated by the PreparePatch method
   *
   * \param trimesh referrence to a Trimesh object, that contains the entire 
   *        surface
   * \param patch marked list of surface vertices (see also
   * surfacePatch::PreparePatch)
   */
  void getTriangles(const Trimesh& trimesh, const vector<patchflag>& patch);
  
  /**
   * \brief removes tangling triangles
   *
   * Dangling triangles are triangles, that have one or more edgepoints whith
   * less than 3 neighbors. Usually dashing triangles should be removed, because
   * they destroy smooth and closed patches.
   */
  void removeDanglings(const Trimesh& trimesh, vector<patchflag>& patch);

public:
	
	//*****************************************************************************
	// CONSTRUCTORS
	//*****************************************************************************

	/**
   * \brief constructor for surface patches
   *
   * The constructore generates a surface patch of the triangulated Mesh 
   * \c trimesh with the central vertex \c center and the the "on-surface" radius
   * \c r. To obtain smoother patches you can specify a tolerance margin relative
   * to the patch radius \c margin (the default value is 0.05).
   */
  surfacePatch(const Trimesh& trimesh, const vertex_t center,
	       const double radius, const double margin = 0.05);

  /**
   * \brief standard constructor
   */
  surfacePatch()
			:Trimesh(), patchflags(), lBorder(), 
			 fstatus(ps_ok), surface_vertices(), dist(), fRadius(0.0), fCenter(0) { }

	/**
	 * \brief copy constructor
	 */
	surfacePatch(const surfacePatch& patch);

	//*****************************************************************************
	// PUBLIC METHODS operating on the patch
	//*****************************************************************************
  
  /**
   * \brief generates a surface patch based on the surface mesh \c trimesh.
   * 
   * This method reinitilizes the object and generates a surface patch based on
   * he \p trimesh surface mesh. The central vertex of the patch is the vertex
   * \p center (a vertex of the \p trimesh). The "on surface" diameter of the 
   * surface is twice the \p radius plus vertices, which are within a certain
   * \c margin of the border are included as well.
   *
   * \note The patch is smoothened by the margin parameter and all its dangling
   * triangles are removed by the method, but the boundary is not searched fo
   * concarve regions, and the patchflag for all the boundary values is
   * \c patch_convexborder.
   *
   * \param trimesh referrence to a Trimesh object, that contains the entire 
   *        surface
   * \param center  central vertex of the patch on the entire surface
   * \param radius on surface radius of the patch
   * \param margin tolerance margin: vertices who are lying within the radius
   *        plus the margin and are adjacent to the boundary of the patch are
   *        included.
   */
  void generatePatch(const Trimesh& trimesh, const vertex_t center,
		     const double radius, const double margin = 0.05);

//   /**
//    * \brief investigates the convex hull of the patch to determine the radius
//    * and interior border (convex/concarve boundary vertices
// 	 *
// 	 * This method sets the vertex kind of the border to either CONCARVE or CONVEX,
// 	 * and returns the convex hull of the patch if \p getHull is \c true.
// 	 *
// 	 * \param getHull if \c true the triangulated mesh of the convex hull is
// 	 *        returned.
//    */
//   Trimesh* investigateBorder(const bool getHull);

	//*****************************************************************************
	// PUBLIC PROPERTY METHODS
	//*****************************************************************************

	/**
	 * \brief retrieves the on-surface distances for all vertices on the mother
	 * surface to the central vertex
	 *
	 * This method has been inserted to avoid double calculations of the distance 
	 * matrix of trimesh objects.
	 */
	const vector<double>& central_distances() const { return dist; }

  /**
   * \brief retrieves the status of the surface patch
   */
  status getStatus() const {  return fstatus;  }

	/**
	 * \brief iterator pointing to the first selected vertex on the surface.
	 */
	vl_cit surfvertices_begin() const { return surface_vertices.begin(); }

	/**
	 * \brief iterator pointing behind the last selected vertex on the surface.
	 */
	vl_cit surfvertices_end() const { return surface_vertices.end(); }

	/**
	 * \brief retrieves the patch radius
	 */
	double getRadius() const { return fRadius; }

	/**
	 * \brief retrieves the central vertex
	 */
	vertex_t getCenter() const { return fCenter; }

	//@{
	///\name border iterators
	/**
	 * \brief border begin end end iterators
	 *
	 * \return the begin or end of the border list
	 */
	list<Trimesh::vertex_t>::const_iterator borderBegin() const { return lBorder.begin(); }
	list<Trimesh::vertex_t>::const_iterator borderEnd() const { return lBorder.end(); }
	//@}



};

#endif
