/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "patchmaker.hh"
#include "trimesh_algorithms.hh"
#include <float.h>
#include <math.h>

//*******************************************************************************
patchMaker::patchMaker(const int v_distance, const double r, 
											 const Trimesh* surface_obj)
//*******************************************************************************
:vertex_distance(v_distance), radius(r), surface(surface_obj),
 central_vertices(), patches()
{
	//check the parameters:
	if (vertex_distance<-1) vertex_distance = 0;
	if (radius < 0) radius *= -1;

	//call the find centers routine
  //if (surface) findCenters();

	patches_valid = true;
}

//*******************************************************************************
surfacePatch::vl_cit patchMaker::vlistBegin(int cv)
//*******************************************************************************
{
	Trimesh::vertex_t  v = surface->getVertex(cv);
	return vlistBegin(v);
}

//*******************************************************************************
surfacePatch::vl_cit patchMaker::vlistEnd(int cv)
//*******************************************************************************
{
	Trimesh::vertex_t  v = surface->getVertex(cv);
	return vlistEnd(v);
}

//*******************************************************************************
surfacePatch::vl_cit patchMaker::vlistBegin(Trimesh::vertex_t cv)
//*******************************************************************************
{
	if (central_vertices.find(cv)!=central_vertices.end())
		throw cv_notinlist();

	if ((patches.find(cv)==patches.end()) || !patches_valid)
	{
		surfacePatch* new_patch = new surfacePatch(*surface, cv, radius);
		patches[cv] = new_patch;
	}
	
	return patches[cv]->surfvertices_begin();
}

//*******************************************************************************
surfacePatch::vl_cit patchMaker::vlistEnd(Trimesh::vertex_t cv)
//*******************************************************************************
{
	if (central_vertices.find(cv)!=central_vertices.end())
		throw cv_notinlist();

	if ((patches.find(cv)==patches.end()) || !patches_valid)
	{
		surfacePatch* new_patch = new surfacePatch(*surface, cv, radius);
		patches[cv] = new_patch;
	}
	
	return patches[cv]->surfvertices_end();
}

//*******************************************************************************
size_t patchMaker::addVertex(Trimesh::vertex_t v)
//*******************************************************************************
{
	central_vertices.insert(v);
	return central_vertices.size();
}

//*******************************************************************************
size_t patchMaker::addVertex(int v)
//*******************************************************************************
{
	Trimesh::vertex_t  vert = surface->getVertex(v);
	return addVertex(vert);
}

//*******************************************************************************
surfacePatch* patchMaker::patch(const int cv, const bool add_vertex)
//*******************************************************************************
{
	Trimesh::vertex_t  v = surface->getVertex(cv);
	return patch(v, add_vertex);
}
	
//*******************************************************************************
surfacePatch* patchMaker::patch(const Trimesh::vertex_t cv, const bool add_vertex)
//*******************************************************************************
{
	//look in the central vertex list and add the vertex if demanded
	if (central_vertices.find(cv)==central_vertices.end()) 
		if (add_vertex) 
			central_vertices.insert(cv);	
		else 
			return (surfacePatch*) 0;

	if ((patches.find(cv)!=patches.end()) && !patches_valid)
		delete patches[cv];
	
	//get the patch (create it if necessary)
	if ((patches.find(cv)==patches.end()) || !patches_valid)
	{
		surfacePatch* new_patch = new surfacePatch(*surface, cv, radius);
		patches[cv] = new_patch;
		return patches[cv];
	}
	else 
		return patches[cv];

}

//*******************************************************************************
int patchMaker::set_vertexDistance(const int v_distance)
//*******************************************************************************
{
	int old = vertex_distance;
	
	if ((vertex_distance != v_distance) &&  (vertex_distance>=-1))
	{
		vertex_distance = v_distance;
		centers_valid = false;
	}

	return old;
}

//*******************************************************************************
double patchMaker::set_radius(const double r)
//*******************************************************************************
{
	double old = radius;

	if ((radius-r)<DBL_EPSILON)
	{
		radius = r;
		patches_valid = false;
	}

	return old;
}

//*******************************************************************************
const Trimesh* patchMaker::set_surface(const Trimesh* surface_obj)
//*******************************************************************************
{
	const Trimesh* old = surface;

	if (surface!=surface_obj)
	{
		surface = surface_obj;
		centers_valid = false;
		patches_valid = false;
	}

	return old;
}

//*******************************************************************************
void patchMaker::clearPatches()
//*******************************************************************************
{ 
	for(patches_it i=patches.begin();i!=patches.end();++i)
		delete (*i).second;
	patches.clear(); 
}

//*******************************************************************************
int patchMaker::calc_all()
//*******************************************************************************
{
	clearPatches();
	for(centers_cit i=central_vertices.begin();i!=central_vertices.end();++i)
	{
		surfacePatch* new_patch = new surfacePatch(*surface, *i, radius);
		patches[*i] = new_patch;
	}
	
	patches_valid = true;
}

//*******************************************************************************
const vector<double>& patchMaker::distances(int cv)
//*******************************************************************************
{
	Trimesh::vertex_t  v = surface->getVertex(cv);
	return distances(v);	
}

//*******************************************************************************
const vector<double>& patchMaker::distances(Trimesh::vertex_t cv)
//*******************************************************************************
{
	if (central_vertices.find(cv)==central_vertices.end())
		throw cv_notinlist();

	
	if (patches.find(cv)==patches.end())
	{
		surfacePatch* new_patch = new surfacePatch(*surface, cv, radius);
		patches[cv] = new_patch;
	}
	
	return patches[cv]->central_distances();	
}

//*******************************************************************************
int patchMaker::findCenters(int v_distance)
//*******************************************************************************
{
	if (!surface) return -1;
	
	// get the v_distance from the property vertex_distance
	if ((v_distance!=-10) && (v_distance>0)) vertex_distance =v_distance;

	//a vertex_distance of 0 prohibits the finding of suitable central vertices
	if (vertex_distance == 0) return 0;
	
	//if the vertex distance is -1 the minimal vertex distance is set to the radius
	//length
	if (vertex_distance == -1)
	{
		//calculate the average edge length and find out the optimal distance for the
		//dispersion
			 
		double sum_d = sum_of_weights(const_cast<Trimesh*>(surface)->Mesh,0.0);
		double av_weight = sum_d/surface->nEdges();
				
		//the optimal distance is the number of edges that represent in average the 
		//length of the radius of the patches
		vertex_distance  = (int) (radius/av_weight);
	}

	int nv = 0;
	// 1 is a special case and can be calculated very fast
	if (vertex_distance == 1)
	{
		central_vertices.insert(surface->begin(),surface->end());
		nv = central_vertices.size();
	}
	else
	{
		//create the set of central vertices and return the number of selected 
		//vertices
		list<Trimesh::vertex_t> vlist;
		Trimesh::vertex_t start = vertex(0,surface->Mesh);
		//assign the connectivity graph

		//look for disperse vertices
		nv = disperseVertices(const_cast<Trimesh*>(surface)->Mesh, start, vertex_distance, 
													back_inserter(vlist));

		//copy the vertices in the central_vertices set
		central_vertices.insert(vlist.begin(),vlist.end());
	}
	
	return nv;
}
