/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef PATCHMAKER_HH
#define PATCHMAKER_HH

#include "trimesh.hh"
#include "patch.hh"
#include "patchgraph.hh"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <set>
#  else
#    include <map.h>
#    include <set.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#    include <map>
#    include <set>
#endif //HAVE_CONFIG_H

/**
 * \brief patchMaker creates and manages a set of vertices, which could be used
 * as the centers of surface patches
 *
 * Only the patch vertices are generated during the Initialization the patches 
 * itself can be retrieved by the methods patchMaker::all_patches(),
 * patchMaker::nearest_patch() and patchMaker::patch_at().
 */
class patchMaker
{
public:
	
	//*****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//*****************************************************************************

	/**
	 * \brief type for the set of vertices
	 */
	typedef set<Trimesh::vertex_t> centers_t;

	/**
	 * \brief type of an iterator and constant iterator for the central vertex set.
	 */
	typedef centers_t::iterator centers_it;
	typedef centers_t::const_iterator centers_cit;


	/**
	 * \brief type for the patch map
	 */
	typedef map<Trimesh::vertex_t, surfacePatch*> patches_t;

	/**
	 * \brief type of an iterator and constant iterator for the patch map.
	 */
	typedef patches_t::iterator patches_it;
	typedef patches_t::const_iterator patches_cit;

	//*****************************************************************************
	// EXCEPTIONS
	//*****************************************************************************
	
	/**
	 * \brief this exception is thrown when a patch from a central vertex not in
	 * the central vertex list is demanded
	 *
	 * This exception is thrown by the vlist and distances methods but not by the 
	 * patchMaker::patch() methods, where the central vertex can be added to the 
	 * list of central vertices.
	 */
	class cv_notinlist { };

private:

	//*****************************************************************************
	// PRIVATE MEMBERS
	//*****************************************************************************
	
	/**
	 * \brief denotes the minimum vertex distance
	 *
	 * Any positive number \em n indicates, that the central vertices are at least
	 * \em n edges away from each other. Nevertheless there are 2 special values
	 * for the minimal vertex distance:
	 *
	 * \li \em 0, no minimal vertex distance is specified, the vertices in  central
	 *     vertices are userdefined.
	 * \li \em -1, the vertices are selected so, that the neighbors of one central
	 *     vertex is roughly placed at the radius of these patch. After the
	 *     patchMaker::findCenters method has been invoked the value is updated to
	 *     the actual minimal vertex distance calculated by the average edge length
	 *     together with the patch radius.
	 */
	int vertex_distance; 
	
	/**
	 * \brief denotes the patch radius in surface length units
	 */
	double radius;

	/**
	 * \brief pointer to the Trimesh structure representing the surface.
	 */
	const Trimesh* surface;

	/**
	 * \brief flag indicating, if the central vertices are valid or not.
	 *
	 * The central vertices become invalid, if the surface is changed. If the min.
	 * \c vertex_distance is changed, the central vertices are still valid, because
	 * they represent valid vertices on the underlying surface.
	 */
	bool centers_valid;

	/**
	 * \brief flag indicating, if the allready created surface patches are valid
	 * or not.
	 *
	 * The allready created surface patches (the patches available in the 
	 * patchMaker::patches map) become invalid, if the patch radius, or the surface
	 * object have been changed.
	 * 
	 * \note If the patches are invalid, and you create a new surface patch by 
	 * request, this patch is valid, because he is generated with either the new
	 * radius or the new surface object. 
	 */
	bool patches_valid;
	
	/**
	 * \brief set of selected central vertices
	 */
	centers_t central_vertices;

	/**
	 * \brief map of surface patches.
	 *
	 * The key value is the index of the central vertex in the surface vertex list.
	 */
	patches_t patches;

	/**
	 * \brief graph containing the topological information about the patches on the
	 * surface.
	 */
	patchgraph_traits::patchgraph_t topology;

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

public:
	//*****************************************************************************
	// CONSTRUCTORS & DESTRUCTORS
	//*****************************************************************************

	/**
	 * \brief standard constructor initializing the patchMaker with invalid values
	 */
	patchMaker(): vertex_distance(0), radius(0), surface(0), central_vertices(), 
								patches() { centers_valid = patches_valid = false; }

	~patchMaker() { clearPatches(); }

	/**
	 * \brief Preferred constructor, intitilizing and allready computing the set of
	 * central vertices.
	 *
	 * \param v_distance minimum vertex distancem (used to compute the central
	 *        vertex list
	 * \param r radius of the created surface patches
	 * \param surface_obj pointer to a Trimesh structure, referencing the surface 
	 *        object.
	 */
	patchMaker(const int v_distance, const double r, const Trimesh* surface_obj);

	
	//*****************************************************************************
	// ITERATOR METHODS
	//*****************************************************************************

	//CENTRAL VERTICES

	/**
	 * \brief iterator pointing to the first central vertex
	 */
	centers_it centerBegin() { return central_vertices.begin(); }

	/**
	 * \brief constant iterator pointing to the first central vertex
	 */
	centers_cit centerBegin() const { return central_vertices.begin(); }

	/**
	 * \brief iterator pointing behind the last central vertex
	 */
	centers_it centerEnd() { return central_vertices.end(); }

	/**
	 * \brief constant iterator pointing behind the last central vertex
	 */
	centers_cit centerEnd() const { return central_vertices.end(); }

	//PATCHES

	/**
	 * \brief iterator pointing to the map entry of the first created patch
	 */
	patches_it patchBegin() { return patches.begin(); }
	
	/**
	 * \brief constant iterator pointing to the map entry of the first created patch
	 */
	patches_cit patchBegin() const { return patches.begin(); }

	/**
	 * \brief iterator pointing to the map entry behind the last created patch
	 */
	patches_it patchEnd() { return patches.end(); }

	/**
	 * \brief constant iterator pointing to the map entry behind the last created 
	 * patch
	 */
	patches_cit patchEnd() const { return patches.end(); }

	//PATCH VERTEX LISTS

  /**
	 * \brief returns a constant iterator pointing to the first surface vertex of
	 * the patch with the central vertex index \p cv
	 *
	 * \param cv index of the patch's central vertex
	 * \exception cv_notinlist is thrown if the vertex \p cv is not in the central
	 *            vertex list.
	 */
	surfacePatch::vl_cit vlistBegin(int cv);
	
	/**
	 * \brief returns a constant iterator pointing behind the last surface vertex 
	 * ofthe patch with the central vertex index \p cv
	 *
	 * \param cv index of the patch's central vertex
	 * \exception cv_notinlist is thrown if the vertex \p cv is not in the central
	 *            vertex list.
	 */
	surfacePatch::vl_cit vlistEnd(int cv);
	
  /**
	 * \brief returns a constant iterator pointing to the first surface vertex of
	 * the patch with the central vertex \p cv
	 *
	 * \param cv the patch's central vertex
	 * \exception cv_notinlist is thrown if the vertex \p cv is not in the central
	 *            vertex list.
	 */
	surfacePatch::vl_cit vlistBegin(Trimesh::vertex_t cv);
	
	/**
	 * \brief returns a constant iterator pointing behind the last surface vertex 
	 * ofthe patch with the central vertex \p cv
	 *
	 * \param cv the patch's central vertex
	 * \exception cv_notinlist is thrown if the vertex \p cv is not in the central
	 *            vertex list.
	 */
	surfacePatch::vl_cit vlistEnd(Trimesh::vertex_t cv);
	
  //*****************************************************************************
	// DATA METHODS
	//*****************************************************************************

	/**
	 * \brief adds the vertex \p v to the list of central vertices
	 *
	 * It does not matter if the vertex \p v is allready a member of the central
	 * vertex list or not. Even an allready created patch for this vertex is 
	 * preserved
	 *
	 * The function returns the new number of central vertices
	 * 
	 * \param v vertex to be added into the central vertex list
	 */
	size_t addVertex(Trimesh::vertex_t v);

	/**
	 * \brief adds the vertex with index \p v to the list of central vertices
	 *
	 * The behaviour is the same as patchMaker::addVertex(Trimesh::vertex_t v).
	 * If the index \p v is invalid no vertex is added to the list, and the 
	 * method returns -1.
	 * 
	 * \param v vertex index to be added into the central vertex list
	 */
	size_t addVertex(int v);
	
	/**
	 * \brief returns a pointer to the patch associated with the vertex index \p cv.
	 * 
	 * If the \p cv is not in the list of central vertices \c NULL is returned, if
	 * the second parameter \p add_vertex is set to \c false (default value).
	 *
	 * \param cv index of the desired central vertex
	 * \param add_vertex \c true effects that the vertex is added to the central
	 *        vertex list, and the patch is created, even if the vertex is not
	 *        contained in the list of central vertices.
	 */
	surfacePatch* patch(const int cv, const bool add_vertex = false);

	/**
	 * \brief returns a pointer to the patch associated with the vertex \p cv.
	 *
	 * see patchMaker::patch(const int, const bool) for further details
	 *
	 * \param cv a desired central vertex on the surface object
	 * \param add_vertex \c true effects that the vertex is added to the central
	 *        vertex list, and the patch is created, even if the vertex is not
	 *        contained in the list of central vertices.
	 */
	surfacePatch* patch(const Trimesh::vertex_t cv, const bool add_vertex = false);

	//*****************************************************************************
	// PROPERTY METHODS
	//*****************************************************************************

	// VERTEX_DISTANCE
	
	/**
	 * \brief retrieves the minimum vertex distance for the central vertices.
	 */
	int get_vertexDistance() const { return vertex_distance; }

	/**
	 * \brief sets the minimum vertex distance for the central vertices.
	 *
	 * 
	 * Returns the old value of \p vertex_distance
	 * 
	 * \param v_distance new value for the vertex distance. Valid values are equal
	 *        or larger than zero. (See also: patchMaker::vertex_distance)
	 */
	int set_vertexDistance(const int v_distance);

	// RADIUS

	/**
	 * \brief retrieves the actual patch radius
	 */
	double get_radius() const { return radius; }

	/**
	 * \brief sets the patch radius to a new value
	 *
	 * Returns the old patch radius
	 *
	 * \param r value for the new radius. Any positive, non-zero double is valid.
	 */
	double set_radius(const double r);


	// SURFACE

	/**
	 * \brief retrieves the pointer to the underlying surface object
	 */
	const Trimesh* get_surface() const { return surface; }
	
	/**
	 * \brief sets the underlying surface object to a new object
	 *
	 * Returns a pointer to the old surface object 
	 *
	 * \param surface_obj pointer to a new surface object
	 */
	const Trimesh* set_surface(const Trimesh* surface_obj);

	//CENTER VALIDITY

	/**
	 * \brief denotes if the list of central vertices is valid
	 */
	bool isCenterValid() const { return centers_valid; }

	//PATCH VALIDITY
	
	/**
	 * \brief denotes if the list of patches is valid
	 */
	bool isPatchValid() const { return patches_valid; }

	//CENTRAL VERTEX DISTANCE VECTORS
	
	/**
	 * \brief retrieves the distances from the central vertex with index \p cv 
	 * to all the other vertices on the mother surface
	 *
	 * See patchMaker::distances(Trimesh::vertex_t cv) for further details
	 *
	 * \param cv index of a central vertex
	 * \exception cv_notinlist is thrown if the vertex \p cv is not in the central
	 *            vertex list.
	 */
	const vector<double>& distances(int cv);

	/**
	 * \brief retrieves the distances from the central vertex  \p cv to all the 
	 * other vertices on the mother surface
	 *
	 * In order to retrieve the distances the patch has to be created, because a 
	 * distance calculation is done at the beginning of the patching process. This
	 * seems to be a little overhead, but in most operations, the patches are
	 * needed anyway.
	 *
	 * \param cv index of a central vertex
	 * \exception cv_notinlist is thrown if the vertex \p cv is not in the central
	 *            vertex list.
	 */
	const vector<double>& distances(Trimesh::vertex_t cv);
	

	//*****************************************************************************
	// PUBLIC METHODS (performing operations on the patch set)
	//*****************************************************************************

	/**
	 * \brief deletes all vertices in the central vertex set.
	 *
	 * The patches are not deleted.
	 */
	void clearCenters() { central_vertices.clear(); }

	/**
	 * \brief deletes all the patches in the patch map.
	 *
	 * The central vertices remain in the patchMaker::central_vertices list.
	 */
	void clearPatches();

	/**
	 * \brief clears the patches as well as the central vertices
	 */
	void clear() { patches.clear(); central_vertices.clear(); }

	/**
	 * \brief recalculates the patches for all the vertices in the central vertex 
	 * list.
	 *
	 * \note All existing patches are cleared and recalculated.
	 */
	int calc_all();

	/**
	 * \brief creates a list of surface vertices which are at least more than
	 * \c vertex_distace vertices away from each other.
	 *
	 * The vertices are added to the list of central vertices, hence all the 
	 * vertices and the corresponding patches allready contained in the list are 
	 * preserved.
	 *
	 * If you want to use a different minimal vertex distance, you can specify it
	 * with the v_distance parameter, which will set the 
	 * patchMaker::vertex_distance parameter automatically to the new value.
	 *
	 * The method returns the number of vertices identified as centers for surface
	 * patches. If no surface is assigned to the patchMaker object, \em -1 is
	 * returned.
	 *
	 * \param v_distance new minimal vertex distance
	 */
	int findCenters(int v_distance = -10);

	//*****************************************************************************
	// PUBLIC OPERATOR METHODS
	//*****************************************************************************

	/**
	 * \brief retrieves the surface patch for the central vertex with index \p cv
	 *
	 * The method returns the surface patch for the \p cv. central vertex only if
	 * the central vertex at index \p cv is a member of the list of central 
	 * vertices, otherwise zero is returned
	 * 
	 * The access operator has the same functionality as the patchmaker::patch() 
	 * methods with the \p add_vertex parameter set to false
	 */
	surfacePatch* operator[](int cv) { return patch(cv, false); }

	/**
	 * \brief retrieves the surface patch for the central vertex \p cv
	 *
	 * The method returns the surface patch for the central vertex \p cv only if
	 * the central vertex at is a member of the list of central vertices, 
	 * otherwise zero is returned
	 * 
	 * The access operator has the same functionality as the patchmaker::patch() 
	 * methods with the \p add_vertex parameter set to false
	 */
	surfacePatch* operator[](Trimesh::vertex_t cv) { return patch(cv, false); }

};

#endif //PATCHMAKER_HH

