/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "povtrimesh.hh"

#include <float.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif //HAVE_CONFIG_H

#ifdef HAVE_STD
#  include <fstream>
#  include <sstream>
#  include <cstdio>
#  include <iostream>
#  include <iomanip>
#  include <algorithm>
#else
#  include <fstream.h>
#  include <sstream.h>
#  include <cstdio.h>
#  include <iostream.h>
#  include <iomanip.h>
#  include <algo.h>
#  include <map.h>
#endif //HAVE_STD

#ifdef HAVE_NAMESPACES
using namespace std;
#endif

#define RAZORBACK_LIB_COMPILE
#include <Distrapprox.hh>
#undef RAZORBACK_LIB_COMPILE

//*******************************************************************************
void standardize(Vector_<double>& data, const unsigned int range = 1)
//*******************************************************************************
{
	double mean = 0.0;
	double stdev = 0.0;

	int n = data.dim();
	vector<double>::iterator i;

	//mean +stddev
	for(i=data.begin();i!=data.end();++i) 
	{
		mean += *i/n;
		stdev += (*i) * (*i)/n;
	}
	stdev -= mean*mean;
	stdev = sqrt(stdev);

	//standardize to mean=0 stdev=1
	for(i=data.begin();i!=data.end();++i) 
	{
	  *i = (*i-mean)/(stdev*range*2) +0.5;

		//clamp between 0 and 1
		if (*i<0) *i = 0;
		if (*i>1) *i = 1;
	}
}

//*******************************************************************************
void cdfying(Vector_<double>& data, const unsigned int bins)
//*******************************************************************************
{
	vector<double>::iterator i;
	double min = DBL_MAX;
	double max = -DBL_MAX;

	//calculate the minimum and maximum
	for(i=data.begin();i!=data.end();++i)
	{
		if (*i<min) min=*i;
		if (*i>max) max=*i;
	}

	//make a distribution of all values and calculate the cdf
  Distrapprox_ dist((max-min)/(1.0*bins));
  for(vector<double>::const_iterator i=data.begin();i!=data.end();++i) 
		dist += *i;
  dist.evaluate();

	for(i=data.begin();i!=data.end();++i)
		*i=dist.cdf(*i);
}

//******************************************************************************
bool povTrimesh::write(const string& aFile)
//******************************************************************************
{
  //open the File
  ofstream stream(aFile.c_str(), ios::out|ios::binary);
  if (!stream)
  {
    cerr << "File: " << aFile << " could not be created" << endl;
    return false;
  }

  bool result = write(stream);  

  stream.close();
  return result;
}


//******************************************************************************
bool povTrimesh::write(ostream& aStream)
//******************************************************************************
{
	bool finish;

	calcUVCoords();
	
  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

	aStream << "//*PMRawBegin\n\n";
	
	aStream << "#declare " << fSurfName << "=mesh2" << endl
					<< "{" << endl;

	/* VERTEX VECTORS */
	
	aStream << "  vertex_vectors" << endl
					<< "  {" << endl
					<< "    " << mesh_structure->nVertices() << "," << endl;

	Trimesh::vertex_it v=mesh_structure->begin();
	finish=false;
	while (!finish)
	{
		aStream << "    <";
		for(int i=0;i<3;++i)
			aStream << (*data_map)[*v].coordinates[i] << ((i<2) ? ", " : ">");

		++v;
		finish=(v==mesh_structure->end());

		aStream << (finish ? "" : ",") << endl;
	}
		
	aStream << "  }" <<  endl << endl;

	/* NORMAL VECTORS */

	aStream << "  normal_vectors" << endl
					<< "  {" << endl
					<< "    " << mesh_structure->nVertices() << "," << endl;

	v=mesh_structure->begin();
	finish=false;
	while (!finish)
	{
		aStream << "    <";
		for(int i=0;i<3;++i)
			aStream << (*data_map)[*v].normal[i] << ((i<2) ? ", " : "> ");

		++v;
		finish=(v==mesh_structure->end());

		aStream << (finish ? "" : ",") << endl;
	}
		
	aStream << "  }" <<  endl << endl;

	/* UV VECTORS */

	aStream << "  uv_vectors" << endl
					<< "  {" << endl
					<< "    " << mesh_structure->nVertices() << "," << endl;

	v=mesh_structure->begin();
	finish=false;
	while (!finish)
	{
		double value = vUVCoords[*v];
		if (fabs(value-1.0)<DBL_EPSILON) value = 0.99;
		aStream << "    <" << value << ",0>";

		++v;
		finish=(v==mesh_structure->end());

		aStream << (finish ? "" : ",") << endl;
	}

	aStream << "  }" <<  endl << endl;

	/* FACE INDICES */
	
	aStream << "  face_indices" << endl
					<< "  {" << endl
					<< "    " << mesh_structure->nFaces() << "," << endl;

	Trimesh::FaceIter f=mesh_structure->beginFace();
	finish=false;
	while (!finish)
	{
		aStream << "    <";
 		for (int i=0;i<3;++i)
			aStream << f->getVertexIndex(i) << ((i<2) ? "," : ">");
		
		
		++f;
		finish=(f==mesh_structure->endFace());

		aStream << (finish ? "" : ",") << endl;
	}

	aStream << "  }" <<  endl
					<< "}" << endl << endl;
		
// 	aStream << "#declare " << fSurfName << "=mesh\n" << "{\n";

//   //write the triangles
//   for(Trimesh::FaceIter f=mesh_structure->beginFace();f!=mesh_structure->endFace();++f)
//   {
// 		int points = 3;

// 		aStream << "  smooth_triangle\n  {\n";

// 		int i;
// 		//vertex vectors and normals
// 		for (i=0;i<points;++i)
// 		{
// 			Trimesh::vertex_t v = mesh_structure->getVertex(f->getVertexIndex(i));
			
// 			aStream << "    <";
// 			for(int j=0;j<3;++j)
// 				aStream << (*data_map)[v].coordinates[j] << ((j<2) ? ", " : ">, ");
// 			aStream << "<";
// 			for(int j=0;j<3;++j)
// 				aStream << (*data_map)[v].normal[j] << ((j<2) ? ", " : ">");
// 			aStream << ((i<2) ? ",\n" : "\n");
// 		}

// 		aStream << "    uv_vectors ";
// 		//uv vectors
// 		for (i=0;i<points;++i)
// 		{
// 			Trimesh::vertex_t v = mesh_structure->getVertex(f->getVertexIndex(i));

// 			double value = vUVCoords[v];
// 			if (fabs(value-1.0)<DBL_EPSILON)
// 				value = 0.99;
// 			aStream << "<" << value << ",0>" << ((i<points-1) ? ", " : "\n");
// 		}

// 		aStream << "  }\n";
//   }

	aStream << "\n//*PMRawEnd" << endl;

	
	Vector_<double> center(3,0);
	Vector_<double> max(3,-DBL_MAX);
	Vector_<double> min(3,-DBL_MAX);

	for(Trimesh::vertex_it v=mesh_structure->begin();v!=mesh_structure->end();++v)
	{
			center += (*data_map)[*v].coordinates;
			
			for(int i=0;i<3;++i)
			{
				if (max[i]<(*data_map)[*v].coordinates[i]) 
					max[i] = (*data_map)[*v].coordinates[i];
				if (min[i]>(*data_map)[*v].coordinates[i]) 
					min[i] = (*data_map)[*v].coordinates[i];
			}
	}
	
	center /= mesh_structure->nVertices();
	
	aStream << "// characteristics:\n";

	aStream << "  // Center: <" << center[0]
					<< ", " << center[1]
					<< ", " << center[2] << ">\n";

	aStream << "  // Extensions: <" << max[0] - min[0]
					<< ", " << max[1] - min[1]
					<< ", " << max[2] - min[2] << ">\n";

  return true;
}

//*******************************************************************************
int povTrimesh::setColorColumn(const int col)
//*******************************************************************************
{
	int oldValue = fColorColumn;
	fColorColumn = col;
	return oldValue;
}


//*******************************************************************************
string povTrimesh::setSurfName(const string& surfname)
//*******************************************************************************
{
	string old = fSurfName;
	fSurfName = surfname;
	return old;
}

//*******************************************************************************
povTrimesh::colorMapping_t povTrimesh::setColorMapping(colorMapping_t mapping)
//*******************************************************************************
{
	colorMapping_t old = fColorMapping;
	fColorMapping = mapping;
	return old;
}	

//*******************************************************************************
void povTrimesh::calcUVCoords()
//*******************************************************************************
{
	if ((fData.cno()>0) && (fColorColumn>-1))
	{
		vUVCoords = fData.col(fColorColumn);
		
// 		for(int i=0;i<vUVCoords.dim();++i)
// 			cerr << i << " = " << vUVCoords[i] << endl; 
		
		switch (fColorMapping)
		{
			case cmSimple:      break;
			case cmCDF:         cdfying(vUVCoords, 100); break;
			case cmStandardize: standardize(vUVCoords, 2); break;
 		}
	}
}
	
