/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef POVMESH_H
#define POVMESH_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_STD
#  include <string>
#else
#  include <string>
#endif

#include "trimeshio.hh"
#include <Vector.hh>

/**
 * \brief trimesh class, able to write triangulated mesh objects in
 * the POV-Ray format
 *
 * The \ref povformat_page "POV-Ray format" is suitable for passing the surfaces
 * the povray raytracer
 */

class povTrimesh : public trimeshIO
{
public:

	enum colorMapping_t { cmSimple, cmStandardize, cmCDF };

  /**
   * \brief constructor initializing msmsTrimesh with a Trimesh object
   */
  povTrimesh(Trimesh* aTrimesh): 
			trimeshIO(aTrimesh), vUVCoords(aTrimesh->nVertices(), 0),
			fColorColumn(-1), fColorMapping(cmStandardize), fSurfName("molsurface") { }

  /** 
   * \brief there is a stream handling for library own format files.
   */
  bool handleStreams() const { return true; }


  /**
   * \brief povray meshes can only be written
   *
   */
  bool read(const string& aFile) { return false; }
  
  /**
   * \brief povray meshes can only be written
   *
   */
  bool read(istream& aStream) { return false; }

  /**
   * \brief writes a surface mesh object in POV-Ray format into a file 
   * specified in \p aFile
   *
   */
  bool write(const string& aFile);

  /**
   * \brief writes a surface mesh object in POV-Ray format into the specified \p
   * aStream
   *
   */
  bool write(ostream& aStream);

	//*****************************************************************************
	// PROPERTY METHODS
	//*****************************************************************************

	/**
	 * \brief retrieves the relevant column for the vertex coloring
	 *
	 * \return the index of the matrix column used for vertex coloring
	 */
	int getColorColumn() const { return fColorColumn; }

  /**
	 * \brief sets the relevant color for the vertex coloring
	 *
	 * \return the index of the old relevant column
	 *
	 * \param col new relevant color column.
	 */
	int setColorColumn(const int col);

	/**
	 * \brief retrieves the relevant column for the vertex coloring
	 *
	 * \return the index of the matrix column used for vertex coloring
	 */
	const string& getSurfName() const { return fSurfName; }

  /**
	 * \brief sets the surface name
	 *
	 * \return the old name of the surface
	 *
	 * \param surfname new name of the surface
	 */
	string setSurfName(const string& surfname);

	/**
	 * \brief retrieves the type of the colormapping
	 *
	 * The colormapping is either \c cmStandardize for linear interpolationb
	 * between the smallest and the largest value or cmCDF for a commulated
	 * density function mapping.
	 * 
	 * \return the type of the colormapping
	 */
	colorMapping_t getColorMapping() const { return fColorMapping; }

  /**
	 * \brief sets the type of the color mapping
	 *
	 * \return the old color mapping type
	 *
	 * \param mapping new color mapping.
	 */
	colorMapping_t setColorMapping(colorMapping_t mapping);
	
private:

	//*****************************************************************************
	// PRIVATE METHODS
	//*****************************************************************************

	void  calcUVCoords();

	//*****************************************************************************
	// PRIVATE DATA
	//*****************************************************************************

	/**
	 * \brief denotes the relevant data column for the vertex coloring
	 */
	int fColorColumn ;

	/**
	 * \brief the name of the surface in the povray code
	 */
	string fSurfName;

	/**
	 * \brief specifies the kind of color mapping
	 */
	colorMapping_t fColorMapping;

	/**
	 * \brief the x uv coordinates for all vertices
	 */
	Vector_<double> vUVCoords;
	
};

/**
 * \page povformat_page Povray surface format
 *
 * To be written
 */


#endif // POVMESH_H

