/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include <sldtrimesh.hh>

#include "string.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif //HAVE_CONFIG_H

#ifdef HAVE_STD
#  include <fstream>
#  include <sstream>
#  include <cstdio>
#  include <iostream>
#  include <iomanip>
#  include <algorithm>
#  include <vector>
#else
#  include <fstream.h>
#  include <sstream.h>
#  include <cstdio.h>
#  include <iostream.h>
#  include <iomanip.h>
#  include <algorithm.h>
#  include <map.h>
#  include <vector.h>
#endif //HAVE_STD


#define FILEID -4712
#define REVERSE -1729232897
#define SLD_HASNORMALS 0x00000001
#define SLD_HASATOMS   0x00000002

//******************************************************************************
bool sldTrimesh::read(const string& aFile)
//******************************************************************************
{
  //open the File
  ifstream stream(aFile.c_str(), ios::in|ios::binary);
  if (!stream)
  {
    cerr << "SLD file: " << aFile << " not found" << endl;
    return false;
  }

  bool result = read(stream);  

  stream.close();
  return result;
}

//******************************************************************************
bool sldTrimesh::read(istream& aStream)
	//******************************************************************************
{
	//check the file ID
	int id = 0;
	bool reverse = false;

	readInt(aStream, id);
	reverse = (id == REVERSE);

	if ((id==FILEID) || (reverse))
	{
		int i;
	
		mesh_t& Mesh = getMesh();
		vector<Triangle>& Triangles = getTriangles();
		map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

		map< size_t, size_t> doublettix;
  
		//header variables: header size, size of a triangle, size of a vertex,
    //                  number of trianlges, dots, connections and props.
		//                  t-Flag
		
		int head_size, triangle_size, vertex_size;
		int nvertex, nfaces, nedges, natoms, nprops;
		int t_flag;

		/* Read header information */
		readInt(aStream, head_size, reverse);
		readInt(aStream, triangle_size, reverse);
		readInt(aStream, vertex_size, reverse);
		
		readInt(aStream, nfaces, reverse);
		readInt(aStream, nvertex, reverse);
		readInt(aStream, nedges, reverse);
		readInt(aStream, natoms, reverse);
		readInt(aStream, nprops, reverse);

		readInt(aStream, t_flag, reverse);

		/* Read property name information */
		fColumnHeads.resize(nprops);
		fData.size(nvertex, nprops);
				
		for(i=0;i<nprops;i++)
		{
			char propname[30];
			aStream.read((char*) propname, 30);          //read property names
			char* colon = strchr(propname, ':');         //truncate before the colon
			if (colon) (*colon) = 0;
			
			fColumnHeads[i] = propname;
		}

		/* Read vertices */
		//shift to the vertex entries
		aStream.seekg(head_size+triangle_size * nfaces, ios::beg);

		int ix = 0;
		for (i=0;i<nvertex;i++)
		{
			//vertex information
			Trimesh::vertex_t v;

			float xyz[3];        // coordinates
			float nxyz[3];       // normals
			float* props = new float[nprops]; // properties
			short int csphere;   // index of the closest VdW sphere = atom index
			short int dummy;     // some dummy value as spacer

			readFloatArr(aStream, xyz, 3, reverse);
			if (t_flag & SLD_HASNORMALS) 
				readFloatArr(aStream, nxyz, 3, reverse);
			
			readShort(aStream, csphere, reverse);
			
			if (nprops > 0)
				readFloatArr(aStream, props, nprops, reverse);
			
			//copy coords and normals to Vector_'s
			Vector_<double> xyz_(3);
			Vector_<double> nxyz_(3);
			int j = 0;

			for(j=0;j<3;++j)
			{
				xyz_[j] = xyz[j];
				nxyz_[j] = nxyz[j];
			}

			//checking for doubletts
			int doublett;
			if (removeDoubletts)
				doublett = mesh_structure->getDoublett(xyz_);
			else
				doublett = -1;

			// if a doublett is found set a reference to it if not set a selfreference
			if (doublett==-1) 
				doublett = ix;
			else
				cout << "Found doublett : " << i << " " << doublett << endl;
			doublettix[i] = doublett;

			//if the vertex is unique add him to the list
			if (doublett == ix)
			{
				//copy the data properties to the matrix
				for(j=0;j<nprops;++j)
					fData(ix,j) = props[j];

				//add the vertex
				v = add_vertex(Mesh);
    
				// Store vertex coordinates and normals and the other flags
				vertexdata_t data;
				data.masterIndex = ix;
				data.ownIndex    = ix;
				data.kind        = PATCH_INSIDE;
				data.coordinates = xyz_;
				data.normal      = nxyz_;
				data.cSphere    = csphere;

				put(*data_map, v, data);
				
				++ix;
			}

			delete[] props;
		}

		//shrink the data matrix to its actual size
		fData.size(ix, nprops);

		Triangle atriangle;  // a triangle for the triangle list

		/* Read triangles */
		//shift back to the triangle entries
		aStream.seekg(head_size, ios::beg);

		ix = 0;
		for (i=0;i<nfaces;i++)
		{
			int j;

			int nvertices = num_vertices(Mesh);
			int triangle[3];
    
			readIntArr(aStream, triangle, 3, reverse);
			
			// Check if vertices of the triangle exist, than check if edges are not
			// already added to the graph than add edge to the graph

			if (doublettix.find(triangle[0]) == doublettix.end())
			{
				cerr << "Error: Vertex " << triangle[0] << " in triangle " << i+1
						 << " has no coordinates!" << endl;

				Mesh.clear();
				Triangles.clear();
				return false;
			}
			if (doublettix.find(triangle[1]) == doublettix.end())
			{
				cerr << "Error: Vertex " << triangle[1] << " in triangle " << i+1
						 << " has no coordinates!" << endl;

				Mesh.clear();
				Triangles.clear();
				return false;
			}

			if (doublettix.find(triangle[2]) == doublettix.end())
			{
				cerr << "Error: Vertex " << triangle[2] << " in triangle " << i+1
						 << " has no coordinates!" << endl;

				Mesh.clear();
				Triangles.clear();
				return false;
			}

			//seting the triangle indices to the unique vertices
			for(int j=0;j<3;++j)
				triangle[j] = doublettix[triangle[j]];

			bool real = true;
			int p=0;
			while ((p<2) && real)
			{
				int q = p+1;
				
				while ((q<3) && real)
					real = real && (triangle[p]!=triangle[q++]);
				++p;
			}
			
			//add the triangle only if all three vertices are different
			if (real)
			{

				// Get the nth vertex in the graph's vertex list.
				Trimesh::vertex_t v1 = vertex(triangle[0],Mesh);
				Trimesh::vertex_t v2 = vertex(triangle[1],Mesh);
				Trimesh::vertex_t v3 = vertex(triangle[2],Mesh);

				//create the triangle object and add it to the trianglelist
				Triangle atriangle(edgepoint(triangle[0],&((*data_map)[v1].coordinates)),
													 edgepoint(triangle[1],&((*data_map)[v2].coordinates)),
													 edgepoint(triangle[2],&((*data_map)[v3].coordinates)));

				Triangles.push_back(atriangle);
				//adding the index of the face to the vertfaceIndex entries of all three
				//vertices
				for(j=0;j<=2;++j)
					vertfaceIndex[triangle[j]].push_back(ix);


				//check if any edges are allready there
				bool has_edge1 = edge(v1,v2, Mesh).second;
				bool has_edge2 = edge(v2,v3, Mesh).second;
				bool has_edge3 = edge(v3,v1, Mesh).second;

				//add every edge, that does not exist yet, and calculate its length
				pair<Trimesh::edge_t, bool> e_b;   //edge, boolean pair (returned from add_edge

				if (!has_edge1)
				{
					e_b = add_edge(v1,v2, Mesh);
					put(*edgelengths_map, e_b.first, 
							calculate_length((*data_map)[v1].coordinates, 
															 (*data_map)[v2].coordinates));

				}
				if (!has_edge2)
				{
					e_b = add_edge(v2,v3, Mesh);
					put(*edgelengths_map, e_b.first, 
							calculate_length((*data_map)[v2].coordinates, 
															 (*data_map)[v3].coordinates));
				}
				if (!has_edge3)
				{
					e_b = add_edge(v3,v1, Mesh);
					put(*edgelengths_map, e_b.first, 
							calculate_length((*data_map)[v3].coordinates, 
															 (*data_map)[v1].coordinates));
				}

				++ix;
			}
		}
		return true;
	}
	else
	{
		cerr << "Wrong binary file format!" << endl;
		return false;
	}
}

//******************************************************************************
bool sldTrimesh::write(const string& aFile)
//******************************************************************************
{
  //open the File
  ofstream stream(aFile.c_str(), ios::out|ios::binary);
  if (!stream)
  {
    cerr << "File: " << aFile << " could not be created" << endl;
    return false;
  }

  bool result = write(stream);  

  stream.close();
  return result;
}


//******************************************************************************
bool sldTrimesh::write(ostream& aStream)
//******************************************************************************
{
	bool reverse = true;
	int i =0;

  mesh_t& Mesh = getMesh();
  vector<Triangle>& Triangles = getTriangles();
  map< int, vector<int> >& vertfaceIndex = getvertfaceIndex();

	//counts
	int nvertex = num_vertices(Mesh);
	int nfaces  = Triangles.size();
	int nedges  = num_edges(Mesh);
	int natoms  = mesh_structure->nSpheres();
	int nprops   = fColumnHeads.size();

	//sizes
	int head_size = 10 * sizeof(int) + nprops * 30;
	int triangle_size = 3* sizeof(int);
	int vertex_size = 6 * sizeof(float) + sizeof(short) + nprops * sizeof(float);

	int t_flag = SLD_HASNORMALS | SLD_HASATOMS;

	//write file ID
	int dummy = FILEID;
	writeInt(aStream, dummy, reverse);

  //#vertices, #faces, #edges
	writeInt(aStream, head_size, reverse);
	writeInt(aStream, triangle_size, reverse);
	writeInt(aStream, vertex_size, reverse);
		
	writeInt(aStream, nfaces, reverse);
	writeInt(aStream, nvertex, reverse);
	writeInt(aStream, nedges, reverse);
	writeInt(aStream, natoms, reverse);
	writeInt(aStream, nprops, reverse);

	writeInt(aStream, t_flag, reverse);

	for(i=0;i<nprops;i++)
	{
		char propname[30] = "";
		size_t namelen = fColumnHeads[i].size();
		if (namelen>29) namelen = 29;

		//copy the column name and null terminate it
		strncpy(propname,fColumnHeads[i].c_str(), namelen);
		propname[namelen] = 0;

		aStream.write((char*) propname, 30);          //read property names
	}

  //write the triangles
  for(Trimesh::FaceIter f=mesh_structure->beginFace();f!=mesh_structure->endFace();++f)
  {
		int points = 3;
		int* vixs = new int[points];

		for (int i=0;i<points;++i)
		{
			vixs[i] = f->getVertexIndex(i);
		}

		writeIntArr(aStream, vixs, 3, reverse);
		
		delete[] vixs;
  }
	
	
  //write the vertices
  for(Trimesh::vertex_it v=mesh_structure->begin();v!=mesh_structure->end();++v)
  { 
		float xyz[3];
		float nxyz[3];
		short atix = (*data_map)[*v].cSphere;

		int i;

		for(i=0;i<3;++i)
		{
			xyz[i]  = (*data_map)[*v].coordinates[i];
			nxyz[i] = (*data_map)[*v].normal[i];
		}

		
		writeFloatArr(aStream, xyz, 3, reverse);
		writeFloatArr(aStream, nxyz, 3, reverse);
		writeShort(aStream, atix, reverse);
		if (nprops>0)
		{
			float* props = new float[nprops];
			
			for(i=0;i<nprops;++i)
				props[i] = fData(*v,i);

			writeFloatArr(aStream, props, nprops, reverse);

			delete[] props;
		}
  }

  return true;
}
