/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef SLDTRIMESH_HH
#define SLDTRIMESH_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <string>
#    include <vector>
#  else
#    include <vector.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#    include <string>
#    include <vector>
#endif //HAVE_CONFIG_H


#include "trimeshio.hh"

/**
 * \brief trimesh class, able to read and write triangulated mesh objects in
 * the SLD object format
 *
 * The \ref sldformat_page "SLD object format" is used by some molecular 
 * modelling packages like MOLCAD(tm) or MOLEKEL(tm).
 */

class sldTrimesh : public trimeshIO
{
public:

  /**
   * \brief constructor initializing msmsTrimesh with a Trimesh object
   */
  sldTrimesh(Trimesh* aTrimesh): 
			trimeshIO(aTrimesh) { }

  /** 
   * \brief there is a stream handling for library own format files.
   */
  bool handleStreams() const { return true; }


  /**
   * \brief reads a surface mesh object in library own format from a file 
   * specified in \p aFile
   *
   */
  bool read(const string& aFile);
  
  /**
   * \brief reads a surface mesh object in library own format from the stream
   * specified in \p aStream
   */
  bool read(istream& aStream);

  /**
   * \brief writes a surface mesh object in library own format into a file 
   * specified in \p aFile
   *
   */
  bool write(const string& aFile);

  /**
   * \brief writes a surface mesh object in library own format into a stream 
   * specified in \p aStream
   *
   */
  bool write(ostream& aStream);
};

#endif 


/**
 * \page sldformat_page SLD object format
 *
 * \note I do not understand the SLD format completely. That is mainly because 
 * the only information I coudl get about the structure of this file format was
 * the source code of MOLEKEL, which was provided by Stefan Portmann.Together 
 * with the code and Emacs hexl-mode I could decode the binary structure of the
 * MOLCAD SLD files, which was my main concern.
 *
 * The SLD format is a binary file format describing triangulated meshes together
 * with vertex properties. It is used by molecular graphics and modelling
 * to describe the various kinds of molecular surfaces.
 *
 * \subsection sldfile_header File Header:
 *
 * The file header contains information about the surface structure and the
 * vertex properties. Its lengths differs with the number of attached properties.
 *
 * The following list contains the meanings of each double word in the header:
 *
 * <table>
 * <tr><td>0000-0003</td>
 *     <td>SLD file format ID code: -4712</td>
 * <tr><td>0004-0007</td>
 *     <td>size of header: 10 * sizeof(int) + number of properties * 30</td>
 * <tr><td>0008-000B</td>
 *     <td>size of triangle description (usually 3 * sizeof(int)</td>
 * <tr><td>000C-000F</td>
 *     <td>size of vertex description</td>
 * <tr><td>0010-0013</td>
 *     <td>number of triangles</td>
 * <tr><td>0014-0017</td>
 *     <td>number of vertices</td>
 * <tr><td>0018-001B</td>
 *     <td>number of connections</td>
 * <tr><td>001C-001F</td>
 *     <td>number of atoms</td>
 * <tr><td>0020-0023</td>
 *     <td>number of properties</td>
 * <tr><td>0024-0027</td>
 *     <td>t-Flag (I do not understand the definition of this element)</td>
 * 
 * The rest of the header, if any, contains 30 character long descriptions of any
 * attached property. Thus the file header size has allways the formula:
 * \f$ 40 + n \cdot 30 \f$, where n is the number of properties.
 *
 * After the triangle and vertex section is a section of readable comment lines
 * which can contain some useful information for the user.
 * 
 * \subsection sldfile_faces Triangle Section:
 * 
 * The triangle information consists of three 4-byte integers representing the
 * indices of the triangles vertices.

 * \subsection sldfile_vertices Vertex Section:
 *
 * Each vertex is described by the following sequence:
 *
 * X Y Z NX NY NZ (float) ATOMID (short int) 0e (?) VERTEX PROPERTIES (one float
 * per prop.) 
 * 
 */
