/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include <triangle.hh>

//******************************************************************************
Vector_<double> Triangle::operator[](const int ix) const
//******************************************************************************
{
	Vector_<double> result(3);
	
	if ((ix>-1) && (ix<3))
		return *(positions[ix]);
	else
		return Vector_<double>(3, -1);
	
}

//******************************************************************************
Triangle::Triangle(const Edgepoint& e1, const Edgepoint& e2, const Edgepoint& e3)
//******************************************************************************
: vertex(3,-1), positions(3)
{
  //copy the positions of the vertices into the positions structure
  positions[0] = e1.second;
  positions[1] = e2.second;
  positions[2] = e3.second;

  //copy the indices of the edge vertices into the vertex structure
  vertex[0] = e1.first;
  vertex[1] = e2.first;
  vertex[2] = e3.first;
}

//******************************************************************************
Triangle::Triangle(const Triangle& source)
//******************************************************************************
: vertex(3), positions(3)
{
	copy(source.vertex.begin(),source.vertex.end(),vertex.begin());
	copy(source.positions.begin(),source.positions.end(),positions.begin());
}

//******************************************************************************
Triangle& Triangle::operator=(const Triangle& source)
//******************************************************************************
{
	copy(source.vertex.begin(),source.vertex.end(),vertex.begin());
	copy(source.positions.begin(),source.positions.end(),positions.begin());	
	
	return *this;
}


//******************************************************************************
Triangle::Edgepoint Triangle::setVertex(const int ix, const Edgepoint& e)
//******************************************************************************
{
  //the old value will be returned as a Edgepoint reference
  Edgepoint old(vertex[ix],positions[ix]);

  if ((ix>=0) && (ix<=2))
  {

    //set the index of the edgepoint as
    vertex[ix] = e.first;
    if (e.second->dim()==3) 
			positions[ix] = e.second;
    else 
      // if the size of the edge vector is not 3 reset to the old vertex
      vertex[ix] = old.first;

  }

  return old;
}

//******************************************************************************
int Triangle::setVertexIndex(const int ix, const int vix)
//******************************************************************************
{
	if ((ix<0) || (ix>2)) throw out_of_range("ix");

	int old = vertex[ix];
	vertex[ix] = vix;
	return old;
}


//******************************************************************************
Vector_<double> Triangle::getVertexPosition(const int ix) const
//******************************************************************************
{
	Vector_<double> result;
	if ((ix>=0) && (ix<=2))
	{
		result = *positions[ix];
		return result;
	}
	else throw out_of_range("ix");

	//should never be returned
	return result;
}


//******************************************************************************
Vector_<double> Triangle::getSide(const int ix) const
//******************************************************************************
{
	Vector_<double> result(3);
	if ((ix>=0) && (ix<=2))
	{
		result = *positions[sideix[ix].second] - *positions[sideix[ix].first];
		return result;
	}
	else 
		throw out_of_range("ix");
	
	//should never be returned
	return result;
}

//******************************************************************************
double Triangle::sideLength(const int ix) const
//******************************************************************************
{
	if ((ix>=0) && (ix<=2))
	{
		Vector_<double> side = *positions[sideix[ix].second]  - 
			                     *positions[sideix[ix].first];
		return side.vec_len();
	}
	else 
		throw out_of_range("ix");
	
	//should never be returned
	return -1;
}


//******************************************************************************
double Triangle::cosangle(int const ix) const
//******************************************************************************
{
  //the angles at each edgepoint are the angles between the two sides that
  //come together at each point

	double cosa = 0;

	double scalarprod = 0;                      //scalar product
	double length_a   = 0;                      //length of first vector
	double length_b   = 0;                      //length of second vector

	//indices to the next points
	int point_a       = ix-1;
	if (point_a<0) point_a = 2;
	int point_b       = ix+1;
	if (point_b>2) point_b = 0;


    //calculating the scalar product and the length of each vector
    for (int j=0;j<=2;j++)
    {
			double a = (*positions[point_a])[j] - (*positions[ix])[j];
			double b = (*positions[point_b])[j] - (*positions[ix])[j];
      scalarprod += a * b;
      length_a   += a * a;
      length_b   += b * b;
    }

    length_a = sqrt(length_a);
    length_b = sqrt(length_b);

    if ((length_a!=0) && (length_b!=0))
      cosa = scalarprod/(length_a*length_b);
    else
      //Remember: the angle of an arbitrary vector with a zero-length vector is
      //90 degrees (pi/2)
      cosa=0;

	return cosa;
}

//******************************************************************************
double Triangle::sinangle(int const ix) const
//******************************************************************************
{
	double cosa = cosangle(ix);
	return sqrt(1-cosa*cosa);
}

//******************************************************************************
double Triangle::cotangle(int const ix) const
//******************************************************************************
{
	//cotangens is cosinus divided by sinus
	double cosa = cosangle(ix);
	double sina = sqrt(1-cosa*cosa);
	
	return cosa/sina; 
}


//******************************************************************************
double Triangle::area() const
//******************************************************************************
{
  //arbitrarily I use the first edgepoint as angle point for A = bc sin(alpha)
	double sina = sinangle(0);
	double len1 = sideLength(1);
	double len2 = sideLength(2);
  return 0.5 * sina * len1 * len2;
}

//******************************************************************************
int Triangle::successor(const int v) const
//******************************************************************************
{
  for(vector<int>::const_iterator i=vertex.begin();i!=vertex.end();++i)
  {
    if (*i == v)
    {
      ++i;
      if (i==vertex.end()) { return *vertex.begin(); } else { return *i; }
    }
  }

  return -1;
}

//******************************************************************************
int Triangle::obtuseAngle() const
//******************************************************************************
{
  for (int i=0;i<=2;++i)
    //if the cosinus of an angle is below zero the angle is between (90,270)
    if (cosangle(i)<0) return i;

  return -1;
}


//******************************************************************************
Triangle::Edgepoint edgepoint(const int index, const Vector_<double>* position)
//******************************************************************************
{
  Triangle::Edgepoint result(index,position);
  return result;
}

