/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef TRIANGLE_H
#define TRIANGLE_H

#include <math.h>
#include <float.h>
#include <algorithm>
#include <Vector.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <stdexcept>
#  else
#    include <stdexception.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <stdexcept>
#endif //HAVE_CONFIG_H

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace RazorBack;
#endif //HAVE_NAMESPACES

typedef pair<int,int> side_t;
const side_t sideix[3] = { side_t(1,2), side_t(2,0), side_t(0,1) };

/**
 * \brief a datatype for the representation of faces with edge vertices, side
 * vectors and angles.
 *
 * The \c Triangle class is a representation for a triangle of a triangulated
 * surface mesh. It contains data about the edgepoints (positions and vertices
 * indices in the mesh), the side vectors and the cosinus of the angles. There
 * are functions retrieving the edgepoints, side directions and lengths, angles
 * and sinus of the angles. The whole triangle area can be calculated too.
 */

class Triangle
{
private:
  /**
   * \brief face flag contained in the MSMS face files
   */
  int faceflag;

  /**
   * \brief analytical face number contained in the MSMS face files
   */
  int facenum;

  /**
   * \brief zero based vertex list containing the three edges of the triangle
   */
  vector<int> vertex;

  /**
   * \brief zero based list of the edgepoint positions
   */
  vector<const Vector_<double>* > positions;

  /**
   * \brief calculates the side vectors of all sides.
   *
   * The vectors of the sides are the differences between the coordinate
   * vectors (s) of the side's edgepoint (e):
   *
   * i.e. \f$ {\bf s}_1 = {\bf e}_3 - {\bf e}_2 \f$
   */
  void calculateSides();

  /**
   * \brief calculates the cosinus of the angles at each edgepoint
   *
   * The angle is calculated via the scalar product of the two adjecent
   * side vectors:
   *
   * \f[
   *     cos(\alpha) = \frac{{\bf a} \cdot {\bf b}}{|{\bf a}| \cdot |{\bf b}|}
   * \f]
   */
  void calculateCosAngles();

public:

  /**
   * \brief pair, describing a triangle edgepoint
   *
   * A triangle Edgepoint consists of an index, referring to a vertex in an
   * triangulated surface Mesh (see Trimesh), and the 3D position of the
   * point. These two values are combined to a std::pair structure to allow
   * easy exchange with other entities.
   */
  typedef pair<int,const Vector_<double>* > Edgepoint;


  /**
   * \brief standard constructor for the Triangle class
   *
   * This constructor is for internal use only, because it does not initialize
   * the edgepoints etc.
   */
  Triangle(): vertex(3,-1), positions(3) { }

  /**
   * \brief constructor for the Triangle class using the edgepoints \c e1 - \c e3
   *
   * The constructor adds the three edgepoints to the triangle and starts the
   * calculation for the side vectors and angles.
   */
  Triangle(const Edgepoint& e1, const Edgepoint& e2, const Edgepoint& e3);

	/**
	 * \brief copy constructor for the Triangle class
	 */
	Triangle(const Triangle&);

	/**
	 * \brief copy operator
	 */
	Triangle& operator=(const Triangle&);

  /**
   * \brief retrieves the edgepoint index of the vertex nr \c ix
   */
  int vertex2edge(const int ix) const
  {
    for(int i=0;i<=2;++i) if (vertex[i]==ix) return i; 
		return -1;
  }

	/**
	 * \brief the index operator retrieves the coordinates of the edge point
	 */
	Vector_<double> operator[](const int ix) const;

  /**
   * \brief function that retrieves the angle with the index \c ix.
   *
   * This is an \c inline method. It returns the radiant value of the angle at
   * the vertex \c ix.
   *
   * \note Do not forget, that the vertex and angle list is zero based!
   */
  double angle(int const ix) const
  {
    return acos(cosangle(ix));
  }

  /**
   * \brief function that retrieves the cosinus of the angle with the index \c ix.
   *
   * This is an \c inline method. It returns the cosinus value of the angle at
   * the vertex \c ix.
   *
   * \note Do not forget, that the vertex and angle list is zero based!
   */
  double cosangle(int const ix) const;

  /**
   * \brief function that retrieves the sinus of the angle with the index \c ix.
   *
   * This is an \c inline method. It returns the sinus value of the angle at
   * the vertex \c ix.
   *
   * \note Do not forget, that the vertex and angle list is zero based!
   */
  double sinangle(int const ix) const;

  /**
   * \brief retrieves the cotangens of the angle at point \c ix
   *
   * The cotangens values are used for the curvature values
   */
  double cotangle(int const ix) const;

  /**
   * \brief sets the \c index and \c position of an vertex to a new \c value
   *
   * If the vertex index is correct (0-2) the function returns the old index, -1
   * otherwise
   */
  Edgepoint setVertex(const int ix, const Edgepoint& e);

  /**
   * \brief retrieves the vertex index of the triangle edgepoint \c ix
   *
   * The index value must be between 0 and 2 otherwise the function will return
   * -1.
   */
  int getVertexIndex(const int ix) const
  {
    if ((ix>=0) && (ix<=2)) return vertex[ix]; else return -1;
  }
  /**
   * \brief sets only the vertex Index of an Edgepoint
   *
   * returns the old vertex index
   *
   * \param ix index of the edgepoint (0, 1 or 2)
   * \param vix new vertex index for edgepoint \p ix
   * 
   * \exception out_of_range thrown, if ix is not valid
   */
  int setVertexIndex(const int ix, const int vix);

  /**
   * \brief retrieves the vertex position in 3D coordinates
   */
  Vector_<double> getVertexPosition(const int ix) const;

  /**
   * \brief retrieves the side vector opposite to the triangle edgepoint \c ix
   *
   * The index value must be between 0 and 2 otherwise the function will return
   * -1.
   */
  Vector_<double> getSide(const int ix) const;

  /**
   * \brief retrieves the length of the side opposite to the triangle edgepoint
   * nr. \c ix
   */
  double sideLength(const int ix) const;

  /**
   * \brief retrieves the faceflag of the triangle
   */
  int getFaceFlag() const
  {
    return faceflag;
  }

  /**
   * \brief sets the faceflag of the triangle
   */
  int setFaceFlag(const int aflag)
  {
    int old = faceflag;
    faceflag = aflag;
    return old;
  }
  
  /**
   * \brief retrieves the analytical face number of the triangle
   */
  int getFaceNum() const
  {
    return facenum;
  }

  /**
   * \brief sets the analytical face number of the triangle
   */
  int setFaceNum(const int afacenum)
  {
    int old = facenum;
    facenum = afacenum;
    return old;
  }

  /**
   * \brief calculates the area of the triangle
   *
   * To calculate the area of the triangle the following formula is used:
   * \f[
   {\cal A} = |{\bf b}| \cdot
   |{\bf c}| \cdot
   sin(\alpha)
   \f]
  */
  double area() const;
  /**
   * \brief retrieves the successor of the vertex \c v in the triangle
   *
   * The successor of an edgepoint is the next edgepoint if you move along
   * the triangle in a counter-clockwise manner. If the vertex nr \c v is not
   * a part of the triangle, the function returns -1;
   */
  int successor(const int v) const;

  /**
   * \brief returns the index of an obtuse angle in the triangle
   *
   * If the triangle has an obtuse angle this function retrieves its index. If
   * there is no obtuse angle the function returns -1
   */
  int obtuseAngle() const;
};


/**
 * \brief helper function to create a Edgepoint pair out of an index and a
 * vector of three elements (the coordinates).
 */
Triangle::Edgepoint edgepoint(const int index, const Vector_<double>* position);

#endif
