/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "trimesh.hh"
#include <boost/graph/dijkstra_shortest_paths.hpp>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <algorithm>
#    include <functional>
#  else
#    include <map.h>
#    include <algo.h>
#    include <function.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#    include <map>
#    include <algorithm>
#    include <functional>
#endif //HAVE_CONFIG_H

/**
 * TYPES
 *
 * - Internal trianlge representation in recalcTriangles: triangle_t
 */

struct triangle_t 
{ 
	Trimesh::vertex_t a, b, c; 
};

/**
 * Helper functions:
 *                    calculate_angle       calculates angles between vertices
 *                    calculate_length      calculates the distance between two
 *                                          adjecent vertices
 */

//*******************************************************************************
inline bool operator<(const triangle_t& x, const triangle_t& y)
//*******************************************************************************
{
	return ((x.a<y.a) || 
					((x.a==y.a) && (x.b<y.b)) ||
					((x.a==y.a) && (x.b==y.b) && (x.c<y.c)));
}

//*******************************************************************************
inline void SwapVertex(Trimesh::vertex_t& a, Trimesh::vertex_t& b)
//*******************************************************************************
{
	Trimesh::vertex_t tmp=a; 
	a=b; 
	b=tmp;	
}

//******************************************************************************
triangle_t sortedTriangle(Trimesh::vertex_t a,Trimesh::vertex_t b,
													Trimesh::vertex_t c)
//******************************************************************************
{
	triangle_t triangle;

	if (b>c) SwapVertex(b,c);
	if (a>c) SwapVertex(a,c);
	if (a>b) SwapVertex(a,b);

	triangle.a = a;
	triangle.b = b;
	triangle.c = c;

	return triangle;
}

//******************************************************************************
Vector_<double> Cross(Vector_<double> a, Vector_<double> b)
//******************************************************************************
{
	Vector_<double> c(3);
	c[0] = a[1]*b[2] - a[2]*b[1];
	c[1] = a[2]*b[0] - a[0]*b[2];
	c[2] = a[0]*b[1] - a[1]*b[0];
	
	return c;
}


//******************************************************************************
double calculate_angle(const Vector_<double>& v1,
											 const Vector_<double>& v2,
											 const Vector_<double>& v3)
//******************************************************************************
/**
 * by Torsten Schindler:
 * Calculate angle at v2 between 3 (3-dim) vectors v1--v2--v3
 */
{
  double a2 = 
		(v3[0]-v1[0]) * (v3[0]-v1[0]) + 
		(v3[1]-v1[1]) * (v3[1]-v1[1]) + 
		(v3[2]-v1[2]) * (v3[2]-v1[2]);
  double b2 = 
		(v2[0]-v1[0]) * (v2[0]-v1[0]) + 
		(v2[1]-v1[1]) * (v2[1]-v1[1]) + 
		(v2[2]-v1[2]) * (v2[2]-v1[2]);
  double c2 = 
		(v2[0]-v3[0]) * (v2[0]-v3[0]) + 
		(v2[1]-v3[1]) * (v2[1]-v3[1]) + 
    (v2[2]-v3[2]) * (v2[2]-v3[2]);

  if (b2*c2 > 0.0) return acos((b2 + c2 - a2)/(2*sqrt(b2*c2))); else return 0.0;
}


//******************************************************************************
double calculate_length(const Vector_<double>& v1,const Vector_<double>& v2)
//******************************************************************************
/**
 * returns the length of the vector from v1 to v2
 */
{
	Vector_<double> delta = v2 - v1;
	return delta.vec_len();
}

//******************************************************************************
int Trimesh::getDoublett(const Vector_<double>& coords) const
//******************************************x************************************
{
  for(vertex_it v=begin();v!=end();++v)
    if (abs(data_map[*v].coordinates[0]-coords[0])<1e-3)
      if (abs(data_map[*v].coordinates[1]-coords[1])<1e-3)
				if (abs(data_map[*v].coordinates[2]-coords[2])<1e-3) 
					return vertIndex(*v);
  return -1;
}


//******************************************************************************
Trimesh::vertex_t Trimesh::getVertex(const int ix, bool ownindex) const
//******************************************************************************
  {
    if (ownindex)
    {
      for(vertex_it v=begin();v!=end();++v)
        if (data_map[*v].ownIndex==ix) return *v;
    }
    else
      return vertex(ix,Mesh);
    
    return 0;
  }

//*******************************************************************************
Trimesh::vertex_t Trimesh::getVertexByMaster(const int mix) const
//*******************************************************************************
{
	for(vertex_it v=begin();v!=end();++v)
		if (data_map[*v].masterIndex==mix) return *v;

	return -1;
}

//******************************************************************************
Trimesh::Trimesh():
//******************************************************************************
Mesh(), Triangles(), vertfaceIndex()
{
  //vertexindices_map = get(vertex_index,Mesh);
	color_map = get(vertex_color, Mesh);
  data_map = get(vertex_data, Mesh);

  edgelengths_map = get(edge_weight, Mesh);
	edgeindex_map = get(edge_index, Mesh);
}

//*******************************************************************************
Trimesh::Trimesh(const Trimesh& source)
//*******************************************************************************
: Mesh(), Triangles(), vertfaceIndex()
{
	typedef pair<vertex_it, vertex_it> vertex_its;
	vertex_its vits = vertices(source.Mesh);

	map<vertex_t, vertex_t> vmapping;
	typedef map<vertex_t, vertex_t>::const_iterator vmapping_it;
	
	color_map = get(vertex_color, Mesh);
  data_map = get(vertex_data, Mesh);

  edgelengths_map = get(edge_weight, Mesh);
	edgeindex_map = get(edge_index, Mesh);

	vertex_it v;
	vertex_it nv; 

	//copy vertices from old trimesh
	for(v=vits.first;v!=vits.second;++v)
	{
		pair<int, vertex_t> new_v = copyVertex(source,*v);
		vmapping[*v] = new_v.second;
	}

	//copy edges
	for(v=vits.first;v!=vits.second;++v)
	{
		pair<edge_oit,edge_oit> v_edges = out_edges(*v, source.Mesh);
		vertex_t s = vmapping[*v];
  
		for(edge_oit e=v_edges.first; e!=v_edges.second;++e)
		{
			vertex_t t = target(*e,source.Mesh);

			//create the edge and add the edgelength if the edge is new
			if (!edge(s, vmapping[t], Mesh).second)
      {
				pair<edge_t, bool> new_e = add_edge(s, vmapping[t], Mesh);
				put(edgelengths_map, new_e.first, source.edgelengths_map[*e]);
				put(edgeindex_map, new_e.first, source.edgeindex_map[*e]);
      }
		}
	}
	
	//recreate the triangles
	Triangles = source.Triangles;
	for(vector<Triangle>::iterator f=Triangles.begin();f!=Triangles.end();++f)
		for(int i=0;i<3;++i)
		{
			int ep= f->getVertexIndex(i);
			vertex_t w = vertex(ep,Mesh);
			f->setVertex(i, edgepoint(ep, &data_map[w].coordinates));
		}

	
	//rebuild vertex face index;
	rebuildVertexFaceIx();

	//copy trimesh properties
	copyProperties(source);
}


//******************************************************************************
Trimesh::ownindexpair_t Trimesh::copyVertex(const Trimesh& trimesh, 
																						const vertex_t v)
//******************************************************************************
{
  vertex_t new_v = add_vertex(Mesh);
  
  // Copy vertex properties
  put(data_map,new_v,trimesh.data_map[v]);
  put(color_map,new_v,trimesh.color_map[v]);

  return ownindexpair_t(trimesh.data_map[v].ownIndex,new_v);
}

//******************************************************************************
void Trimesh::copyVertex(const Trimesh& trimesh, 
												 const vertex_t source, const vertex_t target,
												 const bool takeMaster)
//******************************************************************************
{
	property_map<Graph_t, vertex_index_t>::type 
		indexmap = get(vertex_index, Mesh);
	
  // Copy vertex properties
  put(data_map,target,trimesh.data_map[source]);
  put(color_map,target,trimesh.color_map[source]);

	if (takeMaster)
	  data_map[target].masterIndex = indexmap[source];
	else
	  data_map[target].masterIndex = trimesh.data_map[source].masterIndex;
}

//*******************************************************************************
size_t Trimesh::addTriangle(vertex_t a, vertex_t b, vertex_t c)
//*******************************************************************************
{
	Triangles.push_back(Triangle(edgepoint(a,&data_map[a].coordinates),
															 edgepoint(b,&data_map[b].coordinates),
															 edgepoint(c,&data_map[c].coordinates)));

	size_t newix = Triangles.size() - 1;
	
	add_edge(a, b, Mesh);
	add_edge(b, c, Mesh);
	add_edge(c, a, Mesh);
	
	vertfaceIndex[a].push_back(newix);
	vertfaceIndex[b].push_back(newix);
	vertfaceIndex[c].push_back(newix);

	return newix;
}

//******************************************************************************
int Trimesh::copyEdges(const Trimesh& trimesh, vertex_t v, vertex_t new_v, 
		       const ownindexmap_t& ixmap)
//******************************************************************************
{
  pair<edge_oit,edge_oit> v_edges = out_edges(v, trimesh.Mesh);
  int nedges=0;
  
  for(edge_oit e=v_edges.first; e!=v_edges.second;++e)
  {
    vertex_t t = target(*e,trimesh.Mesh);
    int ownindex = trimesh.data_map[t].ownIndex;
    
    ownindexmap_t::const_iterator new_t = ixmap.find(ownindex);
    if (new_t!=ixmap.end())
    {
      //create the edge and add the edgelength if the edge is new
      pair<edge_t, bool> new_e = add_edge(new_v, new_t->second, Mesh);
      if (new_e.second)
      {
				put(edgelengths_map, new_e.first, trimesh.edgelengths_map[*e]);
				++nedges;
				
				put(edgeindex_map, new_e.first, trimesh.edgeindex_map[*e]);
      }
    }
  }

  return nedges;
}

//******************************************************************************
map<int, vector<int> >& Trimesh::rebuildVertexFaceIx()
//******************************************************************************
{
  map<int, vector<int> > newIndex;

  for(int f=0;f<nFaces();++f)
    for(int i=0;i<3;++i) 
      newIndex[Triangles[f].getVertexIndex(i)].push_back(f);

  vertfaceIndex = newIndex;
  return vertfaceIndex;
}


//*******************************************************************************
int Trimesh::createEdgeIndex()
//*******************************************************************************
{
	int c = 0;
  pair<edge_it,edge_it> eits = edges(Mesh);
	
  for(edge_it e=eits.first; e!=eits.second;++e)
		edgeindex_map[*e] = c++;

	return --c;
}

//******************************************************************************
void Trimesh::calcOnSurfaceDistance(const vertex_t v, vector<double>& dist)
	const
//******************************************************************************
{
  dist.resize(num_vertices(Mesh));
	dijkstra_shortest_paths(Mesh, v, distance_map(&dist[0]));
}

//*******************************************************************************
size_t Trimesh::nSpheres() const
//*******************************************************************************
{
	size_t max_ix = 0;

	for(vertex_it v = begin();v!=end();++v)
		if (data_map[*v].cSphere>max_ix) max_ix = data_map[*v].cSphere;
	
	return max_ix;		
}

//******************************************************************************
void Trimesh::calcSurfaceNormals()
//******************************************************************************
{
  //presized vector to accept all the normals of all triangles
  vector<Vector_<double> > normals(Triangles.size());
  
  //create all the triangle normals
  int i=0;
  for(FaceIter f=beginFace();f!=endFace();++f)
  {
    //resize the surface normal of triangle f
		normals[i] = Cross((*f)[2]-(*f)[1],(*f)[0]-(*f)[1]);
		normals[i].vec_norm();
    ++i;
  }
  
  for(vertex_it v=begin();v!=end();++v)
  {
    //surface normal of vertex v initialized with 0
    Vector_<double> n(3,0);

    //sum over all triangle normals adjacent to the vertex
    for(int i=0;i<vertfaceIndex[*v].size();++i)
    {
      int f = vertfaceIndex[*v][i];
			n += normals[f];
    }
  
    //normalize surface normals
    n.vec_norm();
    //write back the surface normal
    data_map[*v].normal = n;
  }
}

//******************************************************************************
size_t Trimesh::recalcTriangles()
//******************************************************************************
{
	/* TYPE DEFINITIONS */
	typedef set<triangle_t> triangles_t;
	typedef triangles_t::iterator triangles_it;
	typedef pair<adjacency_it,adjacency_it> a_its;

	/* INITIALIZATION */
	triangles_t triangleset;

	//clear all triangles
	vertfaceIndex.clear();
	Triangles.clear();


	/* TRIANGLE BUILD UP */
	//LOOP over all edges and build the triangles
	pair<edge_it,edge_it> eits = edges(Mesh);
	for(edge_it e=eits.first;e!=eits.second;++e)
	{
		list<vertex_t> points;
		vertex_t s = source(*e,Mesh);
		vertex_t t = target(*e,Mesh);

		//set of adjacency vertices for the source and target vertex of the edge
		a_its saits = adjacent_vertices(s, Mesh);
		a_its taits = adjacent_vertices(t, Mesh);
		set<vertex_t> s_adj(saits.first, saits.second);
		set<vertex_t> t_adj(taits.first, taits.second);

		
		//make a set intersection between saits and taits
		set_intersection(s_adj.begin(), s_adj.end(), t_adj.begin(), t_adj.end(),
                     back_inserter(points));
		
		for(list<vertex_t>::const_iterator v=points.begin();v!=points.end();++v)
			triangleset.insert(sortedTriangle(*v,s,t));
	}

	/* GEOMETRIC CHECK AND INSERT INTO MESH*/
	// The following condition must be true:
	//     bc x ba * nb > 0 

	//LOOP over all triangles to check the vertex order and insert into mesh
	for(triangles_it t=triangleset.begin();t!=triangleset.end();++t)
	{
		//triangle points
		Vector_<double> a = vertCoordinates(t->a);
		Vector_<double> b = vertCoordinates(t->b);
		Vector_<double> c = vertCoordinates(t->c);
			
		Vector_<double> n = vertNormals(t->b);             //normal at point b
		Vector_<double> bc = c - b;                        //triangle side a
		Vector_<double> ba = a - b;                        //triangle side b

		//GEOMETRY CHECK and inserting triangle counter clockwise
		if ((n * Cross(bc,ba)) > 0) 
			Triangles.push_back(Triangle(edgepoint(t->a,&(data_map[t->a].coordinates)), 
																	 edgepoint(t->b,&(data_map[t->b].coordinates)),
																	 edgepoint(t->c,&(data_map[t->c].coordinates))
																	 ));
		else
			Triangles.push_back(Triangle(edgepoint(t->c,&(data_map[t->c].coordinates)),
																	 edgepoint(t->b,&(data_map[t->b].coordinates)),
																	 edgepoint(t->a,&(data_map[t->a].coordinates))
																	 ));
			
		//Update the vertex face index
		size_t ix = Triangles.size()-1;
		vertfaceIndex[t->a].push_back(ix);
		vertfaceIndex[t->b].push_back(ix);
		vertfaceIndex[t->c].push_back(ix);
	}

	//from the number of edges and triangles added, calculate the number of missing
	//edges
	return 2 * num_edges(Mesh) - 3 * Triangles.size();
	
}

//******************************************************************************
Vector_<double> Trimesh::setCoordinates(const vertex_t& v, 
																			 const double x, 
																			 const double y, 
																			 const double z)
//******************************************************************************
{
	Vector_<double> old_pos = data_map[v].coordinates;
	
	data_map[v].coordinates[0] = x;
	data_map[v].coordinates[1] = y;
	data_map[v].coordinates[2] = z;

	return old_pos;
}

//******************************************************************************
Vector_<double> Trimesh::setCoordinates(const int & vix, 
																			const double x, 
																			const double y, 
																			const double z)
//******************************************************************************
{
	vertex_t v = vertex(vix, Mesh);

	Vector_<double> old_pos = data_map[v].coordinates;
	
	data_map[v].coordinates[0] = x;
	data_map[v].coordinates[1] = y;
	data_map[v].coordinates[2] = z;

	return old_pos;
}

//******************************************************************************
Vector_<double> Trimesh::setCoordinates(const vertex_t& v,
																			 const Vector_<double>& pos)
//******************************************************************************
{
	const Vector_<double>& old_pos = data_map[v].coordinates;
	
	if (pos.dim()==3)
		data_map[v].coordinates = pos;
	else
		throw out_of_range("ix");

	return old_pos;
}

//******************************************************************************
Vector_<double> Trimesh::setCoordinates(const int & vix,
																			 const Vector_<double>& pos)
//******************************************************************************
{
	vertex_t v = vertex(vix, Mesh);
	const Vector_<double>& old_pos = data_map[v].coordinates;
	
	if (pos.dim()==3)
		data_map[v].coordinates = pos;
	else
		throw out_of_range("ix");

	return old_pos;
}

//*******************************************************************************
Vector_<double> Trimesh::setNormal(const int vix, const Vector_<double>& normal)
//*******************************************************************************
{
	vertex_t v = vertex(vix, Mesh);
	const Vector_<double>& old  = data_map[v].normal;
	
	if (normal.dim()==3)
		data_map[v].normal = normal;
	else
		throw out_of_range("ix");

	return old;
}

//*******************************************************************************
Vector_<double> Trimesh::setNormal(const vertex_t& v, 
																	 const Vector_<double>& normal)
//*******************************************************************************
{
	const Vector_<double>& old  = data_map[v].normal;
	
	if (normal.dim()==3)
		data_map[v].normal = normal;
	else
		throw out_of_range("ix");

	return old;	
}

//*******************************************************************************
int Trimesh::vertIndex(const vertex_t v, bool ownindex) const
//*******************************************************************************
{
	if (ownindex) 
		return data_map[v].ownIndex; 
	else 
	{
		trimesh_traits::VertexIndexMap index_map = get(vertex_index, Mesh);
		return get(index_map, v);
	}
}

//******************************************************************************
int Trimesh::setvertKind(const int& vix, const int& value)
//******************************************************************************
{
	vertex_t v = vertex(vix,Mesh);

	int old_value = data_map[v].kind;
	data_map[v].kind = value;
	
	return old_value;
}

//******************************************************************************
int Trimesh::setvertKind(const vertex_t& v, const int& value)
//******************************************************************************
{
	int old_value = data_map[v].kind;
	data_map[v].kind = value;
	
	return old_value;
}

//*******************************************************************************
int Trimesh::setMasterIx(vertex_t v, int master)
//*******************************************************************************
{
	int old_value = data_map[v].masterIndex;
	data_map[v].masterIndex = master;
	return old_value;
}	

//******************************************************************************
double Trimesh::edgeWeight(const vertex_t v1, const vertex_t v2, double def) const
//******************************************************************************
{
	pair <edge_t, bool> e = edge(v1,v2,Mesh);

	if (e.second)
		return edgelengths_map[e.first];
	else
		return def;

	return -1;
}

//******************************************************************************
double Trimesh::setEdgeWeight(edge_t e, const double value)
//******************************************************************************
{
	double old_value = edgelengths_map[e];
	edgelengths_map[e] = value;
	
	return old_value;
}


//******************************************************************************
double Trimesh::surfaceArea() const
//******************************************************************************
{
  typedef vector<Triangle>::const_iterator iter;
  double surface = 0;

  for (iter i=Triangles.begin();i<Triangles.end();++i)
  {
    surface += i->area();
  }

  return surface;
}

//*******************************************************************************
Trimesh::FaceIxIter Trimesh::beginFaceIx(const int ix) const
//*******************************************************************************
{
	map<int, vector<int> >::const_iterator m = vertfaceIndex.find(ix);
	if (m==vertfaceIndex.end()) 
		throw out_of_range("ix");
	
	return m->second.begin();
}

//*******************************************************************************
Trimesh::FaceIxIter Trimesh::beginFaceIx(const vertex_t& v) const
//*******************************************************************************
{
	int ix = vertIndex(v, false);
	map<int, vector<int> >::const_iterator m = vertfaceIndex.find(ix);
	if (m==vertfaceIndex.end()) 
		throw out_of_range("vertex v");

	return m->second.begin();
}

//*******************************************************************************
Trimesh::FaceIxIter Trimesh::endFaceIx(const int ix) const
//*******************************************************************************
{
	map<int, vector<int> >::const_iterator m = vertfaceIndex.find(ix);
	if (m==vertfaceIndex.end()) 
				throw out_of_range("ix");
	
	return m->second.end();
}

//*******************************************************************************
Trimesh::FaceIxIter Trimesh::endFaceIx(const vertex_t& v) const
//*******************************************************************************
{
	int ix = vertIndex(v, false);
	map<int, vector<int> >::const_iterator m = vertfaceIndex.find(v);
	if (m==vertfaceIndex.end()) 
			throw out_of_range("vertex v");

	return m->second.end();
}

//*******************************************************************************
Trimesh::vertex_t Trimesh::nextVertex(const vertex_t center, const vertex_t out)
//*******************************************************************************
{
	list<int> common;
	commonTriangles(center, out, back_inserter(common));
	vertex_t suc;

	for(list<int>::iterator i=common.begin();i!=common.end();++i)
	{
		suc = Triangles[*i].successor(out);
		if (suc!=center) break; 
	}
	
	return suc;
}
