/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// ADT - Trimesh
//
// Author: Christian Hofbauer - Novartis Forschungsinstitut, Vienna Nov. 2001
//
// This abstract data type represents an inteface for a triangulated mesh.
// The class provides a collection of basic operations like reading it from
// a msms file, colouring the vertices with different properties. And writing
// it back to some different file types, including MSMS files and our own file
// type for storing all the colours
//
// It can also create a NxN distance matrix
//
////////////////////////////////////////////////////////////////////////////////

#ifndef TRIMESH_H
#define TRIMESH_H

#include "trimesh_traits.hh"
#include "triangle.hh"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <string>
#    include <stdexcept>
#    include <algorithm>
#    include <list>
#  else
#    include <string.h>
#    include <stdexception.h>
#    include <algorithm.h>
#    include <list.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <string>
#  include <stdexcept>
#  include <algorithm>
#  include <list>
#endif //HAVE_CONFIG_H

/**
 * The Trimesh class is an encapsulation for a triangulated surface mesh. The
 * class contains information about the vertices of the mesh, their connectivity
 * and the single faces (which are of course triangles). The mesh can be
 * constructed from MSMS (by Micheal Sattler) output files. For the
 * administration of the faces the Trimesh class uses a list of Triangles.
 *
 * The graph declerations for the \c Trimesh class are stored in the 
 * \c trimesh_traits class.
 */
class Trimesh: public trimesh_traits
{
protected:
  /**
   * \brief holds a list of all faces on the mesh
   *
   * The list of faces contains a fully described triangle for all faces on
   * the surface mesh.
   */
  vector<Triangle> Triangles;

  /**
   * \brief retrieves the indices of the faces meeting at a vertex
   */
  map<int, vector<int> > vertfaceIndex;

  typedef pair<int, vector<int> > faceixpair_t;
  typedef map<int, vector<int> >::const_iterator faceix_it;

  /**
   * \brief referrence to the list of vertex coordinates
   */
  VertexColorsMap color_map;

  /**
   * \brief referrence to the list of vertex coordinates
   */
  VertexDataMap data_map;

  /**
   * \brief referrence to the list of edge lengths (weights)
   */
  EdgeLengthMap edgelengths_map;

  /**
   * \brief referrence to the list of edge indices
   */
  EdgeIndexMap edgeindex_map;

	/**
	 * \brief the number of closest spheres
	 */
	int fSpheres;

  string vertexComment;
  string faceComment;

  size_t nsphere;
  double probe_r;
  double density;


public:

  /**
   * \brief graph that contains the vertex and connectivity information for
   * the surface mesh
   */
  Graph_t Mesh;

  /**
   * \brief searches for another vertex with the same coordinates as \c vix
   *
   * Doublett-search is important, because isotopic vertices are leading into 
   * degenerated triangles (containing zero angles). The method searches for
   * the first occurence of the coordinate set coord in the vertex list of the
   * mesh.
   */
  int getDoublett(const Vector_<double>& coords) const;

  /**
   * \brief iterator type for the face lists
   */
  typedef vector<int>::const_iterator FaceIxIter;

  /**
   * \brief iterator type for the list of face lists
   */
  typedef map<int, vector<int> >::const_iterator FaceListIter;

  /**
   * \brief iterator type for the faces
   */
  typedef vector<Triangle>::const_iterator FaceIter;

  /**
   * the standard constructor for the \c Trimesh class. Initilizes the graph
   * and all its internal properties
   */
  Trimesh();

	/**
	 * \brief copy constructor
	 */
	Trimesh(const Trimesh& source);

  /**
   * \brief copies a vertex from another Surface 
	 *
	 * \return the ownindex and the added vertex.
	 *
	 * \param trimesh source trimesh structure
	 * \param v source vertex in the source trimesh structure
   */
  ownindexpair_t copyVertex(const Trimesh& trimesh, const vertex_t v);

	
	/**
	 * \brief copies the vertex \p source from another Surface to \p target on this
	 * surface
	 *
	 * The vertex \p target receives any property of \p source, but no connectivity
	 * information is copied. The treatment of the masterindex property is
	 * specified by the \p takeMaster paramter. If this parameter is \c true 
	 * (default) the index of \p source becomes the masterindex of the \p target
	 * vertex, otherwise the master index of the \p source is copied as any other
	 * property.
	 *
	 * \param trimesh source trimesh structure
	 * \param source source vertex in the source trimesh structure
	 * \param target target vertex in this trimesh structrue
	 * \param takeMaster if \c true assignes the index of \p source as masterindex
	 *                   or copies the masterindex otherwise
	 */
  void copyVertex(const Trimesh& trimesh, 
									const vertex_t source, const vertex_t target,
									const bool takeMaster = true);

	/**
	 * \brief adds a triangle to the mesh
	 *
	 * The method adds a triangle, all its edges and the vertfaceIndex entries to
	 * the mesh. The edgeweights are not set. The method returns the index of the
	 * new triangle
	 *
	 * \return index of the new triangle
	 *
	 * \param a,b,c three vertices forming the triangle
	 */
	size_t addTriangle(vertex_t a, vertex_t b, vertex_t c);

  /**
   * \brief copies the edges of the vertex \p v in \p trimesh to the new vertex
   * \p new_v in this mesh.
   *
   * The method adds all edges leaving the vertex \p v in the surface mesh
   * \p trimesh to the new vertex \p new_v where the target vertices  are 
   * contained in this mesh too.
   *
   * \param trimesh the source surface mesh
   * \param v source vertex
   * \param new_v vertex_t in this mesh, corresponding to the source vertex in the
   *              source mesh.
   * \param ixmap map correlating the ownindices with the new vertices
   */
  int copyEdges(const Trimesh& trimesh, vertex_t v, 
		vertex_t new_v, const ownindexmap_t& ixmap);

  void copyProperties(const Trimesh& trimesh)
  {
    vertexComment = trimesh.vertexComment;
    faceComment = trimesh.faceComment;

    nsphere = trimesh.nsphere;
    probe_r = trimesh.probe_r;
    density = trimesh.density;
  }

  /**
   * \brief rebuilds the vertex face index 
   */
  map<int, vector<int> >& rebuildVertexFaceIx();

	/**
	 * \brief makes an index of all edges
	 *
	 * The function gives every edge in the graph an index, according to their
	 * appereance in the edge list. It returns the last index of the edge list.
	 */
	int createEdgeIndex();

	/**
	 * \brief counts the number of triangles common for two vertices 
	 *
	 * \param v,u  vertices in the mesh
	 */
	template<class Out>
	void commonTriangles(vertex_t v, vertex_t u, Out ins) const;

  /**
   * \brief (re)calculates the "on surface" distance Matrix for the mesh
   *
   * The "on surface" distance between 2 vertices is the sum of the length of
   * all edges along the shortest path between them.
   *
   * \note Don't mix that up with the Euklidean distance which is in fact the
   * distance on a straight line through space and can be retrieved by a
   * trivial calculation, which is performed by TriMesh::eukDistance
   */
  void calcOnSurfaceDistance(const vertex_t v, 
														 vector<double>& distances) const;

  /**
   * \brief recalculates the surface normals for each vertex
   *
   * The surface normals are calculated using the normal vectors of each triangle
   * on the surface. For every triangle on the surface, the normal vector to the 
	 * surface is calculated.
   *
   * For each vertex v on the surface the normal vector is the sum of all normal 
   * vectors of the adjacent surfaces, normalized to a length of 1
   */
  void calcSurfaceNormals();

  /**
   * \brief reestablishes the triangles out of a set of edges
	 *
	 * All triangles, that are formed by the mesh's set of edges are found. If any
	 * edges are missing the triangles are not found. To find missing edges and 
	 * complete the triangulation use correctEdges(size_t).
	 *
	 * \return the number of border edges.
	 *
	 * \sa correctEdges(size_t)
   */
	size_t recalcTriangles();

  /**
   * \brief retrieves the number of vertices in the mesh
   */
  int nVertices() const
  {
    return num_vertices(Mesh);
  }

  /**
   * \brief retrieves the number of edges in the mesh
   */
  int nEdges() const
  {
    return num_edges(Mesh);
  }

	/**
	 * \brief returns the maxium index for the closest spheres
	 *
	 * \return maximum value for the closest spheres
	 */
	size_t nSpheres() const; // { return fSpheres; }

	/**
	 * \brief sets the number of closest spheres
	 */
	

  /**
   * \brief returns the number of faces in the mesh
   */
  int nFaces() const
  {
    return Triangles.size();
  }

  /**
   * \brief returns the number of faces adjacent to a vertex
   */
  int nFaces(const vertex_t& v) const
  {
    int vix = vertIndex(v);
    faceix_it faceix = vertfaceIndex.find(vix);
    if (faceix==vertfaceIndex.end()) throw inconsistent_vertexfaceix();

    return faceix->second.size();
  }

  /**
   * \brief returns the vertex with the number \c ix
   */
  vertex_t getVertex(const int ix, bool ownindex=false) const;

	/**
	 * \brief returns the vertex who's master index is \p mix
	 *
	 * This method retrieves the first vertex who's master index is \p mix.
	 *
	 * \return vertex with master index \p mix
	 *
	 * \param mix master index of the desired vertex
	 */
	vertex_t getVertexByMaster(const int mix) const;

  /**
   * \brief retrieves the index of the Vertex \c v.
   *
   * The indices of the vertices are determined through the sequence of the
   * vertices in the vertex file.
	 *
	 * \return  the index of the vertex \p v
	 *
	 * \param v identifer of a vertex in the trimesh structure
	 * \param ownindex denotes if the own index, or the internal index of the
	 *                 vertex should be returned
   */
  int vertIndex(const vertex_t v, bool ownindex = true) const;

  /**
   * \brief retrieves the coordinates of vertex nr \c vix
   */
  const Vector_<double>& vertCoordinates(const int vix) const
  {
    if ((vix<0) || (vix>=num_vertices(Mesh))) throw out_of_range("vix");
    return data_map[vertex(vix,Mesh)].coordinates;
  }

  const Vector_<double>& vertCoordinates(const vertex_t& v) const
  {
    return data_map[v].coordinates;
  }

	/**
	 * \brief Sets the vertex coordinates of the vertex \p v to the new \p x, 
	 * \p y and \p z value.
	 */
	Vector_<double> setCoordinates(const vertex_t& v, 
																const double x, const double y, const double z);

	/**
	 * \brief Sets the vertex coordinates of the vertex with index \p vix  to the 
	 * new \p x, \p y and \p z value.
	 */
	Vector_<double> setCoordinates(const int & vix, 
																const double x, const double y, const double z);

	/**
	 * \brief Sets the vertex coordinates of the vertex \p v to the new position
	 * \p pos.
	 */
	Vector_<double> setCoordinates(const vertex_t& v,const Vector_<double>& pos);

	/**
	 * \brief Sets the vertex coordinates of the vertex with index \p vix  to the 
	 * new position \p pos.
	 */
	Vector_<double> setCoordinates(const int & vix,const Vector_<double>& pos);

	/**
	 * \brief Sets the normal of the vertex with index \p vix  to the 
	 * new vector \p normal.
	 *
	 * \param vix    the index of the vertex
	 * \param normal the new vertex normal
	 *
	 * \return the components of the old vertex normal vector
	 */
	Vector_<double> setNormal(const int & vix,const Vector_<double>& normal);

	/**
	 * \brief Sets the normal of the vertex \p V to the new vector \p normal.
	 *
	 * \param v      the vertex
	 * \param normal the new vertex normal
	 *
	 * \return the components of the old vertex normal vector
	 */
	Vector_<double> setNormal(const vertex_t& v, const Vector_<double>& normal);

	/**
	 * \brief Sets the normal of the vertex with index \p vix  to the 
	 * new vector \p normal.
	 */
	Vector_<double> setNormal(const int vix, const Vector_<double>& normal);

	/**
	 * \brief retrieves the kind of the vertex \c v
	 */
	int vertKind(const vertex_t& v) const { return data_map[v].kind; }

	/**
	 * \brief retrieves the kind of the vertex with index \c vix
	 */
	int vertKind(const int& vix) const { return data_map[vertex(vix, Mesh)].kind; }

	/**
	 * \brief sets the kind of the vertex with the index \c vix to the value \p value
	 */
	int setvertKind(const int& vix, const int& value);
	
	/**
	 * \brief sets the kind of the vertex \p v to the value \p value
	 */
	int setvertKind(const vertex_t& v, const int& value);

	/**
	 * \brief retrieves the masterindex of the vertex \c v
	 */
	int vertMasterIx(const vertex_t& v) const { return data_map[v].masterIndex; }

	/**
	 * \brief retrieves the masterindex of the vertex with index \c vix
	 */
	int vertMasterIx(const int& vix) const { return data_map[vertex(vix, Mesh)].masterIndex; }

	/**
	 * \brief sets the master index of vertex v
	 *
	 * \param v      the vertex 
	 * \param master the new master index
	 */
	int setMasterIx(vertex_t v, int master);

	/**
	 * \brief retrieves the color of the vertex with index \c vix
	 */
	int vertColor(const vertex_t& v) const { return color_map[v]; }

  /**
   * \brief retrieves the surface normal vector of vertex nr \c vix
   */
  const Vector_<double>& vertNormals(const int vix) const
  {
    if ((vix<0) || (vix>=num_vertices(Mesh))) throw out_of_range("vix");
    return data_map[vertex(vix,Mesh)].normal;
  }

  /**
   * \brief retrieves the surface normal vector of vertex \c v
   */
  const Vector_<double>& vertNormals(const vertex_t& v) const
  {
    return data_map[v].normal;
  }

	/**
	 * \brief retrieves the source vertex of the given edge
	 *
	 * \return source vertex of the given edge
	 *
	 * \param edge considered edge
	 */
	vertex_t sourceVertex(const edge_t edge) const { return source(edge, Mesh); }

	/**
	 * \brief retrieves the target vertex of the given edge
	 *
	 * \return target vertex of the given edge
	 *
	 * \param edge considered edge
	 */
	vertex_t targetVertex(const edge_t edge) const { return target(edge, Mesh); }

	/**
	 * \brief retrieves the edge weight associated with an possible edge between
	 * the vertices \p v1 and \p v2
	 *
	 * If there is an edge between \p v1 and \p v2 the method returns the edge
	 * weight, else the method returns the \p default value which normally is zero.
	 *
	 * \param {v1, v2} vertices of the edge
	 * \param def value returned, if there is no edge between v1 and v2
	 */
	double edgeWeight(const vertex_t v1, const vertex_t v2, double def = 0) const;

	/**
	 * \brief retrieves the edge weight of the edge \p e.
	 *
	 * \param e edge of triangulated mesh
	 */
	double edgeWeight(edge_t e) const { return edgelengths_map[e]; }

	/**
	 * \brief sets the weight of an edge
	 */
	double setEdgeWeight(edge_t e, const double value);

  /**
   * \brief retrieves the surface area of the mesh
   *
   * The surface area is the sum of all triangle areas. The unit is square
   * Angstroms
   */
  double surfaceArea() const;

  /**
   * \brief retrieves the face at index \c ix
   */
  const Triangle& getFace(const int ix) const
  {
    if ((ix<0) || (ix>=Triangles.size())) throw out_of_range("");
    return Triangles[ix];
  }
  /**
   * \brief retrieves the triangle at position \c ix in the facelist of vertex
   * nr. \c vix
   * 
   * \param vix own index of the vertex
   * \param ix index of the face 
   *
   * \exception trimesh_traits::inconsistent_vertexfaceix thrown if the face
   *            index of the vertex with index \p vix could not be found
   * \exception out_of_range thrown if the face index \p ix is not valid
   */
  const Triangle& getFace(const int vix, const int ix) const
  {

    faceix_it faceix = vertfaceIndex.find(vix);
    if (faceix==vertfaceIndex.end()) throw inconsistent_vertexfaceix();
     
    if((ix<0) || (ix>=faceix->second.size())) throw out_of_range("");

    return Triangles[faceix->second[ix]];
  }

  /**
   * \brief retrieves the triangle at position \c ix in the facelist of vertex
   * \c v
   *
   * \param v vertex to which the face belongs
   * \param ix index of the face 
   *
   * \exception trimesh_traits::inconsistent_vertexfaceix thrown if the face
   *            index of the vertex \p v could not be found
   * \exception out_of_range thrown if the face index \p ix is not valid
   */
  const Triangle& getFace(const vertex_t& v, const int ix) const
  {
    int vix = vertIndex(v);

    faceix_it faceix = vertfaceIndex.find(vix);
    if (faceix==vertfaceIndex.end()) throw inconsistent_vertexfaceix();
     
    if((ix<0) || (ix>=faceix->second.size())) throw out_of_range("");

    return Triangles[faceix->second[ix]];
  }

  /**
   * \brief points to the begin of a faceindex list for the vertex nr. \c ix
   *
   * This function, together with \c {endFaceIx(const int ix)} can be used to
   * iterate the face indices for single vertex with nr. \c ix.
   */
  FaceIxIter beginFaceIx(const int ix) const;
  FaceIxIter beginFaceIx(const vertex_t& v) const;

  /**
   * \brief points to the end of a faceindex list for the vertex nr. \c ix
   *
   * This function, together with \c {beginFaceIx(const int ix)} can be used
   * to iterate the face indices for single vertex with nr. \c ix.
   */
  FaceIxIter endFaceIx(const int ix) const;
  FaceIxIter endFaceIx(const vertex_t& v) const;

  /**
   * \brief iterator pointing to the first face in the facelist of the mesh
   *
   * The face iterators provide access to all the faces of the mesh. The begin
   * iterator points to the first face in the list, which corresponds to the
   * first face in the face file of the mesh object. The end of the face list
   * is referrenced by the endFace() method.
   */
  FaceIter beginFace() const { return Triangles.begin(); }

  /**
   * \brief iterator pointing one after the last face in the facelist of the
   * mesh
   *
   * The face iterators provide access to all the faces of the mesh. The begin
   * iterator points to the first face in the list, which corresponds to the
   * first face in the face file of the mesh object. The beginning of the face
   * list is referrenced by the beginFace() method.
   */
  FaceIter endFace() const { return Triangles.end(); }

  /**
   * \brief iterator pointing to the first edge in the mesh
   *
   * With a edge iterator you can access all edges in the mesh. The
   * sequence is sorted by the edge indices. The end of the edge
   * sequence is marked with the end() method.
   */
  edge_it beginEdge() const { return edges(Mesh).first; }

  /**
   * \brief iterator pointing one after the last edge in the mesh
   *
   * With a edge iterator you can access all vertices in the mesh. The
   * sequence is sorted by the edge indices. The begin of the edge
   * sequence is marked with the \c begin() method.
   */
  edge_it endEdge() const { return edges(Mesh).second; }

  /**
   * \brief iterator pointing to the first vertex in the mesh
   *
   * With a vertex iterator you can access all vertices in the mesh. The
   * sequence is sorted by the vertex indices. The end of the vertex
   * sequence is marked with the end() method.
   */
  vertex_it begin() const { return vertices(Mesh).first; }

  /**
   * \brief iterator pointing one after the last vertex in the mesh
   *
   * With a vertex iterator you can access all vertices in the mesh. The
   * sequence is sorted by the vertex indices. The begin of the vertex
   * sequence is marked with the \c begin() method.
   */
  vertex_it end() const { return vertices(Mesh).second; }

  /**
   * \brief iterator pointing to the first adjacent vertex of the vertex \c v
   *
   * An adjacency iterator provides access to the direct neighbors of any vertex.
   * To iterate through the neighborhood of a vertex use the \c beginAdjacency
   * and \c endAdjacency methods.
   */
  adjacency_it beginAdjacency(const vertex_t v) const 
	{
    return adjacent_vertices(v,Mesh).first;
  }

  /**
   * \brief iterator pointing on after the last adjacent vertex of the vertex 
   * \c v
   *
   * An adjacency iterator provides access to the direct neighbors of any vertex.
   * To iterate through the neighborhood of a vertex use the \c beginAdjacency
   * and \c endAdjacency methods.
   */
  adjacency_it endAdjacency(const vertex_t v) const 
  {
    return adjacent_vertices(v,Mesh).second;
  }


	vertex_t nextVertex(const vertex_t center, const vertex_t out);

  /**
   * \brief retrieves a vector with the ordered sequence of the one-ring
   * around a central vertex \c v
   *
   * The one-ring neighborhood (also called the flower) of a vertex \c v is
   * the sequence of all vertices adjecent to that vertex in a counter-clock
   * wise order. The sequence starts at an arbitrary adjecent vertex and moves
   * around the one-ring in the right direction.
   *
   * As a side effect of this procession the triangle index for the vertex v is
   * sorted the same way.
   *
   * \param v vertex for which the onering is demanded
	 * \param ins output iterator accepting the sequence of vertices
   *
   * \exception trimesh_traits::inconsistent_vertexfaceix thrown if an entry in 
   *            the vertex-face index is missing.
   */
	template<class Out>
  size_t onering(const vertex_t& v, Out ins) const;

  /**
   * \brief the IO class trimeshIO is a friend of trimesh
   */
  friend class trimeshIO;

  /**
   * \brief the class Convexhull is a friend
   */
  friend class Convexhull;

};

const int PATCH_INSIDE = 1;	
const int PATCH_CONVEXBORDER = 2;
const int PATCH_CONCARVEBORDER = 3;
const int PATCH_CENTER = 4;

double calculate_angle(const Vector_<double>& v1,
											 const Vector_<double>& v2,
											 const Vector_<double>& v3);

double calculate_length(const Vector_<double>& v1,
												const Vector_<double>& v2);

Vector_<double> Cross(Vector_<double> a, Vector_<double> b);

#ifdef INCLUDE_TMPL_DEFS
  #include "trimesh_templ.cc"
#endif

#endif
