/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef TRIMESH_ALGORITHMS_HH
#define TRIMESH_ALGORITHMS_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#    include <list>
#    include <set>
#    include <vector>
#    include <algorithm>
#  else
#    include <map.h>
#    include <list.h>
#    include <set.h>
#    include <vector.h>
#    include <algorithm.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <map>
#  include <list>
#  include <set>
#  include <vector>
#  include <algorithm>
#endif //HAVE_CONFIG_H

#include <boost/graph/visitors.hpp>
#include <boost/graph/breadth_first_search.hpp>

#ifdef HAVE_NAMESPACES
using namespace boost;
using namespace RazorBack;
#endif //HAVE_NAMESPACES

/**
 * DISTANCE TABLES Trait class
 *
 * a distance table of a vertex is a vector (range 0 to distance) of list<int>
 * its zero layer contains only the own vertex the other layer contain the n'th 
 * ring around the vertex
 */

template<class vertexClass> class distancetable_traits
{
	public:

	typedef set<vertexClass> ring;
	typedef vector<ring> ringtable;
	typedef map<vertexClass,ringtable> distances;
};

/**
 * \brief creates a set of vertices which are at least but not much more than 
 * \p distance edges away from each other.
 *
 * The result of the method is a set of equally dispersed vertices on a graph.
 * The vertices should have a minimum distance of \p distance from each other,
 * but should not be much further away from their neighbors. The output is
 * written in \p ins, which should be an object representing the concept of 
 * \b Inputiterator. 
 *
 * The number of selected vertices is returned.
 * 
 * \param graph the graph containing the vertices (should be representing the 
 *        concept of an adjacency_list graph of the BOOST library
 * \param start index of the starting vertex on the \p graph
 * \param distance number of edges to be the minimum distance on the graph
 * \param ins input iterator expecting the selected vertices
 */
//*******************************************************************************
template<class graph_t, class Out>
size_t disperseVertices(graph_t& graph, 
												const graph_traits<graph_t>::vertex_descriptor start, 
												const int distance,	Out ins)
//*******************************************************************************
{
	//graph types
  typedef graph_traits<graph_t>::vertex_descriptor  vertex_t;
  typedef graph_traits<graph_t>::vertex_iterator vertex_it;
  typedef graph_traits<graph_t>::vertices_size_type vsize_t;
	typedef graph_traits<graph_t>::adjacency_iterator adj_it;

	//property maps
	typedef typename property_map<graph_t, vertex_index_t>::type indexmap_t;
	typedef typename property_traits<indexmap_t>::value_type index_t;
	typedef typename property_map<graph_t, vertex_color_t>::type colormap_t;
	typedef typename property_traits<colormap_t>::value_type color_t;

	//distance table
	typedef typename distancetable_traits<vertex_t>::ring ring_t;
	typedef typename ring_t::iterator ring_it;
	typedef typename distancetable_traits<vertex_t>::ringtable dt_t;
	typedef typename dt_t::iterator dt_it;
	typedef typename distancetable_traits<vertex_t>::distances dts_t;
	typedef typename dts_t::iterator dts_it;

	/* CONCEPT PROOF */
	function_requires< VertexListGraphConcept<graph_t> >();
	function_requires< ReadWritePropertyMapConcept<colormap_t, vertex_t> >();
	function_requires< IntegerConcept<color_t> >();
	function_requires< IntegerConcept<index_t> >();

  /**
   * INITIALIZATION 
   */

  /* preparation of the variables */

  const int UNKNOWN = 0;
  const int POSSIBLE = 1;
  const int IMPOSSIBLE = 2;
  const int SELECTED = 3;

  //number of vertices
  int n_verts = num_vertices(graph);

	colormap_t colormap = get(vertex_color, graph);

  //initialize color map

	dts_t D;

  //initialize color map  and get distance map
	pair<vertex_it, vertex_it> vits = vertices(graph);
  for (vertex_it i=vits.first;i!=vits.second;++i)
  {
		//initialize the color map entry of this vertex
		colormap[*i] = UNKNOWN;

		//initialize the distance map of vertex i with distance layers
		D[*i].resize(distance+1);
		//add itself into the first layer
		D[*i][0].insert(*i);


		//generate one layer above the other
		for(int j=1;j<=distance;++j)
		{
			for(ring_it r=D[*i][j-1].begin();r!=D[*i][j-1].end();++r)
			{
				pair<adj_it,adj_it> adjacency = adjacent_vertices(vertex(*r,graph),graph);
				for(adj_it a=adjacency.first;a!=adjacency.second;++a)
				{
					//check if it is not in the last two layers
					if (D[*i][j-1].find(*a) == D[*i][j-1].end())
					{
						//there could be a conflict with the second layer
						bool ok_2nd = true;
						if (j-2>=0)
							ok_2nd =(D[*i][j-2].find(*a)==D[*i][j-2].end());
						
						if (ok_2nd) D[*i][j].insert(*a);
					}
				}
			}
		}
  }

  //list containing the SELECTED vertices
  list<vertex_t> selected_vs;

  //add start vertex to the SELECTED list
  selected_vs.push_back(start);
  colormap[start] = SELECTED;

  //mark for the start vertex d<distance as IMPOSSIBLE 
	for(int i=1;i<distance;++i)
		for(ring_it r=D[start][i].begin();r!=D[start][i].end();++r)
			colormap[*r] = IMPOSSIBLE;
	
	//and d=distance as POSSIBLE	
	for(ring_it r=D[start][distance].begin();r!=D[start][distance].end();++r)
		if (selected_vs.size()<2)
		{
			//push the first POSSIBLE vertex to the SELECTED list
			selected_vs.push_back(*r);
			colormap[*r] = SELECTED;
		}
		else
			colormap[*r] = POSSIBLE;
	
  //mark all the neighbors of the new vertex IMPOSSIBLE if d<distance
  list<vertex_t>::iterator act_vertex = selected_vs.end();
  --act_vertex;

	for(int i=1;i<distance;++i)
		for(ring_it r=D[*act_vertex][i].begin();r!=D[*act_vertex][i].end();++r)
			colormap[*r] = IMPOSSIBLE;

  /**
   * ITERATION 
   */
  while (act_vertex!=selected_vs.end())
	{
		for(ring_it r=D[*act_vertex][distance].begin();r!=D[*act_vertex][distance].end();++r)
			if (colormap[*r] == POSSIBLE)
			{
				//the vertex is in a POSSIBLE distance to all other SELECTED vertices
				selected_vs.push_back(*r);
				colormap[*r] = SELECTED;
	    
				//set the status of all vertices nearer to i than distance to IMPOSSIBLE
				for(int i=1;i<distance;++i)
					for(ring_it s=D[*r][i].begin();s!=D[*r][i].end();++s)
						colormap[*s] = IMPOSSIBLE;
			}
			else if (colormap[*r] == UNKNOWN)
				colormap[*r] = POSSIBLE;
      
    ++act_vertex;
  }

  /**
   * FINALIZATION
   */
  for(list<vertex_t>::iterator v=selected_vs.begin();v!=selected_vs.end();++v)
    if (colormap[*v] == SELECTED)
      *ins++ = *v;


  return selected_vs.size();
}


/**
 * \brief calculates the sum of all edge weights within a graph
 *
 * The method returns the sum of all edge weight within the graph. Therefore
 * the graph must represent the concept of \c adjacency_list of the BOOST
 * library and must contain an edge_weight property.
 *
 * The type of the weights is denoted with the variable accepting the result
 * of the function
 *
 * \param graph     graph containing the weighted edges.
 * \param old_value receives a previous sum if addition should be continued
 */

//*******************************************************************************
template<class graph_t, typename weight_t>
weight_t sum_of_weights(graph_t& graph, weight_t old_value = 0)
//*******************************************************************************
{
  typedef graph_traits<graph_t>::edge_descriptor edge_t;
  typedef graph_traits<graph_t>::edge_iterator edge_it;

	property_map<graph_t, edge_weight_t>::type weight_map = get(edge_weight, graph);

	pair<edge_it, edge_it> edge_its = edges(graph);
	for(edge_it e=edge_its.first;e!=edge_its.second;++e)
	  old_value += get(weight_map, *e);

	return old_value;
	
}

/**
 * \page disperse_vertex Finding Dispersed Vertices
 *
 * For surface on surface matching of whole objects it is important to cover the
 * whole surface with homogen patches. Hence the central vertices of the surface
 * patches should be equally distributed over the surface.
 * 
 * One method to generate a well distributed set of vertices is selecting those
 * who are in a certain distance from each other. The following 
 * \ref distv_alg "algorithm" produces a such a set of vertices. To perform this
 * operation it uses a so called \ref dist_table "distance table" to access its 
 * neighborhood.
 * 
 *
 * \section dist_table Distance Table
 * A distance table is a datastructure which holds the vertices in a certain
 * distance of a central vertex \em v (a layer) in a seperate container leading 
 * to the following table:
 *
 * <table>
 * <tr><td><b> distance</b></td>
 *     <td><b>vertices in the layer</b></td></tr>
 * <tr><td>1</td>
 *   <td>v<sub>1,1</sub>, v<sub>1,2</sub>, v<sub>1,3</sub>, ... v<sub>1,n</sub></td></tr>
 * <tr><td>2</td>
 *   <td>v<sub>2,1</sub>, v<sub>2,2</sub>, v<sub>2,3</sub>, ... v<sub>2,n</sub></td></tr>
 * <tr><td>...</td>
 *   <td>...</td></tr>
 * <tr><td>n</td>
 *   <td>v<sub>n,1</sub>, v<sub>n,2</sub>, v<sub>n,3</sub>, ... v<sub>n,n</sub></td></tr>
 * </table>
 *
 * \section distv_alg Dispersed Vertices Algorithm
 * The mechanism of this algorithm could be compared with the drawing of circles
 * around each selected vertex and selecting the intersection points with other
 * circles if the lay outside of all circles drawn at that time. The following
 * pseudo code describes the algorithm:
 *
 * \code
 * PROCEDURE DispersedVertices( Graph, start_vertex, distance, selected_list)
 * BEGIN
 * //initialization 
 * create Distance table (by breadth first search)
 *
 * mark v of Graph | d(start_vertex,v)< distance as IMPOSSIBLE
 * mark v of Graph | d(start_vertex,v)== distance as POSSIBLE
 * 
 * act_vertex := first POSSIBLE vertex
 * 
 * selected_list << start_vertex << act_vertex
 * mark act_vertex & start_vertex SELECTED
 *
 * mark v of Graph | d(act_vertex,v) < distance as IMPOSSIBLE
 * 
 * act_vertex := last of selected_list
 * 
 * //loop over all selected vertices
 * WHILE (index of act_vertex < count of selected_list)
 *   BEGIN
 *   FOR EACH v of Graph | d(act_vertex,v) == distance
 *     IF v marked POSSIBLE
 *       THEN BEGIN
 *            selected_list << v 
 *            mark v as SELECTED
 *            mark w of Graph | d(w,v) < distance as IMPOSSIBLE
 *            END
 *     ELSE IF v not marked
 *            mark v as POSSIBLE
 *
 *   act_vertex := next in selected_list
 *   END 
 * 
 * \endcode
 */

#endif //TRIMESH_ALGORITHMS_HH
