/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef TRIMESH_GRAPH_HH
#define TRIMESH_GRAPH_HH

#include <boost/config.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/properties.hpp>
#include <boost/property_map.hpp>

#include <Vector.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#  else
#    include <map.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#    include <map>
#endif //HAVE_CONFIG_H

#ifdef HAVE_NAMESPACES
using namespace std;
using namespace boost;
using namespace RazorBack;
#endif //HAVE_NAMESPACES

class trimeshIO;

struct vertexdata_t
{
	size_t ownIndex;
	size_t masterIndex;

	int kind;

	Vector_<double> coordinates;
	Vector_<double> normal;
	
	int face_num;
	int cSphere;
	int face_flag;

	/**
	 * \brief standard constructor
	 */
	vertexdata_t()
			: ownIndex(-1), masterIndex(0), kind(0),
				coordinates(3), normal(3),
				face_num(0), cSphere(0), face_flag(0) {}

	/**
	 * \brief copy constructor
	 */
	vertexdata_t(const vertexdata_t& src)
			: ownIndex(src.ownIndex), masterIndex(src.masterIndex), kind(src.kind),
				coordinates(src.coordinates), normal(src.normal),
				face_num(src.face_num), cSphere(src.cSphere), face_flag(src.face_flag) {}
			 
};

/**
 * In the boost namespace the enumaration types and property definition for the
 * additional vertex and edge properties are defined.
 *
 * Actually there are two additional vertex properties:
 *
 * \li The vertex 3D coordinates (\c vertex_coordinates)
 * \li The vertex surface normal vector (\c vertex_normals)
 */
namespace boost
{
  enum vertex_data_t { vertex_data = 300 };
  BOOST_INSTALL_PROPERTY(vertex, data);
};

/**
 * \brief the trimesh_traits class contains all the type definitions of the
 * Trimesh class. The main purpose is to maintain the typedefs for the mesh graph
 * acoording to the boost library.
 */
class trimesh_traits
{
protected:
  //Type definition for the mesh graph, its vertices and edges

	/**
	 * \brief a color property for different algorithms
	 *
	 */
	typedef property<vertex_color_t, int> VertexColor;

  /**
   * \brief vectors of the surface normals
   *
   * is a vector (size 3) that contains the 3D-space coordinates for the
   * surface normals at each vertex point.
   */
  typedef property<vertex_data_t, vertexdata_t , VertexColor> VertexData;

  /**
   * \brief index property for each vertex
   *
   * The vertex indices start with 0 and end with numVertices. They are
   * according to the MSMS vertex file.
   */
  typedef property<vertex_index_t, int, VertexData> VertexIndex;

  /**
   * \brief euklidean distance between each adjecent pair of vertices
   *
   * The length of the edges is stored as the weights (colors) of the mesh's
   * edges.
   */
  typedef property<edge_weight_t, double> EdgeLength;
	typedef property<edge_weight2_t, double, EdgeLength> EdgeWeight2;

  /**
   * \brief euklidean distance between each adjecent pair of vertices
   *
   * The length of the edges is stored as the weights (colors) of the mesh's
   * edges.
   */
  typedef property<edge_index_t, int, EdgeWeight2> EdgeIndex;

public:

	/**
	 * \brief standard constructor
	 */
	trimesh_traits() { }

  /**
   * \brief typdefinition for the Mesh graph.
   *
   * The surface mesh is a undirected graph, that is stored as an adjacency
   * list and the vertices and edges are stored as vectors
   */
  typedef adjacency_list<vecS, vecS, undirectedS, VertexIndex, EdgeIndex> Graph_t;

  /**
   * \brief property map for the vertex color
   */
  typedef property_map<Graph_t, vertex_color_t>::type VertexColorsMap;

  /**
   * \brief property map for the surface normal vectors
   */
  typedef property_map<Graph_t, vertex_data_t>::type VertexDataMap;

  /**
   * \brief property map for the edge lengths (weights)
   */
  typedef property_map<Graph_t, edge_weight_t>::type EdgeLengthMap;

  /**
   * \brief property map for the edge lengths (weights)
   */
  typedef property_map<Graph_t, edge_index_t>::type EdgeIndexMap;

  /**
   * \brief property map for the vertex indices
   */
  typedef property_map<Graph_t, vertex_index_t>::type VertexIndexMap;

  /**
   * \brief typedefinition of the vertex descriptor
   */
  typedef graph_traits<Graph_t>::vertex_descriptor vertex_t;

  /**
   * \brief typedefinition of the edge descriptor
   */
  typedef graph_traits<Graph_t>::edge_descriptor edge_t;

  /**
   * \brief iterator type for the vertices
   */
  typedef graph_traits<Graph_t>::vertex_iterator vertex_it;

  /**
   * \brief iterator type for the edges
   */
  typedef graph_traits<Graph_t>::edge_iterator edge_it;
  
  /**
   * \brief out-edge iterator type
   */
  typedef graph_traits<Graph_t>::out_edge_iterator edge_oit;

  /**
   * \brief adjancecy iterator
   */
  typedef graph_traits<Graph_t>::adjacency_iterator adjacency_it;

  /**
   * \brief exception thrown when the Trimesh::vertexFaceIndex is not uptodate
   */
  class inconsistent_vertexfaceix { };
  
  /**
   * \brief thrown if the Trimesh structure is plain. (e.g. Convexhull::generateHull)
   */
  class plain_meshstructure { };

protected:

  /**
   * \brief an ownindex associated with an vertex (represented as a standard 
   * pair)
   */
  typedef pair<int,vertex_t> ownindexpair_t;

  /**
   * \brief a map associting own indices with vertices
   */
  typedef map<int,vertex_t> ownindexmap_t;

  /**
   * \brief iterator type for ownindexmap_t
   */
  typedef ownindexmap_t::iterator ownindexmap_it;

  friend class trimeshIO;

};
#endif
