/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef TRIMESHIO_HH
#define TRIMESHIO_HH

#include "trimesh.hh"

#include <Rectmat.hh>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <iostream>
#    include <string>
#  else
#    include <iostream.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <iostream>
#  include <string>
#endif //HAVE_CONFIG_H

/**
 * \class trimeshIO
 * \brief reader/writer basis class for the Trimesh class
 *
 * If you want to specify your own reader/writer for a specific file type, you
 * have to create a class, inheriting the trimeshIO class. This class has 5  
 * abstract methods, which must be overriden, to obtain a fully working file
 * format reader/writer:
 *
 * \li trimeshIO::handleStreams()
 * \li trimeshIO::read(const string&)
 * \li trimeshIO::read(&istream)
 * \li trimeshIO::write(const string&)
 * \li trimeshIO::write(&ostream)
 */
class trimeshIO 
{
private:
	void reverseBytes(char* seq, const int len);

protected:

  /**
   * \brief flag set \c true if doubletts should be removed during the reading 
   * of objects
   */
  bool removeDoubletts;

  /**
   * \brief doublett index mapping the file vertex indices to the actual vertex 
   * index
   */
  map<int,int> doublettix;

	/**
	 * \brief contains any data that is attached to the vertices
	 *
	 * Allthough the data is stored in floating point format, it is converted to
	 * double precision in the internal representation because of compatibility to
	 * the other modules.
	 */
	RazorBack::Rectmat_<double> fData;

	/**
	 * \brief contains the column headers
	 *
	 * Each column of the fData matrix can be equipped with a column title
	 */
	vector<string> fColumnHeads;


  /**
   * \brief mirror type definition of the mesh structure's graph type
   */
  typedef trimesh_traits::Graph_t mesh_t;
  
  /**
   * \brief reference to a Trimesh structure
   */
  Trimesh* mesh_structure;

  /**
   * \brief provides access to the protected member Mesh of the mesh_structure
   */
  mesh_t& getMesh() { return mesh_structure->Mesh; }

  /**
   * \brief provides access to the protected member Triangles of the mesh_structure
   */
  vector<Triangle>& getTriangles() { return mesh_structure->Triangles; };

  /**
   * \brief provides access to the protected member vertFaceIndex of the 
   * mesh_structure
   */
  map<int, vector<int> >& getvertfaceIndex() 
  { 
    return mesh_structure->vertfaceIndex;
  };

  /**
   * \brief referrence to the list of vertex coordinates
   */
  trimesh_traits::VertexDataMap* data_map;

  /**
   * \brief referrence to the list of edge lengths (weights)
   */
  trimesh_traits::EdgeLengthMap* edgelengths_map;

	//*****************************************************************************
	// STREAM WRITING METHODS
	//*****************************************************************************
	
	/**
	 * \brief writes the basic datatypes into binary streams regarding the
	 * byte order.
	 *
	 * \return number of written bytes
	 *
	 * \param stream the output stream where the data is written.
	 * \param value contains the value to be written
	 * \param reverse indicates, if the value should be written with byte reversal
	 *                order.
	 */
	size_t writeInt(ostream& stream, const int value, bool reverse=false);
	size_t writeShort(ostream& stream, const short value, bool reverse=false);
	size_t writeDouble(ostream& stream, const double value, bool reverse=false);
	size_t writeFloat(ostream& stream, const float value, bool reverse=false);

	/**
	 * \brief writes arrays of the basic datatypes into binary streams regarding 
	 * the byte order.
	 *
	 * \return number of written bytes
	 *
	 * \param stream the output stream where the data is written.
	 * \param value pointer to the array of data
	 * \param len number of elements in the array
	 * \param reverse indicates, if the value should be written with byte reversal
	 *                order.
	 */
	size_t writeIntArr(ostream& stream, const int* value, int len, 
										 bool reverse=false);
	size_t writeShortArr(ostream& stream, const short* value, int len, 
											 bool reverse=false);
	size_t writeDoubleArr(ostream& stream, const double* value, int len,
												bool reverse=false);
	size_t writeFloatArr(ostream& stream, const float* value, int len,
											 bool reverse=false);

	/**
	 * \brief reads the basic datatypes into binary streams regarding the
	 * byte order.
	 *
	 * \return number of read bytes
	 *
	 * \param stream the input stream from which the data should be read.
	 * \param value receives the data value
	 * \param reverse indicates, if the value should be written with byte reversal
	 *                order.
	 */
	size_t readInt(istream& stream, int& value, bool reverse=false);
	size_t readShort(istream& stream, short& value, bool reverse=false);
	size_t readDouble(istream& stream, double& value, bool reverse=false);
	size_t readFloat(istream& stream, float& value, bool reverse=false);

	/**
	 * \brief reads arrays of the basic datatypes into binary streams regarding 
	 * the byte order.
	 *
	 * \return number of read bytes
	 *
	 * \param stream the input stream from which the data should be read.
	 * \param value pointer to the array reveiving the data values
	 * \param len number of elements in the array
	 * \param reverse indicates, if the value should be written with byte reversal
	 *                order.
	 */
	size_t readIntArr(istream& stream, int* value, int len, bool reverse=false);
	size_t readShortArr(istream& stream, short* value, int len,bool reverse=false);
	size_t readDoubleArr(istream& stream, double* value, int len,
											 bool reverse=false);
	size_t readFloatArr(istream& stream, float* value, int len, bool reverse=false);

public:

  trimeshIO();
  /**
   * \brief constructor expacting a reference to a Trimesh structure
   */
  trimeshIO(Trimesh* aTrimesh);

  /**
   * \brief denotes, if the reader/writer has a set of methods to handle streams
   *
   * Some file formats are using different files to store the vertices and faces.
   * These formats can not handle the file opening and closing with a single 
   * read from stream or write to stream method.
   *
   * Reader/writer classes for a specific format have to implement there own
   * version of this method. 
   */
  virtual bool handleStreams() const = 0;
  
  /**
   * \brief reads a mesh object from the input stream \p aStream
   *
   * Reader/writer classes for a specific format have to implement there own
   * version of this method. 
   *
   * \param aStream input stream containing the object in the file format of 
   *        the class.
   */
  virtual bool read(istream& aStream) = 0;

  /**
   * \brief writes a mesh object from the input stream \p aStream
   *
   * Reader/writer classes for a specific format have to implement there own
   * version of this method. 
   *
   * \param aStream output stream receiving the object in the file format of 
   *        the class.
   */
  virtual bool write(ostream& aStream) = 0;

  /**
   * \brief reads a mesh object from a file named \p aFile
   *
   * Reader/writer classes for a specific format have to implement there own
   * version of this method. 
   *
   * \param aFile filename of the file to read from.
   */
  virtual bool read(const string& aFile) = 0;

  /**
   * \brief writes a mesh object to a file named \p aFile
   *
   * Reader/writer classes for a specific format have to implement there own
   * version of this method.
   *
   * \param aFile filename of the file to write to.
   */
  virtual bool write(const string& aFile) = 0;
  
  /**
   * \brief sets the removing doubletts flag of the reader/writer
   *
   * Some file formats can contain doublett vertices (e.g. the MSMS vertex files).
   * To avoid anormal triangles with angles equal to zero , these doubletts have
   * be removed during reading. If there are doubletts in the vertex list, you
   * should turn on the remove doubletts flag. 
   * 
   * Normally it is turned off, because searching for doubletts can be very time
   * consuming.
   */
  bool setDoubletts(const bool value);

  /**
   * \brief retrieves the status of the removing doubletts flag
   *
   * See also : trimeshIO::setDoubletts
   */
  bool isDoublett() const { return removeDoubletts; }

	//*****************************************************************************
	// PUBLIC METHODS regarding the vertex data
	//*****************************************************************************

	/**
	 * \brief provides full access to the data matrix.
	 *
	 * \return a reference to the data matrix.
	 */
	Rectmat_<double>& data() { return fData; }
	
	/**
	 * \brief retrieves the name of the \p ix'th column
	 *
	 * \return name of the \p ix'th column
	 *
	 * \param ix index of the desired column
	 */
	string getColName(const int ix) const;
	
	/**
	 * \brief sets the name of the \p ix'th column in the matrix
	 *
	 * \return old column name
	 *
	 * \param ix index of the desired colum.
	 * \param colname new column title
	 */
	string setColName(const int ix, const string& colname);

};

#endif
