/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file common.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 09 Jun 2004, last CVS commit 2004/06/09 10:43:29
 */

#ifndef XMLUTILS_COMMON_HH
#define XMLUTILS_COMMON_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include <string>
#include <exception>

#include "xmlerrorreporter.hh"
#include "conversion.hh"

#include "xercesns.hh"

/**
 * \brief the standard XML exception
 *
 * The xmlutils exception receives a message which is returned by the what()
 * method. The message can either be a std::string or an XMLCh array.
 */

class XMLError: public std::exception
{
public:
		
	/** 
	 * \brief default constructor
	 * 
	 * \param message the error message
	 */
	XMLError(const std::string& message = "") throw() 
			: exception(), message_(message) { }

	/** 
	 * \brief default constructor with XMLCh* 
	 * 
	 * \param message the error message
	 */
	XMLError(const XMLCh* message) throw() 
			: exception(), message_() { message_ = XMLStr::convert(message); }

	/** 
	 * \brief destructor, inherited from exception
	 */
	~XMLError() throw() { }
	/** 
	 * \brief retrieves the error message
	 * 
	 * This method fullfills the std::exception guideline to return the error
	 * message with the method what().
	 * 
	 * \return the error message
	 */
	virtual const char* what() throw() { return message_.c_str(); }
		
private:

	/** 
	 * \brief the error message
	 */
	std::string message_;
};


/**
 * \brief exception for invalid string conversions
 */
class InvalidConversion: public XMLError
{
public:
  InvalidConversion() throw() 
      : XMLError("XML string conversion not value") { }
};

/**
 * \brief syntax error exception
 *
 * This is the base class for syntax warnings and errors. It owns an
 * ErrorReporter object that includes all the syntax exceptions that happened
 * during a specific reading or writing operation. The reporter object is
 * passed in the constructor and is destroyed upon destruction. getReport()
 * retrieves the error reporter.
 */
class SyntaxException: public XMLError
{
public:

  /** 
   * \brief initializing constructor
   * 
   * \param reporter the reporter object which contains all the collected
   *                 errors
   */
	SyntaxException(ErrorReporter* reporter)
			: XMLError("Syntax problems"), reporter_(reporter) { }

	/** 
   * \brief default destructor
   */
  ~SyntaxException() throw() { delete reporter_; }

  /** 
   * \brief retrieves the error reporter
   * 
   * \return a reference to the error reporter that holds all the data
   */
	const ErrorReporter& getReport() const { return *reporter_; }

private:

	/** 
   * \brief the instance of the error repoter
   */
	ErrorReporter* reporter_;
};

/**
 * \brief representation of a syntax error
 */
class XMLSyntaxError: public SyntaxException
{
public:
	XMLSyntaxError(ErrorReporter* reporter) : SyntaxException(reporter) { }
};

/**
 * \brief representation of a syntax warning
 */
class XMLSyntaxWarning: public SyntaxException
{
public:
	XMLSyntaxWarning(ErrorReporter* reporter) : SyntaxException(reporter) { }
};


#endif // XMLUTILS_COMMON_HH
