/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "conversion.hh"

#include <xercesc/util/XMLUniDefs.hpp>
//#include <iostream>

#include "common.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
XERCESNS_USE
#endif 


string XMLStr::FORMAT_INT    = string("%d");
string XMLStr::FORMAT_DOUBLE = string("%e");

string XMLStr::FORMAT_TRUE   = string("true");
string XMLStr::FORMAT_FALSE  = string("false");

int XMLStr::CONVERSION_BUFFERSIZE = 128;

//******************************************************************************
// POSSIBLE TRUE AND FALSE TOKENS
//******************************************************************************

const int TRUEFALSE_PAIRS = 3;

const XMLCh TRUE_VALUES[TRUEFALSE_PAIRS][5] = 
{
  { chLatin_Y, chLatin_E, chLatin_S, chNull,    chNull },
  { chLatin_O, chLatin_N, chNull,    chNull,    chNull },
  { chLatin_T, chLatin_R, chLatin_U, chLatin_E, chNull }
};

const XMLCh FALSE_VALUES[TRUEFALSE_PAIRS][6] = 
{
  { chLatin_N, chLatin_O, chNull,    chNull,    chNull,    chNull },
  { chLatin_O, chLatin_F, chLatin_F, chNull,    chNull,    chNull },
  { chLatin_F, chLatin_A, chLatin_L, chLatin_S, chLatin_E, chNull }
};
  

//******************************************************************************
XMLStr& XMLStr::operator=(const string& value) 
//******************************************************************************
{ 
	clear();
	if (value!="")
		value_ = convert(value);
	
	return *this;
}

//******************************************************************************
XMLStr& XMLStr::operator=(const XMLCh* value) 
//******************************************************************************
{ 
	clear();

	if (value)
	{
		value_ = new XMLCh[XMLString::stringLen(value)+1];
		XMLString::copyString(value_, value);
	}

	return *this;
}

//******************************************************************************
XMLStr& XMLStr::fromValue(const int& value)
//******************************************************************************
{
	clear();

  char buffer[CONVERSION_BUFFERSIZE];
  sprintf(buffer, FORMAT_INT.c_str(), value);
  value_ = XMLString::transcode(buffer);

  return *this;
}

//******************************************************************************
XMLStr& XMLStr::fromValue(const double& value)
//******************************************************************************
{
	clear();

  char buffer[CONVERSION_BUFFERSIZE];
  sprintf(buffer, FORMAT_DOUBLE.c_str(), value);
  value_ = XMLString::transcode(buffer);

  return *this;
}

//******************************************************************************
XMLStr& XMLStr::fromValue(const bool& value)
//******************************************************************************
{
	clear();

  if (value)
    value_ = convert(FORMAT_TRUE);
  else
    value_ = convert(FORMAT_FALSE);

  return *this;
}

//******************************************************************************
bool XMLStr::operator==(const XMLCh* query) const
//******************************************************************************
{
	return XMLString::compareIString(value_,query)==0;
}

//******************************************************************************
bool XMLStr::operator==(const std::string& query) const
//******************************************************************************
{
	XMLStr qstr = query;
	return XMLString::compareIString(value_,qstr.value_)==0;
}

//******************************************************************************
void XMLStr::collapse()
//******************************************************************************
{
  XMLString::collapseWS(value_);
  adjust();
}

//******************************************************************************
bool XMLStr::wp() const
//******************************************************************************
{
  return XMLString::isAllWhiteSpace(value_);
}

//******************************************************************************
XMLStr::operator int () const
//******************************************************************************
{
  int value;
  char* cstr = XMLString::transcode(value_);
  if (!sscanf(cstr, "%d", &value)) throw InvalidConversion();

  delete[] cstr;
  return value;
}

//******************************************************************************
XMLStr::operator double () const
//******************************************************************************
{
  double value;
  char* cstr = XMLString::transcode(value_);
  if (!sscanf(cstr, "%lf", &value)) throw InvalidConversion();

  delete[] cstr;
  return value;
}

//******************************************************************************
XMLStr::operator bool () const
//******************************************************************************
{
  XMLStr query = value_;
  XMLString::upperCase(query.value_);
  
  for(int i=0;i<TRUEFALSE_PAIRS;++i)
  {
    if (XMLString::compareIString(query.value_,TRUE_VALUES[i])==0)  
      return true;
    if (XMLString::compareIString(query.value_,FALSE_VALUES[i])==0) 
      return false;
  }

  //no conversion performed
  throw InvalidConversion();
}

//******************************************************************************
void XMLStr::clear()
//******************************************************************************
{
	if (value_) 
	{
		delete[] value_; 
		value_ = 0;
	}
}

//******************************************************************************
void XMLStr::adjust()
//******************************************************************************
{
  XMLCh* old = value_;
  value_ = new XMLCh[XMLString::stringLen(old)+1];
  XMLString::copyString(value_, old);
  
  delete[] old;
}

//******************************************************************************
string XMLStr::convert(const XMLCh* xmlstring)
//******************************************************************************
{
  if (xmlstring)
  {
    char*  buffer = XMLString::transcode(xmlstring);
    string result = buffer;
    delete[] buffer;
    
    return result;
  }
  else 
    return "";
}

//******************************************************************************
XMLCh* XMLStr::convert(const std::string& stdstring)
//******************************************************************************
{
	return XMLString::transcode(stdstring.c_str());
}

//******************************************************************************
ostream& operator<<(ostream& stream, const XMLStr& str)
//******************************************************************************
{
  stream << XMLStr::convert(str.value_);
  return stream;
}

//******************************************************************************
istream& operator>>(istream& stream, XMLStr& str)
//******************************************************************************
{
  string value;
  stream >> value;
  str = value;

  return stream;
}

