/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file conversion.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 09 Jun 2004, last CVS commit 2004/06/09 09:54:37
 */

#ifndef CONVERSION_HH
#define CONVERSION_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#else
#endif

#include <xercesc/util/XMLString.hpp>
#include <string>

#include "xercesns.hh"

/**
 * \brief a conversion class between XMLCh* and std::string
 * 
 * This class provides a convenient access to Xerces-c XML character
 * strings. The class stores internally a character sequence as a XMLCh*
 * pointer but provides read and write access for XMLCh* and std::string
 * types.
 *
 * The class can be initialized via its constructors XMLStr(const XMLCh*),
 * XMLStr(const std::string&) and XMLStr() while accessed to the character
 * sequence is provided via the conversion operators. The conversion operator
 * for XMLCh* returns a pointer to the character sequence, while the conversion
 * for std::string provides a transcoded copy of the xml string. If the string
 * is empty or not can be queried by the empty() method.
 *
 * The conversion between XMLCh* and std::string is performed by the public
 * static methods convert(const XMLCh*) and convert(const std::string&). Beside
 * that the class provides bidirectional conversion with integer, floating
 * point and boolean values. How the elements are transformed into characters
 * can be controlled by the FORMAT_INT, FORMAT_DOUBLE, FORMAT_TRUE and
 * FORMAT_FALSE static members which are initialized with meaningfull values,
 * but can be adjusted by the user. Values can be assigned by the
 * fromValue(const int&), fromValue(const& double) and fromValue(const& bool)
 * methods. For the readout conversion operations are provided.
 *
 * The XMLStr provides the possibility to collapse the whitspaces in the
 * character sequence (collapse()). If the string consist of white spaces only
 * can be checked with wp().
 */

class XMLStr
{
public:

	//****************************************************************************
	// CONSTRUCTORS, DESTRUCTOR and ASSIGNMENT
	//****************************************************************************

	/** 
	 * \brief default constructor
	 */
	XMLStr() : value_(0) { }
	/** 
	 * \brief constructor, initializing from a XMLCh*
	 * 
	 * \param value a pointer to a XML character sequence
	 *
	 * \sa XMLStr(const std::string&), 
   *     operator=(const XMLCh*)
	 */
	XMLStr(const XMLCh* value) : value_(0) { operator=(value); }
	/** 
	 * \brief constructor, initializing from std::string
	 * 
	 * \param value a std::string containing the data
	 *
	 * \sa XMLStr(const XMLCh* value), 
   *     operator=(const std::string&)
	 */
	XMLStr(const std::string& value) : value_(0) { operator=(value); }

	/** 
	 * \brief destructor
	 * 
	 * The destructor calls clean() to delete the data from the memory
	 */
	~XMLStr() { clear(); }

	/** 
	 * \brief assignment from a XMLCh* pointer
	 * 
	 * \param value a pointer to a XML character sequence
	 * 
	 * \return a reference to \c this object
	 *
	 * \sa operator=(const std::string&),
   *     fromValue(const int&),
   *     fromValue(const double&),
   *     fromValue(const bool&),
	 *     XMLStr=(const XCh* value)
	 */
	XMLStr& operator=(const XMLCh* value); 
	/** 
	 * \brief assignment from a std::string
	 *
	 * Assigns a copy of the \p value string. If the string is empty (="") the
	 * internal data is set to NULL.
	 * 
	 * \param value a std::string containing the data
	 * 
	 * \return a reference to \c this object
	 *
	 * \sa operator=(const XMLCh* value), 
   *     fromValue(const int&),
   *     fromValue(const double&),
   *     fromValue(const bool&),
   *     XMLStr(const std::string&)
	 */
  XMLStr& operator=(const std::string& value); 

	/** 
	 * \brief assignment from an integer
	 *
	 * Assigns a copy of the \p value integer.
	 * 
	 * \param value an integer that should be stored by the string
	 * 
	 * \return a reference to \c this object
	 *
	 * \sa operator=(const XMLCh* value), 
   *     operator=(const std::string&),
   *     fromValue(const double&),
   *     fromValue(const bool&),
	 */
  XMLStr& fromValue(const int& value); 
	/** 
	 * \brief assignment from a double
	 * 
	 * \param value a double value that should be stored in the string
	 * 
	 * \return a reference to \c this object
	 *
	 * \sa operator=(const XMLCh* value), 
   *     operator=(const std::string&),
   *     fromValue(const int&),
   *     operator=(const bool&),
	 */
  XMLStr& fromValue(const double& value);
	/** 
	 * \brief assignment from a boolean
	 * 
	 * \param value a boolean value that should be stored in the string
	 * 
	 * \return a reference to \c this object
	 *
	 * \sa operator=(const XMLCh* value), 
   *     operator=(const std::string&),
   *     fromValue(const int&),
   *     fromValue(const double&),
	 */
  XMLStr& fromValue(const bool& value);

	//****************************************************************************
	// CONVERSION
	//****************************************************************************

	/** 
	 * \brief conversion into a XMLCh* string
	 * 
	 * \return the pointer to the XMLCh* data
	 * 
	 * \sa operator std::string(), 
   *     operator int (), operator double (), operatpr bool ()
	 */
	operator const XMLCh* () const { return value_; }
	/** 
	 * \brief conversion into a XMLCh* string
	 * 
	 * \return a copy of the xml character data as std::string
	 * 
	 * \sa operator XMLCh*(),
   *     operator int (), operator double (), operatpr bool ()
	 */
  operator std::string  () const { return convert(value_); }

  /** 
   * \brief conversion into int
   *
   * \return the content of the XML string as an integer value
   *
   * \exception InvalidConversion if the string does not represent an integer 
   *                              value
	 * \sa operator XMLCh*(), operator std::string(),
   *     operator double (), operatpr bool ()
   */
  operator int () const;
  /** 
   * \brief conversion into double
   *
   * \return the content of the XML string as an double value
   *
   * \exception InvalidConversion if the string does not represent an double 
   *                              value
	 * \sa operator XMLCh*(), operator std::string(),
   *     operator int (), operatpr bool ()
   */
  operator double () const;
  /** 
   * \brief conversion into bool
   *
   * \return the content of the XML string as an boolean value
   *
   * \exception InvalidConversion if the string does not represent an boolean 
   *                              value
	 * \sa operator XMLCh*(), operator std::string(),
   *     operator int (), operator double ()
   */
  operator bool () const;
	
	/** 
	 * \brief conversion from XMLCh* to std::string
	 * 
	 * \param xmlstring a XML character sequence
	 * 
	 * \return a std::string that contains a transcoded copy of the XMLCh* string.
	 */
	static std::string convert(const XMLCh* xmlstring);
	/** 
	 * \brief conversion from std::string to XMLCh*
	 * 
	 * The method returns a freshly allocated XMLCh* string. The caller of the
	 * method has to take care of the deallocation.
	 * 
	 * \param stdstring a std::string
	 * 
	 * \return a XMLCh* string that contains a transcoded copy of the std::string.
	 */
	static XMLCh*  convert(const std::string& stdstring);

  //****************************************************************************
	// COMPARISON
	//****************************************************************************

	/** 
	 * \brief comparing with XMLStr
	 * 
	 * \param query the XMLStr object that should be compared with this one
	 * 
	 * \return \c true if both strings contain the same data
	 */
	bool operator==(const XMLStr& query) const { return *this == query.value_; }
	/** 
	 * \brief comparing with XMLCh*
	 * 
	 * \param query the XMLCh* string that should be compared with this one
	 * 
	 * \return \c true if both strings contain the same data
	 */
	bool operator==(const XMLCh* query) const;
	/** 
	 * \brief comparing with std::string
	 * 
	 * \param query the std::string string that should be compared with this one
	 * 
	 * \return \c true if both strings contain the same data
	 */
	bool operator==(const std::string& query) const;

  //****************************************************************************
  // OPERATIONS
  //****************************************************************************

  /** 
   * \brief removes all trailing and appended white spaces
   * 
   * Whitespaces are the following characters: space, tab, carriage return,
   * line feed.
   */
  void collapse();

	//****************************************************************************
	// PROPERTY METHODS
	//****************************************************************************

	/** 
	 * \brief indicates if the XMLStr is empty or not
	 * 
	 * \return \c true if the XMLStr contains no data, \c false otherwise
	 */
	bool empty() const { return !value_; }

	/** 
	 * \brief indicates if the XMLStr contains only white spaces
	 * 
	 * \return \c true if the XMLStr contains only white spaces, \c false otherwise
	 */
	bool wp() const;

  /**
   * \brief the format specifier for a int -> XMLStr conversion
   *
   * The default value is \c \%d.
   */
  static std::string FORMAT_INT;
  /**
   * \brief the format specifier for a double -> XMLStr conversion
   *
   * The default value is \c \%e.
   */
  static std::string FORMAT_DOUBLE;

  /**
   * \brief the string that represents a \c true value
   *
   * The default value is \c TRUE.
   */
  static std::string FORMAT_TRUE;
  /**
   * \brief the string that represents a \c false value
   *
   * The default value is \c FALSE.
   */
  static std::string FORMAT_FALSE;

  static int CONVERSION_BUFFERSIZE;

private:
	
  //****************************************************************************
  // PRIVATE METHODS
  //****************************************************************************

	/** 
	 * \brief cleans the allocated memory
	 * 
	 * The method deletes the character sequence and sets the value_ parameter to
	 * NULL. Thus the XMLStr is empty after this operation.
	 */
	void clear();

  /** 
   * \brief adjust the string length to the current requirements
   * 
   * This method copies the new string up to its current length and releases
   * the old one.
   * 
   * \return the number of deleted characters
   */
  void adjust();
	
  //****************************************************************************
  // PRIVATE MEMBERS
  //****************************************************************************

	/**
	 * \brief the internal XML character sequence
	 */
	XMLCh* value_;

  friend std::ostream& operator<<(std::ostream&, const XMLStr&);

};

std::ostream& operator<<(std::ostream& stream, const XMLStr& str);
std::istream& operator>>(std::istream& stream, XMLStr& str);

#endif // CONVERSION_HH
