/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file xmlerrorreporter.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 09 Jun 2004, last CVS commit 2004/06/09 09:41:38
 */

#ifndef XMLERRORREPORTER_HH
#define XMLERRORREPORTER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <list>
#else
#  include <list.h>
#endif

#include <string>

#include <xercesc/util/XercesDefs.hpp>
#include <xercesc/sax/ErrorHandler.hpp>
#include <xercesc/dom/DOMErrorHandler.hpp>

#include "xercesns.hh"

/** 
 * \brief a type for a syntax error
 * 
 * A syntax error nows the position in the text (row, col) and the cause of the
 * error (message). It has a standard constructor that accepts a row, column
 * and the message (SyntaxError(int,int, const std::string&).
 *
 */
struct SyntaxError
{
	int row;
	int col;
	std::string message;

	SyntaxError(int r, int c, const std::string& msg) 
			: row(r), col(c), message(msg) { }
};

class ErrorReporter : public XERCESNS ErrorHandler, 
											public XERCESNS DOMErrorHandler
{
public:

	//****************************************************************************
	// PUBLIC TYPE DEFINITION
	//****************************************************************************

	//@{
	/**
	 * \brief Syntax error list type + iterators
	 */
	typedef std::list<SyntaxError> errors_t;
	typedef errors_t::iterator errors_it;
	typedef errors_t::const_iterator errors_cit;
	//@}

	//****************************************************************************
	// CONSTRUCTOR
	//****************************************************************************

  /** 
	 * \brief default constructor
	 */
	ErrorReporter() 
			: ErrorHandler(), DOMErrorHandler(), errors_(), warnings_() { }
	
	//****************************************************************************
	// ERROR HANDLER INTERFACE
	//****************************************************************************

	/** 
	 * \brief catches all the warnings
	 * 
	 * \param toCatch the parsing error exception
	 *
	 * \sa error(const SAXParseException&), fatalError(const SAXParseException&)
	 */
	void warning(const XERCESNS SAXParseException& toCatch);
	/** 
	 * \brief catches all the errors
	 * 
	 * \param toCatch the parsing error exception
	 *
	 * \sa fatalError(const SAXParseException&), warning(const SAXParseException&)
	 */
	void error(const XERCESNS SAXParseException& toCatch);
	/** 
	 * \brief catches all the fatal errors
	 * 
	 * \param toCatch the parsing error exception
	 *
	 * \sa error(const SAXParseException&), warning(const SAXParseException&)
	 */
	void fatalError(const XERCESNS SAXParseException& toCatch);

	/** 
	 * \brief handles DOM errors
	 * 
	 * \param domError the last DOM error
	 * 
	 * \return \c true if the method could resolve the error, or otherwise 
	 *         \c false.
   */
	bool handleError(const XERCESNS DOMError& domError);

	/** 
	 * \brief resets the warnings and errors
	 */
	void resetErrors();
	
	//****************************************************************************
	// PROPERTY METHODS
	//****************************************************************************

	/** 
	 * \brief counts the reported errors
	 * 
	 * \return the number of reported errors.
	 *
	 * \sa errors_
	 */
	bool getNErrors() const { return errors_.size(); }

	//@{
	/** 
	 * \brief begin and end iterators for the errors
	 * 
	 * \return the begin and end of the error sequence.
	 *
	 * \sa errors_
	 */
	errors_cit beginError() const { return errors_.begin(); }
	errors_cit endError()   const { return errors_.end(); }
	//@}

	/** 
	 * \brief counts the reported warnings
	 * 
	 * \return the number of reported warnings.
	 *
	 * \sa warnings_
	 */
	bool getNWarnings() const { return warnings_.size(); }

	//@{
	/** 
	 * \brief begin and end iterators for the warnings
	 * 
	 * \return the begin and end of the warning sequence.
	 *
	 * \sa warnings_
	 */
	errors_cit beginWarning() const { return warnings_.begin(); }
	errors_cit endWarning()   const { return warnings_.end(); }
	//@}

private:
	
  /**
	 * \brief list of errors
	 */
	errors_t errors_;
  /**
	 * \brief list of warnings
	 */
	errors_t warnings_;
};


#endif //XMLERRORREPORTER
