/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file xmlreader.hh
 *
 * \author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 09 Jun 2004, last CVS commit 2004/06/09 09:41:38
 */

#ifndef XMLCONVENIENTREADER_HH
#define XMLCONVENIENTREADER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <iostream>
#else
#  include <iostream.h>
#endif

#include <string>
#include <xercesc/parsers/XercesDOMParser.hpp>

#include <xercesns.hh>

XERCES_FORWARD(InputSource);
XERCES_FORWARD(DOMDocument);  

/**
 * \brief a convenient XML DOM document reader
 *
 * The XMLConvenientReader class provides the means to read and parse XML
 * documents from various input sources (files, URLs, input streams and C
 * strings). It is also possible to set some flags that control the process of
 * the XML parsing: 
 * - handling of namespaces (doesNameSpaces(), doNameSpaces(bool))
 * - xml document validation (doesValidation(), doValidation(bool))
 * - xml scheme processing (doesSchema(), doSchema(bool))
 *
 * After successfull processing and parsing the DOM document is available via
 * getDocument, which automatically transfers the parsed document structure
 * into the responsibility of the user. A general piece of code that uses
 * XMLConvenientReader to read an XML document is something like ...
 * 
 * \code
 *  ...
 * 
 * XMLConvenientReader reader;
 * reader.setSourceFile("myxmldocument.xml");
 * // or for URL sources
 * reader.setSourceURL("http://myhost.net/myxmldocument.xml");
 * // or for istream sources
 * reader.setSource(cout, 4200);
 * // or for target strings
 * reader.setSource("<answer>42</answer>", 19);
 *
 * try
 * {
 *   reader.parse();
 *   DOMDocument* document = reader.getDocument();
 *   
 *   ... 
 *
 *   document->release();
 * }
 * catch(XMLException& xmlerr)
 * {
 *   cerr << XMLStr::convert(xmlerr.getMessage()) << endl;
 * }
 * catch(DOMException& domerr)
 * {
 *   cerr << domerr.code << endl;
 * }
 * catch(SyntaxException& synerr)
 * {
 *		cerr << synerr.what());
 *    cerr << XMLUtils::printSyntaxError(synerr.getReport());
 * }
 * \endcode
 * 
 * 
 */
 class XMLConvenientReader
{
public:

	//****************************************************************************
	// CONSTRUCTOR
	//****************************************************************************

	/** 
	 * \brief default constructor
	 * 
	 * \param donamespaces turnes checking of namespaces on/off
	 * \param doschema turnes checking of schema on/off
	 * \param dovalidation turnes validation on/off
	 */
	XMLConvenientReader(bool donamespaces=false, bool doschema=false, 
						bool dovalidation=false);

	~XMLConvenientReader();

	//****************************************************************************
	// OPERATIONS
	//****************************************************************************

	/** 
	 * \brief parses the XML source
	 * 
	 * This method invokes the parser on the given source. The resulting document
	 * can be retrieved from getDocument();
	 *
	 * \exception XMLException thrown if an XML error occurs
	 * \exception DOMException thrown if an DOM error occurs
	 * \exception SyntaxException thrown if any syntax errors have been 
	 *                            detected
	 *
	 * \sa getDocument()
	 */
	void parse();

	//****************************************************************************
	// PROPERTY METHODS
	//****************************************************************************

	/** 
	 * \brief retrieves the root XML node
	 * 
   * After the call to this method, the document is in the responsibilty of the
   * caller.
   *
	 * \return the root XML node of the read structure
	 *
	 * \sa parse()
	 */
	XERCESNS DOMDocument* getDocument() { return parser_->adoptDocument(); }
	
	/** 
	 * \brief queries the handling of namespaces
	 * 
	 * \return \c true if the parser will consider namespaces
	 *
	 * \sa doNamespaces(bool)
	 */
	bool doesNamespaces() const { return parser_->getDoNamespaces(); }
	/** 
	 * \brief defines the handling of namespaces
	 * 
	 * \param on \c true if the parser should consider namespaces
	 *
	 * \sa doesNamespaces()
	 */
	void doNamespaces(bool on) { parser_->setDoNamespaces(on); }

	/** 
	 * \brief queries the handling of schema
	 * 
	 * \return \c true if the parser will consider schema
	 *
	 * \sa doSchema(bool)
	 */
	bool doesSchema() const { return parser_->getDoSchema(); }
	/** 
	 * \brief defines the handling of schema
	 * 
	 * \param on \c true if the parser should consider schema
	 *
	 * \sa doesSchema()
	 */
	void doSchema(bool on) { parser_->setDoSchema(on); }

	/** 
	 * \brief queries the handling of validation
	 * 
	 * \return \c true if the parser will consider validation
	 *
	 * \sa doValidation(bool)
	 */
	bool doesValidation() const { return parser_->getDoValidation(); }
	/** 
	 * \brief defines the handling of validation
	 * 
	 * \param on \c true if the parser should consider validation
	 *
	 * \sa doesValidation()
	 */
	void doValidation(bool on) { parser_->setDoValidation(on); }

	/** 
	 * \brief sets the input source to an URL
	 * 
	 * \param url the name of the input URL
	 * 
	 * \return \c true if the source could be set
	 *
	 * \sa setSourceFile(const std::string&), setSource(std::istream&), 
	 *     setSource(const char*)
	 */
	bool setSourceURL(const std::string& url);
	/** 
	 * \brief sets the input source to a filename
	 * 
	 * \param filename the name of the input file
	 * 
	 * \return \c true if the source could be set
	 *
	 * \sa setSourceURL(const std::string&), setSource(std::istream&), 
	 *     setSource(const char*)
	 */
	bool setSourceFile(const std::string& filename);
	/** 
	 * \brief sets the input source to an standard input stream
	 * 
	 * \param stream a standard input stream
	 * \param size   the number of characters that should be taken from the 
	 *               stream. -1 means read to the end
	 * 
	 * \return \c true if the source could be set
	 *
	 * \sa setSourceURL(const std::string&), setSourceFile(const std::string&) 
	 *     setSource(const char*)
	 */
	bool setSource(std::istream& stream, int size = -1);
	/** 
	 * \brief sets the input source to an XML text
	 * 
	 * \param xmltext C string that represents the content of the XML input
	 * \param size   the number of characters that should be taken
	 * 
	 * \return \c true if the source could be set
	 *
	 * \sa setSourceURL(const std::string&), setSourceFile(const std::string&) 
	 *     setSource(std::istream&)
	 */
	bool setSource(const unsigned char* xmltext, int size);

private:

	//****************************************************************************
	// PRIVATE METHODS
	//****************************************************************************

	/** 
	 * \brief removes the input source if assigned
	 */
	void clearSource();

	//****************************************************************************
	// PRIVATE MEMBERS
	//****************************************************************************

	/**
	 * \brief a pointer to the input source
	 */
	XERCESNS InputSource* source_;
	/**
	 * \brief the input parser
	 */
	XERCESNS XercesDOMParser* parser_;
};

#endif // XMLCONVENIENTREADER_HH
