/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "xmlutils.hh"

#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/dom/DOM.hpp>

#ifdef HAVE_STD
#  include <sstream>
#else
#  include <sstream.h>
#endif


#ifdef HAVE_NAMESPACES
using namespace std;
XERCESNS_USE
#endif 

struct XercesInitializer
{
	XercesInitializer()  { 
		try
		{
			XMLUtils::xmlInit();
		}
		catch (XMLError err)
		{
			cerr << err.what() << endl;
		}
	}

	~XercesInitializer() { XMLUtils::xmlTerminate(); }
};

static XercesInitializer xerces_init;

//#include <iostream>

//******************************************************************************
void XMLUtils::xmlInit()
//******************************************************************************
{
	try
	{
		XMLPlatformUtils::Initialize();
	}
	catch(const XMLException& toCatch)
	{
		throw XMLError(toCatch.getMessage());
	}
}

//******************************************************************************
void XMLUtils::xmlTerminate()
//******************************************************************************
{
	XMLPlatformUtils::Terminate();
}

//******************************************************************************
DOMDocument* XMLUtils::createDocument(const std::string& root)
//******************************************************************************
{
  XMLStr root_str = root;
	return DOMImplementationRegistry::
		getDOMImplementation(XMLStr("LS"))->createDocument(0, root_str, 0);
}

//*******************************************************************************
DOMElement* XMLUtils::createElement(DOMElement* root, const string& name)
//*******************************************************************************
{
	DOMElement* node = root->getOwnerDocument()->createElement(XMLStr(name));
	root->appendChild(node);

	return node;
}

//*******************************************************************************
DOMElement* XMLUtils::createTextNode(DOMElement* root, 
																		 const string& name, const string& contents)
//*******************************************************************************
{
	DOMDocument* doc = root->getOwnerDocument();

	DOMElement* node = doc->createElement(XMLStr(name));
	node->appendChild(doc->createTextNode(XMLStr(contents)));

	root->appendChild(node);
	return node;
}

//******************************************************************************
const DOMElement* XMLUtils::findElement(const XERCESNS DOMElement* root,
																				const string& name)
//******************************************************************************
{
  const DOMElement* result = 0;
  XMLStr tagname = name;

  DOMNode* child = root->getFirstChild();
  while ((child) && (!result))
  {
    result = elementByName(child, name);
    child = child->getNextSibling();
  }

  return result;
}

//******************************************************************************
const DOMElement* XMLUtils::elementByName(const DOMNode* node, 
																					const string& name)
//******************************************************************************
{
  if ((node->getNodeType()==DOMNode::ELEMENT_NODE) &&
      (XMLStr(node->getNodeName())==name))  
    return dynamic_cast<const DOMElement*>(node);
  else
    return 0;
}

//******************************************************************************
XMLStr XMLUtils::getAttribute(const DOMElement* node,
															const std::string& name, const XMLStr& def)
//******************************************************************************
{
	XMLStr attname = name;
  if (node->hasAttribute(attname))
		return XMLStr(node->getAttribute(attname));
	else
		return def;
	
}

//*******************************************************************************
XMLStr XMLUtils::getText(const DOMNode* node)
//*******************************************************************************
{
	DOMNode* child = node->getFirstChild();
	while (child)
	{
		if (child->getNodeType()==DOMNode::TEXT_NODE)
			return XMLStr(child->getNodeValue());
		
		child = child->getNextSibling();
	}

	return XMLStr();
}

//******************************************************************************
void XMLUtils::collapseTextNodes(DOMNode* node)
//******************************************************************************
{
  DOMDocument* doc = node->getOwnerDocument();

  DOMNode* child = node->getFirstChild();
  while (child)
    if (child->getNodeType()==DOMNode::TEXT_NODE)
    {
      DOMText* text = dynamic_cast<DOMText*>(child);
      child = child->getNextSibling();

      XMLStr data = text->getNodeValue();
      data.collapse();
      DOMText* new_text = doc->createTextNode(data);
      
      node->replaceChild(new_text,text)->release();
    }
    else
    {
      if (child->hasChildNodes()) collapseTextNodes(child);
      child = child->getNextSibling();
    }
}

//******************************************************************************
void XMLUtils::ignoreWPTextNodes(DOMNode* node)
//******************************************************************************
{
  DOMDocument* doc = node->getOwnerDocument();

  DOMNode* child = node->getFirstChild();
  while (child)
    if (child->getNodeType()==DOMNode::TEXT_NODE)
    {
      DOMText* text = dynamic_cast<DOMText*>(child);
      child = child->getNextSibling();

      if (XMLString::isAllWhiteSpace(text->getNodeValue()))
        node->removeChild(text)->release();
    }
    else
    {
      if (child->hasChildNodes()) ignoreWPTextNodes(child);
      child = child->getNextSibling();
    }
  
}

//******************************************************************************
string XMLUtils::printSyntaxError(const ErrorReporter& reporter)
//******************************************************************************
{
  stringstream out;

  out << reporter.getNErrors()   << " errors and " 
      << reporter.getNWarnings() << " warnings detected:" << endl;

  ErrorReporter::errors_cit e;
  for(e=reporter.beginError();e!=reporter.endError();++e)
    out << "  " << e->row << "/" << e->col << ": "
        << e->message << endl;

  ErrorReporter::errors_cit w;
  for(w=reporter.beginWarning();w!=reporter.endWarning();++w)
    out << "  " << w->row << "/" << w->col << ": "
        << w->message << endl;

  return out.str();
}
