/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file xmlutils.hh
 *
 * Author : Christian Hofbauer (christian.hofbauer@pharma.novartis.com)\n
 * started at 09 Jun 2004, last CVS commit 2004/06/09 09:41:17
 */

#ifndef XMLUTILS_HH
#define XMLUTILS_HH

#include "common.hh"
#include "conversion.hh"
#include "xmlreader.hh"
#include "xmlwriter.hh"
#include "xmlerrorreporter.hh"

XERCES_FORWARD(DOMDocument);

/** 
 * \brief a static class that provides many XML functions for conveniency
 *
 * The XMLUtils class is the central class of this library, except for the
 * XMLConvenientReader and XMLConvenientWriter classes. It provides the following static
 * methods :
 * 
 * - initialization (xmlInit()) and termination (xmlTerminate()) of the
 * Xerces-C engine.
 * - convenient factory methods for the creation of DOM documents
 * (createDocument()), elements (createElement()) and text nodes
 * createTextNode().
 * - whitespace collapsing (collapseTextNodes()) and removing of white-space
 * only text nodes (ignoreWPTextNodes()).
 * - syntax error reporting (printSyntaxError())
 */
struct XMLUtils
{
	//****************************************************************************
  // ENGINE INITIALIZATION
  //****************************************************************************

  /** 
   * \brief initializes the Xerces-c XML enviroment
   *
   * \exception XMLError if the initialization went wrong
   */
  static void xmlInit();

  /** 
   * \brief terminates the Xerces-c enviroment
   */
  static void xmlTerminate();

	//****************************************************************************
  // NODE CREATION
  //****************************************************************************

  /** 
	 * \brief creates a new XML document structure
	 *
	 * \brief root the tag name of the root tag
	 * 
	 * \return a new XML DOM document object
	 */
	static XERCESNS DOMDocument* createDocument(const std::string& root);

	/** 
	 * \brief creates a new DOM element
	 * 
	 * The method creates a new DOM element and appends it as a child to the \p
	 * root element.
	 * 
	 * \param root the parent element where the new element is added as a child
	 * \param name the name of the new element
	 * 
	 * \return a pointer to the new DOM element
	 */
	static XERCESNS DOMElement* createElement(XERCESNS DOMElement* root, 
																						const std::string& name);

	/** 
	 * \brief creates a new DOM text node
	 * 
	 * The method creates a new DOM element and appends it as a child to the \p
	 * root element. The \p contents is automatically added as a text node to the
	 * \p name element.
	 * 
	 * \param root the parent element where the new text node is added as a child
	 * \param name the name of the new text node
	 * \param contents the text
	 * 
	 * \return a pointer to the new DOM text node
	 */
	static XERCESNS DOMElement* createTextNode(XERCESNS DOMElement* root, 
																						 const std::string& name, 
																						 const std::string& contents);

  //****************************************************************************
  // NODE PARSING
  //****************************************************************************

  /** 
   * \brief searches for a child node by name
   * 
   * This method parses all the childs of the given parent \p root until it
   * finds the first occurance of a node with the tag name \p name. This
   * element is returned. If no child with the proper name could be found \c
   * NULL is returned.
   *
   * \param root the parent were the child element is expected
   * \param name the name of the child element
   * 
   * \return a pointer to the child element or \c NULL if no child could be
   * found.
   */
  static const XERCESNS DOMElement* findElement(const XERCESNS DOMElement* root,
																								const std::string& name);

  /** 
   * \brief checks if a DOM node is a DOMElement of the given name
   * 
   * \param node the XML DOM node that should be checked
   * \param name the name of the desired node
   * 
   * \return the DOMElement* pointer of this node, if it has the given tagname
   */
  static const XERCESNS DOMElement* elementByName(const XERCESNS DOMNode* node,
																									const std::string& name);

	/** 
	 * \brief retrieves the value of an attribute
	 * 
	 * \param node the node that contains the attribute
	 * \param name the name of the attribute
	 * \param def  a default value if the attribute was not found.
	 * 
	 * \return an XMLStr that contains the content of the attribute or \p def
	 * if no content was found
	 */
	static XMLStr getAttribute(const XERCESNS DOMElement* node,
														 const std::string& name, 
														 const XMLStr& def=XMLStr());

	/** 
	 * \brief retrieves the corresponding text for a given node
	 * 
	 * \param node the node that contains the text as a sub text-node
	 * 
	 * \return the contents of the text sub-node
	 */
	static XMLStr getText(const XERCESNS DOMNode* node);
  
  //****************************************************************************
  // OPERATIONS
  //****************************************************************************

  /** 
   * \brief collapses all text nodes in a DOM document structure
   * 
   * This method collapses the XML strings of every text node in the DOM
   * structure beneath \p node.
   *
   * \param node the root node of the document structure that should be 
   *             collapsed
   */
  static void collapseTextNodes(XERCESNS DOMNode* node);

  /** 
   * \brief removes all white-space only text nodes
   * 
   * This method investigates every text node in the DOM structure beneath \p
   * node and removes all text entries that contain only white spaces.
   *
   * \param node the root node of the document structure that should be 
   *             cleaned from wp-only text nodes
   */
  static void ignoreWPTextNodes(XERCESNS DOMNode* node);

	//****************************************************************************
	// ERROR MANAGEMENT
	//****************************************************************************

  /** 
   * \brief pretty print a syntax error
   * 
   * \param reporter the error reporter
   *
   * \return the pretty printed error message as a string 
   */
  static std::string printSyntaxError(const ErrorReporter& reporter);
  
};

#endif // XMLUTILS_HH
