/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "xmlwriter.hh"

#ifdef HAVE_STD
#  include <memory>
#else
#  include <memory.h>
#endif

#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMWriter.hpp>
#include <xercesc/framework/MemBufFormatTarget.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>
#include <xercesc/framework/StdOutFormatTarget.hpp>
#include <xercesc/util/XMLUniDefs.hpp>

#include "common.hh"

#ifdef HAVE_NAMESPACES
using namespace std;
XERCESNS_USE
#endif 

const XMLCh EOLS[3][3]= 
{ 
	{ chLF, chNull, chNull }, 
	{ chCR, chNull, chNull },
	{ chCR, chLF,   chNull } 
};

//******************************************************************************
void XMLConvenientWriter::initialize(const std::string& encoding, 
                                     const eol_t& eol)
//******************************************************************************
{
	setEncoding(encoding);
	setEOLSequence(eol);

  DOMWriter* writer = createWriter();

	if (writer->getEncoding()) encoding_ = writer->getEncoding();
	if (writer->getNewLine())  eolsequence_ = writer->getNewLine();

  writer->release();
}

//******************************************************************************
bool XMLConvenientWriter::write() const
//******************************************************************************
{
	auto_ptr<StdOutFormatTarget> target(new StdOutFormatTarget());
	return write(target.get()); 
}

//******************************************************************************
bool XMLConvenientWriter::writeFile(const std::string& filename) const
//******************************************************************************
{
	auto_ptr<LocalFileFormatTarget> 
		target(new LocalFileFormatTarget(filename.c_str()));
  
	return write(target.get());
}

//******************************************************************************
size_t XMLConvenientWriter::write(std::ostream& stream) const
//******************************************************************************
{
  auto_ptr<MemBufFormatTarget> target(new MemBufFormatTarget());
 
	if (write(target.get()))
	{
		int length = target->getLen();
		stream.write((const char*) target->getRawBuffer(), length);
		
		return length;
	}
	else
		return 0;
}

//******************************************************************************
size_t XMLConvenientWriter::write(char* xmltext) const
//******************************************************************************
{
  auto_ptr<MemBufFormatTarget> target(new MemBufFormatTarget());
 
	if (write(target.get()))
	{
		int length = target->getLen();
		xmltext = new char[length+1];
		XMLString::copyString(xmltext, (const char*) target->getRawBuffer());
			
		return length;
	}
	else
		return 0;
}

//******************************************************************************
XMLConvenientWriter::eol_t XMLConvenientWriter::getEOLSequence() const
//******************************************************************************
{ 
	for(int i=0;i<3;++i)
		if (eolsequence_==EOLS[i]) return eol_t(i);
}

//******************************************************************************
void XMLConvenientWriter::setEOLSequence(const eol_t& eol)
//******************************************************************************
{
	eolsequence_ = EOLS[int(eol)];
}


//******************************************************************************
bool XMLConvenientWriter::write(XMLFormatTarget* target) const
//******************************************************************************
{
	if (node_)
	{
		ErrorReporter* reporter = new ErrorReporter();
		DOMWriter* writer = createWriter();
		
		writer->setErrorHandler(reporter);
		writer->writeNode(target, *node_);

		if (reporter->getNErrors()>0)
		{
			writer->release();
			throw XMLSyntaxError(reporter);
		}
		else if (reporter->getNWarnings()>0)
		{
			writer->release();
			throw XMLSyntaxWarning(reporter);
		}
		else
		{
			delete reporter;
			writer->release();
			return true;
		}
	}
	else
		return false;
}

//******************************************************************************
DOMWriter* XMLConvenientWriter::createWriter(const std::string& implname) const
//******************************************************************************
{
  DOMWriter* writer = DOMImplementationRegistry::
		getDOMImplementation(XMLStr("LS"))->createDOMWriter();
  
  if (!encoding_.empty()) writer->setEncoding(encoding_);
  if (!eolsequence_.empty()) writer->setNewLine(eolsequence_);

	if (writer->canSetFeature(XMLUni::fgDOMWRTFormatPrettyPrint, true))
		writer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint, true);

// 	if (writer->canSetFeature(XMLUni::fgDOMWRTWhitespaceInElementContent, true))
// 		writer->setFeature(XMLUni::fgDOMWRTWhitespaceInElementContent, true);

  return writer;
}
