/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

/**
 * \file xmlwriter.hh
 *
 * Author : Christian Hofbauer (chofbaue@qspr03.tuwien.ac.at)\n
 * started at 16 Jun 2004, last CVS commit 2004/06/16 16:05:05
 */

#ifndef XMLWRITER_HH
#define XMLWRITER_HH

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef HAVE_STD
#  include <iostream>
#else
#  include <iostream.h>
#endif

#include <string>
#include "xercesns.hh"
#include "conversion.hh"


XERCES_FORWARD(DOMNode);
XERCES_FORWARD(DOMWriter);
XERCES_FORWARD(XMLFormatTarget);

/**
 * \brief a convenient XML file/stream writer
 *
 * This file/stream writer takes care of all the necessary settings and
 * transformations to write simple (not heavily encoded) XML files. Its main
 * purpose is the support of machine written and readed XML documents that are
 * used as an exchange between different computer processes.
 */
class XMLConvenientWriter
{
public:

	//****************************************************************************
	// TYPE DEFINITIONS
	//****************************************************************************

	/** 
	 * \brief possible end-of-line sequences
	 */
	enum eol_t 
	{ eolLF,  ///< standard UNIX linefeed
		eolCR,  ///< a single carriage return
		eolCRLF ///< standard DOS carriage return  + linefeed
	};

	//****************************************************************************
	// CONSTRUCTOR
	//****************************************************************************

  /** 
   * \brief default constructor
   */
  XMLConvenientWriter() 
      : encoding_(), eolsequence_(), node_(0) { initialize(); }
  /** 
   * \brief initializing constructor
   * 
   * \param encoding the desired XML character encoding
   * \param eol the necessary end-of-line sequence
   */
  XMLConvenientWriter(const std::string& encoding, const eol_t& eol)
      :encoding_(), eolsequence_(), node_(0) { initialize(encoding, eol); }
  
  /** 
   * \brief initializes the class
   *
   * This method recognizes the new values for the character encoding and the
   * end-of-line sequence.
   * 
   * \param encoding the desired XML character encoding
   * \param eol the necessary end-of-line sequence
   */
  void initialize(const std::string& encoding = "", const eol_t& eol = eolLF);

  //****************************************************************************
  // OPERATIONS
  //****************************************************************************

	/** 
	 * \brief writes the XML document structure to cout
	 * 
	 * \return \c true if the document could be written
	 *
	 * \exception SyntaxException thrown if any syntax errors have been 
	 *                            detected
	 *
	 * \sa write(std::ostream&), writeFile(const std::string&), write(char* const)
	 */
	bool write() const;
	/** 
	 * \brief writes the XML document structure into a file
	 * 
	 * \param filename the name of the output file
	 * 
	 * \return \c true if the document could be written
	 *
	 * \exception SyntaxException thrown if any syntax errors have been 
	 *                            detected
	 *
	 * \sa write(), write(std::ostream&) write(char* const)
	 */
	bool writeFile(const std::string& filename) const;
	/** 
	 * \brief writes the XML document structure into an output stream
	 * 
	 * \param stream a standard output stream
   *
	 * \return \c the number of characters written into the stream
	 *
	 * \exception SyntaxException thrown if any syntax errors have been 
	 *                            detected
	 *
	 * \sa write(), writeFile(const std::string&), write(char* const)
	 */
	size_t write(std::ostream& stream) const;
	/** 
	 * \brief writes the XML structure into a character buffer
	 * 
	 * \param xmltext C string that receives the XML output. The string must be
	 *                empty!
   * 
	 * \return the number of characters of the new buffer
	 *
	 * \exception SyntaxException thrown if any syntax errors have been 
	 *                            detected
	 *
	 * \sa write(), writeFile(const std::string&), write(std::ostream&)
	 */
	size_t write(char* xmltext) const;

  //****************************************************************************
  // PROPERTY METHODS
  //****************************************************************************

  /** 
   * \brief sets the reference to the XML document structure
   * 
   * Sets the reference to the XML document structure that should be written.
   *
   * \param node the XML document structure
   * 
   */
  void setNode(const XERCESNS DOMNode* node) { node_ = node; }

  /** 
   * \brief retrieves the active XML encoding 
   * 
   * \return the chosen XML encoding as a std::string
   *
   * \sa setEncoding(const std::string&), encoding_
   */
  std::string getEncoding() const { return encoding_; }
  /** 
   * \brief assigns a new XML encoding
   * 
   * \param encoding the new XML encoding for the writer
   *
   * \sa getEncoding(), encoding_
   */
  void setEncoding(const std::string& encoding) { encoding_ = encoding; }

  /** 
   * \brief retrieves the active EOL  sequence
   * 
   * \return the active end-of-line character (either "CR" or "LF" or both)
   *
   * \sa setEOLSequence(const std::string&), eolsequence_
   */
  eol_t getEOLSequence() const;
  /** 
   * \brief assigns a new EOL character
   * 
   * \param eol the new end-of-line character sequence (either "CR" or "LF" or 
	 *            both)
   *
   * \sa getEncoding(), encoding_
   */
  void setEOLSequence(const eol_t& eol);

private:

  //****************************************************************************
  // PRIVATE METHODS
  //****************************************************************************

  /** 
   * \brief internal write method
   *
   * This method is used by the public write methods, which are just wrappers
   * for this function.
   *
   * \param target the XML output target
   * \return \c true if the write process was successfull, \c false otherwise.
   *
   * \sa write(), 
   *     writeFile(const std::string&), write(std::ostream&), write(char* const)
   */
  bool write(XERCESNS XMLFormatTarget* target) const;

  /** 
   * \brief creates an instance of a DOMWriter
   * 
   * \param implname the name of the DOM implementation
   * 
   * \return a valid and initialized DOMWriter object
	 *
	 * \exception SyntaxException thrown if any syntax errors have been 
	 *                            detected
   */
  XERCESNS DOMWriter* createWriter(const std::string& implname = "LS") const;

  //****************************************************************************
  // PRIVATE MEMBERS
  //****************************************************************************

  /**
   * \brief the pointer to the XML document structure
   */
  const XERCESNS DOMNode* node_;

  /**
   * \brief the active XML character encoding
   */
  XMLStr encoding_;
  /**
   * \brief the active end-of-line sequence
   */
  XMLStr eolsequence_;
};
#endif // XMLWRITER_HH
