/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "conversion_test.hh"

#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <xercesc/util/XMLUniDefs.hpp>

#include <conversion.hh>
#include <common.hh>

#ifdef HAVE_STD
#  include <iostream>
#else
#  include <iostream.h>
#endif

#include "templates.hh"

#ifdef HAVE_NAMESPACES
using namespace boost::unit_test_framework;
using namespace std;
XERCESNS_USE
#endif 

//******************************************************************************
void conversion_test::test_conversion()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " XMLCh <-> string conversion test " << "\n");

	XMLCh testxmlch[] = { 
		chLatin_Q, chLatin_u, chLatin_e, chLatin_s, chLatin_t, chLatin_i, 
		chLatin_o, chLatin_n, chDigit_4, chDigit_2, chNull };
	string teststring = "Question42";

	BOOST_MESSAGE(" ** check conversion between XMLCh* and std::string ");
	BOOST_CHECK(XMLStr::convert(testxmlch) == teststring);

	BOOST_MESSAGE(" ** check conversion between std::string and XMLCh*");
	XMLCh* test = XMLStr::convert(teststring);
	BOOST_CHECK(XMLString::compareIString(test,testxmlch)==0);
	delete[] test;

	XMLStr testxmlstr = teststring;
	string teststr = testxmlstr;
}

//******************************************************************************
void conversion_test::test_numeric()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " numeric conversion test " << "\n");

  const int    TESTINT    = 42;
  const double TESTDOUBLE = 6.4607;
  const bool   TESTBOOL   = true;

  XMLStr::FORMAT_DOUBLE = "%.4f";
  XMLStr::FORMAT_TRUE   = "on";

  XMLStr int_str;
  XMLStr double_str;
  XMLStr bool_str;

  BOOST_MESSAGE(" ** check conversion between integer and XMLStr");
  int_str.fromValue(TESTINT);
	BOOST_CHECK(int_str == string("42"));

  BOOST_MESSAGE(" ** check conversion between double and XMLStr");
  double_str.fromValue(TESTDOUBLE);
	BOOST_CHECK(double_str == string("6.4607"));

  BOOST_MESSAGE(" ** check conversion between boolean and XMLStr");
  bool_str.fromValue(TESTBOOL);
	BOOST_CHECK(bool_str == string("on"));

  BOOST_MESSAGE(" ** check conversion between XMLStr and integer");
  int test_int = int_str;
	BOOST_CHECK(test_int == TESTINT);

  BOOST_MESSAGE(" ** check conversion between XMLStr and double");
  double test_double = double_str;
	BOOST_CHECK_CLOSE(test_double, TESTDOUBLE, 0.0001);

  BOOST_MESSAGE(" ** check conversion between XMLStr and boolean");
  bool test_bool = bool_str;
	BOOST_CHECK(test_bool == true);
  test_bool = XMLStr("false");
	BOOST_CHECK(test_bool == false);

  XMLStr adams_str("ADAMS");

  try
  {
    int adams_int = adams_str;
    BOOST_ERROR(" EE conversion error not thrown!");
  }
  catch ( InvalidConversion err )
  {
    BOOST_MESSAGE(" ** conversion errror test passed for int");
  }

  try
  {
    double adams_double = adams_str;
    BOOST_ERROR(" EE conversion error not thrown!");
  }
  catch ( InvalidConversion err )
  {
    BOOST_MESSAGE(" ** conversion errror test passed for double");
  }

  try
  {
    bool adams_bool = adams_str;
    BOOST_ERROR(" EE conversion error not thrown!");
  }
  catch ( InvalidConversion err )
  {
    BOOST_MESSAGE(" ** conversion errror test passed for boolean");
  }
}

//******************************************************************************
void conversion_test::test_collapse()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " XMLStr collapse test " << "\n");
  
  string full = "   y  \n\t";
  string collapsed = "y";

  XMLStr test = full;
  test.collapse();

	BOOST_MESSAGE(" ** check if white spaces are correctly collapsed");
  BOOST_CHECK(test==collapsed);

  string wp_only = " \t\n \t\n";
  
	BOOST_MESSAGE(" ** check the wp flag of XStr");
  BOOST_CHECK(XMLStr(wp_only).wp());
}

//******************************************************************************
conversion_test_suite::conversion_test_suite()
//******************************************************************************
: test_suite("conversion_test_suite") 
{
	// add member function test cases to a test suite
	boost::shared_ptr<conversion_test> instance(new conversion_test);
			 
	test_case* conversion_test_case = 
		BOOST_CLASS_TEST_CASE(&conversion_test::test_conversion, instance); 
	test_case* numeric_test_case = 
		BOOST_CLASS_TEST_CASE(&conversion_test::test_numeric, instance); 
	test_case* collapse_test_case = 
		BOOST_CLASS_TEST_CASE(&conversion_test::test_collapse, instance); 
	
	//test case, number of tests that should fail
	add(conversion_test_case, 0);
	add(numeric_test_case, 0);
	add(collapse_test_case, 0);
}
