/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "writer_test.hh"

#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <xercesc/dom/DOM.hpp>

#include <xmlutils.hh>
#include <conversion.hh>
#include <common.hh>

#ifdef HAVE_STD
#  include <iostream>
#  include <fstream>
#else
#  include <iostream.h>
#  include <fstream.h>
#endif

#include "templates.hh"
//#include "utils.hh"

#ifdef HAVE_NAMESPACES
using namespace boost::unit_test_framework;
using namespace std;
XERCESNS_USE
#endif 

//******************************************************************************
writer_test::writer_test()
//******************************************************************************
: document(0)
{
	document = XMLUtils::createDocument("miracle");

	XMLUtils::createTextNode(document->getDocumentElement(), "question", "???");
  XMLUtils::createTextNode(document->getDocumentElement(), "answer", "42");
}

//******************************************************************************
writer_test::~writer_test()
//******************************************************************************
{
  if (document) document->release(); 
}

//******************************************************************************
XMLConvenientWriter* writer_test::prepare_writing()
//******************************************************************************
{
	XMLConvenientWriter* writer = new XMLConvenientWriter();
	writer->setNode(document);//->getDocumentElement());

	return writer;
}

//******************************************************************************
void writer_test::test_output(const string& name, 
                              const DOMDocument* const tdoc)
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " Compare "<< name << " with the given tdoc\n");
  
  XMLConvenientReader reader;
  reader.setSourceFile(name);

	try
	{
		reader.parse();

    //make some priliminary checks
    DOMDocument* qdoc  = reader.getDocument();
    BOOST_MESSAGE(" ** check if the document is really a document");
    BOOST_CHECK(XMLStr::convert(qdoc->getNodeName())=="#document");

    BOOST_CHECK(qdoc!=tdoc);
    //test_nodes(qdoc->getDocumentElement(), tdoc->getDocumentElement());

		qdoc->release();
	}
	catch(XMLException& xmlerr)
	{
		BOOST_ERROR(" EE received xml error:\n"
								<< XMLStr::convert(xmlerr.getMessage()));
	}
	catch(DOMException& domerr)
	{
		BOOST_ERROR(" EE received DOM error:\n"
								<< domerr.code);
	}
	catch(SyntaxException& synerr)
	{
		BOOST_ERROR(" EE received Syntax error:\n"
								<< synerr.what());
    cerr << XMLUtils::printSyntaxError(synerr.getReport());
	}
}

//******************************************************************************
void writer_test::test_writing_file()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " XMLConvenientWriter file test " << "\n");

	BOOST_MESSAGE(" !! create writer object");
	XMLConvenientWriter* writer = prepare_writing();
	bool ok = false;

	BOOST_MESSAGE(" ** check if universe.xml can be written");
	try
	{
		ok = writer->writeFile("universe.xml");
	}
	catch(XMLException& xmlerr)
	{
		BOOST_ERROR(" EE received xml error:\n"
								<< XMLStr::convert(xmlerr.getMessage()));
	}
	catch(DOMException& domerr)
	{
		BOOST_ERROR(" EE received DOM error:\n"
								<< domerr.code);
	}
	catch(SyntaxException& synerr)
	{
		BOOST_ERROR(" EE received Syntax error:\n"
								<< synerr.what());
    cerr << XMLUtils::printSyntaxError(synerr.getReport());
	}
	delete writer;

  test_output("universe.xml", document);
}

//******************************************************************************
void writer_test::test_writing_stream()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " XMLConvenientWriter stream test " << "\n");

	BOOST_MESSAGE(" !! create writer object");
	XMLConvenientWriter* writer = prepare_writing();
	bool ok = false;

	BOOST_MESSAGE(" ** check if a stream to universe.xml can be written");
	try
	{
    ofstream stream("universe.xml");
    if (stream)
    {
      ok = writer->write(stream);
      stream.close();
    }
	}
	catch(XMLException& xmlerr)
	{
		BOOST_ERROR(" EE received xml error:\n"
								<< XMLStr::convert(xmlerr.getMessage()));
	}
	catch(DOMException& domerr)
	{
		BOOST_ERROR(" EE received DOM error:\n"
								<< domerr.code);
	}
	catch(SyntaxException& synerr)
	{
		BOOST_ERROR(" EE received Syntax error:\n"
								<< synerr.what());
    cerr << XMLUtils::printSyntaxError(synerr.getReport());
	}
	delete writer;

  test_output("universe.xml", document);
}

//******************************************************************************
void writer_test::test_writing_char()
//******************************************************************************
{
}

//******************************************************************************
writer_test_suite::writer_test_suite()
//******************************************************************************
: test_suite("writer_test_suite") 
{
	// add member function test cases to a test suite
	boost::shared_ptr<writer_test> instance(new writer_test);
			 
	test_case* filewriting_test_case = 
		BOOST_CLASS_TEST_CASE(&writer_test::test_writing_file, instance); 
	test_case* streamwriting_test_case = 
		BOOST_CLASS_TEST_CASE(&writer_test::test_writing_stream, instance); 
	test_case* charwriting_test_case = 
		BOOST_CLASS_TEST_CASE(&writer_test::test_writing_char, instance); 
	
	//test case, number of tests that should fail
	add(filewriting_test_case, 0);
	add(streamwriting_test_case, 0);
	add(charwriting_test_case, 0);
}
