/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "xmlutils_test.hh"

#include <boost/test/unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>

#include <xercesc/dom/DOM.hpp>

#include <xmlutils.hh>
#include <conversion.hh>
#include <common.hh>

#ifdef HAVE_STD
#  include <iostream>
#  include <fstream>
#else
#  include <iostream.h>
#  include <fstream.h>
#endif

#include "templates.hh"
#include "utils.hh"

#ifdef HAVE_NAMESPACES
using namespace boost::unit_test_framework;
using namespace std;
XERCESNS_USE
#endif 

//******************************************************************************
void xmlutils_test::test_collapsenodes()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " Test XMLUtils::collapseTextNodes\n");

  BOOST_MESSAGE(" !! read wp-containing file document");
  DOMDocument* wp_doc = XMLUtils::createDocument("miracle");
  XMLUtils::createTextNode(wp_doc->getDocumentElement(),
                           "question", "live\nuniverse\teverything  and more");

  BOOST_MESSAGE(" !! read non-wp-containing file document");
  DOMDocument* nowp_doc = XMLUtils::createDocument("miracle");
  XMLUtils::createTextNode(nowp_doc->getDocumentElement(),
                           "question", "live universe everything and more");

  BOOST_MESSAGE(" ** check if both strings are not equal");
  BOOST_CHECK(!(XMLUtils::getText(nowp_doc->getDocumentElement()->
																	getFirstChild()) == 
								XMLUtils::getText(wp_doc->getDocumentElement()->
																	getFirstChild())));

  BOOST_MESSAGE(" !! collapse text nodes in wp-document");
  XMLUtils::collapseTextNodes(wp_doc->getDocumentElement());

  BOOST_MESSAGE(" ** check if both strings are equal");
  BOOST_CHECK((XMLUtils::getText(nowp_doc->getDocumentElement()->
																 getFirstChild()) == 
							 XMLUtils::getText(wp_doc->getDocumentElement()->
																 getFirstChild())));

  wp_doc->release();
  nowp_doc->release();
  
}

//******************************************************************************
void xmlutils_test::test_ignorewpnodes()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " Test XMLUtils::ignoreWPTextNodes\n");

  BOOST_MESSAGE(" !! read wp-containing file universe_wp.xml");
  XMLConvenientReader read_wp;
  read_wp.setSourceFile("universe_wp.xml");
  read_wp.parse();
  DOMDocument* wp_doc = read_wp.getDocument();

  BOOST_MESSAGE(" !! read non-wp-containing file universe_nowp.xml");
  XMLConvenientReader read_nowp;
  read_nowp.setSourceFile("universe_nowp.xml");
  read_nowp.parse();
  DOMDocument* nowp_doc = read_nowp.getDocument();

  BOOST_MESSAGE(" !! ignore whitespace nodes in wp-document");
  XMLUtils::ignoreWPTextNodes(wp_doc->getDocumentElement());

  test_nodes(wp_doc->getDocumentElement(), nowp_doc->getDocumentElement());
  
  wp_doc->release();
  nowp_doc->release();
}

//******************************************************************************
void xmlutils_test::test_findelement()
//******************************************************************************
{
	BOOST_MESSAGE("\n" << DELIMITER << "\n" 
								<< " Test XMLUtils::findElements\n");

  BOOST_MESSAGE(" !! read file universe.xml");
  XMLConvenientReader read;
  read.setSourceFile("universe.xml");
  read.parse();
  DOMDocument* doc = read.getDocument();

  BOOST_MESSAGE(" !! ignore whitespace nodes in document");
  XMLUtils::ignoreWPTextNodes(doc->getDocumentElement());

  BOOST_MESSAGE(" !! retrieve the question node");
  const DOMElement* question = XMLUtils::findElement(doc->getDocumentElement(),
																										 "question");
  BOOST_MESSAGE(" ** check if any node was found");
  BOOST_CHECK(question);
  
  BOOST_MESSAGE(" ** check if question is the right node");
  BOOST_CHECK(XMLStr(question->getTagName()) == XMLStr("question"));
  
  doc->release();
}

//******************************************************************************
xmlutils_test_suite::xmlutils_test_suite()
//******************************************************************************
: test_suite("xmlutils_test_suite") 
{
	// add member function test cases to a test suite
	boost::shared_ptr<xmlutils_test> instance(new xmlutils_test);
			 
	test_case* collapsenodes_test_case = 
		BOOST_CLASS_TEST_CASE(&xmlutils_test::test_collapsenodes, instance); 
	test_case* ignorewpnodes_test_case = 
		BOOST_CLASS_TEST_CASE(&xmlutils_test::test_ignorewpnodes, instance); 
	test_case* findelement_test_case = 
		BOOST_CLASS_TEST_CASE(&xmlutils_test::test_findelement, instance); 
	
	//test case, number of tests that should fail
	add(ignorewpnodes_test_case, 0);
	add(collapsenodes_test_case, 0);
	add(findelement_test_case, 0);
}
