/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#include "fuzzyfilter.hh"

#include <math.h>
#include <float.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#  ifdef HAVE_STD
#    include <map>
#  else
#    include <map.h>
#  endif //HAVE_STD
#else //HAVE_CONFIG_H
#  include <map>
#endif //HAVE_CONFIG_H

#include <Vector.hh>
#include <utils.hh>

typedef map<Trimesh::vertex_t, Vector_<double> > lvmap_t;
typedef lvmap_t::iterator lvmap_it;
typedef lvmap_t::const_iterator lvmap_cit;

using namespace RazorBack;

//*******************************************************************************
Vector_<double> assignClass(double val, const FuzzyFilter::classscheme_t scheme)
//*******************************************************************************
{
	Vector_<double> lv((scheme==FuzzyFilter::ctExtended) ? 5 : 3, 0);
	
	switch (scheme)
	{
		// EXTENDED CLASSIFICATION SCHEME
		case FuzzyFilter::ctExtended:
		{
			if (val < -2)
			{
				lv[0] = 1;
			}
			else if (val < -1)
			{
				lv[0] = -(val + 1);
				lv[1] = val + 2;
			}
			else if (val < 0)
			{
				lv[1] = -val;
				lv[2] = 1 + val;
			}
			else if (val < 1)
			{
				lv[2] = 1 - val;
				lv[3] = val;
			}
			else if (val < 2)
			{
				lv[3] = 2 - val;
				lv[4] = val - 1;
			}
			else 
				lv[4] = 1;

			break;
		}

		// EXTENDED CLASSIFICATION SCHEME
		case FuzzyFilter::ctCompact:
		{
			if (val<-1)
				lv[0] = 1;
			else if (val<0)
			{
				lv[0] = -val;
				lv[1] = 1 + val;
			}
			else if (val<1)
			{
				lv[1] = 1 - val;
				lv[2] = val;
			}
			else 
				lv[2] = 1;
			
			break;		
		}
	}

	return lv;
}

//*******************************************************************************
double calcDissimilarity(const Vector_<double>& lvA, const Vector_<double>& lvB)
//*******************************************************************************
{
	double delta = 0.0;
	double sum = 0.0;
	
	if (lvA.dim()==lvB.dim())
	{
		for(int i=0;i<lvA.dim();++i)
		{
			delta += fabs(lvA[i] - lvB[i]);
			sum   += lvA[i] + lvB[i];
		}

		return delta/sum;
	}
	else
		return 1.0;
}

//*******************************************************************************
FuzzyFilter::FuzzyFilter(Trimesh** surfaces, distancemap_t** distances,
												 const Rectmat_<double>** data,
												 double threshold, classscheme_t classscheme)
//*******************************************************************************
:Filter(), fClassScheme(ctExtended), fThreshold(0)
{
	init(surfaces, distances, data, threshold, classscheme);
}

//*******************************************************************************
void FuzzyFilter::init(Trimesh** surfaces, distancemap_t** distances,
											 const Rectmat_<double>** data,
											 double threshold, classscheme_t classscheme)
//*******************************************************************************
{
	Filter::init(surfaces, distances);

	fClassScheme = classscheme;
	fThreshold = fabs(threshold);
	
	for(int i=0;i<2;++i)
		pSurfData[i] = data[i];
}	

//*******************************************************************************
size_t FuzzyFilter::operator()(assocgraph_t& assocgraph)
//*******************************************************************************
{
	/* INITIALIZATION */

	//necessary propertymaps
	assocgraph_traits::pairmap_t pairmap = get(vertex_pair, assocgraph);

	//linguistic variables for both surfaces
	lvmap_t lvmapA, lvmapB;

	int i, c=0;

	/* ALGORITHM */

	//LOOP over all columns in the data matrix
	for(int col=0;col<pSurfData[0]->cno();++col)
	{
		lvmapA.clear();
		lvmapB.clear();

		//copy the column vectors and standardize
		Vector_<double> column[2];
		for(i=0;i<2;++i)
		{
			column[i] = pSurfData[i]->col(col);
			standardize(column[i].begin(), column[i].end());
		}
		
		/* GET THE LINGUISTIC VARIABLES FOR ALL PATCHPAIRS AND CALCULATE THE SIM. */

		//LOOP over all patch pairs
		pair<assocgraph_traits::vertex_it,assocgraph_traits::vertex_it>
			vits = vertices(assocgraph);
		for(assocgraph_traits::vertex_it vit=vits.first;vit!=vits.second;)
		{
			//get the association graph vertex and the patch centers on the surfaces
			assocgraph_traits::vertex_it v=vit++;
			Trimesh::vertex_t a = pairmap[*v].first;
			Trimesh::vertex_t b = pairmap[*v].second;
			
			//LV A
			lvmap_it lvA = lvmapA.find(a);
			if (lvA==lvmapA.end())	
				lvA = lvmapA.insert(
					make_pair(a, assignClass(column[0][a], fClassScheme))).first;

			//LV B
			lvmap_it lvB = lvmapB.find(b);
			if (lvB==lvmapB.end())	
				lvB = lvmapB.insert(
					make_pair(b, assignClass(column[1][b], fClassScheme))).first;

			/* FILTER */
			double dissimilarity  = calcDissimilarity(lvA->second, lvB->second);
			
			//remove the vertex if the dissimilarity is larger than the threshold
			if (fThreshold<dissimilarity)
			{
			//delete the vertex pair
			removeVertex(*v, assocgraph);
			++c;
			}
		}
	}

	return c;
}
