/*******************************************************************************
 Copyright (C) 2001-2004 by Novartis Institutes of Biomedical Research
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of the Novartis Open Source License as published by
 Novartis AG; either version 1, or (at your option) any later version of
 the license.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 Novartis Open Source License for more details.
 
*******************************************************************************/

#ifndef FUZZYFILTER_HH
#define FUZZYFILTER_HH

#include <Rectmat.hh>

#include "filter.hh"

/**
 * \brief a fuzzy filter for the possible pairs of vertices
 *
 * The fuzzy filter calculates a dissimilarity for all patchpairs in the
 * associative graph based on a linguistic variable with 3 or 5 classes. The
 * classes denote high, medium and low property values (with a very high and
 * very low class alternatively). Only those patchpairs remain in the
 * associative graph which show a certain similarity.
 *
 * The only threshold of this filter is the dissimilarity value that must not be
 * exceeded (fDissimil). The filter takes the surface date from two matrices
 * (pSurfaceData). If more than one column is present in these matrices the
 * filter criterion is fullfilled if the dissimilarity values for all columns
 * are below the threshold.
 *
 * The filtering is invoked by the operator()(assocgraph_t&) operator and
 * performs the follwing steps:
 *
 * - Take the ith columns of the data matrices of both surfaces, and calculate
 * the mean and standard deviation for each of them.
 *
 * - For every patchpair in the association graph calculate a set of membership
 * functions for vertex A and B using the standardized value \f$x_{Si}\f$ of the
 * data column: 
 * \f[ 
 *     x_{Si} = \frac{p_{Si} - \mu_S}{\sigma_S} 
 * \f]
 * \f[ 
 *     LV_S = \left\{ MF_i(x_{Si}) \forall i \in classes \right\}
 * \f]
 *
 * - Calculate the dissimilarity for the two lingustic variables according to 
 *  \f[ 
 *      D_LV(v_A, v_B) = \frac{\sum\limits_{i}\limits^N 
 *                                 \left| MF_i(v_A) - MF_i(v_B) \right|}
 *                            {\sum\limits_{i}\limits^N 
 *                                 \left( MF_i(v_A) - MF_i(v_B) \right)}
 *   \f]
 *
 * - remove those pairs from assocgraph which are more then fDissimil different
 * from each other.
 * 
 */
class FuzzyFilter: public Filter
{
public:

	//*****************************************************************************
	// PUBLIC TYPE DEFINITIONS
	//*****************************************************************************

	/**
	 * \brief describes the classification scheme used for the fuzzy sets
	 *
	 * The linguistic variables can either consist of three or five fuzzy sets.
	 */
	enum classscheme_t {
		ctCompact,   /*!< linguistic variables are based on three sets (low, medium,
									* high) */
		 ctExtended  /*!< linguistic variables are based on five sets (very low,
									* low, medium, high, very high)*/
	};

	
	//*****************************************************************************
	// CONSTRUCTOR AND DESTRUCTOR
	//*****************************************************************************

	/**
	 * \brief default constructor
	 */
	FuzzyFilter() : Filter() {}

	/**
	 * \brief initializing constructor creating a working instance of FuzzyFilter
	 *
	 * The constructor calls the init() to setup the filter parameters.
	 *
	 * \param surfaces    a two membered field of pointers to trimesh structures
	 * \param distances   a two membered field of constant pointers to distance maps
	 *                    for surface 0 and 1 (or A and B).
	 * \param data        two membered field of Rectmat_<double> pointers pointing
	 *                    to the vertex data for both surfaces
	 * \param threshold   the dissimilarity threshold
	 * \param classscheme the type of classification (3 or 5 (default) fuzzy sets)
	 *
	 * \sa FuzzyFilter(), Filter::Filter(Trimesh**, distancemap_t*)
	 */
	FuzzyFilter(Trimesh** surfaces, distancemap_t** distances,
							const RazorBack::Rectmat_<double>** data,
							double threshold, classscheme_t classscheme = ctExtended);

	/**
	 * \brief initializing routine
	 *
	 * This routine sets all the necessary parameters to the given values
	 * \param surfaces    a two membered field of pointers to trimesh structures
	 * \param distances   a two membered field of constant pointers to distance maps
	 *                    for surface 0 and 1 (or A and B).
	 * \param data        two membered field of Rectmat_<double> pointers pointing
	 *                    to the vertex data for both surfaces
	 * \param threshold   the dissimilarity threshold
	 * \param classscheme the type of classification (3 or 5 (default) fuzzy sets)
	 *
	 * \sa Filter::init(Trimesh**, const distancemap_t*)
	 */
	void init(Trimesh** surfaces, distancemap_t** distances,
						const RazorBack::Rectmat_<double>** data,
						double threshold, classscheme_t classscheme = ctExtended);

	//*****************************************************************************
	// OPERATORS
	//*****************************************************************************

	/**
	 * \brief the filter operator
	 *
	 * For further details see Filter::operator().
	 *
	 * \param assocgraph a reference to an association graph, that should be
	 *                   processed.
	 * 
	 * \return the number of modified graph elements.
	 */
	size_t operator()(assocgraph_t& assocgraph);

private:

	//*****************************************************************************
	// PRIVATA DATA MEMBERS
	//*****************************************************************************

	/**
	 * \brief flag indicating which classification scheme is in use
	 */
	classscheme_t fClassScheme;
	
	/**
	 * \brief the threshold parameter
	 *
	 * The threshold parameter defines the allowed degree of dissimilarity between
	 * two patches of a patchpair.
	 */
	double fThreshold;

	/**
	 * \brief two pointers to the surface data matrices of this experiment
	 */
	const RazorBack::Rectmat_<double>* pSurfData[2];

};

#endif //FUZZYFILTER_HH
